package nxm.database.lib;

import java.sql.*;
import nxm.sys.lib.*;

/**
  The generic super class to describe a database in java. When all of the
  necessary fields are defined it can execute basic SQL functions. While
  not an abstract class it is not really meant to be instantiated but is
  extended by the specific database classes.
  @version $Id: NxmDB.java,v 1.2 2012/09/06 18:25:13 ntn Exp $
 */
public class NxmDB {

  protected Driver      myDriver;
  protected Connection  myConnection;
  protected String      myHost;
  protected int         myPort;
  protected String      myName;
  protected String      myUserName;
  protected String      myUserPassword;
  protected boolean     verbose;
  protected String      myDrivStr;
  protected String      myURLStr;

  /** establish a connection with the database */
  public void connect() throws DatabaseException {
    try {
      /** since this class has no knowledge of what type of db it is accessing, everything
          needs to be explicitly defined in order to connect
       */
      myDriver = (Driver)Class.forName(myDrivStr).newInstance();
      myConnection = DriverManager.getConnection(myURLStr, myUserName, myUserPassword);
      if (verbose) {
        SQLWarning warn = myConnection.getWarnings();
        while (warn != null) {
          System.out.println("SQLState = " + warn.getSQLState());
          System.out.println("Message  = " + warn.getMessage());
          System.out.println("Vendor   = " + warn.getErrorCode());
          System.out.println("");
          warn = warn.getNextWarning();
        }
      }
    } catch (Exception e) {
      throw(new DatabaseException(e.getMessage()));
    }

  }

  /** generic constructor */
  public NxmDB() {
    myPort = 0;
    myDriver = null;
    myConnection = null;
    myHost = null;
    myName = null;
    myUserName = null;
    myUserPassword = null;
    verbose = false;
  }

  /** more detailed constructor */
  public NxmDB(String host, int port, String dbName,
      String user, String passwd) {

    this();
    myHost = new String(host);
    myName = new String(dbName);
    myUserName = new String(user);
    myUserPassword = new String(passwd);
    myPort = port;

  }

  /** public interface functions */
  public void setVerbose(boolean v) { verbose = v; }
  public void setDriverString(String s) { myDrivStr = s; }
  public void setURLString(String s) { myURLStr = s; }
  public void setHost(String s) { myHost = s; }
  public void setName(String s) { myName = s; }
  public void setUserName(String s) { myUserName = s; }
  public void setUserPassword(String s) { myUserPassword = s; }
  public void setPort(int i) { myPort = i; }

  /** Getting a list of tables is database dependent, must be done in subclass */
  public ResultSet getTables() {
    if (verbose) System.out.println("getTables() has not been implemented");
    return (null);
  }

  /** return the fields and types of a table but since this is dependent on
         the driver we will let the subclass implement this...could probably
         implement it here but we will be safe... */
  public ResultSet describeTable(String t) {
    if (verbose) System.out.println("describeTable(String table) has not been implemented");
    return (null);
  }

  /** run a generic SELECT SQL statement, returns a ResultSet object */
  public ResultSet selectQuery(String q) {
    ResultSet rs;
    try {
      Statement stmt = myConnection.createStatement();
      rs = stmt.executeQuery(q);
    } catch (SQLException e) {
      String errStr = "SQL Statement '" + q + "'\nProduced Exception: " + e;
      if (verbose) { System.out.println(errStr);}
      rs = null;
    }
    return(rs);
  }

  /** run a generic UPDATE SQL statement, returns number of rows updated as int */
  public int updateQuery(String q) {
    int numUpdate;
    try {
      Statement stmt = myConnection.createStatement();
      numUpdate = stmt.executeUpdate(q);
    } catch (SQLException e) {
      String errStr = "SQL Statement '" + q + "'\nProduced Exception: " + e;
      if (verbose) { System.out.println(errStr);}
      numUpdate = 0;
    }
    return(numUpdate);
  }

  /** runs a generic SQL statement. Depending on the type of statement it will
      either return a ResultSet or an Integer
   */
  public Object executeQuery(String q) {
    boolean ret;
    ResultSet rs;
    int numUpdate;
    Statement stmt;

    try {
      stmt = myConnection.createStatement();
      ret = stmt.execute(q);
      if (ret) {
        rs = stmt.getResultSet();
        return (rs);
      } else {
        numUpdate = stmt.getUpdateCount();
        return (new Integer(numUpdate));
      }

    } catch (SQLException e) {
      String errStr = "SQL Statement '" + q + "'\nProduced Exception: " + e;
      if (verbose) { System.out.println(errStr);}
      return (null);
    }
  }

  /**  disconnect and close the database connection  */
  public void disConnect() throws SQLException {
    if (myConnection != null) myConnection.close();
  }

  public Table newTableEntry(String t) {
    Table retVal = new Table();
    ResultSet rs = describeTable(t);
    String resName;
    int resType;
    try {
      ResultSetMetaData md = rs.getMetaData();
      int cols = md.getColumnCount();
      while (rs.next()) {
        Table col = new Table();
        resName = "";
        resType = Types.NULL;
        for (int i = 1; i <= cols; i++) {
          String colName = md.getColumnName(i).toUpperCase();
          if (colName.equals("COLUMN_NAME")) { resName = rs.getString(i);}
          if (colName.equals("DATA_TYPE")) {resType =  rs.getInt(i);}
        }
        switch (resType) {
          case java.sql.Types.INTEGER:
            int ival = 0;
            col.put("VALUE", ival);
            break;
          case java.sql.Types.FLOAT:
            float fval = new Float(0.0).floatValue();
            col.put("VALUE", fval);
            break;
          case java.sql.Types.DECIMAL:
          case java.sql.Types.NUMERIC:
          case java.sql.Types.DOUBLE:
            double dval = 0.0;
            col.put("VALUE", dval);
            break;
          case java.sql.Types.CHAR:
          default:
            String sval = "";
            col.put("VALUE", sval);
        }
        col.put("NAME", resName.toUpperCase());
        col.put("COLUMN_NAME", resName);
        retVal.put(resName.toUpperCase(), col);
      }
    } catch (Exception e) {
      System.out.println("Caught exception in newTableEntry(String): " + e);
      retVal = null;
    }
    return (retVal);
  }

  public boolean isClosed() {
    boolean retVal;
    if (myConnection == null) return (true);
    try {
      retVal = myConnection.isClosed();
    } catch (Exception e) {
      System.out.println("Caught exception in isClosed " + e);
      retVal = true;
    }

    return (retVal);
  }

  public void commit() {
    try {
      myConnection.commit();
    } catch (SQLException e) {
      System.out.println("Problem with commit: " + e);
    }
  }
}
