package nxm.dsp.prim;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

/**
  Performs Frame Synchronous Video Integration functions.

  @author Jeff Schoen
  @version $Id: fsvi.java,v 1.5 2012/05/02 17:47:50 jgs Exp $
*/
public class fsvi extends Primitive {

  public static final String modeList = "Integrate,Average,Exponetial,Maximum";
  public static final int M_INT=1, M_AVG=2, M_EXP=3, M_MAX=4;

  private DataFile hi,ho;
  private Data dbi,dbo,dbf;

  private int mag;  // 
  private boolean in2d; // input is a type 2000 file
  private boolean cxm;  // input is a complex file
  private boolean out1d;// output is a type 1000 file

  private int nout;     // output size
  private int mode;     // integration mode
  private int frame;    // frame size
  private int nblk,iblk;// frame block factor
  private double pri;   // pulse repitition rate
  private double delx;  // sample time delta
  private double size;  // file size
  private double factor;// averaging factor

  public int open () {

    hi = MA.getDataFile ("IN","1000,2000","S#,C#",0);
    hi.open(hi.WARNEMPTY);

    in2d = (hi.typeClass==2);
    cxm  = (hi.spa==2);
    delx = hi.getXDelta();
    size = hi.getSize();

    nblk = MA.getL("BLOCKS");
    pri  = MA.getD("PRI");
    mode = MA.getSelectionIndex("MODE",modeList,0);
    mag  = MA.getL("/MAG",cxm?1:0);
    out1d= MA.getState("/1D",(nblk<0));

         if (pri>0) frame = (int)Math.round(pri/delx);
    else if (pri<0) frame = (int)Math.max(1.0,-pri);
    else if (in2d)  frame = hi.getSubsize();
    else M.error("Illegal PRI value = "+pri);

    if (in2d) size *= frame;
    if (nblk<0) nblk = (int)Math.floor(size/frame);
    factor = 1.0/nblk;
    int nblko = (mode==M_EXP)? 1:nblk;
    size = Math.floor(size/frame/nblko);
    
    ho = MA.getDataFile("OUT",hi,0);
    if (in2d) {
      ho.setYDelta(hi.getYDelta()*nblko);
    } else {
      ho.setType(2000);
      ho.setSubSize(frame);
      ho.setYStart(hi.getXStart());
      ho.setYDelta(delx*frame*nblko);
      ho.setYUnits(Units.TIME);
      ho.setXStart(0.0);
    }
    ho.setSize(size);
    byte dtype = FLOAT;
    ho.setFormatType(dtype);
    if (mag>0) ho.setFormatMode(SCALAR);
    ho.open(DataFile.WARNEMPTY);

    int fsi = (in2d && frame==hi.getSubsize())? 1 : frame;
    dbi = hi.getDataBuffer(fsi,dtype);
    dbo = ho.getDataBuffer(1,dtype);
    dbf = ho.getDataBuffer(1,dtype);

    iblk=1;
    return NORMAL;
  }

  public int process () {

    int n = hi.read(dbi);
    if (n<0) return(FINISH); // EOF

         if (mag==2) DataOp.mag2(dbi,dbf); 
    else if (mag==1) DataOp.mag(dbi,dbf); 
    else dbf = dbi;

    if (mode==M_EXP && iblk>1) {
      DataOp.scale(dbf,dbf,frame,factor);
      DataOp.scale(dbo,dbo,frame,1.0-factor);
    }

    if (iblk==1) DataOp.copy(dbf,dbo);
    else if (mode==M_MAX) DataOp.max(dbf,dbo,dbo);
    else DataOp.add(dbf,dbo,dbo);

    if (++iblk > nblk) {  
      if (mode==M_AVG) DataOp.scale(dbo,dbo,frame,factor);
      ho.write (dbo); 
      if (mode!=M_EXP) iblk=1; 
    }
    return NORMAL;
  }

  public int close () {
    if (out1d) ho.setFS(0);
    hi.close();
    ho.close();
    return NORMAL;
  }

  /** get the number of input frames to block average for each output frame */
  public int getBlock () { return nblk; }

  /** get the fractional overlap, zero for no overlap, negative to skip */
  public double getPRI () { return pri; }

  /** set the number of input frames to block average for each output frame */
  public void setBlock (int value) {
    nblk = value;
    factor = 1.0/nblk;
  }

  /** set the fractional overlap, zero for no overlap, negative to skip */
  public void setPRI (double value) {
    if (value==pri) return;
    MA.put("PRI",""+value);
    setState(RESTART);
  }

}
