package nxm.sys.inc;

/** Definitions for the {@link #draw(int)} method implemented by all Drawable objects. <br>
    <br>
    Summary of draw methods:
    <table width="80%" border="1" align="center">
      <tr>
        <th rowspan="2">Draw Mode</th>
        <th colspan="4">Functionality</th>
      </tr>
      <tr>
        <th>Clear Plot First</th>
        <th>Plot Data</th>
        <th>Replot Axis</th>
      </tr>
      <tr>
        <td align="center">{@link #DONE}</td>
        <td align="center">&nbsp;</td>
        <td align="center">&nbsp;</td>
        <td align="center">&nbsp;</td>
      </tr>
      <tr>
        <td align="center">{@link #FULL}</td>
        <td align="center"><b>X</b></td>
        <td align="center">All Data</td>
        <td align="center"><b>X</b></td>
      </tr>
      <tr>
        <td align="center">{@link #DATA}</td>
        <td align="center"><b>X</b></td>
        <td align="center">All Data</td>
        <td align="center">&nbsp;</td>
      </tr>
      <tr>
        <td align="center">{@link #DATAX}</td>
        <td align="center">&nbsp;</td>
        <td align="center">Additional Data Only</td>
        <td align="center">&nbsp;</td>
      </tr>
      <tr>
        <td align="center">{@link #FAST}</td>
        <td align="center">X</td>
        <td align="center">"Draft" Quality Plot</td>
        <td align="center">X</td>
      </tr>
    </table>

    @author Jeff Schoen
    @version $Id: Drawable.java,v 1.12 2012/12/17 15:25:34 ntn Exp $
*/
public interface Drawable {

  /** The drawing mode indicating that nothing needs to be drawn.
      @see Drawable
   */
  public static final int DONE=-1;

  /** The drawing mode indicating that everything should be drawn including axis.
      @see Drawable
   */
  public static final int FULL=0;

  /** The drawing mode indicating that data objects should be drawn from scratch.
      @see Drawable
   */
  public static final int DATA=1;

  /** The drawing mode indicating that additional data should be drawn. This
      implies that there is more data available to plot and that it should
      be plotted on top of any data that has already been plotted.
      @see Drawable
   */
  public static final int DATAX=2;

  /** The drawing mode indicating that a "fast drawn" should be done. This is
      provided such that complex plotting algorithms can skip data (e.g. plot
      every tenth point) to give a "draft" quality view of the data that would
      be suitable for circumstances where plot quality is less important than
      computational speed (e.g. showing a plot while dragging a window).
      @see Drawable
   */
  public static final int FAST=3;

  /** Draw the object.
      @param flag The draw mode to use. This should be one of:
                  {@link #DONE}, {@link #FULL}, {@link #DATA}, {@link #DATAX}, or {@link #FAST}.
   */
  public void draw (int flag);

  /** Refreshes the drawable object. <br>
      <br>
      Different implementations may interpret this differently. Some may
      immediately refresh while others may do a partial refresh or just
      mark them selves as needing a refresh in the future.
      @since NeXtMidas 1.9.5
   */
  public void refresh ();

  /** Resets the drawable object. This is only applicable to a {@link nxm.sys.libg.Layer} (see
      {@link nxm.sys.libg.Layer#reset()} for details) for all others this is a noop method that
      returns false.
      @return true if a reset was done, false otherwise.
   */
  public boolean reset ();

  /** Indicates if the drawable needs to be refreshed.
      @return true if the drawable needs to be refreshed, false if it does not.
      @since NeXtMidas 2.5.4
   */
  public boolean getNeedRefresh ();

  /** Called by {@link nxm.sys.libg.MPlot} to clear the refresh flag(s) without actually drawing
      the drawable. This should clear all of the flags checked in {@link #getNeedRefresh()}. This
      method is also called at the top of {@link nxm.sys.libg.Layer#draw(int)}.
      @since NeXtMidas 2.5.4
   */
  public void clearNeedRefresh ();

  /** Gets the name of this drawable object. <br>
      <br>
      Implementations that have no name should be implemented as:
      <pre>
        public String getName() {
          return toString();
        }
      </pre>
      @return The name of the object.
      @since NeXtMidas 1.9.5
   */
  public String getName ();

  /** Sets the enable flags for the layer. Outside of the Layers and {@link nxm.sys.libg.MPlot} the
      {@link #setEnable(String)} method should be used in place of this one.
      @param value The flags to enable (see {@link nxm.sys.libg.MPlot#enableList}).
   */
  public void setEnable (int value);

  /** Sets the enable flags for the layer.
      @param arg The flags to enable (see {@link nxm.sys.libg.MPlot#enableList}).
   */
  public void setEnable (String arg);

  /** Gets the enable flags for the layer. Outside of the Layers and {@link nxm.sys.libg.MPlot} the
      {@link #getEnableString()} method should be used in place of this one.
      @return The flags that are enabled.
   */
  public int getEnable ();

  /** Gets the enable flags for the layer.
      @return The flags that are enabled.
   */
  public String getEnableString ();

  /** Indicates if the drawable is enabled and/or set in backing-store --
      <b>note that this does not check to see if a flag from {@link #getEnable()}
      is set as may be implied by the name</b>. The following "truth table"
      applies to the checks against the {@link nxm.sys.libg.MPlot#EN_GLOBAL} and
      {@link nxm.sys.libg.MPlot#EN_BSTORE} flags done by this method.
      <pre>
        flag | EN_GLOBAL | EN_BSTORE | isEnabled(flag)
        -----+-----------+-----------+----------------
         +1  |   false   |   false   |    false
             |   false   |   true    |    false
             |   true    |   false   |    true
             |   true    |   true    |    false
        -----+-----------+-----------+----------------
          0  |   false   |   false   |    false
             |   false   |   true    |    false
             |   true    |   false   |    true
             |   true    |   true    |    true
        -----+-----------+-----------+----------------
         -1  |   false   |   false   |    false
             |   false   |   true    |    false
             |   true    |   false   |    false
             |   true    |   true    |    true
        -----+-----------+-----------+----------------
        Note: Although other values for flag may be passed in, their behavior
              is considered "undocumented" and may change in the future.
      </pre>
      Prior to NeXtMidas 3.3.0 this was not marked {@link InternalUseOnly} and
      was documented to imply that any flag from {@link nxm.sys.libg.MPlot#enableList}
      could be passed in with a behavior that effectively checked to see that the
      given flag and {@link nxm.sys.libg.MPlot#EN_GLOBAL} were set. Since the
      functionality never matched this, the NeXtMidas 3.3.0 release began marking
      this as {@link InternalUseOnly}. Users who implement {@link Drawable} by
      extending {@link nxm.sys.libg.Layer} or {@link nxm.sys.libg.Feature} should
      simply treat the version of {@link #isEnabled(int)} in the parent class as
      if it was declared <tt>final</tt> and not override it. <br>
      <br>
      Users looking to check to see if a flag is set within the layer should
      use the {@link #getEnable()} method.
      @param flag One of the following:
                  <pre>
                    +1 (check EN_GLOBAL &amp; !EN_BSTORE)
                     0 (check EN_GLOBAL),
                    -1 (check EN_GLOBAL &amp; EN_BSTORE)
                  </pre>
      @return true or false (as shown in the "truth table" above)
      @since NeXtMidas 2.5.4
      @see #setEnable(int)
      @see #getEnable()
   */
  @InternalUseOnly
  public boolean isEnabled (int flag);
}
