package nxm.sys.inc;

import nxm.sys.lib.Shell;
import nxm.sys.lib.StringUtil;

/** Definitions of installation specific constants.

    @author Jeff Schoen
    @version $Id: Installation.java,v 1.120 2012/12/21 17:57:11 lsn Exp $
 */
public interface Installation {
  /** Indicates if the current version is a pre-release version. <br>
      <br>
      If this is <code>true</code> then the this represents a pre-release of NeXtMidas that will be
      released as version {@link #NM_VERSION} in the near future. If this is <code>false</code> then
      this represents an official release of NeXtMidas {@link #NM_VERSION}. (Note: This value makes
      no indication as whether or not this version has been modified by a customer, in general all
      customer modifications should be listed in <code>$NMROOT/nxm/sys/updates.txt</code> or
      <code>$NMROOT/updates.txt</code>.)<br>
      <br>
      This value will be set to <code>true</code> for all CVS snapshots and development
      baselines/branches in the NeXtMidas CVS.
   */
  static final boolean NM_PRE_RELEASE = true;

  /** Version number for this release of NeXtMidas. <br>
      <br>
      Value <code>{@value}</code>
   */
  static final String NM_VERSION = "3.3.2";

  /** Indicates if native methods are supported. */
  static final boolean NATIVE = true;

  /** Indicates if the X-Midas interface is included. */
  static final boolean XMIDASIF = false;

  /** Indicates if the M2k interface is included. */
  static final boolean M2KIF = false;

  /** Indicates if NeXtMidas should run in "Debug Mode". */
  static final boolean DEBUG = false;

  /** Indicates if NeXtMidas is using a debugger. */
  static final boolean DEBUGGER = false;

  /** Indicates if Java's Swing packages should be used for graphics. */
  static final boolean SWING = true;

  /** Enables or disables special bug fixes. All of these constants should be assumed to be
      deprecated since we will remove them as soon as the special work-around is no longer required.
      <br>
      NOTE: This has not really been used by users, but provides a mechanism for baseline developers to:<br>
           1. Provide a way to workaround some COTS or internal Bug, until it is no longer needed or properly fixed.<br>
           2. Provide a way to back-out/revert certain NeXtMidas DR fixes that changes behavior for a stable series or two.
      @since NeXtMidas 2.0.1
   */
  @InternalUseOnly
  public static final class ApplyBugFix {
    /** Enable the fix to GCONTROL to have /WTHEME= override /THEME= switch.
        To revert to previous behavior, set the BUGFIX.GCONTROL_WTHEME_OVERRIDES_THEME
        system property to false.
        <pre>
          NeXtMidas Bug ID: 2543 GCONTROL: Ignores /WTHEME= if /THEME= is set
        </pre>
        @since NeXtMidas 3.1.1
     */
    @InternalUseOnly
    public static final boolean GCONTROL_WTHEME_OVERRIDES_THEME
                   = getStatus("GCONTROL_WTHEME_OVERRIDES_THEME", true);

    /** Disable the fix to KEYWORD to have the FSET function overwrite duplicate
        keys in the MAIN header (for a single equal "=") as well as adding to the EXTENDED header
        (for double equals "=="). <br>
        To revert to previous behavior (appending key/value pairs only into the extended header),
        set the BUGFIX.KEYWORD_FSET_INTO_EXTENDED_ONLY system property to true.
        <pre>
          NeXtMidas Bug ID: 2599 KEYWORD: incorrectly puts keywords in extended header with FSET
        </pre>
        @since NeXtMidas 3.1.2
     */
    @InternalUseOnly
    public static final boolean KEYWORD_FSET_INTO_EXTENDED_ONLY
                   = getStatus("KEYWORD_FSET_INTO_EXTENDED_ONLY", false);

    /**  Here is am example special JVM bug workaround. Work-around fix is enabled for Java
         versions prior to 6.0.
         <pre>
           Broken in:        Java 5.0
           Fixed in:         Java 6.0
           Java Bug ID:      ?
           NeXtMidas Bug ID: ?
           Affected Classes: ?
         </pre>
     */
    @InternalUseOnly
    /* public */ static final boolean EXAMPLE_BUG_WORKAROUND_BEFORE_JAVA6
             = getStatus("EXAMPLE_BUG_WORKAROUND_BEFORE_JAVA6", !javaVerIs("1.6.0"));

    /** Indicates if the current Java version is greater than or equal to the given version.
        @param ver The given version number.
        @return <code>false</code> if the Java VM version is earlier than the given version
                number, <code>true</code> if the Java VM version is later than or equal to
                the given version number.
     */
    private static final boolean javaVerIs (String ver) {
      return StringUtil.compareVersions(ver, Shell.getJavaVersion()) <= 0;
    }

    /** Gets the status of the bug fix.
        @param name The name of the fix.
        @param def  The default value for the fix.
        @return Indicates if the bug fix should be used.
     */
    private static final boolean getStatus (String name, boolean def) {
      String force = Shell.getProperty("BUGFIX." + name);

      return ((force==null)||force.equalsIgnoreCase("AUTO"))? def: StringUtil.isTrue(force);
    }
  }
}
