package nxm.sys.inc;

import nxm.sys.libg.Layer;

/** A class to define all file methods needed by Plot.

    @author  Jeff Schoen
    @version $Id: PlotFile.java,v 1.14 2008/02/16 00:23:38 ntn Exp $
 */
public interface PlotFile extends FrameFile {
  /** Initialize the data header fields.
      @param ref      The reference to use (usually an instance of {@link nxm.sys.lib.Midas} or
                      {@link nxm.sys.lib.Primitive}).
      @param filename The name of the file to use.
   */
  public void init (Object ref, Object filename); // remove?

  /** Get the value of a specific qualifiers from the file's file name.  Returns the qualifier
      value associated with the key.
      @param key The qualifier's key.
      @return The value of the qualifier or null if not found.
   */
  public Object getQualifier (String key); // remove?

  /** Get the time in seconds since 1 January 1950 at a given offset.
      @param offset The offset at which to get the time. <i>For piped data, this can be negative, -N,
                    in order to get the time N frames prior to the current pipe position (e.g. if
                    the most recent element was 1000, <tt>getTimeAt(-10)</tt> is the same as
                    <tt>getTimeAt(990)</tt>). The use of -N is an advanced feature used in
                    conjunction with NeXtMidas pipes since it allows traversing pipe restarts
                    (this is handled in {@link nxm.sys.lib.DataFile#getTimeAt(double)}).
      @return The time in seconds since 1 Jan 1950.
      @see nxm.sys.lib.Time
   */
  public double getTimeAt (double offset);

  /** Test for the existence of the file for a given read/write direction.
      @param dir The {@link nxm.sys.lib.IOResource} direction value (see comments at top of
                 {@link nxm.sys.lib.IOResource} for details.
      @return <code>true</code> if the file exists, <code>false</code> if not.
   */
  public boolean find (int dir);

  /** Modify the connection (attach/detach from pipe). In cases where a pipe is not involved,
      this is simply an empty method (i.e. only {@link nxm.sys.lib.DataFile} implements this method).
      @param mode See {@link nxm.sys.lib.PipeResource#connect(int)} for details.
   */
  public void connect (int mode);

  /** Gets the file type for a BLUE file.
      @return The file type (e.g. 1000, 2000, etc.)
   */
  public int getType ();

  /** Get scalars per atom (also called SPA or mode) for the file.
      @return The number of scalars per atom for the file.
      @see nxm.sys.lib.Data#getSPA(byte)
   */
  public int getMode ();

  /** Get the file tag to used.
      @return The "TAG" for the file (this is usually the root name of the file or specified using
              the <tt>{TAG=...}</tt> qualifier).
   */
  public String getTag ();

  /** Get the default {@link Layer} on which this file is plotted.
      @return The default layer to use, or <tt>null</tt> to let <tt>PLOT</tt> decide.
   */
  public Layer getLayer ();

  /** Tells if the I/O resource being used is a stream. Use {@link #isStreaming()} to determine if
      a stream is flowing.
      @return true if a stream is being used, false otherwise.
   */
  public boolean isStream ();

  /** Tells if the I/O resource is streaming data.
      @return true if the resource is streaming, false otherwise. (This will always be false if
              {@link #isStream()} is false.)
   */
  public boolean isStreaming ();

  /** Get the start value for the data from the file's header. Usually the same as {@link #getXStart()};
      in cases where Y is dominant (e.g. a Type 2000 BLUE File) {@link #getYStart()} is returned.
      @see nxm.sys.lib.DataFile#getStart()
   */
  public double getStart ();

  /** Get the delta value for the data from the file's header. Usually the same as {@link #getXDelta()};
      in cases where Y is dominant (e.g. a Type 2000 BLUE File) {@link #getYDelta()} is returned.
      @see nxm.sys.lib.DataFile#getDelta()
   */
  public double getDelta ();

  /** Get the units for the data from the file's header. Usually the same as {@link #getXUnits()};
      in cases where Y is dominant (e.g. a Type 2000 BLUE File) {@link #getYUnits()} is returned.
      @see nxm.sys.lib.DataFile#getUnits()
      @see Units
   */
  public int getUnits ();

  /** Get the start value for the x-axis data from the file's header. */
  public double getXStart ();

  /** Get the delta value for the x-axis data from the file's header. */
  public double getXDelta ();

  /** Get the units for the x-axis data from the file's header.
      @see Units
  */
  public int getXUnits ();

  /** Get the number of elements in the X-direction. */
  public int getXFrame ();

  /** Get the start value for the y-axis data from the file's header. */
  public double getYStart ();

  /** Get the delta value for the y-axis data from the file's header. */
  public double getYDelta ();

  /** Get the units for the y-axis data from the file's header.
      @see Units
   */
  public int getYUnits ();

  /** Get the number of elements in the Y-direction. */
  public int getYFrame ();
}
