package nxm.sys.intr;

import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.sys.lib.Table;
import nxm.sys.lib.FileUtil.FnFilter;
import nxm.sys.libg.MJGenerateJnlp;
import static nxm.sys.lib.Generate.genExplainFromSource;
import static nxm.sys.lib.Generate.genExplainFromComments;

import tools.BuildProps;

import java.io.*;
import java.io.File;
import java.util.*;
import java.util.jar.Attributes;
import java.util.jar.Manifest;

/** Generate or modifies files of various formats (text, jars, etc.) for a named
    command or library function.
    @author Jeff Schoen / Neon Ngo
    @version $Id: Generate.java,v 1.90 2012/09/14 20:52:47 ntn Exp $
*/
public class Generate extends Intrinsic implements AsciiMap {
  /** skip options enum for "JARS" function: <pre>
      JARSTANDARD - skip standard jar file creation (.class, .mm, data, etc.)
      JARNATIVE   - skip jarring up native library files for current OS (machine)
      COPYJARS    - skip copying and signing jar files from JAR area (directory) to destination
    For "BUNDLE" function: these options are available: (Since NeXtMidas 3.3.0)
      BUNSTANDARD - Skip standard OSGi bundle file creation
      BUNNATIVE   - Skip native libraries OSGi bundle file creation
      BUNSOURCE   - Skip source code OSGi bundle file creation
      </pre>
      WARNING: Rename/remove the constants in this enum with care as it will "break" previously
               valid options for the /SKIP= switch.
      @since NeXtMidas 2.9.0
   */
  private static enum SkipOptions { JARSTANDARD, JARNATIVE, COPYJARS, BUNSTANDARD, BUNNATIVE, BUNSOURCE};
  /** default skip options enum mask for "JARS" function */
  private static final MaskValue<SkipOptions> DEF_SKIP_OPTIONS = MaskValue.constant(SkipOptions.class); // default to NONE

  // Since NeXtMidas 3.3.0: these constant strings will be included in verbose info message in doJarFile(..)
  private static final String JAR_STANDARD    = "standard JAR";    // Standard JAR file
  private static final String JAR_NATIVE      = "native   JAR";    // Native JAR file (for all OS)
  private static final String BUNDLE_STANDARD = "standard bundle"; // Standard OSGi bundle JAR
  private static final String BUNDLE_NATIVE   = "native   bundle"; // Native OSGi bundle JAR
  private static final String BUNDLE_SOURCE   = "source   bundle"; // Source OSGi bundle JAR

  private boolean doWarn       = true;  // Issue warnings if true
  private boolean debug        = false; // Show debug output?
  private boolean waitingOnGui = false; // Indicates if waiting for a GUI to make it's call back.
  private File    tmpdir       = null;  // Temp dir for files going into the jar file.
  private Map<String,Table> optInfoCache = new HashMap<String,Table>(); // since 3.3.0

  public int open () {
    String  jarsDir = FileName.terminatePath(Shell.getNmRoot()+"jars"); // The $NMROOT/jars/ dir
    String  func  = MA.getU("FUNC");
    String  name  = MA.getU("NAME");
    String  ffilt = MA.getCS("/FFILT");
    boolean sign  = MA.getState("/SIGN");
    boolean keep  = MA.getState("/KEEP");
    doWarn = MA.getState("/WARN", doWarn);
    debug  = MA.getState("/DEBUG",debug);

    // These features have been deprecated since NeXtMidas 2.5.0 ...
    boolean gmiller = false; // Use gordon code?
    if (func.equals("JAR") || func.equals("NJAR")) {
      M.deprecate("Since NeXtMidas 2.5.0: GENERATE "+func+" has been deprecated, use JARS.");
      gmiller = true;
    }
    else if (func.equals("SIGN") && MA.isPresent("P3")) {
      M.deprecate("Since NeXtMidas 2.5.0: GENERATE SIGN <name> <dest> has been deprecated, "
                + "see explain file for details.");
      gmiller = true;
    }

    if (func.startsWith("JNLP")) {
      name = MA.getCS("NAME");
      if (func.equals("JNLP/GUI") || func.equals("JNLP/TABLE")) {
        nxm.sys.lib.Table table = MA.getTable("P3");
        if (func.indexOf("/GUI") >= 0) {
          MJGenerateJnlp.createJnlpFile(this, name, null, table);
          waitingOnGui = true;
          pollTime = args.getD("/POLL", 0.25); // default to a quarter second poll time, since nothing is done in process()
        } else {
          writeJnlpFile(name, null, table);
        }
      } else {
        String opts = MA.getCS("P3","SYS");
        String dest = MA.getCS("P4", jarsDir);
        String codebase = MA.getCS("P5", "http://localhost/nmroot/jars");
        String href = MA.getCS("P6", null);
        // 2010-05-10 TODO: explain file says that <name> is case-sensitive under JNLP section
        //                  BUT it is always getting lower cased!! I think we should conform to
        //                  the explain file in the next BETA (2.9.0)
        doJnlp(name.toLowerCase(), opts, dest, codebase, href);
      }
    }
    else if (func.equals("JARS")) {
      String dest = MA.getCS("P3", "jars");
      if (dest.equals("jars")) dest = jarsDir;

      MaskValue<SkipOptions> skipMask = args.getMaskValue("/SKIP", null, DEF_SKIP_OPTIONS);

      if (!skipMask.isSet(SkipOptions.JARSTANDARD)) {
        doJarFile(name,dest,ffilt,JAR_STANDARD);
      }
      if (!skipMask.isSet(SkipOptions.JARNATIVE)) {
        doJarFile(name,dest,ffilt,JAR_NATIVE); // Since 2.9.2 (replaces JAR_ALL)
      }
      if (!skipMask.isSet(SkipOptions.COPYJARS)) {
        copyJarFiles(name,dest);
      }
    }
    else if (func.equals("JAR")) { // DEPRECATED VERSION (gmiller) since 2.5.0
      String dest = MA.getCS("P3", M.io.getAuxPath("HOME"));
      if (dest.equals("jars")) dest = jarsDir;
      String exts = MA.getCS("P4", "class|mm*|inc|cnf|tbl|props|prm|shp|shx|dbf");
      doJar(name.toLowerCase(), dest, exts, sign, keep);
    }
    else if (func.equals("NJAR")) { // DEPRECATED VERSION (gmiller) since 2.5.0
      String dest = MA.getCS("P3", M.io.getAuxPath("HOME"));
      if (dest.equals("jars")) dest = jarsDir;
      doNativeJar( name.toLowerCase(), dest, sign, keep );
    }
    else if (func.equals("SIGN") && gmiller) { // DEPRECATED VERSION (gmiller) since 2.5.0
      String dest = MA.getCS("P3", M.io.getAuxPath("HOME"));
      if (dest.equals("jars")) dest = jarsDir;
      String jarPath = getJarFileName(dest, name.toLowerCase());
      signJar (jarPath, dest);
    }
    else if (func.equals("SIGN")) {
      JarFile.signJarFile(M, MA.getFileName("NAME"));
    }
    else if (func.equals("OPER")) {
      String path = M.io.name("libm","","","SYS");
      nxm.sys.libm.Generate.doOpers (path,name,true);
    }
    else if (func.startsWith("EXP")) {
      Args    targs   = Shell.parseCommand(M,name);
      boolean doStub  = (func.indexOf("/S") > 0);
      String  expName = MA.getCS("P3", null);
      boolean edit    = MA.getState("/EDIT");

      doExplain(targs.name.toLowerCase(), expName, targs, doStub, edit);
    }
    else if (func.startsWith("BUN")) { // BUNdle - since NeXtMidas 3.3.0 for Bug 2484
      String dest = MA.getCS("P3", "bundle");
      if (dest.equals("bundle")) {
        dest = FileName.terminatePath(Shell.getNmRoot()+"bundle");
      }
      else if (dest.equals("jars")) {
        dest = jarsDir;
      }
      File file = new File(dest);
      if (!file.exists()) { // if directory does not exist
        file.mkdirs();      // create directory and any necessary parent dirs
      }

      MaskValue<SkipOptions> skipMask = args.getMaskValue("/SKIP", null, DEF_SKIP_OPTIONS);
      if (!skipMask.isSet(SkipOptions.BUNSTANDARD)) {
        doJarFile(name, dest, ffilt, BUNDLE_STANDARD);
      }
      if (!skipMask.isSet(SkipOptions.BUNNATIVE)) {
        doJarFile(name, dest, ffilt, BUNDLE_NATIVE);
      }
      if (!skipMask.isSet(SkipOptions.BUNSOURCE)) {
        doJarFile(name, dest, ffilt, BUNDLE_SOURCE);
      }

    } else {
      throw new MidasException("GENERATE: Unknown function FUNC='"+func+"'.");
    }
    return NORMAL;
  }

  public int process () {
    return (waitingOnGui)? NOOP: FINISH; // waitingOnGui is used by JNLP/GUI
  }

  /** Generates an explain file. */
  private void doExplain (String name, String expName, Args targs, boolean doStub, boolean edit) {
    String option = targs.option.toLowerCase();
    String editor = MR.getString("ENV.EDITOR");
    String source = (new FileName(targs.getDefaultSourceName())).toString();

    boolean overWrite = MA.getState("/NEW");
    // Generate an explain STUB or Generate the explain file from comments
    // in the source code
    if (expName == null) expName = M.io.name("exp",name.toLowerCase(),"exp",option);
    int status = -1;

    if (doStub) { status = genExplainFromSource (source, name, expName, overWrite, doWarn); }
    else        { status = genExplainFromComments (source, expName, overWrite); }
    if (status < 0) {
      if (status == -2) {
        if (doWarn) M.warning ("Explain file for "+name+" exists. Use /NEW switch to overwrite");
      } else {
        if (doWarn) M.warning ("Unable to create explain file for "+name+" ,status = "+status);
      }
    }

    if (edit) Shell.runCommand (M,editor+" "+source);
  }

  /** Generates a JNLP file based on a table of properties. <br>
      <br>
      The table should specify the following properties.
      <pre>
        JNLPHREF   - The "href" for the JNLP file.          [Required]
        CODEBASE   - The "codebase" for the JNLP jars.      [Required]
        COMMAND    - The name of the command to run.        [Required]
        USENATIVE  - Enables native libraries (TRUE/FALSE). [Default=FALSE]
        OPTS       - The option trees required.             [Default=SYS]
        TITLE      - Title of the application.              [Default="MyApp"]
        DESCR      - Description of the application.        [Default=title]
        SHORTDESCR - Short description of the application.  [Default=descr]
        VENDOR     - The ID of the vendor.                  [Default="ICE"]
        HOMEPAGE   - The homepage of the application.       [Default="nextmidas.html"]
        JAVAVER    - The Java versions that can be used.    [Default="1.5+"]
      </pre>

      @param fileName     The file name of the JNLP file (can not be null).
      @param templateName The name of the template file to save (usually null).
      @param table        The table of properties.
  */
  public void writeJnlpFile (String fileName, String templateName, nxm.sys.lib.Table table) {
    if ((fileName == null) && (templateName == null)) {
      M.info("No file written.");
    }

    if (fileName != null) {
      String  codebase   = table.getS("CODEBASE");
      String  href       = table.getS("JNLPHREF");
      String  command    = table.getS("COMMAND");
      boolean useNative  = table.getState("USENATIVE", false);
      String  opts       = table.getS("OPTS",       "SYS");
      String  title      = table.getS("TITLE",      "MyApp");
      String  descr      = table.getS("DESCR",      title);
      String  shortDescr = table.getS("SHORTDESCR", descr);
      String  vendor     = table.getS("VENDOR",     "ICE");
      String  homepage   = table.getS("HOMEPAGE",   "nextmidas.html");
      String  javaVer    = table.getS("JAVAVER",    "1.5+");

      try {
        ////////////////////////////// List the Jars
        String[] optionTrees = opts.split("[|]"); // Split on '|'
        ArrayList<File> jars = new ArrayList<File>(); // List of the required jars.
        File jarDir = new File(Shell.getNmRoot(), "jars");

        // Always do SYS first
        jars.add(new File(jarDir, "nxm-sys.jar"));
        if (useNative) { jars.add(new File(jarDir, "nxm-sys-all.jar")); } // TODO: fix this
        jars.addAll(Arrays.asList(OptionTree.getJarFiles("SYS")));

        for(int i = 0; i < optionTrees.length; i++) {
          String opt = optionTrees[i].toLowerCase().trim();

          if (!opt.equals("sys") && (opt.length() > 0)) {
            jars.add(new File(jarDir, "nxm-"+opt+".jar"));
            if (useNative) { jars.add(new File(jarDir, "nxm-"+opt+"-all.jar")); } // TODO: fix this

            jars.addAll(Arrays.asList(OptionTree.getJarFiles(opt)));
          }
        }

        ////////////////////////////// Write the file
        PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(fileName)));

        out.println("<?xml version=\"1.0\" encoding=\"utf-8\"?>");
        out.println ("<!-- This file is generated using NeXtMidas "+Installation.NM_VERSION+" -->");
        out.println("<jnlp spec=\"1.0+\" href=\""+href+"\"");
        out.println("      codebase=\""+codebase+"\">");
        out.println();
        out.println("  <information>");
        out.println("    <title>"+title+"</title>");
        out.println("    <vendor>"+vendor+"</vendor>");
        out.println("    <homepage href=\""+homepage+"\"/>");
        out.println("    <description>"+descr+"</description>");
        out.println("    <description kind=\"short\">"+shortDescr+"</description>");
        out.println("    <offline-allowed/>");
        out.println("  </information>");
        out.println();
        out.println("  <!-- Set the permissions that we'll request from the host. -->");
        out.println("  <security>");
        out.println("    <all-permissions/>");
        out.println("  </security>");
        out.println();
        out.println("  <resources>");
        out.println("    <j2se version=\""+javaVer+"\"/>");
        for (File jarfile : jars) {
          if ("nxm-sys.jar".equals(jarfile.getName())) {
            out.println("    <jar href=\""+jarfile+"\" main=\"true\"/>");
          } else {
            out.println("    <jar href=\""+jarfile+"\"/>");
          }
        }
        out.println("    <property name=\"OPTS\" value=\""+opts+"\"/>");
        out.println("    <property name=\"NONATIVE\" value=\""+useNative+"\"/>");
        out.println("  </resources>");
        out.println();
        out.println("  <application-desc main-class=\"nxm.sys.lib.Webstart\">");
        out.println("    <argument>"+command+"</argument>");
        out.println("  </application-desc>");
        out.println("</jnlp>");

        out.close();

        M.info("JNLP file written to " + fileName);
      } catch(Exception e) {
        throw new MidasException("Error writing file "+fileName, e);
      }
    }

    if (templateName != null) {
//TODO write table to file
      M.warning("WRITING OF TEMPLATE TO FILE NOT IMPLEMENTED");
    }

    waitingOnGui = false;
  }

  /** Write A JNLP file for the given command.
      @param cmdname  The NeXtMidas macro that will be run as a Java Web Start (JNLP) application.
      @param optStr   A String specifying the names of the option trees (other than SYS) that are
                      to be included in the application. Multiple option trees are to be separated
                      by the pipe character (<code>'|'</code>).
      @param dest     The directory where the generated files are to be stored.
      @param codebase The URL for the JNLP codebase. Codebase is also used as the base URL for all
                      relative URLs in href attributes.
      @param jnlphref The URL for the JNLP file (optional).
                      If null, the "href" attribute will be the set to "<codebase>/<cmdName>.jnlp".
                      If blank, the "href" attribute will NOT be written (useful when your jnlp
                      file is created dynamically AND arguments or properties change from
                      request to request (user to user)). Note, that Java Web Start needs href to
                      list your application in the Web Start application cache viewer.
  */
  private void doJnlp (String cmdname, String optStr, String dest, String codebase, String jnlphref) {
    //jph -- TBD separate READ and WRITE aux setting
    String auxStr = "HOME HOME|RAM|DAT"; // WRITE READ aux list

    msg("Generating JNLP files for " + cmdname+" in " + dest);

    // See if the dest directory exists, if it doesn't create it.
    File destFile = new File(dest);
    if ( !destFile.exists() ) {
      destFile.mkdir();
    }

    String jnlpName = cmdname + ".jnlp";
    FileName jnlpPath = new FileName (dest, cmdname, "jnlp");

    String href = (jnlphref==null)? codebase + "/" + jnlpName : jnlphref;
    boolean hasHref = href.trim().length() > 0; // if not blank after trimming

    String[] opts = StringUtil.cleanSplit ( optStr, "|" );

    msg("Writing " + jnlpPath );

    try {
      PrintWriter jnlp = new PrintWriter(new FileWriter(jnlpPath.getFullName()));

      jnlp.println ("<?xml version=\"1.0\" encoding=\"utf-8\"?>");
      jnlp.println ("<!-- This file was autogenerated using NeXtMidas -->");
      jnlp.println ("<jnlp spec=\"1.0+\" codebase=\"" + codebase + "\"" + (!hasHref? ">": ""));
      if (hasHref) { jnlp.println ("      href=\"" + href + "\">"); }
      jnlp.println ("");
      jnlp.println ("  <information>");
      jnlp.println ("    <title>" + cmdname + "</title>");
      jnlp.println ("    <vendor>ICE</vendor>");
      jnlp.println ("    <homepage href=\"../htdocs/index.html\"/>");
      jnlp.println ("    <description>Run " + cmdname + "</description>");
      jnlp.println ("    <description kind=\"short\">"+" Run " + cmdname + " </description>");
      jnlp.println ("    <offline-allowed/>");
      jnlp.println ("  </information>");
      jnlp.println ("");
      jnlp.println ("  <security>");
      jnlp.println ("    <all-permissions/>");
      jnlp.println ("  </security>");
      jnlp.println ("");
      jnlp.println ("  <resources>");
      jnlp.println ("    <j2se version=\"1.5+\"/>");
      jnlp.println ("    <jar href=\"nxm-sys.jar\" main=\"true\"/>");
      boolean doNative = false; //jph -- tbd make this a switch     // TODO: fix this
      if (doNative) {
        jnlp.println ("    <nativelib href=\"nxm-sys-all.jar\"/>"); // TODO: fix this
      }
      for (String optname : opts) {
        if ( !"SYS".equalsIgnoreCase(optname)) {
          String jarname = "nxm-" + optname.toLowerCase() + ".jar";
          jnlp.println ("    <jar href=\"" + jarname + "\"/>");

          if ( hasNativeCode( optname ) ) {
            String njarname = "nxm-" + optname.toLowerCase() + "-all.jar"; // TODO: fix this
            if (doNative) {
              jnlp.println ("    <nativelib href=\"" + njarname + "\"/>");
            }
          }
        }
      }
      jnlp.println ("    <property name=\"NONATIVE\" value=\""+(!doNative)+"\"/>");
      jnlp.println ("    <property name=\"OPTS\" value=\"" + optStr + "\"/>" );
      jnlp.println ("    <property name=\"AUX\" value=\"" + auxStr + "\"/>" );
      jnlp.println ("  </resources>");
      jnlp.println ("");
      jnlp.println ("  <application-desc main-class=\"nxm.sys.lib.Webstart\">");
      jnlp.println ("    <argument>" + cmdname + "</argument>");
      jnlp.println ("  </application-desc>");
      jnlp.println ("</jnlp>");

      jnlp.close();

    } catch (IOException e) {
      throw new MidasException("GENERATE: Error writing JNLP file.", e);
    }
  }

// JC4: Just something I was trying out...
//
//  /** Write A JNLP file for the given command.
//      @param cmdname The NeXtMidas macro that will be run by the JNLP app.
//      @param optStr  A String specifying the names of the option trees (other
//                     than sys) that are to be included in the app. multiple
//                     option trees are to be separated by the pipe character
//                     (<code>'|'</code>).
//      @param dest    The directory where the generated files are to be stored.
//      @param codebase    The URL for the JNLP codebase
//  */
//  private void writeJnlp (String opt, String dest, String name, String sJarName,
//                          File[] normalJars, File[] nativeJars) {
//    // See if the dest directory exists, if it doesn't create it.
//    File destFile = new File(dest + name);
//
//    try {
//      PrintWriter jnlp = new PrintWriter(new FileWriter(destFile));
//
//      jnlp.println("<?xml version=\"1.0\" encoding=\"utf-8\"?>");
//      jnlp.println("<!-- This file was autogenerated using nextmidas -->");
//      jnlp.println("<jnlp spec=\"1.0+\">");
//      jnlp.println("");
//      jnlp.println("  <information>");
//      jnlp.println("    <title>Required Resources and Native Code for " + opt + "</title>");
//      jnlp.println("    <vendor>ICE</vendor>");
////      jnlp.println("    <homepage href=\"../htdocs/nextmidas.html\"/>");
//      jnlp.println("    <offline-allowed/>");
//      jnlp.println("  </information>");
//      jnlp.println("  ");
//      jnlp.println("  <!-- RESOURCES FOR THE "+opt.toUpperCase()+" OPTION TREE -->");
//      jnlp.println("  <resources>");
//      jnlp.println("    ");
//      if (opt.equals("sys")) {
//        jnlp.println("    <jar href=\""+sJarName+"\ main=\"true\"/>");
//      } else {
//        jnlp.println("    <jar href=\""+sJarName+"\"/>");
//      }
//      for (int i = 0; i < normalJars.length; i++) {
//        String jarName = normalJars[i].getName();
//        jnlp.println("    <jar href=\""+jarName+"\"/>");
//      }
//      jnlp.println("  </resources>");
//
//      String prefix  = "nxm-"+opt+"-native-";
//      String suffix  = ".jar";
//      for (int i = 0; i < nativeJars.length; i++) {
//        String jarName = nativeJars[i].getName();
//        String params  = jarName.substring(prefix.length(), jarName.length()-suffix.length());
//        int    dash    = params.indexOf('-');
//        String os      = params.substring(0,dash).replace('_', ' ');
//        String arch    = params.substring(dash+1).replace('_', ' ');
//
//        jnlp.println("  ");
//        jnlp.println("  <resources os=\""+os+"\" arch=\""+arch+"\">");
//        jnlp.println("    <nativelib href=\""+jarName+"\"/>");
//        jnlp.println("  </resources>");
//      }
//      jnlp.println("  ");
//      jnlp.println("  <component-desc/>");
//      jnlp.println("  ");
//      jnlp.println("</jnlp>");
//      jnlp.close();
//
//    } catch (IOException e) {
//      throw new MidasException("GENERATE: Error writing JNLP file.", e);
//    }
//  }

  /** Trys to load a manifest from disk, if not found, it creates a new blank manifest. */
  private Manifest loadManifest (FileName fname) {
    Manifest manifest = null;

    if (fname != null) {
      BaseFile bf = new BaseFile(M,fname);
      if (bf.exists()) {
        try {
          if (bf.open()) { // open file
            manifest = new Manifest(bf.getResource().getInputStream());
            bf.close();    // close file
          }
        } catch (Exception e) {
          M.printStackTrace("Unable to read manifest file '"+bf.getName()+"', using default manifest.", e);
        }
      }
    }
    if (manifest == null) {
      manifest = new Manifest();
      manifest.getMainAttributes().put(Attributes.Name.MANIFEST_VERSION, "1.0"); // REQUIRED!
    }
    return manifest;
  }

  /** Lists the jar files for a given option tree. */
  private File[] listJarFiles (String opt) {
    opt = opt.toLowerCase();
    File     dir  = new File(OptionTree.getOptPath(opt,"jars"));
    FnFilter filt = FileUtil.createFnFilter(null,"*.jar",null);
    return FileUtil.findFiles(dir,filt);
  }

  /** Lists the jar files for a given option tree. */
  private File[] listNativeJarFiles (String opt, String dest) {
    opt = opt.toLowerCase();
    File     dir  = new File(dest);
    FnFilter filt = FileUtil.createFnFilter(null,"nxm-"+opt+"-native-*.jar",null);
    return FileUtil.findFiles(dir,filt);
  }

  /** Copies the required jars files from nxm/<opt>/jars to the dest directory and signs them. */
  private void copyJarFiles (String opt, String dest) {
    File[] jars = listJarFiles(opt);
    for (File file : jars) {
      String   name    = file.getName();
      FileName inName  = new FileName(file.getPath(),true);
      FileName outName = new FileName(dest, name, null, true);

      if (verbose) M.info("Copying & Signing "+opt+"'s "+name+" to "+outName);
      FileUtil.copyBaseFile(M, inName, outName);
      JarFile.signJarFile(M,outName);
    }
  }

  /** Only used by doJarFile(..) */
  private String toClasspath (File[] jars) {
    StringBuilder jarFiles = new StringBuilder();

    for (File jar : jars) {
      jarFiles.append(jar.getName()).append(' ');
    }

    return jarFiles.toString();
  }

  /** Create JAR file for specified option tree.
      Since NeXtMidas 3.3.0, this can create OSGi bundles.
      Note: Info from option tree's cfg/manifest.mf takes precedence over gleaned/parsed
            info from version.txt or generated).
      @param opt    option tree
      @param dest   destination path (SHOULD be path terminated).
      @param filter filename filter (to put in jar file)
      @param type   type of jar task (JAR_STANDARD, JAR_NATIVE, BUNDLE_STANDARD, BUNDLE_NATIVE, or BUNDLE_SOURCE)
   */
  private void doJarFile (String opt, String dest, String filter, String type) {
    opt = opt.toLowerCase();

    Manifest manifest = loadManifest(new FileName("nxm."+opt+".cfg.manifest.mf"));

    String jarFiles = null;
    String jarName = null;
    String os   = "";
    String arch = "";

    Table versionInfo = determineVersion(opt, manifest.getMainAttributes());
    String version    = versionInfo.getS("_VERSION_");

    if (type == JAR_STANDARD) {
//    String   jnlpName = "nxm-"+opt+".jnlp";
      jarName  = "nxm-"+opt+".jar";
      jarFiles = toClasspath(listJarFiles(opt));
//    writeJnlp(opt, dest, jnlpName, listJarFiles(opt), listNativeJarFiles(opt,dest));
    }
    else if (type == BUNDLE_STANDARD) { // since 3.3.0 for BUG 2484
      jarName  = "nxm."+opt+"_"+version+".jar";
    }
    else if (type == JAR_NATIVE || type == BUNDLE_NATIVE) {
      manifest = loadManifest(null);    // start with empty Manifest for native libraries jar/bundle
      filter = "*"+Shell.getSharedObjExt();
      os     = getOsName(type);
      arch   = getOsArch(type);
      if (type == BUNDLE_NATIVE) {      // since 3.3.0 for BUG 2484
        jarName = "nxm."+opt+"."+os+"."+arch+"_"+version+".jar";
      }
      else { // JAR_NATIVE
        jarName = "nxm-"+opt+"-native-"+os+"-"+arch+".jar";
      }
    }
    else if (type == BUNDLE_SOURCE) {   // since 3.3.0 for BUG 2484
      manifest = loadManifest(null);    // start with empty Manifest for source code bundle
      filter   = "*.java";
      jarName    = "nxm."+opt+".source"+"_"+version+".jar";
    }
    else {
      throw new IllegalArgumentException("Unknown JAR type '"+type+"'");
    }

    Attributes mfMainAttributes = manifest.getMainAttributes();
    if (jarFiles != null) {
      // 2010-11-02: append jar files to existing "Class-Path" entry (if exists) for BUG 2277
      String manifestClasspath = mfMainAttributes.getValue(Attributes.Name.CLASS_PATH);
      String classpath = (manifestClasspath==null)? jarFiles : manifestClasspath + " " + jarFiles;
      mfMainAttributes.put(Attributes.Name.CLASS_PATH, classpath);
    }

    // add Implementation Version metadata into manifest if not present
    putIfNotPresent(mfMainAttributes, "Implementation-Version", version);

    // since 3.3.0 for BUG 2484
    if (type == BUNDLE_STANDARD || type == BUNDLE_NATIVE || type == BUNDLE_SOURCE) {
      // add OSGi bundle specification version
      putIfNotPresent(mfMainAttributes, "Bundle-ManifestVersion", "2");
      putIfNotPresent(mfMainAttributes, "Bundle-Version", version);
      String symname = putIfNotPresent(mfMainAttributes, "Bundle-SymbolicName", "nxm."+opt); // +"; singleton:=true"?
      int iSemicolon = symname.indexOf(';');
      if (iSemicolon > 0) {
        symname = symname.substring(0, iSemicolon); // prune directives and attributes after bundle symbolic name
      }
      putIfNotPresent(mfMainAttributes, "Bundle-Description", versionInfo.getS("DESCRIPTION"));
      putIfNotPresent(mfMainAttributes, "Bundle-Name", "NeXtMidas Option Tree (" + type +") - "+opt.toUpperCase());
      putIfNotPresent(mfMainAttributes, "Bundle-ActivationPolicy", "lazy");
      if (type == BUNDLE_NATIVE || (type == BUNDLE_SOURCE)) {
        String fragmentBSN; // BSN = bundle symoblic name
        if (type == BUNDLE_NATIVE) {
          fragmentBSN = symname+"."+os+"."+arch;
          mfMainAttributes.putValue("Bundle-SymbolicName", fragmentBSN);
          // The fragment host SHOULD be the primary plugin's BSN
          mfMainAttributes.putValue("Fragment-Host", symname+"; bundle-version=\""+version+"\"");
          // Add the Eclipse Platform-Filter
          mfMainAttributes.putValue("Eclipse-PlatformFilter", "(&(osgi.os="+os+")(osgi.arch="+arch+"))");
        }
        else if (type == BUNDLE_SOURCE) {
          fragmentBSN = symname+".source";
          mfMainAttributes.putValue("Bundle-SymbolicName", fragmentBSN);
          // Eclipse-SourceBundle: <bundle-id>;version=<version>;roots:="root1, root2"
          //   where the bundle-id and version indicate the bundle that the included source code corresponds to.
          mfMainAttributes.putValue("Eclipse-SourceBundle", symname + "; version=\""+version+"\"; roots:=\".\"");
        }
      }
    }

    String     defFilter = BuildProps.getPropFor(opt, "jnlp.file_filter", Shell.getCurrent());
    String     optPath   = OptionTree.getOptPath(opt,null);
    FnFilter   filt      = FileUtil.createFnFilter(optPath,filter,defFilter);
    FileName   jarFile   = new FileName(dest, jarName, null);
    File       dir       = new File(optPath);
    File[]     files     = FileUtil.findFiles(dir, filt);
    ArrayList<String> namesInJar = new ArrayList<String>(files.length);   // name of file(s) to put in jar file (i.e. jar entry name)
    ArrayList<FileName>   fnames = new ArrayList<FileName>(files.length); // actual filename(s) on disk to put in jar file
    StringBuilder nativeCode     = new StringBuilder(); // since 3.3.0 for OSGi native bundle
    StringBuilder exportPackages = new StringBuilder(); // since 3.3.0 for OSGi standard bundle

    if (type == JAR_STANDARD || type == BUNDLE_STANDARD || type == BUNDLE_SOURCE) {
      // need directory structure in jar, so entries can be search at that level
      namesInJar.add("nxm/");
      fnames.add(null);
      namesInJar.add("nxm/"+opt+"/");
      fnames.add(null);
      exportPackages.append("nxm."+opt);
    }

    for (File file : files) { // loop through files to add to JAR

      fnames.add(new FileName(file));
      if (type == JAR_STANDARD || type == BUNDLE_STANDARD || type == BUNDLE_SOURCE) {
        String  relDir = FileUtil.getRelativeDir(dir,file.getParentFile()).replace('\\', '/');
        boolean isDir  = file.isDirectory();
        if (isDir) {
          namesInJar.add("nxm/"+opt+"/"+relDir + file.getName() + "/");
          exportPackages.append("; nxm."+opt+"."+file.getName());
        } else {
          namesInJar.add("nxm/"+opt+"/"+relDir + file.getName());
        }
      }
      else if (type == JAR_NATIVE || type == BUNDLE_NATIVE) { // since 3.3.0
        String filename = file.getName(); // get filename with extension
        if (filename.endsWith(".so") && filename.endsWith(Shell.getSharedObjExt())) {
          // Within Java Web Start and OSGi for System.loadLibrary(libname) to work the actual file
          // must be lib<libname>.so for UNIX platforms
          filename = "lib"+filename;
        } // else - don't need to prepend "lib" for Windows DLLs
        namesInJar.add(filename);
        if (type == BUNDLE_NATIVE) { // since 3.3.0: add native libraries into native bundle metadata
          nativeCode.append(filename).append("; ");
        }
      }
      else { // all other types
        namesInJar.add(file.getName());
      }
    } // end loop to add option tree files to <opt> jar file

    if ("sys".equals(opt) ) { // since 3.3.0: add $NMROOT/tools/BuildProps* to generated SYS jar file
      if (type == JAR_STANDARD || type == BUNDLE_STANDARD || type == BUNDLE_SOURCE) {
        // need directory structure in jar, so entries can be search at that level
        namesInJar.add("tools/");
        fnames.add(null);

        File   toolsDir   = new File(Shell.getNmRoot()+"tools/");
        File[] toolsFiles = FileUtil.findFiles(toolsDir, FileUtil.createFnFilter("BuildProps*.class","*.class"));
        for (File file : toolsFiles) {
          fnames.add(new FileName(file));
          namesInJar.add("tools/"+file.getName()); // two class files for BuildProps.java
        }
      }
    } // end loop to add tools/files to SYS jar file

    // since 3.3.0: more metadata for OSGi bundles (BUG 2484)
    if (type == BUNDLE_STANDARD) {
      // specify Export-Package (calculated above)
      putIfNotPresent(mfMainAttributes, "Export-Package", exportPackages.toString());
      // This Eclipse bundle buddy policy allows NXM to load classes from the option trees
      if ("sys".equals(opt)) {
        mfMainAttributes.putValue("Eclipse-BuddyPolicy", "registered");
      } else {
        mfMainAttributes.putValue("Eclipse-RegisterBuddy", "nxm.sys");
        // import main SYS packages for option trees
        putIfNotPresent(mfMainAttributes, "Import-Package",
            "nxm.sys.inc; nxm.sys.lib; nxm.sys.libg; nxm.sys.libm; nxm.sys.net; nxm.sys.test.junit; resolution:=optional");
      }
      putIfNotPresent(mfMainAttributes, "DynamicImport-Package", "*"); // cheat by dynamically importing every package

      if (MA.getState("/EMBEDJARS",true)) {
        // embed <opt>/jars/*.jar into this bundle
        File[] jarfiles = listJarFiles(opt);
        if (jarfiles.length > 0) {
          StringBuilder bundleClasspath = new StringBuilder("."); // include "." default
          for (File jarfile : jarfiles) {
            String jarfilename = jarfile.getName();
            fnames.add(new FileName(jarfile));
            namesInJar.add(jarfilename);
            bundleClasspath.append(',').append(jarfilename);
          }
          mfMainAttributes.putValue("Bundle-ClassPath", bundleClasspath.toString());
        }
      }
    }
    else if (type == BUNDLE_NATIVE) { // since 3.3.0: specify NativeCode libraries bundle metadata
      if (nativeCode.length() > 0) {  // if there are any native libraries
        nativeCode.append("osname=").append(os).append("; ");
        nativeCode.append("processor=").append(arch);
        mfMainAttributes.putValue("Bundle-NativeCode", nativeCode.toString());
      } else { // no native libraries found
        if (verbose) {
          M.info("Skipping "+type+" (no native libraries found) for "+opt.toUpperCase());
        }
        return;
      }
    }

    if (namesInJar.size() == 0) {
      // JAR must have 1 entry, so create a blank ".keepme" file for this case. This is common
      // for native JARs when an option tree has no native files. However, create the JAR anyhow
      // so that users don't need to modify their JNLP files if they add native files later on.
      fnames.add(null);
      namesInJar.add(".keepme");
    }

    if (verbose) {
      M.info("Creating "+type+" file "+jarFile);
      if (debug) {
        M.info("  file filter      : "+filt);
        M.info("  number of files  : "+namesInJar.size());
      }
    }
    String[]   nameArray  = namesInJar.toArray(StringUtil.EMPTY_STRING_ARRAY);
    FileName[] fnameArray = fnames.toArray(new FileName[0]);
    JarFile.createJarFile(M,jarFile,manifest,nameArray,fnameArray,debug,true);
  }

  /** Put into specified attribute name/value into specified Manifest Attribute object if it does
      exists presently.
      @param mfAttributes manifest attributes to check/put attribute into (cannot be null)
      @param name name of attribute to check/put into manifest attributes (cannot be null)
      @param value value to put into specified manifest attribute object
      @return value in manifest attribute (either previous value or input value if attribute
              name was not present).
      @since NeXtMidas 3.3.0
   */
  private String putIfNotPresent (Attributes mfAttributes, String name, String value) {
    String currentVal = mfAttributes.getValue(name);
    if (currentVal == null) {
      mfAttributes.putValue(name, value);
      return value;
    }
    return currentVal;
  }

  /** Determine version of option tree based. Here is the precedence order:<pre>
      1st: manifest.cf: Bundle-Version (OSGi key/value)
      2nd: manifest.cf: Implementation-Version (JAR package version info)
      3rd: version.txt: parsed from free form text file
      4th: none specified: defaults to {@link OptionTree#DEAFULT_VERSION}
      </pre>
      This will also add "Bundle-Version" (if not set) and "Implementation-Version" (if not set)
      with the determined version string in the specified mfMainAttributes object.
      @param opt option tree name
      @param mfMainAttributes main manifest's attributes object for this option tree (cannot be null) (IN/OUT)
      @return Table with version string to use in "_VERSION_" key, with other keys/values
              as documented in {@link OptionTree#getInfo(String, Midas, boolean, boolean)}.
      @see OptionTree#getInfo(String, Midas, boolean, boolean)
   */
  private synchronized Table determineVersion (String opt, Attributes mfMainAttributes) {
    String version;
    String bundleVersion = mfMainAttributes.getValue("Bundle-Version");
    String implVersion   = mfMainAttributes.getValue("Implementation-Version");

    // get option tree info (including version info)
    boolean warn = doWarn && (bundleVersion == null) && (implVersion == null);
    Table optInfo = optInfoCache.get(opt); // using cached optInfo entry prevents multiple warnings
    if (optInfo == null) {
      optInfo = OptionTree.getInfo(opt, M, verbose, warn);
      optInfoCache.put(opt, optInfo);
    }

    if (bundleVersion != null) {            // 1st precedence for version
      version = bundleVersion;
    } else if (implVersion != null) {       // 2nd precedence for version
      version = implVersion;
    } else {
      version = optInfo.getS("VERSION");    // 3rd & 4th precedence for version
    }
    if (version == null) { // this should NEVER be true (but just in case)
      version = OptionTree.DEAFULT_VERSION; // 4th precedence for version
    }

    if (implVersion == null) {   // if not specified in manifest, set it
      mfMainAttributes.putValue("Implementation-Version", version);
    }

    optInfo.put("_VERSION_", version);
    return optInfo;
  }

  /** Get current operating system (OS) name from "os.name" system property.
      This will replace spaces with underscores since spaces in file names.
      If format is {@link #BUNDLE_NATIVE}, this converts to Eclipse/OSGi standard by lower casing.
      @param format output naming convention/type
      @return "os.arch" system property normalized to specified format
      @since NeXtMidas 3.3.0
   */
  private String getOsName (String format) {
    String osname = Shell.getProperty("os.name").replace(' ', '_');
    if (format == BUNDLE_NATIVE) {
      osname = osname.toLowerCase();
    }
    return osname;
  }

  /** Get current operating system (OS) architecture from "os.arch" system property.
      Spaces (' ') are replaced with underscores ('_').<br>
      If format is {@link #BUNDLE_NATIVE}, this converts to Eclipse/OSGi standard
      naming convention: 1. lower case 2. converts "i386" to "x86" and "amd64" to "x86_64".
      @param format output naming convention/type
      @return "os.arch" system property normalized to specified format
      @since NeXtMidas 3.3.0
   */
  private String getOsArch (String format) {
    String arch = Shell.getProperty("os.arch").replace(' ', '_');
    if (format == BUNDLE_NATIVE) {
      if ("i386".equals(arch)) {
        arch = "x86";
      }
      else if ("amd64".equals(arch)) {
        arch = "x86_64";
      }
      else {
        arch = arch.toLowerCase();
      }
    }
    return arch;
  }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // "Jar Dir" Methods written by gmiller... these need to be rewritten!
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** Creates and signs jar files. Returns the name of the generated jar file or null if there is
      an error.
   */
  private String doJar (String name, String dest, String exts, boolean sign, boolean keep) {
    String jarPath = null;

    File keystore = new File( getKeystorePath() );

    try {
      msg ( "Destination is " + dest);

      boolean madeJar = false;
      boolean isFile = name.endsWith(".jar");
      jarPath = (isFile)? name : getJarFileName(dest, name);

      // If the name does not end in .jar then we take this as a command to
      // create the jar.

      if ( !isFile ) {

         msg ("creating jar for " + name);

        // Make the temporary directory that will contain the contents of the
        // jar file.
        boolean goon = makeJarDir(name, exts);
        if ( !goon ) return null;

        // Delete any old copy of the jar file.  The jar command won't delete
        // old copies, it'll simply update them.

        File file = new File ( jarPath );
        if ( file.exists() ) { msg ("Deleting " + file); file.delete(); }


        // JC4: If there is a manifest.mf file in the option tree's cfg area
        //      then use it as the manifest for the jar.
        File manifest = new File(Shell.getOptPath(name, "cfg") + File.separator + "manifest.mf");

        // Now run the jar command to create the file.
        String[] args;
        if (manifest.exists()) {
          args = new String[] {"cmf", manifest.getAbsolutePath(),
                                      jarPath, "-C", "" + tmpdir, "."};
        } else {
          args = new String[] {"cf", jarPath, "-C", "" + tmpdir, "."};
        }
        if ( !runCommand ("jar", args, 0, null) ) {
          M.warning("Error creating " + jarPath);
          return null;
        }
        madeJar = true;
        M.info ("Created " + jarPath);
      }

      if (sign) {
        signJar (jarPath, keystore.getAbsolutePath(), (madeJar ? null : dest));
      }
    }
    catch ( Exception excep ) {
      excep.printStackTrace();
    }
    finally {
      if ( !keep && tmpdir != null ) {
        msg ("deleting temporary directory " + tmpdir);
        recursiveDelete(tmpdir);
      }
    }

    return new File(jarPath).getName();
  }

  /** Invoked by the shell to create and sign the jar file containing the
      native code for an option tree. Return the name of the generated jar
      file or null if there is an error.
      NOTE: Since NeXtMidas 3.1.0 this method is now private, prior to that it was public.
            It was deprecated in NeXtMidas 2.5.0.
      @param opt  The name of the option tree to create. Unlike the {@link #doJar}
                  method, this method only operates on option trees.
      @param dest The destination directory for the created jar file. If this
                  directory does not exist the directory will be created.
      @param sign Set to true to sign the jar file using the Java&nbsp;1.2
                  jarsigner tool.
      @param keep Do not delete the temporary files after creating the jar.
      @return The full path to the created jar file.
    */
  private String doNativeJar (String opt, String dest, boolean sign, boolean keep) {
    String jarPath = null;

    msg ("creating native jar for " + opt);
    msg ( "Destination is " + dest);

    try {
      jarPath = getNativeJarFileName(dest, opt);

      if ( !makeNativeJarDir(opt) ) return null;

      // Delete any old copy of the jar file.  The jar command won't delete
      // old copies, it'll simply update them.
      File file = new File ( jarPath );
      if ( file.exists() ) { msg ("Deleting " + file); file.delete(); }

      String[] args = new String[] {"cf", jarPath, "-C", "" + tmpdir, "."};
      if ( !runCommand ("jar", args, 0, null) ) {
        M.warning("Error creating " + jarPath);
        return null;
      }
      M.info ("Created " + jarPath);

      if ( sign ) signJar (jarPath, dest);
    }
    catch ( Exception excep ) {
      excep.printStackTrace();
    }
    finally {
      if ( !keep && tmpdir != null ) {
        msg ("deleting temporary directory " + tmpdir);
        recursiveDelete(tmpdir);
      }
    }

    return new File(jarPath).getName();
  }


  /** Create and populate a temporary directory with all of the files that are
      going into this jar file.  The directory will be created in the
      temporary directory identified by the <code>java.io.tmpdir</code>.<br>
      <br>
      Example:
      <pre>
        makeJarDir ("sys", "ALL|-java|-html")
      </pre>

      This command will copy all files from the sys option tree EXCEPT java
      and html files.

      @param opt   The option tree that is going to be jarred.
      @param exts  A string specifying the extensions that are or are NOT to be
                   copied. The string must be a pipe separated list of extensions.
  */
  protected boolean makeJarDir (final String opt, final String exts) {
    String name = null;
    String stripOff = null;
    File dir = null;
    if ( opt.startsWith("./") ) {
      name = opt.substring(2);
      dir = new File(name);
      stripOff = "" + new File(dir.getAbsolutePath()).getParent();
      msg ("Directory  " + dir);
    }
    else if (OptionTree.getNxmRoot(M,opt) != null) {
      stripOff = OptionTree.getNxmRoot(M,opt);
      dir = new File(new File(stripOff, "nxm"), opt);
      name = opt;
      msg ("Option tree " + opt);
    }
    else {
      msg ("Cannot find " + opt );
      return false;
    }

    String tmp = System.getProperty("java.io.tmpdir");
    tmpdir = new File(tmp,"nxm-"+name+"-jar");

    msg ("Making temp dir       " + tmpdir);
    msg ("Copying files from    " + dir);
    msg ("Using pattern         " + exts );
    msg ("Stripping off leading " + stripOff);
    tmpdir.mkdir();

    FilenameFilter filter = new CoreIO.MatchCriteria("*",exts,true);
    recursiveCopy (stripOff,filter,dir,tmpdir,verbose,false);

    return true;
  }

  /** Create and populate a temporary directory with all of the native
      libraries that will go into a native jar file. The directory will be
      created in the temporary directory identified by the java.io.tmpdir. The
      function will copy all .so, .dll, and .exe files. After copying the
      files it will then go and rename the .so files so that they are lib*.so
      @param opt The option tree that is going to be jarred.
      @return At this time this method always returns <code>true</code>.
  */
  protected boolean makeNativeJarDir (final String opt) {
    String stripOff = OptionTree.getNxmRoot(M,opt);
    File dir = new File(new File(stripOff, "nxm"), opt);

    tmpdir = new File( Shell.getTempDir(), "nxm-"+name+"-native-jar");
    tmpdir.mkdir();

    FilenameFilter filter = new CoreIO.MatchCriteria("*_*","dll|so",true);
    recursiveCopy (stripOff,filter,dir,tmpdir,verbose,true);

    // Now go through every so file in the directory and rename it to lib*
    // so it will get picked up by the loader properly (exception for *.dll)
    File[] flist = tmpdir.listFiles();
    for (File file : flist) {
      String fname = file.getName();
      if ( !fname.startsWith("lib") && !fname.endsWith(".dll")) {
        File newFile = new File( tmpdir, "lib" + fname);
        file.renameTo ( newFile );
      }
    }

    return true;
  }

  private static boolean DEBUG = false;
  public static void setDebug (boolean state) { DEBUG = state; }
  public static boolean getDebug () { return DEBUG; }

  /** Since NeXtMidas 3.1.0 this method is now private, prior to that it was public.
      It was deprecated in NeXtMidas 2.5.0.
   */
  private String getKeystorePath () {
    return Shell.getOptPath("sys", "cfg") + "keystore";
  }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Methods deprecated since 2.5.0 will be removed after release 2.8.0
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** Since NeXtMidas 3.1.0 this method is now private, prior to that it was public.
      It was deprecated in NeXtMidas 2.5.0. Instead use {@link JarFile#signJarFile(Midas,FileName)}.
   */
  private boolean signJar11 (String jarfname, String dir, String destdir, String keystore) {

    String certFile = "nmcert.crt";

    msg ("signJar11.jarfname = " + jarfname );
    msg ("signJar11.dir      = " + dir );
    msg ("keystore           = " + keystore );
    msg ("certFile           = " + certFile );

    String[] args = new String[] { "-k", certFile,
                                   "--verbosity", "-1",
                                   "-Z", jarfname, dir };

    boolean result = runCommand ("signtool", args, 0, null);
    if (result) msg("Signed " + jarfname);
    else        M.warning("Error signing "+jarfname);

    return result;
  }

  /** Since NeXtMidas 3.1.0 this method is now private, prior to that it was public.
      It was deprecated in NeXtMidas 2.5.0. Instead use {@link JarFile#signJarFile(Midas,FileName)}.
   */
  private boolean signJar (String jarfname, String destdir) {
    File keystore = new File( getKeystorePath() );
    return signJar (jarfname, keystore.getAbsolutePath(), destdir);
  }

  /** Since NeXtMidas 3.1.0 this method is now private, prior to that it was public.
      It was deprecated in NeXtMidas 2.5.0. Instead use {@link JarFile#signJarFile(Midas,FileName)}.
   */
  private boolean signJar (String jarfname, String keystore, String destdir) {

    msg ("Signing  " + jarfname);
    msg ("Keystore " + keystore);
    msg ("Destdir  " + destdir );

    String signedJar = null;
    try {
      if ( destdir != null ) {
        File f1 = new File(jarfname);
        File f2 = new File(destdir);
        File f3 = new File(destdir, f1.getName());

        signedJar = "" + f3;
        if ( signedJar.equals (jarfname) ) {
          signedJar = null;
        }
        msg ("Destfile " + signedJar);
      }
    }
    catch ( Exception excep ) {
      M.warning (excep);
    }

    Vector<String> args = new Vector<String>();
    args.addElement ( "-keystore" );
    args.addElement ( keystore );
    args.addElement ( "-storepass" );
    args.addElement ( "nmsign" );
    if ( signedJar != null ) {
      args.addElement ( "-signedjar" );
      args.addElement ( signedJar );
    }
    args.addElement ( jarfname );
    args.addElement ( "nmsign" );

    boolean success = false;
    try {
      success = runCommand ("jarsigner", args, 0, null);
    }
    catch (Exception excep) {
      System.out.println (excep);
      excep.printStackTrace();
      success = false;
    }

    if (success) msg("Successfully signed " + jarfname);
    else         msg("Error signing " + jarfname);

    return success;
  }

  /** Since NeXtMidas 3.1.0 this method is now private, prior to that it was public.
      It was deprecated in NeXtMidas 2.5.0. Instead use {@link Foreign#runInternal(Midas,String,boolean)}.
   */
  private static boolean runCommand (final String cmd, final Vector<String> args,
                                      int successCode, String input) {
    String[] argStr = new String [ args.size() ];
    for (int i = 0; i < args.size(); i++) {
      argStr[i] = args.elementAt(i);
    }
    return runCommand(cmd,argStr,successCode, input);
  }

  /** Since NeXtMidas 3.1.0 this method is now private, prior to that it was public.
      It was deprecated in NeXtMidas 2.5.0.  Instead use {@link Foreign#runInternal(Midas,String,boolean)}.
   */
  private static boolean runCommand (String cmd, String[] args, int successCode, String input) {
    String[] transArgs = new String[ args.length ];
    for (int i = 0; i < args.length; i++) {
      transArgs[i] = args[i];
      if ( args[i].indexOf(" ") > -1 ) transArgs[i] = "\"" + args[i] + "\"";
    }

    String[] javaCmds = new String[] { "javac", "jar", "java", "jarsigner" };
    String cmdWithPath = cmd;
    for (String javacmd : javaCmds) {
      if ( cmd.equals(javacmd) ) {
        File path = new File(
          new File(System.getProperty("JAVAHOME"), "bin"),
          cmd);

        cmdWithPath = path.getAbsolutePath();
      }
    }

    String fullCmd = cmdWithPath + " " + StringUtil.join(transArgs, " ");
    if (DEBUG) System.out.println("Executing '" + fullCmd + "'");

    try {
      Process proc = Runtime.getRuntime().exec(fullCmd);
      if ( input != null ) {
        OutputStream os = proc.getOutputStream();
        os.write( input.getBytes() );
        os.flush();
      }
      int exitCode = proc.waitFor();
      return (exitCode == successCode);
    }
    catch ( Exception e ) {
      e.printStackTrace();
      return false;
    }
  }

  /** Since NeXtMidas 3.1.0 this method is now private, prior to that it was public.
      It was deprecated in NeXtMidas 2.5.0. gmiller code... do not use! */
  private void msg (String txt) { if (DEBUG || verbose) M.info (txt); }

  /** Since NeXtMidas 3.1.0 this method is now private, prior to that it was public.
      It was deprecated in NeXtMidas 2.5.0. gmiller code... do not use! */
  private String getJarFileName (final String dest, final String name) {
    String base = "nxm-" + name;
    if ( name.startsWith("./") ) base = name.substring(2);
    return "" + new File(dest, base + ".jar");
  }

  /** Since NeXtMidas 3.1.0 this method is now private, prior to that it was public.
      It was deprecated in NeXtMidas 2.5.0. gmiller code... do not use! */
  private String getNativeJarFileName( final String dest, final String opt) {
    String base = "nxm-" + opt + "-all.jar";
    return "" + new File(dest, base);
  }

  /** Since NeXtMidas 3.1.0 this method is now private, prior to that it was public.
      It was deprecated in NeXtMidas 2.5.0. gmiller code... do not use! */
  private boolean hasNativeCode (String opt) {
    class NativeCodeChecker extends FileVisitor {

      public boolean hasNative = false;

      public NativeCodeChecker() {
        super ( new CoreIO.MatchCriteria("*","dll|so",true) );
      }

      public void visit (final String name) { hasNative = true; }
    }

    NativeCodeChecker ncc = new NativeCodeChecker();
    ncc.execute(OptionTree.getNxmRoot(M,opt));
    return ncc.hasNative;
  }

  /** calls the deprecated method in CoreIO w/o compile warnings */
  private void recursiveDelete (java.io.File f) {
    try {
      Object[] params = { f };
      Util.getMethod("nxm.sys.lib.CoreIO.recursiveDelete(java.io.File)",true).invoke(null, params);
    } catch (Exception e) {
      throw new MidasException(e);
    }
  }

  /** calls the deprecated method in CoreIO w/o compile warnings */
  private void recursiveCopy (String r, FilenameFilter f, File d, File t, boolean v, boolean n) {
    try {
      Object[] params = { r, f, d, t, Boolean.valueOf(v), Boolean.valueOf(n) };
      Util.getMethod("nxm.sys.lib.CoreIO.recursiveCopy(java.lang.String,java.io.FilenameFilter,"
                            +"java.io.File,java.io.File,boolean,boolean)",true).invoke(null, params);

    } catch (Exception e) {
      throw new MidasException(e);
    }
  }
}
