package nxm.sys.intr;

import nxm.sys.inc.Installation;
import nxm.sys.lib.*;
import nxm.sys.lib.Build;        /* Needed to distinguish from using nxm.sys.intr.Build */
import nxm.sys.lib.FileUtil.FnFilter;
import nxm.sys.libg.MBox;
import nxm.sys.libg.MJBrowser;
import nxm.sys.libg.MJFrame;
import nxm.sys.libg.MJHelp;

import java.io.File;             /* Needed for File.separator */
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;         /* Needed to sort List */
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;            /* Needed to distinguish from using nxm.sys.intr.Set */
import java.util.regex.Pattern;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.StopFilter;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.demo.HTMLDocument;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.CorruptIndexException;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.queryParser.MultiFieldQueryParser;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.BooleanQuery;
import org.apache.lucene.search.Hits;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.store.AlreadyClosedException;

import tools.BuildProps;         /* Used to access build properties. */ // TODO: do NOT use BuildProps

/** Displays HELP for a given keyword or query string. HELP can create the help index via the /INDEX
    switch. HELP displays a graphical user interface to query the help index via the /GUI switch.
    Otherwise HELP is an interactive command that users can list the search results and the result
    contents in a NeXtMidas shell.

    @author unknown / J. Czechowski IV / Jay Hallam / Neon Ngo
    @version $Id: Help.java,v 1.141 2013/01/03 17:46:00 ntn Exp $
 */
public class Help extends Intrinsic {
  /** NMROOT when this class was loaded */
  private static final String NMROOT = Shell.getNmRoot();

  /** This is the default href for the style sheet used by help files.  */
  private static final String STYLE_SHEET_HREF = "../../nxm/sys/docs/style.css";

  /** The location of the javadoc directory. */
  private static final String JAVADOC_PATH = NMROOT + "htdocs" + File.separator + "api" + File.separator;

  /** The location of the help directory. */
  private static final String HELP_PATH = NMROOT + "htdocs" + File.separator + "help" + File.separator;

  private static final String mainFileName   = "mainhelp.hlp";
  private static final String configFileName = "mainhelp.cnf";

  /** properties from sys/cfg/build.props */
  private BuildProps sysBuildProps = BuildProps.getPropsFor("SYS"); // TODO: do NOT use BuildProps
  /** main help config file */
  private TextFile fcnf;
  /** main help search file */
  private TextFile fmhs;
  /** all help/commands/documents page (HTML) for a particular option tree */
  private TextFile fall;
  /** The help directory ($NMROOT/htdocs/help) */
  private String   jhdir;

  private boolean inTable = false;
  private boolean debug   = false;

  /** Help index searcher instance */
  private Searcher searcher    = null;
  /** Swing Frame containing the Help GUI */
  private MJFrame helpGUIFrame = null;
  /** reference to the Help GUI */
  private MJHelp  mjhelp       = null;
  /** default Close operation for /GUI or /LITEGUI */
  private int defaultCloseOp   = MJFrame.DISPOSE_ON_CLOSE;
  /** have lite GUI been started? */
  private boolean liteGUIRunning = false;

  /** pre-compiled pattern to test if valid command - alpha + (alphanumeric or ;) */
  private static final Pattern VALID_COMMAND_REGEX  = Pattern.compile("\\p{Alpha}[\\w;]*");
  /** pre-compiled pattern for HTML <pre> tag */
  private static final Pattern PRE_OPEN_REGEX  = Pattern.compile("(?i)<pre>");
  /** pre-compiled pattern for HTML </pre> tag */
  private static final Pattern PRE_CLOSE_REGEX  = Pattern.compile("(?i)</pre>");
  /** Option tree containing help item */
  private String opt;
  /** option trees to build help files or for interactive CLI searching */
  private String[] opts;

  /** execution mode enum */
  private enum RunMode { INDEX, BUILD, GUI, LiteGUI, CLI, LegacyCLI };
  private RunMode runMode;
  /** is Shell running under one-shot mode ie. {@link Shell#isOneShot()} */
  boolean isOneShot;

  public int open () {
    // get parameters
    opt         = MA.getU("OPT", "ALL");
    debug       = MA.getState("/DEBUG");

    // determine option tree search path
    if (opt.equals("ALL")) {
      String path = (String)M.results.get("OPT.PATH");
      opts = StringUtil.cleanSplit(path,",");
    } else {
      opts = new String[] { opt };
    }

    isOneShot = Shell.isOneShot();
    if (MA.getState("/INDEX")) {         // build help index for searching
      runMode = RunMode.INDEX;
    }
    else if (MA.getState("/BUILD")) {
      runMode = RunMode.BUILD;
    }
    else if (MA.getState("/GUI")) {      // Interactive HelpGUI
      if (MA.find("/JAVADOC")) M.deprecate("Since NeXtMidas 2.5.0 /JAVADOC is no longer needed.");
      runMode = RunMode.GUI;
      if (isOneShot) defaultCloseOp = MJFrame.EXIT_ON_CLOSE;
    }
    else if (MA.getState("/LITEGUI")) {  // Lite interactive HelpGUI
      runMode = RunMode.LiteGUI;
      if (isOneShot) defaultCloseOp = MJFrame.EXIT_ON_CLOSE;
    }
    else if (MA.getState("/LEGACY", true)) {
      runMode = RunMode.LegacyCLI;
    }
    else {
      runMode = RunMode.CLI;
    }

    // default to a quarter second poll time, since nothing is done in process() for /GUI and /LITEGUI
    if (isOneShot) pollTime = args.getD("/POLL", 0.25);

    return NORMAL;
  }

  public int process () {
    if (runMode == RunMode.INDEX) {
      // index help files using Lucene and exit
      String filter   = MA.getCS("/FILTER", null);
      String moreDirs = MA.getCS("/MOREDIRS", null);
      if (!indexFiles(filter, moreDirs)) M.error("Run of Help Indexer failed!");
      return FINISH;
    }
    else if (runMode == RunMode.GUI) {
      if (helpGUIFrame == null) { // Start Interactive HelpGUI (it will stay up even after HELP command
        helpGUI();                // FINISHes). HelpGUI exits when user closes GUI or when Shell exits.
      }
      if (isOneShot) return NOOP; // need to return NOOP under one-shot mode otherwise GUI closes immediately
      else           return FINISH;
    }
    else if (runMode == RunMode.LiteGUI) {
      if (!liteGUIRunning) {
        // launch the "lite" HelpGUI w/o the MJBrowser (viewer) and exit
        try {
          String indexPath = sysBuildProps.getPath("help.index.dir");
          MJHelp.runAsApplication("NeXtMidas Search", indexPath, 500, 400, Integer.valueOf(defaultCloseOp));
          liteGUIRunning = true;
        } catch (Exception e) {
          M.error("Aborted run of Lite HelpGUI (MJHelp). "+e);
        }
      }
      if (isOneShot) return NOOP; // need to return NOOP under one-shot mode otherwise GUI closes immediately
      else           return FINISH;
    }
    else if (runMode == RunMode.BUILD) {
      // build the help system for the option tree (or ALL)
      String jddir = null;
      String[] pathList = { Shell.getNmRoot(),"htdocs","api" };

      if (MA.find("/LOCAL")) {
        if (opt.equals("ALL")) {
          M.error("Option name must accompany /LOCAL");
        }

        pathList[0] = Shell.getOptPath(opt,"");
        pathList[1]="docs";
        FileName fn = new FileName (pathList,"","");
        jddir = fn.getFullName();
      }

      if(jddir == null) {
        M.info("Building JavaDocs for opt=" + opt + " in " + JAVADOC_PATH);
      } else {
        M.info("Building JavaDocs for opt=" + opt + " in " + jddir);
      }
      Build.setMidasContext(M);
      Build.compileDocs(opt,jddir);

      pathList[2] = "help";
      FileName fn = new FileName (pathList,"","");
      jhdir = fn.getFullName();
      M.info("Building help for opt=" + opt + " in " + jhdir);

      int status = 0;
      for(int i=0; i<opts.length; i++) { status += buildMainHelp(opts[i]); }
      for(int i=0; i<opts.length; i++) { status += buildHelp(opts[i]);     }

      status += buildMasterIndex();
      if (status<0) {
        M.warning("Encountered " + (-status) + " problems while building Help.");
      }

      return FINISH;
    }
    else if (runMode == RunMode.LegacyCLI) {
      // handle interaction with user on terminal (CLI)
      String helpItemName = MA.getU("NAME","MAINHELP");
      doLegacySearch(helpItemName, opt, opts);
    } else {
      // handle interaction with user on terminal (CLI)
      String queryStr = MA.getU("NAME","");
      // handle user's interaction with search results
      doSearchInteraction(queryStr);
    }

    return FINISH;
  }

  public int close () {
    if (helpGUIFrame != null) {
      mjhelp = null;
      helpGUIFrame = null;
    }
    return NORMAL;
  }

  /** Create and Initial HELP GUI graphic components (Frame, JSplitPane, MJHelp, MJBrowser).
      Does not set the frame to visible.
    */
  private void initHelpGUIComponents (String indexPath) {
    helpGUIFrame = new MJFrame("NeXtMidas Search", new MBox(20, 30, 800, 600));
    javax.swing.JSplitPane jSplitPane = new javax.swing.JSplitPane();
    mjhelp = new MJHelp();
    MJBrowser browser = new MJBrowser();
    mjhelp.setViewer(browser);

    jSplitPane.setDividerLocation(220);
    jSplitPane.setDividerSize(8);

    jSplitPane.setLeftComponent(mjhelp);
    jSplitPane.setRightComponent(browser);

    helpGUIFrame.getContentPane().add(jSplitPane, java.awt.BorderLayout.CENTER);
    helpGUIFrame.pack();
    helpGUIFrame.setDefaultCloseOperation(defaultCloseOp);

    mjhelp.setIndexPath(indexPath);
  }

  /** Run HELP GUI */
  private void helpGUI () {
    try {
      if (helpGUIFrame == null) initHelpGUIComponents(sysBuildProps.getPath("help.index.dir"));
      helpGUIFrame.setVisible(true);
      mjhelp.requestFocusForSearch(); // request focus on search text input field
    } catch (Throwable t) {
      throw new MidasException("Aborted run of HelpGUI.", t);
    }
  }

  /** Handle terminal search functionality prior to moving to searching Lucene Help index.
      @param name Name of help item to search for
      @param opt Option tree containing help item
      @param opts
   */
  private void doLegacySearch (String name, String opt, String[] opts) {
    ArrayList<Key> list = new ArrayList<Key>(128); // init List for search list

    // search for an explain file
    Args args = Shell.parseCommand (M,name);
    if (!args.name.equals("") && (opt.equals("ALL") || args.option.equals(opt))) {
      Key key = new Key();
      key.type = "EXPLAIN";
      key.label = args.name;
      key.opt = args.option;
      list.add(key);
    }

    // if this is MAINHELP splash screen reset default opts list
    if (name.equals("MAINHELP") && opt.equals("ALL")) opts = new String[] { "SYS" };

    // search the MainHelp files
    for (int i=0; i<opts.length; i++) searchMainHelp(opts[i],name,list);

    // output the list of matches
    M.terminal.writeln("Search for key="+name+" in opt="+opt+" yields "+list.size()+" entries:");
    if (list.size()>1) listEntries(list);

    // loop on what to do next
    for (int ilist=0; list.size()>0; ilist++) {
      String ans="N";
      if (list.size()>1)
        ans = M.terminal.getInput("Enter item number or [N]ext/Quit/List ?: ","");
      if (ans.length()==0) ans="N";
      else ans = ans.trim().toUpperCase();

      if (ans.startsWith("Q")) break;
      else if (ans.startsWith("L")) { listEntries(list); continue; }
      else if (ans.startsWith("N")) { if (ilist==list.size()) break; }
      else ilist = Convert.s2l(ans);
      ilist %= list.size();
      Key key = list.get(ilist);
      if (key.type.equals("EXPLAIN")) doExplainFileEntry (key.label,key.opt);
      else doHelpTreeEntry (key.file,key.label,key.opt);
    }
  }

  // look for a keyword match in the main help file
  private void searchMainHelp (String opt, String keyword, List<Key> list) {
    String fn = M.io.name("HLP","MAINHELP","HLP",opt);
    TextFile tf = new TextFile(fn);
    if (!tf.find(IOResource.DIR_INPUT)) return;
    if (!tf.open(TextFile.NOABORT)) return;
    String text,filename=null;
    while ((text=tf.read())!=null) {
      Key key = parseKey(text);
      if (key==null) continue;
      else if (key.type.equals("FILE")) {
        filename=key.label;
        key.file="MAINHELP"; key.opt=opt;
        if (key.label.indexOf(keyword)>=0) {
          // Added to prevent duplicate entries like ~File FLAGS in MAINHELP of ICE tree
          int iDupMain = mainHelpInList(list);
          if (iDupMain>=0) {
//System.out.println("MAINHELP ALREADY IN, NOT ADDING key->"+key);
          } else {
            list.add(key);
          }
        }
      }
      else if (key.type.equals("ENTRY")) {
        key.file=filename; key.opt=opt;
        if (key.label.indexOf(keyword)>=0) {
          // Added to prevent duplicate entries like ~File FLAGS in MAINHELP
          // of ICE tree
          if (key.file.equals("MAINHELP") && mainHelpInList(list) >=0) {
//System.out.println("MAINHELP ALREADY IN, NOT ADDING key->"+key);
          } else {
            list.add(key);
          }
        }
      }
    }
    tf.close();
  }

// START - Added to prevent duplicate entries like ~File FLAGS in MAINHELP of ICE tree
  /** Determine if MAINHELP is already in the list */
  private int mainHelpInList (List<Key> list) {
    int index = -1;
    for (int i=0; i<list.size(); i++) {
      Key key = list.get(i);
      if (key.file != null && key.file.equals("MAINHELP")) {
        index = i;
        break;
      }
    }
    return index;
  }
// END - Added to prevent duplicate entries like ~File FLAGS in MAINHELP of ICE tree

  // list numbered list of search entries to the screen
  private void listEntries (List<Key> list) {
    M.terminal.more(Terminal.INIT);
    for (int i=0; i<list.size(); i++) {
      Key key = list.get(i);
      String type = "Help Item  ";
      if (key.type.equals("FILE")) type = "Help File  ";
      if (key.type.equals("EXPLAIN")) type = "Explain On ";
      String file = (key.file==null)? "": " file="+key.file;
      String index = " "+i+") "; if (i<10) index = " "+index;
      M.terminal.more (index+type+key.label+"   in"+file+" opt="+key.opt);
    }
    M.terminal.more(Terminal.EXIT);
  }

  // display an explain file entry
  private void doExplainFileEntry (String name, String opt) {
    String text, fname=M.io.name("EXP",name,"EXP",opt);
    TextFile tf = new TextFile (fname);
    if (!tf.open(TextFile.NOABORT)) return;
    M.terminal.more(Terminal.INIT);
    while (M.terminal.more() && (text=tf.read())!=null) M.terminal.more (text);
    M.terminal.more(Terminal.EXIT);
    tf.close();
  }

  // display a help tree entry
  private int doHelpTreeEntry (String filename, String keyword, String opt) {
    String fname, text=null, infname=filename;
    ArrayList<Key> list = new ArrayList<Key>(16);

    // construct the help file name
    if (filename.equals("DOCUMENTS")) {
      fname = M.io.name("HLP",keyword,"DOC",opt);
      if (!M.io.find(fname)) fname = M.io.name("DOCS",keyword,"DOC",opt);
    }
    else if (filename.equals("INCLUDES"))
      fname = M.io.name("INC",keyword,"INC",opt);
    else
      fname = M.io.name("HLP",filename,"HLP",opt);

    // open the designated help file
    TextFile tf = new TextFile (M,fname);
    if (!tf.open(BaseFile.INPUT|BaseFile.NOABORT)) {
      M.warning("HELP: Could not find "+tf.getName());
      return 0;
    }

    // advance to section line of interest
    if (keyword.equals("")) text = tf.read();
    else if (filename.equals("INCLUDES")) text = "Include File: "+fname;
    else if (filename.equals("DOCUMENTS")) text = "Document File: "+fname;
    else {  // build list of keys while looking for the section of interest
      while ((text=tf.read())!=null) {
        Key key = parseKey(text);
        if (key==null) continue;
        else if (key.type.equals("FILE") && infname.equals("MAINHELP")) {
          filename = key.label;
          if (key.label.indexOf(keyword)>=0)  {
            text="Help Category: "+opt+"_"+key.label; break; }
        }
        else if (key.type.equals("SECTION") || key.type.equals("FUNCTION")) {
          if (key.label.indexOf(keyword)>=0)  {
            text="Help Entry: "+opt+"_"+filename+"_"+key.label; break; }
        }
      }
    }

    // section not found - malformed help file
    if (text==null) M.error("Help Format Err: Section "+keyword+" not found in "+infname);

    // print this entry to screen
    M.terminal.more(Terminal.INIT);
    for ( M.terminal.more(text); M.terminal.more(); ) {
      text=tf.read();
      if (text==null) break;
      Key key = parseKey(text);
      if (key==null);
      // build new sub-level search list from this section while printing
      else if (key.type.equals("ENTRY")) { key.file=filename; list.add(key); }
      else if (key.type.equals("INCLUDE")) { doHelpTreeEntry(key.file, "", opt);  } // Added for BUG 2147
      else break;
      M.terminal.more (text);
    }
    M.terminal.more(Terminal.EXIT);
    if (list.size()>0) {   // query for further probing ?
      text = M.terminal.getInput("Item in "+keyword+" to explore ?: ","");
      text = text.toUpperCase();
    }
    tf.close();

    if (text==null || text.length()==0) return 0;  // no interest in continuing

    // we have a list of keys, check for interest and search
    int stat=0;
    for (Key key : list) {
      if (key.label.indexOf(text)>=0) {
        String ans = M.terminal.getInput("Explore "+key.label+" in "+
           opt+"_"+key.file+" [Y]/N/Q ?: ","");
        ans = ans.toUpperCase();
        if (ans.startsWith("N")) continue;
        if (ans.startsWith("Q")) return -1;
        stat = doHelpTreeEntry(key.file,key.label,opt);
      }
    }
    return stat;
  }

  // parse a line of help text for possible special keys
  private Key parseKey (String text) {
    text = text.replaceAll("\\s+", " "); // JC4: Replace all blocks of whitespace with a single space

    int i=0, j, len=text.length();
    for (; i<len && text.charAt(i)<=' '; i++);
    if (i==len) return null;
    Key key=null;
    if (text.charAt(i)=='~') {      // tilde section header
      key = new Key();
      for (j=++i; i<len && text.charAt(i)>' '; i++);
      key.type = text.substring(j,i).toUpperCase();
      for (; i<len && text.charAt(i)==' '; i++);
      for (j=i; j<len && text.charAt(j)!=' '; j++);
      if (key.type.equals("INCLUDE")) { key.file=text.substring(i,j); }  // JWB- ADDED FOR BUG 2147
      key.label = text.substring(i,j).toUpperCase();
      for (; j<len && (text.charAt(j)==' ' || text.charAt(j)=='-'); j++);
      key.text = (j<len)? text.substring(j):"";

      if(key.type.trim().length() == 0) {
        key = null;
      }
    } else if ((i=text.indexOf('-'))>0) {    // dash separated keyword entry
      key = new Key();
      key.type = "ENTRY";
      key.label = text.substring(0,i).trim().toUpperCase();
      for (j=0; j<key.label.length(); j++) // multiple words before dash ?
        if (key.label.charAt(j) <= ' ') return null;
      key.text = text.substring(i+1).trim();
    }
    return key;
  }

  private class Key {
    String type;  // type of key (FILE,SECTION,ENTRY,EXPLAIN...)
    String file;  // filename of help file containing info
    String label;  // section header or other keyword of interest
    String text;  // additional text for label description
    String opt;    // Option tree this was found in

    public String toString() {
      return "Key: type="+type+" file="+file+" label="+label+" text="+text+" opt="+opt;
    }
  }

   /** Build the Help support files for a given option tree. The result of
      this method is the generation of a number of files into the hlp
      directory of the option tree.

      In addition, a topic directory will be created in the hlp area which
      contains the HTML files for the individual help topics.  These are
      referenced by the nxmMaster files.

      @param opt The name of a single option tree to build the Java Help files for.
  */
  private int buildMainHelp (String opt) {

    String path = M.io.name("HLP","","",opt);
    ArrayList<Key> list = new ArrayList<Key>(16);
    String text;

    fmhs = new TextFile( path+mainFileName );
    fcnf = new TextFile( path+configFileName );

    if (fcnf.find(IOResource.DIR_INPUT)) {
      M.info("File mainhelp.cnf found, generating mainhelp.hlp for "+opt);
      if (!fcnf.open(BaseFile.INPUT|BaseFile.NOABORT)) return -1;
      if (!fmhs.open(BaseFile.OUTPUT|BaseFile.NOABORT)) return -1;
      fmhs.writeln("~File MAINHELP");
      while ((text=fcnf.read())!=null) {
        fmhs.writeln(text);
        Key key = parseKey(text);
        if (key==null) continue;
        list.add(key);
      }
      for (Key key : list) {
        buildMainHelpFile(opt,key.label);
      }
    }
    else if (fmhs.find(IOResource.DIR_INPUT)) {
      M.info("File mainhelp.cnf not found, using mainhelp.hlp for "+opt);
    }
    else {
      M.info("No mainhelp.cnf or mainhelp.hlp found for "+opt);
    }
    fmhs.close();
    fcnf.close();
    return 0;
  }

  // process one hlp file for buildMainHelp
  private int buildMainHelpFile (String opt, String name) {
    String text; boolean header=true, needOneLiner=false;

    String fn = M.io.name("HLP",name,"HLP",opt);
    TextFile tf = new TextFile(fn);
    if (!tf.open(TextFile.NOABORT)) return -1;
    tf.setIncludeString("~Include");

    fmhs.writeln("~File "+name);
    fmhs.writeln("");

    while ((text=tf.read())!=null) {
      Key key = parseKey(text);
      if (key==null) {
        if (needOneLiner && text.length()>0 && text.charAt(0)!=' ') {
          int i=text.indexOf(". ");  if (i>0) text = text.substring(0,i);
          fmhs.writeln(text); needOneLiner=false;
        }
      }
      else if (key.type.equals("MAIN")) {
        header=true; text=null;
      }
      else if (key.type.equals("SECTION") || key.type.equals("FUNCTION")) {
        if (needOneLiner) fmhs.writeln("unknown");
        needOneLiner = (key.text.length()==0);
        if (needOneLiner) fmhs.write("  "+key.label+" - ");
        else fmhs.writeln("  "+key.label+" - "+key.text);
        header = false;
      }
      else header = false;
      if (header && text!=null) fmhs.writeln(text);
    }
    fmhs.writeln("");
    return 0;
  }

  /** Build all of the Help and Html pages for specified option tree. */
  private int buildHelp (String opt) {

    M.info("Building Help/Explain/Documents for "+opt);

    String path = M.io.name("HLP","","",opt);
    String optLC = opt.toLowerCase();
    ArrayList<Key> list = new ArrayList<Key>(16);
    String text, target;

    // TODO: NTN: I don't think we need the next two lines of Gordon code
    // if main help file exists, then make sure we can open it for read, otherwise continue processing
    fmhs = new TextFile( path+mainFileName );
    if (fmhs.find(IOResource.DIR_INPUT) && !fmhs.open(TextFile.INPUT|TextFile.NOABORT)) return -1;

    // build the top level splash screen
    target = optLC+"_tree_index";
    fall = getHelpFile(target);
    writeHTMLHeader (fall,opt+" Top Help",opt+" Option Tree",null);
    TextFile fver = new TextFile( M.io.name(".","version","txt",opt) );
    if (fver.open(TextFile.INPUT|TextFile.NOABORT)) {
      fall.writeln("<p align='center'>");
      while ((text=fver.read())!=null) {
        if(!text.startsWith("#")) {
          fall.writeln("  <code>" + text.replaceAll("  ", " &nbsp;") + "</code><br>");
        }
      }
      fall.writeln("</p>");
      fall.writeln("");
      fver.close();
    }
    writeHTMLTableHeader (fall);
    // help areas
    writeHTMLBullet (fall,"CONCEPTS",optLC+"_hlp_all.html","Operational Concepts - Help Tree");
    writeHTMLBullet (fall,"COMMANDS",optLC+"_exp_all.html","Configured Commands - Explain Tree");
    writeHTMLBullet (fall,"DOCUMENTS",optLC+"_doc_all.html","Supporting Documents - Docs Tree");
    // javadoc areas
    text="../api/nxm/"+optLC;
    if (M.io.fexists(jhdir+"../api/nxm/"+optLC+"/lib",true))
      writeHTMLBullet (fall,"LIBRARIES",text+"/lib/package-summary.html","Libraries for Basic Functions");
    if (M.io.fexists(jhdir+"../api/nxm/"+optLC+"/libg",true))
      writeHTMLBullet (fall,"GRAPHICS",text+"/libg/package-summary.html","Libraries for Graphics Functions");
    if (M.io.fexists(jhdir+"../api/nxm/"+optLC+"/libm",true))
      writeHTMLBullet (fall,"MATH",text+"/libm/package-summary.html","Libraries for Mathematical Functions");
    if (M.io.fexists(jhdir+"../api/nxm/"+optLC+"/lib",true))
      writeHTMLBullet (fall,"LIBRARIES",text+"/lib/package-summary.html","Libraries for Basic Functions");
    // release notes
    if (M.io.fexists(M.io.name("hlp","release","hlp",opt)))
      writeHTMLBullet (fall,"RELEASE",optLC+"_hlp_release.html","Release Notes for Current Version");
    // homepage bullet
    String homepage = M.io.name("docs","index","html",opt);
    if (M.io.fexists(homepage)) homepage="../../nxm/"+optLC+"/docs/index.html"; else homepage="";
    homepage = MA.getCS("/HOMEPAGE",homepage);
    if (homepage.length()>0) writeHTMLBullet (fall,"HOMEPAGE",homepage,"Option Tree Home Page");
    writeHTMLTableFooter (fall);
    // copyright/licensing notice
    TextFile fin,fout; int ok=0;
    fin = new TextFile( M.io.name(".","copyright","txt",opt) );
    fout = new TextFile( jhdir+optLC+"_copyright.txt");
    if (fin.find(IOResource.DIR_INPUT)) ok += fin.copy(fout);
    fin = new TextFile( M.io.name(".","license","txt",opt) );
    fout = new TextFile( jhdir+optLC+"_license.txt");
    if (fin.find(IOResource.DIR_INPUT)) ok += fin.copy(fout);
    if (ok>0) {
      fall.writeln("<br><center><a href='"+optLC+"_copyright.txt'>Copyright</a>");
      fall.writeln(" / <a href='"+optLC+"_license.txt'>License</a></center>");
    }
    writeHTMLFooter (fall);
    fall.close();

    // build list of help files
    buildMainHelpFiles(opt,list);

    // build help for all help files
    target = optLC+"_hlp_all";
    fall = getHelpFile(target);
    writeHTMLHeader (fall,"All Concepts",opt+" Concept Index",null);
    writeHTMLTableHeader (fall);
    for (Key key : list) {
      buildHelpFile(opt,key);
    }
    writeHTMLTableFooter (fall);
    writeHTMLFooter (fall);
    fall.close();

    // build help for all commands
    target = optLC+"_exp_all";
    fall = getHelpFile(target);
    writeHTMLHeader (fall,"All Commands",opt+" Command Index",null);
    writeHTMLTableHeader (fall);
    writeHTMLCommandTableTitle(fall);
    for (int i=0; i<M.dictionary.size; i++) {
      Parser command = new Parser ( M.dictionary.get(i) );
      buildExplainFile(opt,command);
    }
    writeHTMLTableFooter (fall);
    writeHTMLFooter (fall);
    fall.close();

    // build help for all documents
    String[] files; String lpath;
    target = optLC+"_doc_all";
    fall = getHelpFile(target);
    writeHTMLHeader (fall,"All Documents",opt+" Document Index",null);
    writeHTMLTableHeader (fall);
    path = M.io.name("HLP","","",opt);
    files = M.io.listFiles(path,"*","doc");
    lpath = "../../nxm/"+optLC+"/hlp";
    if (!M.io.fexists(jhdir+lpath,true)) lpath = null;
    for (int i=0; files!=null && i<files.length; i++)
      buildDocumentFile(opt,path,files[i],lpath);
    path = M.io.name("DOCS","","",opt);
    files = M.io.listFiles(path,"*","*");
    lpath = "../../nxm/"+optLC+"/docs";
    if (!M.io.fexists(jhdir+lpath,true)) lpath = null;
    for (int i=0; files!=null && i<files.length; i++)
      buildDocumentFile(opt,path,files[i],lpath);
    writeHTMLTableFooter (fall);
    writeHTMLFooter (fall);
    fall.close();

    fmhs.close();
    return 0;
  }

  // process the mainhelp.hlp file for top level screen
  private int buildMainHelpFiles (String opt, List<Key> list) {

    String path = M.io.name("HLP","MAINHELP","HLP",opt);
    fmhs = new TextFile(path);
    if (!fmhs.find(IOResource.DIR_INPUT)) return 0;
    if (!fmhs.open(TextFile.NOABORT)) return -1;

    String optLC = opt.toLowerCase();
    String target = optLC+"_hlp_main";
    TextFile fhtm = getHelpFile(target);
    writeHTMLHeader (fhtm,null,"Help: "+opt+"_MAIN",null);
    fhtm.writeln("<pre>");
    String text, section=null;

    fmhs.read(); // read past ~File MAINHELP
    while ((text=fmhs.read())!=null) {
      Key key = parseKey(text);
      if (key==null) fhtm.writeln(textToHtml(text,opt,"HLP"));
      else if (key.type.equals("FILE")) {
        section = key.label;
        fhtm.writeln("</pre>");
        writeHTMLFooter(fhtm);
        fhtm.close();
        target = optLC+"_hlp_"+key.label.toLowerCase();
        fhtm = getHelpFile(target);
        writeHTMLHeader (fhtm,null,"Help: "+opt+"_"+key.label,key.text);
        fhtm.writeln("<pre>");
      }
      else if (key.type.equals("ENTRY")) {
        if (section==null) list.add(key);
        String starget = target+"_"+key.label.toLowerCase();
        writeHTMLBullet (fhtm,key.label,starget+".html",key.text);
      }
    }
    fhtm.writeln("</pre>");
    writeHTMLFooter(fhtm);
    fhtm.close();
    fmhs.close();
    return 0;
  }

  // process one hlp file for buildHelp
  private int buildHelpFile (String opt, Key key) {

    String text,starget="";
    boolean section=false;
    String name = key.label;
    String optLC = opt.toLowerCase();
    TextFile fhtms;

    String fn = M.io.name("HLP",name,"HLP",opt);
    TextFile tf = new TextFile(fn);
    if (!tf.open(TextFile.NOABORT)) return -1;
    tf.setIncludeString("~Include");

    // tag the file created in buildMainHelpFiles
    String target = optLC+"_hlp_"+name.toLowerCase();
    writeHTMLBullet (fall,name,target+".html",key.text);
    TextFile fhtm = getHelpFile(target, false);

    while ((text=tf.read())!=null) {
      key = parseKey(text);
      if (key==null);
      else if (key.type.equals("ENTRY"));
      else if (key.type.equals("END")) text=null;
      else if (key.type.equals("COMMENT")) text=null;
      else if (key.type.equals("FILE")) text="";
      else if (key.type.equals("MAIN")) {
        if (section) buildHelpFileCloseSection(fhtm,/*ftoc*/null);
        section = false;
      }
      else if (key.type.equals("SUBSECTION")) {
        fhtm.writeln("<a name=\""+key.label+"\">");
        text = key.label + key.text;
      }
      else {
        if (section) buildHelpFileCloseSection(fhtm,/*ftoc*/null);
        starget = target+"_"+key.label.toLowerCase();
        fhtm = getHelpFile(starget);
        if (key.text.length()==0) key.text = getHelpKeyText(opt,name,key.label);
        writeHTMLHeader (fhtm,null,"Help: "+opt+"_"+name+"_"+key.label,key.text);
        fhtm.writeln("<pre>");
        if (key.type.equals("FUNCTION")) text = text.substring(1);
        else text = null;
        section = true;
      }
      if (section && text!=null) fhtm.writeln(textToHtml(text,opt,"HLP"));
    }
    if (section) buildHelpFileCloseSection(fhtm,/*ftoc*/null);
    tf.close();
    return 0;
  }

  // helper repeat function for buildHelpFile
  private void buildHelpFileCloseSection (TextFile fhtm, TextFile ftoc) {
    fhtm.writeln("</pre>");
    writeHTMLFooter(fhtm);
    fhtm.close();
  }

  // build HTML version of an explain file
  private int buildExplainFile (String opt, Parser command) {

    String name = command.get(1); // name looks like COMMAND;OPT:n
    char support = command.get(2).charAt(0);
    int lname = name.indexOf(';');
    int lendName = name.indexOf(':');
    if (lendName < 0) lendName = name.length();
    String option = name.substring(lname+1,lendName);
    if (!option.equals(opt)) return 0;

    name = name.substring(0,lname);
    TextFile fexp = new TextFile( M.io.name("EXP",name,"exp",opt) );
    if (!fexp.open(BaseFile.INPUT|BaseFile.NOABORT)) return 0;
    fexp.setIncludeString("~Include");
    String optLC = opt.toLowerCase();
    String nameLC = name.toLowerCase();
    String target = optLC+"_exp_"+nameLC;
    TextFile fhtm = getHelpFile(target);

    ////////////////////////////////////////////////////////////////////////////////////////////////
    //String text = fexp.read();
    //Key key = parseKey(text);
    //writeHTMLHeader (fhtm,"Command Help -"+name,name,key.text);
    //fhtm.writeln("<pre>");
    //fhtm.writeln(textToHtml(text,opt,"EXP"));
    //while ( (text=fexp.read()) != null) {
    //  fhtm.writeln(textToHtml(text,opt,"EXP"));
    //}
    ArrayList<String> fileText = new ArrayList<String>();
    String text;
    while ((text=fexp.read()) != null) {
      fileText.add(text);
    }

    Key key = null;
    try {
      key = parseKey(fileText.get(0));
      /* Try to catch multi-line headers, e.g.:
           APPSERVER - Server macro allowing clients to start/stop/monitor Midas macros
                       on this node.
       */
      final String leadingSpace = StringUtil.padRight("", key.label.length());
      for (int i = 1; i < fileText.size(); i++) {
        final String txt = fileText.get(i).toString();

        if (txt.startsWith(leadingSpace) && (!txt.trim().startsWith("<"))
                                        && (!txt.trim().startsWith("["))) {
          key.text += " " + txt.trim();
        } else {
          break;
        }
      }
      /* Make it a proper sentence: first letter capitalized, ends with a period. */
      key.text = key.text.trim();
      if (key.text.length() > 1) {
        key.text = key.text.substring(0,1).toUpperCase() + key.text.substring(1)
                 + ((key.text.endsWith(".")? "": "."));
      }
    } catch (Exception ex) {
      M.warning("Unable to build html from explain file for "+name+" command");
      //ex.printStackTrace();
    }
    String fileTextString = "Unknown - invalid or missing explain file";
    if (key != null) {
      fileTextString = key.text;
    }
    /* Write the header. */
    writeHTMLHeader (fhtm,"Command Help -"+name,name,fileTextString);
    /* Write the text. */
    fhtm.writeln("<pre>");
    for (String str : fileText) {
      fhtm.writeln(textToHtml(str,opt,"EXP"));
    }
    ////////////////////////////////////////////////////////////////////////////////////////////////
    String srcCodeName   = null;            // Name of the source file (e.g. "temp.java").
    String testMacroName = "test_"+nameLC;  // Name of the test macro - NO EXTENSION (e.g. "test_temp").
    String className     = null;            // The name of the class (null for non-Java).
    String area          = null;            // The option tree area for the command.
    boolean noTest = false;

    switch (support) {
      case 'M': area = "mcr";  srcCodeName = nameLC+".mm";  break;
      case 'H': area = "host"; srcCodeName = nameLC+".for"; break;
      case 'I': area = "intr"; className = name.charAt(0)+nameLC.substring(1); break;
      case 'P': area = "prim"; className = nameLC; break;
      default:  area = "prim"; className = nameLC; break;
    }

    /////////// CHECK FOR TEST MACRO
    TextFile configFile = new TextFile(M, "nxm."+opt+".test.test_config.tbl");
    String   testConfig = null;
    String   optOfTestMacro = opt;
    if (configFile.exists()) {
      nxm.sys.lib.Table configTbl = new nxm.sys.lib.Table(configFile);
      testConfig = configTbl.getS(name);
    }
    if (testConfig != null) {
      if (testConfig.equals("NOTEST")) {
        noTest = true;
      } else if (testConfig.equals("WEB") || testConfig.equals("BROKEN")) {
        // Test is fine, just not normally run.
      } else {
        testMacroName = "test_" + testConfig.toLowerCase(); // Tested by other macro.
        int iSemicolon = testMacroName.lastIndexOf(';');    // since 3.1.1 to support COMMAND;OTHER_OPT syntax
        if (iSemicolon > 0) {
          int optIndex = iSemicolon + 1;
          if (testMacroName.length() > optIndex) optOfTestMacro = testMacroName.substring(optIndex);
          testMacroName  = testMacroName.substring(0, iSemicolon);
        }
      }
    }

    /////////// SET PATH VARIABLES
    if (className != null) { srcCodeName = className+".java"; } // File name based on class name
    String srcPath  = relativeToDocDir(OptionTree.getOptPath(opt,area),   srcCodeName);
    String testPath = OptionTree.getOptPath(optOfTestMacro,"test");
    String relativeToTestPath = relativeToDocDir(testPath, testMacroName+".mm");
    String expPath  = opt.toLowerCase()+"_exp_"+name.toLowerCase()+".html";
    String docPath  = null;

    if (className != null) { // If Java, set names based on name of class.
      String dir = getJavadocDir(opt) + File.separator + "nxm"
                                      + File.separator + opt.toLowerCase()
                                      + File.separator + area;
      srcCodeName = className+".java";
      docPath = relativeToDocDir(dir, className+".html");
    }
    if (!noTest && !M.io.exists(relativeToTestPath + testMacroName+".mm")) {
      relativeToTestPath = null; // No test macro found
    }

    fhtm.writeln("</pre>");
    writeHTMLFooter(fhtm);
    fall.writeln("<tr>");
    fall.writeln("  <td><a href='"+expPath+"'>"+name+"</a></td>");
    fall.writeln("  <td>"+fileTextString+"</td>");
    fall.writeln("  <td align='center'>"+support+"</td>");
    if (docPath != null) {
      fall.writeln("  <td align='center'><a href='"+docPath+"'>doc</a></td>");
    } else {
      fall.writeln("  <td align='center'><a href='"+expPath+"'>doc</a></td>"); // Default to exp file.
    }
    fall.writeln("  <td align='center'><a href='"+srcPath+"'>src</a></td>");
    if (noTest) {
      fall.writeln("  <td align='center'>n/a</td>");
    } else if (relativeToTestPath != null) {
      fall.writeln("  <td align='center'><a href='"+relativeToTestPath+"'>"+testMacroName+"</a></td>");
    } else {
      fall.writeln("  <td align='center'><i>missing</i></td>");
    }
    fall.writeln("</tr>");
    fhtm.close();
    fexp.close();
    return 0;
  }

  /** Gets a relative link to destDir from the documentation dir. */
  private String relativeToDocDir (String destDir, String fname) {
    java.io.File dest   = new java.io.File(destDir);
    java.io.File docDir = new java.io.File(getHelpDir());
    String dir = FileUtil.getRelativeDir(docDir, dest);

    dir = dir.replace(java.io.File.separatorChar, '/'); // Convert to URL syntax

    return dir + fname;
  }

  // build Help links for the document files
  private int buildDocumentFile (String opt, String path, String dfname, String lpath) {
    FileName fn = new FileName(path+dfname,true);
    String name = fn.getRoot()+"."+fn.getExt();
    String url = fn.getFullName();
    if (lpath!=null) url = lpath + "/" + name;
    writeHTMLBullet (fall,name,url,"document file");
    return 0;
  }

  /** Find the help file reference for an embedded see also help entry. */
  private String getHelpTarget (String name) {
    String retValue = "sys_hlp_help_notfound";
    String text, section=null;
    String path = (String)M.results.get("OPT.PATH");
    String[] opts = StringUtil.cleanSplit(path,",");
    for (int i=opts.length-1; i>=0; i--) {
      String opt = opts[i];
      String fn = M.io.name("HLP","MAINHELP","HLP",opt);
      TextFile tf = new TextFile(fn);
      if (!tf.find(IOResource.DIR_INPUT)) continue;
      if (!tf.open(TextFile.NOABORT)) continue;
      while ((text=tf.read())!=null) {
        Key key = parseKey(text);
        if (key==null);
        else if (key.type.equals("FILE")) {
          section = key.label;
          if (name.equalsIgnoreCase(key.label))
            retValue = opt+"_hlp_"+section;
        }
        else if (key.type.equals("ENTRY")) {
          if (name.equalsIgnoreCase(key.label))
            retValue = opt+"_hlp_"+section+"_"+key.label;
        }
      }
    }
    return retValue;
  }

  /** Find the help file reference for a Help Key without text.

      @param opt  The NeXtMidas option.
      @param file The file name.
      @param name The the label in the file.

      @todo Better explain this method.
   */
  private String getHelpKeyText (String opt, String file, String name) {
    String text; boolean section=false;
    String fn = M.io.name("HLP","MAINHELP","HLP",opt);
    TextFile tf = new TextFile(fn);
    if (!tf.open(TextFile.NOABORT)) return "noMainHelp";
    while ((text=tf.read())!=null) {
      Key key = parseKey(text);
      if (key==null);
      else if (key.type.equals("FILE")) {
        section = file.equalsIgnoreCase(key.label);
      }
      else if (key.type.equals("ENTRY")) {
        if (section && name.equalsIgnoreCase(key.label)) return key.text;
      }
    }
    return "unknown";
  }

  /** Build a master index pointing to all configured option trees.

      @return Always returns <code>0</code>.
   */
  private int buildMasterIndex () {
    String[] options = getKnownOptionTrees();
    fall = getHelpFile("index");
    writeHTMLHeader (fall,null,"NeXtMidas Master Help Index","Configured Option Trees");
    fall.writeln("<center><a href='../search.html' target='_top'>Show Search Frame</a> &nbsp; " +
                 "        <a href='index.html' target='_top'>Hide Search Frame</a></center>");

    writeHTMLTableHeader (fall);
    for (int i=0; i < options.length; i++) {
      String fileName = options[i].toLowerCase() + "_tree_index.html";
      writeHTMLBullet (fall, options[i], fileName,
                       "The " + options[i] + " option tree");
    }
    writeHTMLTableFooter (fall);

    // JC4: START HACK: Add in documentation links.
    final String verClass    = Installation.NM_VERSION.substring(0,
                                Installation.NM_VERSION.lastIndexOf('.')+1)+"X";
    final String releaseFile = "ReleaseNotes_" + verClass.replace('.','_') + ".html";
    final String releaseName = "Release Notes " + verClass;

    fall.writeln("<h3 align='center'>General NeXtMidas Documentation</h3>");
    fall.writeln("<table border='1' cellpadding='2' cellspacing='0' align='center'>");
    fall.writeln("  <tr>");
    fall.writeln("    <td><a href='../../nxm/sys/docs/" + releaseFile + "'>RN</a></td>");
    fall.writeln("    <td>" + releaseName + "</td>");
    fall.writeln("  </tr>");
    fall.writeln("  <tr>");
    fall.writeln("    <td><a href='../api/index.html'>API</a></td>");
    fall.writeln("    <td>Source Code Documentation</td>");
    fall.writeln("  </tr>");
    fall.writeln("  <tr>");
    fall.writeln("    <td><a href='../usersguide/index.html'>NMUG</a></td>");
    fall.writeln("    <td><i>NeXtMidas User's Guide</i></td>");
    fall.writeln("  </tr>");
    fall.writeln("  <tr>");
    fall.writeln("    <td><a href='../training/index.html'>TRAINING</a></td>");
    fall.writeln("    <td><i>NeXtMidas Training Class</i></td>");
    fall.writeln("  </tr>");
    fall.writeln("  <tr>");
    fall.writeln("    <td><a href='../../nxm/sys/docs/InstallationGuide.html'>NMIG</a></td>");
    fall.writeln("    <td><i>NeXtMidas Installation Guide</i></td>");
    fall.writeln("  </tr>");
    fall.writeln("</table>");
    // JC4: END HACK

    writeHTMLFooter (fall);

    fall.close();
    return 0;
  }


  /** Open a file in the JavaHelp area for writing.
      @param targetName The name of the file to open.
      @param open       If it should be opened.
      @return The text file.
   */
  private TextFile getHelpFile (String targetName, boolean open) {
    TextFile tf = new TextFile (jhdir+targetName+".html");

    if (open) tf.open(BaseFile.OUTPUT|BaseFile.MKDIR|BaseFile.FORCEABORT);
    return tf;
  }


  /** Open a file in the JavaHelp area for writing. This is the same as
      <code>getHelpFile(target,true)</code>.

      @param target The name of the file to open.
      @return The text file.
      @see #getHelpFile(String,boolean)
   */
  private TextFile getHelpFile (String target) {
    return getHelpFile(target, true);
  }

  /** Translate a line of text to html.

      @param text The initial text.
      @param opt  The NeXtMidas option area.
      @param area The NeXtMidas area.

      @return The translated string.
   */
  private String textToHtml (String text, String opt, String area) {
    try {
      text = StringUtil.toHTML(text);
      text = translateSeeAlso(text,opt,area);
    } catch (Exception e) {
      // ignore and just use the inputted text
      M.warning("Help.textToHtml - Cannot translate the following text: " + text);
      if (debug) M.warning(e);
    }
    return text;
  }

  /** Search for a "See Also" and its synonyms and make the appropriate translation.
      @param text The text to search.
      @param opt  <i>This is not used as this time.</i>
      @param area The area this is in.
      @return The results.
      @todo Better explain this method.
   */
  private String translateSeeAlso (String text, String opt, String area) {
    //JC4 - Next line removed to maintain consistency with raw explain file
    //text = text.trim();
    int i,j,k=text.length();
    i=text.toUpperCase().indexOf("SEE");
    if (i<0) return text;
    i+=3;                                              // past "SEE"
    if (text.indexOf(':',i)==i) i++;                   // past ':' for for "SEE:"
    while(text.indexOf(' ',i)==i) i++;                 // past spaces
    if (text.toUpperCase().indexOf("ALSO",i)==i) i+=4; // past ALSO
    if (text.indexOf("the ",i)==i) i+=4;               // past "the"
    if (i>=k) return text;                             // nothing after "see also" (return)
    if (text.indexOf(':',i)==i) i++;                   // past ':' again for "SEE ALSO:"
    if (i>=k) return text;                             // nothing after "see also:" (return)

    j=text.indexOf(". ",i); if (j<i) j = text.length();
    for (;text.charAt(j-1)==' '|| text.charAt(j-1)=='.'; j--);
    String stext = text.substring(0,i);
    for (k=j; i<k; i=j) {
      // find the end of this parse
      for (;i<k && text.charAt(i)==' '; i++,stext+=" "); // past spaces
      if (i>=k) break;  // nothing else after see also
      j=text.indexOf(',',i); if (j<0||j>k) j=k;  // to comma or end of line

      String[] tokens = StringUtil.cleanSplit(text.substring(i,j)," ");

      // handle special cases: see command explain file, or see entry help file
      if (tokens.length>=3 && tokens[2].equalsIgnoreCase("FILE")) { //Made >= instead of >
        if (tokens[1].equalsIgnoreCase("EXPLAIN")) area = "EXP";
        else if (tokens[1].equalsIgnoreCase("EXP")) area = "EXP";
        else if (tokens[1].equalsIgnoreCase("HELP")) area = "HLP";
        else break;
        name = tokens[0];
        j=k=i+tokens[0].length()+tokens[1].length()+tokens[2].length()+2;
      }
      // handle special cases: see explain on command, or see help on entry
      else if (tokens.length>=3 && tokens[1].equalsIgnoreCase("ON")) { //Made >= instead of >
        if (tokens[0].equalsIgnoreCase("EXPLAIN")) area = "EXP";
        else if (tokens[0].equalsIgnoreCase("EXP")) area = "EXP";
        else if (tokens[0].equalsIgnoreCase("HELP")) area = "HLP";
        else break;
        name = tokens[2];
        j=k=i+tokens[0].length()+tokens[1].length()+tokens[2].length()+2;
      }
      else if (tokens.length==3) {
        if (tokens[1].equalsIgnoreCase("EXPLAIN")
            || tokens[1].equalsIgnoreCase("EXP")) {
          area = "EXP";
        }
        else if (tokens[1].equalsIgnoreCase("HELP")){
          area = "HLP";
        }
        else break;
        name=tokens[2]; opt=tokens[0];
        j=k=i+tokens[0].length()+tokens[1].length()+tokens[2].length()+2;
      }
      else if (tokens.length==2) {
        if (tokens[0].equalsIgnoreCase("EXPLAIN")
            || tokens[0].equalsIgnoreCase("EXP")) {
          area = "EXP";
        }
        else if (tokens[0].equalsIgnoreCase("HELP")){
          area = "HLP";
        }
        else break;
        name=tokens[1];
        j=k=i+tokens[0].length()+tokens[1].length()+1;
      }
      else {
        boolean resolved = false;
        if (tokens.length>3) {
          if (tokens[0].equalsIgnoreCase("EXPLAIN")
              || tokens[0].equalsIgnoreCase("EXP")) {
            area = "EXP";
            resolved = true;
          }
          else if (tokens[0].equalsIgnoreCase("HELP")){
            area = "HLP";
            resolved = true;
          }
          if (resolved)
          {
            name=tokens[1];
            j=k=i+tokens[0].length()+tokens[1].length()+1;
          }
        }
        if (tokens.length>=1 && !resolved) {
          name=tokens[0];
          if (tokens.length>1) j=k=i+tokens[0].length();
          resolved = true;
          //Moved this code down after stripping punctuation
          //if (name.toUpperCase().indexOf("NXM.") >= 0) {
          // Take care of nxm.<opt>.<area>.<name>.[<ext>]
          //  subtokens = StringUtil.cleanSplit(text.substring(i,j),".");
          //}
        }//try to resolve see also when many tokens follow
        if (!resolved)
        {
          break;  // wrong number of tokens, ignore
        }
      }

      //Finally, remove trailing periods and spaces from name
      name = name.trim();
      //while (name.endsWith(".")){
      //  name = name.substring(0,name.length()-1);
      //}

      //remove trailing punctuation from name--1
      String[] parts = name.split("\\p{Punct}");

      int punct = 0;
      if((parts.length > 0) && (parts[0].length() != name.length())) {
        int end = name.lastIndexOf(parts[parts.length-1]) + parts[parts.length-1].length();
        punct = name.length() - end;
        name = name.substring(0,end);
        j=j-punct;
      }

      String[] subtokens = null; // Tokens for nxm.<opt>.<area>.<name>.[<ext] syntax
      if (name.toUpperCase().indexOf("NXM.") >= 0) {
        // Take care of nxm.<opt>.<area>.<name>.[<ext>]
        subtokens = StringUtil.cleanSplit(text.substring(i,j),".");
      }

      if (area.equalsIgnoreCase("EXPL")) area = "EXP";
      if (area.equalsIgnoreCase("HELP")) area = "HLP";
      if (area.equalsIgnoreCase("EXP")) {   // expand possible abbreviation
        if (VALID_COMMAND_REGEX.matcher(name).matches()) { // if name is a valid command
          Args args = Shell.parseCommand (M,name);         // lookup the command
          name=args.name; opt= args.option;
        }
      }

      String target = opt+"_"+area+"_"+name;
      if (area.equalsIgnoreCase("HLP")) {   // find help target
        target = getHelpTarget (name);
      }

      if (subtokens != null) {
        if (subtokens.length > 3 ) {
          // Take care of nxm.<opt>.<area>.<name>.[<ext>]
          String subExt  = "html";
          String subOpt  = subtokens[1].toLowerCase();
          String subArea =  subtokens[2].toLowerCase();
          String subName =  subtokens[3];
          String subPath = "../api/"+subOpt+"/nxm/";
          if (subtokens.length > 4 ) {
            // When an extension is included a source file is assumed and
            // we build a source URL
            subExt = subtokens[4];
            subPath = "../../nxm/";
          }
          target = subPath+subOpt+"/"+subArea+"/"+subName+"."+subExt;
        } else {
          M.warning("Unknown See Reference string = "+text.substring(i,j));
        }

      } else {
        target = target.toLowerCase()+".html";
      }
      //if (opt.equals("") && subtokens != null && subtokens.length <= 3)
      if(target.endsWith("_"+area.toLowerCase()+"_.html") || target.endsWith("notfound.html"))
      {
        break;
      }
      stext += "<a href='"+target+"'>"+text.substring(i,j)+"</a>";
      if (j<k) stext += text.charAt(j++);
    } // end for loop
    if (i<text.length()) stext += text.substring(i);
    return stext;
  }

  /** Writes the header of a HTML file to a text file.
      @param tf    The file to write the tags to.
      @param title The title of the screen.
      @param head1 The first heading.
      @param head2 The second heading.
      @see #writeHTMLFooter
   */
  private void writeHTMLHeader (TextFile tf, String title, String head1, String head2) {
    if (title==null) title=head1;
    tf.writeln("<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">");
    tf.writeln("<html>");
    tf.writeln("<head>");
    tf.writeln("  <title>"+title+"</title>");
    tf.writeln("  <link href='" + STYLE_SHEET_HREF + "' type='text/css' rel='stylesheet'>");
    tf.writeln("</head>");
    tf.writeln("<body><div>");
    if (head1!=null) tf.writeln("<h2 align='center'>"+head1+"</h2>");
    if (head2!=null) tf.writeln("<h3 align='center'>"+head2+"</h3>");
  }
  /** Writes the HTML for the Command table titles
      @param tf    The file to write the tags to.
      @see #writeHTMLHeader
   */
  private void writeHTMLCommandTableTitle (TextFile tf) {
    tf.writeln("<tr>");
    tf.writeln("  <td>Command/Explain</td>");
    tf.writeln("  <td>Description</td>");
    tf.writeln("  <td align='center'>SUP</td>");
    tf.writeln("  <td align='center'>API</td>");
    tf.writeln("  <td align='center'>Source</td>");
    tf.writeln("  <td align='center'>Test Macro</td>");
    tf.writeln("</tr>");
  }

  /** Writes the header of a HTML table to a text file.
      @param tf The file to write the tags to.
      @see #writeHTMLTableFooter
   */
  private void writeHTMLTableHeader (TextFile tf) {
    tf.writeln("<table border='1' cellpadding='2' cellspacing='0' align='center'>");
    inTable=true;
  }

  /** Writes the footer of a HTML table to a text file.
      @param tf The file to write the tags to.
      @see #writeHTMLTableHeader
   */
  private void writeHTMLTableFooter (TextFile tf) {
    tf.writeln("</table>");
    inTable=false;
  }

  /** Writes a HTML table row with a link to a text file.
      @param tf   The file to write the tags to.
      @param name The name of the entry.
      @param ref  The link/URL reference.
      @param text The description of the entry.
      @see #writeHTMLTableHeader
   */
  private void writeHTMLBullet (TextFile tf, String name, String ref, String text) {
    if (inTable) {
      tf.writeln("<tr><td><a href='"+ref+"'>"+name+"</a></td>"
               + "<td>"+text+"</td></tr>");
    } else {
      tf.writeln("  <a href='"+ref+"'>"+name+"</a> - "+text);
    }
  }

  /** Writes the footer of an HTML file to a text file.
      @param tf The file to write the tags to.
      @see #writeHTMLHeader
   */
  private void writeHTMLFooter (TextFile tf) {
    tf.writeln("</div></body>");
    tf.writeln("</html>");
  }

  /** Determines where the help directory is. This will <b>not</b> include
      any trailing path separator.

      @return The location of the help directory.

      @since NeXtMidas Version 1.8.0

      @see #getJavadocDir(String)

      @todo Move the definition for the path to the build.props file.
   */
  public String getHelpDir () {
    final int delimLength = FileName.LOCAL_PATH_DELIM.length();

    return HELP_PATH.substring(0, (HELP_PATH.length() - delimLength));
  }

  /** Determines where the javadoc directory is. This is the directory in which
      javadoc places the generated HTML files. This will <b>not</b> include any
      trailing path separator.

      @param opt The option tree to get the javadoc directory for.

      @return The location of the javadoc directory.

      @since NeXtMidas Version 1.8.0

      @see #getHelpDir()
   */
  public String getJavadocDir (String opt) {
    return JAVADOC_PATH + opt.toLowerCase();
  }

  /** This will get a list of all the option trees that have had their documents
      compiled at some point in the past. <br>
      <br>
      This works by listing all of the directories (each representing an option
      tree) that are in the <code>htdocs/api</code> directory.

      @return The list of option trees that have had their documents compiled.
              All values in this array will be capitalized.

      @since NeXtMidas 1.8.0
   */
  public static String[] getKnownOptionTrees () {
    String[] options = new String[0];

    try {
      /* This will filter out all non-directories. */
      final FilenameFilter filter = new FilenameFilter() {
        public boolean accept(java.io.File dir, String name) {
          return (new java.io.File(dir, name)).isDirectory();
        }
      };
      options = (new java.io.File(JAVADOC_PATH)).list(filter);

    } catch(Exception exception) {
      options = new String[0];
    }

    /* Capitalize the names. */
    for(int i = 0; i < options.length; i++) {
      options[i] = options[i].toUpperCase();
    }

    return options;
  }

  // list numbered list of search entries to the screen
  private void listResults (List<Document> list, String query) {
    boolean showpath = MA.getState("/SHOWPATH");
    int  len  = list.size();
    String statusMessage = len + " result";
    if (len == 0 || len == 1) statusMessage += " found for: ";
    else                      statusMessage += "s found for: ";
    M.terminal.more(Terminal.INIT);
    M.terminal.more(statusMessage + query);
    for (int i=0; i<len && M.terminal.more(); i++) {
      Object obj = list.get(i);
      if (obj instanceof Document) {
        Document doc = (Document) obj;
        String doctitle = doc.get("title"); // get its title
        String path     = doc.get("path");  // get its path field
        String filename = path.substring(path.lastIndexOf('/') + 1);
        String pathFromNMROOT = filename;
        int nmrootIndex = path.indexOf(NMROOT);
        if (nmrootIndex != -1) pathFromNMROOT = path.substring(NMROOT.length());    // move past NMROOT in path
        if ((doctitle == null) || doctitle.equals("")) { //use the path if it has no title
          doctitle = pathFromNMROOT;
        } else if (showpath) {
          doctitle += " (" + pathFromNMROOT +")";
        } else {
          doctitle += " (" + filename +")";
        }
        String index = " "+i+") "; if (i<10) index = " "+index;
        M.terminal.more (index + " " + doctitle);
      }
    } // end for loop
    M.terminal.more(Terminal.EXIT);
  }

  /** Handle user's interaction with search results from terminal, such as selecting and
      viewing the search result contents.
     @param queryStr search query string
   */
  private void doSearchInteraction (String queryStr) {
    if (queryStr==null || queryStr.trim().length() == 0) {
      queryStr = M.terminal.getInput("Enter search string: ", "").trim();
      if (queryStr.length()==0) return; // nothing to search for
    }

    // search help index for specified query string
    List<Document> results = searchIndex(queryStr);

    int len = results.size();
    if (len > 1) listResults(results, queryStr); // output the list of matches if more than 1 result

    // loop on what to do next
    for (int ilist=0; ilist<len; ilist++) {
      String ans="N";
      if (len>1) ans = M.terminal.getInput("Enter item number or [N]ext/Quit/List ?: ","");
      if (M.terminal.getInterrupt()>0) {               break; }

      if (ans.length()==0) ans="N";
      else ans = ans.trim().toUpperCase();

      if (ans.startsWith("Q")) {
                                                       break;
      }
      else if (ans.startsWith("L")) {
        listResults(results, queryStr);
        ilist--;                                       continue;
      }
      else if (ans.startsWith("N")) {
        if (ilist==len) {                              break; }
      }
      else {
        try {
          ilist = Integer.parseInt(ans);
        } catch (NumberFormatException nfe) {
          ilist--;                                     continue;
        }
      }

      if (ilist >= len) {                               break; }
      Object obj = results.get(ilist);

      // display result content if it is a Document Hit
      if (obj instanceof Document) {
        Document doc   = (Document) obj;
        String docPath = doc.get("path");  // get its path field
        TextFile tf    = new TextFile (docPath);

        if (!tf.open(TextFile.NOABORT)) continue;
        M.terminal.more(Terminal.INIT);
        M.terminal.more("Displaying file: " + docPath);
        String text;
        boolean insidePreTag = false;
        while (M.terminal.more() && (text=tf.read())!=null) {
          if (!insidePreTag && PRE_OPEN_REGEX.matcher(text).matches())  insidePreTag = true;
          if ( insidePreTag && PRE_CLOSE_REGEX.matcher(text).matches()) insidePreTag = false;
          text = StringUtil.html2Text(text);                      // convert HTML to text
          if (!insidePreTag && text.trim().length()==0) continue; // ignore empty lines inside <pre> tags
          M.terminal.more(text);
        }
        M.terminal.more(Terminal.EXIT);
        tf.close();
        M.terminal.writeln("For better formatting of the contents, view the following in your browser.\n "+docPath);
      }
    } // end for loop
  }

  /** Search the help index for matches for the specified query string.
      @param queryStr query string to search for (this cannot be null)
      @return Empty List if no results are found, otherwise a List of Hits.
      @since NeXtMidas 2.5.0
   */
  private List<Document> searchIndex (String queryStr) {
    List<Document> results = new ArrayList<Document>(0);
    if (searcher == null) {
      String indexPath = sysBuildProps.getPath("help.index.dir");
      if (indexPath == null || indexPath.length()==0) M.error("help.index.dir not set in build.props!");
      try {
         searcher = new Searcher(indexPath);
      } catch (IOException ioe) {
        M.warning("Cannot initialize help index search. " + ioe);
      }
    }

    try {
      if (searcher != null) results = searcher.query(queryStr, 100); // execute search (limit to 100 results)
    } catch (ParseException pe) {
      String message = "Invalid search string. " + pe.getMessage();
      M.warning(message);
    }
    return results;
  }

  /** Index files for help and searches
      @param filter Filename filter to use specified the documents to index
      @param additionalDirs '|' separated list of additional directories to
                            include in the list of locations to index documents.
      @return true if indexing is successful, otherwise false
      @since NeXtMidas 2.5.0
   */
  private boolean indexFiles (String filter, String additionalDirs) {
    boolean result         = false;
    String  indexPath      = sysBuildProps.getPath("help.index.dir");
    int     maxFieldLength = Integer.parseInt(sysBuildProps.getProperty("help.maxFieldLength", "100000"));
    int     numDocs        = 0; // total number of documents indexed
    String  locations      = sysBuildProps.getProperty("help.index.locations");
    String  useParentDirs  = sysBuildProps.getProperty("help.index.useparent");
            filter         = sysBuildProps.getProperty("help.index.filter", filter);

    if (indexPath==null || indexPath.length()==0) M.error("help.index.dir not set in build.props!");
    if (additionalDirs!=null) locations += "|" + additionalDirs;

    M.info("Creating help index at " + indexPath + " filter=" + filter +
        "\nThis may take a minute or two...");
    long startTime = System.currentTimeMillis();

    try {
      Indexer indexer = new Indexer(indexPath, true, maxFieldLength, filter);
      indexer.debug   = debug; // pass on debug flag to Indexer

      if (locations!=null && locations.length()>0) {
        indexer.setFilter(filter);
        if (debug) M.info("indexing dirs = "+locations+"  -  useparentDirs = "+useParentDirs);
        String[] dirs = locations.split("[|]");
        for (int i=0; i<dirs.length; i++) {
          String dir = dirs[i].trim();
          if (dir.length()>0) {
            boolean useParentDir = (useParentDirs.indexOf(dir) != -1); // false for most cases
            indexer.indexDocs(new File(dir), true, useParentDir);
          }
        }
      }

      numDocs = indexer.getDocCount(); // get total # docs indexed
      indexer.close();                 // optimize and close indexer
      indexer = null;                  // remove reference indexer to help GC free this object
      result  = true;                  // index is a success
    } catch (Exception e) {
      M.warning("Problem trying to index documents for help! " + e);
      if (debug) e.printStackTrace();
    }
    long endTime = System.currentTimeMillis();
    M.info("Help index created. Indexed "+ numDocs +" files in "+ (endTime - startTime)/1000 +" seconds.");

    return result;
  }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Inner classes
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** Help indexer to create lucene index for searching help content.
      This class is only used internally by Help.
      @author Neon Ngo
      @since NeXtMidas 2.5.0
   */
  private static class Indexer {
    private IndexWriter writer;    // Lucene index writer to create and add documents to index
    private FnFilter    fnFilter;  // filename filter for files to index
    protected static Analyzer analyzer = new CustomAnalyzer(); // Lucene analyzer for creating/searching help index.
    private boolean debug = false; // enable/disable debugging outputs

    /** Construct a help indexer to create lucene index for searching help content.
       @param indexPath path to directory where index will be stored
       @param overwrite true to create the index or overwrite the existing one; false to append to
                        the existing index
       @param maxFieldLength maximum number of terms that will be indexed for a single field in a document
       @param filter Filename filter for files to index. If null the default filter
                     ("*.html|*.htm|*.txt") will be used. Note: A blank (empty) string, means NONE.
                     See {@link FileUtil#createFnFilter(String, String)} for filename filter format.
       @throws IOException
       @since NeXtMidas 2.5.0
     */
    private Indexer (String indexPath, boolean overwrite, int maxFieldLength, String filter)
      throws IOException {
      writer = new IndexWriter(indexPath, analyzer, overwrite);
      writer.setMaxFieldLength(maxFieldLength);
      fnFilter = FileUtil.createFnFilter(filter,  "*.html|*.htm|*.txt");
    }

    /** Make sure the indexer writer data is flush to disk and committed by calling
        {@link #close()} when this object is garbage collected. */
    protected void finalize () throws Throwable {
      super.finalize();
      close(); // make sure writer has been closed
    }

    /** Optimize and close index writer (i.e. commit changes to disk) for search to use.
        If index has been closed already, this method will ignore that exception. */
    private void close () {
      if (writer!=null) {
        try {
          writer.optimize();
          writer.close();
        } catch (AlreadyClosedException ace) {
          // It is safe to ignore this exception, writer has been close already
        } catch (Exception e) {
          Shell.printStackTrace("Exception caught in Help.Indexer.close() ",e);
        }
      }
    }

    /** Set the filename filter for files to index.
        @param filter Filename filter for files to index. A blank (empty) string, means NONE.
                      See {@link FileUtil#createFnFilter(String, String)} for filename filter format.
     */
    private void setFilter (String filter) {
      fnFilter = FileUtil.createFnFilter(filter, "");
    }

    /** Return the number of documents currently in the help index writer.
        @return number of documents currently in the help index writer
        @throws AlreadyClosedException if close() has been called before this.
      */
    private int getDocCount () throws AlreadyClosedException {
      int count = 0;
      if (writer!=null) count = writer.docCount();
      return count;
    }

    /** This is the same as indexDocs(file, recursive, false).
        See {@link #indexDocs(File, boolean, boolean)}.
     */
    private void indexDocs (File file, boolean recursive) throws Exception {
      indexDocs(file, recursive, false);
    }

    /** Index specified file or files in the specified path
        @param file file or path to directory to index files for help and searches.
        @param recursive set to true if index should goto subdirectories recursively
        @param useParentDir set to true if the stored path in the index for this file or files
               under this directory should point to the parent directory. This is needed for the
               training and user's guides files under the rawfiles directory.
        @throws Exception
     */
    private void indexDocs (File file, boolean recursive, boolean useParentDir) throws Exception {
      if (file.isDirectory()) {                  // if a directory
        String[] files = file.list();            // list its files
        Arrays.sort(files);                      // sort the files
        for (int i = 0; i < files.length; i++) { // add files to index
          File afile = new File(file, files[i]);
          if (file.isDirectory() && recursive) indexDocs(afile, recursive, useParentDir); // recursively index subdirectory
          else if (fnFilter.accept(afile))     addDocToIndex(afile, useParentDir);
        }
      } else if (fnFilter.accept(file)) {
        addDocToIndex(file, useParentDir);
      }
    }

    /** Add the following file to the help index using the index writer.
     @param file the file to add to the help index
     @param useParentDir set to true if want to point file path to existing file in parent directory.
                         if file does not exist in parent directory, this flag has no effect.
     @throws IOException
     @throws InterruptedException
     @throws CorruptIndexException
     */
    private final void addDocToIndex (File file, boolean useParentDir) throws IOException, InterruptedException {
      if (debug) System.out.println("Help.Indexer: parsing " + file);
      Document doc = HTMLDocument.Document(file); // parse document
      String path  = file.getPath();
      String name  = file.getName();

      if (useParentDir) {
        int pos = path.lastIndexOf(File.separatorChar);
        pos = (pos<=0 ? -1 : path.lastIndexOf(File.separatorChar, pos-1));
        if (pos != -1) {
          String fileInParentDir = path.substring(0, pos+1) + name;
          File fileInParent = new File(fileInParentDir);
          if (fileInParent.exists()) {
            // point file path to file in parent directory (since it exists)
            String newValue = fileInParentDir.replace(File.separatorChar, '/');
            doc.removeField("path"); // remove previously added path
            doc.add(new Field("path", newValue, Field.Store.YES, Field.Index.UN_TOKENIZED));
          }
        }
      }

      // get base filename (which would normally be Javadoc class API or HTML generated version of explain file)
      int dotIndex = name.lastIndexOf('.');
      if (dotIndex != -1) name = name.substring(0, dotIndex); // strip off file extension

      // if we have HTML explain file, strip out option tree and area from it's filename to get command name
      int expIndex = name.indexOf("_exp_");
      if (expIndex != -1) {
        name = name.substring(expIndex + "_exp_".length());
        // add command name in field "explain" to boost the search results for Commands
        doc.add(new Field("explain", name, Field.Store.YES, Field.Index.TOKENIZED));
      }

      // add class/command name in field "midas" to boost the search results for Commands & Classes
      doc.add(new Field("midas", name, Field.Store.YES, Field.Index.TOKENIZED));

      if (debug) System.out.println("Help.Indexer: adding " + path);
      writer.addDocument(doc);                    // add document to index
    }
  } // end class Indexer

  /** Help index searcher for querying/finding content in the lucene help index.
      This class public because it is used by {@link nxm.sys.libg.MJHelp}.
      @author Neon Ngo
      @since NeXtMidas 2.5.0
   */
  public static class Searcher {
    private IndexSearcher searcher;    // lucene index searcher
    private QueryParser   queryParser; // lucene query parser
    /** search the following fields in the index */
    private static final String[] search_fields = {"midas", "explain", "title", "contents"};

    /** Construct a help searcher for querying the content in the specified lucene help index.
        @param indexPath path to directory to lucene help index
        @throws IOException
        @since NeXtMidas 2.5.0
     */
    public Searcher (String indexPath) throws IOException {
      Map<String,Float> boost = new HashMap<String,Float>(); // NTN: I got key/value type from MultiFieldQueryParser source code

      boost.put("explain", Float.valueOf(6)); // boost results for the explain field by 6
      boost.put("midas",   Float.valueOf(4)); // boost results for the midas field by 4
      boost.put("title",   Float.valueOf(2)); // boost results for the title field by 2
      queryParser  = new MultiFieldQueryParser(search_fields, Indexer.analyzer, boost);
      // set default query terms to be ANDed together
      queryParser.setDefaultOperator(QueryParser.Operator.AND);
      queryParser.setAllowLeadingWildcard(true);
      setIndexPath(indexPath);
      BooleanQuery.setMaxClauseCount(2048); // increase from default of 1024 so we can do query for "n*"; TODO: move to config file and/or add a switch to change this
    }

    /** Use Shell to display info/warn messages if Shell.getNmRoot() is set,
        otherwise display to standard output. This method is needed because
        when used in applet context, NMROOT may not be defined which causes
        the Shell to have trouble initializing which causes
        @param msg String to display as an info message
        @param ex  if exception exists then this becomes a warning message
     */
    void displayMessage (String msg, Exception ex) {
      if (Shell.getNmRoot()!=null) {
        if (ex!=null) Shell.printStackTrace(msg, ex);
        else          Shell.info(msg);
      } else {
        String prepend = (ex!=null) ? "WARN: " : "INFO: ";
        System.out.println(prepend+msg);
        if (ex!=null) ex.printStackTrace();
      }
    }

    /** Set the path to help index for searches.
        @param indexDir path to directory containing the help index
    */
   private void setIndexPath (String indexDir) throws IOException {
     displayMessage("HELP: setting indexPath to " + indexDir, null);
     try {
       searcher = new IndexSearcher(indexDir); // NOTE: this operation is slow for large
                                               // indices (much slower than the search itself)
     } catch (IOException e) {
       // any error that happens is probably due to a permission problem or
       // non-existent index path or a corrupt index
       String msg = "Could not initialize help for index path='"+ indexDir +
                    "'. Did you run 'HELP/index' or do you have the appropriate file/dir permissions? " + e;
       displayMessage(msg, e);
       throw (e); // re-throw Exception up the chain
     }
   }

   /** Same as {@link #query(String, int)} with a max search result limit of 100.
       @deprecated Since NeXtMidas 2.7.1 use {@link #query(String, int)} instead
    */
   @Deprecated public List<Document> query (String queryString) throws ParseException {
     return query(queryString, 100);
   }

   /** Query the help index to find matches for the specified query string.
       @param queryString search string to find in the help index
       @param maxResults  max search results to return. Use 0 or -1 for unlimited.
       @return List containing {@link Document} of the matching results. Returns an empty list if
               there are no matching results in the help index.
       @throws ParseException query string is invalid
    */
   public List<Document> query (String queryString, int maxResults) throws ParseException {
     ArrayList<Document> list = new ArrayList<Document>(); // list of results to return
     final int AUTO_WILDCARD_LEN = 5; // TODO make this configurable
     StringBuilder querySB = new StringBuilder(64);
     if (queryString != null && queryString.length() > 0) {
       // if only one search term, add a wildcard search for this term in the explain and midas fields
       // e.g. exp => exp || explain:exp* || midas:exp*
       if (queryString.indexOf(" ") == -1 && !queryString.endsWith("~") && !queryString.endsWith("^")) {
//       queryString = queryString + " || explain:" + queryString + "* || midas:" + queryString + "*"; // ORIG
         String term1 = (queryString.length() < AUTO_WILDCARD_LEN)? queryString : queryString + "*";
         querySB.append(term1).append(" || explain:").append(queryString).append("* || midas:").append(queryString).append('*');
       }
       else {
         String terms[] = queryString.split("\\s");
         for (String term : terms) {
           if (term.length() >= AUTO_WILDCARD_LEN && !term.endsWith("~") && !term.endsWith("^")) {
             querySB.append(term).append("* "); // auto wildcard terms AUTO_WILDCARD_LEN or longer
           } else {
             querySB.append(term).append(' ');
           }
         }
       }
       try {
         Query query = queryParser.parse(querySB.toString()); // parse the query string using the queryParser
         Hits hits   = searcher.search(query);         // run the query
         int length  = hits.length();                  // number of hits for the query
         if (maxResults>0) {
           length = Math.min(maxResults, length);      // limit hits to max results
         }
         list.ensureCapacity(length);                  // ensure capacity on results list so it does not have to resize

         for(int i=0; i<length; i++) {
           Document doc = hits.doc(i);                 // get the next document
           float score  = hits.score(i);               // get the documents search score
           doc.add(new Field("score", Float.toString(score), Field.Store.NO, Field.Index.UN_TOKENIZED));
           list.add(doc);
         } // end for loop
       } catch (IOException ioe) {
         displayMessage("Caught Exception while searching for '" + queryString + "' ", ioe);
       }
     }
     return list;
   }
  } // end class Searcher


  /** CustomerAnalyzer that is the same as the StandardAnalyzer, except it uses single English
      alphabetical letters as stop words (i.e. "a", "b", "c", ..., "z").
      This class is only used internally by {@link Indexer} and {@link Searcher}.
      @since NeXtMidas 2.5.0
    */
  private static class CustomAnalyzer extends StandardAnalyzer {
    /** A set containing single English letters that are usually not useful for searching. */
    protected static final Set<?> stopWords = StopFilter.makeStopSet(new String[] {
      "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m",
      "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z"
    });

    public CustomAnalyzer () {
      super(stopWords);
    }
  } // end class CustomAnalyzer

} // end class Help
