package nxm.sys.intr;

import nxm.sys.lib.*;
import nxm.sys.libg.Theme;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;

/** A tool for changing the look and feel for all Swing components used by the system.

    @author  J. Czechowski IV
    @version $Id: Lookandfeel.java,v 1.11 2013/01/03 17:41:21 ntn Exp $
    @since   NeXtMidas 2.1.0 (Moved from UCL)
 */
@SuppressWarnings({"rawtypes","unchecked"}) // <-- NOTE: JDK7 has some rawtypes & unchecked Xlint warnings that CANNOT be fixed since we still support JDK6
public final class Lookandfeel extends Intrinsic {
  /** The status of the primitive. */
  private int currentStatus = NOOP;

  public int open () {
    final String lookAndFeel = MA.getS("LAF", "").trim();

    if (MA.getState("/GUI")) {
      pollTime = args.getD("/POLL", 0.25); // default to a quarter second poll time, since nothing is done in process()
      Theme.initialize(); // 2011-03-23 changes to Theme class requires us to call this first to get the correct initial theme for the GUI
      (new PlafChooser()).setVisible(true);
    }
    else if (lookAndFeel.trim().length() == 0) {
      String[] themeNames = Theme.getThemeNames();
      M.type("Current Theme:     "+Theme.getGlobalTheme());
      M.type("Available Themes:");
      for (String themeName : themeNames) {
        String description = Theme.getTheme(themeName).getDescr();
        M.type("  " + StringUtil.padRight(themeName,10) + " - " + description);
      }
      return FINISH;
    }
    else {
      setLookAndFeel(lookAndFeel);
      return FINISH;
    }

    return NORMAL;
  }

  public int process () {
    return currentStatus;
  }

  /** Sets the look and feel.
      @param name The name of the look and feel.
  */
  private void setLookAndFeel (String name) {
    try {
      Theme.setGlobalTheme(name);
    }
    catch (IllegalArgumentException e) {
       throw e;
    }
    catch (UnsupportedLookAndFeelException e) {
      throw new MidasException("The '" + name + "' Look & Feel is not supported on this platform.");
    }
    catch (Exception e) {
      throw new MidasException("Unable to set Look & Feel to '" + name + "': " + e);
    }
  }

  /****************************************************************************/
  /* GUI SELECTOR - Most of this was auto-generated.                          */
  /****************************************************************************/

  /** The GUI used for choosing the Look & Feel. Plaf = Pluggable look-and-feel. */
  @SuppressWarnings("serial")
  private class PlafChooser extends JFrame implements ActionListener, WindowListener {

    /** Creates new instance. */
    public PlafChooser () {
      super("Choose a Look & Feel");
      getContentPane().setPreferredSize(new Dimension(330, 240));
      initComponents();

      this.lookAndFeels.setModel(new DefaultComboBoxModel(Theme.getThemeNames()));
      this.lookAndFeels.addActionListener(this);
      this.done.addActionListener(this);
      addWindowListener(this);
    }

    public void actionPerformed (ActionEvent evt) {
      if (evt.getSource() == lookAndFeels) {
        String lookAndFeel = lookAndFeels.getSelectedItem().toString();

        try {
          setLookAndFeel(lookAndFeel);
          SwingUtilities.updateComponentTreeUI(this);
          pack();

        } catch(Exception e) {
          final String msg = "Error setting Look & Feel: " + e;
          JOptionPane.showMessageDialog(this, msg, "Error", JOptionPane.ERROR_MESSAGE);
        }
      } else if (evt.getSource() == this.done) {
        setVisible(false);
        dispose();
        currentStatus = FINISH;
      }
    }

    /** This method is called from within the constructor to initialize the form. */
    private void initComponents() {
      jPanel0 = new javax.swing.JPanel();
      jPanel0.setBounds(new Rectangle(2, 2, 324, 40));
      jLabel1 = new javax.swing.JLabel();
      jLabel1.setBounds(0, 0, 110, 30);
      lookAndFeels = new javax.swing.JComboBox();
      lookAndFeels.setBounds(110, 0, 210, 30);
      done = new javax.swing.JButton();
      done.setBounds(250, 210, 67, 25);
      jPanel1 = new javax.swing.JPanel();
      jPanel1.setBounds(10, 80, 310, 120);
      jButton2 = new javax.swing.JButton();
      jButton2.setBounds(10, 20, 130, 25);
      jButton2.setToolTipText("a tool tip for a text button (JButton)");
      jCheckBox1 = new javax.swing.JCheckBox();
      jCheckBox1.setBounds(170, 20, 103, 23);
      jCheckBox1.setToolTipText("a tool tip for a check box (JCheckBox)");
      jRadioButton1 = new javax.swing.JRadioButton();
      jRadioButton1.setBounds(170, 50, 119, 23);
      jRadioButton1.setToolTipText("a tool tip for a radio button (JRadioButton)");
      jComboBox2 = new javax.swing.JComboBox();
      jComboBox2.setBounds(10, 50, 130, 24);
      jTextField1 = new javax.swing.JTextField();
      jTextField1.setBounds(10, 80, 130, 20);
      jTextField1.setToolTipText("a tool tip for a text field (JTextField)");

      setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
      setResizable(false);
      getContentPane().setLayout(null);
      done.setText("Done");
      getContentPane().add(done);
      jPanel0.setLayout(null);

      jLabel1.setText("Look And Feel:");
      jPanel0.add(jLabel1);

      jPanel0.add(lookAndFeels);

      getContentPane().add(jPanel0);

      jPanel1.setBorder(new javax.swing.border.TitledBorder("Sample"));
      jPanel1.setLayout(null);
      jButton2.setText("A Button");
      jPanel1.add(jButton2);

      jCheckBox1.setText("A Check Box");

      jPanel1.add(jCheckBox1);

      jRadioButton1.setText("A Radio Button");
      jPanel1.add(jRadioButton1);

      jComboBox2.setModel(new javax.swing.DefaultComboBoxModel(
                                               new String[] { "A Combo Box" }));
      jPanel1.add(jComboBox2);

      jTextField1.setText("A Text Field");
      jPanel1.add(jTextField1);

      getContentPane().add(jPanel1);

      pack();
    }

    public void windowOpened(WindowEvent e) {
      // ignore
    }

    public void windowClosing(WindowEvent e) {
      currentStatus = FINISH;
    }

    public void windowClosed(WindowEvent e) {
      // ignore
    }

    public void windowIconified(WindowEvent e) {
      // ignore
    }

    public void windowDeiconified(WindowEvent e) {
      // ignore
    }

    public void windowActivated(WindowEvent e) {
      // ignore
    }

    public void windowDeactivated(WindowEvent e) {
      // ignore
    }

    // Variables declaration - do not modify
    private javax.swing.JButton done;
    private javax.swing.JButton jButton2;
    private javax.swing.JCheckBox jCheckBox1;
    private javax.swing.JComboBox jComboBox2;
    private javax.swing.JLabel jLabel1;
    private javax.swing.JPanel jPanel0;
    private javax.swing.JPanel jPanel1;
    private javax.swing.JRadioButton jRadioButton1;
    private javax.swing.JTextField jTextField1;
    private javax.swing.JComboBox lookAndFeels;
    // End of variables declaration
  } // end private class PlafChooser
}
