package nxm.sys.intr;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

/**
  Displays the contents of a Midas file header.

  @author Jeff Schoen
  @version $Id: Status.java,v 1.29 2012/11/16 21:44:40 ntn Exp $
*/
public class Status extends Intrinsic {

  static final String positionalKeys = "FILE,SZ,XS,XD,XU,FS,YS,YD,YU,TYPE,FORMAT";

  public int open () {
    // Open the file
    BaseFile file = MA.getFile ("FILE");
    file.open();

    // Extract named fields into results parameters
    for (int n=2; n<=MA.numberOf(); n++) {
      String key = MA.getKey(n);
      String label = MA.getU(key);
      if (label.length()==0) continue;

      // if no key given, treat as X-Midas positional syntax
      if (key.equals("P"+n)) key = Parser.get(positionalKeys,n);

      Object value = null;
      if (file instanceof DataFile) {
        // Special cases for X-Midas backwards compatibility
        if      (key.equalsIgnoreCase("SZ")) key = "SIZE";
        else if (key.equalsIgnoreCase("DS")) key = "DATASTART";
        else if (key.equalsIgnoreCase("XU")) key = "XUNITS";
        else if (key.equalsIgnoreCase("YU")) key = "YUNITS";

        // Get record definitions for type 3000 or 5000 files
        if (key.startsWith("SR(") || key.startsWith("C(")) {
          value = getSubRecord((ListFile)file, key);
        }
      }
      if (value==null) value = KeyObject.getKey(file,key);

      if (value!=null) MR.put (label,value);
      else M.terminal.writeln ("Bad parameter syntax: "+key+"="+label);
    }

    // display the Header status to the user terminal
    if (MA.numberOf() == 1) M.terminal.write( file.listHeader() );

    // display the keyword status to the user terminal
    if (MA.getState("/K") || MA.getState("/KEY")) M.terminal.write( file.listKeywords() );

    // display the timeline status to the user terminal
    if (MA.getState("/T") || MA.getState("/TIME") && file instanceof DataFile) {
      DataFile df = (DataFile)file;
      if (df.getTimeLineHandler() == null) {
        M.warning("There are no TIMELINE keywords in "+df.getURL());
      } else {
        int decimalPlaces = MA.getL("/PRECISION",MA.getL("/PREC",3));
        M.terminal.write(df.getTimeLineHandler().list(decimalPlaces));
      }
    }
    // display the data status to the user terminal
    if (MA.find("/D") || MA.find("/DATA")) {
      String dataElsToList = MA.getS("/D", MA.getS("/DATA"));
      Shell.runCommand(M,"DATALIST,"+file.fn.toString()+",0,"+dataElsToList);
    }

    file.close ();

    return FINISH;
  }

  /** Helper wrapper to support the old X-Midas syntax of C(n) or SR(n) for record-based
      3000 or 5000 files.  Number parsed from the key is a one-based number to best match
      what is already implemented in HEADERMOD.
   */
  private nxm.sys.lib.Table getSubRecord(ListFile listFile, String key) {
    // The KEY can be:
    //  1) a number like SR(1) or
    //  2) a result like like SR(myres) or
    //  3) a record name like SR(NAME) or SR("NAME")
    nxm.sys.lib.Table retVal = null;
    int iLParen = key.indexOf("(");
    int iRParen = key.indexOf(")");
    String numOrString = key.substring(iLParen+1,iRParen);
    double dIndex = -1;

    // Get the string as a number, pure string or result containing either a number or a string
    Object recObject = Convert.o2o(numOrString, '_', this);
    if (recObject != null) {
      if (StringUtil.isNumber(""+recObject)) dIndex = Convert.o2d(recObject);
      else numOrString = ""+recObject;
    }

    // Now we know it is a number or a String. Form table like...
    //    {NAME=POS,FORMAT=VD,TYPE=GEODETIC,UNITS=DISTANCE}
    if (dIndex != -1) {  // ... a record number
      retVal = listFile.getRecordDef((int)(dIndex-1)); // Need to use an offset 0..n-1
    } else {
      try {              // ... a record name
        nxm.sys.lib.Table allRecDefs = listFile.getRecordDefs();
        if (allRecDefs.containsKey(numOrString)) { retVal = allRecDefs.getTable(numOrString); }
        else                                     { M.warning("STATUS: Unable to find key '"+numOrString+"' in file"); }
      } catch (Exception e) {
        M.warning("Unable to find subrecord in string "+key);
      }
    }

    return retVal;
  }

}
