package nxm.sys.lib;

import java.applet.Applet;
import java.awt.Color;
import java.io.File;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import nxm.sys.libg.MApplet;
import nxm.sys.libg.MColor;

import static nxm.sys.lib.Util.getArchList;
import static nxm.sys.lib.Util.getOSList;

/** Manages the creation of an applet-launcher based application. <br>
 *  <br>
 *  Typical usage of this class:
 *  <pre>
 *    AppletLauncher a = new AppletLauncher("APPLET",
 *                                          <font color="blue">"My Applet"</font>,
 *                                          <font color="blue">"My Company"</font>,
 *                                          <font color="blue">"http://www.test.com/"</font>,
 *                                          <font color="blue">"http://www.test.com/code/"</font>,
 *                                          "1.6+",
 *                                          "512m",
 *                                          "512m");
 *    a.setOptionTrees(<font color="blue">"MYOPT", ...</font>, "SYS");
 *    a.setNmStartup("nxm.<font color="blue">myopt</font>.cfg.nmstartup.mm");
 *    a.setCommand(<font color="blue">"MYCOMMAND/SWITCH ARG1 ARG2"</font>);
 *    a.writeFiles(<font color="blue">"/some/dir/"</font>,
 *                 <font color="blue">"my_applet.html"</font>);
 *  </pre>
 *  @author  J. Czechowski IV
 *  @version $Id: AppletLauncherGen.java,v 1.3 2012/09/28 14:01:16 ntn Exp $
 */
public final class AppletLauncherGen {

  private static final String ALT_TEXT = "Your browser understands the &lt;APPLET&gt; "
                                       + "tag but is not running the Java applet (is "
                                       + "Java enabled in your preferences?)";
  /** The standard requires a single-line archive= entry (ONE_LINE_ARCHIVE=true),
   *  but nearly all example use a multi-line form (ONE_LINE_ARCHIVE=false).
   */
  private static final boolean ONE_LINE_ARCHIVE = true;

  private       String                    appletID    = "APPLET";
  private       String                    appletName  = "NeXtMidas Applet";
  private       boolean                   appletDebug = false;
  private       String                    pageTitle   = null;
  private       Class<? extends Applet>   mainClass   = MApplet.class;
  private       String                    nmStartup   = null;
  private       String                    command     = null;
  private       String                    image       = null;
  private       int                       width       = 640;
  private       int                       height      = 480;
  private       Color                     boxBgColor  = MColor.getColor("#333333");
  private       Color                     boxFgColor  = MColor.getColor("#FFFFFF");
  private final List<String>              optionTrees = new ArrayList<String>();
  private       String                    writeAux    = null;
  private final List<String>              readAux     = new ArrayList<String>();
  private final Map<String,String>        params      = new LinkedHashMap<String,String>();
  private final JnlpFile                  jnlp        = new JnlpFile();
  private final JnlpFile.AbstractJnlpJ2SE java;       // Set in constructor

  /** Creates a new instance with the three required fields.
   *  @param appletID     The applet ID   (see {@link #setAppletID}).
   *  @param appletName   The applet name (see {@link #setAppletName}).
   *  @param vendor       The vendor      (see {@link JnlpFile.JnlpInformation#setVendor}).
   *  @param homepage     The homepage    (see {@link JnlpFile.JnlpInformation#setHomepage}).
   *  @param codebase     The codebase    (see {@link #setCodebase}).
   *  @param javaVersion  The Java version required (see {@link JnlpFile.JnlpJ2SE#setVersion}).
   *  @param javaInitHeap The initial Java heap required (see {@link JnlpFile.JnlpJ2SE#setInitialHeapSize(String)}).
   *  @param javaMaxHeap  The maximum Java heap required (see {@link JnlpFile.JnlpJ2SE#setMaxHeapSize(String)}).
   */
  public AppletLauncherGen (String appletID, String appletName, String vendor,
                         String homepage, String codebase,
                         String javaVersion, String javaInitHeap, String javaMaxHeap) {
    this.java = new JnlpFile.JnlpJ2SE(javaVersion, javaInitHeap, javaMaxHeap);
    setAppletID(appletID);
    setAppletName(appletName);
    setCodebase(codebase);

    jnlp.setHref("##TBD##");
    jnlp.getInfo().setTitle("##TBD##");
    jnlp.getInfo().setVendor(vendor);
    jnlp.getInfo().setHomepage(homepage);
    jnlp.setAllPermissions(true);
  }

  /** The web page title. */
  public String getPageTitle () { return (pageTitle == null)? getAppletName() : pageTitle; }
  /** The web page title. */
  public void setPageTitle (String title) { this.pageTitle = title; }

  /** The applet name. */
  public String getAppletName () { return (appletName == null)? getAppletID() : appletName; }
  /** The applet name. */
  public void setAppletName (String name) { this.appletName = name; }

  /** The applet ID. */
  public String getAppletID () { return appletID; }
  /** The applet ID. */
  public void setAppletID (String id) {
    if (id == null) throw new NullPointerException("Applet ID can not be null.");
    this.appletID = id;
  }

  /** Turn NeXtMidas applet debugging? */
  public boolean getAppletDebug () { return appletDebug; }
  /** Turn NeXtMidas applet debugging? */
  public void setAppletDebug (boolean debug) { this.appletDebug = debug; }

  /** The main class for the applet (typically {@link MApplet}. */
  public Class<? extends Applet> getMainClass () { return mainClass; }
  /** The main class for the applet (typically {@link MApplet}. */
  public void setMainClass (Class<? extends Applet> mainClass) {
    if (mainClass == null) throw new NullPointerException("Main class can not be null.");
    this.mainClass = mainClass;
  }

  /** The codebase to use. */
  public String getCodebase () { return jnlp.getCodebase(); }
  /** The codebase to use. */
  public void setCodebase (String codebase) { jnlp.setCodebase(codebase); }

  /** The applet width in pixels. */
  public int getWidth () { return width; }
  /** The applet width in pixels. */
  public void setWidth (int width) { this.width = width; }

  /** The applet height in pixels. */
  public int getHeight () { return height; }
  /** The applet height in pixels. */
  public void setHeight (int height) { this.height = height; }

  /** The background color for the box containing the applet. */
  public Color getBoxBgColor () { return boxBgColor; }
  /** The background color for the box containing the applet. */
  public void setBoxBgColor (Color color) {
    if (color == null) throw new NullPointerException("Color can not be null.");
    this.boxBgColor = color;
  }

  /** The foreground color for the box containing the applet. */
  public Color getBoxFgColor () { return boxFgColor; }
  /** The foreground color for the box containing the applet. */
  public void setBoxFgColor (Color color) {
    if (color == null) throw new NullPointerException("Color can not be null.");
    this.boxFgColor = color;
  }

  /** The URL of the image to use while the applet is loading. */
  public String getImage () { return image; }
  /** The URL of the image to use while the applet is loading. */
  public void setImage (String image) { this.image = image; }

  /** The list of option trees used. */
  public List<String> getOptionTrees () { return optionTrees; }

  /** The list of option trees used in the order they will appear on the path. */
  public void setOptionTrees (List<? extends String> opts) {
    optionTrees.clear();
    optionTrees.addAll(opts);
  }
  /** The list of option trees used in the order they will appear on the path. */
  public final void setOptionTrees (String ... opts) {
    setOptionTrees(Arrays.asList(opts));
  }
  /** The list of option trees used in the order they will appear on the path. */
  private String _getOpts () {
    StringBuilder str = new StringBuilder();
    boolean first = true;
    for (String opt : getOptionTrees()) {
      if (!first) str.append(',');
      str.append(opt.toUpperCase());
      first = false;
    }
    return str.toString();
  }

  /** The NeXtMidas command to run. */
  public String getCommand () { return command; }
  /** The NeXtMidas command to run. Setting this to non-null sets the main class
   *  to {@link MApplet}.
   */
  public void setCommand (String cmd) {
    this.mainClass = MApplet.class;
    this.command   = cmd;
  }

  /** The NeXtMidas READ AUX. */
  public List<String> getReadAux () {
    if ((readAux != null) || (writeAux == null)) {
      return Collections.unmodifiableList(readAux);
    }
    return Arrays.asList(writeAux);
  }

  /** The NeXtMidas READ AUX. */
  public final void setReadAux (String ... readAux) {
    setReadAux(Arrays.asList(readAux));
  }

  /** The NeXtMidas READ AUX. */
  public void setReadAux (List<String> readAux) {
    this.readAux.clear();
    this.readAux.addAll(readAux);
  }

  /** The NeXtMidas WRITE AUX. */
  public String getWriteAux () {
    if ((writeAux != null) || readAux.isEmpty()) {
      return writeAux;
    }
    return readAux.get(0);
  }

  /** The NeXtMidas WRITE AUX. */
  public void setWriteAux (String writeAux) {
    this.writeAux = writeAux;
  }

  /** The NeXtMidas WRITE + READ AUX. */
  public final void setAux (String writeAux, String ... readAux) {
    setWriteAux(writeAux);
    setReadAux(readAux);
  }

  /** The NeXtMidas WRITE + READ AUX. */
  private String getAux () {
    String        write = getWriteAux();
    List<String>  read  = getReadAux();
    boolean       first = true;
    StringBuilder str   = new StringBuilder();

    if ((write == null) && read.isEmpty()) {
      return null;
    }
    str.append(write);
    for (String aux : read) {
      if (first) str.append(' ').append(aux);
      else       str.append('|').append(aux);
      first = false;
    }
    return str.toString();
  }

  /** The NeXtMidas startup script to run. */
  public String getNmStartup () { return nmStartup; }
  /** The NeXtMidas startup script to run. */
  public void setNmStartup (String mcr) { this.nmStartup = mcr; }

  /** The list of JARs used as they must be listed for the archive= parameter. */
  private CharSequence getArchive () {
    StringBuilder str = new StringBuilder();
    str.append("applet-launcher.jar");  // <-- MUST COME FIRST!

    for (String jar : getOptionTreeJars(true)) {
      if (jar.trim().equals("applet-launcher.jar")) continue;
      if (ONE_LINE_ARCHIVE) {
        str.append(",").append(jar.trim());
      }
      else {
        str.append(",\n                     ").append(jar);
      }
    }
    return str;
  }

  /** Gets a listing of the jar files used for all of the current option trees. */
  public List<String> getOptionTreeJars () {
    return getOptionTreeJars(false);
  }

  /** Gets a listing of the jar files used for all of the current option trees. */
  private List<String> getOptionTreeJars (boolean indent) {
    List<String> jars = new LinkedList<String>();

    for (String opt : getOptionTrees()) {
      jars.add("nxm-"+opt.toLowerCase()+".jar");

      if (indent) {
        for (File f : OptionTree.getJarFiles(opt)) {
          jars.add("  "+f.getName());
        }
      }
      else {
        for (File f : OptionTree.getJarFiles(opt)) {
          jars.add(f.getName());
        }
      }

    }
    return jars;
  }

  /** The template JNLP file to use. */
  public JnlpFile getJnlpTemplate () { return jnlp; }

  /** The template Java/J2SE parameters to use. */
  public JnlpFile.AbstractJnlpJ2SE getJava () { return java; }

  /**The user-specified applet parameters. */
  public Map<String,String> getParams () { return params; }

  /** Creates a *new* map with the internally-generated applet parameters. */
  public Map<String,String> getInternalParams () {
    Map<String,String> map = new LinkedHashMap<String,String>();

    // STANDARD APPLET PARAMS
    map.put("codebase_lookup",       "false");
    map.put("boxbgcolor",            MColor.toString(getBoxBgColor()));
    map.put("boxfgcolor",            MColor.toString(getBoxFgColor()));
    if (getImage() != null) {
      map.put("image",               getImage());
    }
    map.put("subapplet.classname",   getMainClass().getName());
    map.put("subapplet.displayname", getAppletName());
    map.put("noddraw.check",         "true");
    map.put("progressbar",           "true");

    // JNLP EXTENSION PARAMS
    map.put("jnlpNumExtensions", Integer.toString(getOptionTrees().size()));

    int i = 0;
    for (String opt : getOptionTrees()) {
      String key   = "jnlpExtension" + (++i);
      String value = getCodebase() + "nxm-" + opt.toLowerCase() + "-native.jnlp";
      map.put(key, value);
    }

    // NEXTMIDAS APPLET PARAMS
    map.put("DEBUG", Boolean.toString(getAppletDebug()));
    map.put("WIN",                    getAppletID());
    map.put("OPTS",                   _getOpts());
    map.put("PATH",                   _getOpts());
    if (getNmStartup() != null) {
      map.put("NMSTARTUP.USER",       getNmStartup());
    }
    if (getCommand() != null) {
      map.put("CMD",                  getCommand());
    }
    if (getAux() != null) {
      map.put("AUX",                  getAux());
    }
    return map;
  }

  /** Writes the files out to the give directory. The JNLP files will be written
   *  to the specified directory using file names "nxm-opt-native.jnlp" where "opt" is
   *  the name of the option tree.
   *  @param ref         The Midas reference.
   *  @param dir         The directory to write the files to. Typically this
   *                     will be equal to "$NMROOT/jars/".
   *  @param webPageName The name of the web page to write.
   *  @return The list of files written.
   */
  public List<String> writeFiles (Object ref, String dir, String webPageName) {
    List<String> fileNames = new LinkedList<String>();
    Midas        midas     = Convert.ref2Midas(ref);

    if (midas == null) midas = Shell.getMidasContext();
    dir = FileUtil.terminatePath(dir);

    // WRITE THE HTML FILE
    String   htmlName = dir+webPageName;
    TextFile htmlFile = new TextFile(ref, htmlName);
    htmlFile.open(BaseFile.OUTPUT);
    htmlFile.write(toHTML().toString());
    htmlFile.close();
    fileNames.add(htmlName);


    // WRITE THE COMPONENT JNLP FILE FOR EACH OPTION TREE
    for (String opt : getOptionTrees()) {
      // OPEN THE OUTPUT FILE --------------------------------------------------
      String   jnlpName = dir+"nxm-"+opt.toLowerCase()+"-native.jnlp";
      JnlpFile jnlpFile = new JnlpFile(ref, jnlpName);
      jnlpFile.open(BaseFile.OUTPUT);

      jnlpFile.fromTable(getJnlpTemplate().toTable());
      jnlpFile.setHref(null);
      jnlpFile.getInfo().setTitle("NeXtMidas "+opt.toUpperCase()+" Option Tree");

      // ADD THE JAVA RESOURCES ------------------------------------------------
//      JnlpFile.JnlpResources r = new JnlpFile.JnlpResources();
//      r.setOS(null);
//      r.setArch(null);
//      r.setLocale(null);
//      r.getJava().add(getJava());
//
//      r.getJars().add(new JnlpFile.JnlpJar("nxm-"+opt.toLowerCase()+".jar"));
//      for (File f : OptionTree.getJarFiles(opt)) {
//        r.getJars().add(new JnlpFile.JnlpJar(f.getName()));
//      }
//      jnlpFile.getResources().add(r);

      // ADD THE NATIVE RESOURCES ----------------------------------------------
      for (Map.Entry<String,List<String>> e : findNativeJars(midas, opt).entrySet()) {
        String       osArch   = e.getKey();
        int          i        = osArch.lastIndexOf('-');
        List<String> osList   = getOSList(osArch.substring(0,i));
        List<String> archList = getArchList(osArch.substring(i+1));
        List<String> jars     = e.getValue();

        for (String os : osList) {
          for (String arch : archList) {
            JnlpFile.JnlpResources r = new JnlpFile.JnlpResources();
            r.setOS(os);
            r.setArch(arch);
            r.setLocale(null);

            for (String jar : jars) {
              r.getJars().add(new JnlpFile.JnlpNativeLib(jar));
            }
            jnlpFile.getResources().add(r);
          }
        }
      }

      // WRITE THE FILE AND CLOSE ----------------------------------------------
      jnlpFile.close();
      fileNames.add(jnlpName);
    }
    return fileNames;
  }

  /** Find all the native JARs for a given option tree. */
  private Map<String,List<String>> findNativeJars (Midas midas, String opt) {
    final String prefix  = "nxm-" + opt.toLowerCase() + "-native-";
    final String suffix  = ".jar";

    // FIND THE JARS
    String         nmroot = FileUtil.terminatePath(Shell.getNmRoot());
    String         dir    = FileUtil.terminatePath(nmroot + "jars");
    FilenameFilter filter = new FilenameFilter() {
      @Override
      public boolean accept (File dir, String name) {
        return name.startsWith(prefix) && name.endsWith(suffix);
      }
    };

    // PARSE THE NAMES
    Map<String,List<String>> jars = new LinkedHashMap<String,List<String>>();
    String[] files = midas.io.listFiles(dir, filter);

    if (files != null) {
      for (String fname : files) {
        String osArch = fname.substring(prefix.length(), fname.length()-suffix.length());

        if (!jars.containsKey(osArch)) {
          jars.put(osArch, new LinkedList<String>());
        }
        jars.get(osArch).add(fname);
      }
    }
    return jars;
  }


  /** Creates the HTML code for a stand-alone applet-containing web page. */
  public CharSequence toHTML () {
    StringBuilder html = new StringBuilder();
    html.append("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">\n");
    html.append("<html>\n");
    html.append("  <head>\n");
    html.append("    <meta http-equiv='Content-Type' content='text/html'>\n");
    html.append("    <title>").append(getPageTitle()).append("</title>\n");
    html.append("  </head>\n");
    html.append("  <body style='margin: 0px;'>\n");

    toAppletHTML(html);

    html.append("  </body>\n");
    html.append("</html>\n");
    return html;
  }

  /** Creates the HTML code for the APPLET section of the web page. */
  public CharSequence toAppletHTML () {
    return toAppletHTML(new StringBuilder());
  }

  /** Creates the HTML code for the APPLET section of the web page. */
  private StringBuilder toAppletHTML (StringBuilder html) {
    html.append("    <!-- Notes:                                                              -->\n");
    html.append("    <!--   1. When listing the archives used, applet-launcher.jar must be    -->\n");
    html.append("    <!--      the FIRST entry in the list.                                   -->\n");
    html.append("    <!--                                                                     -->\n");
    html.append("    <!--   2. The way applet-launcher.jar is signed and the  presence (or    -->\n");
    html.append("    <!--      absence) of the 'Trusted-Library: true' entry in the JAR       -->\n");
    html.append("    <!--      manifest is extremely important.                               -->\n");
    html.append("    <!--                                                                     -->\n");
    html.append("    <!--        [If you get this step wrong, you will be pulling your hair   -->\n");
    html.append("    <!--         out for hours trying to figure out what is wrong. The       -->\n");
    html.append("    <!--         errors the Java plug-in provides in this situation are      -->\n");
    html.append("    <!--         cryptic (at best) and you can even get into a situation     -->\n");
    html.append("    <!--         with a blank applet box and no indication as to what the    -->\n");
    html.append("    <!--         cause is.]                                                  -->\n");
    html.append("    <!--                                                                     -->\n");
    html.append("    <!--   3. In theory it is possible to do something like:                 -->\n");
    html.append("    <!--        <applet height='600' width='400'>                            -->\n");
    html.append("    <!--           <param name='jnlp_href' value='myfile.jnlp'>              -->\n");
    html.append("    <!--        </applet>                                                    -->\n");
    html.append("    <!--      But even among modern browsers with the Oracle Java 7 and/or   -->\n");
    html.append("    <!--      OpenJDK 6 (ICEDTEA) plug-in, support for that appears          -->\n");
    html.append("    <!--      extremely limited on Linux, hence why that is not used here.   -->\n");
    html.append("    <!--                                                                     -->\n");
    html.append("    <applet id='").append(getAppletID()).append("'\n");
    html.append("            code='org.jdesktop.applet.util.JNLPAppletLauncher'\n");
    html.append("            codebase='").append(getCodebase()  ).append("'\n");
    html.append("            archive='" ).append(getArchive()   ).append("'\n");
    html.append("            width='"   ).append(getWidth()     ).append("'\n");
    html.append("            height='"  ).append(getHeight()    ).append("'\n");
    html.append("            name='"    ).append(getAppletName()).append("'\n");
    html.append("            alt='"     ).append(ALT_TEXT       ).append("'\n");
    html.append("            mayscript>\n");
    for (Map.Entry<String,String> e : getInternalParams().entrySet()) {
      html.append("      <param name='").append(e.getKey())
                    .append("' value='").append(e.getValue()).append("'>\n");
    }
    for (Map.Entry<String,String> e : getParams().entrySet()) {
      html.append("      <param name='").append(e.getKey())
                    .append("' value='").append(e.getValue()).append("'>\n");
    }
    html.append("      Your browser is completely ignoring the &lt;APPLET&gt; tag or\n");
    html.append("      you don't have the Java Plug-in installed/setup for your browser.\n");
    html.append("    </applet>\n");
    return html;
  }
}
