package nxm.sys.lib;

import nxm.sys.inc.*;

/**
  A class to define a Midas Data pipe.

  @author Jeff Schoen
  @version $Id: DataPipe.java,v 1.8 2004/07/15 15:46:52 jxc Exp $

*/
public class DataPipe extends DataFile {

  public final String propertyList = super.propertyList +
	",Inlet,Outlets,InByte,OutByte,PipeSize,Avail,Rate";

  public int pipeInlet=0, pipeOutlets=0, pipeOutMask=0, pipeSize=131072;
  public double pipeInPtr=0, pipeOutPtr[]=new double[9];
  public byte[] pipeBuffer;
  public DataPipe next=null;

  public DataPipe (Midas M, String fullname) {
    this.M = M;
    this.filename = fullname;
  }

  /** attach to pipe */
  public synchronized int attach (int dir) {
    int outlet=0;
    if (dir>0) {
      isOpen = true;
      pipeInlet = 1;
    } else {
      for (int i=0; i<8 && outlet==0; i++) {
        if ( (pipeOutMask&(1<<i)) == 0) {
	  outlet = i+1;
	  pipeOutMask |= (1<<i);
	  pipeOutlets = Math.max(pipeOutlets,outlet);
	  pipeOutPtr[outlet] = pipeInPtr;
	}
      }
    }
    return outlet;
  }

  /** detach from pipe */
  public synchronized int detach (int outlet) {
    if (outlet==0) {
      isOpen = false;
      pipeInlet = 0;
    } else {
      int i=outlet-1;
      pipeOutMask &= ~(1<<i);
      for (i=pipeOutlets-1; i>=0 && (pipeOutMask&(1<<i))==0; i--); 
      pipeOutlets = i+1;
      outlet = -2;
    }
    return outlet;
  }

  /** update the combined read pointers */
  public double lastPipeOut () {
    double pipeOut = pipeInPtr;
    for (int i=0; i<pipeOutlets; i++) {
      if ((pipeOutMask&(1<<i)) != 0 && pipeOutPtr[i+1]<pipeOut) 
	pipeOut=pipeOutPtr[i+1];
    }
    pipeOutPtr[0] = pipeOut;
    return pipeOut;
  }

  public double avail (int mode, int outlet) {
    double status=0.0;
    if (mode>0) status = pipeSize - (pipeInPtr-lastPipeOut());
    else if (outlet<0 || outlet>8);
    else status = pipeInPtr-pipeOutPtr[outlet];
    return status;
  } 
      
  /** write to pipe section */
  public int write (byte[] buf, int boff, int bytes, int outlet) {
    if (M.pipeMode==Midas.PPAUSE) return 0;
    double pipeOut = lastPipeOut();
    int poff = (int)(pipeInPtr % (double)pipeSize);
    int bmax = pipeSize-poff;
    if (bmax<bytes) bytes=bmax;
    int pmax = pipeSize - (int)(pipeInPtr-pipeOut);
    if (pmax<bytes) bytes=pmax;
    if (bytes>0) {
      System.arraycopy (buf,boff,pipeBuffer,poff,bytes);
      pipeInPtr += bytes; 
    } else bytes=0;
    if (M.pipeMode==Midas.PSTOP || M.pipeMode==Midas.POFF) return -1;
    return bytes;
  } 

  /** read from pipe section */
  public int read (byte[] buf, int boff, int bytes, int outlet) {
    if (M.pipeMode==Midas.PPAUSE) return 0;
    double pipeOut = pipeOutPtr[outlet];
    int poff = (int)(pipeOut % (double)pipeSize);
    int bmax = pipeSize-poff;
    if (bmax<bytes) bytes=bmax;
    int pmax = (int)(pipeInPtr-pipeOut);
    if (pmax<bytes) bytes=pmax;
    if (bytes<=0 && !isOpen) {
      if (next!=null) return (CoreIO.REOPEN); 
      else return (CoreIO.CLOSED);
    }
    if (bytes>0) {
      if (bytes>pipeSize) bytes=pipeSize;
      System.arraycopy (pipeBuffer,poff,buf,boff,bytes);
      pipeOut += bytes; 
    } else bytes=0;
    pipeOutPtr[outlet] = pipeOut;
    if (M.pipeMode==Midas.PSTOP || M.pipeMode==Midas.POFF) return -1;
    return bytes;
  } 

  /** {@inheritDoc} */
  public short getInlet() { return (short)pipeInlet; }
  
  /** {@inheritDoc} */
  public short getOutlets() { return (short)pipeOutlets; }
  
  /** {@inheritDoc} */
  public int getOutMask() { return pipeOutMask; }
  
  /** {@inheritDoc} */
  public int getPipeSize() { return pipeSize; }
  
  /** {@inheritDoc} */
  public double getInByte() { return pipeInPtr; }
  
  public double getOutByte() { return pipeOutPtr[0]; }
  
  /** {@inheritDoc} */
  public double getOutByte(int i) { return pipeOutPtr[i]; }
  
  public double getAvail() { return (pipeInPtr-pipeOutPtr[0]); }
  
  private double timeLast=0,pipeInPtrLast=0;
  
  public double getRate() { 
    double time = Time.current();
    double avail = (pipeInPtr-pipeOutPtr[0]);
    double xfer = (pipeInPtr-pipeInPtrLast);
    double rate = xfer/(time-timeLast);
    timeLast = time;
    pipeInPtrLast = pipeInPtr;
    return rate;
  }
  

  /** Allow dynamic setting of the pipe size for tuning.
   
      @param value {@inheritDoc}
   */
  public void setPipeSize(int value) { 
    if (pipeBuffer==null || value!=pipeSize) {
      int mode = M.pipeMode;
      
      if (mode==Midas.PRUN) {
        M.pipeMode=Midas.PPAUSE;
        Time.sleep(0.2);
      }
      
      pipeBuffer = new byte[value];
      pipeSize = value;
      
      if (M.pipeMode!=mode) {
        M.pipeMode=mode;
      }
    }
  }
}
