package nxm.sys.lib;

import java.io.*;
import java.net.*;
import java.util.Map;

import nxm.sys.inc.AsciiMap;

/**
  Utility class for the manipulation of filenames. Methods are provided for
  combining path components into filenames, translating between different OS
  path separators, etc.

  Filenames within Midas can contain far more information than simply the
  path to the file. Midas filenames also contain "qualifiers" and "trimmers"
  which are indicated by curly braces and parentheses respectively.

  Some examples of strings that are suitable file names:


  Filename that doesn't contain any qualifiers or trimmers.
  <blockquote> nxm/sys/dat/world.prm </blockquote>

  Filename that contains qualifiers.
  <blockquote> nxm/sys/dat/world.prm{PS=1K} </blockquote>

  Filename that contains trimmers.
  <blockquote> nxm/sys/dat/world.prm(0:1M) </blockquote>

  Filename that contains both trimmers and qualifiers.
  <blockquote> nxm/sys/dat/world.prm(0:1M){PS=1K} </blockquote>

  @version $Id: FileName.java,v 1.79 2013/01/03 19:30:43 ntn Exp $
*/
public class FileName implements AsciiMap,Cloneable {

  /** OS type MNEMONICS used as the path syntax indicator */
  public static final int OS_UNKNOWN=-1,OS_LOCAL=1,OS_UNIX=2,OS_DOS=3,OS_VMS=4;

  /** The local path separator (DOS, UNIX or VMS) */
  public static final String LOCAL_PATH_DELIM = File.separator;
  public static final String DOS_PATH_DELIM = "\\";
  public static final String UNIX_PATH_DELIM = "/";
  public static final String VMS_PATH_DELIM = ".";

  private String path,root,ext,tag,qualifiers,trimmers;
  private String fullname;
  private Table qualTable=null;

  /** Initialize a new instance according to the filename given.  The filename
      can be in any of the supported OS formats (DOS, UNIX, or VMS). <br><br>
      This is the same as <code>FileName(filename,false)</code>.
      @param filename The file name.
   */
  public FileName (String filename) {
    this (filename,false);
  }

  /** Create a new filename that is relative to the existing context. <br><br>
      This is the same as <code>FileName(context,fname,false)</code>.
      @param context The context for the file name. The file name is referenced
                     relative to this.
      @param fname   The file name.
   */
  public FileName (FileName context, String fname) {
    this (context,fname,false);
  }

  /** Constructor for a file name. Takes a path, root and extension. <br><br>
      This is the same as <code>FileName(inpath,inroot,inext,false)</code>.
      @param inpath The directory components of the path.
      @param inroot The root name of the file within its directory.
      @param inext The extension of the file.
  */
  public FileName (String inpath, String inroot, String inext) {
    this (inpath,inroot,inext,false);
  }

  /** Constructor for a file name.  Takes an array of path name strings,
      a single root and extension. <br><br>
      This is the same as <code>FileName(pathList,inroot,inext,false)</code>.
      @param pathList The array of strings making up the path list. To make the
                      path each of these will be joined by inserting the
                      appropriate platform-dependent file separator.
      @param inroot   The root name of the file within its directory.
      @param inext    The extension of the file.
   */
  public FileName (String pathList[], String inroot, String inext) {
    this (pathList,inroot,inext,false);
  }

  /** Initialize a new instance according to the given {@link File}. This is the same as
      <code>Filename(file.toString(),true)</code>
      @param file The file.
      @since NeXtMidas 2.5.0
   */
  public FileName (File file) {
    this(file.toString(),true);
  }

  /** Initialize a new instance according to the filename given.  The filename
      can be in any of the supported OS formats (DOS, UNIX, or VMS).
      @param filename The file name.
      @param keepCase Indicates the original case of the file name should be
                      kept. If this is false any part of a file name that is
                      in all upper case will be converted to lower case.
      @since NeXtMidas 1.9.2
   */
  public FileName (String filename, boolean keepCase) {
    parseName(filename,keepCase);
  }

  /** Create a new filename that is relative to the existing context.
      @param context  The context for the file name. The file name is referenced
                      relative to this.
      @param fname    The file name.
      @param keepCase Indicates the original case of the file name should be
                      kept. If this is false any part of a file name that is
                      in all upper case will be converted to lower case.
      @since NeXtMidas 1.9.2
   */
  public FileName (FileName context, String fname, boolean keepCase) {
    try {
      parseName(context.relativeTo(fname).getFullName(), keepCase);
    }
    catch (MalformedURLException e) {
      Shell.printStackTrace(e);
    }
  }

  /** Constructor for a file name. Takes a path, root and extension. <br><br>
      This is the same as <code>FileName(context,fname,false)</code>.
      @param inpath   The directory components of the path.
      @param inroot   The root name of the file within its directory.
      @param inext    The extension of the file.
      @param keepCase Indicates the original case of the file name should be
                      kept. If this is false any part of a file name that is
                      in all upper case will be converted to lower case.
      @since NeXtMidas 1.9.2
  */
  public FileName (String inpath, String inroot, String inext, boolean keepCase) {
    parseName( addPathRootExt(inpath,inroot,inext), keepCase );
  }

  /** Constructor for a file name.  Takes an array of path name strings,
      a single root and extension.
      @param pathList The array of strings making up the path list. To make the
                      path each of these will be joined by inserting the
                      appropriate platform-dependent file separator (cannot be null or empty)
      @param inroot   The root name of the file within its directory.
      @param inext    The extension of the file.
      @param keepCase Indicates the original case of the file name should be
                      kept. If this is false any part of a file name that is
                      in all upper case will be converted to lower case.
      @since NeXtMidas 1.9.2
   */
  public FileName (String pathList[], String inroot, String inext, boolean keepCase) {
    String inpath = pathList[0];
    for (int i=1; i<pathList.length; i++)
      inpath = addPathDir(inpath,pathList[i]);
    parseName( addPathRootExt(inpath,inroot,inext), keepCase );
  }

  /** Parses the file name into the various class members. <br> <br>
      This is the same as <code>parseName(filename,false)</code>.
      @param filename Full file name with optional delimiters, qualifiers and trimmers.
  */
  public void parseName (String filename) {
    parseName(filename,false);
  }

  /** Parses the file name into the various class members.
      @param filename Full file name with optional delimiters, qualifiers and trimmers.
      @param keepCase Indicates the original case of the file name should be
                      kept. If this is false any part of a file name that is
                      in all upper case will be converted to lower case.
      @since NeXtMidas 1.9.2
  */
  public void parseName (String filename, boolean keepCase) {
    int i,j,k,ls;
    if (filename==null || filename.length()==0||filename.equals("NULL")) {
      path=root=ext=tag="";
      return;
    }
    boolean quoted = (filename.charAt(0)==QUOTE);
    if (quoted) {   // strip quotes
      i = filename.lastIndexOf(QUOTE);
      if (i<=0) i=filename.length();
      filename = filename.substring(1,i);
    }
    i = getQualifiersIndex(filename); // file qualifiers
    if (i>=0) { qualifiers = filename.substring(i); filename = filename.substring(0,i); }
    i = getTrimmersIndex(filename); // file trimmers
    if (i>=0) { trimmers = filename.substring(i); filename = filename.substring(0,i); }
    i = filename.lastIndexOf(SCOLON); // VMS version number
    if (i>=0) ls = i; else ls = filename.length();
    k = filename.indexOf("://");
    i = filename.lastIndexOf(SLASH);
    if (k>0 && i<k+3) i=-1;  // not the slash in the URL separator
    if (i<0) i = filename.lastIndexOf(BSLASH);
    if (i<0) i = filename.lastIndexOf(RBRACE);
    if (i<0) i = filename.lastIndexOf(COLON);
    if (k>0 && i<k+3) i=-1; // ignore chars in URL tag
    if (i<0 && k>0) i=k+2; // just URL tag
    if (i>=0) path = filename.substring(0,i+1); else path = "";
    // check for nxm.opt.area or nxm.aux.name (NXM dot) syntax
    if (i<0 && (filename.startsWith("nxm.")||filename.startsWith("NXM."))) {
      int ii = filename.indexOf('.',4);
      int jj = filename.indexOf('.',ii+1);
      if (ii>0 && jj>0)  { i=jj;
        String opt = filename.substring(4,ii);
        String area = filename.substring(ii+1,jj);
        if (opt.toLowerCase().equals("aux"))
          path = Shell.getAuxPath(area);
        else
          path = Shell.getOptPath(opt,area);
      }
    }
    j = filename.lastIndexOf(PERIOD);
    if (j>=0 && j>i) {
      root = filename.substring(i+1,j);
      ext = filename.substring(j+1,ls);
    } else {
      root = filename.substring(i+1);
      ext = "";
    }
    if (!keepCase && !quoted && !path.endsWith("/RPF/")) {
      path = fixCase(path);
      root = fixCase(root);
      ext  = fixCase(ext);
    }
    if (qualifiers!=null) qualifiers = Parser.toUpperCaseWithQuotes(qualifiers);
    if (trimmers!=null) trimmers = Parser.toUpperCaseWithQuotes(trimmers);
    if (tag==null||tag.length()==0) tag = root.toUpperCase();
  }

  /** if all UpperCase return all LowerCase, else return as is */
  private static String fixCase (String name) {
    for (int i=0; i<name.length(); i++) {
      char c = name.charAt(i);
      if (c>=AsciiMap.LOWERA && c<=AsciiMap.LOWERZ) return name;
    }
    return name.toLowerCase();
  }

  /** Return the full name (path, filename and extension) as specified. This does NOT
      resolve the AUX used (e.g. "myfile.prm{AUX=HOME}" becomes "myfile.prm").
      The {@link #toURL()} and {@link #toLocalFileName()} functions will resolve
      the AUX (e.g. "myfile.prm{AUX=HOME}" becomes "file:///home/username/myfile.prm"
      or "/home/username/myfile.prm").
   */
  public String getFullName() {
    if (fullname==null) {
      fullname = path+root;
      if (ext.length()>0) fullname+="."+ext;
    }
    return fullname;
  }

  /** Return the full name with trimmers and qualifiers */
  public String getAllName() {
    String s = getFullName();
    if (hasTrimmers()) s += getTrimmers();
    if (hasQualifiers()) s += getQualifiers();
    return s;
  }

  /** Return the path component of the current filename. */
  public String getPath() { return path; }
  public void setPath (String value) { path=value; fullname=null; }

  /** Return a string array containing the file's path elements using the
      local system's path delimiter. */
  public String[] getPathElements () {
    String fname = getFullName();
    String delim = "/";
    if (fname.indexOf(BSLASH)>=0) delim = "\\";
    return StringUtil.cleanSplit(fname, delim);
  }

  /** Get the filename root (without extension) from a full path filename -or-
      from the NeXtMidas (Java like) NXM dot syntax (with filename extension!).
      NOTE1: This assumes the input filename has NO trimmers and qualifiers.
      NOTE2: path should have consistent path separators (behavior is undefined otherwise).
      For example:
      <pre>
        getRoot("myfile.prm")                    = myfile
        getRoot("/home/smith/myfile")            = myfile
        getRoot("/home/smith/myfile.prm")        = myfile
        getRoot("\data\myfile.tmp")              = myfile
        getRoot("home/smith/MyFile.abc")         = MyFile
        getRoot("file://home/smith/myfile.txt")  = myfile
        getRoot("http://server/data/myfile.tbl") = myfile
        getRoot("nxm.sys.test.MyFile.xyz")       = MyFile
        getRoot("nxm.sys.test.sub.myFILE.dat")   = myFILE
        getRoot("nxm.sys.test.zzz")              = test   (extension here is "zzz")
        getRoot("/home/smith/mydir/")            = "" (empty string, this is a directory)
        getRoot("file:/c:\home\myfile.prm")      = (INVALID! output is undefined!)
      </pre>
     @param path full path to filename -or- "nxm" dot syntax (with file extension)
     @return root of filename with no extension. null if path is null.
             Empty string if path ends with separator char (e.g. it is a directory).
     @see #getBasename(String)
     @see #getExt(String)
     @since NeXtMidas 3.1.2
   */
  public static String getRoot (String path) {
    if (path == null) return path;
    String root = getBasename(path);

    int ii = root.lastIndexOf('.'); // find last dot to get filename extension
    if (ii >= 0) root = root.substring(0, ii);

    return root;
  }

  /** Return the root name of the current filename.  The root name of a
      filename is the component following the path but preceding the extension.
      @return The string containing the file's root name.
  */
  public String getRoot() { return root; }

  /** Set the root name of the instance. This will change the file that this
      instance is pointing to.
      @param value The new root name for the instance.
   */
  public void setRoot (String value) { root=value; fullname=null; }


  /** Return the final extension of the given pathname. The extension is
      defined as everything following the final period ('.') in the name.
      NOTE: This assumes the input filename has NO trimmers and qualifiers.
      In 3.1.2:
        This returns an empty string ("") if filename has NO extension to match behavior of
          {@link #getExt()} & {@link #getRoot()} (after {@link #parseName(String)}),
          {@link #getBasename()}, {@link #getRoot(String)}, etc. Previously, this would
          have return back null.
        Also, if name is null, then null is returned. Previously, this would have
          caused a NullPointerException.
        Use {@link #getExtLegacy(String)} for previous behavior.

      @param name The name of the file to return the extension for.
      @return The file extension or empty string ("") if it does not have one or
              null if name is null.
      @see #getBasename(String)
      @see #getRoot(String)
  */
  public static String getExt (String name) {
    if (name == null) return name;
    int pos = name.lastIndexOf('.');
    return pos >= 0 ? name.substring(pos + 1) : "";
  }

  /** @return The file extension or empty string ("") if it does not have one.
      @deprecated Since it was initial added in NeXtMidas 3.1.2
   */
  @Deprecated
  public static String getExtLegacy (String name) {
    int pos = name.lastIndexOf('.');
    return pos >= 0 ? name.substring(pos + 1) : null;
  }

  /** Return the extension of the current filename.  */
  public String getExt() { return ext; }
  public void setExt (String value) { ext=value; fullname=null; }

  /** Get the basename (filename root + extension, i.e. no path) from a file path -or-
      from the NeXtMidas (Java like) NXM dot syntax (with filename extension!).
      NOTE1: This assumes the input filename has NO trimmers and qualifiers.
      NOTE2: Path should have consistent path separators (behavior is undefined otherwise).
      For example:
      <pre>
        getBasename("myfile.prm")                    = myfile.prm
        getBasename("/home/smith/myfile")            = myfile
        getBasename("/home/smith/myfile.prm")        = myfile.prm
        getBasename("\data\myfile.tmp")              = myfile.tmp
        getBasename("home/smith/MyFile.abc")         = MyFile.abc
        getBasename("file://home/smith/myfile.txt")  = myfile.txt
        getBasename("http://server/data/myfile.tbl") = myfile.tbl
        getBasename("nxm.sys.test.MyFile.xyz")       = MyFile.xyz
        getBasename("nxm.sys.test.sub.myFILE.dat")   = myFILE.dat
        getBasename("nxm.sys.test.zzz")              = test.zzz
        getBasename("/home/smith/mydir/")            = "" (empty string, this is a directory)
        getBasename("file:/c:\home\myfile.prm")      = (INVALID! output is undefined!)
      </pre>
     @param path full path to filename -or- "nxm" dot syntax (with file extension)
     @return basename (root + ext) of filename. null if path is null.
             Empty string if path ends with separator char (e.g. it is a directory).
     @see #getRoot(String)
     @see #getExt(String)
     @since NeXtMidas 3.1.2
   */
  public static String getBasename (String path) {
    if (path == null) return path;
    String basename = path;
    int ii = path.lastIndexOf('/');               // check for Unix/Web style paths
    if (ii < 0) ii = path.lastIndexOf('\\');      // otherwise check for Windows style paths
    if (ii >=0 ) basename = path.substring(ii+1); // trim head off path (keeping root+ext)
    else if (path.startsWith("nxm.") || path.startsWith("NXM.")) {
      // we have NXM dot syntax (nxm.<opt>.<area>.<filename>.<ext>, etc.)
      int iLastDot = path.lastIndexOf('.');   // find last dot
      ii = path.lastIndexOf('.', iLastDot-1); // next 2nd to last dot
      if (ii >= 0) basename = path.substring(ii+1);
      return basename;
    }
    return basename;
  }

  /** Return the basename of the current filename.  The base name of a
      filename is the root name with file extension (i.e. no path).
      @return The string containing the file's basename.
      @since NeXtMidas 3.1.1
  */
  public String getBasename () {
    return root + ((ext!=null && ext.length()>0)? '.'+ext : "");
  }

  /** Return the tag for the current filename. */
  public String getTag() { return tag; }
  public void setTag (String value) { tag=value; }

  /** Return if there are any qualifiers (non empty) */
  public boolean hasQualifiers() { return (getQualifierTable().size() > 0); }

  /** Return the qualifiers of the current filename.  */
  public String getQualifiers() {
    if (qualTable!=null) return qualTable.toString();
    else return qualifiers;
  }

  /** Return the qualifiers of the current filename.
      Note: prior to NeXtMidas 3.1.0, this returned null vs an empty Table
            when no qualifiers are specified. See BUG 2497 for details.
      @return Table object (empty Table if there are no qualifiers specified).
   */
  public Table getQualifierTable () {
    if (qualTable!=null) return qualTable;
    if (qualifiers==null) {
      return (qualTable = new Table()); // empty Table (instead of null) since 3.1.0 for BUG 2497
    }
    if (qualifiers.charAt(0)==LCURLY) qualTable = new Table(qualifiers);
    else if (qualifiers.charAt(0)==LPAREN) {	  // search for old style qualifiers
      qualTable = new Table();
      for (int pos=0;;) {
        int i = qualifiers.indexOf(LPAREN,pos);
        if (i<0) break;
        int j = qualifiers.indexOf(EQUALS,i);
        int k = StringUtil.getClosingFensePos(qualifiers, "()", i);
        if (j>i && j<k) qualTable.put(qualifiers.substring(i+1,j),qualifiers.substring(j+1,k));
        pos = k;
      }
    }
    return qualTable;
  }

  /** Set the qualifiers for the instance.  The parameter is expected to be
      the full string, complete with the open and close brackets.   This will
      overwrite any previous string.
      @param quals The full qualifier string.
  */
  public void setQualifiers (String quals) {
    qualifiers = quals;
    qualTable = null;
  }

  /** set a qualifier byte name and value */
  public void setQualifier (String key, Object value) {
    // if (qualTable==null) qualTable = new Table(qualifiers); // commented out as part of BUG 2497 workoff
    // qualifiers = null;                                      // moved to end as part of BUG 2497 workoff
    // qualTable.put(key,value);                               // switched to below line as part of BUG 2497 workoff
    getQualifierTable().put(key, value);
    qualifiers = null;                                         // qualifiers string is now invalid
  }
  /** get a qualifier by name */
  public Object getQualifier (String key) {
    return getQualifierTable().get(key);
  }
  /** get a qualifier by name with default value if not found
      @since NeXtMidas 3.1.0
   */
  public Object getQualifier (String key, Object defValue) {
    return getQualifierTable().get(key,defValue);
  }

  /** Add the specified default qualifier (key/value) in this FileName.
      Any existing entries in the qualifiers table (e.g. user specified or previous added) will NOT be overridden.
      @param key   The qualifier's key.
      @param value The qualifier's (new) value.
      @return true if key/value is added to the file's qualifiers table, otherwise false
                   (e.g. there is an existing user specified qualifier for that key).
      @see Table#addIfNotPresent(String,Object)
      @since NeXtMidas 3.1.0
    */
  public boolean setDefaultQualifier (String key, Object value) {
    return getQualifierTable().addIfNotPresent(key, value);
  }

   /** Add following Map of key/value,...,keyN,valueN to the default qualifiers in this FileName.
       Any existing entries in the qualifiers table (e.g. user specified or previous added) will NOT be overridden.
       @param defQualsMap default qualifiers Map (e.g. Table) to add to file
       @return true if any key/value from defQualsMap is added to the file's qualifiers table.
       @see #setDefaultQualifier(String, Object)
       @see Table#addIfNotPresent(Map)
       @since NeXtMidas 3.1.0
   */
  public boolean setDefaultQualifiers (Map<String,Object> defQualsMap) {
    return getQualifierTable().addIfNotPresent(defQualsMap);
  }

  /** Return if there are any qualifiers */
  public boolean hasTrimmers() { return (trimmers!=null); }

  /** Return the trimmers of the current filename.  */
  public String getTrimmers() { return trimmers; }

  /** add trimmers or qualifiers by string */
  public void append (String filename) {
    int i = getQualifiersIndex(filename); // file qualifiers
    if (i>=0) {
      if (qualTable!=null) qualTable.append (filename.substring(i));
      else if (qualifiers!=null) qualifiers =
        qualifiers.substring(0,qualifiers.length()-1)+","+filename.substring(i+1);
      else qualifiers = filename.substring(i);
      filename = filename.substring(0,i);
    }
    i = getTrimmersIndex(filename); // file trimmers
    if (i>=0) {
      trimmers = filename.substring(i);
//    filename = filename.substring(0,i); // 2011-12-09 NTN: commented out, since not used afterwards
    }
  }

  /** Dump the class members to {@link System#out} for debugging */
  public void dump () {
    System.out.println("Path = "+path);
    System.out.println("Root = "+root);
    System.out.println("Ext  = "+ext);
    System.out.println("Full = "+getFullName());
    System.out.println("Trim = "+trimmers);
    System.out.println("Tag  = "+tag);
    System.out.println("Qual (String) = "+qualifiers);
    System.out.println("Qual (Table)  = "+qualTable);
  }

  /** Return the full path string */
  public String toString () {
    String tmp = getFullName();
    if (trimmers!=null) tmp+=trimmers;
    if (qualifiers!=null) tmp+=qualifiers;
    return tmp;
  }

  /** Converts the file name to a local file name using {@link FileUtil#toLocalFileName(URL)}
      @param dir I/O direction. See "Direction Values" at the top of {@link IOResource} for details
      @return a local file name or null if the file is a remote resource
      @throws MalformedURLException if a URL cannot be formed
      @since NeXtMidas 3.1.0
   */
  public String toLocalFileName (int dir) throws MalformedURLException {
    return FileUtil.toLocalFileName(toURL(dir));
  }

  /** Converts the file name to a local file name using {@link FileUtil#toLocalFileName(URL)}
      @return A local file name or null if the file is a remote resource
      @throws MalformedURLException if a URL cannot be formed
      @since NeXtMidas 3.1.0
   */
  public String toLocalFileName () throws MalformedURLException {
    return FileUtil.toLocalFileName(toURL());
  }

  /** Converts an object into a file name. If the object is a {@link FileName}, the given object is
      just returned. If the object is not a {@link FileName} it is converted to a string and then a
      new {@link FileName} is created.
      @param obj An object.
      @return A file name.
      @since NeXtMidas 2.3.1
   */
  public static FileName toFileName (Object obj) {
    FileName fn;

    if (obj instanceof FileName) {
      fn = (FileName)obj;
    } else {
      fn = new FileName(Convert.o2s(obj));
    }
    return fn;
  }

  /** Converts an object into a URL. If the object is a {@link URL}, the given object is just
      returned; if the object is a {@link URI} it is converted to a {@link URL}, otherwise the
      the NeXtMidas file naming conventions are assumed (AUXs, qualifiers, etc) and this parses
      the name based on them. <br>
      <br>
      Unlike simply calling <tt>new URL(..)</tt>, this knows how to resolve Midas'isms like the
      "nxm." syntax and knows to omit any file qualifiers or trimmers.
      @param ref The reference to use (usually an instance of {@link Midas} or {@link Command}).
      @param obj The filename to parse.
      @param dir I/O direction. See "Direction Values" at the top of {@link IOResource} for details.
      @return A file name.
      @throws MalformedURLException if a URL can not be formed.
      @since NeXtMidas 2.7.0
   */
  public static URL toURL (Object ref, Object obj, int dir) throws MalformedURLException {
    if (obj instanceof URL) return (URL)obj;
    if (obj instanceof URI) return ((URI)obj).toURL();

    BaseFile bf  = new BaseFile(ref, obj);
    String   url = bf.findResource(dir).getURL();

    return new URL(url);
  }

  /** Converts the file name to a URL using <tt>toURL(null,this,dir)</tt>.
      @param dir I/O direction. See "Direction Values" at the top of {@link IOResource} for details.
      @return A URL.
      @throws MalformedURLException if a URL can not be formed.
      @since NeXtMidas 2.7.0
   */
  public URL toURL (int dir) throws MalformedURLException {
    return toURL(null, this, dir);
  }

  /** Converts the file name to a URL using {@link #toURL(int)} with a dir of 0
      @return a URL
      @throws MalformedURLException if a URL cannot be formed
      @since NeXtMidas 3.1.0
   */
  public URL toURL () throws MalformedURLException {
    return toURL(0);
  }

  /** Return a new filename given a name relative to an original context. This
      method assumes that the context is a fully qualified path name to a
      file of some sort (not just a directory). The operation of this method
      is easier to illustrate by example than by algorithm.   In all of these
      examples, the context is 'nxm/sys/hlp/concepts.hlp' and the token is '/'.
      <br>
      All files within NeXtMidas have extensions.  If the name parameter does
      not have an extension on it then the returned name will have the same
      extension as this object.
      <pre>
        relativeTo("plot.hlp")          --&gt "nxm/sys/hlp/plot.hlp"
        relativeTo("exp/fft.exp")       --&gt "nxm/sys/hlp/exp/fft.exp"
        relativeTo("jay/exp/foo.exp")   --&gt "nxm/sys/hlp/jay/exp/foo.exp"
        relativeTo("include_data")      --&gt "nxm/sys/hlp/include_data.hlp" (no extension is NOT RECOMMENDED!)
        relativeTo("file.ext area")     --&gt "$NMROOT/nxm/sys/area/file.ext"
        relativeTo("file.ext area opt") --&gt "$NMROOT/nxm/<opt>/area/file.ext" (option tree under $NMROOT only!)
      </pre>
      @param name The name of the file relative to this one.
      @return The full name of the file relative to this one.
  */
  public FileName relativeTo (final String name) throws MalformedURLException {
    String context = getFullName();

    // If the name has spaces then assume it is specifying areas, option trees (under $NMROOT),
    // and files.  If there are no spaces then assume it's a relative filename.
    String[] fields = StringUtil.cleanSplit(name, " ");
    String nmroot = Shell.getNmRoot(); // 2010-01-08 NTN: changed to work under Java Web Start for Bug 1463
    if ( nmroot == null ) nmroot = "";

    URL newUrl = null;

    switch ( fields.length ) {
      case ( 2 ): { // <filename> <area> of SYS option tree
        String path = getOptionPath(nmroot, "sys", fields[1]);
        // 2010-01-08 NTN TODO: the next line DOES NOT WORK under Windows (since it creates URL with forward slashes)
        newUrl = new URL (new URL("file:" + path), fields[0]);
        break;
      }
      case ( 3 ): { // <filename> <area> <optionTree under $NMROOT>
        String path = getOptionPath(nmroot, fields[2], fields[1]);
        // 2010-01-08 NTN TODO: the next line DOES NOT WORK under Windows (since it creates URL with forward slashes)
        newUrl = new URL (new URL("file:" + path), fields[0]);
        break;
      }
      default: { // <filename>
        if ( !context.startsWith("file:") ) context = "file:" + context;
        newUrl = new URL (new URL(context), name);
        break;
      }
    }

    // Using the newly constructed URL, determine if we need to add an extension.
    String retval = newUrl.getFile();
    // NTN: fix for Bug 1463 e.g. when retval = "jar:http://server/nxm/jars/nxm-sys.jar!/nxm/sys/test/include_data"
    FileName retfn = new FileName(retval, true); // true to keep case
    String ext = retfn.getExt();
    if (ext == null || ext.equals("")) {
      retfn.setExt(getExt());
    }
    return retfn;
  }

  public FileName cloneOf() {
    try {
      return (FileName)this.clone();
    } catch (CloneNotSupportedException e) {
      System.out.println("Clone not supported");
      return this;
    }
  }

// ********************************************************************** //
//             PUBLIC STATIC METHODS
// ********************************************************************** //

  /** Add required terminating directory separator if not already present
      The end terminator is gleaned from the input path
      @param inpath Input path
      @return The terminated path.
   */
  public static String terminatePath (String inpath) {
    return FileUtil.terminatePath(inpath);
  }

  /** Return the index to the [start, end] of the qualifier information in the given
      URL. If there are no qualifiers on the URL or the URL is null then this
      method will return -1.  Qualifiers are indicated on the url by the
      presence of a left curly brace ({).
      @param url The url to search for the qualifiers.
      @return The position of the [start, end] of the qualifier information in the
              url or -1 if there is no qualifier information.
   */
  private static int[] getQualIndex (String url) {
    if (url == null) return null;

    // url = trim(url); // ORIG 2.7.3 and prior
    url = StringUtil.rtrim(url); // remove trailing whitespace (TODO: JC4, why not trim leading whitespaces?)
    if (url.length() == 0) return null;
    if (!url.endsWith("}") && !url.endsWith(")")) {
      return null;
    }

    int end   = url.length();
    int index = -1;

    while (url.length() > 0) {
      int i = -1;

      if (url.endsWith("}")) { // standard NeXtMidas qualifier table
        i = StringUtil.getOpeningFensePos(url, "{}", url.length()-1);
      }
      else if (url.endsWith(")")) { // old-style single-enrty X-Midas qualifier
        i = url.lastIndexOf('('); // no nesting with this style
        if ((i > 0) && url.substring(i).indexOf('=') < 0) i = -1;
      }

      if (i < 0) break;

      index = i;
      url = url.substring(0,i);
    }
    return new int[] { index, end };
  }

  /** Return the index to the start of the qualifier information in the given
      URL. If there are no qualifiers on the URL or the URL is null then this
      method will return -1.  Qualifiers are indicated on the url by the
      presence of a left curly brace ({).
      @param url The url to search for the qualifiers.
      @return The position of the start of the qualifier information in the
              url or -1 if there is no qualifier information.
   */
  public static int getQualifiersIndex (String url) {
    int[] index = getQualIndex(url);
    return (index == null)? -1 : index[0];
  }

  /** Return the string of the qualifier data that is contained in the url.
      If there is no qualifier string present, the method will return null.
      For example, if url is "http://localhost/foo.tmp{exist=true}", then this
      method will return "{exist=true}".
      @param url The url to check for qualifiers
      @return The qualifier string (including ending and closing qualifiers)
              or null if no qualifiers exist.
   */
  public static String getQualifierString (String url) {
    int[] index = getQualIndex(url);
    return (index == null)? null : url.substring(index[0], index[1]);
  }

  /** Get the trimmers start index - JPH -- why know this outside of this class ? */
  public static int getTrimmersIndex (String url) {
    // url = trim(url); // ORIG 2.7.3 and prior
    url = StringUtil.rtrim(url); // remove trailing whitespace (TODO: JC4, why not trim leading whitespaces?)

    int index = getQualifiersIndex(url);
    if (index > 0) url = url.substring(0,index); // ignore any qualifiers at the end

    if (!url.endsWith(")")) return -1;

    int i = url.lastIndexOf('('); // no nesting with trimmers
    if ((i > 0) && url.substring(i).indexOf(':') < 0) i = -1;
    return i;
  }

  /** Creates an option tree path in a system independent manner.  Uses the rootpath
      to determine the file system type.  The root path must have a delimiter.
      Note: opt and area will be converted to lower case to match NeXtMidas naming
      convention when forming path on OS file system.
      @param rootpath the path to start from (cannot be null!)
      @param opt      the name of the option tree
      @param area     the area in the option tree (to append to return value if NOT null)
      @throws IllegalArgumentException if rootpath is null (Since NeXtMidas 3.1.2)
   */
  public static String getOptionPath (String rootpath, String opt, String area) {
    if (rootpath == null) throw new IllegalArgumentException("rootpath cannot be null!");
    String path = addPathDir(rootpath,"nxm");
    if (opt  != null) path = addPathDir(path,opt.toLowerCase());
    if (area != null) path = addPathDir(path,area.toLowerCase());
    return path;
  }

  /** Appends a directory to an existing path. The path (directory) separator is
      based on input path string and will always be included in return value.
      This will handle case where path does not have path terminator by adding it
      before appending the specified directory.
      @param path existing path (cannot be null or empty)
      @param dir  directory to append to path (if null or empty, then path is returned)
      @return new path string with a terminating path separator based on input path.
      @throws IllegalArgumentException if path is null AND dir is not null or empty
      @throws IllegalArgumentException if path is blank/empty (i.e. length of zero)
   */
  public static String addPathDir (String path, String dir) {
// 2010-01-08 NTN: TODO: Jeff, what should be returned when path is null or empty? (vs the above exceptions?)
    if (dir==null || dir.length()==0) return path;
    if (path==null) throw new IllegalArgumentException("path cannot be null!");
    int lp = path.length();
    if (lp==0) throw new IllegalArgumentException("path cannot be blank!");
    char term = path.charAt(lp-1);
         if (term==SLASH) path = path + dir + SLASH;
    else if (term==RBRACE) path = path.substring(0,lp-1) + PERIOD + dir + RBRACE; // TODO: Jeff why is a period inserted for this special case?
    else if (term==BSLASH) path = path + dir + BSLASH;
    else if (path.indexOf(BSLASH)>=0) path = path + BSLASH + dir + BSLASH;
    else path = path + SLASH + dir + SLASH;
    return path;
  }

  /** appends a filename root to an existing path */
  public static String addPathRoot (String path, String root) {
    if (root==null || root.length()==0) return path;
    int lp = path.length();
    if (lp==0) return root;
    char term = path.charAt(lp-1);
         if (term==SLASH) path += root;
    else if (term==RBRACE) path += root;
    else if (term==BSLASH) path += root;
    else if (path.indexOf(BSLASH)>=0) path += BSLASH + root;
    else path += SLASH + root;
    return path;
  }

  /** appends a filename extension to an existing path */
  public static String addPathExt (String path, String ext) {
    if (ext==null || ext.length()==0) return path;
    return path + PERIOD + ext;
  }

  /** appends a filename root and extension to an existing path */
  public static String addPathRootExt (String path, String root, String ext) {
    return addPathExt( addPathRoot(path,root), ext);
  }
}
