package nxm.sys.lib;

import nxm.sys.inc.*;
import nxm.sys.libg.Layer;
import java.io.FileFilter;
import java.io.FilenameFilter;
import java.io.File;
import java.io.IOException;
import java.net.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.regex.Pattern;

/** A collection of utility methods useful when working with files.

    @author  J. Czechowski IV
    @version $Id: FileUtil.java,v 1.41 2012/11/29 20:03:34 jxc Exp $
    @since   NeXtMidas 2.3.0
 */
public final class FileUtil {
  /** comma (',') or OS path-separator (Unix=':' or DOS/Window=';') regex (compiled) */
  public static final Pattern COMMA_OR_PATH_SEPARATOR = Pattern.compile("[,"+File.pathSeparator+"]");

  /** A file name separator '/' or '\' regex (compiled) */
  private static final Pattern FILE_SEPARATOR = Pattern.compile("[/\\\\]");
  /** Starts with Windows drive letter regex (compiled) */
  private static final Pattern WINDOWS_DRIVE  = Pattern.compile("^[a-zA-Z]:.*");
  /** Starts with '/' or '\' then Windows drive letter regex (compiled)*/
  private static final Pattern SLASH_WINDOWS_DRIVE = Pattern.compile("^[/\\\\][a-zA-Z]:.*");

  /** Only used by getLocalPrefixList() */
  private static String[] localPrefixList = null;

  /** Used with {@link #copyListFile(Midas,Object,Object,Object[])} to indicate that column
      definitions should be propagated from the input file to the output file.
   */
  public static final Object[] PROPAGATE = new Object[0];


  /** Prevent instantiation since all methods are static. */
  private FileUtil() { }

  /** This will perform a binary copy from one file to another. This opens (if necessary) and
      closes the input and output files.
      Note: Since NeXtMidas 2.9.1 the inFile and outFile can be either a file name or {@link BaseFile}.
      @param midas The midas context to use.
      @param inFile  The file name (or {@link BaseFile} instance) of the source file to read the data from.
      @param outFile The file name (or {@link BaseFile} instance) of the destination file to write/copy the data to.
   */
  public static void copyBaseFile (Midas midas, Object inFile, Object outFile) {
    copyBaseFile(midas, inFile, outFile, BaseFile.BUFFER_SIZE);
  }

  /** This will perform a binary copy from one file to another. This opens (if necessary) and
      closes the input and output files.
      Note: Since NeXtMidas 2.9.1 the inFile and outFile can be either a file name or {@link BaseFile}.
      @param midas The midas context to use.
      @param inFile  The file name (or {@link BaseFile} instance) of the source file to read the data from.
      @param outFile The file name (or {@link BaseFile} instance) of the destination file to write/copy the data to.
      @param bufSize transfer length (buffer size) in bytes
      @since NeXtMidas 3.1.2
   */
  public static void copyBaseFile (Midas midas, Object inFile, Object outFile, final int bufSize) {
    byte[]   buf   = new byte[bufSize];
    boolean  done  = false;
    BaseFile in    = (inFile  instanceof BaseFile)? (BaseFile) inFile  : new BaseFile(midas, inFile);
    BaseFile out   = (outFile instanceof BaseFile)? (BaseFile) outFile : new BaseFile(midas, outFile);
    double   pause = (midas != null)? midas.pause : 0.0125;

    if (!in.isOpen() && !in.open(BaseFile.INPUT)) {
      throw new MidasException("BaseFile: Error opening URL="+in.getURL()+" for INPUT.");
    }
    if (!out.isOpen() && !out.open(BaseFile.OUTPUT)) {
      throw new MidasException("BaseFile: Error opening URL="+out.getURL()+" for OUTPUT.");
    }

    // since 3.1.2: Pre-allocate output file length to input file's length.
    //              This can be disabled via {PREALLOCATELENGTH=false} qualifier
    if (out.getQualifiers().getState("PREALLOCATELENGTH", true)) {
      long inFileLen = in.getResource().getLength();
      IOResource ores = out.getResource();
      ores.setLength(inFileLen);
      if (ores.getLength() < inFileLen) {       // if output file length does not match
        out.getResource().setLength(inFileLen); // setLength again (e.g. for RamFile)
      }
    } // end since 3.1.2

    while (!done) {
      int bytesRead = in.read(buf, 0, buf.length);

      if (bytesRead == 0) {
        Time.sleep(pause);
      }
      else if (bytesRead < 0) {
        done = true;
      }
      else {
        out.write(buf,0,bytesRead);
      }
    } // end while loop

    in.close();
    out.close();
  }

  /** This will copy the data from one {@link ListFile} to another. This is the same as
      <code>copyListFile(midas,inFile,outFile,PROPAGATE)</code>.
      @param midas    The midas context to use.
      @param inFile   The file name of the source file to read the data from.
      @param outFile  The file name of the destination file to write/copy the data to.
      @throws ClassCastException If either of the file names does not match an instance of
                                 {@link ListFile}.
   */
  public static void copyListFile (Midas midas, Object inFile, Object outFile) {
    copyListFile(midas,inFile,outFile,PROPAGATE);
  }

  /** This will copy the data from one {@link ListFile} to another. Since NeXtMidas 3.3.0,
      this is the same as <code>copyListFile(midas,inFile,outFile,PROPAGATE,0)</code>.
      @param midas   The midas context to use.
      @param inFile  The file name of the source file to read the data from.
      @param outFile The file name of the destination file to write/copy the data to.
      @param cols    The names of the columns to include (or one of the special values listed above).
      @throws ClassCastException If either of the file names does not match an instance of
                                 {@link ListFile}.
   */
  public static void copyListFile (Midas midas, Object inFile, Object outFile, Object[] cols) {
    // Note: It is intentional that this is not a vararg function.
    copyListFile(midas, inFile, outFile, cols, 0);
  }

  /** This will copy the data from one {@link ListFile} to another. This is the same as:
      <pre>
        int propmask = Parser.mask(DataFile.propagationMaskList, hdrPropMask, Parser.MASK_ALL);
        copyListFile(midas,inFile,outFile,PROPAGATE,propmask);
      </pre>
      @param midas       The midas context to use.
      @param inFile      The file name of the source file to read the data from.
      @param outFile     The file name of the destination file to write/copy the data to.
      @param cols        The names of the columns to include (or one of the special values listed above).
      @param hdrPropMask The set of flags controlling header propagation to the output file.
      @throws ClassCastException If either of the file names does not match an instance of
                                 {@link ListFile}.
      @since NeXtMidas 3.3.0
   */
  public static void copyListFile (Midas midas, Object inFile, Object outFile, Object[] cols,
                                   String hdrPropMask) {
    int propmask = Parser.mask(DataFile.propagationMaskList, hdrPropMask, Parser.MASK_ALL);
    copyListFile(midas, inFile, outFile, cols, propmask);
  }

  /** This will copy the data from one {@link ListFile} to another. Unlike {@link #copyBaseFile}
      this is NOT a binary copy and can be used to copy data between two different file types (e.g.
      {@link DataFile} to {@link CsvFile}). <br>
      <br>
      There are two special values used when defining the columns in the output file:
      <pre>
        cols = null      - Columns are already defined (e.g. via a qualifier on fn2).
        cols = {@link #PROPAGATE} - Propagate column definitions from input file to output file.
                           (Since NeXtMidas 3.3.0, if TIME_NAME=, ABSC_NAME=, or INDEX_NAME= are
                           specified, they will be "virtually" propagated as required.)
      </pre>
      Since NeXtMidas 3.3.0 it is possible to propagate the time, abscissa, and index values
      to the output file. This is done by adding the following qualifiers to the output file:
      <pre>
        TIME_NAME=[col]  - Name of the column to be populated with the computed
                           time values. The time value inserted will be a {@link Time}
                           object (which will convert to String/number as needed).
                           This is only applicable when the input is:
                             - {@link DataFile} (time value taken from {@link DataFile#getTimeAt}), or
                             - {@link PlotFile} with {@link PlotFile#getUnits()} equal to
                               {@link Units#TIME} or {@link Units#TIMECODE} (value is
                               computed based on {@link PlotFile#getStart()} and
                               {@link PlotFile#getDelta()})

        ABSC_NAME=[col]  - Name of the column to be populated with the computed
                           abscissa values. This is only applicable when the input
                           is:
                             - {@link DataFile} or {@link PlotFile} (value is
                               computed based on {@link PlotFile#getStart()} and
                               {@link PlotFile#getDelta()})

        INDEX_NAME=[col] - Name of the column to be populated with the computed
                           index value. This applies to all file types and is
                           automatically computed based on the (zero-based) row
                           number.
      </pre>
      The primary use of <tt>TIME_NAME=</tt>, and <tt>ABSC_NAME=</tt> is in a {@link DataFile}
      to {@link CsvFile} conversion where this information would otherwise be lost. If
      <tt>TIME_NAME=</tt>, or <tt>ABSC_NAME=</tt> are used in a situation where they are not
      applicable, the default value for that column (0 or "") will generally be inserted. <br>
      <br>
      When using <tt>TIME_NAME=</tt>, <tt>ABSC_NAME=</tt> or <tt>INDEX_NAME=</tt>, if the
      column identified already exists in the input file, the value from the input file
      will be ignored in favor of the computed value (where applicable) -- EXCEPT when
      the column name is given with a "+" prefix (e.g. <tt>TIME_NAME=+TIME</tt>) in which
      case the value from the input file will be kept. <i>(The use of the "+" prefix to
      indicate optional column data is patterned off of <tt>M$TRANSFORM</tt> in X-Midas.)</i>
      @param midas       The midas context to use.
      @param inFile      The file name of the source file to read the data from.
      @param outFile     The file name of the destination file to write/copy the data to.
      @param cols        The names of the columns to include (or one of the special values listed above).
      @param hdrPropMask The set of flags controlling header propagation to the output file. This
                         takes the same values as used with {@link DataFile#propagationMaskList}.
                         If the input and output files are both instances of {@link DataFile} the
                         {@link DataFile#propagate} function is used for this. Otherwise if both
                         are instances of {@link PlotFile} the following fields will be set via
                         a set-keys call: XDELTA, XFRAME, XSTART, XUNITS, YDELTA, YFRAME, YSTART,
                         YUNITS, <b>TYPE, DELTA, START, UNITS</b>. The propagation of additional
                         fields and/or for additional file-type combinations may be added in the
                         future without warning, so it is important to set this to 0 if such is
                         not desired.
      @throws ClassCastException If either of the file names does not match an instance of
                                 {@link ListFile}.
      @since NeXtMidas 3.3.0
   */
  public static void copyListFile (Midas midas, Object inFile, Object outFile, Object[] cols,
                                   int hdrPropMask) {
    ListFile in  = (ListFile)BaseFile.getInstanceFor(midas,inFile);  // The input file
    ListFile out = (ListFile)BaseFile.getInstanceFor(midas,outFile); // The output file

    // OPEN INPUT FILE
    if (!in.open(BaseFile.INPUT)) {
      throw new MidasException("BaseFile: Error opening URL="+in.getURL()+" for INPUT.");
    }

    // PROPAGATE HEADER & KEYWORDS
    // This is done between the opening of the input and output files so that any user-specified
    // qualifiers controling the output file will be applied AFTER the propagated settings.
    if (hdrPropMask == 0) {
      // nothing to propagate
    }
    else if ((in instanceof DataFile) && (out instanceof DataFile)) {
      DataFile df = (DataFile)out;
      df.propagate((DataFile)in, hdrPropMask);
      df.setSubRecords(""); // <-- do not keep old definitions
      df.setSize(0.0);      // <-- set size=0 since we use append calls below
    }
    else if ((in instanceof PlotFile) && (out instanceof PlotFile)
                                      && ((hdrPropMask&DataFile.PM_HEADER) != 0)) {
      Table quals = new Table();
      quals.put("XDELTA", ((PlotFile)in).getXDelta());
      quals.put("XFRAME", ((PlotFile)in).getXFrame());
      quals.put("XSTART", ((PlotFile)in).getXStart());
      quals.put("XUNITS", ((PlotFile)in).getXUnits());
      quals.put("YDELTA", ((PlotFile)in).getYDelta());
      quals.put("YFRAME", ((PlotFile)in).getYFrame());
      quals.put("YSTART", ((PlotFile)in).getYStart());
      quals.put("YUNITS", ((PlotFile)in).getYUnits());

      quals.put("TYPE",   ((PlotFile)in).getType());
      quals.put("DELTA",  ((PlotFile)in).getDelta());
      quals.put("START",  ((PlotFile)in).getStart());
      quals.put("UNITS",  ((PlotFile)in).getUnits());

      KeyObject.setKeys(out, quals, null, KeyObject.SOFT);
    }

    // OPEN OUTPUT FILE
    if (!out.open(BaseFile.OUTPUT)) {
      throw new MidasException("BaseFile: Error opening URL="+out.getURL()+" for OUTPUT.");
    }

    // INITIALIZE COLUMN DEFINITIONS
    Table colDefs;

    if (cols == PROPAGATE) {
      colDefs = in.getRecordDefs();  // <-- taken from input file
    }
    else if (cols == null) {
      colDefs = out.getRecordDefs(); // <-- taken from output qualifiers
    }
    else {
      Table inDefs = in.getRecordDefs();

      colDefs = new Table();
      for (int i = 0; i < cols.length; i++) {
        String col = cols[i].toString();
        Table  def = inDefs.getTable(col);
        if (def == null) def = recDef(col, null, null, null, null);
        colDefs.put(col, def);
      }
    }

    // DETERMINE COMPUTED COLUMNS
    // Use "" rather than null so we don't need additional "if" checks the code
    // below loop code.
    String time  = Convert.o2s(out.getName().getQualifier("TIME_NAME",  ""));
    String absc  = Convert.o2s(out.getName().getQualifier("ABSC_NAME",  ""));
    String index = Convert.o2s(out.getName().getQualifier("INDEX_NAME", ""));

    if (time.startsWith("+")) {
      time = time.substring(1);
      if (in.getRecordDefs().containsKey(time)) time = ""; // don't override
    }
    if (absc.startsWith("+")) {
      absc = absc.substring(1);
      if (in.getRecordDefs().containsKey(absc)) absc = ""; // don't override
    }
    if (index.startsWith("+")) {
      index = time.substring(1);
      if (in.getRecordDefs().containsKey(index)) index = ""; // don't override
    }

    if (cols == PROPAGATE) {
      // "virtual" propagation of the special TIME, ABSC, and INDEX values
      if (!time.isEmpty() && !colDefs.containsKey(time)) {
        String form = (out instanceof DataFile)? "SD" : null;
        colDefs.put(time, recDef(time, form, null, Units.SCALAR, Units.TIME));
      }
      if (!absc.isEmpty() && !colDefs.containsKey(absc)) {
        Object units = (in instanceof PlotFile)? ((PlotFile)in).getUnits() : null;
        colDefs.put(absc, recDef(absc, "SD", null, Units.SCALAR, units));
      }
      if (!index.isEmpty() && !colDefs.containsKey(index)) {
        String form = (in.getNumberOfRows() <= Integer.MAX_VALUE)? "SL" : "SX";
        colDefs.put(index, recDef(index, form, null, Units.SCALAR, null));
      }
    }

    // SET COLUMN DEFINITIONS
    out.setRecordDefs(colDefs);

    // COPY DATA OVER
    // For performance reasons the main part of this is split into three blocks based on
    // input file type (DataFile/PlotFile/other) rather than using a single loop with
    // a large number of "if" statements.
    if (in instanceof DataFile) {
      DataFile df   = (DataFile)in;
      Table    data = new Table();
      DataFile dftc = null;
      
      // TODO: Rethink this check... goal is to prevent creating a TimeLine in
      // the output file when one did not exist in the input file.
      if (((hdrPropMask&DataFile.PM_TIMECODE)!=0) && (out instanceof DataFile)
          && (df.getTimeLineHandler() != null)) {
        dftc = (DataFile)out;
      }
      
      df.seek(0.0);
      for (long row = 0; true; row++) {
        Time   _time   = df.getTimeAtCurrent();
        double _absc   = df.getStart() + df.getDelta() * row;
        int    numRead = df.readDataTable(data);

        if (numRead < 0) break;

        data.put(time,  _time);
        data.put(absc,  _absc);
        data.put(index, row);

        if (dftc != null) {
          dftc.setTimeAt(_time);
        }
        out.insertData(-1.0, data);
      }
    }
    else if (in instanceof PlotFile) {
      PlotFile pf = (PlotFile)in;

      if ((pf.getUnits() != Units.TIME) || (pf.getUnits() != Units.TIMECODE)) {
        time = ""; // absc value != time value, don't include
      }

      pf.seek(0.0);
      for (long row = 0; row < in.getNumberOfRows(); row++) {
        double _absc = pf.getStart() + pf.getDelta() * row;

        Table data = in.getDataTable((double)row);

        data.put(time,  new Time(_absc));
        data.put(absc,  _absc);
        data.put(index, row);

        out.insertData(-1.0, data);
      }
    }
    else {
      for (long row = 0; row < in.getNumberOfRows(); row++) {
        Table data = in.getDataTable((double)row);

        data.put(index, row);

        out.insertData(-1.0, data);
      }
    }

    // CLOSE FILES
    in.close();
    out.close();
  }

  /** Creates a record definition for use with {@link ListFile} or {@link DataFile}.
      @param name   The name of the column.            (required)
      @param format The format (e.g. "SF" or "VD").    (optional, but strongly encouraged)
   */
  public static Table recDef (String name, String format) {
    if (name == null) throw new NullPointerException("Name required, given null.");
    Table def = new Table();
    def.put("NAME", name);
    if (format != null) def.put("FORMAT", format);
    return def;
  }

  /** Creates a record definition for use with {@link ListFile} or {@link DataFile}.
      The 'type' and 'units' can either be the formal names as strings or the
      numeric constant from {@link Units}.
      @param name   The name of the column.            (required)
      @param format The format (e.g. "SF" or "VD").    (optional, but strongly encouraged)
      @param offset The sub-record offset.             (optional, only applies to Type 3000 files)
      @param type   The sub-record type.               (optional, only applies to Type 5000/6000 files)
      @param units  The sub-record units.              (optional, only applies to Type 5000/6000 files)
   */
  public static Table recDef (String name, String format, Integer offset, Object type, Object units) {
    Table def = recDef(name, format);
    if (offset != null) def.put("OFFSET", offset);
    if (type   != null) def.put("TYPE",   type);
    if (units  != null) def.put("UNITS",  units);
    return def;
  }

  /** Makes a file accessible via the {@link ListFile} interface.
      @param file The file to use via the {@link ListFile} interface.
      @return Returns the following...
              <pre>
                (1) If file implements {@link ListFile}, file is returned, cast to
                    {@link ListFile}.
                (2) if file implements {@link PlotFileNH}, a read-only view of the
                    file is returned that conforms to the {@link ListFile} interface.
                (3) Returns null if unable to convert to a {@link ListFile}.
              </pre>
   */
  public static ListFile asListFile (Object file) {
    if (file instanceof ListFile  ) return (ListFile)file;
    if (file instanceof PlotFileNH) return new PlotFileAsListFile((PlotFileNH)file);

    return null;
  }

  /** Makes a file accessible via the {@link ListFile} interface.
      @param file The file to use via the {@link ListFile} interface.
      @return Returns the following...
              <pre>
                (1) If file implements {@link ListFile}, file is returned, cast to
                    {@link ListFile}.
                (2) if file implements {@link PlotFileNH}, a read-only view of the
                    file is returned that conforms to the {@link ListFile} interface.
                (3) Returns null if unable to convert to a {@link ListFile}.
              </pre>
   */
  public static PlotFileNH asPlotFileNH (Object file) {
    if (file instanceof PlotFileNH) return (PlotFileNH)file;

    if (file instanceof PlotFile  ) {
      if (file instanceof ListFile) return new PlotFileWrapper2((PlotFile)file, (ListFile)file);
      else                          return new PlotFileWrapper((PlotFile)file);
    }

    return null;
  }

  /** Resolves the canonical file name. Unlike {@link File#getCanonicalFile()} this will fall-back
      and use {@link File#getAbsoluteFile()} if there is an {@link IOException} thrown by the
      {@link File#getCanonicalFile()}. <br>
      <br>
      The canonical file name is the absolute file name after all symbolic links, <code>/./</code>
      and <code>/../</code> entries have been removed. See {@link File#getCanonicalFile()} for
      more details.
      @param file The file to get the canonical name for.
      @return The canonical name for the given file.
      @throws SecurityException If either the {@link File#getCanonicalFile()} or
              {@link File#getAbsoluteFile()} can not execute due to security limitations.
   */
  public static File getCanonicalFile (File file) {
    File canonical;
    try {
      // The next line resolves any symbolic links and any ../ entries, it may fail on some systems.
      canonical = file.getCanonicalFile();
    } catch (IOException e) {
      // Unable to resolve links, just fix any ../ entries.
      canonical = file.getAbsoluteFile();
    }
    return canonical;
  }

  /** Resolves the canonical URL. This is simply...
      <pre>
        URI uri = new URI(url.toString());
        uri = uri.normalize(); // Removes any ./ or ../ entries
        url = uri.toURL();
      </pre>
      ...except that in the event of an exception, the original URL is just returned. <br>
      <br>
      The canonical URL is the absolute file name after all <code>/./</code> and <code>/../</code>
      entries have been removed. See {@link URI#normalize()} for more details.
      @param url The URL to get the canonical form for.
      @return The canonical form of the URL.
      @since NeXtMidas 2.5.3
   */
  public static URL getCanonicalUrl (URL url) {
    try {
      URI uri = new URI(url.toString());
      uri = uri.normalize(); // Removes any ./ or ../ entries
      url = uri.toURL();
    } catch(URISyntaxException e) {
      // Unable to convert URL->URI
    } catch(MalformedURLException e) {
      // Unable to convert URI->URL
    }
    return url;
  }

  /** Resolves the canonical path. If the path is a local path (see {@link #isLocalFileName})
      then {@link #getCanonicalFile} is used; otherwise {@link #getCanonicalUrl} is used. In
      either case, if an error occurs, the original path is returned.
      @param path The path to get the canonical form for.
      @return The canonical form of the path.
      @since NeXtMidas 2.5.3
   */
  public static String getCanonicalPath (String path) {
    try {
      String local = toLocalFileName(path);

      if (local != null) {
        path = getCanonicalFile(new File(local)).toString();
      }
      else {
        path = getCanonicalUrl(new URL(path)).toString();
      }
    } catch(Exception e) {
      // Unable to convert
    }
    return path;
  }


  /** Gets the relative directory for src relative to dest. For example:
      <pre>
        getRelativeDir("/home/smith/nxm230/nxm/sys/lib/",
                       "/home/smith/nxm230/nxm/sys/lib/")    = ""

        getRelativeDir("/home/smith/nxm230/nxm/",
                       "/home/smith/nxm230/nxm/sys/lib/")    = "sys/lib/"

        getRelativeDir("/home/smith/nxm230/nxm/sys/lib/",
                       "/home/smith/nxm230/nxm/")            = "../../"
      </pre>
      @param src  The source DIRECTORY (the return value is relative to this).
      @param dest The destination DIRECTORY (this is the directory we are trying to get to
                  from src).
      @return The relative directory.
   */
  public static String getRelativeDir(File src, File dest) {
    return getRelativeDir(FileName.terminatePath(src.toString()),
                          FileName.terminatePath(dest.toString()),
                          File.separatorChar);
  }


  /** Gets the relative directory for src relative to dest. For example:
      <pre>
        getRelativeDir("http://alice.techma.com/nextmidas/",
                       "http://alice.techma.com/nextmidas/")            = ""

        getRelativeDir("http://alice.techma.com/nextmidas/snapshots/",
                       "http://alice.techma.com/nextmidas/")            = "../"

        getRelativeDir("http://alice.techma.com/nextmidas/",
                       "http://alice.techma.com/nextmidas/snapshots/")  = "snapshots/"

        getRelativeDir("http://alice.techma.com:8080/nextmidas/",
                       "http://alice.techma.com/nextmidas/")
                                          = "http://alice.techma.com/nextmidas/"

        getRelativeDir("http://68.167.166.136/nextmidas/",
                       "http://alice.techma.com/nextmidas/")
                                          = "http://alice.techma.com/nextmidas/"

        getRelativeDir("http://alice.techma.com/nextmidas/",
                       "ftp://alice.techma.com/nextmidas/")
                                          = "ftp://alice.techma.com/nextmidas/"

        getRelativeDir("http://alice.techma.com/nextmidas/",
                       "http://www.techma.com/")
                                          = "http://www.techma.com/"
      </pre>
      At this time no attempt is made to resolve host names (e.g. 68.167.166.136 and
      alice.techma.com are treated as unique although they point to the same system).
      This may be changed in the future.
      @param src  The source DIRECTORY (the return value is relative to this).
      @param dest The destination DIRECTORY (this is the directory we are trying to get to
                  from src).
      @return The relative directory.
   */
  public static String getRelativeUrl(URL src, URL dest) {
    String relDir;
    int    srcPort  = src.getPort();
    int    destPort = dest.getPort();
    String srcStr   = src.toString();
    String destStr  = dest.toString();

    if (!srcStr.endsWith("/"))  { srcStr  = srcStr  + "/"; }
    if (!destStr.endsWith("/")) { destStr = destStr + "/"; }

    if (src.getProtocol().equals(dest.getProtocol()) && (srcPort == destPort)
     && src.getHost().equals(dest.getHost())) {
      relDir = getRelativeDir(srcStr, destStr, '/');
    } else {
      relDir = destStr;
    }
    return relDir;
  }


  /** Gets the relative directory for src relative to dest. This is only for use by the other
      getRelativeDir(..) methods.
   */
  private static String getRelativeDir(Object src, Object dest, char separator) {
    StringBuilder relDir  = new StringBuilder(); // dest relative to content
    try {
      String[] srcPath   = splitPath(src.toString());
      String[] destPath  = splitPath(dest.toString());
      int      sameCount = 1; // Use sameCount=1 to skip "/" directory

      while ((sameCount < srcPath.length) && (sameCount < destPath.length)
             && srcPath[sameCount].equals(destPath[sameCount])) {
        sameCount++;
      }

      for (int i = sameCount; i < srcPath.length; i++) {
        relDir.append("..").append(separator);
      }

      for (int i = sameCount; i < destPath.length; i++) {
        relDir.append(destPath[i]).append(separator);
      }

    } catch (Exception e) {
      Shell.printStackTrace(e);
    }
    return relDir.toString();
  }

  /** Converts a given URL to a local file name
      @param url The given URL
      @return The file name without any "file:" prefix or null if it is not a local file
      @since NeXtMidas 3.1.0
   */
  public static String toLocalFileName (URL url) {
    return toLocalFileName(url.toString());
  }

  /** Indicates if a given URL represents a local file name
      @param url The given URL
      @return true if it is a local file name, false if not
      @see #isLocalFileName(String)
      @since NeXtMidas 3.1.0
   */
  public static boolean isLocalFileName (URL url) {
    return isLocalFileName(url.toString());
  }

  /** Splits a string that contains items that are comma-separator of OS path-separator separated
      into an array of Strings. This never returns null so that client usage of this API is simplified.
      @param entries comma or OS path-separator separated string list of entries
      @return string array of entries (empty if entries is null or empty)
      @since NeXtMidas 3.0.1
   */
  public static String[] splitEntries (String entries) {
    if (entries == null) return StringUtil.EMPTY_STRING_ARRAY;
    return COMMA_OR_PATH_SEPARATOR.split(entries, 0);
  }

  /** Splits a path into its component parts.
      @param path is first converted to a String (cannot be null!) then split up
   */
  private static String[] splitPath(Object path) {
    return FILE_SEPARATOR.split(path.toString());
  }

  /** Gets the path separator used in the given path based on the first
      path separator character occurrence in the given path.
      @param path A path.
      @return The path separator used in path or null if unknown.
      @since NeXtMidas 2.5.3
   */
  public static String getPathSeparator (String path) {
    if (path.indexOf('/')  >= 0) return "/";  // UNIX or URL
    if (path.indexOf('\\') >= 0) return "\\"; // DOS
    if (path.indexOf('[')  >= 0) return "[";  // VMS
    return null;
  }

  /** Add required terminating directory separator if not already present.
      The end terminator is gleaned from the input path. If the separator
      cannot be gleaned from input path, the local system's file separator
      (i.e. {@link File#separator}) will be used.
      @param path The path to terminate.
      @return The terminated path or path if already terminated.
      @since NeXtMidas 2.5.3
  */
  public static String terminatePath (String path) {
    String term = getPathSeparator(path);

    if (term == null) {
      // term = "/";  // DEFAULTS to UNIX prior to NeXtMidas 2.9.0
      term = File.separator; // use local system's file/directory separator
      // 2010-07-02 NTN: I commented out the below warning since Shell.warning() tries to get a Midas context
      //                 which caused it to boot up a NeXtMidas session when using this class as a library.
      // TODO: we should use a more standard logging framework like SLF4J to log messages at the API level
      // Shell.warning("No known delimiter in path '"+path+"', using UNIX");
    }

    return (path.endsWith(term))? path : path+term;
  }

  /** Converts the separators in a file path to the correct version. This changes any slashes
      ('/') or back-slashes ('\') to a given character. This is commonly used in one of two ways:
      <pre>
        changePathSeparator(path, File.pathSeparator)   - Converts to local file system
        changePathSeparator(path, "/")                  - Converts to URL style
      </pre>
      @param path      The file or directory path.
      @param separator The separator to use.
      @return The path with the correct separator in use.
   */
  public static String changePathSeparator (String path, String separator) {
    String repl = StringUtil.textToReplacement(separator);
    return FILE_SEPARATOR.matcher(path).replaceAll(repl);
  }

  /** Converts a given path to a local file name. For example:
      <pre>
        toLocalFileName("/home/smith/myfile.prm")             = "/home/smith/myfile.prm"
        toLocalFileName("file:/home/smith/myfile.prm")        = "/home/smith/myfile.prm"
        toLocalFileName("file:///home/smith/myfile.prm")      = "/home/smith/myfile.prm"
        toLocalFileName("ramd:/midas/icedisk/myfile.prm")     = "/midas/icedisk/myfile.prm"
        toLocalFileName("http://alice.techma.com/myfile.prm") = null
        toLocalFileName("c:\data\myfile.prm")                 = "c:\data\myfile.prm"
        toLocalFileName("file:/c:/data/myfile.prm")           = "c:\data\myfile.prm" (Since 2.9.0)
      </pre>
      Note that the local file name will use the local path separator:
      <pre>
        Unix:
          toLocalFileName("file:/home/smith/myfile.prm")      = "/home/smith/myfile.prm"
        Windows:
          toLocalFileName("file:/home/smith/myfile.prm")      = "\home\smith\myfile.prm"
      </pre>
      @param path The file name either as a normal path or with the "file:" prefix.
      @return The file name without any "file:" prefix or null if it is not a local file.
      @see #isLocalFileName(String)
   */
  public static String toLocalFileName (String path) {
    String prefix = getLocalPrefix(path);
    String fname  = null;

    if (prefix != null) {
      fname = path.substring(prefix.length()); // remove the "file:"

      if (fname.startsWith("///")) {
        fname = fname.substring(2); // if it was "file:///home/junk" make it "/home/junk"
      }
      else if (SLASH_WINDOWS_DRIVE.matcher(fname).matches()) {
        fname = fname.substring(1); // remove slash before windows drive letter
      }
      fname = changePathSeparator(fname, File.separator);
    }
    return fname;
  }

  /** Indicates if a given path to a local file name. For example:
      <pre>
        isLocalFileName("/home/smith/myfile.prm")             = true
        isLocalFileName("file:/home/smith/myfile.prm")        = true
        isLocalFileName("file:///home/smith/myfile.prm")      = true
        isLocalFileName("http://alice.techma.com/myfile.prm") = false
        isLocalFileName("c:\data\myfile.prm")                 = true
        isLocalFileName("file:/c:/data/myfile.prm")           = true
        // NOTE: "ramd:" is a special case for ICE/Midas RamDisk
        isLocalFileName("ramd:/midas/icedisk/myfile.prm")     = true
      </pre>
      @param path The file name either as a normal path or with the "file:" prefix.
      @return true if it is a local file name, false if not.
      @see #toLocalFileName(String)
   */
  public static boolean isLocalFileName(String path) {
    return (getLocalPrefix(path) != null);
  }

  /** Determine which local prefix is used.
      @param path The path.
      @return null=NotLocal, ""=AlreadyLocal, "xxx:"=Prefix is "xxx:"
   */
  private static String getLocalPrefix(String path) {
    String prefix = null;
    if (path.indexOf(':') < 0) {                        // No ':' in the path name, assume local
      prefix = "";
    } else if (WINDOWS_DRIVE.matcher(path).matches()) { // Starts with a Windows drive letter
      prefix = "";
    } else {                                            // Check to see if it begins with "file:",...
      String[] list = getLocalPrefixList();
      for (int i = 0; (prefix == null) && (i < list.length); i++) {
        if (path.startsWith(list[i])) {
          prefix = list[i];
        }
      }
    }
    return prefix;
  }

  /** Indicates if a given path starts with a local path separator character(s)
      (aka the root directory).
      For example:
      <pre>
        startsWithRootDir("/home/smith/myfile.prm")            = true only for unix
        startsWithRootDir("\data\myfile.prm")                  = true only for windows
        startsWithRootDir("home/smith/myfile.prm")             = false
        startsWithRootDir("data\myfile.prm")                   = false
        startsWithRootDir("file:/home/smith/myfile.prm")       = false
        startsWithRootDir("http://alice.techma.com")           = false
        startsWithRootDir("dd:\data\myfile.prm")               = false
        // NOTE: the two below is true under unix, but
        startsWithRootDir("c:\data\myfile.prm")                = true for both unix/windows
        startsWithRootDir("x:\data\myfile.prm")                = true for both unix/windows
        // NOTE: the below path usage is not recommended without the "\" after the drive letter
        startsWithRootDir("c:data\myfile.prm")                  = true
        startsWithRootDir("ramd:/midas/icedisk/myfile.prm")     = false
      </pre>
      @param path The file path to check
      @return true it it starts with the local path separator, otherwise false.
   */
  public static boolean startsWithRootDir(String path) {
    return (path.startsWith(File.separator))         // it starts with a '/' or a '\'
        || WINDOWS_DRIVE.matcher(path).matches();    // if starts with a Windows drive letter
  }

  /** Gets the list of prefixes (e.g. "file:", "ramd:") that are used with local files. */
  private static String[] getLocalPrefixList() {
    if (localPrefixList == null) {
      // JC4: This is where we should load the list of local file prefixes allowed. For now
      //      I've just hard-coded in the two most common ones.
      localPrefixList = new String[] { "file:", "ramd:", "ice:" };
    }
    return localPrefixList;
  }

  /** Recursively finds files on the local disk. The include/exclude filters allow filtering of
      the files. If both filters are null, all files will be included. Note that the filters are
      also applied to directory names even though the directories will not be included in the list
      (this allows exclusions of given directories while searching, e.g. "CVS").
      @param dir    The directory to search in.
      @param filter The filter that indicates the files to include (null=Include All).
      @return The files found.
      @since NeXtMidas 2.5.0
   */
  public static File[] findFiles (File dir, FileFilter filter) {
    ArrayList<File> allFiles = new ArrayList<File>();

    if (dir.isDirectory() && dir.exists()) {
      HashMap<File,File> dirMap = new HashMap<File,File>();
      File can = getCanonicalFile(dir);
      dirMap.put(can, dir); // put canonical dir path into Map to prevent going into backlinks (BUG 2444)
      findFiles(dirMap, allFiles, dir, filter);
    }
    return allFiles.toArray(new File[allFiles.size()]);
  }

  /** Recursive method used to find files. Only used by findFiles(..). */
  private static void findFiles (HashMap<File,File> dirMap, List<File> allFiles, File dir, FileFilter filter) {
    File[] files = dir.listFiles();

    for (int i = 0; i < files.length; i++) {
      File file = files[i];
      if ((filter == null) || filter.accept(file)) {
        allFiles.add(file);
      }
      if (file.isDirectory()) {
        File can = getCanonicalFile(file);
        if (dirMap.get(can) == null) { // never revisit a directory (in case of back links, etc)
          dirMap.put(can,file);
          findFiles(dirMap,allFiles,file,filter);
        }
      }
    }
  }

  /** Provide a platform neutral way to delete a file, because not all Operating
      Systems provide file deletion capabilities outside of the OS shell, and
      macros cannot run OS shell commands (only native OS commands can be run
      from macros).  This method is recursive because it deletes directories
      as well as child files under directories. This method will NOT follow
      symbolic links, but will delete it.
      @param file file or directory to recursively delete
      @return count of items deleted (zero for none or when file is null)
      @since NeXtMidas 3.1.0
   */
  public static int recursiveDelete (File file) {
    if (file == null) return 0;

    int count = 0;
    if (file.isDirectory()) {
      for (File child : file.listFiles()) {
        count += recursiveDelete(child);
      }
    }
    if (file.delete()) count++;
    // else  System.err.println("Could not delete: " + file);
    return count;
  }

  /** Provide a platform neutral way to recursively delete specified path.
      @param path path to file or directory to recursively delete (cannot be null)
      @return count of items deleted (zero for none)
      @see #recursiveDelete(File)
      @since NeXtMidas 3.1.0
   */
  public static int recursiveDelete (String path) {
    return recursiveDelete(new File(path));
  }

  /** Creates a file name filter. This is a short form of {@link #createFnFilter(String,String,String)}
      that only supports file name comparisons.
      @param filter The filter string passed in by the user.
      @param def    The default filter.
      @return The filter that accepts/rejects the files given.
      @throws NullPointerException If the filter string has a '/' in it.
      @since NeXtMidas 2.5.0
   */
  public static FnFilter createFnFilter (String filter, String def) {
    return createFnFilter(null,filter,def);
  }

  /** Creates a file name filter. By default, all files not matching one of the filters given
      are excluded. There are special keywords/characters in the filter string to allow a user
      to easily include/exclude files.
      <pre>
        Special Characters (must be first character in an entry in filter string):
          '-'    - Exclude any file that matches this.
          '+'    - Include any file that matches this.

        Special Keywords (must be first entry in filter string):
            ALL      - Same as "*"
            DEF      - Include the defaults
            DEFAULT  - Include the defaults
            NONE     - Does nothing (all files excluded by default)

        Note: If filter is blank ("") or null the default filter will be used.
      </pre>
      Some examples:
      <pre>
        createFnFilter("/data/",   "*.xml|*.csv",     "*.tmp|*.prm")   -> Accepts *.xml or *.csv
        createFnFilter("/data/",   "+*.xml",          "*.tmp|*.prm")   -> Accepts *.xml, *.tmp, or *.prm
        createFnFilter("/data/",   "DEF|*.xml",       "*.tmp|*.prm")   -> Accepts *.xml, *.tmp, or *.prm
        createFnFilter("/data/",   "NONE|*.xml",      "*.tmp|*.prm")   -> Accepts *.xml only
        createFnFilter("/data/",   "-*.tmp",          "*.tmp|*.prm")   -> Accepts *.prm only
        createFnFilter("/data/",   "-*.xml",          "*.tmp|*.prm")   -> Accepts *.tmp or *.prm
        createFnFilter("/data/",   "-*.xml",          "")              -> All files rejected
        createFnFilter("/data/",   "",                "")              -> All files rejected
        createFnFilter("/data/",   "-*",              "*.tmp|*.prm")   -> All files rejected
        createFnFilter("/data/",   "NONE",            "*.tmp|*.prm")   -> All files rejected
        createFnFilter("/data/",   "*",               "*.tmp|*.prm")   -> Accepts all files
        createFnFilter("/data/",   "*|-*.xml",        "*.tmp|*.prm")   -> Accepts all files that are not *.xml
        createFnFilter("/data/",   "*|*.xml",         "*.tmp|*.prm")   -> Accepts all files
        createFnFilter("/data/",   "foo/*.txt|*.prm", "*.tmp|*.prm")   -> Accepts all *.prm files and all
                                                                          *.txt files in /data/foo/
        createFnFilter("c:\data\", "foo/*.txt|*.prm", "*.tmp|*.prm")   -> Accepts all *.prm files and all
                                                                          *.txt files in c:\data\foo\
      </pre>
      @param root   The root directory (should use '\' on Unix, '/' on Windows).
      @param filter The filter string passed in by the user (always uses '/'). (null=UseDefault)
      @param def    The default filter (null="*").
      @return The filter that accepts/rejects the files given.
      @throws NullPointerException If root is null and the filter string has a '/' in it.
      @since NeXtMidas 2.5.0
   */
  public static FnFilter createFnFilter (String root, String filter, String def) {
    List<String> incPath = new ArrayList<String>();
    List<String> excPath = new ArrayList<String>();
    List<String> incFile = new ArrayList<String>();
    List<String> excFile = new ArrayList<String>();
    boolean  includeDef  = false;
    boolean  includeAll  = false;

    if ((filter == null) || (filter.length() == 0)) {
      filter = (def != null)? def : "*";
    }

    String[] filtList = filter.split("[|]");
    int index = 0;

    if (filtList[0].equalsIgnoreCase("ALL") || filtList[0].equalsIgnoreCase("*")) {
      includeAll = true;
      index++;
    } else if (filtList[0].equalsIgnoreCase("NONE")) {
      index++;
    } else if (filtList[0].equalsIgnoreCase("DEF") || filtList[0].equalsIgnoreCase("DEFAULT")) {
      includeDef = true;
      index++;
    } else if (filtList[0].startsWith("+") || filtList[0].startsWith("-")) {
      includeDef = true;
    }

    if (includeDef) {
      if (def == null) {
        includeAll = true;
      } else {
        String[] defList = def.split("[|]");
        includeAll = makeFilterList(incPath,incFile,excPath,excFile,defList,0) || includeAll;
      }
    }
    includeAll = makeFilterList(incPath,incFile,excPath,excFile,filtList,index) || includeAll;

    if (includeAll) { incFile = null; incPath = null; } // make it faster for this special case

    return new FnFilterImpl(root,incPath,incFile,excPath,excFile);
  }

  /** Only used by createFileNameFilter(..) */
  private static boolean makeFilterList (List<String> incPath, List<String> incFile,
                                         List<String> excPath, List<String> excFile, String[] list, int index) {
    boolean includeAll = false;

    for (; index < list.length; index++) {
      String  filt = list[index];
      boolean excl = false;

           if (filt.startsWith("-")) { filt = filt.substring(1); excl = true;  }
      else if (filt.startsWith("+")) { filt = filt.substring(1); excl = false; }

      if (filt.length() == 0) {
        // ignore
      } else if (excl) {
        if (filt.indexOf('/') >= 0) excPath.add(filt);
        else                        excFile.add(filt);
      } else if (filt.equals("*")) { // must be after excl but before incl
        includeAll = true;
      } else {
        if (filt.indexOf('/') >= 0) incPath.add(filt);
        else                        incFile.add(filt);
      }
    }
    return includeAll;
  }

  /** A filter that is compatible with both {@link FilenameFilter}, and {@link FileFilter}.
      @since NeXtMidas 2.5.0
   */
  public static interface FnFilter extends FilenameFilter, FileFilter {
    @Override
    public String toString();
  }

  /** Extract the modified date from a file using the HTTP header info. If the
      file is a local file, this method will use {@link File#lastModified()}
      method. If the file is not local, a URL connection is made
      and {@link URLConnection#getLastModified()} is used.
      @param url The URL of the file to check
      @return Seconds since 1970 to the URL modified date, or -1 if not found
      @since NeXtMidas 2.7.1
    */
  public static long lastModified (URL url) {
    long dateInSeconds = -1;

    if (isLocalFileName(url.toString())) {
      dateInSeconds = (long) ((new File(url.getPath())).lastModified() * .001);
    } else {
      try {
        URLConnection conn;
        conn = url.openConnection();
        dateInSeconds = (long)(conn.getLastModified() * .001);
      }
      catch (IOException e) {
        Shell.printStackTrace(e);
      }
    }

    return dateInSeconds;
  }



  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Nested Classes
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** A private implementation of {@link FileUtil.FnFilter} */
  private static final class FnFilterImpl implements FnFilter {
    private final StringBuffer description;  // description for toString()
    private final Pattern[]    pathAccept;   // Acceptance based on full path
    private final Pattern[]    pathReject;   // Rejection based on full path
    private final Pattern[]    fileAccept;   // Acceptance based on name only
    private final Pattern[]    fileReject;   // Rejection based on name only

    /** Only used by createFileNameFilter(..) */
    FnFilterImpl (String root, List<String> accPath, List<String> accFile, List<String> rejPath, List<String> rejFile) {
      this.pathAccept  = (accPath == null)? null           : new Pattern[accPath.size()];
      this.fileAccept  = (accFile == null)? null           : new Pattern[accFile.size()];
      this.pathReject  = (rejPath == null)? new Pattern[0] : new Pattern[rejPath.size()];
      this.fileReject  = (rejFile == null)? new Pattern[0] : new Pattern[rejFile.size()];
      this.description = new StringBuffer();
      description.append("FnFilter: (accept: ");

      if (pathAccept == null) {
        description.append("* ");
      } else {
        compileRegex(description, fileAccept, accFile, "");
        compileRegex(description, pathAccept, accPath, root);
      }
      description.append("reject:");
      compileRegex(description, fileReject, rejFile, "");
      compileRegex(description, pathReject, rejPath, root);
      description.append(")");
    }

    private void compileRegex (StringBuffer descr, Pattern[] pat, List<String> wild, String root) {
      for (int i = 0; i < pat.length; i++) {
        String wc = wild.get(i);
        if (wc.indexOf('/') >= 0) {
          if ((root == null) || root.isEmpty()) {
            throw new NullPointerException("FileUtil.createFnFilter: found '"+wc+"', but root was null.");
          }
          wc = terminatePath(root) + wc;
        }
        wc = wc.replace('/', File.separatorChar);
        pat[i] = Pattern.compile(StringUtil.wildcardsToRegex(wc));
        descr.append(wc).append(' ');
      }
    }

    @Override
    public boolean accept (File dir, String name) {
      return accept(new File(dir,name));
    }

    @Override
    public boolean accept (File file) {
      String  path = file.getPath();
      String  name = file.getName();
      boolean ok   = (fileAccept == null);

      if (file.isDirectory()) {
        path = path + File.separator;
        for (int i = 0; !ok && (i < pathAccept.length); i++)   ok =  pathAccept[i].matcher(path).matches();
        for (int i = 0;  ok && (i < pathReject.length); i++)   ok = !pathReject[i].matcher(path).matches();
      } else {
        for (int i = 0; !ok && (i < fileAccept.length); i++)   ok =  fileAccept[i].matcher(name).matches();
        for (int i = 0; !ok && (i < pathAccept.length); i++)   ok =  pathAccept[i].matcher(path).matches();
        for (int i = 0;  ok && (i < fileReject.length); i++)   ok = !fileReject[i].matcher(name).matches();
        for (int i = 0;  ok && (i < pathReject.length); i++)   ok = !pathReject[i].matcher(path).matches();
      }
      return ok;
    }

    @Override
    public String toString() {
      return description.toString();
    }
  }


  /** Wraps a PlotFileNH as a read-only ListFile. */
  private static class PlotFileAsListFile implements ListFile {
    private final PlotFileNH pf;

    /** Creates a new instance. */
    PlotFileAsListFile (PlotFileNH pf) {
      this.pf = pf;
    }

    @Override public boolean  isOpen()                 { return pf.isOpen();            }
    @Override public boolean  open(int flags)          { return pf.open(flags);         }
    @Override public String   getURL()                 { return pf.getURL();            }
    @Override public FileName getName()                { return pf.getName();           }
    @Override public double   getNumberOfRows()        { return pf.getSize();           }
    @Override public boolean  getProtected()           { return true;                   } // Always read-only
    @Override public int      getRecordDefCount()      { return pf.getRecordDefCount(); }
    @Override public Table    getRecordDef(int i)      { return pf.getRecordDef(i);     }
    @Override public Table    getRecordDefs()          { return pf.getRecordDefs();     }
    @Override public Table    getDataTable(double off) { return pf.getDataTable(off);   }
              public int      seek (double off)        { return pf.seek(off);           }
              public double   seek()                   { return pf.seek();              }
              public double   avail()                  { return pf.avail();             }
    @Override public void     close()                  { pf.close();                    }
              public int      readDataTable(Table tbl) { return pf.readDataTable(tbl);  }
    @Override public void     removeData (double off)  { removeData(off,1);             }

    @Override
    public void setRecordDefs (Table defs) {
      throw new UnsupportedOperationException("Can not write to "+getName()+" file is read-only.");
    }
    @Override
    public int setData (double offset, Table table) {
      throw new UnsupportedOperationException("Can not write to "+getName()+" file is read-only.");
    }
    @Override
    public void setData (double element, String fieldName, Object obj) {
      throw new UnsupportedOperationException("Can not write to "+getName()+" file is read-only.");
    }
    @Override
    public void insertData (double offset, Object obj) {
      throw new UnsupportedOperationException("Can not write to "+getName()+" file is read-only.");
    }
    @Override
    public void removeData (double offset, long count) {
      throw new UnsupportedOperationException("Can not write to "+getName()+" file is read-only.");
    }
    public int writeDataTable (Table tbl) {
      throw new UnsupportedOperationException("Can not write to "+getName()+" file is read-only.");
    }
    @Override
    public void update() { // since 3.1.1
      throw new UnsupportedOperationException("Can not write to "+getName()+" file is read-only.");
    }
  }

  /** Wraps a PlotFileNH as a read-only ListFile. */
  private static class PlotFileWrapper implements PlotFileNH, ListFile {
    protected final PlotFile pf;

    /** Creates a new instance. */
    protected PlotFileWrapper (PlotFile pf) {
      this.pf = pf;
      if (pf == null) throw new NullPointerException("pf can not be null");
    }

    @Override public String   getURL()       { return pf.getURL();       }
    @Override public String   getTag()       { return pf.getTag();       }
    @Override public double   getSize()      { return pf.getSize();      }
    @Override public String   getFormat()    { return pf.getFormat();    }
    @Override public double   getStart()     { return pf.getStart();     }
    @Override public double   getDelta()     { return pf.getDelta();     }
    @Override public int      getUnits()     { return pf.getUnits();     }
    @Override public double   getXStart()    { return pf.getXStart();    }
    @Override public double   getXDelta()    { return pf.getXDelta();    }
    @Override public int      getXUnits()    { return pf.getXUnits();    }
    @Override public int      getXFrame()    { return pf.getXFrame();    }
    @Override public double   getYStart()    { return pf.getYStart();    }
    @Override public double   getYDelta()    { return pf.getYDelta();    }
    @Override public int      getYUnits()    { return pf.getYUnits();    }
    @Override public int      getYFrame()    { return pf.getYFrame();    }
    @Override public int      getType()      { return pf.getType();      }
    @Override public int      getMode()      { return pf.getMode();      }
    @Override public Layer    getLayer()     { return pf.getLayer();     }
    @Override public FileName getName()      { return pf.getName();      }
    @Override public boolean  isStream()     { return pf.isStream();     }
    @Override public boolean  isStreaming()  { return pf.isStreaming();  }
    @Override public boolean  isOpen()       { return pf.isOpen();       }
    @Override public double   seek()         { return pf.seek();         }
    @Override public double   avail()        { return pf.avail();        }

    @Override public Data    getDataBuffer (int size)   { return pf.getDataBuffer(size); }
    @Override public double  getTimeAt (double offset)  { return pf.getTimeAt(offset);   }
    @Override public Object  getQualifier (String key)  { return pf.getQualifier(key);   }
    @Override public boolean find (int dir)             { return pf.find(dir);           }
    @Override public boolean open (int flags)           { return pf.open(flags);         }
    @Override public int     read (Data data, int elem) { return pf.read(data,elem);     }
    @Override public int     seek (double off)          { return pf.seek(off);           }
    @Override public double  getNumberOfRows()          { return pf.getSize();           }
    @Override public boolean getProtected()             { return true;                   }
    @Override public int     getRecordDefCount()        { return pf.getXFrame();         }

    @Override public void    init (Object ref, Object fname) { pf.init(ref,fname); }
    @Override public void    connect (int mode)              { pf.connect(mode);   }
    @Override public void    close()                         { pf.close();         }
    @Override public void    removeData (double off)         { removeData(off,1);  }

    @Override
    public void setRecordDefs (Table defs) {
      throw new UnsupportedOperationException("Can not write to "+getName()+" file is read-only.");
    }
    @Override
    public int setData (double offset, Table table) {
      throw new UnsupportedOperationException("Can not write to "+getName()+" file is read-only.");
    }
    @Override
    public void setData (double element, String fieldName, Object obj) {
      throw new UnsupportedOperationException("Can not write to "+getName()+" file is read-only.");
    }
    @Override
    public void insertData (double offset, Object obj) {
      throw new UnsupportedOperationException("Can not write to "+getName()+" file is read-only.");
    }
    @Override
    public void removeData (double offset, long count) {
      throw new UnsupportedOperationException("Can not write to "+getName()+" file is read-only.");
    }
    public int writeDataTable (Table tbl) {
      throw new UnsupportedOperationException("Can not write to "+getName()+" file is read-only.");
    }
    @Override
    public void update() {  // since 3.1.1
      throw new UnsupportedOperationException("Can not write to "+getName()+" file is read-only.");
    }

    @Override
    public Table getRecordDef (int i) {
      Table def = null;

      if ((i >= 0) && (i < getRecordDefCount())) {
        def = new Table();
        def.put("NAME",   "F"+(i+1));
        def.put("FORMAT", getFormat());
      }
      return def;
    }

    @Override
    public Table getRecordDefs() {
      Table tbl = new Table();

      for (int i = 0; i < getRecordDefCount(); i++) {
        tbl.put(Integer.toString(i), getRecordDef(i));
      }
      return tbl;
    }

    @Override
    public Table getDataTable (double offset) {
      double seek = seek();
      Table  tbl  = null;

      seek(offset);
      Data buf = getDataBuffer(1);
      int  num = pf.read(buf, 1);

      if (num != 1) {
        tbl = new Table();
        for (int i = 0; i < buf.getAPE(); i++) {
          tbl.put(("F"+(i+1)), buf.getAtomAt(i));
        }
      }
      seek(seek+1);
      return tbl;
    }

    @Override
    public int readDataTable (Table tbl) {
      Table row = getDataTable(seek());
      if (row == null) return -1;

      tbl.merge(row);
      return 1;
    }
  }

  /** Wraps a PlotFileNH as a read-only ListFile. */
  private static class PlotFileWrapper2 extends PlotFileWrapper {
    protected final ListFile lf;

    /** Creates a new instance. */
    PlotFileWrapper2 (PlotFile pf, ListFile lf) {
      super(pf);
      this.lf = lf;
      if (lf == null) throw new NullPointerException("lf can not be null");
      if (lf != pf  ) throw new IllegalArgumentException("lf must match pf");
    }

    @Override public double  getNumberOfRows()                   { return lf.getNumberOfRows();     }
    @Override public boolean getProtected()                      { return lf.getProtected();        }
    @Override public int     getRecordDefCount()                 { return lf.getRecordDefCount();   }
    @Override public Table   getRecordDef(int i)                 { return lf.getRecordDef(i);       }
    @Override public Table   getRecordDefs()                     { return lf.getRecordDefs();       }
    @Override public Table   getDataTable (double off)           { return lf.getDataTable(off);     }
    @Override public int     setData (double off, Table tbl)     { return lf.setData(off,tbl);      }

    @Override public void setRecordDefs (Table defs)                    { lf.setRecordDefs(defs);   }
    @Override public void removeData (double off)                       { lf.removeData(off);       }
    @Override public void insertData (double off, Object obj)           { lf.insertData(off,obj);   }
    @Override public void removeData (double off, long count)           { lf.removeData(off,count); }
    @Override public void setData (double off, String name, Object obj) { lf.setData(off,name,obj); }

    @Override
    public int writeDataTable (Table tbl) {
      double seek = seek();
      lf.setData(seek,tbl);
      seek(seek+1);
      return 0;
    }

    @Override
    public int readDataTable (Table tbl) {
      double seek = seek();
      Table row = getDataTable(seek);
      if (row == null) return -1;

      tbl.merge(row);
      seek(seek+1);
      return 1;
    }
  }
}
