package nxm.sys.lib;

import nxm.sys.inc.*;

import java.io.*;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.jar.*;
import java.util.zip.*;

/**
  A class to handle JAR files.

  @author Jay Hallam
  @version $Id: JarFile.java,v 1.13 2013/01/03 18:31:34 ntn Exp $

*/
public class JarFile extends ZipFile {
  Manifest manifest; // The JAR file manifest

  public JarFile () { }

  public JarFile (Object ref, Object filename) {
    init(ref,filename);
  }

  public String getMimeType () {
    return "application/x-java-archive";
  }

  ZipEntry[] readZipFile () throws IOException {
    String fname = FileUtil.toLocalFileName(getURL());
    java.util.jar.JarFile jarFile = new java.util.jar.JarFile(fname);
    manifest = jarFile.getManifest();
    return (fname == null)? null : readZipFile(jarFile);
  }

  ZipEntry[] readZipInputStream () throws IOException {
    JarInputStream in = new JarInputStream(io.getInputStream());
    manifest = in.getManifest();
    return readZipInputStream(in);
  }

  public String listHeader() {
    return "JarFile          : "+getURL()+"\n"
         + "Entries          : "+getNumEntries()+"\n"
         + "Bytes            : "+getSize()+"\n"
         + "Main Attributes  : "+listMainAttributes()
         + "Manifest Entries : "+listManifestEntries();
  }

  /** Used only by listHeader() */
  private String listMainAttributes () {
    StringBuilder attr = new StringBuilder();
    if (manifest == null) {
      attr.append("(not available)\n");
    } else {
      attr.append("\n");

      Iterator<Entry<Object, Object>> iter = manifest.getMainAttributes().entrySet().iterator();
      while (iter.hasNext()) {
        Map.Entry<Object,Object> entry = iter.next();
        String    key   = entry.getKey().toString();
        String    value = entry.getValue().toString();

        attr.append("  ").append(StringUtil.padRight(key,32));
        attr.append(" : ").append(value).append('\n');
      }
    }
    return attr.toString();
  }

  /** Used only by listHeader() */
  private String listManifestEntries () {
    StringBuilder ent = new StringBuilder();

    if (manifest == null) {
      ent.append("(not available)\n");
    } else {
      ent.append("(excludes SHA1-Digest)\n");

      Iterator<String> entries = manifest.getEntries().keySet().iterator();
      while (entries.hasNext()) {
        String   name    = entries.next().toString();
        boolean  anyAttr = false;
        Iterator<Entry<Object, Object>> attIter = manifest.getAttributes(name).entrySet().iterator();
        while (attIter.hasNext()) {
          Map.Entry<Object,Object> entry = attIter.next();
          String    key   = entry.getKey().toString();
          String    value = entry.getValue().toString();

          if (!key.equals("SHA1-Digest")) {
            if (!anyAttr) {
              anyAttr = true;
              ent.append("  ").append(StringUtil.padRight(name,32)).append(" :");
            }
            ent.append(' ').append(key).append('=').append(value);
          }
        }
        if (anyAttr) ent.append('\n');
      }
    }
    return ent.toString();
  }

  /** @deprecated Since NeXtMidas 2.5.0: Use {@link #getNumEntries()}. */
  @Deprecated public double getElements() { return getNumEntries(); }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Methods to implement ListFile
  //////////////////////////////////////////////////////////////////////////////////////////////////
  void notEditable () { throw new UnsupportedOperationException("JarFile is not editable."); }

  public Table getDataTable (double offset) {
    if (offset < 0) return null;
    if (offset >= zipEntries.length) return null; // end-of-file

    Table    t = super.getDataTable(offset);
    JarEntry e = (JarEntry)zipEntries[(int)offset];
    StringBuilder attr = new StringBuilder();
    String       cert = "";

    try {
      Attributes a = e.getAttributes();
      Object[]   c = e.getCertificates();
      //Object[]   s = e.getCodeSigners(); // Since Java 5

      if (a != null) {
        Iterator<Entry<Object, Object>> iter = a.entrySet().iterator();
        while (iter.hasNext()) {
          Map.Entry<Object,Object> ent = iter.next();
          if (attr.length() != 0) attr.append(", ");
          attr.append(ent.getKey()).append(":").append(ent.getValue());
        }
      }
      if (c != null) cert = StringUtil.join(c);
      //if (s != null) sign = StringUtil.join(s);  // Since Java 5
    } catch (IOException ex) {
      // ignore it (no manifest available)
    }
    t.put("ATTRIBUTES",   attr);
    t.put("CERTIFICATES", cert);
    //t.put("CODE_SIGNERS", sign);  // Since Java 5

    return t;
  }

  public int getRecordDefCount () {
    //return super.getRecordDefCount() + 3; // Since Java 5
    return super.getRecordDefCount() + 2;
  }

  public Table getRecordDef (int i) {
    Table t;
    switch (i - super.getRecordDefCount()) {
      case 0:  t = new Table();  t.put("NAME", "ATTRIBUTES");   break;
      case 1:  t = new Table();  t.put("NAME", "CERTIFICATES"); break;
      case 2:  t = new Table();  t.put("NAME", "CODE_SIGNERS"); break; // Since Java 5
      default: t = super.getRecordDef(i);                       break;
    }
    return t;
  }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Methods to create a JarFile
  //////////////////////////////////////////////////////////////////////////////////////////////////

  /** Signs a JAR file. Note that the JAR file must be on the local disk. This method is usually
      called by the <code>GENERATE SIGN</code> command, but can be called directly if needed.
      Example:
      <pre>
        FileName fname = new FileName(Shell.getOptPath("jars","myopt")+"myjar.jar");
        signJarFile(midas, fname);
      </pre>
      @param midas The current midas context.
      @param fname The name of the JAR file.
   */
  public static void signJarFile (Midas midas, FileName fname) {
    if (midas == null) { midas = Shell.getMidasContext(); }

    // JC4: While it is possible to sign a JAR file programmatically
    //     (see http://www.onjava.com/pub/a/onjava/2001/04/12/signing_jar.html),
    //     it is not clean or easy. This is due to the fact that allowing users to
    //     potentially alter security privilages at run-time is generally discouraged.
    if (!FileUtil.isLocalFileName(fname.toString())) {
      throw new MidasException("JarResource.signJar(..): Can only sign JAR files on the local "
                             + "disk; given '"+fname+"'.");
    }
    String sep  = java.io.File.separator; // path separator '/' on unix '\' on windows
    String name = FileUtil.toLocalFileName(fname.toString());

    if (name.length() == 0) {
      throw new MidasException("JarResource.signJar(..): Can only sign JAR, no file name given.");
    }

    String javahome = Shell.getProperty("JAVAHOME");
    String cmdline  = javahome+sep+"bin"+sep+"jarsigner "
                    + "-keystore "+Shell.getOptPath("sys","cfg")+"keystore "
                    + "-storepass nmsign "
                    + name + " "
                    + "nmsign";

    if (javahome == null) {
      throw new MidasException("JAVAHOME OS env not set to JDK directory, cannot run: "+cmdline);
    }

    if (Foreign.runInternal(midas,cmdline,true) != 0) {
      throw new MidasException("JarFile: Unknown error while signing JAR, filename='"+name
                             + "' command='"+cmdline+"'.");
    }
  }

  /** Creates a new jar file, but does not sign it. This is the same as
      <code>createJarFile(midas,jarFile,null,names,files,false)</code>. The signature of this
      method matches the {@link ZipFile#createZipFile(Midas,FileName,String[],FileName[],boolean)}
      method.
      @param midas   The midas context to use.
      @param jarFile The name of the jar file to create.
      @param names   The names of the files to write, as shown in the zip file. Names must use
                     forward slashes "/" not backward slashes "\" on all systems.
      @param files   The names of the files to write, as on disk. (Must be same length as names).
      @param verbose Show verbose output.
      @see ZipFile#createZipFile(Midas,FileName,String[],FileName[],boolean)
   */
  public static void createJarFile (Midas midas, FileName jarFile, String[] names, FileName[] files,
                                   boolean verbose) {
    createJarFile(midas,jarFile,null,names,files,verbose,false);
  }

  /** Creates a new jar file. The files added to the jar are listed twice using the same syntax as
      {@link ZipFile#createZipFile(Midas,FileName,String[],FileName[],boolean)}. <br>
      <br>
      <b>Important Note:</b> If you are creating a {@link Manifest} to pass in to this method, be
      sure that the "Main Attribute" {@link java.util.jar.Attributes.Name#MANIFEST_VERSION} is set
      properly, otherwise the manifest will not be written to the file (this is a limitation is
      imposed by the {@link JarOutputStream} class).
      @param midas    The midas context to use.
      @param jarFile  The name of the jar file to create.
      @param manifest The JAR file manifest to include.
      @param names    The names of the files to write, as shown in the zip file. Names must use
                      forward slashes "/" not backward slashes "\" on all systems.
      @param files    The names of the files to write, as on disk. (Must be same length as names).
      @param verbose  Show verbose output.
      @param sign     Should the jar file be signed (using <code>signJarFile(midas,jarFile)</code>)
                      after it is created?
   */
  public static void createJarFile (Midas midas, FileName jarFile, Manifest manifest, String[] names,
                                   FileName[] files, boolean verbose, boolean sign) {
    try {
      FileOutputStream     dest = new FileOutputStream(jarFile.toString());
      BufferedOutputStream bos  = new BufferedOutputStream(dest);
      JarOutputStream      out  = (manifest==null)? new JarOutputStream(bos)
                                                  : new JarOutputStream(bos,manifest);

      ZipFile.createZipFile(midas,out,names,files,true,verbose);
    } catch (Exception e) {
      throw new MidasException("JarFile: Error unable to create jar file",e);
    }
    if (sign) {
      if (verbose) {
        midas.info("Signing JAR file "+jarFile);
      }
      signJarFile(midas,jarFile);
    }
  }
}
