package nxm.sys.lib;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.jar.*;
import java.util.jar.JarFile;


/** Resource for interacting with entries in a Jar File.

    @author  Jeff Schoen
    @author  J. Czechowski IV
    @version $Id: JarResource.java,v 1.20 2012/08/22 13:37:20 ntn Exp $
 */
public class JarResource extends IOResource {
  private URL                 url; // Use of Jar/File (special syntax, see setURL(..))
  private BufferedInputStream in;  // Current input stream.

  /** Return a string containing a description of this object. */
  public String toString () {
    return "JarResource: " + getURL();
  }

  /** Sets the fully qualified URL of the resource.
      <pre>
        {@link java.net.JarURLConnection} requires a very specific format for specifying
        a file/dir in a URL. Some examples:

            jar:http://www.foo.com/bar/baz.jar!/COM/foo/Quux.class    ! Jar entry
            jar:file:/home/duke/duke.jar!/                            ! Jar File
            jar:http://www.foo.com/bar/baz.jar!/COM/foo/              ! Jar Directory
            jar:file:c:\nextmidas\jars\nxm-sys.jar!/nxm/sys/dat/      ! Jar Directory on Windows

        NeXtMidas is a little less restrictive and we try to fix the URL where possible.
        Some things we *try* to fix:
            - Omission of the '!' after the ".jar"
            - Omission of the root-level '/' (needs to have "!/")
            - Omission of "file:" for local JAR files
            - Omission of "jar:"
      </pre>
      @param urlstr The fully qualified URL of the resource.
   */
  public void setURL (String urlstr) {
    super.setURL("jar:", ".jar", new String[] { "!/" }, urlstr);

    try {
      this.url = new URL(getURL());
    } catch (MalformedURLException e) {
      throw new MidasException("JarResource: Error accessing JAR file URL="+getURL(), e);
    }
  }

  /** Gets a connection to the JAR. */
  protected JarURLConnection getConnection () {
    URLConnection conn = null;

    try {
      conn = url.openConnection();
      return (JarURLConnection)conn;

    } catch (ClassCastException e) {
      // This is extremely unlikely with a Java VM from Sun Microsystems, but could potentially
      // happen with another vendor's Java VM.
      throw new MidasException("JarResource: Unable to convert "+conn.getClass()
                             + " to JarURLConnection when given URL="+url, e);
    } catch (Exception e) {
      throw new MidasException("JarResource: Error connecting to URL="+url, e);
    }
  }

  /** Gets a entry in the JAR.
      @return JarEntry representing a JAR file entry,
              or null if unable to get JarURLConnection or JarEntry from the connection.
      @see JarURLConnection#getJarEntry()
   */
  protected JarEntry getEntry () {
    try {
      JarURLConnection conn = getConnection();
      return (conn != null)? conn.getJarEntry(): null;
    } catch (IOException e) {
      // Error in getJarEntry() or the entry does not exist (or was not specified).
      return null;
    }
  }

  /** Gets the JAR file.
      @return JarFile to read the contents of a jar file,
              or null if unable to get JarURLConnection or JarFile from the connection
      @see JarURLConnection#getJarFile()
   */
  protected java.util.jar.JarFile getFile () {
    try {
      JarURLConnection conn = getConnection();
      return (conn != null)? conn.getJarFile(): null;
    } catch (IOException e) {
      // Error in getJarFile().
      return null;
    }
  }

  public boolean exists () {
    return (getEntry() != null) || (getFile() != null);
  }

  public boolean open () {
    isOpen = false;
    try {
      JarURLConnection conn = getConnection();

      if (dir <= DIR_INOUT) { // Input
        in = new BufferedInputStream(conn.getInputStream());
      }
      if (dir >= DIR_INOUT) { // Output
        // Output to a JAR file is not as simple as opening an output stream. It requires
        // a linear write of all data (no going back to write a BLUE File header) for all
        // files with a special ZipEntry added before each file. Oh, and it can only be
        // done on local random-access files.
        throw new MidasException("JarResource: Output not supported at this time, URL="+url);
      }

      isOpen = true;

      if (offset != 0) {
        seek(0L,offset);
      }
    } catch (Exception e) {
      // Same as FileResuorce - Ignore the exception and return isOpen
    }
    return isOpen;
  }

  public boolean close () {
    boolean stat = true;

    stat  &= closeIO(in);
    offset = 0;
    isOpen = false;
    in     = null;

    return stat;
  }

  public long getLength() {
    JarEntry entry = getEntry();
    return (entry != null)? entry.getSize() : 0L;
  }

  /** Seek ahead in the inpur stream.
      @param offset Current offset.
      @param off    New offset (must be greater than offset).
   */
  private void seek (long offset, long off) {
    try {
      if (in != null) { in.skip(off-offset); }
      offset = off;
    } catch (IOException e) {
      throw new MidasException("JarResource: Unable to seek to "+off+" URL="+url,e);
    }
  }

  public boolean seek (long off) {
    if (isOpen && off!=offset) {
      // Streams are not random-access. Need to close and then reopen.
      if (off > offset) {
        seek(offset,off);
      } else {
        close();
        offset = off;
        open();
      }
    }
    return true;
  }

  public long avail () {
    if (in == null) return 0L;

    try {
      return in.available();
    } catch (IOException e) {
      return 0L;
    }
  }

  public int read (byte[] buf, int boff, int bytes) {
    int status=0, off=boff;
    while (bytes>0) {
      try {
        status = in.read(buf,off,bytes);
      } catch (IOException e) {
        status = -1;
      }
      if (status>0) { bytes-=status; off+=status; }
      else break;
    } bytes = off-boff;
    if (status<0 && bytes==0) return status;
    offset += bytes;
    return bytes;
  }

  public int write (byte[] buf, int boff, int bytes) {
    throw new MidasException("JarResource: Output not supported at this time, URL="+url);
  }

  /** Return the list of entries in the jar file that pass the given filter.*/
  public String[] getEntryList (FilenameFilter filter) {
    JarURLConnection      conn    = getConnection();
    ArrayList<String>     list    = new ArrayList<String>();
    JarFile               jarFile = getFile();
    Enumeration<JarEntry> entries = (jarFile != null) ? jarFile.entries() : null;

    // NOTE: Each entry in the JAR is listed as <path>/<filename>. For example, assume the AUX
    //       is pointing to "nxm-sys.jar/nxm/sys/dat/" then the "world.prm" will be listed as
    //       "nxm/sys/dat/world.prm". We need to exclude files in other directories and then
    //       trim the path off of the name.
    String entryName = conn.getEntryName();
    int    lenEntName;

    if (entryName == null) { // Flat ZIP file with no directory structure.
      entryName = "";
      lenEntName = 0;
    } else {
      lenEntName = entryName.length();
    }

    while ((entries != null) && entries.hasMoreElements()) {
      JarEntry entry = entries.nextElement();
      String   name  = entry.getName();

      if (name.startsWith(entryName)) {
        String fname = name.substring(lenEntName);

        if ((fname.length() > 0) && (fname.indexOf('/') < 0) && filter.accept(null, fname)) {
          list.add(fname);
        }
      }
    }

    String[] files = list.toArray(EMPTY_STRING_ARRAY);
    Arrays.sort(files);
    return files;
  }

  /** @since NeXtMidas 3.3.0 */
  @Override
  public long lastModified () {
    long utcSecs = 0;
    JarEntry jarEntry = getEntry();
    if (jarEntry != null) {
      utcSecs = jarEntry.getTime();
      if (utcSecs < 0 ) utcSecs = 0;
    }
    return utcSecs;
  }

}
