package nxm.sys.lib;

import nxm.sys.inc.*;

import java.lang.reflect.*;
import java.util.*;

/** A class to handle accessing Objects from the macro language.
    <br><br>
    This class is used by the results table to access methods and fields of objects. <br>
    <br>
    For read access, the algorithm first looks for a getNAME() method, where NAME is the key name
    for the attribute.  The search is case insensitive. If this fails, a case insensitive search
    for a field named NAME is performed. <br>
    <br>
    For write access, the algorithm first looks for a setNAME(arg) method, where NAME is the key
    name for the attribute.  The search is case insensitive. If this fails, it tries without the
    set for a NAME(arg) method. <br>
    <br>
    If these fail and the root object implements Chainable, the access request is passed on to
    the next object in the chain. <br>
    <br>
    A complete state diagram for this class is provided in the "Using KeyObject" section of the
    <i>NeXtMidas User's Guide</i>.

    @see nxm.sys.inc.Keyable
    @see nxm.sys.inc.Chainable

    @author  Jeff Schoen (Original version)
    @author  J. Czechowski IV (Revised version) / Neon Ngo
    @version $Id: KeyObject.java,v 1.101 2013/01/03 18:21:46 ntn Exp $
 */
public class KeyObject extends Object implements DataTypes {
  /** immutable zero-length Object array used by findAndInvoke method. */
  static final Object[] EMPTY_OBJECT_ARRAY = new Object[0];
  /** immutable zero-length Member array */
  private static final Member[] EMPTY_MEMBER_ARRAY = new Member[0];
  /** immutable zero-length Method array */
  private static final Method[] EMPTY_METHOD_ARRAY = new Method[0];
  /** immutable zero-length Constructor array */
  private static final Constructor<?>[] EMPTY_CONSTRUCTOR_ARRAY = new Constructor<?>[0];
  /** immutable zero-length Field array */
  private static final Field[] EMPTY_FIELD_ARRAY = new Field[0];

  /** JC4: Enables public variable warnings. */
  private static final boolean BE_ANNOYING  = false;
  /** JC4: Enables ambiguous key warnings.  */
  private static final boolean WARN_AMBIG   = true;
  /** JC4: Enables new method finder debug.  */
  private static final boolean DEBUG_FINDER = false;
  /** JC4: Enables extra debug information inside the new method finder. This prints out a TON of
           information, so use it sparingly. (This requires DEBUG_FINDER to be true.)
   */
  private static final boolean EXTRA_DEBUG  = false;
  /** NTN: Enables debug for getKey() */
  private static boolean DEBUG_GETKEY       = false;

  /** Search methods by name.             */
  private static final int NAMED_METHODS   = 0;
  /** Search accessors by parameter name. */
  private static final int NAMED_ACCESSORS = 1;
  /** Search modifiers by parameter name. */
  private static final int NAMED_MODIFIERS = 2;
  /** Search fields by name.              */
  private static final int NAMED_FIELDS    = 3;

  /** Default search order for a GET key. */
  private static final int[] SEARCH_ORDER_GET
          = { NAMED_ACCESSORS, NAMED_METHODS, NAMED_FIELDS, NAMED_MODIFIERS};

  /** Default search order for a SET key. */
  private static final int[] SEARCH_ORDER_SET
          = { NAMED_MODIFIERS, NAMED_METHODS, NAMED_FIELDS, NAMED_ACCESSORS};

  // JC4: Profiling showed that the getNamedMembers(..) method was the most time consuming method in
  //      this class. To reduce the number of calls to the getNamedMembers(..) method a set of look-
  //      up tables are used. Since the members are statically defined for a given class the entries
  //      in these tables should remain valid over time. The keys in these tables are strings in
  //      the form <className>#<keyName> (e.g. "java.lang.String#length"), the values are the member
  //      arrays.
  private static Cache accessors    = new Cache(32); // Used by getNamedAccessors(..)
  private static Cache modifiers    = new Cache(32); // Used by getNamedModifiers(..)
  private static Cache methods      = new Cache(32); // Used by getNamedMethods(..)
  private static Cache fields       = new Cache(32); // Used by getNamedFields(..)
  private static Cache constructors = new Cache(32); // Used by getConstructors(..)

  /** Flag for setKeys(..): Perform a soft set, do not error out and do not issue a warning if
                            no applicable set method is found.
   */
  public static final int SOFT=0x1;

  /** Flag for setKeys(..): *** Not Used *** */
  public static final int RESOLVE=0x2;

  /** Flag for setKeys(..): Issue a warning if no applicable set method is found. This is private
                            since it should always be combined with SOFT (see SOFT_WARN).
   */
  private static final int WARN=0x4;

  /** Flag for setKeys(..): Perform a soft set with a warning, do not error out and but issue a
                            warning if no applicable set method is found. (This includes {@link #SOFT}).
   */
  public static final int SOFTWARN=SOFT | WARN;

  /** get a keyed value from an object */
  public static Object getKey (Object root, String key) {
    return getKey(root,key,null,null);
  }

  /** Get a keyed value from an object.
      @param root  The object to get keyed field in
      @param key   The key to get
      @param value Extra arguments to pass when invoking the method to get the keyed value
      @param ref   Reference object (to get Midas context) used when getting keys
    */
  public static Object getKey (Object root, String key, Object value, Object ref) {
    if (DEBUG_GETKEY) { Shell.writeln("KeyObject.getKey(....) ENTER; root="+Util.getClassName(root)
        +" key="+key+" value="+Util.getClassName(value)+" ref="+Util.getClassName(ref));
    }
    for (int kp; (kp=key.indexOf('.'))>0; ) {  // kp = index of next dot in key
      int klp = key.indexOf('(');              // index of first left paren (open)
      // if (klp>0&&klp<kp) break;             // ORIG 2.6.x
      if (klp>0 && klp<kp) {                   // if left paren is before dot (fix for Bug 1647 & 1840)
        int lpcount = 1;                       // count of left parens including first left paren
        int rpcount = 0;                       // count of right parens after first left paren
        boolean doublequote = false;           // true if within a double quote
        boolean singlequote = false;           // true if within a single quote
        int klen = key.length();               // length of current key (different for each loop iteration)
        int ii=klp+1;                          // index of to start looking for left and right parens
        for (; ii<klen; ii++) {
          char ch = key.charAt(ii);
          if (ch == ')' && !doublequote && !singlequote) {    // right paren outside of single and/or double quotes
            rpcount++;
            if (ii+1 == kp && rpcount==lpcount) break;        // right paren just before dot & parens balanced, break from inner for loop
          }
          else if (ch=='(' && !doublequote && !singlequote) { // left paren outside of single and double quotes
            lpcount++;
          }
          else if (ch=='"')  doublequote = !doublequote;
          else if (ch=='\'') singlequote = !singlequote;
        } // end inner-for loop
        if (DEBUG_GETKEY) Shell.writeln("KeyObject.getKey(....) ==> key="+key+" klp="+klp+" kp="+kp+" key.len="+klen+" rpcount="+rpcount+" lpcount="+lpcount + " ii="+ii);
        if      (rpcount >  lpcount) throw new MidasException("Extra right-parenthesis. key="+key);
        else if (rpcount == lpcount) { // balanced left and right parens
          if (ii == klen) break;       // right paren at end of string (so break out)
                                       // otherwise continue processing
        }
        else throw new MidasException("Unbalanced parenthesis. key="+key);
      } // end if (klp>0 && klp<kp)

      String subKey = key.substring(0,kp);
      if (DEBUG_GETKEY) Shell.writeln("KeyObject.getKey(....)   calling getKey(....) subKey="+subKey+" klp="+klp+" kp="+kp);
      root=getKey(root,subKey,null,ref);
      key=key.substring(kp+1);
    } // end for loop

    if (DEBUG_GETKEY) Shell.writeln("KeyObject.getKey(....)   after for loop root="+Util.getClassName(root)+" key="+key);
    if (root==null) return null;
    Object[] args;
    boolean hasArgs = (value instanceof Object[]);
    int ip = key.indexOf('(');
    if (ip>0) {
      args = parseArgs(key,ip,key.length(),value,ref);
      key = key.substring(0,ip);
      hasArgs = true;
    } else if (hasArgs) {
      args = (Object[])value;
    } else {
      args = null;
    }
    return findAndInvoke(root, key, args, SEARCH_ORDER_GET, ref);
  }

  /** set a key in an object */
  public static Object setKey (Object root, String key, Object value) {
    return setKey(root,key,value,null);
  }

  /** set a key in an object */
  public static Object setKey (Object root, String key, Object value, Object ref) {
    for (int kp; (kp=key.indexOf('.'))>0; ) {
      int klp = key.indexOf('('); if (klp>0&&klp<kp) break;
      root=getKey(root,key.substring(0,kp),null,ref);
      key=key.substring(kp+1);
    }
    if (root==null) return null;
    if (value==null) return getKey (root,key,null,ref); // just run Method
    Object[] args;
    boolean hasArgs = (value instanceof Object[]);
    int ip = key.indexOf('(');
    if (ip>0) {
      args = parseArgs(key,ip,key.length(),value,ref);
      key = key.substring(0,ip);
      hasArgs = true;
    } else if (hasArgs) {
      args = (Object[])value;
    } else {
      args = new Object[1];
      args[0] = value;
    }
    value = findAndInvoke(root, key, args, SEARCH_ORDER_SET, ref);
    return (value == null)? args[args.length-1]: value; // <-- To appease Results.java
  }

  /** Get a member of an indexable object. <br>
      <br>
      This is used to get values out of an array or array-like object. Currently this supports the
      following array-like types:
      <ul>
        <li><i>Java Arrays</i></li>
        <li>Instances of {@link List}</li>
        <li>Instances of {@link Data}</li>
        <li>Instances of {@link KeyVector}</li>
      </ul>
      @param root  The array or array-like object.
      @param index The index into the array-like object. Note that some array-like objects are
                   one-based while the majority tend to be zero-based.
      @return An element from the array-like object.
      @throws IllegalArgumentException If unable to get the element from the array-like object.
      @see #setIndexed(Object,int,Object)
   */
  public static Object getIndexed (Object root, int index) {
    boolean notIndexable = false;
    Object  value        = null;

    try {
      // JC4: Changed use of Vector to use List since Vector is a List
      //      and List includes more types.
           if (root instanceof Data     ) { value = ((Data)root).getIndex(index);      }
      else if (root instanceof Indexable) { value = ((Indexable)root).getIndex(index); }
      else if (root instanceof List     ) { value = ((List)root).get(index);           }
      else if (root instanceof KeyVector) { value = ((KeyVector)root).get(index);      }
      else if (root.getClass().isArray()) { value = Array.get(root,index);             }
      else                                { notIndexable = true; }

    } catch (Exception e) {
      throw new IllegalArgumentException("KeyObject.getIndexed(): Error in accessing element "
                                       + index + " from " + root + ": " + e);
    }

    if (notIndexable) {
      throw new MidasException("Object is not indexable.");
    }
    return value;
  }

  /** Set a member of an indexable object. <br>
      <br>
      This is used to get values out of an array or array-like object. This supports the same object
      types as are supported by {@link #getIndexed(Object,int)}.
      @param root  The array or array-like object.
      @param index The index into the array-like object. Note that some array-like objects are
                   one-based while the majority tend to be zero-based.
      @param value The value to set.
      @return <code>value</code> (This may change in the future.)
      @throws IllegalArgumentException If unable to set the element in the array-like object.
      @see #getIndexed(Object,int)
   */
  @SuppressWarnings("unchecked") // suppress warn for (List)root
  public static Object setIndexed (Object root, int index, Object value) {
    boolean notIndexable = false;

    try {
      // JC4: Changed use of Vector to use List since Vector is a List
      //      and List includes more types.
           if (root instanceof Data     ) { ((Data)root).setIndex(index,value);      }
      else if (root instanceof Indexable) { ((Indexable)root).setIndex(index,value); }
      else if (root instanceof List     ) { ((List)root).set(index,value);           }
      else if (root instanceof KeyVector) { ((KeyVector)root).set(index,value);      }
      else if (root.getClass().isArray()) { Array.set(root,index,value);             }

    } catch (Exception e) {
      throw new IllegalArgumentException("KeyObject.setIndexed(): Error in setting element "
                                       + index + " in " + root + " to " + value + ": " + e);
    }

    if (notIndexable) {
      throw new MidasException("Object is not indexable.");
    }
    return value;
  }

  /** Get the index value associated with the named field. This is used for legacy functionality
      that used arrays with a given ordering rather than using an object. Use of this is discouraged
      since many array values do not correspond to these names and using these names with those
      arrays will give the wrong results. <br>
      <br>
      The currently supported <i>Key--&gt;Index</i> are as follows:
      <pre>
         "X"  --&gt; 0    "Y"  --&gt; 1    "Z"  --&gt; 2    "T"  --&gt; 3
         "R"  --&gt; 0    "I"  --&gt; 1
        "ALT" --&gt; 0   "LAT" --&gt; 1   "LON" --&gt; 2
      </pre>
      If the key is a number then that number is used as the index.
      @param key The named key.
      @return The index associated with that key or <code>-1</code> if none is found.
   */
  public static int getKeyIndex (String key) {
    int i=-1, kl=key.length();
    if (StringUtil.isNumber(key)) {
      i = Convert.o2l(key);
    }
    else if (kl==1) {
      char c = key.charAt(0);
      if (c>='0' && c<='9') i = (c-'0');
      else if (c=='X') i=0;
      else if (c=='Y') i=1;
      else if (c=='Z') i=2;
      else if (c=='T') i=3;
      else if (c=='R') i=0;
      else if (c=='I') i=1;
    }
    else if (kl!=3) { }
    else if (key.equals("ALT")) i=0;
    else if (key.equals("LAT")) i=1;
    else if (key.equals("LON")) i=2;
    return i;
  }

  /** Sets keys in an object from a table of key=value pairs. This is identical to
      <tt>setKeys(root,tab,null,0)</tt>.
      @param root  The object to set fields in.
      @param tab   The table of sets.
      @return Number of keys set. Returns -1 if root or tab is null.
      @see #setKeys(Object, Table, Object, int)
   */
  public static int setKeys (Object root, Table tab) {
    return setKeys(root,tab,null,0);
  }

  /** sets keys in an object from a table of key=value pairs
      @param root  The object to set fields in.
      @param tab   The table of sets.
      @param ref   Reference used when setting keys.
      @param flags Optional flags to control how sets are applied (0=Throw exception on error,
                   {@link #SOFT}=Ignore all errors (no warning), {@link #SOFTWARN}=Warn on error
                   but continue on).
      @return Number of keys set. Returns -1 if root or tab is null. Returns 0 if tab is empty.
   */
  public static int setKeys (Object root, Table tab, Object ref, int flags) {
    if ((root==null) || (tab==null) || (tab.size()==0)) return -1;
    if (tab.size()==0) return 0; // optimization since 3.1.0
    int nkey=0;
    for (Table.Iterator ti=tab.iterator(); ti.getNext(); ) {
      String key = ti.key;
      Object xarg = ti.value;
      if (xarg instanceof Table) {
        try {
          setKey (root,key,xarg,ref); nkey++; continue;  // nkey++ and continue added in 3.1.1 to prevent double adding (BUG 2563)
        } catch (MidasException ex) {
          Object troot=null;
          try { troot = getKey(root,key); } catch (MidasException e) { }
          if (troot!=null) { setKeys(troot,(Table)xarg,ref,flags); continue; }
        }
      }
      if (xarg instanceof String) xarg = Convert.s2o((String)xarg,'_',ref);
      if ((flags&SOFT)==0) {
        setKey(root,key,xarg,ref); nkey++;
      } else {
        try {
          setKey(root,key,xarg,ref); nkey++;
        } catch (MidasException e) {
          if ((flags&WARN) != 0) {
            ref2Midas(ref).printStackTrace("KeyObject: Can not set "+key+"="+xarg,e);
          }
        }
      }
    }
    return nkey;
  }

  /** Converts a reference to a Midas instance, never returns null. */
  private static Midas ref2Midas(Object ref) {
    Midas midas = Convert.ref2Midas(ref);
    return (midas!=null)? midas: Shell.getMidasContext();
  }

  /** Parses an arguments string into a parameters array
      Since 3.1.0 this method is private, this was deprecated in 2.7.0 as this is for internal use only.
      @param argstr The argument string
      @param i1 Start offset into argstr INCLUSIVE (0..len-1)
      @param i2 End offset into argstr EXCLUSIVE (1..len)
      @param xarg eXtra arguments to be added
      @param ref Midas reference
      @return Arguments as an array of Objects
   */
  private static Object[] parseArgs (String argstr, int i1, int i2, Object xarg, Object ref) {
// jph -- 23AUG2004 - This method appears to be a "quick and dirty" way to parse
// arguments... it probably should be deprecated/replaced.  I added the ability
// for it to handle QUOTES but it may have problems with other more complicated
// arg patterns

    int i,nargs=0;
    // Count the number of arguments
    int quotes = 0;
    for (i=++i1,i2--; i<i2; i++) {
      if (argstr.charAt(i) == '"') quotes = 1 - quotes;
      // Ignore commas inside quotes...
      if (quotes == 0) {
        i=argstr.indexOf(',',i); if (i<0) i=i2; nargs++;
      }
    }
    if (xarg!=null) nargs++;
//System.out.println("ParseArgs: '"+argstr+"' nargs "+nargs);
    Object[] args = new Object[nargs];
    int iQuote = -1;
    int iQuoteStart = -1;
    for (nargs=0; i1<i2; i1=i+1) {
      // Is the current character a QUOTE?
      if (argstr.charAt(i1) == '"') { iQuote = i1; }
      else { iQuote = -1; }

//System.out.println("i1 "+i1+" char "+argstr.charAt(i1)+" i "+i+" iQuote "+iQuote);
      if (iQuote>=0) { if (iQuoteStart < 0) iQuoteStart = iQuote; }
      if (iQuoteStart < 0) {
        i=argstr.indexOf(',',i1); if (i<0) i=i2;
//System.out.println("parseArgs: convert s20 on arg # "+(nargs+1)+" = '"
//                  +argstr.substring(i1,i)+"'");
        args[nargs++] = Convert.s2o(argstr.substring(i1,i),'_',ref);
      } else {
        // Look for second QUOTE
        i=argstr.indexOf('"',iQuoteStart+1); if (i<0) i=i2; else i=i+1;
//System.out.println("parseArgs: convert s20 on QUOTED arg # "+(nargs+1)+" = '"
//                  +argstr.substring(iQuoteStart,i)+"'");
        args[nargs++] = Convert.s2o(argstr.substring(iQuoteStart,i),'_',ref);
        iQuoteStart = -1;
      }
    }
    if (xarg!=null) args[nargs++]=xarg;
//for (i=0;i<nargs;i++) System.out.println("Arg"+i+"="+args[i]);
    return args;
  }

  public static final int KEY_EXACT=0,KEY_CAPS=1,KEY_ABBR=2,KEY_NONE=3;
  public static final int ARG_EXACT=0,ARG_NUM=1,ARG_NUMO=2,ARG_CVT=3,ARG_NONE=4;

  public static Object invokeStaticMethod (String className, String methodName, Object[] args) {
    try {
      Class<?> clas = Class.forName(className);
      Method methods[] = clas.getMethods();
      for (Method method : methods) {
        String name = method.getName();
        if (!name.equals(methodName)) continue;
        Class<?>[] params = method.getParameterTypes();
        if (params.length!=args.length) continue;
        boolean okArgs = true;
        for (int j=0; okArgs && j<params.length; j++) {
          String pname = params[j].getName();
          Class<?> aclass = args[j].getClass();
          String aname = aclass.getName();
          if (aname.equals(pname)) continue;
          okArgs=false;
          Class<?> sclass = aclass.getSuperclass();
          while (!okArgs && sclass != null) {
            okArgs = sclass.getName().equals(pname);
            sclass = sclass.getSuperclass();
          }
          if (okArgs) continue;
          Class<?>[] classes = aclass.getInterfaces();
          for (int k=0; classes!=null && k<classes.length && !okArgs; k++) {
            okArgs = classes[k].getName().equals(pname);
          }
        }
        if (okArgs) return method.invoke (null,args);
      }
    } catch (Exception e) {
      throw new MidasException("invokeStaticMethod: "+e);
    }
    throw new MidasException("No match to "+className+"."+methodName+" with "+args);
  }


  /** Invoke an instance method on an object and return the result. This method is very useful when
      you want to use classes that you don't want to import or otherwise reference explicitly.
      Examples include using the JNLP and netscape classes. In JDK1.4 the java.beans.Statement
      object can be used in place of this method.
      @param obj      The object to invoke the method on.
      @param methName The name of the static method to invoke.
      @param args     An array of objects to pass to the method.  If the method
                      takes no arguments, this can be null.
      @return The return value from invoking the method.
   */
  public static Object invokeInstanceMethod (Object obj, String methName, Object[] args)
                                            throws IllegalAccessException, NoSuchMethodException,
                                                   InvocationTargetException  {

    int nargs = (args != null) ? args.length : 0;

    Class<?>[] argTypes = new Class<?>[ nargs ];
    for (int i = 0; i < nargs; i++) {
      argTypes[i] = args[i].getClass();
    }

    Method meth = obj.getClass().getMethod( methName, argTypes );
    return meth.invoke( obj, args );
  }

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

  /** Selects the best method form a set of available methods to which a given set of arguments will
      be applied. This method tries to maintain the spirit of section 15.12.2 and 5.1.2 of <i>The
      Java Language Specification, Second Edition</i> (hereafter <i>JLS</i>), but has several
      allowances for NeXtMidas-specific (and Midas-specific) conversions.
      @param methods         The members methods available (all of which should have the same method name).
      @param name            The most specific member name to use.
      @param types           The types of the arguments (classes) to be applied to the methods.
      @param allowBoxing     Indicates if boxing or primitives parameters is allowed.
      @param allowConversion Indicates if the conversion of primitive types is allowed for any
                             widening conversions.
      @param allowMidasConv  Indicates if Midas-specific conversions are permitted.
      @return The best method for the given types.
   */
  private static Member getBestMember(Member[] methods, String name, Class[] types,
                                      boolean allowBoxing,
                                      boolean allowConversion, boolean allowMidasConv) {
    Member[] goodMethods = getGoodMembers(methods, types, false, false, false);

    if ((goodMethods.length == 0) && allowBoxing) {
      goodMethods = getGoodMembers(methods, types, true, false, false);
    }
    if ((goodMethods.length == 0) && allowConversion) {
      goodMethods = getGoodMembers(methods, types, false, true, false);
    }
    if ((goodMethods.length == 0) && allowBoxing && allowConversion) {
      goodMethods = getGoodMembers(methods, types, true, true, false);
    }
    if ((goodMethods.length == 0) && allowMidasConv) {
      goodMethods = getGoodMembers(methods, types, true, true, true);
    }

    return getMostSpecificMember(name, goodMethods);
  }

  /** Chooses the good methods (those both applicable and accessible) from a set of methods.
      This should only be called from getBestMethod(..).
      @param members         The members available (all of which should have the same method name).
      @param types           The types of the arguments to be applied to the methods.
      @param allowBoxing     Indicates if boxing or primitives parameters is allowed.
      @param allowConversion Indicates if the conversion of primitive types is allowed for any
                             widening conversions.
      @param allowMidasConv  Indicates if Midas-specific conversions are permitted.
      @return The methods that are both applicable and accessible. This may be an array of length
              zero if none are found.
   */
  private static Member[] getGoodMembers(Member[] members, Class[] types, boolean allowBoxing,
                                         boolean allowConversion, boolean allowMidasConv) {
    Member[] retVal = members;

    if (types != null) {
      ArrayList<Member> goodMethods = new ArrayList<Member>();
      Class<?>[] paramTypes;

      for (Member member : members) {
        if (member instanceof Method) {
          paramTypes = ((Method)member).getParameterTypes();
        } else if (member instanceof Constructor){
          paramTypes = ((Constructor<?>)member).getParameterTypes();
        } else if (member instanceof Field) {
          paramTypes = new Class<?>[] { ((Field)member).getType() };
        } else {
          paramTypes = new Class<?>[0];
        }

        if (types.length == paramTypes.length) {
          boolean match = true;

          for (int t = 0; (match) && (t < types.length); t++) {
            if (types[t] != null) {
              match = match & isMatchingType(paramTypes[t], types[t],
                                             allowBoxing, allowConversion, allowMidasConv);
            }
          }
          if (match) {
            goodMethods.add(member);
          }
        }
      }

      if (members instanceof Method[]) {
        retVal = goodMethods.toArray(EMPTY_METHOD_ARRAY);
      } else {
        retVal = goodMethods.toArray(EMPTY_CONSTRUCTOR_ARRAY);
      }
    }

    return retVal;
  }

  /** Chooses the good methods (those both applicable and accessible) from a set of methods.
      This should only be called from getBestMethod(..).
      @param fields          The fields available (all of which should have the same method name).
      @param types           The types of the arguments to be applied to the methods.
      @param allowBoxing     Indicates if boxing or primitives parameters is allowed.
      @param allowConversion Indicates if the conversion of primitive types is allowed for any
                             widening conversions.
      @param allowMidasConv  Indicates if Midas-specific conversions are permitted.
      @return The fields that are both applicable and accessible. This may be an array of length
              zero if none are found.
   */
  private static Field[] getGoodFields (Field[] fields, Class[] types, boolean allowBoxing,
                                        boolean allowConversion, boolean allowMidasConv) {
    Field[] retVal = fields;

    if (types != null) {
      ArrayList<Field> goodFields = new ArrayList<Field>();

      for (Field field : fields) {
        Class<?> type = field.getType();
        boolean match = true;

        if (types.length == 0) {
          match = true; // true for a get

        } else if (types.length == 1) {

          if (Modifier.isFinal(field.getModifiers())) {
            match = false; // Can't set a final value.

          } else if (types[0] != null) {
            match = isMatchingType(type, types[0], allowBoxing, allowConversion, allowMidasConv);
          }

        } else {
          match = false; // not a get or a set
        }

        if (match) {
          goodFields.add(field);
        }
      }

      retVal = goodFields.toArray(EMPTY_FIELD_ARRAY);
    }

    return retVal;
  }

  /** Indicates if two types are matching. This should only be called from getGoodMethods(..).
      @param givenType       The given type.
      @param sampleType      The type to test for a match with the given type.
      @param allowBoxing     Indicates if boxing or primitives parameters is allowed.
      @param allowConversion Indicates if the conversion of primitive types is allowed for any
                             widening conversions.
      @param allowMidasConv  Indicates if Midas-specific conversions are permitted.
      @return An indication of whether or not they match.
   */
  private static boolean isMatchingType(Class<?> givenType, Class<?> sampleType, boolean allowBoxing,
                                        boolean allowConversion, boolean allowMidasConv) {
    boolean match = givenType.isAssignableFrom(sampleType);

    if (!match && allowBoxing && (givenType.isPrimitive() || sampleType.isPrimitive())) {

      match = isMatchingType(getBoxedType(givenType), getBoxedType(sampleType), false,
                             allowConversion, allowMidasConv);
    }

    if (!match && allowConversion) {
      if (Number.class.isAssignableFrom(givenType)) {
        match = Number.class.isAssignableFrom(sampleType) || sampleType.equals(Character.class);
      }

      if (givenType.equals(Character.class)) {
        match = match || Number.class.isAssignableFrom(sampleType);
      }
    }

    if (!match && allowMidasConv) {
      match = givenType.equals(String.class) || givenType.equals(Data.class);

      if (sampleType.equals(String.class)) {
        match = match || Number.class.isAssignableFrom(givenType)
                      || Boolean.class.isAssignableFrom(givenType)
                      || Character.class.isAssignableFrom(givenType);
      }

      if (!sampleType.isPrimitive()) {
        match = match || givenType.equals(Table.class);
      }
    }

    return match;
  }

  /** Chooses the most specific method from a set of methods. This method follows the rules in
      <i>JLS</i> 15.12.2.2. This should only be called from getBestMethod(..).
      @param name    The name as entered.
      @param members The available Members.
      @return The most specific Member.
   */
  private static Member getMostSpecificMember(String name, Member[] members) {
    // System.out.println("DEBUG: getMostSpecificMember for ["+name+"] =\n"+StringUtil.toStringBuilder(members," "));
    Member mostSpecific = null;
    if (members.length > 0) {
      mostSpecific = members[0];

      for (int i = 1; i < members.length; i++) {
        boolean enableError = (i+1 == members.length); // since 3.3.0: enable error on last entry (BUG 2570)
        if (mostSpecific.equals(members[i])) {
          // do nothing - both are the same
        }
        else if (isMoreSpecificName(name, mostSpecific, members[i], true, enableError)) {
          mostSpecific = members[i];
        }
        else if (isMoreSpecificName(name, members[i], mostSpecific, true, enableError)) { // since 3.1.1:
          // do nothing - current 'mostSpecific' has a more specific name than members[i] (BUG 2570)
        }
        else if (isMoreSpecific(mostSpecific, members[i])) {
          mostSpecific = members[i];
        }
        else if (isMoreSpecific(members[i], mostSpecific)) {
          // do nothing - mostSpecific is still correct
        }
        else {
          Shell.warning("KeyObject: Ambiguous member reference: key '"+name+"' "
                      + "matches\n '"+mostSpecific+"' and\n '"+members[i]+"'\n using '"
                      + mostSpecific+"'.");
        }
      }
    }
    return mostSpecific;
  }

  /** Indicates if the second method is more specific than the first. This method follows the
      rules in <i>JLS</i> 15.12.2.2. This should only be called from getMostSpecificMember(..).
      @param first  The first method.
      @param second The second method.
      @return <code>true</code> is the second is more specific, <code>false</code> if the first is.
   */
  private static boolean isMoreSpecific(Member first, Member second) {
    Class<?>[] firstParams  = null;
    Class<?>[] secondParams = null;

    if (first instanceof Method) {
      firstParams  = ((Method)first).getParameterTypes();
      secondParams = ((Method)second).getParameterTypes();
    } else {
      firstParams  = ((Constructor)first).getParameterTypes();
      secondParams = ((Constructor)second).getParameterTypes();
    }

    if (firstParams.length != secondParams.length) {
      throw new RuntimeException("KeyObject.isMoreSpecific(Member,Member): array length error.");
    }

    /* The two tests must pass -- see JLS 15.2.2 */
    boolean classConv = first.getDeclaringClass().isAssignableFrom(second.getDeclaringClass());
    boolean paramConv = true;

    for (int i = 0; i < firstParams.length; i++) {
      paramConv = paramConv && firstParams[i].isAssignableFrom(secondParams[i]);
    }

    return classConv && paramConv;
  }

  /** Indicates if the second method is more specific than the first. This method follows the
      rules in <i>JLS</i> 15.12.2.2. This should only be called from getMostSpecificMethod(..).
      @param name The member name.
      @param firstMethod  The first member.
      @param secondMethod The second member.
      @param enableWarn   enable warning message for "Ambiguous member definition" (since 3.1.1 for BUG 2570)
      @param enableError  enable error for when neither firstMethod or secondMethod matches (since 3.3.0 for BUG 2570)
      @return <code>true</code> is the second is more specific, <code>false</code> if the first is.
   */
  private static boolean isMoreSpecificName (String name, Member firstMethod, Member secondMethod,
                                             boolean enableWarn, boolean enableError) {
    String first    = firstMethod.getName();
    String second   = secondMethod.getName();
    String nameUC   = name.toUpperCase();
    String firstUC  = first.toUpperCase();
    String secondUC = second.toUpperCase();

    boolean useSecond = false;
    boolean warn      = false;
    boolean error     = false;

    boolean bothIdentical    = first.equals(second);
    boolean firstMatchExact  = first.equals(name);
    boolean secondMatchExact = second.equals(name);
    boolean firstMatchUC     = firstUC.equals(nameUC);
    boolean secondMatchUC    = secondUC.equals(nameUC);
    boolean firstMatchGIS    = firstUC.equals("GET"+nameUC)  || firstUC.equals("IS"+nameUC)
                                                             || firstUC.equals("SET"+nameUC);
    boolean secondMatchGIS   = secondUC.equals("GET"+nameUC) || secondUC.equals("IS"+nameUC)
                                                             || secondUC.equals("SET"+nameUC);

         if (bothIdentical   ) { useSecond=false;                         } // Overloaded method name
    else if (firstMatchExact ) { useSecond=false;                         }
    else if (secondMatchExact) { useSecond=true;                          }
    else if (firstMatchUC    ) { useSecond=false; warn  = secondMatchUC;  } // Error if both UC versions match
    else if (secondMatchUC   ) { useSecond=true;                          }
    else if (firstMatchGIS   ) { useSecond=false; error = secondMatchGIS; } // Error if both get/is/set versions match
    else if (secondMatchGIS  ) { useSecond=true;                          }
    else                       {    error = enableError && true;          } // Error different names (usually
                                                                            // abbreviation/acronym matches
                                                                            // two different method names)
    if ((error || warn) && WARN_AMBIG) {
      String msg = "KeyObject: Ambiguous member definition: name '"+name+"' matches\n '"
                 + firstMethod+"' and\n '"+secondMethod+"'";

      if (error) {
        throw new MidasException(msg);
      } else if (enableWarn) { // since 3.1.1: check enableWarn argument to prevent warning twice (BUG 2570)
        Shell.warning(msg+",\n using "+firstMethod);
      }
    }

    return useSecond;
  }

  /** Gets the class for a given object. The class for the given object is determined as follows:
      <pre>
        (1) If obj==null then null will be returned.
        (2) If obj is an instance of Data...
               (2a) ...and the data is an ASCII value then String will be returned.
               (2b) ...and the data is a numerical scalar then the primitive type for the
                    data value will be returned (e.g. int).
               (2c) ...and the data is a numerical array then an array of the primitive type
                    for the data value will be returned (e.g. "VD" = double[]).
        (3) For all others, obj.getClass() will be returned.
      </pre>
      @param obj The object to get the class for.
      @return The object's class.
   */
  private static Class<?> getClassOf (Object obj) {
    if (obj == null) { return null; }
    if (obj instanceof Data) {
      Data data = (Data)obj;
      int  spa  = data.getSPA();

      if (data.getFormat().charAt(1) == 'A') { return String.class; }

      switch (data.getFormat().charAt(1)) {
        case 'A': return String.class;
        case 'D': return (spa == 1)? Double.TYPE  : (new double[spa]).getClass();
        case 'F': return (spa == 1)? Float.TYPE   : (new float[spa] ).getClass();
        case 'X': return (spa == 1)? Long.TYPE    : (new long[spa]  ).getClass();
        case 'L': return (spa == 1)? Integer.TYPE : (new int[spa]   ).getClass();
        case 'I': return (spa == 1)? Short.TYPE   : (new short[spa] ).getClass();
        case 'B': return (spa == 1)? Byte.TYPE    : (new byte[spa]  ).getClass();
      }
    }
    return obj.getClass();
  }

  /** Gets the class for each of a given set of objects.
      @param objs The Object array to get the class for each item.
      @return The object's class(es).
   */
  private static Class<?>[] getClassOf (Object[] objs) {
    Class<?>[] classes = new Class<?>[objs.length];

    for (int i = 0; i < classes.length; i++) {
      classes[i] = getClassOf(objs[i]);
    }

    return classes;
  }

  /** Converts a set of parameters to the appropriate types for invocation on a method.
      @param types  The types of the parameters the method takes in.
      @param params The parameters for the method, but not necessarily in the right form.
      @return The parameters to pass to the method.
   */
  private static Object[] convertParams (Class<?>[] types, Object[] params) {
    Object[] newParams = new Object[params.length];

    for (int i = 0; i < params.length; i++) {
      newParams[i] = convertParam(types[i], params[i]);

      if ((newParams[i] == null) && (params[i] != null)) {
        Shell.warning("KeyObject: Error converting "+params[i]+" to type "+types[i].getName()+".");
      }
    }

    return newParams;
  }

  /** Converts a parameter to the appropriate type for invocation on a method.
      @param type  The type of the parameters the method takes in.
      @param param The parameter for the method, but not necessarily in the right form.
      @return The parameters to pass to the method, returns <code>null</code> on error or if the
              parameter was <code>null</code>.
   */
  private static Object convertParam (Class<?> type, Object param) {
    if (type.isPrimitive()) { type = getBoxedType(type); } /* Long.TYPE -> Long.class */
    Object newParam = null;

    if ((param == null) || type.isInstance(param)) {
      newParam = param; // No conversion required

    } else if (type.isArray() && param.getClass().isArray()) {
      Object[] array = new Object[Array.getLength(param)];
      Class<?> clazz;

      try {
        clazz = Class.forName(type.getName().substring(1)); /* int[] -> int */
      } catch(ClassNotFoundException e) {
        throw new IllegalArgumentException("KeyObject: Unable to instantiate class "
                                         + type.getName().substring(1) + ": " + e);
      }

      for (int i = 0; i < Array.getLength(param); i++) {
        array[i] = convertParam(clazz, Array.get(param, i));
      }

      newParam = array;

    } else if (Number.class.isAssignableFrom(type) && (param instanceof Character)) {
      char charVal = ((Character)param).charValue();

      newParam = convertParam(type, new Integer(charVal));

    } else {
           if (type.equals(Double.class))    { newParam = new Double( Convert.o2d(param)); }
      else if (type.equals(Float.class))     { newParam = new Float(  Convert.o2f(param)); }
      else if (type.equals(Long.class))      { newParam = new Long(   Convert.o2x(param)); }
      else if (type.equals(Integer.class))   { newParam = new Integer(Convert.o2l(param)); }
      else if (type.equals(Short.class))     { newParam = new Short(  Convert.o2i(param)); }
      else if (type.equals(Byte.class))      { newParam = new Byte(   Convert.o2b(param)); }

      else if (type.equals(Character.class)) { newParam = new Character((char)Convert.o2l(param));  }
      else if (type.equals(Boolean.class))   { newParam = new Boolean(StringUtil.isTrue(""+param)); }

      else if (type.equals(String.class))    { newParam = Convert.o2s(param);  }
      else if (type.equals(Data.class))      { newParam = new Data(param);     }
      else if (type.equals(Table.class))     { newParam = Convert.o2t(param);  }
    }

    return newParam;
  }

  /** Gets the boxed type for a primitive.
      @param type The primitive type.
      @return The boxed type.
   */
  private static Class<?> getBoxedType (Class<?> type) {
    if (type.isPrimitive()) {
      if (type.equals(Double.TYPE))    { return Double.class;    }
      if (type.equals(Float.TYPE))     { return Float.class;     }
      if (type.equals(Long.TYPE))      { return Long.class;      }
      if (type.equals(Integer.TYPE))   { return Integer.class;   }
      if (type.equals(Short.TYPE))     { return Short.class;     }
      if (type.equals(Byte.TYPE))      { return Byte.class;      }
      if (type.equals(Character.TYPE)) { return Character.class; }
      if (type.equals(Boolean.TYPE))   { return Boolean.class;   }
      if (type.equals(Void.TYPE))      { return Void.class;      }

      throw new RuntimeException("KeyObject: Unknown Java type " + type.getName() + " (only those "
                               + "types supported in Java 1.4 are supported at this time, enums "
                               + "and generics are not yet supported).");
    } else {
      return type;
    }
  }

  /** Gets a list of public accessors matching a given name for a class.
      @param clazz      The class.
      @param name       The name of the field.
      @param staticOnly Only get static members.
      @return All matching methods.
   */
  private static Method[] getNamedAccessors (Class<?> clazz, String name, boolean staticOnly) {
    String   tag    = (staticOnly)? "#static#": "#";
    String   key    = clazz.getName() + tag + name;
    Method[] retVal = (Method[])accessors.get(key);

    if (retVal == null) {
      Method[] methods = getMethods(clazz, false);

      Vector<Member> matches = getNamedMembers(methods,"is",name,true,true,true);
      matches.addAll(getNamedMembers(methods,"get",name,true,true,true));

      retVal = matches.toArray(EMPTY_METHOD_ARRAY);
    }
    accessors.put(key, retVal); // Always to the put so it will be marked as most recent.
    return retVal;
  }

  /** Gets a list of public modifiers matching a given name for a class.
      @param clazz      The class.
      @param name       The name of the field.
      @param staticOnly Only get static members.
      @return All matching methods.
   */
  private static Method[] getNamedModifiers (Class<?> clazz, String name, boolean staticOnly) {
    String   tag    = (staticOnly)? "#static#": "#";
    String   key    = clazz.getName() + tag + name;
    Method[] retVal = (Method[])modifiers.get(key);

    if (retVal == null) {
      Method[] methods = getMethods(clazz, false);

      retVal = getNamedMembers(methods,"set",name,true,true,true).toArray(EMPTY_METHOD_ARRAY);
    }
    modifiers.put(key, retVal); // Always to the put so it will be marked as most recent.
    return retVal;
  }

  /** Gets a list of public methods matching a given name for a class.
      @param clazz      The class.
      @param name       The name of the field.
      @param staticOnly Only get static members.
      @return All matching methods.
   */
  private static Method[] getNamedMethods (Class<?> clazz, String name, boolean staticOnly) {
    String   tag    = (staticOnly)? "#static#": "#";
    String   key    = clazz.getName() + tag + name;
    Method[] retVal = (Method[])methods.get(key);

    if (retVal == null) {
      Method[] methods = getMethods(clazz, staticOnly);
      retVal = getNamedMembers(methods,"",name,true,false,false).toArray(EMPTY_METHOD_ARRAY);
    }
    methods.put(key, retVal); // Always to the put so it will be marked as most recent.
    return retVal;
  }

  /** Gets a list of all public constructors for a class.
      @param clazz The class to get the constructors for.
      @return All public constructors
   */
  private static Constructor<?>[] getConstructors (Class<?> clazz) {
    String           key    = clazz.getName() + "#<init>";
    Constructor<?>[] retVal = (Constructor[])constructors.get(key);

    if (retVal == null) {
      if (Modifier.isPublic(clazz.getModifiers())) {
        retVal = clazz.getConstructors();
      } else {
        retVal = EMPTY_CONSTRUCTOR_ARRAY;
      }
    }
    constructors.put(key, retVal); // Always to the put so it will be marked as most recent.
    return retVal;
  }

  /** Gets a list of public fields matching a given name for a class.
      @param clazz      The class.
      @param name       The name of the field.
      @param staticOnly Only get static members.
      @return All matching fields.
   */
  private static Field[] getNamedFields (Class<?> clazz, String name, boolean staticOnly) {
    String   tag    = (staticOnly)? "#static#": "#";
    String   key    = clazz.getName() + tag + name;
    Field[]  retVal = (Field[])fields.get(key);

    if (retVal == null) {
      Field[] fields = getFields(clazz, staticOnly);

      retVal = getNamedMembers(fields,"",name,true,false,false).toArray(EMPTY_FIELD_ARRAY);
    }
    fields.put(key, retVal); // Always to the put so it will be marked as most recent.
    return retVal;
  }

  /** Gets the public methods in a class. If the class is not public then get the public methods
      form the superclass or implemented interfaces.
      @param clazz      The class.
      @param staticOnly Only get static members.
      @return The methods.
   */
  private static Method[] getMethods (Class<?> clazz, boolean staticOnly) {
    ArrayList<Method> members = new ArrayList<Method>();
    Method[] methods = null;

    if (Modifier.isPublic(clazz.getModifiers())) {
      members.addAll(Arrays.asList(clazz.getMethods()));

    } else {
      Class<?>   superclass = clazz.getSuperclass();
      Class<?>[] interfaces = clazz.getInterfaces();

      if (superclass != null) {
        members.addAll(Arrays.asList(getMethods(superclass, staticOnly)));
      }
      for (int i = 0; i < interfaces.length; i++) {
        members.addAll(Arrays.asList(getMethods(interfaces[i], staticOnly)));
      }
      methods = members.toArray(EMPTY_METHOD_ARRAY);
    }

    if (staticOnly) {
      ArrayList<Method> staticMembers = new ArrayList<Method>();
      for (Method method : members) {
        if (Modifier.isStatic(method.getModifiers())) {
          staticMembers.add(method);
        }
      }
      methods = staticMembers.toArray(EMPTY_METHOD_ARRAY);
    } else {
      methods = members.toArray(EMPTY_METHOD_ARRAY);
    }

    return methods;
  }

  /** Gets the public fields in a class. If the class is not public then get the public fields
      form the superclass or implemented interfaces.
      @param clazz      The class.
      @param staticOnly Only get static members.
      @return The methods.
   */
  private static Field[] getFields (Class<?> clazz, boolean staticOnly) {
    Field[] fields  = null;
    ArrayList<Field> fieldsList = new ArrayList<Field>();

    if (Modifier.isPublic(clazz.getModifiers())) {
      fieldsList.addAll(Arrays.asList(clazz.getFields()));

    } else {
      Class<?>   superclass = clazz.getSuperclass();
      Class<?>[] interfaces = clazz.getInterfaces();

      if (superclass != null) {
        fieldsList.addAll(Arrays.asList(getFields(superclass, staticOnly)));
      }
      for (int i = 0; i < interfaces.length; i++) {
        fieldsList.addAll(Arrays.asList(getFields(interfaces[i], staticOnly)));
      }
      fields = fieldsList.toArray(EMPTY_FIELD_ARRAY);
    }

    if (staticOnly) {
      ArrayList<Field> staticFields = new ArrayList<Field>();
      for (int i = 0; i < fieldsList.size(); i++) {
        if (Modifier.isStatic(((Member)fieldsList.get(i)).getModifiers())) {
          staticFields.add(fieldsList.get(i));
        }
      }
      fields = staticFields.toArray(EMPTY_FIELD_ARRAY);
    } else {
      fields = fieldsList.toArray(EMPTY_FIELD_ARRAY);
    }

    return fields;
  }

  /** Gets a list of the members that match a given name. This should only be used by
      getNamedAccessors(Object,String), getNamedModifiers(Object,String),
      getNamedMethods(Object,String)}, and getNamedFields(Object,String).
      @param members     All possible members.
      @param prefix      The prefix that is on the member name (can be the empty string).
      @param name        The name of the member less the prefix.
      @param ignoreCase  Indicates if the tests are case-insensitive.
      @param allowAbbr   Indicates if the name may be considered to be an abbreviation
                         for a member name while comparing.
      @param allowAccronym Indicates if the name may be considered to be an acronym
                           for a member name while comparing.
      @return The set of conforming members.
   */
  private static Vector<Member> getNamedMembers (Member[] members, String prefix, String name,
                                         boolean ignoreCase, boolean allowAbbr, boolean allowAccronym) {
    Vector<Member>  matches    = new Vector<Member>(); // The matching methods.
    Vector<Member>  matchesUC  = new Vector<Member>(); // Case-insensitive matches.
    Vector<Member>  shortAbbr  = new Vector<Member>(); // Matching abbreviations (short).
    Vector<Member>  abbr       = new Vector<Member>(); // Matching abbreviations (full).
    String  fullName   = name;
    if (prefix != null &&  prefix.length() > 0) {
       // Add prefix only if given and upcase first letter match Java bean syntax
      fullName = prefix + name.substring(0,1).toUpperCase() + name.substring(1);
    }
    String  fullNameUC = fullName.toUpperCase();
    boolean abbrOk     = allowAbbr     && (name.length() > 1) && (name.length() <= 4);
    boolean accrOk     = allowAccronym && (name.length() > 1) && (name.length() <  4);
    boolean hasPrefix  = (prefix.length() > 0);

    for (Member member: members) {
      if (Modifier.isPublic(member.getModifiers())) {
        String memberName   = member.getName();
        String memberNameUC = memberName.toUpperCase();
        String accronym     = StringUtil.getAcronym(memberName);

        if (memberName.startsWith(prefix)) {
          if (memberName.equals(fullName)) {                              // Exact match
            matches.add(member);
          } else if (memberName.equalsIgnoreCase(fullName)) {             // Case-insensitive match
            matchesUC.add(member);
          } else if (abbrOk && memberNameUC.startsWith(fullNameUC)) {     // Abbreviation match
            if ((accronym.length() == 1) && hasPrefix) {
              shortAbbr.add(member);
            } else {
              abbr.add(member);
            }
          } else if (accrOk && accronym.equalsIgnoreCase(name)) {         // Acronym match
            matches.add(member);
          }
        }
      }
    }
    Vector<Member> retVal = matches;

    if ((retVal.size() == 0) && ignoreCase   ) { retVal = matchesUC; }
    if ((retVal.size() == 0) && allowAbbr    ) { retVal = shortAbbr; }
    if ((retVal.size() == 0) && allowAbbr    ) { retVal = abbr;      }
    //if ((retVal.size() == 0) && allowAccronym) { retVal = accr;      }

    return retVal;
  }

  /** Generate a warning message that includes the signature for the required method.
      @param clazz      The class the filed is in.
      @param field      The deprecated field.
      @param isModifier Indicates if a modifier is required and not an accessor.
      @return The message to output.
   */
  private static StringBuilder getDeprecatedFieldWarning (Class<?> clazz, Field field, boolean isModifier) {
    char[] fieldName = field.getName().toCharArray();
    StringBuilder sig = new StringBuilder(255);

    sig.append("Direct use of public fields is deprecated, please add a method '");

    fieldName[0] = Character.toUpperCase(fieldName[0]);

    sig.append("public ");
    if (Modifier.isStatic(field.getModifiers())) { sig.append("static "); }
    sig.append(isModifier? "void": field.getType().getName());   /* Return type */
    sig.append(isModifier? " set": " get");                      /* Name prefix */
    sig.append(fieldName);                                       /* Name        */
    sig.append("(");
    sig.append(isModifier? field.getType().getName(): "");       /* Parameters  */
    sig.append(")");

    sig.append("' to ").append(clazz).append('.');
    return sig;
  }

  /** Finds a named method or property.
      @param clazz       The class to check in.
      @param name        The name of the property.
      @param types       The argument types to be applied. If this is <code>null</code> no argument
                         checks will be applied (used with the <code>QUERY</code> command) if this
                         is an array of length zero than the methods are assumed to accept zero
                         arguments.
      @param staticOnly  Only get static members.
   */
  public static Member[] findMembers(Class<?> clazz, String name, Class<?>[] types, boolean staticOnly) {
    return findMembers(clazz, name, types, staticOnly, false, null);
  }

  /** Finds a named method or property.
      @param clazz       The class to check in.
      @param name        The name of the property.
      @param types       The argument types to be applied. If this is <code>null</code> no argument
                         checks will be applied (used with the <code>QUERY</code> command) if this
                         is an array of length zero than the methods are assumed to accept zero
                         arguments.
      @param staticOnly  Only get static members.
      @param tryIndexed  Indicates if the indexed version should be tried (not applicable unless
                         <code>types.length==1</code>).
      @param searchOrder The member type search order.
   */
  private static Member[] findMembers(Class<?> clazz, String name, Class<?>[] types, boolean staticOnly,
                                      boolean tryIndexed, int[] searchOrder) {
    Member[] members = new Method[0];

    if (searchOrder == null) {
      ArrayList<Member> allMembers = new ArrayList<Member>();
      Constructor<?>[] constructors = null;
      Field[]  fields  = null;
      Method[] methods = null;

      if (!staticOnly && ((name == null) || (name.equalsIgnoreCase("<init>")))) {
        constructors = getConstructors(clazz);
      }
      else {
        constructors = EMPTY_CONSTRUCTOR_ARRAY;
      }

      if (name == null) {
        fields  = getFields(clazz,staticOnly);
        methods = getMethods(clazz,staticOnly);
      }
      else {
        fields  = getNamedFields(clazz,name,staticOnly);
        methods = getNamedMethods(clazz,name,staticOnly);

        List<Method> methodsFound = Arrays.asList(methods);
        Method[] accessors = getNamedAccessors(clazz,name,staticOnly);
        Method[] modifiers = getNamedModifiers(clazz,name,staticOnly);

        for (Method accessor : accessors) {
          if (!methodsFound.contains(accessor)) {
            methodsFound.add(accessor);
          }
        }

        for (Method modifier : modifiers) {
          if (!methodsFound.contains(modifier)) {
            methodsFound.add(modifier);
          }
        }

        methods = methodsFound.toArray(EMPTY_METHOD_ARRAY);
      }

      constructors = (Constructor<?>[]) getGoodMembers(constructors,  types, true, true, true);
      fields  = (Field[]) getGoodMembers(fields,  types, true, true, true);
      methods = (Method[]) getGoodMembers(methods, types, true, true, true);

      allMembers.addAll(Arrays.asList(constructors));
      allMembers.addAll(Arrays.asList(fields));
      allMembers.addAll(Arrays.asList(methods));

      members = allMembers.toArray(EMPTY_MEMBER_ARRAY);

    }
    else if (name.equalsIgnoreCase("<init>")) {
      members = getConstructors(clazz);
      members = getGoodMembers(members, types, true, true, true);

    }
    else {
      for (int i = 0; (i < searchOrder.length) && ((members.length == 0) || (name == null)); i++) {
        if (searchOrder[i] == NAMED_FIELDS) {
          Field[] fields = getNamedFields(clazz, name, staticOnly);

          members = getGoodFields(fields, types, true, true, true);

          if ((members.length == 0) && tryIndexed) {
            members = getGoodFields(fields, new Class<?>[0], true, true, true);
          }
        } else {
          Method[] methods = null;

          switch (searchOrder[i]) {
            case NAMED_METHODS:   methods = getNamedMethods(  clazz, name, staticOnly); break;
            case NAMED_ACCESSORS: methods = getNamedAccessors(clazz, name, staticOnly); break;
            case NAMED_MODIFIERS: methods = getNamedModifiers(clazz, name, staticOnly); break;
            default: throw new RuntimeException("Illegal search order param ("+i+")");
          }

          members = getGoodMembers(methods, types, true, true, true);

          if ((members.length == 0) && tryIndexed) {
            members = getGoodMembers(methods, new Class<?>[0], true, true, true);
          }
        }
      }
    }
    return members;
  }

  /** Finds a named method and invokes it.
      @param obj  The object to invoke the method on (may be a class if the method is static).
      @param key  The name of the method or name of a fields (with accessor/modifier).
      @param args Arguments to pass to the method (may be an array of length zero).
      @param ref  Reference object (to get Midas context)
      @return The return value from the method (may be <code>null</code>) or <code>null</code> if
              there is no applicable return value.
   */
  private static Object findAndInvoke (Object obj, String key, Object[] args, int[] searchOrder, Object ref) {
    StringBuffer debugInfo = (DEBUG_FINDER)? new StringBuffer(): null;

    if (args == null) args = EMPTY_OBJECT_ARRAY;

    Object value = null; /* The return value from an accessor or null if using a modifier. */

    if (DEBUG_FINDER) {
      debugInfo.append("KeyObject debug info:\n");
      debugInfo.append("  obj       = " + obj  + "\n");
      if (obj!=null) debugInfo.append("  obj.class = " + obj.getClass() + "\n");
      debugInfo.append("  key       = '" + key + "'\n");
      debugInfo.append("  args      = { " + StringUtil.join(args) + " }\n");
    }

    boolean hasOneArg  = (args != null) && (args.length == 1);
    boolean valueSet = false;
    boolean isStatic = ((obj != Class.class) && (obj instanceof Class)) || (obj == null);
    Class<?>   clazz    = (isStatic)? (Class<?>)obj: obj.getClass();

    if (obj == null) {
      value    = null;
      valueSet = true;
    }
    else if ((obj instanceof Table) || (obj instanceof KeyVector)) {
      Keyable keyable = (Keyable)obj;

      Object iarg = (args!=null && args.length==2)? iarg=args[0] : null;  // prep for indexed key
      if (searchOrder == SEARCH_ORDER_SET && iarg != null) {              // since 2.7.4 to fix Bug 2068
        if (DEBUG_FINDER) debugInfo.append("  (Object is a Table/KeyVector, using setIndexed(getKey(String),index,args[1])\n");
        value = setIndexed(keyable.getKey(key),Convert.o2l(iarg),args[1]);
      }
      else if (!hasOneArg || (searchOrder == SEARCH_ORDER_GET)) {
        if (DEBUG_FINDER) debugInfo.append("  (Object is a Table/KeyVector, using getKey(String))\n");
        value = keyable.getKey(key);

        // This next section used when getting "X(0)" from the results table where X is a VD.
        if (hasOneArg) value = getIndexed(value, getKeyIndex(args[0].toString()));
      }
      else {
        if (DEBUG_FINDER) debugInfo.append("  (Object is a Table/KeyVector, using setKey(String,Object))\n");
        value = keyable.setKey(key, args[0]);
      }
      valueSet = true; // Never check fields/methods/chainable/etc. with a Table or KeyVector

    } else if (obj instanceof XMValue) {
      XMValue xmvalue = (XMValue)obj;

      if (key.equalsIgnoreCase("GETSIZE")) {
        if (DEBUG_FINDER) debugInfo.append("  (Object is a XMValue, getting size)\n");
        value = xmvalue.getSize();
      }
      if (!hasOneArg || (searchOrder == SEARCH_ORDER_GET)) {
        if (DEBUG_FINDER) debugInfo.append("  (Object is a XMValue, using get(String))\n");
        value = xmvalue.get(key);

      } else {
        if (DEBUG_FINDER) debugInfo.append("  (Object is a XMValue, using set(String,Object))\n");
        value = xmvalue.set(key, args[0]);
      }
      valueSet = true; // Never check fields/methods/chainable/etc. with a Table or KeyVector

    } else if (obj instanceof Keyable) {
      if (!hasOneArg && (searchOrder == SEARCH_ORDER_GET)) {
        if (DEBUG_FINDER) {
          debugInfo.append("  (Object is a Keyable, using getKey(String))\n");
        }
        value    = ((Keyable)obj).getKey(key);
        valueSet = (value != null);

      } else if (hasOneArg && (args.length == 1) && (searchOrder == SEARCH_ORDER_SET)) {
        if (DEBUG_FINDER) {
          debugInfo.append("  (Object is a Keyable, using setKey(String,Object))\n");
        }
        value    = ((Keyable)obj).setKey(key, args[0]);
        valueSet = (value != null);
      }

    } else if (clazz.isArray()) {
      int index = getKeyIndex(key);

      if (key.equalsIgnoreCase("SIZE") || key.equalsIgnoreCase("GETSIZE")) {
        if (DEBUG_FINDER) {
          debugInfo.append("  member   = null (getting size of array)\n");
        }

        if (args.length == 0) {
          value = new Integer(Array.getLength(obj));
        } else {
          error(key,obj,null,debugInfo,"Unable to set size of array, array size is immutable.");
        }

      } else if (args.length < 2) {
        if (DEBUG_FINDER) {
          debugInfo.append("  member   = null (using '"+key+"' as array index)\n");
        }

        int length = Array.getLength(obj);

        if ((index >= 0) || (index < Array.getLength(obj))) {
          if (args.length == 0) {
            value = getIndexed(obj, index);
          } else {
            value = setIndexed(obj, index, args[0]);
          }
        } else {
          error(key,obj,null,debugInfo,"Unable to access array index "+key+": given array has "
                  + "only "+length+" elements numbered 0.."+(length-1)+".");
        }
      } else {
        error(key,obj,null,debugInfo,"Unable to set array index "+key+" when given "+args.length
                 + " arguments.");
      }
      valueSet = true;
    }

    if (!valueSet) {
      //// BASIC SETUP /////////////////////////////////////////////////////////////////////////////
      Class[]  types    = getClassOf(args);
      Object   nextLink = (obj instanceof Chainable)? ((Chainable)obj).getNextLink(): null;
      int      index    = getKeyIndex(key);
      boolean  tryIndex = (searchOrder == SEARCH_ORDER_GET) && (args.length == 1)
                       && (getKeyIndex(args[0].toString()) >= 0);

      //// GET THE METHOD/FIELD TO ACCESS //////////////////////////////////////////////////////////
      //Member[] members = findMembers(clazz,key,types,isStatic,tryIndex,searchOrder);
      Member[] members = findMembers(clazz,key,types,false,tryIndex,searchOrder);
      Member   member  = null; /* The specific member to access. */

      if (DEBUG_FINDER) {
        debugInfo.append("  arg types = { " + StringUtil.join(types) + " }\n");
        debugInfo.append("  members   = { " + StringUtil.join(members) + " }\n");
      }

      //// ACCESS THE METHOD/FIELD, IF NONE TRY KEYABLE/CHAINABLE/INDEXED //////////////////////////
      if (members.length > 0) {
        if (members instanceof Method[]) {
          Method[] methods = (Method[])members;
          Method   method  = null;
// if ("SR".equals(key) || "SETPAGE".equals(key)) System.out.println("DEBUG: methods =\n"+StringUtil.toStringBuilder(methods," "));
          try {
            if (methods[0].getParameterTypes().length != args.length) {
              //// INVOKE THEN GET INDEXED /////////////////////////////////////////////////////////
              method = (Method)getBestMember(methods, key, new Class<?>[0], true, true, true);

              if (DEBUG_FINDER) {
                debugInfo.append("  (Accessing value of '"+key+"' via index '"+args[0]+"')\n");
                debugInfo.append("  member    =  " + method + "\n");
              }

              value = getIndexed(method.invoke(obj,EMPTY_OBJECT_ARRAY), getKeyIndex(args[0].toString()));

            } else {
              //// NORMAL INVOKE ///////////////////////////////////////////////////////////////////
              method = (Method)getBestMember(methods, key, types, true, true, true);

              if (DEBUG_FINDER) {
                debugInfo.append("  member    = " + method + "\n");
              }

              // since 3.1.1: Work around for Java BUG 4071957 (reflect) Method.invoke access control
              // does not understand inner class scoping. For BUG 2552.
              // Otherwise, we get an IllegalAccessException trying to invoke a "public abstract" method
              // from an inner interface (e.g. MJList.CachedDataFile implements ListFile -> RecordFileRW -> RecordFile.getURL())
              // nM> res reg.list2.listfile.url  ! to see problem
              if (Modifier.isAbstract(method.getModifiers())) {
                try {
                  method.setAccessible(true); // try to allow access to public abstract methods
                } catch (SecurityException se) {
                  // ignore and continue (for inside Applet or restricted Web Start sandbox)
                }
              } // end since 3.1.1 for BUG 2552

              value = method.invoke(obj, convertParams(method.getParameterTypes(), args));
            }
          } catch(Exception e) {
            error("An exception occurred while trying to run "+method,e);
          }
        } else if (members instanceof Constructor[]) {
          Constructor<?>[] constructors = (Constructor[])members;
          Constructor<?>   constructor  = null;

          try {
            constructor = (Constructor<?>)getBestMember(constructors, key, types, true, true, true);

            if (DEBUG_FINDER) {
              debugInfo.append("  member    = " + constructor + "\n");
            }

            value = constructor.newInstance(convertParams(constructor.getParameterTypes(), args));

          } catch(Exception e) {
            error("An exception occurred while trying to run "+constructor,e);
          }
        } else if (members instanceof Field[]) {
          Field field = (Field)members[0];

          if (BE_ANNOYING && !Modifier.isStatic(field.getModifiers())
                          && !Modifier.isFinal(field.getModifiers())) {
            /* Output a warning message that includes the signature for the required method. */
            ref2Midas(ref).warning(getDeprecatedFieldWarning(clazz, field, (args.length == 1)));
          }

          if (args.length > 1) {
              error(key,obj,types,debugInfo,"Too many parameters given ("+args.length+") when trying "
                      + "to set " + field);
          } else {
            try {
              if (args.length == 0) { //// GET FIELD ////////////////////////////////////////////////
                value = field.get(obj);
              } else if ((searchOrder == SEARCH_ORDER_GET) && (args.length == 1)) {
                                     //// INVOKE THEN GET INDEXED ///////////// ////////////////////
                value = getIndexed(field.get(obj), getKeyIndex(args[0].toString()));
              } else {               //// SET FIELD ////////////////////////////////////////////////
                field.set(obj, convertParam(field.getType(), args[0]));
              }

            } catch(Exception e) {
              error("An exception occurred while trying to "+((args.length==1)? "set": "get")
                   +" the value of "+field,e);
            }
          }
        } else {
          // This should be impossible at this time.... will add ability to work with constructors
          // in the future
          throw new RuntimeException("Variable member must be an instance of either Method or "
                                   + "Field, given instance of " + toString(member.getClass()));
        }

      } else if (nextLink != null) {
        if (DEBUG_FINDER) {
          debugInfo.append("  member   = null (using Chainable interface)\n");
        }
        value = findAndInvoke(nextLink, key, args, searchOrder, ref);

      } else if ((index >= 0) && (args.length < 2)) {
        if (DEBUG_FINDER) {
          debugInfo.append("  member   = null (using '"+key+"' as index value)\n");
        }

        try {
          value = (args.length == 0)? getIndexed(obj, index): setIndexed(obj, index, args[0]);

        } catch(ArrayIndexOutOfBoundsException e) {
          error("Unable to access index "+index, e);
        }

      } else {
        if ((obj instanceof Keyable) || (clazz.isArray())) {
          /* Handled above but returned null... not an error. */

        } else {
          if (DEBUG_FINDER) {
            debugInfo.append("  member   = null");
          }
          error(key,obj,types,debugInfo, "No methods found.");
        }
      }
    }
    if (EXTRA_DEBUG && DEBUG_FINDER) { System.out.println(debugInfo); }

    return value;
  }

  private static final String[] COMMON_PACKAGES = { "java.lang.", "java.util.", "nxm.sys.lib." , "nxm.sys.libg." };

  /** Converts a class to a string and trims off common package names. */
  private static String toString (Class<?> clazz) {

    String str = (clazz == null)? "null": clazz.getName();

    for (String packageName : COMMON_PACKAGES) {
      if (str.startsWith(packageName) && (str.indexOf('.', packageName.length()) < 0)) {
        str = str.substring(packageName.length());
        break;
      }
    }
    return str;
  }

  private static void error (String key, Object obj, Class[] types, StringBuffer debugInfo, String error) {
    StringBuffer str = new StringBuffer();
    String className = (obj==null)? null: obj.getClass().getName();

    str.append("KeyObject: Can not access '"+key+"' in "+className);

    if (types == null) {
      /* Ignore type list */
    } else if (types.length == 0) {
      str.append(" when given no arguments");

    } else if (types.length == 1) {
      str.append(" when given one argument of type (");
      str.append(toString(types[0]));
      str.append(")");

    } else {
      str.append(" when given arguments of type (");
      str.append(toString(types[0]));
      for (int i = 1; i < types.length; i++) {
        str.append(", ");
        str.append(toString(types[i]));
      }
      str.append(")");
    }

    str.append(": ");
    str.append(error);

    if (DEBUG_FINDER) {
      str.append("\n");
      str.append(debugInfo);
    }

    throw new MidasException(str.toString());
  }

  private static void error (String error, Throwable throwable) {
    // If the first thing is an InvocationTargetException then it usually isn't particularly
    // interesting unless it's the only thing (possibly a bad type conversion). So immediately
    // strip it off to get the real cause of the exception (typically an error in the method
    // called that is unrelated to it being called via invoke).
    if ((throwable instanceof InvocationTargetException) && (throwable.getCause() != null)) {
      throwable = throwable.getCause();
    }
    MidasException e = new MidasException("KeyObject: "+error+": "+throwable, throwable);
    e.setStackTrace(new StackTraceElement[0]); // only include the nested exception's stack trace
    throw e;
  }
}
