package nxm.sys.lib;

import nxm.sys.inc.*;
import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Map;

import static nxm.sys.lib.DataFile.getUnitsID;

/** A class to perform Keyword head block manipulation.
    <pre>
      Format {
        L:length - Total length of keyword (lkey) including header (8), data value, tag name, and padding.
        I:lext   - Length of extra bytes = length - length of data
        B:ltag   - Length of keyword tag name
        B:type   - A,B,I,L,...
        data     - Data (has associated REP). length of data = (length - lext)
        tag      - Tag name is after data
        pad      - PAD bytes (to reach a total keyword structure size that is a multiple of 8 bytes)
      }
    </pre>
    @author Jeff Schoen / Neon Ngo
    @version $Id: Keywords.java,v 1.80 2012/12/20 22:33:58 jxc Exp $
*/
public class Keywords extends Object implements Keyable,DataTypes,AsciiMap {
  /** These are the only keywords that should appear in the MAIN header. This is as per version 1.1
      of the BLUE ICD. Any keyword not in this list, is subject to automatic deletion. (BLUE 1.1
      requires that the <tt>VER=</tt> and <tt>IO=</tt> keywords be written, {@link DataFile} will
      be forced to delete any extra keywords in the main header if it needs to make room for these.
      <pre>
        Main Header Keywords Listed in ICD:
          CREATOR=        - Identifier for the system that created the file.
          IO=             - Name of the I/O library that wrote the file (see {@link DataFile#getBlueIO}).
          PACKET=         - Used by NeXtMidas to indicate that a separate file contains packetized data
                            that controls the interpretation of data in this file. The value identifies a
                            packet handler and any switches that modify its behavior (e.g. ICE and SDDS).
          PKT_BYTE_COUNT= - Used by X-Midas to indicate that a file contains packetized data. The value
                            is the total number of bytes of packetized data contained in the file.
          TC_PREC=        - Precision time code (see section 3.4.1.1 of the BLUE ICD).
          VER=            - Version of BLUE ICD used when writing file (see {@link DataFile#getBlueVer}).
      </pre>
      @since NeXtMidas 2.5.3
   */
  @InternalUseOnly // since 3.3.0
  public static final String[] MAIN_HEADER_KEYWORDS
          = { "CREATOR", "IO", "PACKET", "PKT_BYTE_COUNT", "TC_PREC", "VER" }; // Must be in alphabetical order!!!
  /** Keywords permitted in the MAIN header without warning on insertion. This list complements
      the {@link #MAIN_HEADER_KEYWORDS} list with some entries that are permitted by convention,
      even though they are not listed in the BLUE 1.0 or BLUE 1.1. ICD.
      <pre>
        ACQDATE=          - Acquisition date
        ACQTIME=          - Acquisition time
        COMMENT=          - User-supplied comment
      </pre>
      Unlike the entries in {@link #MAIN_HEADER_KEYWORDS} the entries here are not given
      "priority use" of the MAIN header any may be removed if necessary to accommodate
      one of the entries in {@link #MAIN_HEADER_KEYWORDS} (e.g. a long COMMENT may be
      removed to permit insertion of TC_PREC).
      @since NeXtMidas 3.3.0
   */
  @InternalUseOnly
  public static final String[] MAIN_HEADER_IGNORE
          = { "ACQDATE", "ACQTIME", "COMMENT" };  // Must be in alphabetical order!!!

  /** byte offset in header of keywords string */
  private static final int IKOFF=164;
  /** Internal Keywords KMAX                   */
  private static final int IKMAX=92;
  /** Allocation size                          */
  private static final int EXTSIZE=2048;
  /** keyword header length of binary keyword block in Midas BLUE EXTENDED header */
  public  static final int HEADER_LEN = 8;

  /** Keyword scopes: SCOPE (only in current scope), MAIN (only in main header), or EXTENDED (only in extended header) */
  public static final int SCOPE=0, MAIN=0x1, EXTENDED=0x2;
  /** SET keyword flags: ADD (allow duplicate keys), REPLACE (overwrite existing key in current scope) */
  public static final int ADD=0x1, REPLACE=0x2;
  
  private static final int ADD_REPLACE = ADD|REPLACE; // mask value for ADD|REPLACE
  
  public static final boolean EXTRACTINPLACE=false;

  /** The comma-separated list of possible Keyword flags. (since NeXtMidas 3.1.1) */
  public static final String flagsList = "NoAbort";
  /** Keywords flag: should not abort (throw Exception) when an invalid keyword block is found (since NeXtMidas 3.1.1) */
  public static final int NOABORT = 0x01;
  /** default Keywords flags (since 3.1.1) */
  private static final int DEF_FLAGS = 0; // none

  /** Buffer for extended header                       */
  public  byte[] buf;
  /** Buffer for main header (keywords start at IKOFF) */
  public  byte[] ibuf;
  /** Size of extended header                          */
  public  int    size;
  /** Size of main header keywords (0..IKMAX)          */
  public  int    isize;
  /** Keyword Byte ordering ('I'=IEEE,'E'=EEEI,...). In some cases this is not initialized,
      hence it was deprecated in NeXtMidas 2.5.2. Instead, use {@link DataFile#getHeadRep()}.
   */
  private byte     rep;
  /** Current scope string                                */
  private String scope="";
  /** Start of scope (-isize..size, -N is in main header) */
  private int scope_start;
  /** End of scope   (-isize..size, -N is in main header) */
  private int scope_end;
  /** current keyword returned from {@link #getNext()}    */
  private Key cur=new Key();
  /** DataFile the keywords are associated with           */
  private DataFile hcb;
  /** flags mask , see {@value #flagsList} (since NeXtMidas 3.1.1) */
  private int flags = 0;
  /** set of places (methods) where we have warned about invalid keyword blocks (since 3.1.1) */
  private HashSet<String> warnedForInvalidBlockSet = new HashSet<String>();

  /** String versions of "(" and ")", really belongs in AsciiMap */
  static final String LPAREN_S = "(", RPAREN_S = ")";

  /** Flags that control the handling of units. Values include:
      <pre>
        UNITS_WARN        - Issue warning on units mismatch.
        UNITS_LIST        - Show units with {@link #list(Map.Entry,String,int,int)}.
        UNITS_PREFER_UNIT - Prefer use of ".UNIT" rather than ".UNITS".
        UNITS_MATCH       - Match the choice of ".UNIT" or ".UNITS" based on
                            existing keyword if {@link #REPLACE} is used and
                            exactly one of them already exists.
      </pre>
   */
  @InternalUseOnly
  public static final int UNITS_WARN        = 0x0100,
                          UNITS_LIST        = 0x0200,
                          UNITS_PREFER_UNIT = 0x0400,
                          UNITS_MATCH       = 0x0800;

  /** Initialize a Keyword object */
  public Keywords () {
    super();
    size=isize=0;
    cur.offset=0;
    hcb=null;
  }

  public Keywords (DataFile hcb) {
    this();
    this.hcb   = hcb;
    this.ibuf  = hcb.hb;
    this.isize = -1;
  }

  public void clear () {
    size=isize=cur.offset=scope_start=scope_end=0;
  }

  /** Clear scope to look in main and extended header for keyword.  This can
      be achieved by setting scope to an empty string
  */
  public void clearScope () {
    scope_start = -isize;
    scope_end = this.size;
  }

  /** Only used by DataFile: Initialize the keywords, static file. */
  public void init (byte[] buf, int size, byte rep) {
    this.size = size;
    this.buf = buf;
    if (hcb!=null) {
      ibuf = hcb.hb; // keywords in DataFile's header (in main header)
      isize = hcb.getKeyLength();
    }
    if (rep!=0) this.rep = rep;  // <-- DEPRECATED since 2.5.3
    else this.rep = Shell.rep;   // <-- DEPRECATED since 2.5.3
    if (!chk(1024)) {
      String msg = "Bad Extended Header on URL="+((hcb!=null)? hcb.getURL() : "");
      warn(msg);
      this.size = 0;
    }
    clearScope();
  }

  /** Only used by DataFile: Initialize the keywords, pipe. */
  public void init (Keywords kw) {
    buf=null;
    if (kw.size>0) {
      buf = new byte[kw.buf.length];
      System.arraycopy (kw.buf,0, buf,0, kw.size); // copy keyword buffer (in extended header)
    }
    init(buf,kw.size,(byte)0);
  }

  /** get current Keyword flags. (Since NeXtMidas 3.1.1) */
  public int getFlags () { return flags; }

  /** Set Keyword flags. (Since NeXtMidas 3.1.1) */
  public void setFlags (int newFlags) { flags = newFlags; }

  /** Set Keyword flags using String. (Since NeXtMidas 3.1.1) */
  public void setFlags (String newFlagsStr) {
    setFlags(Parser.mask(flagsList, newFlagsStr, DEF_FLAGS));
  }

  /** Get Keyword flags as a String. (Since NeXtMidas 3.1.1) */
  public String getFlagsString () {
    return Parser.mask2s(flagsList, flags);
  }

  /** Get the current scope string */
  public String getScope () { return scope; }

  /** Set the scope string. This only applies to the extended header as there is no scoping concept
      in the main header. However, there are three special scopes "MAIN" (scope to main header),
      "EXT" (scope to entire extended header), and "ALL" (scope to everything in main and extended header).
      <pre> Handles input strings of the form
        TAG=VAL or
        "TAG1=VAL1,TAG2=VAL2" or
        (TAG=VAL1,TAG=VAL2)
      </pre>
      @param newScope Input scope or "MAIN" or "EXT" or "ALL". A null, blank,
                      or white space only String is the same as "ALL".
      @return The number of levels (this will always be 1 for "MAIN", "EXT", or "ALL").
  */
  public int setScope (String newScope) {
    int levels=0;
    scope = (newScope==null)? "" : newScope.trim();
    if (scope.equals("ALL") || scope.length()==0) { clearScope(); return 1; }
    scope_start=-isize; scope_end=0;
    if (scope.equals("MAIN")) { return 1; }
    scope_start=0; scope_end=size;
    if (scope.equals("EXT")) { return 1; }
    Parser parser;
    // Handle string of the form (TAG=VAL1,TAG=VAL2)
    if (scope.startsWith(LPAREN_S)) {
      if (scope.endsWith(RPAREN_S)) {
        parser = new Parser(scope.substring(1,scope.length()-1), AsciiMap.COMMA);
      } else {
        warn("Unbalanced parenthesis in '"+scope+"'");
        parser = new Parser(scope.substring(1), AsciiMap.COMMA);
      }
    } else {
      parser = new Parser(scope, AsciiMap.COMMA);
    }
    while (parser.more()) {
      String str = parser.next();
      int iEqual=str.indexOf('=');
      String name = str.substring(0,iEqual);
      String value = str.substring(iEqual+1);
      int found = setScope(name,value);
      // Any keyword not found for scope causes error
      if (found == 1) { levels++; }
      else            { levels = 0; break; }
    }
    return levels;
  }

  /* The real scope setting method, not published to the outside world */
  private int setScope (String name, String value) {
    boolean found=false;
    int i; Key k=new Key();
    byte[] bname = name.getBytes();
    int blen = bname.length;
    boolean skiptag = (bname[blen-1]=='-' || bname[blen-1]=='+');
    boolean nexttag = bname[blen-1]=='+';
    if (nexttag||skiptag) blen--;
    for (k.offset=scope_start; k.offset<scope_end; k.offset+=k.length) {
      String invalidDetails = unpack(k);
      if (invalidDetails != null) { // since 3.1.1 (will get here is NoAbort is set)
        if (k.length >= HEADER_LEN) continue; // try next keyword block
        else                        break;    // bad length, can't recover from this (break out of loop)
      }
      if (k.ltag!=blen) continue;
      for (i=0; i<blen; i++)
        if (bname[i] != buf[k.offset + HEADER_LEN + k.ldat + i]) break;
      if (i!=blen) { } // no match
      else if (found) {
        scope_end=k.offset;
        if (nexttag) scope_end += k.length;
        return 1;
      } else {
        extractData(k);
        if (k.type != 'A') {
          // Numbers like 2 and 2.0 have to be treated as equal to match legacy X-Midas behavior
          if (Convert.o2d(k.value) == Convert.o2d(value)) {
            found=true; scope_start=k.offset;
            if (skiptag) scope_start += k.length;
          }
        } else {
          if (value.equals(k.value.toString())) {
            found=true; scope_start=k.offset;
            if (skiptag) scope_start += k.length;
          }
        }
      }
    }
    if (found) return 1;
    else return 0;
  }

  /** Get the next keyword in the MAIN or extended header. This method should NOT be used to iterate
      through all of the keywords (or all the keywords in scope), if used this way it may result a
      StringIndexOutOfBoundsException for some (but not all) files. The proper way to iterate over
      the keywords is as follows:
      <pre>
        if (keywds.setScope("FIELD="+i) == 1) { // setScope is optional but frequently used
          Keywords.Iterator kwi = keywds.iterator();
          Keywords.key      key = null;

          while(kwi.hasNext()) {
            key = kwi.next();

            // Process keyword here
          }
        }
      </pre>
      @return next Key, null if at end or if invalid block is found
  */
  public Key getNext () {
    cur.offset += cur.length;
    if (cur.offset>=scope_end) return null;
    String invalidDetails = unpack(cur);
    if (invalidDetails != null) { // since 3.1.1 (will get here is NoAbort is set)
      return null;                // return null if invalid keyword block
    }
    extractName(cur);
    extractData(cur);
    return cur;
  }

  /** Unpack keyword header block into specified Key object.
      Since 3.1.1: If NoAbort is NOT set in the flags mask AND an invalid keyword block is found
                   then a MidasException will be thrown with the details of the bad block.
      @param k Key to unpack keyword block into
      @return null if successful, otherwise a String with more details of invalid keyword block (since 3.1.1)
      @throws MidasException when an invalid block has been unpacked, unless the {@link #NOABORT}
              flag has been set (since 3.1.1)
      @see #handleBadKeywordBlock(Key, String, CharSequence)
   */
  private String unpack (Key k) {
    int off = k.offset;
    short lext;
    if (off < 0) {  // negative offset is to unpack keyword block in BLUE file's main header
      int i;
      off += isize; // translate offset to internal buffer
      for (i=off; i<isize && ibuf[IKOFF+i]!=0; i++) { }             // find offset of NUL char
      k.length = i-off;
      for (i=off; i<off+k.length && ibuf[IKOFF+i]!=EQUALS; i++) { } // find offset of '=' char
      k.ltag = i-off;
      k.ldat = k.length-k.ltag-1;
      k.type = ASCII;                           // all internal fields in main header are ASCII
      if (ibuf[IKOFF+off+1]==COLON) k.type=ibuf[IKOFF+off];
      if (off+k.length < isize) k.length++;     // skip \0 delimiter
      else k.length = isize-off;
    }
    else { // else unpack keyword block in extended header
      byte _rep = (hcb==null)? Shell.rep: hcb.headRep;
      k.length = Convert.unpackL(buf, off,   _rep); // int_4 for lkey
      lext     = Convert.unpackI(buf, off+4, _rep); // int_2 for lext
      k.ltag = buf[off+6]; // int_1
      k.type = buf[off+7]; // char
      k.ldat = k.length - lext;

      // since 3.1.1, check if unpacked keyword block is valid (in extended header) for BUG 2548
      if ( k.length < HEADER_LEN   // 1. lkey (in extended header) cannot be < 8 bytes (keyword header size)
          || lext    >  k.length   // 2. lext cannot be >  than lkey (this also covers when lext is negative)
          || k.ltag  >= k.length   // 3. ltag cannot be >= lkey
          || k.ldat  >= k.length   // 4. ldat cannot be >= lkey (this also covers when ldat is negative)
          ) {
        byte count = 0;
        StringBuilder details = new StringBuilder(88); // 88 is about the max size for all the error details below
        if (k.length < HEADER_LEN) details.append((++count>1? "," : "")).append(" lkey < ").append(HEADER_LEN); // 1
        if (lext > k.length)       details.append((++count>1? "," : "")).append(" lext > lkey");      // 2a
        if (lext < 0)              details.append((++count>1? "," : "")).append(" lext is negative"); // 2b
        if (k.ltag >= k.length)    details.append((++count>1? "," : "")).append(" ltag >= lkey");     // 3
        if (k.ldat >= k.length)    details.append((++count>1? "," : "")).append(" ldat >= lkey");     // 4a
        if (k.ldat < 0)            details.append((++count>1? "," : "")).append(" ldat is negative"); // 4b
        details.append(' ');
        handleBadKeywordBlock(k, "unpack", details);
        return details.toString();
      }
    }
    return null;
  }

  /** find the named key in the current scope */
  public Key find (String name) {
    return find(name,SCOPE);
  }

  /** find the named key in the specified scope
    @param name  The keyword name
    @param where The keyword scope: {@link #SCOPE} (default), {@link #MAIN}, or {@link #EXTENDED}
    @return Key if found, null if not found
  */
  public Key find (String name, int where) {
    int i,off,len; Key k=new Key();
    byte[] tmp, bname=name.getBytes();
    int start=scope_start, end=scope_end;
    if (where==MAIN) { start=-isize; end=0; }
    else if (where==EXTENDED) { start=0; end=size; }
    for (k.offset=start; k.offset<end; k.offset+=k.length) {
      String invalidDetails = unpack(k);
      if (invalidDetails != null) { // since 3.1.1 (will get here is NoAbort is set)
        if (k.length >= HEADER_LEN) continue; // try next keyword block
        else                        break;    // bad length, can't recover from this (break out of loop)
      }
      if (k.offset<0) { tmp=ibuf; off=IKOFF+k.offset+isize; len=k.ltag; }
      else { tmp=buf; off=k.offset+HEADER_LEN+k.ldat; len=k.ltag; }
      if (tmp[off+1]==COLON) { off+=2; len-=2; }
      if (len!=bname.length) continue;
      for (i=0; i<bname.length; i++)
        if (bname[i]!=tmp[off+i]) break;
      if (i==bname.length) return k;
    }
    return null;
  }

  /** check validity of this block up to the specified length.
      @return true is valid (up to specified size), otherwise false for invalid block(s)
    */
  private boolean chk (int maxlen) {
    byte _rep = (hcb==null)? Shell.rep: hcb.headRep;
    Key k=new Key();
    if (size>8 && Convert.unpackL(buf,4,_rep)<200) convertMartez();
    int end = (maxlen>=0)?  Math.min(size,maxlen) : size;
    for (k.offset=0; k.offset<end; k.offset+=k.length) {
      String invalidDetails = unpack(k);
      if (invalidDetails != null) {  // since 3.1.1 (will get here is NoAbort is set)
        return false;                                // 1. since 3.1.1, invalid blocks for BUG 2548
      }
      if ((k.offset + k.length) > size ||            // 2. bad length (beyond size of extended header)
          (k.offset + HEADER_LEN + k.ldat + k.ltag) > size) { // 3. bad tag length and/or data length (beyond size of extend header)
        handleBadKeywordBlock(k, "chk", "bad tag length and/or data length");
        return false;
      }
      for (int i=0,j=k.offset+HEADER_LEN+k.ldat; i<k.ltag; i++) {
        char c = (char)buf[j+i];
        if (c<' ' || c>'~') {                        // 4. bad char in (tag) name
          handleBadKeywordBlock(k, "chk", "bad char in tag name");
          return false;
        }
      }
    }
    return true;
  }

  private void convertMartez() {
    byte _rep = (hcb==null)? Shell.rep: hcb.headRep;
    int off,next;
    byte type;
    byte[] tmp = new byte[256];
    Key k=new Key(); k.offset=0;
    for (off=0; off<size; off=next) {
      next = Convert.unpackL(buf,off+0,_rep);
      int ltag = Convert.unpackL(buf,off+4,_rep);
      int ldat = Convert.unpackL(buf,off+8,_rep);
      int idat = 12 + (ltag+3)&0xFC;
      // type is before colon if in name, else default to ASCII
      type = (buf[off+13]==COLON)? buf[off+12] : ASCII;
      System.arraycopy(buf,off+12, tmp,0, ltag);  // save off name
      k.length = (8+ldat+ltag+7)&0xFFFFFFF8; // round to next 8 byte boundary
      Convert.packL(buf,k.offset+0,k.length);
      Convert.packI(buf,k.offset+4,(short)(k.length-ldat));
      if (_rep!=Shell.rep) {
        Convert.rep(buf,k.offset+0, LONG,1, Shell.rep,_rep);
        Convert.rep(buf,k.offset+4, INT,1, Shell.rep,_rep);
      }
      buf[k.offset+6] = (byte)ltag;
      buf[k.offset+7] = type;
      System.arraycopy (buf,off+12+((ltag+3)&0xFC), buf,k.offset+8, ldat);
      System.arraycopy (tmp,0, buf,k.offset+8+ldat, ltag);
      k.offset += k.length;
    }
    size = k.offset;
  }

  public void extractName (Key k) {
    if (k.ltag < 0) { // since 3.1.1 for BUG 2548
      handleBadKeywordBlock(k, "extractName", "negative ltag");
      return;
    }
    if (k.offset<0) k.name = new String(ibuf,IKOFF+k.offset+isize,k.ltag);
    else k.name = new String(buf, k.offset+HEADER_LEN+k.ldat, k.ltag);
    if (k.ltag>1 && k.name.charAt(1)==':') k.name = k.name.substring(2);
  }

  public void extractData (Key k) {
    byte _rep = (hcb==null)? Shell.rep: hcb.headRep;
    int offset = k.offset + HEADER_LEN;
    if (k.offset<0) k.value = new String(ibuf,IKOFF+k.offset+isize+k.ltag+1,k.ldat);
    else if (k.type==ASCII) k.value = new String(buf,offset,k.ldat);
    else if (EXTRACTINPLACE && (_rep==Shell.rep||k.type==BYTE)) {
        k.value = new Data(buf,offset,k.ldat,k.type);
    } else {
      byte[] buffer = new byte[k.ldat];
      System.arraycopy(buf,offset, buffer,0, k.ldat);
      if (_rep!=Shell.rep&&k.type!=BYTE) {
        Convert.rep (buffer,0, k.type, k.ldat/Data.getBPS(k.type), _rep,Shell.rep);
      }
      k.value = new Data(buffer,0,k.ldat,k.type);
    }
  }
  
  /** Gets all of the keys in a collection (setting scope to "ALL"). This is
      identical to using <tt>getAll("ALL",false)</tt>.
      @since NeXtMidas 3.3.0
   */
  public final Collection<Map.Entry<String,Object>> getAll () {
    return getAll("ALL",false);
  }
  
  /** Gets all of the keys in a collection. <br>
      <br>
      If a corrupted keyword block is detected, an exception will be thrown,
      except if the {@link #NOABORT} flag is set, in which case this function
      will return all of the keywords it was able to extract before encountering
      the error (sorted if applicable) followed by an entry with the key name
      <tt>"&lt;error&gt;"</tt>. Note that in the error case, calling this function
      to separately extract the main and extended header keywords (e.g. "MAIN",
      "EXT") *may* yield more keywords than calling this function with a scope of
      "ALL".
      @param scope The scope to use, one of: "ALL", "MAIN", "EXTENDED", "&lt;scope&gt;".
      @param sort  Sort the entries before returning?
      @since NeXtMidas 3.3.0
   */
  public final Collection<Map.Entry<String,Object>> getAll (String scope, boolean sort) {
    Iterator iter = iterator(scope);
    boolean  err  = false;
    
    ArrayList<Map.Entry<String,Object>> keys = new ArrayList<Map.Entry<String,Object>>();
    try {
      while (iter.hasNext()) {
        Key k = iter.next();
        keys.add(new AbstractMap.SimpleEntry<String,Object>(k.name, k.value));
      }
    }
    catch (RuntimeException e) {
      // Note that we don't print out the exception since it should already have
      // been printed as a warning.
      if ((flags & NOABORT) == 0) throw e;
      err = true;
    }
    
    if (sort) {
      Collections.sort(keys, StringUtil.KEY_COMPARATOR);
    }
    
    if (err) {
      keys.add(new AbstractMap.SimpleEntry<String,Object>("<error>", "<corrupted keyword block>"));
    }
    return keys;
  }

  /** Gets (extracts) the value for the specified keyword name.
      @param name keyword name to get
      @return value for specified keyword name, null if keyword is not found.
   */
  public Object get (String name) {
    Key k = find(name);
    if (k==null) return null;
    extractData(k);
    return k.value;
  }

  /** Gets (extracts) the value for the specified keyword name as an integer.
      @param name keyword name to get
      @param def  default value to return if keyword is not found
      @return integer value for specified keyword name
   */
  public int getL (String name, int def) {
    Object value = get(name);
    if (value==null) return def;
    else return Convert.o2l(value);
  }

  /** Gets (extracts) the value for the specified keyword name as an double.
      @param name keyword name to get
      @param def  default value to return if keyword is not found
      @return double value for specified keyword name
   */
  public double getD (String name, double def) {
    Object value = get(name);
    if (value==null) return def;
    else return Convert.o2d(value);
  }

  /** <b>Work in progress may change within this beta series:</b>
      Gets (extracts) the .UNIT or .UNITS value for the specified keyword name
      as an integer. This is identical to <tt>getUnitsFor(name,null,UNITS_WARN)</tt>.
      @param name  keyword name to get the units for.
      @return Integer value for the units specifier or null if not found.
      @since NeXtMidas 3.3.0
   */
  public final Integer getUnitsFor (String name) {
    return getUnitsFor(name,null,UNITS_WARN);
  }

  /** <b>Work in progress may change within this beta series:</b>
      Gets (extracts) the .UNIT or .UNITS value for the specified keyword name
      as an integer. This is identical to <tt>getUnitsFor(name,def,UNITS_WARN)</tt>.
      @param name  keyword name to get the units for.
      @param def   default value to return if keyword is not found
      @return Integer value for the units specifier or def if not found.
      @since NeXtMidas 3.3.0
   */
  public final Integer getUnitsFor (String name, Integer def) {
    return getUnitsFor(name,def,UNITS_WARN);
  }

  /** <b>Work in progress may change within this beta series:</b>
      Gets (extracts) the .UNIT or .UNITS value for the specified keyword name
      as an integer. The Platinum ICD Vol. 1 calls out support for ".UNITS" in
      section 1.8.1 and ".UNITS" in section 1.8.2; as such this function checks
      for both of them. If both are present and don't match a warning may be
      issued.
      @param name  keyword name to get the units for (e.g. use name="ABC" to get
                   the value of "ABC.UNITS" or "ABC.UNIT").
      @param def   default value to return if keyword is not found
      @param flags Controls the handling of mismatch between ".UNITS" and ".UNIT"
                   in the event that both are present (see {@link #UNITS_WARN}
                   and {@link #UNITS_PREFER_UNIT}).
      @return Integer value for the units specifier (see {@link nxm.sys.inc.Units})
              or def if not found.
      @since NeXtMidas 3.3.0
   */
  public Integer getUnitsFor (String name, Integer def, int flags) {
    Object units = get(name+".UNITS");
    Object unit  = get(name+".UNIT");

    if ((unit == null) && (units == null)) return def;
    if ((unit == null) && (units != null)) return Convert.o2l(units);
    if ((unit != null) && (units == null)) return Convert.o2l(unit);
    
    int _units = Convert.o2l(units);
    int _unit  = Convert.o2l(unit);
    if (_units == _unit) return _units; // both match
    
    if ((flags&UNITS_WARN) != 0) {
      warn("Found keyword mismatch "+name+".UNIT="+_unit+" and "
          +name+".UNITS="+_units+" using "+name+".UNITS="+_units);
    }
    return ((flags&UNITS_PREFER_UNIT) != 0)? _unit : _units;
  }

  /** Gets (extracts) the value for the specified keyword name as an integer.
      @param name keyword name to get
      @param def  default value to return if keyword is not found
      @return integer value for specified keyword name
      @since NeXtMidas 3.3.0
   */
  public Integer getL (String name, Integer def) {
    Object value = get(name);
    if (value==null) return def;
    else return Convert.o2l(value);
  }

  /** Gets (extracts) the value for the specified keyword name as an double.
      @param name keyword name to get
      @param def  default value to return if keyword is not found
      @return double value for specified keyword name
      @since NeXtMidas 3.3.0
   */
  public Double getD (String name, Double def) {
    Object value = get(name);
    if (value==null) return def;
    else return Convert.o2d(value);
  }

  /** Adds the specified keyword name/value to this (allowing duplicate keyword name).
      @param name keyword name to add
      @param obj value for specified keyword name to add
   */
  public void add (String name, Object obj) {
    set(name,obj,ADD);
  }

  /** <b>Work in progress may change within this beta series:</b>
      Adds the specified keyword name/value/units to this based on the specified
      flags. This is identical to: <tt>addWithUnits(name,value,units,0)</tt>.
      @param name  keyword name to set.
      @param value value for specified keyword name to set.
      @param units The units accompanying the value.
      @since NeXtMidas 3.3.0
   */
  public final void addWithUnits (String name, Object value, Object units) {
    addWithUnits(name,value,units,0);
  }

  /** <b>Work in progress may change within this beta series:</b>
      Adds the specified keyword name/value/units to this based on the specified
      flags.
      @param name  keyword name to set.
      @param value value for specified keyword name to set.
      @param units The units accompanying the value. This can either be the integer
                   value from {@link nxm.sys.inc.Units} or the corresponding string.
      @param flags The flags specifying the choice of ".UNIT" or ".UNITS" (see
                   {@link #UNITS_PREFER_UNIT}).
      @since NeXtMidas 3.3.0
   */
  public final void addWithUnits (String name, Object value, Object units, int flags) {
    setWithUnits(name,value,units,flags|ADD);
  }

  /** Copies all of the mappings from the specified map (allowing duplicate keyword names)
      to this using (i.e. using flags=ADD).
      @param map Mappings to be copied in this.
      @since NeXtMidas 3.1.2
   */
  public void addAll (Map<String,Object> map) {
    setAll(map, ADD);
  }

  /** Puts the specified keyword name/value to this (replacing any existing keyword name).
      @param name keyword name to put
      @param obj value for specified keyword name to put
   */
  public void put (String name, Object obj) {
    set(name,obj,REPLACE);
  }

  /** <b>Work in progress may change within this beta series:</b>
      Puts the specified keyword name/value/units to this based on the specified
      flags. This is identical to: <tt>putWithUnits(name,value,units,UNITS_MATCH)</tt>.
      @param name  keyword name to set.
      @param value value for specified keyword name to set.
      @param units The units accompanying the value.
      @since NeXtMidas 3.3.0
   */
  public final void putWithUnits (String name, Object value, Object units) {
    putWithUnits(name,value,units,UNITS_MATCH);
  }

  /** <b>Work in progress may change within this beta series:</b>
      Puts the specified keyword name/value/units to this based on the specified
      flags. Note that the behavior of this method is undefined if multiple
      matches already exist.
      @param name  keyword name to set.
      @param value value for specified keyword name to set.
      @param units The units accompanying the value. This can either be the integer
                   value from {@link nxm.sys.inc.Units} or the corresponding string.
      @param flags The flags specifying the choice of ".UNIT" or ".UNITS" (see
                   {@link #UNITS_MATCH} and {@link #UNITS_PREFER_UNIT}).
      @since NeXtMidas 3.3.0
   */
  public final void putWithUnits (String name, Object value, Object units, int flags) {
    setWithUnits(name,value,units,flags|REPLACE);
  }

  /** Copies all of the mappings from the specified map. Duplicate keyword
      names will replace existing entry (i.e. FLAGS=REPLACE).
      @param map Mappings to be copied in this.
      @since NeXtMidas 3.1.2
   */
  public void putAll (Map<String,Object> map) {
    setAll(map, REPLACE);
  }

  /** Copies all of the mappings from the specified collection. Duplicate keyword
      names will replace existing entry (i.e. FLAGS=REPLACE).
      @param entries Mappings to be copied in this.
      @since NeXtMidas 3.1.2
   */
  public void putAll (Collection<Map.Entry<String,Object>> entries) {
    setAll(entries, REPLACE);
  }

  /** Copies all of the mappings from the specified map to this using the specified flags.
      @param map Mappings to be copied in this.
      @since NeXtMidas 3.1.2
   */
  public void setAll (Map<String,Object> map, int flags) {
    if (map==null) return; // if map is null, don't need to copy to prevent NPE
    setAll(map.entrySet(), flags);
  }

  /** Copies all of the mappings from the specified collection to this using the specified flags.
      @param entries Mappings to be copied in this set of keywords.
      @since NeXtMidas 3.3.0
   */
  public void setAll (Collection<Map.Entry<String,Object>> entries, int flags) {
    if (entries==null) return; // if map is null, don't need to copy to prevent NPE
    for (Map.Entry<String,Object> e : entries) {
      set(e.getKey(), e.getValue(), flags);
    }
  }

  /** <b>Work in progress may change within this beta series:</b>
      Sets (add or put) the specified keyword name/value/units to this based on
      the specified flags. Note that when using {@link #REPLACE}, the behavior
      of this method is undefined if multiple matches already exist.
      @param name  keyword name to set.
      @param value value for specified keyword name to set.
      @param units The units accompanying the value. This can either be the integer
                   value from {@link nxm.sys.inc.Units} or the corresponding string.
      @param flags The flags specifying {@link #ADD} or {@link #REPLACE} plus the
                   choice of ".UNIT" or ".UNITS" (see {@link #UNITS_MATCH} and
                   {@link #UNITS_PREFER_UNIT}). If neither {@link #ADD} or
                   {@link #REPLACE} are set, {@link #REPLACE} is assumed.
      @since NeXtMidas 3.3.0
   */
  @InternalUseOnly
  private void setWithUnits (String name, Object value, Object units, int flags) {
    String  kwdName    = (name.length()>2 && name.charAt(1)==':')? name.substring(2) : name;
    boolean preferUnit = ((flags&UNITS_PREFER_UNIT)!=0);
    
    // Check the ADD/REPLACE flags
    if ((flags&(ADD_REPLACE)) == 0) {
      flags = flags & REPLACE;
    }
    else if ((flags&(ADD_REPLACE)) == ADD_REPLACE) {
      throw new MidasException("ADD and REPLACE flags are mutually exclusive, "
                             + "but both were given");
    }
    
    // If doing a REPLACE check for old entries
    if ((flags & REPLACE) != 0) {
      Key oldUnits = find(kwdName+".UNITS");
      Key oldUnit  = find(kwdName+".UNIT");
      
      // Match units if applicable
      if ((flags&UNITS_MATCH) != 0) {
        if ((oldUnits != null) && (oldUnit == null)) preferUnit = false;
        if ((oldUnits == null) && (oldUnit != null)) preferUnit = true;
      }
      
      // Delete the one we aren't going to use
      if ((oldUnits != null) &&  preferUnit) delete(oldUnits);
      if ((oldUnit  != null) && !preferUnit) delete(oldUnit);
    }
    
    // Set the keywords (note that only the ADD_REPLACE get passed down) the actual
    // type used for the units is arbitrary with "X:", "L:", "I:", and "B:" all
    // valid ('I:" and "L:" tend to be the most common), in this case we use "B:"
    // since it is the most space-efficient and matches the size to encode
    // sub-record units in Type 5000 files.
    if (preferUnit) {
      set(name,                  value,             flags&ADD_REPLACE);
      set("B:"+kwdName+".UNIT",  getUnitsID(units), flags&ADD_REPLACE);
    }
    else {
      set(name,                  value,             flags&ADD_REPLACE);
      set("B:"+kwdName+".UNITS", getUnitsID(units), flags&ADD_REPLACE);
    }
  }

  /** Sets (i.e. add or put) the specified keyword name/value to this based on the  specified flags.
      @param name keyword name to set
      @param obj value for specified keyword name to set
   */
  public void set (String name, Object obj, int flags) {
    char type = ' ';
    if (name.length()==0) return;
    if (name.length()>2 && name.charAt(1)==':') {
      type = name.charAt(0); name = name.substring(2);
      obj = Convert.o2o(obj,type,null);
    }
    else {
      // This section added in NeXtMidas 3.3.0
      if ((obj instanceof Data) || (obj instanceof String) || (obj instanceof Table)) {
        // keep as-is
      }
      else {
        // try to convert
        obj = Convert.o2o(obj,type,null);
      }
    }

    if (obj instanceof Data) {
      Data data = (Data)obj;
      if (data.type == ASCII) {
        String string = ""+obj;
        set(name, ASCII, string.getBytes(), string.length(), flags);
      } else {
        set(name, data.type, data.buf, data.size*data.bpe, flags);
      }
    }
    else if (obj instanceof String) {
      String string = (String)obj;
      set(name, ASCII, string.getBytes(), string.length(), flags);
    }
    else if (obj instanceof Table) {
      byte[] bbuf = ((Table)obj).toBytes();
      set(name, ASCII, bbuf, bbuf.length, flags);
    }
    else {
      throw new MidasException ("Illegal Keyword Data type: "+obj);
    }
  }

  public void set (String name, byte type, byte[] data, int len, int flags) {
    byte  _rep   = (hcb==null)? Shell.rep: hcb.headRep;
    int   off    = scope_end, olength = 0;    // default no replace
    byte  ltag   = (byte)name.length();
    int   length = (len+ltag+15)/8*8;
    short lext   = (short)(length-len);
    // find current location for replacement
    if ((flags&REPLACE)!=0) {
      Key k = find(name,SCOPE);
      if (k != null) { off = k.offset; olength = k.length; }
    }
    // expand buffer if necessary
    if (buf==null) buf = new byte[EXTSIZE];
    if (size+length-olength > buf.length) {
      byte[] bufo = buf;
      buf = new byte[ (size+length+EXTSIZE)/EXTSIZE*EXTSIZE ];
      System.arraycopy (bufo,0, buf,0, size);
    }
    // create/collapse the hole
    if (length!=olength && off<size) {
      System.arraycopy (buf,off+olength, buf,off+length, size-(off+olength));
    }
    if (type==STRING) type=ASCII;
    Convert.packL(buf,off+0,length);
    Convert.packI(buf,off+4,lext);
    buf[off+6] = ltag;
    buf[off+7] = type;
    if (_rep!=Shell.rep) {
      if (type!=ASCII && type!=BYTE) Convert.rep(data,0,type,1,Shell.rep,_rep);
      Convert.rep(buf,off,LONG,1,Shell.rep,_rep);
      Convert.rep(buf,off+4,INT,1,Shell.rep,_rep);
    }
    System.arraycopy (data,0, buf,off+HEADER_LEN, len);                            // set data
    for (int i=0,j=off+HEADER_LEN+len; i<ltag;) buf[j++] = (byte)name.charAt(i++); // set tag name
    scope_end += length-olength;
    size += length-olength;
  }

  public int delete (String name) {
    if (name.equals("TIMELINE") && hcb!=null) {
      hcb.setTimeLineHandler(null); // <-- Since 2.5.0 (DataFile will re-insert it otherwise)
    }
    return delete(find(name));
  }
  
  /** Deletes a key. */
  @InternalUseOnly
  private int delete (Key k) {  // Since NeXtMidas 3.3.0
    if (k==null) return 0;
    if (k.offset<0) return deleteMain(k.name);
    size -= k.length;  // new length, close the gap
    System.arraycopy(buf,k.offset+k.length, buf,k.offset, size-k.offset);
    if (scope_start>k.offset) scope_start -= k.length;
    if (scope_end>k.offset) scope_end -= k.length;
    return 0;
  }

  /** Deletes all keywords within a given scope. Before returning this method will always call
      {@link #clearScope()}.
      @param scope The scope to delete.
      @return true if deleted, false if not found.
      @since NeXtMidas 2.7.0
   */
  public boolean deleteScope (String scope) {
    int i = setScope(scope);
    if (i > 0) {
      int len = size - scope_end;
      System.arraycopy(buf, scope_start, buf, scope_end, len);
      size = len;
    }
    clearScope();
    return true;
  }

  /** Put a keyword into the MAIN header. If the keyword specified is included in
      {@link #MAIN_HEADER_KEYWORDS}, and the header is full, any keywords in the MAIN header that
      are not in {@link #MAIN_HEADER_KEYWORDS} are subject to automatic deletion as is needed to
      add the new keyword. If a keyword is deleted and <tt>warn=true</tt>, a warning will be issued.
      @param name  The keyword name.
      @param value The value of the keyword.
      @param warn  Issue warnings if (1) the main header is full, or (2) if inserting a keyword in
                   the main header that belongs in the extended header (see {@link #MAIN_HEADER_KEYWORDS}).
   */
  public void putMain (String name, String value, boolean warn) {
    if (name==null || name.length()<=0) return; // check for degenerate cases

    // strip type casting - internal keywords must be Ascii anyway
    if (name.length()>2 && name.charAt(1)==':') name = name.substring(2);

    if (!isOkMain(name) && warn) {
      warn(name+"="+value+" should only be set in the EXTENDED header");
    }
    if (isize<0) { ibuf=hcb.hb; isize=hcb.getKeyLength(); } // auto attach
    deleteMain(name); // no duplicates in main header keywords
    // form the new entry
    String knv = name + "=" + value + NULL;
    int off = isize;
    int len = knv.length();
    if ((off+len>IKMAX) && isReqMain(name)) {
      // Delete as many keywords from MAIN as are necessary to make room. Keep only those listed
      // in MAIN_HEADER_KEYWORDS. Note that it is (remotely) possible that this will not find enough
      // room in the header.
      Table    main = makeTable(true,MAIN);
      String[] keys = main.getKeys();
      for (int i = keys.length-1; (i >= 0) && (off+len>IKMAX); i--) {
        if (!isReqMain(keys[i])) {
          if (warn) warn("No room in MAIN header for "+name+"='"+value+"', "+ "deleting "
                        + keys[i]+"='"+main.getS(keys[i])+"' to make room");
          deleteMain(keys[i]);
          off = isize;
        }
      }
    }
    if (off+len>IKMAX) {
      if (warn) warn("No room in MAIN header keywords for entry: "+name+"='"+value+"'");
      return;
    }
    System.arraycopy (knv.getBytes(),0, ibuf,IKOFF+off, len);
    isize += len;
    hcb.setKeyLength(isize);
  }

  /** Put a keyword into the MAIN header. This is the same as <tt>putMain(name,value,true)</tt>.
      @param name  The keyword name.
      @param value The value of the keyword.
   */
  public void putMain (String name, String value) {
    putMain(name,value,true);
  }

  /** Indicates if the given name is in MAIN_HEADER_KEYWORDS. */
  private static boolean isReqMain (String name) {
    return (Arrays.binarySearch(MAIN_HEADER_KEYWORDS, name) >= 0);
  }

  /** Indicates if the given name is in MAIN_HEADER_KEYWORDS or MAIN_HEADER_IGNORE. */
  private static boolean isOkMain (String name) {
    return isReqMain(name)
        || (Arrays.binarySearch(MAIN_HEADER_IGNORE, name) >= 0);
  }

  /** Handle how a bad keyword block found (e.g. throw MidasException unless
      the {@link #NOABORT} flag has been set. If not aborting, then display
      a warning once per fromMethod.
      @param key         keyword block that is invalid
      @param fromMethod  method where the warning is coming from
      @param moreDetails any additional details/information
      @throws MidasException unless the {@link #NOABORT} flag has been set.
      @since NeXtMidas 3.1.1
   */
  private void handleBadKeywordBlock (Key key, String fromMethod, CharSequence moreDetails) throws MidasException {
    if ((flags&NOABORT)==0) {
      throw new MidasException(fromMethod+" found invalid keyword block ("+moreDetails+") " + key.headerInfo());
    }
    else if (!warnedForInvalidBlockSet.contains(fromMethod)) { // otherwise only warn once per fromMethod
      warnedForInvalidBlockSet.add(fromMethod);
      warn(fromMethod + " invalid block (" + moreDetails + ") "+ key.headerInfo());
    }
  }

  /** Prints out a warning. */
  private void warn (String msg) {
    // String wmsg = "Keywords: " + msg + " file='" + (hcb==null? null : hcb.getName()) + "'."; // prior to 3.1.1
    String wmsg = "Keywords: " + msg + " file='" + hcb + "'."; // since 3.1. since DataFile has more informative toString()
    if ((hcb != null) && (hcb.M != null)) { hcb.M.warning(wmsg); }
    else                                  { Shell.warning(wmsg); }
  }

  /** Get a keyword from the MAIN header.
      Returns the FIRST occurrence of the keyword or NULL if not found. */
  public String getMain (String name) {
    if (isize<0) { ibuf=hcb.hb; isize=hcb.getKeyLength(); } // auto attach
    // check for degenerate cases
    if (isize<=0 || name==null || name.length()==0) return null;
    // strip type casting - internal keywords must be ASCII anyway
    if (name.length()>2 && name.charAt(1)==':') name = name.substring(2);
    Key k = find(name,MAIN);
    if (k==null) return null;
    extractData(k);
    return (String)k.value;
  }

  /** Delete a keyword from the MAIN header. */
  public int deleteMain (String name) {
    if (isize<0) { ibuf=hcb.hb; isize=hcb.getKeyLength(); } // auto attach
    // check for degenerate cases
    if (isize<=0 || name==null || name.length()==0) return 0;
    // strip type casting - internal keywords must be ASCII anyway
    if (name.length()>2 && name.charAt(1)==':') name = name.substring(2);
    Key k = find(name,MAIN);
    if (k==null) return 0;
    int off=k.offset+isize;
    int len=k.ltag+1+k.ldat+1;
    if (off+len<isize) System.arraycopy(ibuf,IKOFF+off+len, ibuf,IKOFF+off, isize-off-len);
    isize -= len;
    hcb.setKeyLength(isize);
    return len;
  }

  /** List a given keyword.
      @param key Keyword to list
      @deprecated Since NeXtMidas 3.3.0 - This function was intended for internal
                  use only and should not be public.
   */
  @Deprecated
  public String list (Key key) {
    // Legacy X-Midas: Use '=' for main and '==' for extended to distinguish the two
    String sep = (key.offset <0) ? "=" : "==";
    if (key.value instanceof String)
      return "S:"+key.name+sep+key.value;
    else if (key.value instanceof Data) {
      Data data = (Data)key.value;
      return (char)data.type+":"+key.name+sep+key.value;
    }
    return "?:"+key.name+"=unknown (invalid keyword block at offset="+key.offset+")";
  }
  
  /** List a given keyword. Intended for use by KEYWORD command and {@link DataFile}.
      @param kwd   Keyword to list
      @param scope The scope used.
      @param pad   The "pretty print" padding value (0=disable).
      @param flags Listing flags (e.g. {@link #UNITS_LIST}).
      @since NeXtMidas 3.3.0
   */
  @InternalUseOnly
  public static CharSequence list (Map.Entry<String,Object> kwd, String scope, int pad, int flags) {
    // Notes:
    //   1. Use '=' for MAIN and '==' for EXT to match X-Midas
    //   2. The type determination is taken from the old list(Key) function
    StringBuilder str = new StringBuilder();
    String        sep = (scope.equals("MAIN"))? "=" : "==";
    String        key = kwd.getKey();
    Object        val = kwd.getValue();
    
         if (val instanceof String) str.append('S');
    else if (val instanceof Data  ) str.append(((Data)val).getFormatTypeChar());
    else                            str.append('?');
    
    str.append(':');
    str.append(key);
    if (pad > 0) {
      while (str.length() < pad) str.append(' ');
      str.append(' ').append(sep).append(' ');
    }
    else {
      str.append(sep);
    }
    str.append(val);
    
    if (((flags&UNITS_LIST)!=0) && (key.endsWith(".UNIT") || key.endsWith(".UNITS"))) {
      // Show units name in parenthesis (the use of ".UNIT" or ".UNITS" comes
      // from section 1.8.1 is the Platinum ICD Vol. 1). It this fails for any
      // reason, just omit it.
      try {
        String units = DataFile.getUnitsVarName(Convert.o2i(val));
        str.append("\t(").append(units).append(')');
      }
      catch (Exception e) {
        // ignore
      }
    }
    return str;
  }

  /** Dump the keywords in the MAIN and EXTENDED header to a text file */
  public void dump (TextFile tf) {
    for (Iterator kwi=iterator(); kwi.getNext();) tf.writeln( kwi.list() );
  }

  /** Make a flat table out of the keywords */
  public Table makeTable () {
    return makeTable(true);
  }

  /** Make a table out of the keywords in the MAIN and EXTENDED header.  Table can be flat or
      have inner tables MAIN and EXTENDED
      @param flat Put main and extended keywords at the top level if true
      @return A table containing all main and extended header keywords
   */
  public Table makeTable (boolean flat) {
    return makeTable(flat,SCOPE);
  }

  /** Make a table out of the keywords in the MAIN and EXTENDED header.  Table can be flat or
      have inner tables MAIN and EXTENDED
      @param flat  Put main and extended keywords at the top level if true.
      @param scope {@link #SCOPE}=Include Current Scope Only, {@link #MAIN}=Include only MAIN header,
                   {@link #EXTENDED}=Include only EXTENDED header, or <tt>MAIN|EXTENDED</tt>=Include
                   both MAIN and EXTENDED header (i.e. include everything).
      @return A table the keywords.
   */
  private Table makeTable (boolean flat, int scope) {
    Table tb = new Table(Table.KV, Table.ADD);
    Table mainTab = null;
    Table extendTab = null;
    if (!flat) {
      mainTab   = new Table(Table.KV, Table.ADD); tb.put("MAIN", mainTab);
      extendTab = new Table(Table.KV, Table.ADD); tb.put("EXTENDED",extendTab);
    }
    for (Iterator kwi=iterator(scope); kwi.getNext();) {
      if (flat) {
        tb.putx(kwi.name, kwi.value);
      } else {
        if (kwi.curentKey.offset < 0) { mainTab.put(kwi.name, kwi.value);   }
        else                          { extendTab.put(kwi.name, kwi.value); }
      }
    }
    return tb;
  }

  /** Load the keywords from a text file using = for main == or extended */
  public void load (TextFile tf) {
    int ie,off,len; char type; String text,name;
    while ( (text=tf.readProper()) != null) {
      off=0; len=text.length(); type='_';
      if (text.charAt(1) == ':') {
        type = text.charAt(0);
        if (type=='A') type='S';
        off = 2;
      }
      ie = text.indexOf('=');
      // prior to 3.1.2 this check was not present and an ArrayIndexOutOfBoundsexception would be thrown
      if (ie<0) throw new MidasException("Invalid entry "+text+" found in "+tf.getName());
      name = text.substring(off,ie);
      if (((ie+1)<len) && (text.charAt(ie+1)=='=')) {	// extended header
        text = text.substring(ie+2);
        if (type=='S') add(name,text);
        else add(name,Convert.s2o(text, type, null));
      }
      else {						// main header
        text = text.substring(ie+1);
        putMain(name,text);
      }
    }
  }

  /** Load the keywords from a text file */
  public void loadExt (TextFile tf) {
    int ie,off,len; char type; String text,name;
    while ( (text=tf.readProper()) != null) {
      off=0; len=text.length(); type='_';
      if (text.charAt(1) == ':') {
        type = text.charAt(0);
        if (type=='A') type='S';
        off = 2;
      }
      ie = text.indexOf('=');
      if (ie<0) throw new MidasException("Invalid entry "+text+" found in "+tf.getName());
      name = text.substring(off,ie);
      if (++ie<len && text.charAt(ie)=='=') ie++;
      if (ie>=len) text=""; else text = text.substring(ie);
      if (type=='S') add (name,text);
      else add(name, Convert.s2o(text,type,null) );
    }
  }

  /** @deprecated Since NeXtMidas 2.5.3: This method should not be called directly,
                  use {@link DataFile#convertHeaderRep}.
   */
  @Deprecated public void convertRep (byte from, byte to) {
    convertExtRep(from,to);
    rep = to;
  }

  /** Convert the representation of the keywords in the extended header.
      @param from From representation
      @param to   To representation
   */
  void convertExtRep (byte from, byte to) { // package-private, used by DataFile
    Key k=new Key();
    for (k.offset=0; k.offset<size; k.offset+=k.length) {
      String invalidDetails = unpack(k);
      if (invalidDetails != null) { // since 3.1.1 (will get here is NoAbort is set)
        if (k.length >= HEADER_LEN) continue; // try next keyword block
        else                        break;    // bad length, can't recover from this (break out of loop)
      }
      int off=k.offset;
      Convert.rep(buf,off,LONG,1,from,to);
      Convert.rep(buf,off+4,INT,1,from,to);
      if (k.type!=ASCII && k.type!=BYTE)
      Convert.rep(buf, off+HEADER_LEN, k.type,k.ldat/Data.getBPS(k.type),from,to);
    }
  }

  /** Always returns null at this time. */
  @Override
  public String[] getKeys() { return null; }

  /** Sets the scope (if name="SCOPE") otherwise simply does a put.
      @param name  Key name
      @param value Key value
   */
  @Override
  public Object setKey (String name, Object value) {
    if (name.equals("SCOPE")) setScope((String)value);
    else put(name,value);
    return value;
  }

  /** Gets the scope (if name="SCOPE") otherwise simply does a get.
      @param name  Key name
   */
  @Override
  public Object getKey (String name) {
    if (name.equals("SCOPE")) return scope;
    else return get(name);
  }

  /** Class to hold a keyword */
  public static class Key {
    /** type of data in keyword data/value  */
    byte type;

    /** if >=0, offset is relative to start of extended header;
        if < 0, offset is relative to end of keywords string in main header.
     */
    int  offset;

    /** Total length of "total" keyword block (lkey) */
    int  length;
    /** Length of keyword tag */
    int  ltag;
    /** Length of keyword data (value) */
    int  ldat;

    /** The keyword name/tag */
    public String name;

    /** The keyword value/data */
    public Object value;

    /** Gets the keyword type. */
    public char getType() {
      return (char)type;
    }

    /** String representation of keyword */
    @Override
    public String toString() {
      String t = (type >= 'A' && type <= 'Z')? Character.toString((char)type): "n/a";
      return "Key: type="+t+" offset="+offset+" length="+length+" ltag="+ltag
            +" ldat="+ldat+" name="+name+" value="+value;
    }

    /** String representation of keyword header info (+offset) for warning/errors messages
        @since 3.1.1
     */
    private String headerInfo () {
      String t = (type >= 'A' && type <= 'Z')? Character.toString((char)type): "n/a";
      return "Key: lkey="+length +" lext="+(length-ldat) +" ltag="+ltag +" type="+t +" offset="+offset +" ldat="+ldat;
    }
  } // end inner class Key

  /** String representation of all keywords */
  @Override
  public String toString() {
    StringBuilder keywordsString = new StringBuilder();
    for (Iterator kwi=iterator(); kwi.getNext();) {
      if (kwi.value instanceof String)
        keywordsString.append("S:");
      else if (kwi.value instanceof Data) {
        Data data = (Data)kwi.value;
        keywordsString.append((char)data.type).append(":");
      }
      keywordsString.append(kwi.name).append("=").append(kwi.value).append("\n");
    }
    return keywordsString.toString();
  }

  /** Gets an iterator that iterates over the keywords within the current scope. */
  public Iterator iterator () {
    return iterator(SCOPE);
  }

  /** Gets an iterator that iterates over the keywords within a specified scope. This is effectively
      the same as doing a <tt>setScope(..)</tt> followed by a call to <tt>iterator()</tt>.
      @param scope The scope to get an iterator for.
      @return The applicable iterator or null if the given scope is not found.
      @since NeXtMidas 2.7.0
   */
  public Iterator iterator (String scope) {
    int i = setScope(scope);
    return (i <= 0)? null : iterator(SCOPE);
  }

  /** Gets an iterator that iterates over the keywords.
      @param scope {@link #SCOPE}=Include Current Scope Only, {@link #MAIN}=Include only MAIN header,
                   {@link #EXTENDED}=Include only EXTENDED header, or <tt>MAIN|EXTENDED</tt>=Include
                   both MAIN and EXTENDED header (i.e. include everything).
      @since NeXtMidas 2.5.3
   */
  private Iterator iterator(int scope) {
    int start = 0;
    int end   = 0;

    if (scope == SCOPE)      { start = scope_start;             end = scope_end;           }
    if ((scope&MAIN)!=0)     { start = Math.min(start, -isize); end = Math.max(end, 0);    }
    if ((scope&EXTENDED)!=0) { start = Math.min(start, 0);      end = Math.max(end, size); }

    return new Iterator(this, start, end);
  }

  /** Iterator through the keyword block */
  public static class Iterator implements java.util.Iterator<Key> {
    public Keywords kw;
    public String name;
    public Object value;
    private int end;
    /** current keyword block location (0 for first one) */
    private int atBlock = 0;

    private Key curentKey = new Key();

    public Iterator (Keywords kw) {
      this(kw, kw.scope_start, kw.scope_end);
    }

    Iterator (Keywords kw, int start, int end) {
      this.kw          = kw;
      this.end         = end;
      curentKey.offset = start;
      curentKey.length = 0;
    }
    @Override
    public boolean hasNext() {
      if (atBlock > 0 && curentKey.length < 1) {
        return false; // since 3.1.1 for BUG 2548 (break out if lkey is zero or negative
      }
      return curentKey.offset+curentKey.length < end;
    }
    /** @return next keyword block, null if at end -OR- if invalid block is found */
    @Override
    public Key next() {
      curentKey.offset += curentKey.length;
      if (curentKey.offset>=end) return null;
      String invalidDetails = kw.unpack(curentKey);
      atBlock++;
      if (invalidDetails != null) { // since 3.1.1 (will get here is NoAbort is set)
        curentKey.name  = null; // reset field
        curentKey.value = null; // reset field

        if (curentKey.length >= HEADER_LEN) {
          // might be able to recover from this, so let it continue to below
        }
        else { return null; } // bad length, can't recover from this (return)
      }
      kw.extractName(curentKey);
      kw.extractData(curentKey);
      return curentKey;
    }
    @Override
    public void remove() {
      kw.delete(name);
    }
    public boolean getNext() {
      if (!hasNext()) return false;
      next();
      name  = curentKey.name;
      value = curentKey.value;
      return true;
    }
    public String list() {
      return kw.list(curentKey);
    }
  }
}
