package nxm.sys.lib;

import nxm.sys.inc.DataTypes;

/** A class to interface to Native function libraries.
    Allows for the turning OFF native library use.  Turning ON use of native
    libraries will have no affect if we can not load them anyway!

  @author Jeff Schoen/ Jay Hallam
  @version $Id: Native.java,v 1.37 2012/01/24 16:21:53 ntn Exp $
*/
public class Native {

  static final int    ROUND_DOWN        = Convert.ROUND_DOWN;        // Used from native code
  static final int    ROUND_CEILING     = Convert.ROUND_CEILING;     // Used from native code
  static final int    ROUND_FLOOR       = Convert.ROUND_FLOOR;       // Used from native code
  static final int    ROUND_HALF_UP     = Convert.ROUND_HALF_UP;     // Used from native code
  static final int    ROUND_HALF_EVEN   = Convert.ROUND_HALF_EVEN;   // Used from native code
  static final int    ROUND_UNNECESSARY = Convert.ROUND_UNNECESSARY; // Used from native code
  static final int    ROUND_ROUND       = Convert.ROUND_ROUND;       // Used from native code
  static final double DMAX              = Double.MAX_VALUE;          // Used from native code
  static final float  FMAX              = Float.MAX_VALUE;           // Used from native code
  static final long   XMAX              = Long.MAX_VALUE;            // Used from native code
  static final int    LMAX              = Integer.MAX_VALUE;         // Used from native code
  static final short  IMAX              = Short.MAX_VALUE;           // Used from native code
  static final byte   BMAX              = Byte.MAX_VALUE;            // Used from native code
  static final byte   NMAX              = DataTypes.NMAX;            // Used from native code

  static final byte   IEEE              = DataTypes.IEEE;            // Used from native code
  static final byte   EEEI              = DataTypes.EEEI;            // Used from native code
  static final byte   VAX               = DataTypes.VAX;             // Used from native code

  /** comma-separate list of valid native code that can be used (i.e. for {@link #setUse(String)}. */
  public static final String useList = "IO,Math,Cast,CastX,Convert,Cuda,Other";
  /** constant options for which native code that can be used in {@link #setUse(int)} */
  public static final int IO=0x1, MATH=0x2, CAST=0x4, CASTX=0x8, CONV=0x10, CUDA=0x20, OTHER=0x40;
  public static boolean canUse, useIO, useMath, useCast, useCastX, useConvert, useCuda, useOther;

  private static int use, useable;

  public static void setCanUse (boolean state) {
    canUse = state;
    setUse(use);
  }
  public static void setUse (String usestr) {
    setUse( Parser.mask(useList,usestr,use) );
  }
  public static void setUse (int mask) {
    use        = (mask & useable);
    useIO      = canUse && (use&IO)!=0;
    useMath    = canUse && (use&MATH)!=0;
    useCast    = canUse && (use&CAST)!=0;
    useCastX   = canUse && (use&CASTX)!=0;
    useConvert = canUse && (use&CONV)!=0;
    useCuda    = canUse && (use&CUDA)!=0;
    useOther   = canUse && (use&OTHER)!=0;
  }
  public static String getUse() {
    return Parser.mask2s (useList,use);
  }
  public static String getUseable() {
    return Parser.mask2s (useList,useable);
  }

  /** initialize the usage of native method groups */
  public static void initUse() {
    canUse = Shell.getPriv(Privileges.PRIV_NATIVE);
    if (canUse) {
      String noNative = Shell.getProperty("NONATIVE");
      canUse = (noNative == null) || StringUtil.isFalse(noNative);
    }
    canUse = canUse && Shell.loadLibrary ("sys","lib","Native");
    useable = 0;
    if (canUse) {
      useable |= (IO|MATH|CONV|OTHER);
      if (castInit(new float[10],new double[11],new byte[12],new short[13],new int[14],
	           new float[20],new double[21],new byte[22],new short[23],new int[24])==1) useable |= CAST;
      if (cudaInit()==1) useable |= CUDA;
    }
    setUse("ALL");  // initially try for everything
    setUse("-CUDA");  // except CUDA
  }

  public static native double[] castD (byte[] array);
  public static native float[] castF (byte[] array);
  public static native int[] castL (byte[] array);
  public static native short[] castI (byte[] array);

  /** Revert previously cast array to a byte array */
  public static native byte[] uncastD (Object array);
  public static native byte[] uncastF (Object array);
  public static native byte[] uncastL (Object array);
  public static native byte[] uncastI (Object array);

  private static native int castInit (Object F, Object D, Object B, Object I, Object L,
  				      Object f, Object d, Object b, Object i, Object l);
  private static native int cudaInit ();

  /** package-private, only used by Convert, parameters match those in Convert.type. */
  static native int type (Object buf1, int off1, byte type1,
                          Object buf2, int off2, byte type2, int size,int round,boolean saturating);

  /** package-private, only used by Convert, parameters match those in
      {@link Convert#rep(long, int, byte, int, byte, byte)}.
      @since NeXtMidas 2.7.0
   */
  static native int rep (long lbuf, int off, byte type, int n, byte repa, byte repb);

  /** package-private, only used by Convert on sub-record of interest (soi), parameters
      match those in {@link Convert#rep(long, int, byte, int, byte, byte, int, int)}.
      @since NeXtMidas 2.7.0
   */
  static native int repsoi (long lbuf, int off, byte type, int n, byte repa, byte repb, int nn, int inc);

  /** @deprecated Since NeXtMidas 2.5.2: Use the methods in convert rather than calling this method
      directly. This method will cause an error if a native conversion is not available.
   */
  @Deprecated public static int cast (Object buf1, int off1, byte type1,
                          Object buf2, int off2, byte type2, int size) {
    return type(buf1,off1,type1,buf2,off2,type2,size,Convert.ROUND_ROUND,true);
  }

  /** dumps the contents of a class in hex */
  public static native void dumpClass (Object obj, int words);

  /** get the memory address of different areas of an object */
  public static native long getAddress (Object obj, int item);

  /** Converts a float array to 10 bit dB using fast exponent algorithm */
  public static native void dbf (Object array, int size, int flags);

  /** Perform a single precision square root function */
  public static native float sqrtf (float value);

  /** Perform a single precision natural logarithm function */
  public static native float logf (float value);

  /** Copy contents of one array to another */
  public static native void copy (Object array1, Object array2, int bytes);

  /** Copy data from a native pointer to a Java array */
  public static native void p2ja (long pointer1, int off1, Object array2, int off2, int bytes);

  /** Copy data from a Java array to a native pointer */
  public static native void ja2p (Object array1, int off1, long pointer2, int off2, int bytes);

  /** Copy data from a native pointer to a native pointer */
  public static native void p2p (long pointer1, int off1, long pointer2, int off2, int bytes);

  /** Open a file for read|write */
  public static native int fopen (String fname, int dir);

  /** Move current read/write pointer in a file */
  public static native long fseek (int fd, long offset);

  /** Read from a file into a pointer.
      NOTE: The return value intentionally matches the native OS's file read(..) API,
      i.e. it DOES NOT match the return value of the IOResource's read(..).
      @param fd    Native file descriptor (from return value of {@link #fopen(String, int)}.
      @param lbuf  Native pointer to buffer.
      @param boff  The offset into <tt>buf</tt> to start at.
      @param bytes The number of bytes to try reading.
      @return number of bytes read, 0 for End-Of-File (EOF), -1 for error.
    */
  public static native int fread (int fd, long lbuf, int boff, int bytes);

  /** Write to a file from a pointer */
  public static native int fwrite (int fd, long lbuf, int boff, int bytes);

  /** Read from a file into a java array.
      NOTE: The return value intentionally matches the native OS's file read(..) API,
      i.e. it DOES NOT match the return value of the IOResource's read(..).
      @param fd   file descriptor (from return value of {@link #fopen(String, int)}.
      @param buf  byte array buffer to read data into.
      @param boff byte offset into the byte array (buf) to start writing at.
      @param bytes The number of bytes to try reading.
      @return number of bytes read, 0 for End-Of-File (EOF), -1 for error.
   */
  public static native int freadb (int fd, Object buf, int boff, int bytes);

  /** Write to a file from a java array */
  public static native int fwriteb (int fd, Object buf, int boff, int bytes);

  /** Close a native file */
  public static native void fclose (int fd);

  /** Change working directory */
  public static native int chdir (String path);

  /** Get system file system stats.
      @return 0 (zero) on success, -1 on error.
    */
  public static native int statfs (String path, Object buf);

  /** Get system CPU, memory, and process stats */
  public static native int top (int flags, Object buf, int boff, int bytes);

  /** This will copy a Java String on to an existing Fortran string. This works
      by converting the Java String to a UTF-8 C string (<code>char*</code>) and
      then copying each of the letters in the C string to the location of the
      Fortran string; effectively overwriting the existing Fortran string with
      the letters in the C string.<br>
      <br>
      When copying the C string to the location of the Fortran string no more
      than <code>lfstr</code> letters of the C string will be copied over (this
      is to prevent writing outside the allocated memory of the Fortran). Thus
      it is vital that <code>lfstr</code> be less than or equal to the actual
      number of bytes that have been allocated to the Fortran string.<br>
      <br>
      Note that if the actual length of the C string (as determined by the C
      function <code>strlen(char*)</code>) is less than <code>lfstr</code> then
      the number of characters copied from the C string to the Fortran string
      will be equal to the length of the C string.

      @param jstr  The Java String to copy onto the Fortran string.

      @param fstr  The native pointer to the location of the Fortran string.

      @param lfstr The maximum string length to copy to the Fortran
                   string. <b>(This must be less than or equal to the number
                   of bytes allocated to the Fortran string.)</b>

      @return The number of characters actually copied to the Fortran string.
              This will be equal to the lesser of <code>lfstr</code> or the
              number of characters in the String after being converted to a
              UTF-8 C string (as determined by <code>strlen(char*)</code>).
  */
  public static native int jstr2fstr (String jstr, long fstr, int lfstr);

}
