package nxm.sys.lib;

import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;

import nxm.sys.inc.*;
import nxm.sys.libg.MidasDisplay;

/** A class to run NeXtMidas command(s) from outside of a Midas context
    <br>
    Note: This class should be treated as a singleton, since various underlining
          NeXtMidas framework classes (such as the NeXtMidas Shell) uses
          global variables (such as the PATH, AUXes, etc).
          Otherwise multiple instances can step over each other and cause undefined behavior.
    <pre>
    <b>Example Usage</b>:
    One-shot" mode (do not instantiate this class):
       [prompt]$ java -classpath $NMROOT nxm.sys.lib.NeXtMidas -path="UCL,DSP,SYS" -cmd="home test sys" -script="$NMROOT/nxm/sys/test/test_plot.mm" -prompt

    Instantiate (in java) and run multiple commands:
      Method 1:
        NeXtMidas nxm = NeXtMidas.getGlobalInstance();
        nxm.runCommand("SET AUX.MYDATA /path/to/mydata");  // define MYDATA AUX
        nxm.runCommand("AUX HOME HOME|MYDATA|DAT|RAM");    // setup write & read AUX list
        nxm.runCommand("OPTION DSP CLASSPATH");            // find & define DSP from CLASSPATH
        nxm.runCommand("OPTION MYOPT /path/to/nxm/myopt"); // define MYOPT option tree
        nxm.runCommand("PATH ADD DSP MYOPT");              // add option trees to PATH
        nxm.runCommand("DEMO/BG");                         // run command in background (non-blocking)

      Method 2:
        String[] opts = {"-path=UCL,DSP,SYS", "aux.read=HOME|DAT|"};
        NeXtMidas nxm = new NeXtMidas(opts); // this calls parseOptions(.) and openShell() for you
        nxm.run(); // <-- [optional] run commands specified via any -cmd, -command, and -script options
        nxm.runCommand("plot/bg Sunset.jpg"); // run PLOT command in background (non-blocking)
        nxm.runCommand("plot world.shp");     // run another PLOT command in foreground (blocks current thread)

      Method 3:
        NeXtMidas nxm = new NeXtMidas();
        String[] opts = {"-path=UCL,DSP,SYS", "aux.read=HOME|DAT|RAM", "-cmd=plot Sunset.jpg"};  // also gives a command to run later
        nxm.parseOptions(opts);
        nxm.openShell();
        nxm.run(); // run commands specified via any -cmd, -command, and -script options
        nxm.runCommand("plot nxm.sys.dat.world.prm"); // run another PLOT command in foreground (blocks)
</pre>
    @author J. Bauer / Neon Ngo
    @version $Id: NeXtMidas.java,v 1.37 2012/11/05 18:50:33 ntn Exp $
    @since NeXtMidas 2.9.0
 */
public class NeXtMidas implements ShellParent {

  private boolean DEBUG = false;
  private static final String DEF_WRITE_AUX = "HOME";
  private static final String DEF_READ_AUX  = "HOME|RAM|DAT";

  /** global NeXtMidas session instance (using Java Terminal and default options) */
  private static final NeXtMidas INSTANCE = new NeXtMidas(new Terminal(), null); // since 3.3.0

  /** use Native code? (DEF=off) */
  private boolean   noNative      = true;
  private String    path          = "SYS";
  /** The shell driving this NeXtMidas instance */
  private Shell     shell         = null;

  private boolean   runInteractiveTextTerminal = false;

  private List<String> executeList = Collections.synchronizedList(new ArrayList<String>());

  /** display panel (when set by user) to embed graphical commands on */
  private MidasDisplay panel;

  /** encountered "--" so stop processing options */
  private boolean      haltOptionsCheck = false;

  /** String Escape mode (enum):
      Percent - translate escape sequences as defined in section 2.1 Percent-Encoding
                of RFC 3986 Uniform Resource Identifier (URI): General Syntax.
                i.e. %hh (here h is hex digit).
   */
  private static enum EscapeCharMode { Percent };
  EscapeCharMode escapeCharMode = null;

  Map<String,String> auxEntries = new LinkedHashMap<String, String>();
  /** terminal to use, or null to use Terminal determined by Shell when it opens */
  private Terminal terminal = null;

  /** Creates a new instance of this class.
      User <b>SHOULD</b> call {@link #parseOptions(String[])} with any desired settings.
      User <b>MUST</b> call {@link #openShell()} prior to running any commands via {@link #runCommand(String)} or {@link #run()}
      Since 3.3.0: this uses the nxm.sys.lib.Terminal (pure Java) vs trying NTerminal (if -nonative=false)
      which will replace SIGINT (Ctrl+C) signal handler.
   */
  public NeXtMidas () {
    this(null);
  }

  /** Create a new instance of this class, then calls {@link #parseOptions(String[])} and {@link #openShell()}.
      Since 3.3.0: this uses the nxm.sys.lib.Terminal (pure Java) vs trying NTerminal (if -nonative=false)
      which will replace SIGINT (Ctrl+C) signal handler.
      @param options the options to set
  */
  public NeXtMidas (String[] options) {
    this(new Terminal(), options);
  }

  /** Create a new instance of this class, then calls {@link #parseOptions(String[])} and {@link #openShell()}.
      @param terminal the terminal to use (set to null to have Shell determine it when it opens)
      @param options the options to set (null or empty for none)
      @since NeXtMidas 3.3.0
   */
  private NeXtMidas (Terminal terminal, String[] options) {
    DEBUG = Convert.o2z(System.getProperty("nextmidas.debug"));
    noNative = Convert.o2z(System.getProperty("NONATIVE", "TRUE")); // native code is disabled by default until BUG 2124 is fixed
    setNoNative(noNative);
    System.setProperty("NOEXTERNAL", "TRUE"); // ensure OS commands run as internal

    this.shell = new Shell(this);

    // set write and read AUX (MUST be after shell instance is created above)
    setWriteAux(DEF_WRITE_AUX);  // set to default
    setReadAux(DEF_READ_AUX);    // set to default

    this.terminal = terminal; // MUST set terminal before calling openShell()

    if (options != null) {
      parseOptions(options);
    }

    openShell();
  }

  /** Return global NeXtMidas session instance (using default options).
      @since NeXtMidas 3.3.0
   */
  public static NeXtMidas getGlobalInstance () {
    return INSTANCE;
  }

  /** Run a NeXtMidas command.
      @param args The command to run plus any NeXtMidas parameters to be set
   */
  public static void main (String[] args) {
    NeXtMidas nxm = new NeXtMidas(null, args);
    if (nxm.executeList.isEmpty()) { // if no commands are given, then default to -shell for BUG 2317
      nxm.determineInteractiveShell(false);
    }
    nxm.run(); // run with parsed command-line options
  }

  /** run with parsed command-line options (which includes running commands/scripts).
      @see #runCommand(String)
   */
  public void run () {
    for (String cmd : executeList ) {
      runCommand(cmd);
    }
    if (runInteractiveTextTerminal) {
      shell.process(); // this will start the interactive text terminal shell
    }
  }

  /** open/initialize shell. MUST be called before {@link #run()} or {@link #runCommand(String)} */
  public void openShell () {
    // MUST set Shell Privileges: set to ALL for now
    Shell.setPriv(Privileges.PRIV_ALL);  // TODO: extract setting privileges from Shell.run() so that we can use it here?
    shell.open(); // initialize shell (this call's shell.boot())
    if (auxEntries.size() > 0) updateAuxList();   // re-apply any user-set aux's. (Bug 2325)
  }

  /** Run a NeXtMidas command.
      WARNING: DO NOT run a graphical NeXtMidas command like (LIST2 or PLOT) from
               Event-Dispatch-Thread (EDT) without the /BG switch otherwise it can lockup the JVM.
      @param cmd command line to run
      @return reference to the Command that is executed (null if cmdline is empty)
   */
  public Command runCommand (String cmd) {
    return runCommand(cmd, panel);
  }

  /** Run a NeXtMidas command with option to embed a graphical command inside specified panel (@link MidasDisplay).
      WARNING: DO NOT run a graphical NeXtMidas command like (LIST2 or PLOT) from
               Event-Dispatch-Thread (EDT) without the /BG switch otherwise it can lockup the JVM.
      @param cmd command line to run
      @param panel MidasDisplay (panel) to embed graphical command in
      @return reference to the Command that is executed (null if cmdline is empty)
      @since NeXtMidas 3.3.0
   */
   public Command runCommand (String cmd, MidasDisplay panel) {
     if (DEBUG) debug("running command: [["+cmd+"]]"+ (panel==null? "" : " on panel = "+panel));
     //shell.run(this, cmd, false);    // NOTE: CANNOT USE THIS since it calls shell's open()/boot() which causes values to get reset
     //Shell.runNew(this, cmd, false); // NOTE: same as above
     Command command = shell.runCommand(cmd, panel);
     return command;
   }

  public void setNoNative (boolean value) {
    noNative = value;
    System.setProperty("NONATIVE", Boolean.toString(noNative));
  }

  public boolean getNoNative () {
    return noNative;
  }

  /** Replace any of the standard NeXtMidas separators (',' or '|'), Unix
      separators (':') and DOS separators (';') with the specific separator
      required here. This allows the user to use the separator supported by
      their shell without worrying about NeXtMidas not supporting it.
      @param value The value with separator characters in it
      @param sep The separator uses internally
      @return The value with OS separators replaced with the given separator character
   */
  private static String replaceSeparators (String value, char sep) {
    return value.replace('|',sep).replace(',', sep).replace(':', sep).replace(';', sep);
  }

  /** Parse the input for any NeXtMidas context settings (options) and commands
      @param options string array of options (SHOULD NOT contain a null item)
   */
  public void parseOptions (String[] options) {
    int ii = 0; // DEBUG
    for (String option : options) {
      if ("--".equals(option)) {
        haltOptionsCheck = true;
      }
      if (DEBUG) { debug("  parseOptions() checking arg "+ (ii++) + ": [["+option+"]]"); }
      parseOption(option);
    } // end for loop for processing options
    haltOptionsCheck = false;  // reset   TODO: this might not always need/want resetting
    escapeCharMode = null;     // reset   TODO: this might not always need/want resetting

    if (DEBUG) { debug("  Commands to execute list: "+executeList); }
  }

  /** Parse the input for any NeXtMidas context settings (options) and commands
      @param options Table of options (SHOULD NOT contain a null item)
   */
  public void parseOptions (Table options) {
    String[] keys = options.getKeys();
    String[] args = new String[keys.length];
    int index     = 0;

    for (String key : keys) {
      Object value = options.get(key);
      if (value==null) args[index++] = "-"+key;
      else             args[index++] = "-"+key+"="+value;
    }
    parseOptions(args);
  }

  /** Parse a 'single' option or command. Options and commands start with a '-'. <br>
      Example:<br>
      <pre>parseOption("-cmd=plot world")</pre>
      @param option The option
   */
  public void parseOption (String option) {
    if (haltOptionsCheck) {
      // TDB
      // 1. rest of options array (args) would be command and it arguments (for #1aa) ?
      // 2. rest of options array (args) would be arguments (for MyApp.jar for #1b)  ?
    } else {
      String key = option; // name of option
      String value = null; // value (if
      int indexEQ = option.indexOf('=');
      if (indexEQ>0) {                                // have equal sign (get value after it)
        key = option.substring(0,indexEQ).trim();     // trim
        value = option.substring(indexEQ + 1).trim(); // +1 to skip equal sign
      }
      setOption(key, value);
    }
  }

  /** Set a single option or command.
      @param key The supported option or command
      @param value The value to set it to
   */
  public void setOption (String key, String value) {
    if (key.charAt(0)!='-') key = "-"+key;  // input key can use '-' or not

    // decode any escapes in the parameter's value
    if (escapeCharMode == EscapeCharMode.Percent) {
      value = StringUtil.decodePercentEncoding(value);
    }

    // process the parameters
    if      ("-aux.read".equalsIgnoreCase(key)) {
      setReadAux(value);
    }
    else if ("-aux.write".equalsIgnoreCase(key)) {
      setWriteAux(value);
    }
    else if ((key.length() > 5) && "-aux.".equalsIgnoreCase(key.substring(0,5))) {
      String auxName = key.substring(5).toUpperCase();
      auxEntries.put(auxName, value);
    }
    else if ("-nonative".equalsIgnoreCase(key)) {
      setNoNative( (value==null) || Convert.o2z(value) );
    }
    else if ("-headless".equalsIgnoreCase(key)) {
      Boolean headless = ( (value==null) || Convert.o2z(value) );
      try {
        System.setProperty("HEADLESS", headless.toString());
      } catch (Exception e) {
        // NOTE: This is an error instead of a warning because we need to fail-fast
        // if we cannot set headless. This avoids situations later on where the graphics
        // init can hang for 30+ seconds which causes other problems
        throw new RuntimeException("Cannot set HEADLESS value", e);
      }
    }
    else if ("-path".equalsIgnoreCase(key)) {
      path = value;
    }
    else if ("-command".equalsIgnoreCase(key) || "-cmd".equalsIgnoreCase(key)) {
      executeList.add(value);
    }
    else if ("-shell".equalsIgnoreCase(key)) {
      determineInteractiveShell(false);
    }
    else if ("-prompt".equalsIgnoreCase(key)) {
      determineInteractiveShell(true); // force text terminal
    }
    else if ("-script".equalsIgnoreCase(key)) {
      if (value != null)  executeList.add(getMacroCmd(value));
    }
    else if ("-debug".equalsIgnoreCase(key)) {
      executeList.add("debug/quiet off all");   // clear any debug from any previous cmds
      String debugCmd;;
      if (value==null) debugCmd = "debug on deprecate";  // default value
      else debugCmd = "debug/quiet on "+value;
      executeList.add(debugCmd);
    }
    else if ("-encoding".equalsIgnoreCase(key)) {
      if      (value==null || value.length() == 0) escapeCharMode = null;
      else if (value.equalsIgnoreCase("None"))     escapeCharMode = null;
      else if (value.equalsIgnoreCase("Percent"))  escapeCharMode = EscapeCharMode.Percent;
      else if (value.equalsIgnoreCase("%"))        escapeCharMode = EscapeCharMode.Percent;
      else throw new IllegalArgumentException("Illegal parameter " +key+"="+value);
    }
    else if ("-version".equalsIgnoreCase(key) || "-v".equalsIgnoreCase(key)) {
      // Print out NXM version to System.out and exit (similar to "java -version")
      System.out.println("NeXtMidas version "+Installation.NM_VERSION+"");
      System.exit(0);
    }
    else if ("-help".equalsIgnoreCase(key) || "-h".equalsIgnoreCase(key) || "-?".equals(key)) {
      // print usage to System.out and exit (similar to "java -help")
      System.out.println(getUsage());
      System.exit(0);
    }
    else {
      throw new IllegalArgumentException("Illegal parameter " +key+"="+value);
    }
  }

  /** Get help on usage of this class (showing available options). Similar to java -help.
      @return help string on how to use this class with all available options
      @since NeXtMidas 3.1.3
   */
  public static String getUsage () {
    // NOTE: wrap output text at or before 80th character
    return "Usage: [-options]*\n"
         + "\n"
         + "where options include:\n"
         + "  -aux.<auxname>=<value>  Define a disk AUX entry to a path or URL\n"
         + "  -aux.read=<auxlist>     Set the read AUX list\n"
         + "  -aux.write=<aux>        Set the write AUX\n"
         + "  -cmd=<command>          Alias for -command\n"
         + "  -command=<command>      A command with any necessary arguments to execute.\n"
         + "  -debug=<value>          Set debug to masked value (deprecate, trace, etc.)\n"
         + "  -encoding=<type>        Set the type of escape char enocding (\"NONE\", \n"
         + "                          \"Percent\", or \"%\")\n"
         + "  -headless=[true|false]  Set HEADLESS mode\n"
         + "  -help, -h, -?           Print this usage help message and exit\n"
         + "  -nonative=[true|false]  Disable(true)/enable(false) native code\n"
         + "  -path=<path>            Set the PATH\n"
         + "  -prompt                 Enter the NeXtMidas 'nM>' terminal shell prompt\n"
         + "                          (executed after all commands and/or scripts)\n"
         + "  -script=<macro>         Run a NeXtMidas macro from HOMEPATH, unless using an\n"
         + "                          absolute path, \"nxm\" dot syntax, or a URL\n"
         + "  -shell                  Open a NeXtMidas SHELLGUI (or -prompt if\n"
         + "                          graphics is not available)\n"
         + "  -version, -v            Print out the NeXtMidas version and exit\n"
         + "\n"
         + "Note: Multiple -command and -script options can be specified to run a series of\n"
         + "      commands. They will be executed in the order specified (left to right).\n"
         ;
  }

  /** determine which interactive shell to use. If not in headless mode, then use
      SHELLGUI otherwise use OS text terminal shell.
      @param forceTextTerminal true to force using text terminal, false to auto-detect
   */
  private void determineInteractiveShell (boolean forceTextTerminal) {
    if (forceTextTerminal || Shell.isHeadless()) {
      runInteractiveTextTerminal = true;
      executeList.add(""); // since 3.1.3: need to give blank command otherwise -prompt will run SHELLGUI first, then interactive text shell
    }
    else {
      executeList.add("SHELLGUI;SYS");
    }
  }

  /** Set READ AUX list. Requires Shell instance! */
  public void setReadAux (String value) {
    auxEntries.put("READ", replaceSeparators(value, '|'));
    updateAuxList();
  }

  /** Set WRITE AUX. Requires Shell instance! */
  public void setWriteAux (String value) {
    auxEntries.put("WRITE", value);
    updateAuxList();
  }

  /** Update AUX entries. Requires this.shell instance! */
  private void updateAuxList () {
    for (Map.Entry<String, String> e : auxEntries.entrySet() ) {
      // The next line follows what is done for CWD within Shell.configure()
      shell.M.results.put("AUX."+e.getKey(), e.getValue(),Table.ADDROOT);
    }
  }

  private String getMacroCmd (String url) {
//  FileName fn = new FileName(url); // prior to 3.1.2 (does not work for -script=nxm.sys.test... when Shell has not opened.)
//  String macroName = fn.getRoot(); // TODO: remove after 3.2.x
    String macroName = FileName.getRoot(url); // since 3.1.2: fix issue with NXM dot syntax before Shell has opened
    String cmd="%"+macroName+" /MFN="+url;
    return cmd;
  }

  public String getNmRoot () {
    String nmroot = OptionTree.findOptRoot("SYS");
    if (nmroot.startsWith("file:") && !(nmroot.contains(".jar"))) nmroot = nmroot.substring(5,nmroot.length());
    return nmroot;
  }

  public String getAux () {
    String write = Convert.o2s(shell.M.results.get("AUX.WRITE"));
    String read  = Convert.o2s(shell.M.results.get("AUX.READ"));
    return write+" "+read;
  }

  public String getOpts () {
    return OptionTree.getOptsHelper(path, this, false);
  }

  public String getCWD () {
    String cwd = System.getProperty("user.dir");
    if (cwd != null) cwd= FileUtil.terminatePath(cwd);
    return cwd;
  }

  public String getNMS () {
    // return null; // prior to 3.1.3: TODO: cleanup in 3.3.x or later beta series
    return Shell.getProperty("NMSTARTUP.USER"); // since 3.1.3 as part of BUG 1447
  }

  public String getHome () {
    String home = System.getenv("HOME");
    if (home!=null) home = FileUtil.terminatePath(home);
    return home;
  }

  public String getServer () {
    return null;
  }

  /** Gets theme based on "ENV.THEME" system property, otherwise returns null
      to use system defaults.
      @since NeXtMidas 2.9.3
   */
  public String getTheme () {
    return Shell.getProperty("ENV.THEME");
  }

  /** Gets style based on "ENV.STYLE" system property, otherwise returns null
      to use system defaults.
      @since NeXtMidas 2.9.3
   */
  public String getStyle () {
    return Shell.getProperty("ENV.STYLE");
  }

  /** Returns a terminal that NeXtMidas should use, null if the Shell should pick
      an appropriate terminal.
      @since NeXtMidas 3.3.0
   */
  @Override
  public Terminal getTerminal () {
    return terminal;
  }

  private final void debug (String msg) {
    System.out.println("DEBUG: "+msg);
  }

  // Set the window name to use for {@link #setPanel(MidasDisplay)}
  /** @deprecated Since 3.3.0: use {@link #runCommand(String, MidasDisplay)} */
  @Deprecated
  public void setWinID (String name) { // TODO: remove method in 3.5.0+
    return;
    // since 3.3.0 this method has no effect
//    if (name==null || name=="") {
//      this.shell.M.warning("Invalid Window ID : "+name+". Retaining name of: "+winName);
//      return;
//    }
//    this.winName = name;
  }

  /** Set the graphics window panel that graphical commands will use (to embed inside of).
      @deprecated Since 3.3.0: Use {@link #runCommand(String, MidasDisplay)} instead.
   */
  @Deprecated
  public void setPanel (nxm.sys.libg.MidasDisplay panel) {
    this.panel = panel;
  }

  /** @deprecated Since 3.3.0, use {@link #runCommand(String, MidasDisplay)} instead. */
  @Deprecated
  public void clearPanel () { // TODO: remove in 3.7.0+
    this.panel = null;
  }

  public javax.swing.JComponent getPanel () { return panel; }

  /** Get current shell driving this instance.
      @return shell driving this instance
      @since NeXtMidas 3.3.0
   */
  public Shell getShell () {
    return shell;
  }

  /** Get current Midas context for this instance.
      @return Midas context for current shell on this instance
      @since NeXtMidas 3.3.0
   */
  public Midas getMidasContext () {
    return shell.M;
  }

}
