package nxm.sys.lib;

import nxm.sys.inc.*;

/** A class to define a Midas Pipe. <br>
    <br>
    Terms:
    <pre>
      Inlet    - The writer to the pipe (e.g. the {@link DataFile} WAVEFORM
                 uses for output).
      Outlet   - One of the readers from the pipe (e.g. the {@link DataFile}
                 PLOT uses for input).

      dfx[0]   - The current input for the writer. (special)
      dfc[0]   - The current input for the writer. This is what is returned by
                 {@link #getDataFile()} or getDataFile(0).

      dfx[n]   - The current output for reader number n.
      dfc[n]   - The current input  for reader number n. Although normally
                 Input directly maps to Output, during a pipe restart the
                 dfc[n] may point to the previous incarnation of the Inlet
                 until such a time that the reader has read all of the "old"
                 data from the pipe. This is what is returned by getDataFile(n).
    </pre>
    During a restart, this class persists, the inlet does not transition to 0
    when the writer closes. When the writer opens again, the new DataFile class
    is chained to the previous so readers can follow the cascaded pipes.<br>
    <br>
    During a restart, this class persists, the outlets chain to their previous
    incarnations to support look-back of time values, etc. When a {@link TimeLine}
    is used this also has implications for the restart. Although most "normal"
    primitives (e.g. NOOP, FFT) create a new {@link DataFile} instance for their
    input pipe when their <tt>open()</tt> is called in a restart situation, PLOT
    and RMIF are special in that they reuse their current {@link DataFile} instance
    (with the {@link DataFile#ALWAYSREOPEN} flag set) and simply trigger a manual
    reopen of the file. As a result, for PLOT and RMIF the "previous incarnations"
    are actually copies of the instance used.
    <pre>
      TIME               PLOT/RMIF                            NOOP/FFT
       |                (as readers)                        (as readers)
       V       +------------+  +------------+       +------------+  +------------+
          dfx->| DataFile 0 |->| TimeLine 0 |  dfx->| DataFile 0 |->| TimeLine 0 |
               +------------+  +------------+       +------------+  +------------+

       ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PIPE RESTART #1 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

               +------------+  +------------+       +------------+  +------------+
          dfx->| DataFile 0 |->| TimeLine 0 |  dfx->| DataFile 1 |->| TimeLine 1 |
               | (Original) |  |            |       |            |  |            |
               +------------+  +------------+       +------------+  +------------+
                      |previous     ^                      |previous
                      v             |                      v
               +------------+       |               +------------+  +------------+
               | DataFile 0 |-------+               | DataFile 0 |->| TimeLine 0 |
               |  (Copy 1)  |                       |            |  |            |
               +------------+                       +------------+  +------------+

       ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PIPE RESTART #2 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

               +------------+  +------------+       +------------+  +------------+
          dfx->| DataFile 0 |->| TimeLine 0 |  dfx->| DataFile 2 |->| TimeLine 2 |
               | (Original) |  |            |       |            |  |            |
               +------------+  +------------+       +------------+  +------------+
                      |previous     ^  ^                   |previous
                      v             |  |                   v
               +------------+       |  |            +------------+  +------------+
               | DataFile 0 |-------+  |            | DataFile 1 |->| TimeLine 1 |
               |  (Copy 2)  |          |            |            |  |            |
               +------------+          |            +------------+  +------------+
                      |previous        |                   |previous
                      v                |                   v
               +------------+          |            +------------+  +------------+
               | DataFile 0 |----------+            | DataFile 0 |->| TimeLine 0 |
               |  (Copy 1)  |                       |            |  |            |
               +------------+                       +------------+  +------------+
    </pre>
    <b>WARNING:</b> The read/write/avail/etc. methods are NOT synchronized for performance
    reasons. If reading or writing from multiple threads, caller MUST synchronize on this
    object. The attach/detach methods are synchronized since they generally are NOT called
    many times. <br>
    <br>
    <b>WARNING:</b> This class permits extension only for the purposes of supporting
    X-Midas inter-operability (see {@link nxm.sys.libxm.XmPipe}) and should not be
    extended within user-defined option trees.

    @author Jeff Schoen
    @version $Id: Pipe.java,v 1.39 2012/12/03 20:10:29 ntn Exp $
 */
public class Pipe implements Chainable {
  /** Is specialized pipe debug enabled? */
  public static final boolean DEBUG = false;

  /** Outlet number indicating that the pipe has detached. */
  public static final int DISCON = -2;

  // NOTE: A number of the field definitions are a bit "silly" from a pure-Java
  //       standpoint (e.g. inlet as int not boolean) but were intentionally
  //       chosen in the early days of NeXtMidas to mirror the definitions used
  //       in X-Midas (and the BLUE header). One advantage of this approach is
  //       that it simplified some of the work required to map X-Midas pipes
  //       into NeXtMidas (e.g. XmPipe for NeXtOpt).

  public    int        inlet     = 0;       // is the inlet attached (1) or not (0)?
  public    int        outlets   = 0;       // number of currently attached readers
  public    int        outMask   = 0;       // bit mask of readers
  public    int        maxOut    = 8;       // maximum readers (outlets)
  public    int        size      = 131072;  // current pipe size in bytes
  public    double     inPtr     = 0;       // Inlet pointer in bytes
  public    double[]   outPtr;              // Outlet pointers in bytes (0=InletTotal, N=SingleOutlet)
  public    byte[]     buffer;              // The data buffer
  protected Midas      M;                   // Current Midas context (holds pipe references)
  /** @deprecated Since 3.1.1: this field is meant to be protected, use {@link #getName()} instead. */
  @Deprecated
  public    String     name;                // Pipe name
  protected Command[]  refs;                // The reference objects for the inlet/outlets (0=Inlet, N=Outlet)
  protected boolean[]  restart;             // Is a pipe restart needed?
  protected DataFile[] dfc;                 // The data files (source  e.g. WAVEFORM) for the inlet/outlets (0=Inlet, N=Outlet)
  protected DataFile[] dfx;                 // The data files (destination e.g. PLOT) for the inlet/outlets (0=Inlet, N=Outlet)
  protected double     timeLast  = 0;       // Time when rate was last checked (see getRate())
  protected double     inPtrLast = 0;       // Position when rate was last checked (see getRate())
  protected double     lastModified;        // last modified time (J1950) (defaults to time this was created) since 3.3.0

  /** Creates a new instance.
      @param M      Current Midas context (holds pipe references)
      @param name   Name of the pipe
   */
  public Pipe (Midas M, String name) {
    this(M,name,8); // 8 is historic, we should change this.
  }

  /** Creates a new instance.
      @param M      Current Midas context (holds pipe references)
      @param name   Name of the pipe
      @param maxOut Max outlets (i.e. pipe readers)
   */
  public Pipe (Midas M, String name, int maxOut) {
    this.M            = M;
    this.name         = name;
    this.maxOut       = maxOut;
    this.outPtr       = new double[maxOut+1];
    this.refs         = new Command[maxOut+1];
    this.restart      = new boolean[maxOut+1];
    this.dfc          = new DataFile[maxOut+1];
    this.dfx          = new DataFile[maxOut+1];
    this.lastModified = Time.current(); // since 3.3.0
  }

  @Override
  public String toString () {
    return "Raw Pipe Name="+name;
  }

  /** Attach to pipe.
      @param dir Direction (one of {@link IOResource#DIR_INPUT},
                 {@link IOResource#DIR_OUTPUT}, {@link IOResource#DIR_INOUT}, etc.)
      @param df  Associated data file.
      @return Outlet number (1..N)
   */
  public synchronized int attach (int dir, DataFile df) {
    int outlet = 0;
    // if (dir>=0) {                                        // prior to 2.3.0
    if (dir>0 || (dir==0 && df!=null && df.cmd==refs[0])) { // since 2.3.0 to support HEADERMOD on pipes (writer to pipe)
      inlet      = 1;
      df.setDataStart(inPtr);
      if (dfc[0]!=null) { dfc[0] = linkDataFiles(dfc[0], df); dfc[0].close(); }
      if (dfx[0]!=null) { dfx[0] = linkDataFiles(dfx[0], df); }
      dfc[0]     = df; // current inlet (writer to pipe)
      dfx[0]     = df;
      refs[0]    = df.cmd;
      restart[0] = false;
    }
    else {
      for (int i=1; i<=maxOut && outlet==0; i++) {
        // we have N-readers in maxOut-slots and we need to loop through to see if:
        //    1. this is a reconnect and we already have this primitive as a reader
        //       if so, we re-use that slot (RESTART) -- KNOWN DR: this logic
        //       can get confused if a primitive has 2+ readers for the same pipe
        //       (e.g. "MYMUX _out _in _in") and/or the pipe switches input
        //       slots (e.g. "MYMUX _out _in1 _in2" before restart and
        //       "MYMUX _out _in2 _in1" after restart).
        //    2. find the first available slot to attach to
        int mask = 1<<(i-1);
        if (restart[i] && (outMask&mask)!=0 && df.cmd==refs[i]) {  // CASE #1
          outlet      = i;
          dfx[outlet] = linkDataFiles(dfx[outlet], df);
        }
        else if (!restart[i] && (outMask&mask)==0) {               // CASE #2
          outlet         = i;
          outMask        = outMask | mask;
          outlets        = Math.max(outlets,outlet);
          outPtr[outlet] = inPtr;
          dfc[outlet]    = dfc[0];
        }
      }
      if (outlet>0) {
        refs[outlet]    = df.cmd;
        restart[outlet] = false;
        dfx[outlet]     = df;
      }
    }
    if (DEBUG) debug("Attach flags="+df.getFlagsString()+" dir="+IOResource.dirToString(dir),df,outlet);

    return outlet;
  }

  /** Links the old data file to the new one. Note that in the case of PLOT and
      RMIF this will be the same instance of DataFile (see diagram at top of
      class) requiring us to make a copy to use as the "old" file.
      @param old The old DataFile instance used (pre restart) or null for the
                 first connect to the pipe.
      @param df  The new DataFile instance to use (post restart).
      @return The "old" instance to store.
   */
  private DataFile linkDataFiles (DataFile old, DataFile df) {
    if (old != null) {
      if (old == df) { old = df.copy(); } // <-- Special case: DataFile is reused on restart
      old.setNextDataFile(df);
      df.setPreviousDataFile(old);

      // TODO: Look back in the history and null out previous files that have
      // no data that is within the last two pipe buffers. The current code
      // creates a "memory leak" by holding onto unnecessady DataFile references
      // in situations where there has been a number of pope restarts. This
      // will be tricky to implement since we need to consider "look back" both
      // in "normal" usage and for the case of a Layer2D raster requiring time
      // information that may be 2+ restarts ago -- the "last two pipe buffers"
      // mentioned above is merely a guess as to how far back to support, an
      // alternate option would be to hold a maximum of N old references.
    }
    return old;
  }

  /** Detach from a pipe.
      @param outlet The outlet number (1..N) or 0 for inlet.
      @param df     Associated data file.
      @return The new outlet number (usually {@link #DISCON}).
   */
  public synchronized int detach (int outlet, DataFile df) {
    return detach(outlet, df, ((df.flags&DataFile.REOPEN)!=0));
  }

  /** Detach from a pipe.
      @param outlet The outlet number (1..N) or 0 for inlet.
      @param df     Associated data file.
      @return The new outlet number (usually {@link #DISCON}).
   */
  private synchronized int detach (int outlet, DataFile df, boolean reopen) {
    if (outlet==DISCON) return outlet; // already detached
    DataFile next = dfc[outlet].getNextDataFile();
    restart[outlet] = reopen || (next!=null);
    if (DEBUG) debug("Detach flags="+df.getFlagsString()+" reopen="+reopen+" next="+next,df,outlet);

    if (outlet==0) {
      if (!restart[outlet]) inlet = 0;
      restart[outlet] = true;
    }
    else {
      int i=outlet-1;
      if (next != null) {
        dfc[outlet]     = next;
        outPtr[outlet]  = next.getDataStart();
        restart[outlet] = true;
      }
      else if (restart[outlet]) {
        // the input did not chain
      }
      else {
        outMask &= ~(1<<i);
      }
      for (i=outlets-1; i>=0 && (outMask&(1<<i))==0; i--) { }
      outlets = i+1;
      outlet = DISCON;
    }
    return outlet;
  }

  /** Update the combined read pointers and return the new value. Note that this recomputes the
      out pointers and detaches any readers that failed to reconnect to the pipe on restart. Due
      to this, most users should use {@link #getOutByte()} and should NOT call this method.
   */
  public double lastOut () {
    double out = inPtr;
    for (int i=0; i<outlets; i++) {
      if ((outMask&(1<<i)) != 0 && outPtr[i+1]<out) {
        // Disconnect readers that did not reconnect, see %TEST_RESTART_BLOCKING
        if (restart[i+1] && (refs[i+1].state != Command.RESTART)) {
          int flags = dfx[i+1].getFlags();
          if ((flags & BaseFile.ALWAYSREOPEN) == 0) { // Only check those that aren't marked as AlwaysReopen
            // Command restarted, but did not reconnect to this pipe after restart.
            M.info("PIPE: "+Convert.ref2String(refs[i+1])+" restarted, but did not reconnect to "+name);
            detach(i+1, dfx[i+1], false);
          }
        }
        out=outPtr[i+1];
      }
    }
    outPtr[0] = out;
    return out;
  }

  /** Read the pipe header.
      @param hb     Buffer to write header to.
      @param outlet Outlet number of outlet to copy the header from.
      @return Number of bytes copied (0 if outlet is disconnected).
   */
  public int readHeader (byte[] hb, int outlet) {
    if (outlet==DISCON) return 0;
    int bytes = Math.min(hb.length, dfc[outlet].hb.length);
    System.arraycopy(dfc[outlet].hb,0, hb,0, bytes);
    return bytes;
  }

  /** Write the pipe header.
      @param hb     Buffer to read header from.
      @param outlet Outlet number of outlet to copy the header to.
      @return Number of bytes copied.
   */
  public int writeHeader (byte[] hb, int outlet) {
    // its already there in dfc
    int bytes = Math.min(hb.length, dfc[outlet].hb.length);
    System.arraycopy(hb,0, dfc[outlet].hb,0, bytes);
    lastModified = Time.current(); // since 3.3.0
    return bytes;
  }

  /** Find bytes available for this outlet.
      @param outlet The outlet number (1..N) or 0 for inlet.
      @return Number of bytes available.
   */
  public double avail (int outlet) {
    double status;
    if (outlet==0) status = size - (inPtr-lastOut());
    else if (outlet<0 || outlet>maxOut) status=CoreIO.ERROR; // invalid outlet number
    else {
      double in = inPtr;
      DataFile df = dfc[outlet];
      // if cascaded then the rest of this tube is available
      if (df.getNextDataFile()!=null) in = df.getDataStop();
      status = in-outPtr[outlet];
      if (status>0) { } // OK
      else if (status<0) status=0;
      else if (inlet==0) status=CoreIO.CLOSED;
      else if (df.getNextDataFile()!=null) status=CoreIO.REOPEN;
    }
    if (status>size) {
      // This should be impossible, if seen it suggests there is an error in the inPtr and outPtr
      // accounting or that the write was allowed to continue even with a full pipe. It does NOT
      // mean that we wrote outside of the buffer (that would be an ArrayIndexOutOfBoundsException).
      M.warning("Pipe overflow detected for "+name+" closing pipe reader.");
      status = CoreIO.ERROR;
    }
    return status;
  }

  /** Check to see if the specified number of bytes are available for reading.
      @param outlet   The outlet number (1..N).
      @param numBytes The number of bytes to check for reading.
      @return true if the number of bytes specified are available for reading or if the pipe
              ends/restarts before the specified number of bytes will be available; false if
              not near the end of the pipe and the specified bytes are not available yet.
      @throws MidasException if the number of bytes specified exceeds the size of the pipe or if
                             the outlet number is invalid.
   */
  public boolean isReady (int outlet, int numBytes) {
    if (numBytes > size) {
      throw new MidasException("Transfer length of "+numBytes+" bytes exceeds pipe size of "+size+" bytes");
    }
    if ((outlet <= 0) || (outlet > maxOut)) {
      throw new MidasException("Invalid outlet number "+outlet);
    }
    // The code below is largely based on the avail(..) method
    double in = inPtr;
    DataFile df = dfc[outlet];
    if (df.getNextDataFile()!=null) in = df.getDataStop();
    double avail = in-outPtr[outlet];

    return (avail > outlet)              // ready
        || (inlet==0)                    // CLOSED
        || (df.getNextDataFile()!=null); // REOPEN
  }

  /** Write to pipe section.
      @param buf    The data buffer.
      @param lbuf   Native buffer location (0=Disable).
      @param boff   The buffer start offset.
      @param bytes  Number of bytes to write.
      @param outlet (not used, keeps signature same as read)
      @return Number of bytes written.
   */
  public int write (byte[] buf, long lbuf, int boff, int bytes, int outlet) {
    if (M.pipeMode==Midas.PPAUSE) return 0;
    if (M.pipeMode==Midas.PSTOP || M.pipeMode==Midas.POFF) return -1;
    double out = lastOut();
    int poff = (int)(inPtr % (double)size);
    int bmax = size-poff;
    if (bmax<bytes) bytes=bmax;
    int pmax = size - (int)(inPtr-out);
    if (pmax<bytes) bytes=pmax;
    if (bytes>0) {
      if (lbuf!=0) Native.p2ja (lbuf,boff,buffer,poff,bytes);
      else System.arraycopy (buf,boff,buffer,poff,bytes);
      inPtr += bytes;
      lastModified = Time.current(); // since 3.3.0
    }
    else {
      bytes=0;
    }
    return bytes;
  }

  /** Read from pipe section.
      @param buf    The data buffer.
      @param lbuf   Native buffer location (0=Disable).
      @param boff   The buffer start offset.
      @param bytes  Number of bytes to read.
      @param outlet The outlet number the reading is being done on behalf of (1..N).
      @return Number of bytes written.
   */
  public int read (byte[] buf, long lbuf, int boff, int bytes, int outlet) {
    if (M.pipeMode==Midas.PPAUSE) return 0;
    if (M.pipeMode==Midas.PSTOP || M.pipeMode==Midas.POFF) return CoreIO.CLOSED;
    DataFile df = dfc[outlet];
    double out = outPtr[outlet];
    int poff = (int)(out % (double)size);
    int bmax = size-poff;
    if (bmax<bytes) bytes=bmax;
    double in = inPtr;
    if (df.getNextDataFile()!=null) in = df.getDataStop();
    int pmax = (int)(in-out);
    if (pmax<bytes) bytes=pmax;
    if (bytes>0) {
      if (bytes>size) bytes=size;
      if (lbuf!=0) Native.ja2p (buffer,poff,lbuf,boff,bytes);
      else System.arraycopy (buffer,poff,buf,boff,bytes);
      outPtr[outlet] = out + bytes;
    }
    else if (df.getNextDataFile()!=null) {
      return CoreIO.REOPEN;
    }
    else if (inlet==0) {
      return CoreIO.CLOSED;
    }
    else {
      bytes=0;
    }
    return bytes;
  }

  /** Get the inlet connection flag (0 if none). */
  public short getInlet () { return (short)inlet; }

  /** Get the number of outlets. */
  public short getOutlets () { return (short)outlets; }

  /** Get the outlet mask mapping connected outlets. */
  public int getOutMask () { return outMask; }

  /** Indicates if a given outlet is in use.
      @param outlet The outlet number (1..N) or 0 for inlet.
      @since NeXtMidas 2.5.0
   */
  public boolean isAttached (int outlet) {
    if (outlet == 0) {
      return (dfc[0] != null);
    }
    else {
      int mask = 1<<(outlet-1);
      return ((outMask&mask) != 0);
    }
  }

  /** Gets the maximum number of outlets (readers) allowed for this pipe.
      @since NeXtMidas 2.5.0
   */
  public int getMaxOut () {
    return maxOut;
  }

  /** Get the size of the pipe in bytes. */
  public int getSize () { return size; }

  /** Get index of the next byte to be written. */
  public double getInByte () { return inPtr; }

  /** Get over-all outlet byte count. Same as <code>getOutByte(0)</code>. */
  public double getOutByte () { return outPtr[0]; }

  /** Get index of the next byte to be read by a given outlet.
      @param i Outlet number (1..N) or 0 for inlet (inlet uses limit for all outlets).
   */
  public double getOutByte (int i) { return outPtr[i]; }

  /** Get the number of bytes available to be written. */
  public double getAvail () { return (inPtr-outPtr[0]); }

  /** Get the approximate data rate for the pipe (bytes/sec). */
  public double getRate () {
    double time = Time.current();
    double xfer = (inPtr-inPtrLast);
    double rate = xfer/(time-timeLast);
    timeLast = time;
    inPtrLast = inPtr;
    return rate;
  }

  /** Gets the data file for the inlet. Same as <code>getDataFile(0)</code>. */
  public DataFile getDataFile () {
    return dfc[0];
  }

  /** Gets the data file for a given inlet/outlet.
      @param outlet Outlet number (1..N) or 0 for inlet.
   */
  public DataFile getDataFile (int outlet) {
    return (outlet!=DISCON) ? dfc[outlet] : null;
  }

  /** Gets the owner for a given inlet/outlet.
      @param outlet Outlet number (1..N) or 0 for inlet.
   */
  public Object getOwner (int outlet) {
    return (outlet!=DISCON) ? refs[outlet] : null;
  }

  /** Sets the size of the pipe. This allows dynamic setting of the pipe size
      for tuning. These are the steps done by this method:
      <pre>
        (1) Set the shared pipe mode to {@link Midas#PPAUSE}.
        (2) Discard the current pipe buffer (discards any unread data).
        (3) Create a new pipe buffer with the new size.
        (4) Restore the shared pipe mode.
      </pre>
      @param value The new size of the pipe (in bytes).
   */
  public void setSize (int value) {
    if (buffer==null || value!=size) {
      int mode = M.pipeMode;
      if (mode==Midas.PRUN) { M.pipeMode=Midas.PPAUSE; Time.sleep(0.2); }
      buffer = new byte[value];
      size = value;
      if (M.pipeMode!=mode) M.pipeMode=mode;
    }
  }

  /** Set index of the next byte to be written. */
  public void setInByte ( double value) { inPtr=value; }

  /** Set over-all outlet byte count. Same as <code>setOutByte(0,value)</code>. */
  public void setOutByte (double value) { outPtr[0]=value; }

  /** Set index of the next byte to be read by a given outlet.
      @param i     Outlet number (1..N) or 0 for inlet.
      @param value The new index.
   */
  public void setOutByte (int i, double value) { outPtr[i]=value; }

  /** The pipe is always chained to the data file used by the inlet.
      Same as <code>getDataFile()</code>
   */
  @Override
  public Object getNextLink () { return dfc[0]; }

  /** There is no previous link for a pipe, always returns null. */
  @Override
  public Object getPrevLink () { return null; }

  /** method used for signaled (vs polled) operation */
  public synchronized void awaken() { notifyAll(); }

  /** @return this pipe's name
      @since NeXtMidas 3.1.1
   */
  public String getName () {
    return name;
  }

  /** @return Last modified time in seconds since J1950.
      @since NeXtMidas 3.3.0
   */
  public double getLastModified () {
    return lastModified;
  }

  /** Prints a debug message. */
  private void debug (String msg) {
    String message = "Pipe "+name+": "+msg;
    if (M != null) {
      M.info(message);
    } else {
      Shell.info(message);
    }
  }

  /** Prints a debug message along with df and outlet. */
  private void debug (String msg, DataFile df, int outlet) {
    String id = (df == null)? null : Convert.ref2String(df.cmd);

    debug(msg+" owner="+id+" outlet="+outlet);
  }
}
