package nxm.sys.lib;

import nxm.sys.inc.DataTypes;
import java.util.Hashtable;

/** A class to handle accessing the Results table. Interprets the
    <code>L:KEYNAME</code> syntax and reformats as necessary. Implements the
    <code>CLASS.METHOD(ARG1,ARG2).MEMBER.<i>etc</i></code> syntax parsing.

    @author Jeff Schoen
    @version $Id: Results.java,v 1.55 2012/10/04 13:01:38 jgs Exp $
*/
public class Results extends Table {

  /** NOTE: this should be changed if the structure/fields of this class changes. */
  private static final long serialVersionUID = 2011040100292L; // for Serializable interface

  private Object ref;                     // The default reference object to use
  private Hashtable<String,Function> fht; // HashTable for macro functions

  /** Initialize the results set. */
  public Results () {
//    setMode(HT);
    this.ref = this;
  }

  /** Initialize the results set.
      @param ref The reference object to use.
   */
  public Results (Object ref) {
    this();
    this.ref = ref;
  }

  public void put (String key, Object value)                       { putResult(key,value,   0,ref); }
  public void put (String key, Object value, int attr)             { putResult(key,value,attr,ref); }
  public void put (String key, Object value, Object ref)           { putResult(key,value,   0,ref); }
  public void put (String key, Object value, int attr, Object ref) { putResult(key,value,attr,ref); }

  public Object setKey (String key, Object value) {
    // Need to override since putting a result can change it's value in the case where the key
    // includes an implicit type cast, e.g. setKey("L:FOO","3.14")
    return putResult(key,value,0,ref);
  }

  /** Special put method required since putting a result can change it's value in the case
      where the key includes an implicit type cast, e.g. setKey("L:FOO","3.14").
   */
  private Object putResult (String key, Object value, int attr, Object ref) {
    if (key==null || key.length()==0) return null;
    if (key.length()>2 && key.charAt(1)==':') {
      char type = key.charAt(0);
      key = key.substring(2);
      if (type == 'A') {
        Shell.deprecate("Since NeXtMidas 2.5.2: Change use of A:"+key+" to S:"+key+". "
                      + "The behavior of of A: will change in future releases.");
        type = 'S';
      }
      value = Convert.o2o (value,type,ref);
    }
    Function f;
    int i=key.indexOf('.');
    int j=key.indexOf(')');
    if (i<0&&j<0) {						// simple case
      if (fht!=null && (f=fht.get(key))!=null) f.set(value);
      else putx(key,value,attr);
    }
    else if (j>0 && key.startsWith("FILE(")) {			// in-line file
      j = closingParen(key,4);
      DataFile df = inLineFile(ref,key.substring(5,j),true);
      KeyObject.setKey (df,key.substring(j+2),value,ref);
      df.close();
    }
    else if (j<0) {						// quick 1st level
      Object root = getx(key.substring(0,i));
      String subkey = key.substring(i+1);
      if (root==null) {
        if ((attr&ADDROOT)==0) Convert.ref2Midas(ref).error("Unable to create inner table "+key);
        root=new Table();
        putx(key.substring(0,i),root,0);
      }
      if (root instanceof Table) {
        ((Table)root).put(subkey,value,attr,ref);
      } else {
        Object obj = KeyObject.setKey (root,subkey,value,ref);
        if (obj==null) Convert.ref2Midas(ref).warning("Unable to set "+key+" to value "+value);
      }
    }
    else {							// full on setKey
      KeyObject.setKey (this,key,value,ref);
    }
    return value;
  }

  public Object get (String key) {
    char type=' '; Object value=null;
    if (key==null || key.length()==0) return null;
    if (key.length()>2 && key.charAt(1)==':') {
      type = key.charAt(0);
      key = key.substring(2);
    }
    Function f;
    int i=key.indexOf('.');
    int j=key.indexOf(')');
    if (i<0&&j<0) {						// simple case
      if (fht!=null && (f=fht.get(key))!=null) value = f.get();
      else value = getx(key);
    }
    else if (j<0) {						// quick 1st level
      Object root = getx(key.substring(0,i));
      value = KeyObject.getKey (root,key.substring(i+1),null,ref);
    }
    else if (key.startsWith("FILE(")) {				// in-line file
      j = closingParen(key,4);
      DataFile df = inLineFile(ref,key.substring(5,j),false);
      if (j==key.length()-1) value = df;
      else value = KeyObject.getKey (df,key.substring(j+2),null,ref);
      df.close();
    }
    else if (key.startsWith("CALC(")) {				// in-line calc
      value = Convert.d2o(Convert.s2d(key,ref));
    }
    else if (key.startsWith("TEST(")) {				// in-line test
      value = Convert.d2o(Convert.s2l(key,ref));
    }
    else if (key.startsWith("SEDIT(")) {			// in-line SEDIT
      Midas M = nxm.sys.lib.Convert.ref2Midas(ref);
      Parser p = new Parser(key.substring(6,key.length()-1), true);
      Args args = new Args(M,p);
      if (ref instanceof Args) ((Args)ref).copySwitchesTo(args);
      value = nxm.sys.intr.Sedit.processCommandLine(args.getCS(1),M,args,2,false);
    }
    else {							// full on getKey
      value = KeyObject.getKey (this,key,null,ref);
    }
    if (type!=' ') value = Convert.o2o(value,type,ref);
    return value;
  }

  private int closingParen (String str, int j) {
    int ip=0;
    for (;j<str.length();j++) {
      char c=str.charAt(j);
           if (c=='(') ip++;
      else if (c==')') ip--;
      if (ip==0) break;
    }
    return j;
  }

  /** Gets the length to end of the valid results name.
      @param buffer The String buffer containing the results name.
      @param index  The index into the buffer to begin reading from.
      @return The valid name length.
      @see #validNameLength(byte[],int,int)
   */
  public static int validNameLength (String buffer, int index) {
    return validNameLength(buffer,null,index,buffer.length());
  }
  /** Gets the length to end of the valid results name.
      @param buffer The byte buffer containing the results name.
      @param index  The index into the buffer to begin reading from.
      @param size   The maximum size to check within (in the buffer).
      @return The valid name length.
      @see #validNameLength(String,int)
   */
  public static int validNameLength (byte[] buffer, int index, int size) {
    return validNameLength(null,buffer,index,size);
  }

  public static int validNameLength (String sb, byte[] bb, int index, int size) {
    int i=index, ip=0,ips=index,iq=0,ic=0,hd=0; char c,cl=' ';
    for (; i<size; i++,cl=c) {
      c = (sb!=null)? sb.charAt(i) : (char)bb[i];
           if (c>='A' && c<='Z');
      else if (c>='a' && c<='z') c-=32; // for later in-line type casting test
      else if (i==index) return 0;
      else if (c>='0' && c<='9');
      else if (c=='_');
      else if (c=='"' && ip>0) iq=1-iq;
      else if (iq>0);
      else if (cl=='.' && ip<=0) return i-1-index; // next character after dot MUST be alphanumeric 
      else if (c=='.') { if (ip==0) { hd++; ic=0; }} // has dot, clear colon exception, if after
      else if (c=='(') { if (ip==0 && ic==0) ips=i; ip++; }
      else if (c==')' && ip>0) ip--; 
      else if (c==':' && i-ips==1 && DataTypes.CONVERT_TYPE_LIST.indexOf(cl)>=0); // allow colon for in-line type casting
      else if (c==':' && ip==1) ic=i; // disallow colon if for file trimming
      else if (ip>0);
      else break;
    }
//if (sb==null) sb = new String(bb,0,size);
//System.out.println("OK "+sb+" "+i+" "+ips+" "+ic);
    if (cl=='.') i--;
    if (ic>0 && ips>0) i=ips;
    return i-index;
  }

  public static int validNameLength_old (String sb, byte[] bb, int index, int size) {
    int ip=0; char c,cl=' ';
    for (int i=index; i<size; i++,cl=c) {
      c = (sb!=null)? sb.charAt(i) : (char)bb[i];
      if (c>='A' && c<='Z');
      else if (c>='a' && c<='z');
      else if (i==index) return 0;
      else if (c>='0' && c<='9');
      else if (c=='_');
//jph      else if (cl=='.') return i-1-index; // next character after dot MUST be alphanumeric
      else if (cl=='.' && ip<=0) {
        return i-1-index; // next character after dot MUST be alphanumeric
      }
      else if (c=='.');
      else if (c=='(') ip++;
      else if (c==')' && ip>0) ip--;
      else if (c==':' && i-index==1); // in-line type casting
      else if (ip>0);
      else if (cl=='.') return i-1-index; // trailing dot
      else return i-index;
    }
    return size-index;
  }

  /** Process an in-line file access read/write.
      @param ref   The reference object to use.
      @param name  The name of the data file.
      @param write Indicates if the file is intended to be written to.

      @return The data file to use for read/write. If <code>write == true</code>
              then the file will be opened in <code>APPEND | INOUT</code> mode;
              if <code>write == false</code> then the file will be opened in
              <code>INPUT</code>.
      @see DataFile#APPEND
      @see DataFile#INOUT
      @see DataFile#INPUT
      @see DataFile#open(int)
   */
  private DataFile inLineFile (Object ref, String name, boolean write) {
    DataFile df = new DataFile();
    df.init(ref,name,"","",0);
    df.open( write? DataFile.APPEND|DataFile.INOUT : DataFile.INPUT);
    return df;
  }

  public interface Function {
    public void   set (Object value);
    public Object get ();
    public Object exec (Object args);
  }

  public void addFunction (String name, Function f) {
    if (fht==null) fht = new Hashtable<String,Function>();
    fht.put(name,f);
  }

  public Hashtable<String,Function> getFunctions () { return fht; }

}
