package nxm.sys.lib;

import nxm.sys.inc.*;

import java.io.*;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.lang.reflect.Array;

/** A class to perform Table put/get and maintenance functions. This class is used for result tables,
    parameters, switches, and other primitive specific tables. This class pre-dates most of the
    classes in the <tt>java.util</tt> package, and for years, was compatible with Java 1.1. Since
    NeXtMidas 2.2.0, this class has implemented the {@link Map} interface, but with a few caveats
    necessary for backwards-compatibility.<br>
    <br>
    This class will implements the {@link Map} interface, but with the following caveats:
    <pre>
      (1) All keys are converted to strings on insertion.
      (2) Any null values are converted to "NULL" on insertion (exception
          {@link #EXCEPTION_ON_NULL} throws exception rather than allowing
          null.
      (3) Keys with a dot represent a put of a value into a sub-table (or
          sub-object using the methods in {@link KeyObject}). For example
          put("A.B","FOO") gets the table "A" from this table and sets "B"
          to "FOO" in that table.
    </pre>
    This class also extends the beyond the normal {@link Hashtable} class by adding typed methods
    for the {@link Data} and {@link File} classes, a {@link #get(String)} method that allows keys
    to be abbreviated, and follows links to other tables to resolve references in a hierarchy of
    tables. <br>
    <br>
    The serialized form of a table has the format:
    <pre>
      {key1=value1,...,keyN=valueN}
    </pre>
    and can include nested tables as follows:
    <pre>
      {key1=value1,...,keyT={keyT1=valueT1,...,keyTN=valueTN},...,keyN=valueN}
    </pre>
    The current implementation allows initialization from the serialized form directly into a
    byte buffer to provide for fast serialization if no alterations are made. This also preserves
    key ordering. <i>This disables validity checks for the table until a later time.</i>

    <h2><a name="TableModes">Table Modes</a></h2>
    The mode of the table identifies the underlying structure used to hold the keys and values.
    Prior to NeXtMidas 2.7.0, only three modes ({@link #FB}, {@link #KV},and {@link #HT}) were
    supported. Since NeXtMidas 2.7.0, a wide range of table modes are supported. The various
    modes are summarized in the following table.
    <table border="1">
      <tr><th>Mode                                     </th><th>Underlying Class  </th><th>Synchronized</th><th>Iteration Order</th><th>Performance<sup>1</sup></th></tr>
      <tr><td>FromBytes                 ({@link #FB})  </td><td><tt>byte[]</tt>            </td><td>n/a</td><td>n/a            </td><td>n/a             </td></tr>
      <tr><td>KeyVector                 ({@link #KV})  </td><td>{@link KeyVector}          </td><td>Yes</td><td>Insertion order</td><td>O(N)<sup>2</sup></td></tr>
      <tr><td>Hashtable                 ({@link #HT})  </td><td>{@link Hashtable}          </td><td>Yes</td><td>Unordered      </td><td>O(1)            </td></tr>
      <tr><td>HashMap<sup>3</sup>       ({@link #HM})  </td><td>{@link HashMap}            </td><td>No </td><td>Unordered      </td><td>O(1)            </td></tr>
      <tr><td>LinkedHashMap             ({@link #LHM}) </td><td>{@link LinkedHashMap}      </td><td>No </td><td>Insertion order</td><td>O(1)            </td></tr>
      <tr><td>SynchronizedLinkedHashMap ({@link #SLHM})</td><td>{@link LinkedHashMap}      </td><td>Yes</td><td>Insertion order</td><td>O(1)            </td></tr>
      <tr><td>TreeMap                   ({@link #TM})  </td><td>{@link TreeMap}            </td><td>No </td><td>Alphabetical   </td><td>O(LG(N))        </td></tr>
      <tr><td>SynchronizedTreeMap       ({@link #STM}) </td><td>{@link TreeMap}            </td><td>Yes</td><td>Alphabetical   </td><td>O(LG(N))        </td></tr>
      <tr><td>Map                       ({@link #MAP}) </td><td>{@link Map}<sup>4</sup>    </td><td>n/a</td><td>n/a            </td><td>n/a             </td></tr>
      <tr><td>Keyable                   ({@link #KEY}) </td><td>{@link Keyable}<sup>4</sup></td><td>n/a</td><td>n/a            </td><td>n/a             </td></tr>
      <tr><td>ConcurrentHashMap         ({@link #CHM}) </td><td>{@link ConcurrentHashMap}  </td><td>No<sup>5</sup></td><td>Unordered</td><td>O(1)       </td></tr>
      <tr><td colspan="5">
        <sup>1</sup> Average time cost for {@link #get(String)} and {@link #put(String,Object)}.<br>
        <sup>2</sup> The {@link #ADD} flag gives O(1) performance when inserting entries in
                     {@link #KV} mode since it disables checks for duplicate entries, performance
                     of other methods is unaffected by the {@link #ADD} flag.<br>
        <sup>3</sup> Hashtable is the synchronized version of HashMap.<br>
        <sup>4</sup> {@link Map} and {@link Keyable} are interfaces, see {@link #MAP} and {@link #KEY}
                     for details.<br>
        <sup>5</sup> {@link ConcurrentHashMap} is a hash table supporting full concurrency of
                     retrievals and adjustable expected concurrency for updates.
                     See java.util.concurrent.ConcurrentHashMap API for details. (Since NeXtMidas 3.1.1).
      </td></tr>
    </table>
    <br>
    <dl>
    <dt>NOTE1:</dt>
      <dd>The public methods in this class are NOT invokeable (by the INVOKE command) within the macro language.</dd>
    <br>
    <dt><a name="dotsyntax">NOTE2:</a></dt>
      <dd>Most of the getXXX methods (e.g. {@link #getTable(String, boolean)}, {@link #get(String)})
          and putXXX methods (e.g. {@link #put(String, Object)}, {@link #put(String, double)})
          takes in a <b>"key"</b> parameter that is a String which can use the <b>dot syntax</b>
          to drill down/chain into the sub-Tables and/or Object properties/methods (if they exists). <br>
          BUT from this flexibility, those methods can throw a {@link MidasException} when drilling down
          into a Object's property/method that does NOT exist. The MidasException message might
          look something like &quot;<tt>ERROR: KeyObject: Can not access '<value after dot>' in
          <class name> when given no arguments: No methods found.</tt>&quot;.
          So care must be taken account for this or to avoid this scenario.<br>
          <br>
          Since NeXtMidas 2.9.2, the {@link #ONLY_DOT_INTO_SUBTABLES} flag can be used to tell the
          {@link #get(String)}, {@link #get(String, Object, int)}, and hence most of the getXXX methods
          to only to drill down into sub-Tables for the specified dotted key (i.e. do NOT drill down
          to the value object's properties and methods).
      </dd>
     </dl>
    <br>
    Since 3.3.1: Table implements {@link java.lang.Iterable} interface so it can be use to
    iterate over the keys in the table using Java 5's foreach loop.

    @author  Jeff Schoen (initial version)
    @author  J. Czechowski IV (changes to implement {@link Map} and add additional table modes)
    @author  Neon Ngo (additional maintenance and fixes)
    @version $Id: Table.java,v 1.157 2013/01/07 15:49:46 ntn Exp $
 */
public class Table implements KeyableByType, AsciiMap, Externalizable, Map, Tablizable, Iterable<String> {
  // Implementation notes:
  //   (1) An AssertionError is thrown in cases where a genericMode is encountered that should be
  //       impossible. If an AssertionError is found, there is something wrong internal to table.
  //   (2) "Hashtable" was chosen over "HashTable" since it matches java.util.Hashtable. The
  //       abbreviation "HT" is maintained for backwards-compatibility (and because we like it).

  /** NOTE: this should be changed if the structure/fields of this class changes. */
  private static final long serialVersionUID = 2010021000274L; // for Serializable interface (see Bug 2075)

  /** Table Mode: Data is in a byte array ("From Bytes") (see <a href="#TableModes">Table Modes</a>).
      When a table in {@link #FB} mode is accessed (put, get, etc.) it is automatically converted
      to one of the "working" modes as specified by the default table mode.
   */
  public static final int FB=1;
  /** Table Mode: Data is in a {@link KeyVector} (see <a href="#TableModes">Table Modes</a>). */
  public static final int KV=2;
  /** Table Mode: Data is in a {@link Hashtable} (see <a href="#TableModes">Table Modes</a>). */
  public static final int HT=3;
  /** Table Mode: Data is in a {@link HashMap} (see <a href="#TableModes">Table Modes</a>).
      @since NeXtMidas 2.7.0
   */
  public static final int HM=4;
  /** Table Mode: Data is in a {@link LinkedHashMap} (see <a href="#TableModes">Table Modes</a>).
      @since NeXtMidas 2.7.0
   */
  public static final int LHM=5;
  /** Table Mode: Data is in a synchronized {@link LinkedHashMap} (see <a href="#TableModes">Table Modes</a>).
      @since NeXtMidas 2.7.0
   */
  public static final int SLHM=6;
  /** Table Mode: Data is in a {@link TreeMap} (see <a href="#TableModes">Table Modes</a>).
      @since NeXtMidas 2.7.0
   */
  public static final int TM=7;
  /** Table Mode: Data is in a synchronized {@link TreeMap} (see <a href="#TableModes">Table Modes</a>).
      @since NeXtMidas 2.7.0
   */
  public static final int STM=8;
  /** Table Mode: Data is in a {@link Map} (see <a href="#TableModes">Table Modes</a>).
      Since {@link Map} is an interface, when the mode is set to {@link #MAP}, {@link Table} will
      use a concrete implementation of {@link Map} in place of it. The implementation used is
      undefined, but {@link LinkedHashMap} is common.
      @since NeXtMidas 2.7.0
   */
  public static final int MAP=9;
  /** Table Mode: Data is in a {@link Keyable} (see <a href="#TableModes">Table Modes</a>).
      Since {@link Keyable} is an interface, when the mode is set to {@link #KEY}, {@link Table} will
      use a concrete implementation of {@link Keyable} in place of it. The implementation used is
      undefined, but {@link KeyVector} is common.
      @since NeXtMidas 2.7.0
   */
  public static final int KEY=10;
  /** Table Mode: Data is in a {@link ConcurrentHashMap} (see <a href="#TableModes">Table Modes</a>).
      @since NeXtMidas 3.1.1
   */
  public static final int CHM=11;


  private static final int FIRST_MODE = FB;    // Numerically the first valid table mode
  private static final int LAST_MODE  = CHM;   // Numerically the last  valid table mode

  /** The set of table modes in their abbreviated form. <i>This is maintained for
      backwards-compatibility with NeXtMidas 2.6.0 and earlier, most users will want to use
      {@link #modeList}.</i>
      <pre>tableModesList = {@value}</pre>
   */
  public static final String tableModesList = "FB,KV,HT,HM,LHM,SLHM,TM,STM,MAP,KEY,CHM";

  /** The set of table modes (see <a href="#TableModes">Table Modes</a>).
      <pre>modeList = {@value}</pre>
   */
  public static final String modeList = "FromBytes,KeyVector,Hashtable,HashMap,"
                                      + "LinkedHashMap,SynchronizedLinkedHashMap,"
                                      + "TreeMap,SynchronizedTreeMap,Map,Keyable,"
                                      + "ConcurrentHashMap";

  /** Table flag: Automatically read from previous table (see {@link #getPrevious()}). */
  public static final int AUTOREADPREV=0x1;

  /** Table flag: Automatically write to previous table (see {@link #getPrevious()}). */
  public static final int AUTOWRITEPREV=0x2;

  /** Table flag: All calls to <tt>put()s</tt> or <tt>putx()s</tt> may assume that the key being
      added is unique. This shifts the burden of uniqueness checks to the caller in
      cases where required for maximum performance. The behavior of the table is
      undefined if an non-unique key is inserted when the {@link #ADD} flag is set.<br>
      <br>
      Since NeXtMidas 2.7.1, if the ADD flag is specified, it will change this Table to
      KeyVector mode to allow duplicate keys in the Table. This behavior ensures
      backwards-compatibility with legacy applications expecting the ADD flag to
      allow duplicate keys.<br>
      <br>
      Typical usage:
      <pre>
        Table tbl = new Table(Table.KV, Table.ADD);
        for (..) {
          tbl.put(uniqueKey, val); // add unique key to table
        }
        tbl.setFlags(0);
      </pre>
   */
  public static final int ADD=0x4;

  /** Table flag: Table is read only. Any changes must be accompanied by the {@link #FORCE} flag. */
  public static final int READONLY=0x8;

  /** Table flag: Overrides the {@link #READONLY} flag and forces a change to a read-only table
                  (see {@link #put(String,Object,int)}. See <tt>/FORCE</tt> in <tt>RESULTS</tt>.
   */
  public static final int FORCE=0x10;

  /** Table flag: Add table roots as needed. See <tt>/ADDROOT</tt> in <tt>RESULTS</tt>. */
  public static final int ADDROOT=0x20;

  /** Table flag: Do not convert keys to upper case.
      @since NeXtMidas 2.5.0
   */
  public static final int CASESENSITIVE=0x40;

  /** Table flag: Do not insert <tt>"NULL"</tt> in place of <tt>null</tt>, rather throw a
                  {@link NullPointerException} if <tt>null</tt> is inserted.
      @since NeXtMidas 2.7.1
   */
  public static final int EXCEPTION_ON_NULL=0x80;

  /** Table flag: When creating Tables from Strings or writing Tables as Strings, parse the String to handle escape characters
      @since NeXtMidas 2.9.0  (Bug 2292)
   */
  public static final int DISABLE_ESCAPE_SEQUENCES=0x100;

  /** Table flag: When using the <a href="#dotsyntax">dot syntax</a> in the key parameter for the
      {@link #get(String)} and hence most of the <b>getXXX</b> methods, only drill down to sub-Tables
      (i.e. do NOT drill down to object properties and methods).
      @since NeXtMidas 2.9.2 to address BUG 2333 and BUG 2334
  */
  public static final int ONLY_DOT_INTO_SUBTABLES=0x200;

  /** When creating a Table from a String, perform checks on the String to prevent some cases of malformed Tables
      @since NeXtMidas 3.3.0
   */
  public static final int CHECK_MALFORMED = 0x400;  // added for Bug 85 in 3.3.0

  /** List of available table flags.
      <pre>flagsList = {@value}</pre>
      @since NeXtMidas 2.7.0
   */
  public static final String flagsList = "AutoReadPrev,AutoWritePrev,Add,ReadOnly,"     // 0x01 - 0x08
                                       + "Force,AddRoot,CaseSensitive,ExceptionOnNull," // 0x10 - 0x80
                                       + "DisableEscapeSequences,OnlyDotIntoSubTables," // 0x100 - 0x200
                                       + "CheckMalformed";                              // 0x400

  private static int defaultMode  = SLHM;  // Default table mode (can not be set to FB)
  private static int defaultFlags = AUTOREADPREV; // Default table flags

  private int       genericMode  = FB;     // Generic type of table (FB,KV,MAP,KEY)
  private int       specificMode = FB;     // Specific type of table (any of the possible modes)
  private byte[]    fb;                    // byte buffer holding data in FB mode (null=Empty)
  private KeyVector kv;                    // KeyVector holding data in KV mode
  private Keyable   kbl;                   // Keyable holding data in KEY mode (not called key since
                                           //   it is a very common variable name)
  private Map<String,Object> map;          // Hashtable holding data in HT mode
  private Listener  listener;              // Table listener
  private Table     prev;                  // Link to the previous table
  private int       flags;                 // Table flags
  private Enum      en;                    // Local enumeration used

  /** Initialize an empty Table with the default mode and default flags. */
  public Table () {
    this(defaultMode, defaultFlags);
  }

  /** Initialize an empty Table with the given mode. <br><br>
      @param mode The table mode to use (see {@link #tableModesList}).
   */
  public Table (int mode) {
    this(mode,defaultFlags);
  }

  /** Initialize an empty Table.
      @param mode  The table mode to use (see {@link #tableModesList}).
      @param flags The table flags.
      @since NeXtMidas 2.5.0
   */
  public Table (int mode, int flags) {
    // Note that genericMode and specificMode default to FB. If this is not done, setMode(..) will
    // encounter a NullPointerException or similar error.
    setMode(mode);
    setFlags(flags);
  }

  /** Initialize table from a String in serialized form.  If the string is
      null then the constructor ignores the string and initializes itself to an empty table.
      @param string The data for the table (in the same format as used by {@link #toString()}).
  */
  public Table (String string) {
    this(FB);
    if (string!=null) fromBytes(string.getBytes());
  }

  /** Initialize table from a byte array in serialized form.  If the byte
      array is null or zero length then the constructor ignores the argument
      and initializes itself to an empty table.
      @param bytes The bytes that make up the table.
  */
  public Table (byte[] bytes) {
    this(FB);
    if (bytes!=null && bytes.length>0) fromBytes(bytes);
  }

  /** Initialize table from a byte array (specified range) in serialized form.
      If the byte array is null or zero length then the constructor ignores the argument
      and initializes itself to an empty table. If the start index and/or length is
      invalid (i.e. out of bounds) this is also initialize itself to an empty table.
      @param bytes The bytes that make up the table.
      @param start start index in byte array
      @param len   the desired length from start index in byte array
      @since NeXtMidas 2.7.1
  */
  public Table (byte[] bytes, int start, int len) {
    this(FB);
    if (bytes!=null && bytes.length>0 && start>=0 && len<=bytes.length-start) fromBytes(bytes, start, start+len);
  }

  /** Initialize table from a TextFile using default flags.
      Note: this forces this Table into KeyVector mode to allow duplicates to be
            loaded (legacy behavior).
      @param tf The text file defining the table.
      @see #fromTextFile(TextFile)
   */
  public Table (TextFile tf) {
    // fromTextFile(tf, defaultFlags); <-- NOT doing this b/c it would be a change in behavior
    this(FB);
    fromTextFile(tf); // <-- this will force Table to KV mode
  }

  /** Initialize table from a TextFile using specified flags mask and default table mode.
      @param tf The text file defining the table.
      @param flags Table flags mask to use for this
      @since NeXtMidas 3.1.2
   */
  public Table (TextFile tf, int flags) {
    this(FB, flags);
    fromTextFile(tf, flags);
  }

  /** Initialize a table from the contents of a Hashtable.  Upon completion
      this Table will contain references to the same keys and values as the
      given hashtable but will not have a reference to the hash itself.
      @param ht The hashtable file defining the table.
  */
  public Table (Hashtable ht) {
    this(FB);
    fromTable(ht);
  }

  /** Initialize a table with a given String. Unlike most other Table constructors,
      this constructor can create and load the Table right away so as to allow for checking
      the input String for formatting
      @param string  The String to convert to a Table
      @param load  Whether or not to parse the input String immediately and load the Table
      @since NeXtMidas 3.3.0
   */
  public Table (String string, boolean load) {
    this(FB);
    if (string==null) return;
    if (!load) fromBytes(string.getBytes());
    else       fromBytes(string.getBytes(),0, string.length());
  }

  /** Converts a mode string to a number. This accepts either the mode abbreviation or full name. */
  private static int modeStrToNum (String mode) {
    // Check tableModesList first so "KEY" matches "Keyable" not "KeyVector" and "HT" matches
    // "Hashtable". Checking tableModesList first is backwards-compatible with NeXtMidas 2.6.0.
    int m = Parser.find(tableModesList, mode, -1);
    if (m < 0) m = Parser.find(modeList, mode, -1);
    if (m < 0) throw new MidasException("Table: Invalid mode '"+mode+"' expected one of '"+modeList+"'.");
    return m;
  }

  /** Get default mode for tables.
      @return The default table mode to use (see <a href="#TableModes">Table Modes</a>).
      @since NeXtMidas 2.7.0
   */
  public static int getDefaultModeInt () {
    return defaultMode;
  }

  /** Get default mode for tables.
      @return One of the names listed in {@link #tableModesList}.
      @since NeXtMidas 2.7.0
   */
  public static String getDefaultModeName () {
    return Parser.get(tableModesList, getDefaultModeInt());
  }

  /** Get default mode for tables.
      @return One of the names listed in {@link #modeList}.
      @since NeXtMidas 2.7.0
   */
  public static String getDefaultMode () {
    return Parser.get(modeList, getDefaultModeInt());
  }

  /** Set default mode for tables. Note that in the case of {@link #MAP} and {@link #KEY}
      a specific implementation of those interfaces will be used. Note that {@link #FB} can not
      be used as a default Table mode.
      @param newMode The default table mode to use (see <a href="#TableModes">Table Modes</a>).
      @throws MidasException if passed an invalid mode or if passed "FB".
   */
  public static void setDefaultMode (String newMode) {
    setDefaultMode(modeStrToNum(newMode));
  }

  /** Set default mode for tables. Note that in the case of {@link #MAP} and {@link #KEY}
      a specific implementation of those interfaces will be used. Note that {@link #FB} can not
      be used as a default Table mode.
      @param newMode The default table mode to use (see <a href="#TableModes">Table Modes</a>).
      @throws MidasException if passed an invalid mode or if passed {@link #FB}.
   */
  public static void setDefaultMode (int newMode) {
    if (newMode == FB) {
      throw new MidasException("Table: Default table mode can not be set to FB.");
    }
    if ((newMode < FIRST_MODE) || (newMode > LAST_MODE)) {
      throw new MidasException("Table: Unknown table mode "+newMode+".");
    }
    defaultMode = newMode;
  }

  /** Set mode of table (see <a href="#TableModes">Table Modes</a>).
      @param newMode The table mode to use (see {@link #modeList} and {@link #tableModesList}).
   */
  public void setMode (String newMode) {
    setMode(modeStrToNum(newMode));
  }

  /** Get mode of table (see <a href="#TableModes">Table Modes</a>).
      @return The integer mask value corresponding with the names listed in {@link #modeList}.
      @since NeXtMidas 2.7.0
   */
  public int getModeInt () {
    return specificMode;
  }

  /** Get mode of table (see <a href="#TableModes">Table Modes</a>).
      @return One of the names listed in {@link #modeList}.
      @since NeXtMidas 2.7.0
   */
  public String getMode () {
    return Parser.get(modeList, getModeInt());
  }

  /** Get mode of table in its abbreviated form. <i>This is maintained for backwards-compatibility
      with NeXtMidas 2.6.0 and earlier, most users will want to use {@link #getMode()}.</i>
      @return One of the names listed in {@link #tableModesList}.
   */
  public String getModeName () {
    return Parser.get(tableModesList, getModeInt());
  }

  /** Checks to see if the Table is using a given mode (see <a href="#TableModes">Table Modes</a>).
      If mode is {@link #MAP} or {@link #KEY} this will check to see if the mode being used is
      compatible with that mode (e.g. <tt>isMode({@link #MAP})</tt> will return true for a mode
      of {@link #MAP}, {@link #HT}, {@link #HM}, etc. and <tt>isMode({@link #KEY})</tt> will
      return true for a mode of {@link #KEY} or {@link #KV}).
      @param mode The mode to check for (see {@link #modeList} and {@link #tableModesList}).
      @return true if that mode is being used, false otherwise.
      @since NeXtMidas 2.7.0
   */
  public boolean isMode (int mode) {
    if (mode == specificMode) return true;
    if (mode == MAP) return (genericMode == MAP);
    if (mode == KEY) return (genericMode == KEY) || (genericMode == KV);
    return false;
  }

  /** Checks to see if the Table is using a given mode (see <a href="#TableModes">Table Modes</a>).
      If mode is {@link #MAP} or {@link #KEY} this will check to see if the mode being used is
      compatible with that mode (e.g. <tt>isMode({@link #MAP})</tt> will return true for a mode
      of {@link #MAP}, {@link #HT}, {@link #HM}, etc. and <tt>isMode({@link #KEY})</tt> will
      return true for a mode of {@link #KEY} or {@link #KV}).
      @param mode The mode to check for (see {@link #modeList} and {@link #tableModesList}).
      @return true if that mode is being used, false otherwise.
      @since NeXtMidas 2.7.0
   */
  public boolean isMode (String mode) {
    return isMode(modeStrToNum(mode));
  }

  /** Set Mode of Table. Note that in the case of {@link #MAP} and {@link #KEY} a specific
      implementation of those interfaces will be used; as such a subsequent call to
      {@link #getMode()} may return a value other than {@link #MAP} or {@link #KEY}.
      @param newMode The table mode to use (see <a href="#TableModes">Table Modes</a>).
   */
  public void setMode (int newMode) {
    if (isMode(newMode)) return;  // No change

    // Converting to FB is a special case that needs to be handled apart from the rest
    if (newMode == FB) {
      fb = toBytes();
      specificMode = FB;
      genericMode  = FB;
      return;
    }

    int       oldMode = genericMode;
    byte[]    oldFB   = fb;
    KeyVector oldKV   = kv;
    Map<String,Object> oldMAP = map;
    Keyable   oldKEY  = kbl;

    fb  = null;
    kv  = null;
    map = null;
    kbl = null;

    if (newMode == MAP) newMode = SLHM; // Use implementation of Map     (may change in future)
    if (newMode == KEY) newMode = KV;   // Use implementation of Keyable (may change in future)

    specificMode = newMode;
    switch (newMode) {
      case KV:   genericMode = KV;  kv  = new KeyVector(32);                                break;
      case HT:   genericMode = MAP; map = new Hashtable<String,Object>();                   break;
      case HM:   genericMode = MAP; map = new HashMap<String,Object>();                     break;
      case LHM:  genericMode = MAP; map = new LinkedHashMap<String,Object>();               break;
      case SLHM: genericMode = MAP; map = Collections.synchronizedMap(new LinkedHashMap<String,Object>()); break;
      case TM:   genericMode = MAP; map = new TreeMap<String,Object>();                     break;
      case STM:  genericMode = MAP; map = Collections.synchronizedMap(new TreeMap<String,Object>());       break;
      case CHM:  genericMode = MAP; map = new ConcurrentHashMap<String,Object>();           break; // since 3.1.1
      default:   throw new IllegalArgumentException("Unknown table mode "+newMode);
    }

    switch (oldMode) {
      case FB:  if (oldFB != null) fromBytes(oldFB,0,oldFB.length); break;
      case KV:  merge(oldKV);                                       break;
      case MAP: putAll(oldMAP);                                     break;
      case KEY: merge(oldKEY);                                      break;
    }
  }

  public Object getCore() {
    Object obj=null;
    switch (genericMode) {
      case FB:  obj=fb; break;
      case KV:  obj=kv; break;
      case MAP: obj=map; break;
      case KEY: obj=kbl; break;
    }
    return obj;
  }

  /** Ensures that the mode is not FB, if it is, change it. */
  private void checkModeFB () {
    if (genericMode<=FB) setMode(defaultMode);
  }

  /** Return the entry from the table that is located under the given key. <br>
      <br>
      Four particular key names (and their abbreviations) have special significance:
      <pre>
        "GETPREV", "Prev"  - same as {@link #getPrevious()}
        "GETKEYS", "Keys"  - same as {@link #getKeys()}
        "GETLIST", "List"  - same as {@link #getKeyList()}
        "GETSIZE". "Size"  - same as {@link #size()}
      </pre>
      This method is intended for Macro users. API users
      should call get(String) or the above methods directly.<br><br>
      @param key The name of the object to look up.
      @return The object that is currently stored under the given key.
      @throws NullPointerException if <tt>key</tt> is <tt>null</tt>
  */
  public Object getKey (String key) throws NullPointerException {
    if (key.equals("Prev") || key.equals("Keys") || key.equals("List") || key.equals("Size")) {
      Shell.deprecate("getKey('"+key+"') is deprecated. Use getKey('GET" +key.toUpperCase()+"')."); // for Bug 1749
    }

    int len = key.length();
    if (len!=7 && len!=4) return getx(key);
    else if (key.equalsIgnoreCase("GETPREV") || key.equals("Prev")) return getPrevious();
    else if (key.equalsIgnoreCase("GETKEYS") || key.equals("Keys")) return getKeys();
    else if (key.equalsIgnoreCase("GETLIST") || key.equals("List")) return getKeyList();
    else if (key.equalsIgnoreCase("GETSIZE") || key.equals("Size")) return size();
    else if (key.equalsIgnoreCase("GETCORE") || key.equals("Core")) return getCore();
    else return getx(key);
  }

  /** Sets the key in the table. This is the same as <tt>put(key,value)</tt>, except that it
      returns <tt>value</tt> following the put.
      @param key   The key to insert (not <tt>null</tt>). See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param value The value to insert (any <tt>null</tt> values will be changed to <tt>"NULL"</tt>
                   on insertion unless {@link #EXCEPTION_ON_NULL} is set).
      @return <tt>value</tt> (note that this returns *exactly* the value passed in, and does not
              reflect any conversion from <tt>null</tt> to "NULL").
      @throws NullPointerException if <tt>key</tt> is <tt>null</tt> or if <tt>value</tt> is
                                   <tt>null</tt> and {@link #EXCEPTION_ON_NULL} is set.
   */
  public Object setKey (String key, Object value) {
    // no special stuff so use put rather than setKey
    put(key,value);
    return value;	// always succeeds
  }

  /** Remove an entry from the table. This is identical to <tt>remove(key,0)</tt>.
      @param key The name of the key to remove. See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @return oldVal The value previously associated with that key (since NeXtMidas 3.1.0)
      @throws NullPointerException if <tt>key</tt> is <tt>null</tt>
   */
  public Object remove (String key) throws NullPointerException {
    return remove(key,0);
  }

  /** Remove an entry from the table.
      @param key The name of the key to remove. See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param attr  The attributes associated with the remove.
      @return oldVal The value previously associated with that key (since NeXtMidas 3.1.0)
      @throws NullPointerException if <tt>key</tt> is <tt>null</tt>
      @since NeXtMidas 2.5.5
   */
  public synchronized Object remove (String key, int attr) throws NullPointerException {
    // since 3.1.1: made method synchronized for thread safety (BUG 2500)
    if ((flags&READONLY)!=0 && (attr&FORCE)==0) {
      Shell.warning("Must use FORCE flag to remove tag="+key+" from ReadOnly table");
      return null; // Bug 1343
    }
    if (key == null) return null;
    Object oldVal = get(key);
    checkModeFB();
    int i = key.indexOf('.');
    if (i>0) {
      String keyroot = key.substring(0,i);
      Object root = getx(keyroot);
      if (root instanceof Table) {
        ((Table)root).remove(key.substring(i+1), attr);
        return oldVal;
      }
    }
    if (key.length() > 0) {
      switch (genericMode) {
        case KEY: kbl.setKey(key,null);  break;
        case KV:  kv.remove(key);        break;
        case MAP: {
          if (en !=null && en.iter != null) en.iter.remove(); // Bug 1791 Table.remove() incompatible with SLHM
          else map.remove(key);
          break;
        }
        default:  throw new AssertionError("Unknown table mode "+genericMode);
      }
    }
    return oldVal;
  }

  /** Put a generic entry into the table under the given key.
      This is identical to <tt>put(key,value,0,null)</tt>.
      @param key   The key to insert (not <tt>null</tt>). See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param value The value to insert (any <tt>null</tt> values will be changed to <tt>"NULL"</tt>
                   on insertion unless {@link #EXCEPTION_ON_NULL} is set).
      @throws NullPointerException if <tt>key</tt> is <tt>null</tt> or if <tt>value</tt> is
                                   <tt>null</tt> and {@link #EXCEPTION_ON_NULL} is set in flags.
   */
  public void put (String key, Object value) throws NullPointerException {
    put(key,value,0,null);
  }

  /** Put a generic entry into the table under the given key with flag attributes.
      This is identical to <tt>put(key,value,attr,null)</tt>.
      @param key   The key to insert (not <tt>null</tt>). See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param value The value to insert (any <tt>null</tt> values will be changed to <tt>"NULL"</tt>
                   on insertion unless {@link #EXCEPTION_ON_NULL} is set).
      @param attr  The attributes associated with the put.
      @throws NullPointerException if <tt>key</tt> is <tt>null</tt> or if <tt>value</tt> is
                                   <tt>null</tt> and {@link #EXCEPTION_ON_NULL} is set in flags.
      @since NeXtMidas 2.5.4
   */
  public void put (String key, Object value, int attr) throws NullPointerException {
    put(key,value,attr,null);
  }

  /** Put a generic entry into the table under the given key with flag attributes.
      @param key   The key to insert (not <tt>null</tt>). See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param value The value to insert (any <tt>null</tt> values will be changed to <tt>"NULL"</tt>
                   on insertion unless {@link #EXCEPTION_ON_NULL} is set).
      @param attr  The attributes associated with the put.
      @param ref   The reference object for translation.
      @throws NullPointerException if <tt>key</tt> is <tt>null</tt> or if <tt>value</tt> is
                                   <tt>null</tt> and {@link #EXCEPTION_ON_NULL} is set in flags.
   */
  public void put (String key, Object value, int attr, Object ref) throws NullPointerException {
    checkModeFB();
    int i;
    if ((i=key.indexOf('.')) > 0) {
      String keyroot = key.substring(0,i);
      key = key.substring(i+1);
      Object root=getx(keyroot);
      if (root==null) throw new NullPointerException("Table Entry "+keyroot+" not found for put of "+key);
      while ( (i=key.indexOf('.')) > 0 ) {
        Object prevRoot = root;
        keyroot = key.substring(0,i);
        key = key.substring(i+1);
        if (root instanceof Table) root=((Table)root).getx(keyroot);
        else root = KeyObject.getKey (root,keyroot,null,ref);
        if (root==null && (attr&ADDROOT)!=0) {
          root=new Table(specificMode,flags);
          if (prevRoot instanceof Table) ((Table)prevRoot).putx(keyroot,root,attr);
          else throw new MidasException("Could not AddRoot to table at key = "+key);
        }
      }
      if (root instanceof Table) ((Table)root).putx(key,value,attr);
      else KeyObject.setKey (root,key,value);
    }
    else putx(key,value,attr);
  }

  /** Put a generic entry into the table under the given key if if there isn't an existing entry present at that key.
      This is similar to the idea in the {@link java.util.Set#add(Object)} method.
      @param key   The key to insert (not <tt>null</tt>). See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param value The value to insert (any <tt>null</tt> values will be changed to <tt>"NULL"</tt>
               on insertion unless {@link #EXCEPTION_ON_NULL} is set).
      @return true if key/value is added, otherwise false (i.e. there is an existing entry at specified key).
      @throws NullPointerException if <tt>key</tt> is <tt>null</tt> or if <tt>value</tt> is
                               <tt>null</tt> and {@link #EXCEPTION_ON_NULL} is set in flags.
      @see #put(String, Object) for normal put use put(..) and its various overloaded methods.
      @since NeXtMidas 3.1.0
   */
  public boolean addIfNotPresent (String key, Object value) throws NullPointerException {
    if (containsKey(key)) return false; // return false: key/value was NOT added to table
    put(key,value,0,null);
    return true;                        // return true: key/value was added to table
  }

  /** Merge key/values from specified Map into this Table for any key which is NOT already present.
      @param map Map (e.g. Table) to merge in key/values that are not already present.
      @return true if any key/value map is added, otherwise false (i.e. there is an existing entry at specified key).
      @see #addIfNotPresent(String, Object)
      @since NeXtMidas 3.1.0

   */
  public boolean addIfNotPresent (Map<String,Object> map) throws NullPointerException {
    if (map==null) return false; // return false: no keys/values was NOT added to table
    boolean retval = false;
    java.util.Iterator<Map.Entry<String,Object>> iter = map.entrySet().iterator();

    while (iter.hasNext()) {
      Map.Entry<String, Object> entry = iter.next();
      retval |= addIfNotPresent(entry.getKey().toString(), entry.getValue());
    }

    return retval;
  }

  /** Adds a new empty table to this table at the specified key if it is not already present.
      The table added will initially be empty.
      See important note on the <a href="#dotsyntax">dot syntax here</a> for key argument.
      @param key The name of the new table.
      @return A pointer to the new table (after adding it to this table) if key is not already present,
              otherwise a pointer to the existing value at that key in this Table.
   */
  public synchronized Object addTable (String key) {
    Object root = get(key);
    if (root!=null) return root;
    root = new Table(specificMode,flags);
    putx(key,root);
    return root;
  }

  /** Put an entry into *this* table. Unlike the various <tt>put(..)</tt> methods, the key must NOT
      contain any dots.
      @param key   The key to insert (not <tt>null</tt>).
      @param value The value to insert (any <tt>null</tt> values will be changed to <tt>"NULL"</tt>
                   on insertion unless {@link #EXCEPTION_ON_NULL} is set).
      @throws NullPointerException if <tt>key</tt> is <tt>null</tt> or if <tt>value</tt> is
                                   <tt>null</tt> and {@link #EXCEPTION_ON_NULL} is set.
   */
  public void putx (String key, Object value) throws NullPointerException {
    putx(key,value,0);
  }

  /** Put an entry into *this* table. Unlike the various <tt>put(..)</tt> methods, the key must NOT
      contain any dots.
      @param key   The key to insert (not <tt>null</tt>).
      @param value The value to insert (any <tt>null</tt> values will be changed to <tt>"NULL"</tt>
                   on insertion unless {@link #EXCEPTION_ON_NULL} is set).
      @param attr  The attributes associated with the put.
      @throws NullPointerException if <tt>key</tt> is <tt>null</tt> or if <tt>value</tt> is
                                   <tt>null</tt> and {@link #EXCEPTION_ON_NULL} is set.
   */
  public synchronized void putx (String key, Object value, int attr) throws NullPointerException {
    // since 3.1.1: made method synchronized for thread safety (BUG 2500)

    if ((flags&READONLY)!=0 && (attr&FORCE)==0) {
      Shell.warning("Must use FORCE flag to put tag="+key+" into ReadOnly table");
      return; // Bug 1343
    }
    if (listener!=null && listener.putAction(key,value)) return;
    checkModeFB();
    if (value==null) {
      if ((flags & EXCEPTION_ON_NULL) != 0) {
        throw new NullPointerException("Table: Can not insert "+key+"="+value+" into table.");
      }
      value = "NULL";
    }

    // check for global table reference
    if (prev!=null) {
      Object root=null;
      switch (genericMode) {
        case KV:  root = kv.get(key);      break;
        case MAP: root = map.get(key);     break;
        case KEY: root = kbl.getKey(key);  break;
        default:  throw new AssertionError("Unknown table mode "+genericMode);
      }
      if (root instanceof Results) {
        // update value only if differs from original
        if (root != value) ((Table)root).putx(key,value,attr);
        return;
      }
    }

    switch (genericMode) {
      case KV:  if ((flags&ADD)!=0) kv.add(key,value);
                else                kv.put(key,value); break;
      case MAP: map.put(key,value);                    break;
      case KEY: kbl.setKey(key,value);                 break;
      default:  throw new AssertionError("Unknown table mode "+genericMode);
    }
  }

  /** Get a generic entry/value from the table using the specified key.
      @param key      The name of the entry to get. See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param defValue The value to return if specified key is not found.
      @param options  Table flags/options to apply. E.g. {{@link #ONLY_DOT_INTO_SUBTABLES}.
      @return The value associated with the key or <code>defValue</code> if the specified key is not
              part of the table.
      @throws NullPointerException If the key is <code>null</code>.
      @see #containsKey(String)
      @since NeXtMidas 2.9.2
   */
  public Object get (String key, Object defValue, int options) {
    // checkModeFB(); // don't need to call it here, since we call it in getx() below
    Object value = null;
    boolean drillIntoProperties = (options & ONLY_DOT_INTO_SUBTABLES) == 0; // for BUG 2333 and BUG 2334
    int ii;
    if ((ii=key.indexOf('.')) > 0) {
      String keyroot = key.substring(0,ii);
      String subkey = key.substring(ii+1);
      Object root = getx(keyroot);
      if (root==null) return defValue;
      while (root!=null && (ii=subkey.indexOf('.')) > 0 ) {
        keyroot = subkey.substring(0,ii);
        subkey  = subkey.substring(ii+1);
        if (root instanceof Table)    root = ((Table)root).getx(keyroot);
        else if (drillIntoProperties) root = KeyObject.getKey(root, keyroot);
      } // end while loop
      if (root==null) value = defValue;
      else if (root instanceof Table) value = ((Table)root).getx(subkey);
      else if (drillIntoProperties)   value = KeyObject.getKey(root, subkey);
    }
    else value = getx(key);
    if (value==null) value = defValue;
    return value;
  }

  /** Same as <code>get(key, defValue, flags)</code>.
      @see #get(String, Object, int)
      @since NeXtMidas 2.9.2
   */
  public Object get (String key, Object defValue) {
    return get(key, defValue, flags);
  }

  /** Get a generic entry from the table using the specified key.
      This is the same as <code>get(key, null, flags)</code>.
      @param key The name of the entry to get. See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @return The value associated with the key or <code>null</code> if the specified key is not
              part of the table.
      @throws NullPointerException If the key is <code>null</code>.
      @see #get(String, Object, int)
      @see #containsKey(String)
   */
  public Object get (String key) throws NullPointerException {
    return get(key, null, flags);
  }

  /** Gets an entry from *this* table. Unlike the <tt>get(..)</tt> method, the key must NOT
      contain any dots.
      @param key The key to get (not <tt>null</tt>).
      @return The value from the table (null if not found).
      @throws NullPointerException If the key is <code>null</code>.
   */
  public Object getx (String key) throws NullPointerException {
    checkModeFB();
    Object value=null;
    switch (genericMode) {
      case KV:  value = kv.get(key);      break;
      case MAP: value = map.get(key);     break;
      case KEY: value = kbl.getKey(key);  break;
      default:  throw new AssertionError("Unknown table mode "+genericMode);
    }
    if (prev==null);
    else if (value==null && (flags&AUTOREADPREV)!=0) value = prev.getx(key);
    else if (value instanceof Results) value = ((Table)value).getx(key);
    if (listener!=null) listener.getAction(key,value);
    return value;
  }

  /** Find the key matching the given abbreviation. The abbreviation for a key is the first key
      found that starts with the given abbreviation.
      @param abbr The abbreviation.
      @return The name of the key that matches the abbreviation or <code>null</code> if not found.
   */
  public String match (String abbr) {
    if (containsKey(abbr)) return abbr;	// found non-abbreviated
    for (Enumeration<String> e=keys(); e.hasMoreElements(); ) {
      String key = e.nextElement();
      if (key.startsWith(abbr)) return key;
    }
    return null;
  }

  /** Initializes an enumeration list of the table. The enumeration is the accessible using the
      {@link #more()} and {@link #next()} methods.
      @return The current size of the table.
   */
  public int enumeration() {
    en = new Enum(this);
    return size();
  }

  /** Clear the Table of all entries. This is identical to <tt>clear(0)</tt>. */
  public void clear() {
    clear(0);
  }

  /** Clear the Table of all entries.
      @param attr The attributes associated with the clear.
      @since NeXtMidas 2.5.5
   */
  public void clear (int attr) {
    if ((flags&READONLY)!=0 && (attr&FORCE)==0) {
      Shell.warning("Must use FORCE flag to clear ReadOnly table");
      return; // Bug 1343
    }
    if (genericMode == KEY) {
      // Keyable has no clear method, so just iterate over all keys and remove each one.
      String[] keys = kbl.getKeys();
      for (int i = 0; i < keys.length; i++) {
        remove(keys[i],attr);
      }
    }
    else {
      switch (genericMode) {
        case FB:  fb = null;       break;
        case KV:  kv.clear();      break;
        case MAP: map.clear();     break;
        case KEY: throw new AssertionError("This should be impossible."); // KEY handled above
        default:  throw new AssertionError("Unknown table mode "+genericMode);
      }
    }
  }

  /** Indicates if the table is empty. */
  public boolean isEmpty() {
    return size()==0;
  }

  /** Gets the number of elements in the table. */
  public int size() {
    if ((genericMode == FB) && (fb == null)) return 0; // Don't bother converting if it is empty
    checkModeFB();
    switch (genericMode) {
      case KV:  return kv.size();
      case MAP: return map.size();
      case KEY: return kbl.getKeys().length;
      default:  throw new AssertionError("Unknown table mode "+genericMode);
    }
  }

  /** Gets the number of elements in the table. This is identical to <code>size()</code>. */
  public int getSize() { return size(); }

  /** Indicates if the table contains the given key (should NOT contain dots) as this
      queries the underlining container directly.
      @param key The key to look for (not null, and no dots).
      @return <code>true</code> if Table contains this key, <code>false</code> otherwise.
      @see #containsKey(String)
      @since NeXtMidas 2.9.2 this was the original implementation of the <tt>containsKey<tt> method.
   */
  public boolean containsKeyx (String key) {
    checkModeFB();
    switch (genericMode) {
      case KV:  return (kv.findKey(key) >= 0);
      case MAP: return map.containsKey(key);
      case KEY: return (kbl.getKey(key) != null);
      default:  throw new AssertionError("Unknown table mode "+genericMode);
    }
  }

  /** Indicates if the table contains the given key. Since NeXtMidas 2.9.2, the dot syntax can be
      used to drill down into keys for sub-Table(s) at any level. Note that this is different
      than the {@link #get(String)} method's <a href="#dotsyntax">dot syntax</a> behavior,
      since this forces the {@link #ONLY_DOT_INTO_SUBTABLES} option.
      Use {@link #containsKeyx(String)} for old behavior to NOT drill down into sub-Tables.
      @param key The key to look for (not null). The dot syntax can be used to drill
                 down into keys for any sub-Table (i.e. not object's properties/methods).
      @return <code>true</code> if Table contains this key, <code>false</code> otherwise.
      @see #containsValue(Object)
      @see #get(String, Object, int)
      @see #containsKeyx(String)
   */
  public boolean containsKey (String key) {
    Object val = get(key, null, flags|ONLY_DOT_INTO_SUBTABLES);
    return (val != null);
  }

  /** Indicates if the table contains the given value. <br>
      <br>
      <font color="red">Warning: The behavior for {@link #KV} mode is non-standard and does not
      match the {@link Map} interface. If the table is in {@link #KV} mode this does an identity
      check (<tt>==</tt>), rather than an equality (<tt>.equals(..)</tt>) check. (This is likely
      to change in future versions.)</font>
      @param value The value to look for.
      @return <code>true</code> if Table contains this value, <code>false</code> otherwise.
      @see #containsKey(String)
   */
  public boolean containsValue (Object value) {
    checkModeFB();
    if (genericMode == KEY) {
      // No containsValue method in Keyable, need to iterate over each value and check it.
      String[] keys = kbl.getKeys();
      for (int i = 0; i < keys.length; i++) {
        if (value.equals(kbl.getKey(keys[i]))) return true;
      }
      return false;
    }
    else {
      switch (genericMode) {
        case KV:  return (kv.findValue(value) >= 0);
        case MAP: return map.containsValue(value);
        default:  throw new AssertionError("Unknown table mode "+genericMode);
      }
    }
  }

  /** Return true if there are more elements in the enumeration list.
      Must use {@link #enumeration()} to initialize the enumeration first.
      @return <code>true</code> if the enumeration has more elements, <code>false</code> otherwise.
              (Returns <code>false</code> if {@link #enumeration()} has not been called.)
      @see #enumeration()
      @see #next()
   */
  public boolean more () {
    if (en==null) return false;
    return en.hasMoreElements();
  }

  /** Return the next key in the table enumeration list.
      Must use {@link #enumeration()} to initialize the enumeration first.
      @return The next key in the enumeration.
              (Returns <code>null</code> if {@link #enumeration()} has not been called.)
      @see #enumeration()
      @see #more()
   */
  public String next () {
    if (en==null) return null;
    return en.nextElement();
  }

  /** Dump the contents of the table to {@link System#out}. This is identical to
      <code>dump(System.out)</code> and is intended for debugging purposes only.
   */
  public void dump () { dump(System.out); }

  /** Dump the contents of the table to the given print stream. This is intended for debugging
      purposes only.
      @param dev The print stream (device) to dump the table to.
   */
  public void dump (java.io.PrintStream dev) {
    for (Iterator ti=iterator(); ti.getNext();) dev.println (ti.key+" = "+ti.value);
  }

  public String[] getKeys () {
    // JC4: This probably could be more efficient
    Iterator ti = iterator();
    return ti.keys;
  }

  /** Gets the list of keys as a comma separated list.
      This is the same as <code>getKeyList(null)</code>.
      @return The list of keys.
   */
  public String getKeyList() { return getKeyList(null); }

  /** Gets the list of keys matching a filter as a comma separated list.
      @param filter The filter to use to determine which keys to list.
                    (If this is <code>null</code> all keys are listed.)
      @return The list of keys.
   */
  public String getKeyList (Filter filter) {
    String list="";
    for (Iterator ti=iterator(); ti.getNext(); ) {
      if (filter!=null && !filter.accept(ti.value)) continue;
      if (list.length()==0) list = ti.key;
      else list += ","+ti.key;
    }
    return list;
  }

  /** Converts this table to a String.
      @return This table as a string.
   */
  public String toString() {
    if (fb!=null) return new String(fb);
    return toStringBuffer().toString();
  }

  /** Converts this Table to a String. Same as {@link #toString()}, but this method
      translates escape sequences where needed (tabs, quotes, Unicode chars, etc)
      @return  The Table as a String
      @since NeXtMidas 2.9.0
   */
  public String getEscapedString () {
    return new String(toBytes());  // toBytes add the escapes back in
  }

  /** Serialize to a byte array.<br><br>
      @return This table as a byte array.
   */
  public byte[] toBytes() {
    StringBuffer sb = new StringBuffer(1024);
    toStringBuffer(this,sb,(flags&DISABLE_ESCAPE_SEQUENCES)==0);
    byte[] bytes = new byte[sb.length()];
    for (int i=0; i<bytes.length; i++) bytes[i] = (byte)sb.charAt(i);
    return bytes;
  }

  /** Converts the table to a string buffer. <br><br>
      <b>Note</b>: This method does NOT add any escape characters into the String.
      @return This table as a string buffer.
   */
  public synchronized StringBuffer toStringBuffer() {
    StringBuffer sb = new StringBuffer(1024);
    toStringBuffer(this,sb,false);
    return sb;
  }

  /** Adds the serialized string value of a given object to a string buffer. <br>
      <br>
      The object is added to the {@link StringBuffer} using the following formatting:
      <pre>
        <u>Class        Added As...                                  </u>
        <i>null</i>            "null"
        String          The string, enclosed in double-quotes.
        Data            In the format T:VAL, where T is the type
                        (D,F,X,L,I,B) and VAL is the value.
        KeyVector       Uses {@link KeyVector#toString()}.
        Table           Uses {@link KeyVector#toString()}.
        java.util.Map   Similar to that used for a Table.
        java.util.List  Similar to that used for an array.
        <i>array</i>    In the form (VAL0,VAL1,...,VALN)
        <i>all others</i> Uses {@link Object#toString()}.
      </pre>

      @param obj The object to add to it.
      @param sb  The string buffer to add to.
      @param esc Whether to use escape sequences.
   */
  public static void toStringBuffer (Object obj, StringBuffer sb, boolean esc) {
    if (obj==null) {
      sb.append("null");
      return;
    }
    boolean isTable=(obj instanceof Table);
    boolean addTblSig=esc && !isTable;
    if (isTable) {
      Table tab = (Table)obj;
           if (tab.fb!=null)         obj = tab.fb;
      else if (tab.genericMode==KV ) obj = tab.kv;
      else if (tab.genericMode==MAP) obj = tab.map;
      else if (tab.genericMode==KEY) obj = tab.kbl;
      else {
        sb.append("{}"); // An empty table
        return;
      }
    }
    if (obj instanceof KeyVector) {
      KeyVector kv = (KeyVector)obj;
      if (addTblSig) sb.append("O:KV:");
      sb.append(LCURLY);
      for (int i=0; i<kv.size(); ) {
        sb.append(kv.getKey(i));
        sb.append(EQUALS);
        toStringBuffer(kv.get(i),sb,esc);
        if (++i<kv.size()) sb.append(COMMA);
      }
      sb.append(RCURLY);
    }
    else if (obj instanceof Hashtable) {
      Hashtable<?,?> ht = (Hashtable<?,?>)obj;
      if (addTblSig) sb.append("O:HT:");
      sb.append(LCURLY);
      for (Enumeration<?> en=ht.keys(); en.hasMoreElements(); ) {
        String key = (String)en.nextElement();
        sb.append(key);
        sb.append(EQUALS);
        toStringBuffer(ht.get(key),sb,esc);
        if (en.hasMoreElements()) sb.append(COMMA);
      }
      sb.append(RCURLY);
    }
    else if (obj instanceof Map) {
      if (addTblSig) {
             if (obj instanceof TreeMap) sb.append("O:TM:");
        else if (obj instanceof LinkedHashMap) sb.append("O:LHM:");
        else if (obj instanceof ConcurrentHashMap) sb.append("O:CHM:");
        else sb.append("O:HM:");
      }
      final Map<?,?> map = (Map<?,?>)obj;
      final java.util.Iterator<?> entries = map.entrySet().iterator();
      sb.append(LCURLY);
      while (entries.hasNext()) {
        final Map.Entry<?,?> entry = (Map.Entry<?,?>)entries.next();
        final Object    key   = entry.getKey();
        final Object    value = entry.getValue();
        sb.append(key); sb.append(EQUALS); toStringBuffer(value,sb,esc);
        if (entries.hasNext()) sb.append(COMMA);
      }
      sb.append(RCURLY);
    }
    else if (obj instanceof List) {
      final java.util.List<?> list = (List<?>)obj;
      sb.append(LPAREN);
      for (int i=0; i<list.size(); i++) {
        if (i != 0) sb.append(COMMA);
        toStringBuffer(list.get(i),sb,esc);
      }
      sb.append(RPAREN);
    }
    else if (obj instanceof byte[]) {
      byte[] fb = (byte[])obj;
      for (int i=0; i<fb.length; i++) sb.append((char)fb[i]);
    }
    else if (obj instanceof String) {
      String str = (String)obj;
      toStringBuffer (str,sb,esc);
    }
    else if (obj instanceof Data) {
      Data data = (Data)obj;
      String str = data.toString();
      char type = (char)data.type;
      if (type=='A') toStringBuffer (str,sb,esc);
      else sb.append(type).append(':').append(str); // number with type prefix
    }
    else if (obj instanceof Keyable) {
      final Keyable  keyable = (Keyable)obj;
      final String[] keys    = keyable.getKeys();
      sb.append('{');
      for (int ii = 0; ii < keys.length; ii++) {
        final String key   = keys[ii];
        final Object value = keyable.getKey(key);
        if (ii != 0) sb.append(',');
        sb.append(key); sb.append('='); toStringBuffer(value,sb,esc);
      }
      sb.append('}');
    }
    else if (obj.getClass().isArray()) {
      final int length = Array.getLength(obj);
      sb.append('(');
      for (int ii = 0; ii < length; ii++) {
        if (ii != 0) sb.append(',');
        toStringBuffer(Array.get(obj,ii),sb,esc);
      }
      sb.append(')');
    }
    else {
      String str = obj.toString();
      if (StringUtil.isNumber(str)) sb.append(str);
      else toStringBuffer (str,sb,esc);
    }
  }

  private static void toStringBuffer (String str, StringBuffer sb, boolean esc) {
    int nbs=0,nesc=0,nq=0,nc=0,iq=0,ip=0,ic=0,ls=str.length();
    // decide if this string needs to be quoted or escaped
    for (int i=0; i<ls; i++) {
      char c=str.charAt(i);
           if (c==QUOTE) { nq++; iq=1-iq; }
      else if (c==BSLASH) nbs++;
      else if (c==COMMA  && iq==0 && ip==0 && ic==0) nc++;
      else if (c==LPAREN && iq==0) ip++;
      else if (c==RPAREN && iq==0) ip--;
      else if (c==LCURLY && iq==0) ic++;
      else if (c==RCURLY && iq==0) ic--;
      else if (c<SPACE || c>TILDE) nesc++;
    }
    // see if we can avoid escaping internal quotes
    if (nq>0 && nbs==0 && nc==0 && nesc==0) sb.append(str);
    // need to add escapes to quoted string (do here not toBytes() for performance and reuse of StringUtil)
    else if (nbs+nesc>0 && esc) sb.append('"').append(StringUtil.escapeString(str)).append('"');
    // quote all other strings for delimmiters
    else sb.append('"').append(str).append('"');
  }

  /** Abbreviated form of {@link #toString()} used for debugging purposes only. */
  public String toStringAbbr() {
    String str = this.toString();
    if (str.length()<20) return str;
    return str.substring(0,20)+"...}";
  }


  /** De-serialize table entries from a byte array.
      @param bytes The byte array.
   */
  public void fromBytes (byte[] bytes) {
    if (genericMode == FB) fb = bytes;
    else { clear(); fromBytes(bytes,0,bytes.length); }
  }

  /** Add table entries from the given table string.
      @param string The new entries to add.
   */
  public void append (String string) {
    byte[] bytes = string.getBytes();
    fromBytes (bytes, 0, bytes.length);
  }

  /** De-serialize from a byte array (specified range).
      @param bytes  the byte array
      @param index1 start index in byte array
      @param index2 end (exclusive index in byte array
      @return status =0 good &lt;0 bad
   */
  public int fromBytes (byte[] bytes, int index1, int index2) {
    int ik1,ik2;  // index of key (start,end)
    int iv1,iv2;  // index of val (start,end)
    int ibr;      // number of braces
    int iqu;      // 1 for in quotes, 0 for out of quotes
    int i = index1;
    int n = index2;
    boolean escSeqEnabled  = (flags&DISABLE_ESCAPE_SEQUENCES)==0;
    boolean checkMalformed = (flags&CHECK_MALFORMED)!=0;

    // trim off whitespace (before open/close braces) for BUG 2095
    while (i<n-1 && bytes[i]  <=SPACE) i++;                    // remove leading white space (left trim)
    while (i<n-1 && bytes[n-1]<=SPACE) n--;                    // remove trailing white space (right trim)

    // error checking
    if (n-i<2)  return -1;
    if (bytes[i]!=LCURLY || bytes[n-1]!=RCURLY) {              // tables must be bracketed by { }
      // exception/warning below added in 3.3.0 for Bug 85
      StringBuilder errmsg = new StringBuilder("Malformed Table (");
      if (bytes[i]!=LCURLY)   errmsg.append(" missing open '{' ,");
      if (bytes[n-1]!=RCURLY) errmsg.append(" missing close '}'");
      errmsg.append(" ) in table: ").append(new String(bytes));
      if (checkMalformed) throw new MidasException(errmsg.toString());
      else Shell.warning(errmsg);
      return -1;
    }

    // trim off whitespace (within braces)
    while (i<n-1 && bytes[i+1]<=SPACE) i++;                    // remove white space at head (after  {)
    while (i<n-1 && bytes[n-2]<=SPACE) n--;                    // remove white space at tail (before })

    // parse loop
    for (i+=1,n-=1; i<n; ) {
      while (i<n && bytes[i]==SPACE) i++;
      while (i<n && bytes[i]==RCURLY) {				// handle {A=2}{}{B=3,C=4} syntax
        int k = i+1;
        while (k<n && bytes[k]==SPACE) k++;
        if (k>=n || bytes[k]!=LCURLY) break;
        i=k+1;
        while (i<n && bytes[i]==SPACE) i++;
      }
      ik1=i;
      ik2=i;
      int nEq = 0;                                  // Count stray ='s which follow other ='s
      boolean inEscSeq = false;
      for (ibr=0,iqu=0; i<n && (ibr>0 || iqu>0 || bytes[i]!=COMMA); i++) {
        byte ibyte = bytes[i];
        if (ibyte==QUOTE && (!escSeqEnabled || !inEscSeq)) {
          iqu = 1-iqu;         // toggle in/out of quotes
          inEscSeq = false;    // a start or end quote resets in escape seq
        }
        else if (iqu!=0) {     // bypass counters inside quotes
          inEscSeq = escSeqEnabled && !inEscSeq && (ibyte==BSLASH); // figure out if in esc seq
        }
        else if (ibyte==LCURLY) ibr++;
        else if (ibyte==RCURLY) ibr--;
        else if (ibyte==LPAREN) ibr++;
        else if (ibyte==RPAREN) ibr--;
        else if (ibyte==EQUALS && ibr==0) { if (++nEq==1) ik2=i; } // found index for end of key (ik2)
        if (ibr<0) break;
      }
      if (i<=ik1) break;
      if (nEq==0) ik2=i;
      iv2=i;
      iv1=ik2+1;
      while (bytes[ik2-1]<=SPACE) ik2--;                       // trim spaces between key and equals
      if (bytes[ik1]==QUOTE)  { ik1++; ik2--; }                // trim quotes around key names

      if (nEq>1) {
        //jph -- eliminated warning until I can filter valid case such as
        //   {SWITCHES=/VERBOSE=ON/STATE=10}
        //String msg = "Malformed Table (missing comma after key) in table: "+new String(bytes,iv1,len);
        //if (checkMalformed) throw new MidasException(msg);
        //else Shell.warning(msg);
      }

      // handle the key
      char type = '_';
      if (bytes[ik1+1]==COLON) { type = (char)bytes[ik1]; ik1+=2; }
      // using Convert.unpackS is faster than new String, because it assumes ASCII
      String key = Convert.unpackS(bytes,ik1,ik2-ik1);
      if ((flags&CASESENSITIVE) == 0) key = key.toUpperCase(); // not case sensitive - convert to uppercase

      // handle the value
      int len=iv2-iv1; // length of value
      if (len<=0) {
        put(key, "");
      }
      else if (bytes[iv1]==QUOTE) {
        // check below added to prevent malformed Tables when given bad input (Since 3.3.0 - Bug 85)
        if (bytes[iv2-1]!=QUOTE) { // missing closing quote
          String msg = "Malformed Table (improperly quoted) in table: "+new String(bytes,iv1,len);
          if (checkMalformed) throw new MidasException(msg);
          else Shell.warning(msg);
        }
        String res = new String(bytes,iv1+1,len-2);
        if (escSeqEnabled) res = StringUtil.unescapeString(res);
        put(key, res);
      }
      else if (bytes[iv1]==LCURLY) {
        Table sub = new Table(specificMode,flags);
        sub.fromBytes (bytes,iv1,iv2);
        put(key, sub);
      }
      else if (len>2 && bytes[iv1+1]==COLON) {
        String res = new String(bytes,iv1+2,len-2);
        put(key, Convert.s2o(res,(char)bytes[iv1],null) );
      }
      else {
        String res = new String(bytes,iv1,len);
        put(key, Convert.s2o(res,type,null) );
      }

      // special handling for merging tables (e.g. "{A=B,C=D}{E=F}" in table syntax)
      if (ibr<0 && i<n && bytes[i+1]==LCURLY) i++;

      // trim off whitespace (between key=value pairs)
      while (i<n && bytes[++i]==SPACE) { }
    }
    return 0;
  }

  /** Set the contents of this table to the contents of the given hashtable. <br>
      <br>
      The Table will have its mode set to Hashtable mode ({@link #HT}) and its contents will
      be replaced with the contents of the parameter. <br>
      <br>
      This method performs a shallow copy of the parameter contents. Upon returning from this
      method the parameter can be modified without affecting the table. Modifying the contents
      of the hash will affect the table contents.
      @param table The hashtable containing the values to insert.
  */
  public void fromTable (Hashtable table) {
    clear();
    setMode(HT);
    putAll(table);
  }

  /** de-serialize from a text file (allows duplicates by using KeyVector mode
      and adding the "ADD" flag during loading). The original table flags will
      then be restored in the end.
      @see Table#fromTextFile(TextFile, int)
   */
  public void fromTextFile (TextFile tf) {
    int saveFlags = flags; // since 2.7.1: save current flags to restore at end of this method (BUG 1838)
    setMode(KV);           // since 2.7.1: by using KeyVector mode and ADD flag (on next line) (BUG 1839)
    fromTextFile(tf, flags|ADD); // allows duplicate keys in Table (KeyVector mode only)
    setFlags(saveFlags);   // restore flags after loading table to match 2.6.x behavior (BUG 1838)
  }

  /** de-serialize from a text file using specified table flags.
      NOTE: if flags contains the {@link #ADD} option, then table mode will be forced KeyVector
      mode to allow duplicate entries.
      @param tf    TextFile to de-serialize from
      @param flags Table flags to use (set) on this Table
      @since NeXtMidas 3.1.2
   * */
  public void fromTextFile (TextFile tf, int flags) {
    char lastchar=',';
    String line;
    StringBuilder sb = new StringBuilder(4096);
    sb.append('{');

    setFlags(flags);
    tf.open(TextFile.INPUT|TextFile.FORCEABORT);
    while ((line=tf.readProper())!=null) {
      line.trim();      // allow for indents
      // automatically add commas at end-of-previous line if appropriate
      if (line.charAt(0)=='}' || line.charAt(0)=='{');
      else if (lastchar=='{');
      else if (lastchar!=',') sb.append(',');
      sb.append(line); // add to buffer
      lastchar = line.charAt(line.length()-1);
    }
    tf.close();
    sb.append('}');
    String str = sb.toString();
    if (str.startsWith("{{") && str.endsWith("}}")) {
      // This happens when the text file includes the start/end braces (e.g. "{A=1,B2=}"). While
      // NeXtMidas will not write out a file like this, M2K will (Bug 965) and it is a common to
      // see in a hand-edited file.
      str = str.substring(1, str.length()-1);
    }
    fromBytes(str.getBytes());
  }

  /** Writes the table out to a text file.
      @param tf The .tbl file to write to.
   */
  public void toTextFile (TextFile tf) {
    tf.open(TextFile.OUTPUT|TextFile.FORCEABORT);
    toTextFile (tf,this,"");
    tf.close();
  }

  /** Writes the table out to a section of a text file with specified indent. Same as
      calling {@link #toTextFile(TextFile, Table, String)}
      @param tf  The currently opened textfile to add to
      @param t   The table to list
      @param pad The String indent to use (usually spaces)
   */
  public static void toTextFile (TextFile tf, Table t, String pad) {
    Table.Iterator ti = t.iterator();
    for (int i=0; i<ti.size; i++) {
      String key = ti.keys[i];
      Object value = ti.values[i];
      if (value instanceof Table) {
        tf.writeln (pad+key+"={");
        toTextFile (tf,(Table)value,pad+"  ");
        tf.writeln (pad+"}");
      }
      else {
        final StringBuffer sb = new StringBuffer();
        t.toStringBuffer(value,sb,(t.flags&DISABLE_ESCAPE_SEQUENCES)==0);
        tf.writeln (pad+key+"="+sb);
      }
    }
  }

  /** Get a {@link Data} object from the table.
      @param key The name of the entry to get. See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @return The named entry as a {@link Data} object or <code>null</code> if not found.
   */
  public Data getData (String key) {
    Object entry = get(key);
    if (entry == null) return null;
    if (entry instanceof String) entry = new Data((String)entry);
    return (Data)entry;
  }

  /** Get a {@link Table} object from the table. The same as calling {@link #getTable(String, boolean)}
      with false for the converto2t boolean.
      @param key The name of the entry to get. See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @return The named entry as a {@link Table} object or null if not found. Also returns null if
              the entry is found, but is not an instance of {@link Table}.
      @throws MidasException when using the dot syntax in the key to drill down into desired
              object but does NOT exist (unless {@link #ONLY_DOT_INTO_SUBTABLES} flag is set).
      @see #getTable(String, Table)
   */
  public Table getTable (String key) {
    return getTable(key, false);
  }

  /** Get a {@link Table} object from the table. If the object returned from
      the given key is not a {@link Table} and the checkVal param is true,
      this method will attempt to convert the Object to a Table.
      @param key The name of the entry to get. See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param converto2t Whether or not to attempt to convert the returned entry to a {@link Table}
                        via {@link Convert#o2t(Object)} when it is not an instance of a Table.
      @return The named entry value as a {@link Table} object or null if not found. If entry is not
              a Table object and converto2t is true, then convert entry to a Table using
              {@link Convert#o2t(Object)}, otherwise return null.
      @throws MidasException when using the dot syntax in the key to drill down into desired
              object but does NOT exist (unless {@link #ONLY_DOT_INTO_SUBTABLES} flag is set).
      @see #getTable(String, Table)
      @since NeXtMidas 2.8.2
   */
  public Table getTable (String key, boolean converto2t) {  // This method added for BUG 2189
    Object entry = get(key);
    if (entry instanceof Table) return (Table)entry;
    if (converto2t)             return Convert.o2t(entry);
    return null;
  }

  /** Get a {@link Table} object from the table at the specified key. If the object returned from
      the given key is not a {@link Table} object, then the defValue is returned.
      The dot syntax can be used to drill down into keys for sub-Table(s) at any level (but NOT
      into the entry value's properties or methods). <br>
      Note that this is different from the standard <a href="#dotsyntax">dot syntax</a> behavior,
      since this forces the {@link #ONLY_DOT_INTO_SUBTABLES} option via call to
      <code>get(key, defValue, flags|ONLY_DOT_INTO_SUBTABLES)</code>.
      @param key The name of the entry to get Table value (not null). The dot syntax can be used to
                 drill down into keys for any sub-Table (but not object's properties/methods).
      @param defValue The default Table value for the parameter if the entry at the specified key does not exist.
      @return The named entry as a {@link Table} object or defValue if not found. Also returns defValue if
               the entry is found, but is not an instance of {@link Table}.
      @see #getTable(String)
      @see #get(String, Object, int)
      @since NeXtMidas 2.9.2
   */
  public Table getTable (String key, Table defValue) {
    Object entry = get(key, defValue, flags|ONLY_DOT_INTO_SUBTABLES);
    if (entry instanceof Table) return (Table)entry;
    else return defValue;
  }

  /** Get a {@link Table} object from the table at the specified key. Same as
      <code>return getTable(key, null)</code>.
      @param key The name of the entry to get Table value (not null). The dot syntax can be used to
                 drill down into keys for any sub-Table (but not object's properties/methods).
      @return The named entry as a {@link Table} object or null if not found. Also returns null if
               the entry is found, but is not an instance of {@link Table}.
      @see #getTable(String, Table)
      @see #get(String, Object, int)
      @since NeXtMidas 2.9.2
   */
  public Table getTableOrNull (String key) {
    return getTable(key, null);
  }

  /** Get a {@link String} object from the table.
      @param key The name of the entry to get. See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @return The named entry as a {@link String} object or null if not found. Also returns null if
              the entry is found, but is not an instance of {@link String}.
   */
  public String getString (String key) {
    Object entry = get (key);
    if (entry instanceof String) return (String)entry;
    return null;
  }

  /** Get a {@link String} object from the table (drilling down into keys for any
      sub-Table (but not object's properties/methods).
      @param key The name of the entry to get. The dot syntax can be used to
                 drill down into keys for any sub-Table (but not object's properties/methods).
      @param defValue The value to return if specified key is not found.
      @return The named entry as a {@link String} object -or- defValue if key not found -or- null if entry
              value is NOT an instance of {@link String}.
      @see #getString(String)
      @since NeXtMidas 3.1.0
   */
  public String getString (String key, String defValue) {
    Object entry = get(key, defValue, flags|ONLY_DOT_INTO_SUBTABLES);
    if (entry instanceof String) return (String)entry;
    return null;
  }

  /** Get a {@link DataFile} object from the table.
      @param key The name of the entry to get. See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @return The named entry as a {@link DataFile} object or null if not found. Also returns null if
              the entry is found, but is not an instance of {@link DataFile}.
   */
  public DataFile getDataFile (String key) {
    Object entry = get (key);
    if (entry instanceof DataFile) return (DataFile)entry;
    return null;
  }

  /** Get a unique key instance name. <b>WARNING: This method was designed for use on small tables
      (usually less than 32 elements) and is extremely inefficient to use on large tables.</b>
   */
  public String getUniqueKey(String key) {
    String keyi = key;
    for (int i=1; containsKey(keyi); i++) keyi=key+i;
    return keyi;
  }

  /** Get a unique key instance name by adding the specified delimiter and the key number to make
      the key name unique. <b>WARNING: This method was designed for use on small tables
      (usually less than 32 elements) and is extremely inefficient to use on large tables.</b>
   */
  public String getUniqueKey(String key, String delim) {
    String keyi = key;
    for (int i=1; containsKey(keyi); i++) keyi=key+delim+i;
    return keyi;
  }

  /** Renames an existing key. If the oldName is not found nothing is done. Note that this operation
      may or may not affect the iteration order (such is undefined and varies based on mode).
      @param oldName The old key name.
      @param newName The new key name.
   */
  public void rename (String oldName, String newName) {
    checkModeFB();
    if (genericMode == KV) {
      kv.rename(oldName, newName);
    }
    else if (genericMode == KEY) {
      Object val = kbl.getKey(oldName);
      if (val != null) kbl.setKey(newName, val);
    }
    else if (genericMode == MAP) {
      Object val = map.remove(oldName);
      if (val != null) map.put(newName, val);
    }
    else {
      throw new AssertionError("Unknown table mode "+genericMode);
    }
  }

  //////////////////////////////////////////////////////////////////////////////
  // Begin implementation of KeyableByType interface
  // See note on the <a href="#dotsyntax">dot syntax</a> for key parameter in
  // class description above.
  //////////////////////////////////////////////////////////////////////////////
  public final String getS (String key) { return getS(key,null);     }
  public final double getD (String key) { return getD(key,0.0d);     }
  public final float  getF (String key) { return getF(key,0.0f);     }
  public final long   getX (String key) { return getX(key,0L);       }
  public final int    getL (String key) { return getL(key,0);        }
  public final short  getI (String key) { return getI(key,(short)0); }
  public final byte   getB (String key) { return getB(key,(byte)0);  }
  public final Object getO (String key) { return getO(key,null);     }
  public final boolean getState (String key)   { return getState(key,false);   }
  public final boolean getNoState (String key) { return getNoState(key,false); }

  public String getS (String key, String def) {
    Object entry=get(key);
    return (entry==null)? def: Convert.o2s(entry);
  }

  public double getD (String key, double def) {
    Object entry=get(key);
    return (entry==null)? def: Convert.o2d(entry);
  }

  public float getF (String key, float def) {
    Object entry=get(key);
    return (entry==null)? def: Convert.o2f(entry);
  }

  public long getX (String key, long def) {
    Object entry=get(key);
    return (entry==null)? def: Convert.o2x(entry);
  }

  public int getL (String key, int def) {
    Object entry=get(key);
    return (entry==null)? def: Convert.o2l(entry);
  }

  public short getI (String key, short def) {
    Object entry=get(key);
    return (entry==null)? def: Convert.o2i(entry);
  }

  public byte getB (String key, byte def) {
    Object entry=get(key);
    return (entry==null)? def: Convert.o2b(entry);
  }

  public boolean getState (String key, boolean def) {
    String str = getS(key,null);
    if (str == null)             return def;
    if (StringUtil.isTrue(str))  return true;
    if (StringUtil.isFalse(str)) return false;
    return (getD(key) > 0);
  }

  public final boolean getNoState (String key, boolean def) {
    return !getState(key,!def);
  }

  public final Object getO (String key, Object def) {
    Object entry = get(key);
    return (entry==null)? def: entry;
  }
  //////////////////////////////////////////////////////////////////////////////
  // End implement KeyableByType interface
  //////////////////////////////////////////////////////////////////////////////

  /** Puts a numeric value into the table. <br>
      This is the same as <code>put(key,Convert.d2o(value));</code>.
      @param key   The key to insert (not <tt>null</tt>). See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param value The value to insert.
   */
  public void put (String key, double value) { put(key,Convert.d2o(value)); }

  /** Puts a numeric value into the table. <br>
      This is the same as <code>put(key,Convert.f2o(value));</code>.
      @param key   The key to insert (not <tt>null</tt>). See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param value The value to insert.
   */
  public void put (String key, float value) { put(key,Convert.f2o(value)); }

  /** Puts a numeric value into the table. <br>
      This is the same as <code>put(key,Convert.x2o(value));</code>.
      @param key   The key to insert (not <tt>null</tt>). See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param value The value to insert.
   */
  public void put (String key, long value) { put(key,Convert.x2o(value)); }

  /** Puts a numeric value into the table. <br>
      This is the same as <code>put(key,Convert.l2o(value));</code>.
      @param key   The key to insert (not <tt>null</tt>). See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param value The value to insert.
   */
  public void put (String key, int value) { put(key,Convert.l2o(value)); }

  /** Puts a numeric value into the table. <br>
      This is the same as <code>put(key,Convert.i2o(value));</code>.
      @param key   The key to insert (not <tt>null</tt>). See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param value The value to insert.
   */
  public void put (String key, short value) { put(key,Convert.i2o(value)); }

  /** Puts a numeric value into the table. <br>
      This is the same as <code>put(key,Convert.b2o(value));</code>.
      @param key   The key to insert (not <tt>null</tt>). See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param value The value to insert.
   */
  public void put (String key, byte value) { put(key,Convert.b2o(value)); }

  /** Puts a boolean value into the table. <br>
      This is the same as using <code>put(key,Boolean.TRUE)</code> or <code>put(key,Boolean.FALSE)</code>.
      @param key   The key to insert (not <tt>null</tt>). See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param value The value to insert.
      @see #put(String, Object)
   */
  public void put (String key, boolean value) { put(key, (value)? Boolean.TRUE : Boolean.FALSE); }

  /** Starting with the current table, walk backward through the "previous" links until it finds
      the root table (located at the very end of the list because it was the first one created).
      The root table is the one whose "prev" link is set to null.
      @return The root table.
   */
  public Table getRoot() {
    Table tbl = this;
    while (tbl.prev != null) tbl = tbl.prev;
    return tbl;
  }

  /** Update the "previous" link so that this table points back to its predecessor, maintaining
      the linked list of tables that eventually lead back to the global table.
      @param  prevTable to set "previous" link to
   */
  public void setPrevious (Table prevTable) {
    if (prev != null && prevTable != null && prevTable != prev) {
      Shell.warning("Table.setPrevious overwriting existing previous link");
    }
    prev = prevTable;
  }

  /** Gets the previous link, see {@link #setPrevious} for details.
      @return The previous link.
      @see #setPrevious(Table)
   */
  public Table getPrevious() { return prev; }

  /** Sets the default Table operating flags. Multiple flags can be set by OR'ing them together.
      Care must be taken in setting this since any change to the default flags could affect existing
      functionality that expects the old (NeXtMidas 2.6.0) default of {@link #AUTOREADPREV}.
      @param mask The flag mask.
      @since NeXtMidas 2.7.1
   */
  public static void setDefaultFlags (int mask) {
    defaultFlags = mask;
  }

  /** Sets the default Table operating flags. Multiple flags can be set by OR'ing them together.
      Care must be taken in setting this since any change to the default flags could affect existing
      functionality that expects the old (NeXtMidas 2.6.0) default of {@link #AUTOREADPREV}.
      @param mask The flag mask (see {@link #flagsList}).
      @since NeXtMidas 2.7.1
   */
  public static final void setDefaultFlags (String mask) {
    setDefaultFlags(Parser.mask(flagsList, mask, getDefaultFlagsInt()));
  }

  /** Gets the default Table flags.
      @return The flag mask (see {@link #flagsList}).
      @since NeXtMidas 2.7.1
   */
  public static int getDefaultFlagsInt () {
    return defaultFlags;
  }

  /** Gets the Table operating flags.
      @return The flag mask (see {@link #flagsList}).
      @since NeXtMidas 2.7.1
   */
  public static final String getDefaultFlags () {
    return Parser.mask2s(flagsList, getDefaultFlagsInt());
  }

  /** Sets the Table operating flags. Multiple flags can be set by OR'ing them together.
      Since NeXtMidas 2.7.1, if the ADD flag is specified, it will change this Table to
      KeyVector mode to allow duplicate keys in the Table.
      @param mask The flag mask.
   */
  public void setFlags (int mask) {
    flags = mask;
    if ((flags&ADD)!=0) setMode(KV); // ADD flag implies KeyVector mode (Bug 1839)
  }

  /** Sets the Table operating flags.  Multiple flags can be set by OR'ing them together.
      @param mask The flag mask (see {@link #flagsList}).
      @see #setFlags(int)
   */
  public final void setFlags (String mask) {
    setFlags(Parser.mask(flagsList, mask, getFlagsInt()));
  }

  /** Gets the Table operating flags.
      @return The flag mask as an int (see {@link #flagsList}).
      @since NeXtMidas 2.5.0
   */
  public int getFlagsInt () {
    return flags;
  }

  /** Gets the Table operating flags.
      @return The flag mask as a String (see {@link #flagsList}).
      @since NeXtMidas 2.7.0
   */
  public final String getFlags () {
    return Parser.mask2s(flagsList, getFlagsInt());
  }

  /** Adds a Table listener. <br>
      @param listener The Table.Listener object.
   */
  public void addListener (Listener listener) {
    if (this.listener!=null && this.listener!=listener) {
      throw new MidasException("Table currently only supports one listener");
    }
    this.listener = listener;
  }
  /** Removes a Table listener. <br>
      @param listener The Table.Listener object.
   */
  public void removeListener (Listener listener) {
    if (listener==this.listener) listener=null;
  }


  /** Gets an enumeration of the keys in this table.
      @return An enumeration of the keys in this table.
   */
  public Enumeration<String> keys() { return new Enum(this); }

  /** Gets the hashtable that holds the data for this table. <br>
      If the mode is not set to hashtable ({@link #HT}), this method will set it to such.
      @return The hashtable that holds the data for this table.
   */
  public Hashtable<String,Object> getHashtable() {
    setMode(HT);
    return (Hashtable<String,Object>)map;
  }

  /** returns an inverse of this table (ie value=key) */
  public Table getInverse() {
    Table t = new Table(specificMode,flags);
    Iterator ti = this.iterator();
    for (int i=0; i<ti.size; i++) t.put( ti.values[i].toString(), ti.keys[i] );
    return t;
  }

  /** Merges a given keyable object (usually a Table) into this one. This copies all the values
      in the given keyable instance into this table overriding any existing entries (in this table)
      that match new entries in the keyable object. Note that this is a shallow merge in that it
      will not recursively enter any sub-tables. This method was added to speed up merging (vs
      using the {@link #putAll(Map)} method) of keyable objects. This speedup is most notable when
      the given a {@link KeyVector} or a Table in KV mode.
      @param keyable The keyable object to merge into this one.
      @since NeXtMidas 2.3.0
   */
  public void merge(Keyable keyable) {
    // JC4: Note that instances of KeyVector (or tables in KV mode) can be copied
    //      faster if they are accessed via index rather than key.
    KeyVector newKV  = null;
    Table     newTbl = null;

    if (keyable instanceof Table) {
      newTbl = (Table)keyable;
      newKV  = newTbl.kv;
    }
    else if (keyable instanceof KeyVector) {
      newKV = (KeyVector)keyable;
    }

    if (newKV != null) {
      if (newKV.size() == 0) {
        // do nothing
      }
      else if ((getSize() == 0) && (getModeInt() == KV)) {
        // If this is in KV mode and it is empty (e.g. called by copy()), just
        // copy the underlying KeyVector. Doing so is **MUCH** faster.
        this.kv = newKV.copy();
      }
      else {
        for (int i = 0; i < newKV.size(); i++) {
          putx(newKV.getKey(i), newKV.get(i));
        }
      }
    }
    else if (newTbl != null) {
      if (newTbl.size() == 0) {
        // do nothing
      }
      else if (newTbl.fb != null) {
        if ((getModeInt() == FB) && (fb == null)) {
          fb = newTbl.fb.clone();
        }
        else {
          fromBytes(newTbl.fb, 0, newTbl.fb.length);
        }
      }
      else {
        Iterator iter = newTbl.iterator();
        for (int i = 0; i < iter.keys.length; i++) {
          putx(iter.keys[i], iter.values[i]);
        }
      }
    }
    else {
      String[] keys = keyable.getKeys();
      for (int i = 0; i < keys.length; i++) {
        putx(keys[i], keyable.getKey(keys[i]));
      }
    }
  }

  /** Sorts the entries in the Table by key. If the table is not in KV mode, the mode will be
      changed to KV mode before sorting. <b>Since NeXtMidas 2.7.0: Setting the table to be in
      {@link #TM} or {@link #STM} will automatically sort the table by key (alphabetically).
      Using {@link #TM} or {@link #STM} can be significantly faster than using this method.</b>
      @param reverse Set this to true to reverse the sort order.
      @since NeXtMidas 2.3.0
   */
  public synchronized void sortByKey(boolean reverse) {
    if (genericMode != KV) { setMode(KV); }
    kv.sortByKey(reverse);
  }

  /** Makes a copy of the Table. This is a shallow copy in that it will copy the
      reference pointers to any nested Tables rather than copy their values to
      new tables; it does not make copies of the values in the table.
      @return A copy of the table.
      @since NeXtMidas 2.3.0
   */
  public synchronized Table copy() {
    return copyOf(this, specificMode);
  }

  /** Makes a copy of the Table to comply with the {@link Tablizable} interface (see
      {@link Tablizable#toTable()} for details). This is the same as calling {@link #copy()}.
      @return A copy of the table.
      @since NeXtMidas 2.7.1
   */
  public final Table toTable () {
    return copy();
  }

  /** Creates a table that is a copy of the given Keyable object. This is the same
      as <code>copyOf(keyable,KV)</code>.
      @param keyable The keyable object to copy.
      @return A copy of the Keyable as a table.
      @since NeXtMidas 2.3.0
   */
  public static Table copyOf(Keyable keyable) {
    return copyOf(keyable,KV);
  }

  /** Merges two tables into a new output table. Input tables <tt>t1</tt> and <tt>t2</tt> are
      not altered. Entries in <tt>t1</tt> are replaced by entries in <tt>t2</tt> if they have the
      same key. If the recurse flag is true, inner tables that exist in both <tt>t1</tt> and
      <tt>t2</tt> are merged as well (this only affects instances of {@link Table}). If recurse
      is false, inner tables in <tt>t2</tt> replace any existing inner tables of the same name in
      <tt>t1</tt> (this matches the behavior in {@link #merge(Keyable)}. <br>
      <br>
      For example:
      <pre>
        t1 = {ZERO=0,ONE=1,TWO=2,TABLE={A="A",B="B",C="C"}}
        t2 = {ONE="ABC",TWO="DEF",THREE="GHI",TABLE={B=2,C=3,D=4}}

        merge(t1,t2,false) -&gt; {ZERO=0,ONE="ABC",TWO="DEF",THREE="GHI",TABLE={B=2,C=3,D=4}}
        merge(t1,t2,true)  -&gt; {ZERO=0,ONE="ABC",TWO="DEF",THREE="GHI",TABLE={A="A",B=2,C=3,D=4}}
      </pre>
      @param t1 Table for which keys are replaced.
      @param t2 Replace or add these keys to t1.
      @param recurse - Recurse into inner tables if true
      @return A new table
  */
  public static Table merge (Table t1, Table t2, boolean recurse) {
    Table tout = t1.copy();
    if (!recurse) { tout.merge(t2); return tout; }

    for (String key : t2.getKeys()) {
      Object val2 = t2.get(key);
      if (val2 instanceof Table) {
        Object val1 = tout.get(key);
        if (val1 instanceof Table) {
          val2 = merge((Table)val1, (Table)val2, true);
        }
      }
      tout.put(key, val2);
    }
    return tout;
  }

  /** Creates a table that is a copy of the given Keyable object. This is a shallow
      copy in that it will copy the reference pointers to any nested Tables rather
      than copy their values to new tables; it does not make copies of the values in
      the table.
      @param keyable The keyable object to copy.
      @param mode    The table mode (any table mode other than FB).
      @return A copy of the Keyable as a table.
      @since NeXtMidas 2.3.0
   */
  public static Table copyOf(Keyable keyable, int mode) {
    Table tbl = new Table(mode);
    tbl.merge(keyable);
    return tbl;
  }

  /** Wraps a KeyVector as a table. This is ONLY used in a few special cases where it
      is required for performance reasons. The KeyVector must not contain any duplicate keys.
   */
  static Table wrap(KeyVector kv) { // Method is package-private to prevent accidental use.
    Table tbl = new Table(KV);
    tbl.kv = kv;
    return tbl;
  }

  /** Converts a config file to a table of tables. Given a config file of the form:
      <pre>
        [SECTION1]
        KEY1=VAL1
        KEY2=VAL2
        ...

        [SECTION2]
        KEY1=VAL1
        KEY2=VAL2
        ...
      </pre>
      This will produce a table of tables in the form:
      <pre>
        {SECTION1={KEY1=VAL1,KEY2=VAL2,...},
         SECTION2={KEY1=VAL1,KEY2=VAL2,...},
         ...}
      </pre>
      Note that a section named "DEFAULT" will be handled special. Also note that
      all keys and section names will be converted to upper case.

      @param allLines The lines from the configuration file
      @param cs       Ignore normal upper-case-only rules for a Table and bring
                      all of the keys in case-sensitive (does not apply to
                      section names)
      @param merge    If this is true, "DEFAULT" values will be automatically merged
                      into each of the sub-tables. If this is false, the "DEFAULT"
                      values will simply remain in the separate "DEFAULT" table.
      @return         The table of tables based on the configuration file
   */
  public static Table fromConfigFile (String[] allLines, boolean cs, boolean merge) {
    Table tbl     = new Table();
    Table defs    = null;
    Table section = null;

    for (int lineNum = 0; lineNum < allLines.length; lineNum++) {
      String line = allLines[lineNum].trim();

      // Strip comments
      if (line.startsWith("#") || line.startsWith(";")) {
        line = "";
      }
      int i = line.indexOf(" ;");
      if (i > 0) {
        line = line.substring(0,i).trim();
      }

      // Check line contents
      if (line.length()==0) {   // EMPTY LINE
        // do nothing (pass)
      }
      else if (line.startsWith("[") && line.endsWith("]")) {     // SECTION START
        String name = line.substring(1,line.length()-1).trim().toUpperCase();
        section = new Table();

        if (name.length()==0) throw new MidasException("Syntax error near line "+lineNum);
        if (name.equals("DEFAULT")) defs = section;
        tbl.put(name, section);
      }
      else if (section == null) {       // MISSING SECTION START
        throw new MidasException("Syntax error near line "+lineNum);

      }
      else {                          // KEY=VAL DEFINITION
        int eq = line.indexOf("=");
        if (eq < 0) {
          // Missing '=' in KEY=VAL
          throw new MidasException("Syntax error near line "+lineNum);
        }
        String key = line.substring(0,eq).trim();
        String val = line.substring(eq+1).trim();

        if (!cs) key = key.toUpperCase();
        section.put(key,val);
      }
    }

    // Merge in defaults
    if (merge && (defs != null)) {
      for (String key : tbl.getKeys()) {
        if (key.equals("DEFAULT")) continue;
        tbl.put(key, merge(defs,tbl.getTable(key),false));
      }
    }

    return tbl;
  }

  /** Converts a table of tables to a config file. This expects tables to be in
      the form:
      <pre>
        {SECTION1={KEY1=VAL1,KEY2=VAL2,...},
         SECTION2={KEY1=VAL1,KEY2=VAL2,...},
         ...}
      </pre>
      And will produce a configuration file in the form:
      <pre>
        [SECTION1]
        KEY1=VAL1
        KEY2=VAL2
        ...

        [SECTION2]
        KEY1=VAL1
        KEY2=VAL2
        ...
      </pre>
      Note that a section named "DEFAULT" will be handled special and will always
      be written to the top of the file/
      @return The lines of the configuration file
   */
  public List<String> toConfigFile () {
    List<String> outFile = new LinkedList<String>();

    // Always put DEFAULT first
    if (containsKey("DEFAULT")) {
      addSectionToConfigFile(outFile, "DEFAULT", getTable("DEFAULT"));
    }

    // Add all other sections
    for (String sectionName : getKeys()) {
      if (sectionName.equals("DEFAULT")) {
        // already done
      }
      else {
        addSectionToConfigFile(outFile, sectionName, getTable(sectionName));
      }
    }
    return outFile;
  }

  /** internal use only. Called from toConfigFile(..) */
  private void addSectionToConfigFile (List<String> outFile, String name, Table tbl) {
    if (tbl==null) throw new MidasException("Invalid input table. Expected "+name+" to be a sub-table.");

    outFile.add("["+name+"]");
    for (String key : tbl.getKeys()) {
      outFile.add(key + "=" + tbl.getS(key));
    }
    outFile.add("");
  }

  /** The {@link Enumeration} implementation (of keys) used by the {@link Table} class. */
  public class Enum implements Enumeration<String> {
    private java.util.Iterator<String> iter;
    private KeyVector kv;
    private String keys[];
    private int item;

    /** Creates a new instance.
        @param table The table to create a new instance for.
        @see Table#keys()
     */
    public Enum (Table table) {
      table.checkModeFB();
      switch (table.genericMode) {
        case KV:  kv   = table.kv;                      break;
        case MAP: iter = table.map.keySet().iterator(); break;
        case KEY: keys = table.kbl.getKeys();           break;
        default:  throw new AssertionError("Unknown table mode "+genericMode);
      }
    }

    public boolean hasMoreElements() {
           if (iter != null) return iter.hasNext();
      else if (kv   != null) return (item < kv.size());
      else                   return (item < keys.length);
    }

    public String nextElement() {
           if (iter != null) return iter.next();
      else if (kv   != null) return kv.getKey(item++);
      else                   return keys[item++];
    }
  } // end class Enum

  /** Creates an iterator that iterates over the elements in this table.
      The return value of the Iterator's <code>{@link Table.Iterator#next()}<code> method is the key.
      To get the value at that key, use the iterator's <code>{@link Table.Iterator#getValue()}</code> method.
      @return An iterator that iterates over the elements in this table.
      @since NeXtMidas 1.7.0
   */
  public synchronized Iterator iterator() { // since 3.1.1: made method synchronized for thread safety (BUG 2500)
    return new Iterator (this);
  }

  /** An iterator used by the {@link Table} class.
      The return value of the {@link #next()} is the key. To get the value at the key, use {@link #getValue()}.
      Since 3.3.1: implements {@link java.util.Iterator} interface.
      @since NeXtMidas 1.7.0
    */
  public class Iterator implements java.util.Iterator<String> {

    public Table table;
    public String[] keys;
    public Object[] values;
    public int offset,size;
    public String key;
    public Object value;

    protected Iterator (Table table, String[] keys, Object[] values) {
      this.table  = table;
      this.keys   = keys;
      this.values = values;
      this.offset = 0;
      this.size   = keys.length;

      if (values == null) {
        this.values = new Object[keys.length];
        for (int i = 0; i < this.size; i++) {
          this.values[i] = table.getKey(keys[i]);
        }
      }
    }

    public Iterator (Table table) {
      this.table = table;
      table.checkModeFB();
      size = table.size();
      keys = new String[size];
      values = new Object[size];
      if (table.genericMode==MAP) {
        java.util.Iterator<Map.Entry<String, Object>> iter = table.map.entrySet().iterator();
        for (int i=0; i<size; i++) {
          Map.Entry<String,Object> e = iter.next();
          keys[i]=e.getKey();
          values[i]=e.getValue();
        }
      }
      else if (table.genericMode==KV) {
        for (int i=0; i<size; i++) {
          keys[i]=table.kv.getKey(i);
          values[i]=table.kv.get(i);
        }
      }
      else if (table.genericMode==KEY) {
        keys = table.kbl.getKeys();
        for (int i = 0; i < keys.length; i++) {
          values[i] = table.kbl.getKey(keys[i]);
        }
      }
      offset=0;
    }

    @Override
    public boolean hasNext() {
      return offset<size;
    }
    public boolean getNext() {
      if (offset>=size) return false;
      next();
      return true;
    }
    /** Gets the next element in the iteration.
        @return key of next element. Use {@link #getValue()} to get value.
     */
    @Override
    public String next() {
      key = keys[offset];
      value = values[offset];
      offset++;
      return key;
    }
    public String getKey() {
      return key;
    }
    public Object getValue() {
      return value;
    }
    @Override
    public void remove() {
      table.remove(key);
    }
  }

  public interface Listener {
    /** do something before the put, return true if done */
    public boolean putAction (String key, Object value);
    /** do something after the get, return true if done */
    public boolean getAction (String key, Object value);
  }

  /** Writes the current contents of this table to the given output stream.
      Only the data of the table is written to the stream.  This method is
      required by the Externalizable interface and is used when doing
      serialization of the Table.

      @param out The stream to write the instance to.
      @throws IOException If an exception occurs while writing.
      @see #readExternal(ObjectInput)
   */
  public void writeExternal (ObjectOutput out) throws IOException {
    out.writeObject( toString() );
  }

  /** Restore the state of this instance to the values that are contained in
      the input stream.  The input stream is required to have been written by
      the writeExternal method.  The data of the table is reset to the values
      in the stream and the other properties are initialized to their default
      state.

      @param in The stream to read the data from.

      @throws IOException            If an exception occurs while reading.
      @throws ClassCastException     If the input stream contains an object not written by
                                     {@link #writeExternal(ObjectOutput)}.
      @throws ClassNotFoundException If the input stream contains an object not written by
                                     {@link #writeExternal(ObjectOutput)}.

      @see #writeExternal(ObjectOutput)
   */
  public void readExternal (ObjectInput in) throws IOException, ClassNotFoundException,
                                                   ClassCastException {
    String data = (String)in.readObject();
    fromBytes( data.getBytes() );
  }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // METHODS ADDED TO NeXtMidas 2.1.0 TO IMPLEMENT THE java.util.Map INTERFACE
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** This is identical to <tt>containsKey(key.toString())</tt>.
      @param key The key to look for (should not be null).
      @throws NullPointerException if the specified key is null.
      @see #containsKey(String)
      @since NeXtMidas 2.1.0
   */
  public boolean containsKey(Object key) {
    return containsKey(key.toString());
  }

  /** @since NeXtMidas 2.1.0 */
  public Set<Map.Entry<String,Object>> entrySet() {
    final String[]    keys    = getKeys();
    final Map.Entry<String,Object>[] entries = new Entry[keys.length];

    for (int i = 0; i < keys.length; i++) {
      entries[i] = new Entry(this, keys[i]);
    }

    return new SimpleSet<Map.Entry<String, Object>>(this) {
      public int size()               { return entries.length;                    }
      public Map.Entry<String, Object> getItem(int i) { return entries[i];                        }
      public Object removeItem(int i) { return table.remove(entries[i].getKey()); }
    };
  }

  /** @since NeXtMidas 2.1.0 */
  public Set<String> keySet() {
    final String[] keys = getKeys();

    return new SimpleSet<String>(this) {
      public int size()               { return keys.length;                   }
      public String getItem(int i)    { return keys[i];                       }
      public Object removeItem(int i) { return table.remove((Object)keys[i]); }
    };
  }

  /** Copies all of the mappings from the specified map to this map (see {@link Map#putAll(Map)}).
      This method is slower than using {@link #merge(Keyable)} when given a {@link KeyVector} or a
      {@link Table} in {@link #KV} mode.
      @param map Mappings to be stored in this map.
      @since NeXtMidas 2.1.0
   */
  public void putAll(Map map) {
    if (map==null) return; // if map is null, don't need to copy to fix NPE in Bug 2118
    java.util.Iterator<?> iter = map.entrySet().iterator();

    while (iter.hasNext()) {
      Map.Entry<?,?> entry = (Map.Entry<?,?>)iter.next();
      putx(entry.getKey().toString(), entry.getValue());
    }
  }

  /** @since NeXtMidas 2.1.0 */
  public Collection<Object> values() {
    final String[] keys = getKeys();

    return new SimpleSet<Object>(this) {
      public int size()               { return keys.length;                   }
      public Object getItem(int i)    { return table.get((Object)keys[i]);    }
      public Object removeItem(int i) { return table.remove((Object)keys[i]); }
    };
  }

  /** @since NeXtMidas 2.1.0 */
  public boolean equals(Object obj) {
    if (obj == this)           return true;
    if (!(obj instanceof Map)) return false; // BUG 1358 - prevent NullPointerException & ClassCastException

    /* Copied from the Map Javadocs. */
    Map<?,?> t1 = this;
    Map<?,?> t2 = (Map<?,?>)obj;

    return t1.entrySet().equals(t2.entrySet());
  }

  /** @since NeXtMidas 2.1.0 */
  public int hashCode() {
    /* This is as defined in the Map Javadocs. */
    final java.util.Iterator<Map.Entry<String, Object>> entries = entrySet().iterator();

    int hashcode = 0;
    while (entries.hasNext()) { hashcode += entries.next().hashCode(); }

    return hashcode;
  }

  /** Put a generic entry into the table under the given key. The insertion is identical to
      <tt>put(key.toString(),value)</tt> since all keys in a {@link Table} must be strings. <br>
      <br>
      Since NeXtMidas 2.1.0 this returns the previous value associated with the key being set in
      order to match the {@link Map} interface. As such this method is SLOWER (by a factor of
      about 2) than {@link #put(String,Object)} method.
      @param key   The key to insert (not <tt>null</tt>). See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @param value The value to insert (any <tt>null</tt> values will be changed to <tt>"NULL"</tt>
                   on insertion unless {@link #EXCEPTION_ON_NULL} is set).
      @return The value removed.
      @throws NullPointerException if <tt>key</tt> is <tt>null</tt> or if <tt>value</tt> is
                                   <tt>null</tt> and {@link #EXCEPTION_ON_NULL} is set.
   */
  public final Object put (Object key, Object value) throws NullPointerException {
    if (key==null) throw new NullPointerException ("Table Key=null not allowed");
    final Object oldValue = get(key);
    put(key.toString(),value);
    return oldValue;
  }

  /** Get a generic entry from the table using the specified key.
      This is the same as using <code>get(key.toString())</code>
      @param key The name of the entry to get. See note on the <a href="#dotsyntax">dot syntax</a> for this parameter.
      @return The value associated with the key (may be <code>null</code>) or <code>null</code> if
              the specified key is not part of the table.
      @see #put(String, Object)
   */
  public final Object get (Object key) throws NullPointerException {
    if (key==null) return null;
    else return get (key.toString());
  }

  /** Remove the entry indexed under the specified key from the table. Since NeXtMidas 2.1.0 this
      returns an {@link Object} which is the previous value associated with the key being set.
      @param key The name of the key to remove. See note on the <a href="#dotsyntax">dot syntax</a> for key parameter.
      @return The value removed.
   */
  public final Object remove (Object key) throws NullPointerException {
    if (key == null) return null;
    //final Object oldValue = get(key);  // prior to 3.1.0: changed when other remove method started returning an Object (Bug 2455)
    //return oldValue;                   // prior to 3.1.0
    return remove(key.toString());       // since 3.1.0 for BUG 2455
  }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // NESTED CLASSES USED TO IMPLEMENT THE java.util.Map INTERFACE
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** A simple implementation of {@link Set} used in the implementation of {@link Map}. */
  private abstract class SimpleSet<E> extends AbstractSet<E> {
    protected final Table table; // The table

    /** Creates a new instance.
        @param table The table.
     */
    private SimpleSet(Table table) {
      this.table   = table;
    }

    /** Gets the i'th element.
        @param i The index.
        @return The item at the given index.
     */
    public abstract E getItem (int i);

    /** Removes the i'th element.
        @param i The index.
        @return the removed item at the i'th element, null if none removed.
     */
    public abstract Object removeItem (int i);

    public java.util.Iterator<E> iterator() {
      return new java.util.Iterator<E>() {
        private int i = 0;
        public boolean hasNext() { return (i < size()); }
        public E       next()    { i++; return getItem(i-1); }
        public void    remove()  { removeItem(i-1); }
      };
    }
  }

  /** A simple implementation of {@link java.util.Map.Entry}. */
  public static final class Entry implements Map.Entry<String,Object> {
    /** The table backing this Entry */ private final Table  table;
    /** The entry key                */ private final String key;

    /** Creates a new instance.
        @param key The key.
     */
    private Entry (Table table, String key) {
      this.table = table;
      this.key   = key;
    }

    public boolean equals (Object obj) {
      if (obj == this) return true;                  // since 3.1.1
      if (!(obj instanceof Map.Entry)) return false; // since 3.1.1

      /* This is copied directly from the Map.Entry Javadocs */
      Map.Entry<?,?> e1 = this;
      Map.Entry<?,?> e2 = (Map.Entry<?,?>)obj;

      return (e1.getKey()==null   ? e2.getKey()==null   : e1.getKey().equals(e2.getKey()))  &&
             (e1.getValue()==null ? e2.getValue()==null : e1.getValue().equals(e2.getValue()));
    }

    public int hashCode() {
      /* This is copied directly from the Map.Entry Javadocs */
      return (this.getKey()==null   ? 0 : this.getKey().hashCode()) ^
             (this.getValue()==null ? 0 : this.getValue().hashCode());
    }

    public String getKey()   { return key; }
    public Object getValue() { return table.get(key); }
    public Object setValue(Object value) { return table.put((Object)key, value); }
  } // end inner class Entry

}
