package nxm.sys.lib;

import nxm.sys.inc.*;

import java.util.Vector;
import java.util.zip.*;

/** A class to handle Tape Archive (TAR) files. <br>
    <br>
    Basic TAR header (rarely seen):
    <pre>
      OFF SIZE  NAME        DESCRIPTION
      ==============================================
        0 100   name        name of file
      100 8     mode        file mode
      108 8     uid         owner user ID
      116 8     gid         owner group ID
      124 12    size        length of file in bytes
      136 12    mtime       modify time of file
      148 8     chksum      checksum for header
      156 1     typeflag    type of file
      157 100   linkname    name of linked file
      257
    </pre>
    USTAR Header:
    <pre>
      OFF SIZE  NAME        DESCRIPTION
      ==============================================
        0 100   name        name of file              --+
      100 8     mode        file mode                   |
      108 8     uid         owner user ID               |
      116 8     gid         owner group ID              | Same as a
      124 12    size        length of file in bytes     | Basic TAR
      136 12    mtime       modify time of file         | Header
      148 8     chksum      checksum for header         |
      156 1     typeflag    type of file                |
      157 100   linkname    name of linked file       --+
      257 6     magic       USTAR indicator
      263 2     version     USTAR version
      265 32    uname       owner user name
      297 32    gname       owner group name
      329 8     devmajor    device major number
      337 8     devminor    device minor number
      345 155   prefix      prefix for file name
      500
    </pre>
    @author  J. Czechowski IV
    @version $Id: TarFile.java,v 1.6 2012/09/14 13:32:42 jwb Exp $
    @since   NeXtMidas 2.5.0
 */
public class TarFile extends BaseFile implements ListFile {
  /** Length of a standard TAR file header. <tt>TAR_HEAD_LEN={@value}</tt> */
  private static final int TAR_HEAD_LEN   = 256;

  /** Length of a USTAR file header. <tt>USTAR_HEAD_LEN={@value}</tt> <br>
      Note that only 500 bytes are actually used, however headers are 256-byte aligned.
   */
  private static final int USTAR_HEAD_LEN = 512;

  private static final int LINK_NORMAL_ALT        =  0;   // Link Indicator: Normal (same as LINK_NORMAL)
  private static final int LINK_NORMAL            = '0';  // Link Indicator: Normal
  private static final int LINK_HARD_LINK         = '1';  // Link Indicator: Hard Link (link to file in TAR)
  private static final int LINK_SYMBOLIC_LINK     = '2';  // Link Indicator: Symbolic Link
  private static final int LINK_CHARACTER_SPECIAL = '3';  // Link Indicator: Char Special
  private static final int LINK_BLOCK_SPECIAL     = '4';  // Link Indicator: Block Special
  private static final int LINK_DIRECTORY         = '5';  // Link Indicator: Directory
  private static final int LINK_FIFO              = '6';  // Link Indicator: FIFO
  private static final int LINK_CONTIGUIOUS_FILE  = '7';  // Link Indicator: Contig File (unofficial)

  // These are copied directly from the man page for chmod(2)
  private static final int S_IRWXU = 0000700;    /* RWX mask for owner */
  private static final int S_IRUSR = 0000400;    /* R for owner */
  private static final int S_IWUSR = 0000200;    /* W for owner */
  private static final int S_IXUSR = 0000100;    /* X for owner */

  private static final int S_IRWXG = 0000070;    /* RWX mask for group */
  private static final int S_IRGRP = 0000040;    /* R for group */
  private static final int S_IWGRP = 0000020;    /* W for group */
  private static final int S_IXGRP = 0000010;    /* X for group */

  private static final int S_IRWXO = 0000007;    /* RWX mask for other */
  private static final int S_IROTH = 0000004;    /* R for other */
  private static final int S_IWOTH = 0000002;    /* W for other */
  private static final int S_IXOTH = 0000001;    /* X for other */

  private static final int S_ISUID = 0004000;    /* set user id on execution */
  private static final int S_ISGID = 0002000;    /* set group id on execution */
  private static final int S_ISVTX = 0001000;    /* save swapped text even after use */


  private TarEntry entries[] = null;

  public TarFile () { }

  public TarFile (Object ref, Object filename) {
    init(ref,filename);
  }

  public String getMimeType () {
    return "application/x-tar";
  }

  public boolean open ()  {
    if (super.open()) {
      Vector<TarEntry> vect = new Vector<TarEntry>();
      TarEntry entry = null;
      while ((entry = nextEntry()) != null) {
        vect.add(entry);
      }
      entries = vect.toArray(new TarEntry[0]);
    }
    return isOpen;
  }

  /** Gets a named TAR entry. Note that multiple entries in the tar are allowed to have the same
      name. When multiple entries have the same name, the LAST one is used (earlier entries are
      assumed to be out-of-date).
   */
  TarEntry getEntry (String name) { // package-private only used by TarResource
    TarEntry entry = null;
    for (int i = entries.length-1; (entry == null) && (i >= 0); i--) {
      if (entries[i].getName().equals(name)) {
        entry = entries[i];
      }
    }
    return entry;
  }

  /** Gets all the TAR entries. 
      @since NeXtMidas 3.3.0 
   */
  public Object[] getEntries () { 
    return entries;  // added for Bug 2574 to allow for getting reference to entries
  }

  /** Reads the next entry from the file. */
  private TarEntry nextEntry () {
    TarEntry entry   = null;
    byte[]   buf     = new byte[USTAR_HEAD_LEN];
    long     offset  = (long)seek();
    int      numRead = read(buf, 0, buf.length);

    if (numRead > 0) {
      entry = new TarEntry(buf,offset);
      if (entry.isBlank()) {
        // Blank entry, assume end of file
        entry = null;
      } else {
        // Seek to start of next entry, note that entries are aligned to 512-byte blocks.
        long seek = (long)seek() + entry.getEntrySize() - buf.length;
        seek(seek);
      }
    }
    return entry;
  }

  public String toString() {
    String str = "TarFile ";
    if (io!=null) str += io.getTypeString();
    str += " Resource";
    return str;
  }

  public String listHeader() {
    String list =
      "TarFile   :  "+getURL()+"\n"
    + "Size      :  "+(long)getSize()+" bytes\n"
    + "Entries   :  "+entries.length+"\n";
    return list;
  }

  public String listElements (double start, int elements, String format, int flags) {
    int i = (int)start;
    return (i < entries.length)? (entries[i].toString(true)+'\n'): null;
  }

  public int listElementsPerLine (int lineWidth, String form, int flags) {
    return 1;
  }

  /** Gets the number of entries in the file. */
  public int getNumEntries() { return entries.length; }



  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Methods to implement ListFile
  //////////////////////////////////////////////////////////////////////////////////////////////////
  private void notEditable () { throw new UnsupportedOperationException("TarFile is not editable."); }
  public double  getNumberOfRows ()                        { return getNumEntries(); }
  public boolean getProtected ()                           { return true;   } // not editable at this time
  public int     setData    (double d, Table t)            { notEditable(); return 0; }
  public void    setData    (double d, String s, Object o) { notEditable(); }
  public void    insertData (double d, Object o)           { notEditable(); }
  public void    removeData (double d)                     { notEditable(); }
  public void    removeData (double d, long c)             { notEditable(); }
  public void    setRecordDefs (Table t)                   { notEditable(); }

  public Table getDataTable(double offset) {
    if (offset < 0) return null;
    if (offset >= entries.length) return null;
    return entries[(int)offset].toTable();
  }

  public int getRecordDefCount () {
    return ((entries.length == 0) || (entries[0].isUstar()))? 13 : 11;
  }

  public Table getRecordDef (int i) {
    Table t = new Table();
    switch (i) {
      case  0: t.put("NAME", "NAME");                                        break;
      case  1: t.put("NAME", "MODE");                t.put("FORMAT", "SL");  break;
      case  2: t.put("NAME", "MODE_STRING");                                 break;
      case  3: t.put("NAME", "OWNER");               t.put("FORMAT", "SL");  break;
      case  4: t.put("NAME", "GROUP");               t.put("FORMAT", "SL");  break;
      case  5: t.put("NAME", "SIZE");                t.put("FORMAT", "SX");  break;
      case  6: t.put("NAME", "TIME");                t.put("FORMAT", "SD");  break;
      case  7: t.put("NAME", "CHECKSUM");                                    break;
      case  8: t.put("NAME", "LINK_INDICATOR");                              break;
      case  9: t.put("NAME", "LINK_INDICATOR_NAME");                         break;
      case 10: t.put("NAME", "LINKED_FILE");                                 break;
      case 11: t.put("NAME", "OWNER_NAME");                                  break; // USTAR ONLY
      case 12: t.put("NAME", "GROUP_NAME");                                  break; // USTAR ONLY
      // DEVICE
      default: throw new ArrayIndexOutOfBoundsException("Invalid record definition i="+i);
    }
    return t;
  }

  public Table getRecordDefs() {
    Table t = new Table();
    for (int i = 0; i < getRecordDefCount(); i++) {
      t.put(Integer.toString(i), getRecordDef(i));
    }
    return t;
  }

static int xxx = 0;

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Inner class defining an entry in the TAR file
  //  - Made public in 3.3.0 for Bug 2574
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** An entry in a TarFile. */
  public static class TarEntry { 
    final byte[] buf;
    final long   offset;

    public TarEntry (byte[] buf, long off) {
      this.buf      = buf;
      this.offset   = align((isUstar())? off+USTAR_HEAD_LEN : off+TAR_HEAD_LEN);
    }

    /** Aligns an offset value to the next 512-byte blocks. */
    private static long align (long seek) {
      long mod  = seek % 512;
      if (mod != 0) seek = seek + 512 - mod;
      return seek;
    }

    long getOffset() { // package-private, used only by TarFile and TarResource
      return offset;
    }

    long getEntrySize () {
      long size = getSize();

      if (isUstar()) size += USTAR_HEAD_LEN;
      else           size += TAR_HEAD_LEN;

      return align(size);
    }

    /** Indicates if it is a USTAR entry. */
    private boolean isUstar() {
      return (buf[257] == 'u')
          && (buf[258] == 's')
          && (buf[259] == 't')
          && (buf[260] == 'a')
          && (buf[261] == 'r')
          && (buf[262] == ' ')
          && (buf[263] == ' ')
          && (buf[264] == '\0');
    }

    /** Indicates if it is a blank entry. */
    boolean isBlank () {
      boolean blank = true;
      for (int i = 0; blank && (i < buf.length); i++) {
        blank = (buf[i] == 0);
      }
      return blank;
    }

    public String toString () {
      return toString(false);
    }

    String toString (boolean pad) {
      // This syntax matches that used by 'tar tvf <file>' except that pad causes the
      // file sizes to be padded out (like 'ls -al') so things line up nicely.
      StringBuffer buf  = new StringBuffer(32);
      String       size = Long.toString(getSize());

      if (pad) size = StringUtil.padLeft(size, 9); // <-- JC4: 9 seems to match 'ls -al' on my box

      buf.append(getModeString()).append(' ');
      buf.append(getOwnerName()).append('/');
      buf.append(getGroupName()).append(' ');
      buf.append(size).append(' ');
      buf.append(getTime().toString(-1)).append(' '); // <-- Time in tar file has no fractional seconds
      buf.append(getName());

      if (isLink()) {
        buf.append(" -> ").append(getLinkedFile());
      }
      return buf.toString();
    }

    /** Dumps info to a table */
    public Table toTable () {
      String linked = getLinkedFile();
      if (linked == null) linked = "";

      Table tbl = new Table();
      tbl.put("NAME",                getName());
      tbl.put("MODE",                getMode());
      tbl.put("MODE_STRING",         getModeString());
      tbl.put("OWNER",               getOwner());
      tbl.put("GROUP",               getGroup());
      tbl.put("TIME",                getTime().getSec()); // Uses FORM=SD
      tbl.put("CHECKSUM",            getCheckSum());
      tbl.put("LINK_INDICATOR",      getLinkIndicator());
      tbl.put("LINK_INDICATOR_NAME", getLinkIndicatorName());
      tbl.put("LINKED_FILE",         linked);

      if (isUstar()) {
        tbl.put("OWNER_NAME",  getOwnerName());
        tbl.put("GROUP_NAME",  getGroupName());
        //tbl.put("DEVICE",      device);
      }

      return tbl;
    }

    /** Dumps info to a string */
    String dumpToString (String sp) {
      String owner  = getOwnerName();
      String group  = getGroupName();
      String device = null;
      String linked = getLinkedFile();

      if (owner  == null) owner  = Integer.toString(getOwner());
      if (group  == null) group  = Integer.toString(getGroup());
      if (device == null) device = "n/a";
      if (linked == null) linked = "n/a";

      return    sp + "Name           : " + getName() + "\n"
              + sp + "Mode           : " + getModeString() + " (mode=" + getMode() + ")\n"
              + sp + "Owner          : " + owner + " (ID=" + getOwner() + ")\n"
              + sp + "Group          : " + group + " (ID=" + getGroup() + ")\n"
              + sp + "File Size      : " + getSize() + " bytes\n"
              + sp + "Last Mod Time  : " + getTime() + "\n"
              + sp + "Check Sum      : " + getCheckSum() + "\n"
              + sp + "Link Indicator : " + getLinkIndicatorName() + "\n"
              + sp + "Linked File    : " + linked + "\n"
              //+ sp + "Device         : " + device + "\n"
              ;
    }

    /** Gets name of archived file. */
    public String getName () {
      String name;
      if (isUstar()) {
        name = unpackS(345, 155) + unpackS(0, 100);
      } else {
        name = unpackS(0, 100);
      }
      return name;
    }

    /** Sets name of archived file. */
    public void setName (String name) {
      int length = name.length();

      if ((length > 100) && isUstar()) {
        packS(  0, 100, name.substring(length-100));
        packS(345, 155, name.substring(0,length-100));
      } else {
        packS(  0, 100, name);
      }
    }

    /** Gets the file mode (file permissions) as a string. */
    public String getModeString () {
      StringBuffer buf = new StringBuffer(10);
      int link = getLinkIndicator();
      int mode = getMode();

      switch (link) {
        case LINK_NORMAL_ALT:        buf.append('-'); break;
        case LINK_NORMAL:            buf.append('-'); break;
        case LINK_HARD_LINK:         buf.append('l'); break; // <-- matches behavior of tar
        case LINK_SYMBOLIC_LINK:     buf.append('l'); break;
        case LINK_CHARACTER_SPECIAL: buf.append('c'); break;
        case LINK_BLOCK_SPECIAL:     buf.append('b'); break;
        case LINK_DIRECTORY:         buf.append('d'); break;
        case LINK_FIFO:              buf.append('p'); break;
        default:                     buf.append('?'); break; // <-- LINK_CONTIGUIOUS_FILE, other?
      }

      if ((mode & S_IRUSR) != 0) buf.append('r');  else  buf.append('-');
      if ((mode & S_IWUSR) != 0) buf.append('w');  else  buf.append('-');
      if ((mode & S_IXUSR) != 0) buf.append('x');  else  buf.append('-');

      if ((mode & S_IRGRP) != 0) buf.append('r');  else  buf.append('-');
      if ((mode & S_IWGRP) != 0) buf.append('w');  else  buf.append('-');
      if ((mode & S_IXGRP) != 0) buf.append('x');  else  buf.append('-');

      if ((mode & S_IROTH) != 0) buf.append('r');  else  buf.append('-');
      if ((mode & S_IWOTH) != 0) buf.append('w');  else  buf.append('-');
      if ((mode & S_IXOTH) != 0) buf.append('x');  else  buf.append('-');

      if ((mode & S_ISUID) != 0) {
        if ((mode & S_IXUSR) != 0) buf.setCharAt(3, 's');  else  buf.setCharAt(3, 'S');
      }
      if ((mode & S_ISGID) != 0) {
        if ((mode & S_IXGRP) != 0) buf.setCharAt(6, 's');  else  buf.setCharAt(6, 'S');
      }
      if ((mode & S_ISVTX) != 0) {
        if ((mode & S_IXOTH) != 0) buf.setCharAt(6, 't');  else  buf.setCharAt(6, 'T');
      }

      return (link == LINK_CONTIGUIOUS_FILE)? null : buf.toString();
    }

    /** Gets the file mode (file permissions). */
    public int getMode () {
      return unpackL(100, 8);
    }

    /** Sets the file mode (file permissions). */
    public void setMode (int val) {
      packL(val, 100, 8);
    }

    /** Gets the numeric UID of the file. */
    public int getOwner () {
      return unpackL(108, 8);
    }

    /** Sets the numeric UID of the file. */
    public void setOwner (int val) {
      packL(108, 8, val);
    }

    /** Gets the name of the file owner (USTAR ONLY). */
    public String getOwnerName () {
      return (isUstar())? unpackS(265, 32): Integer.toString(getOwner());
    }

    /** Sets the name of the file owner (USTAR ONLY). */
    public void setOwnerName (String val) {
      packS(265, 32, val);
    }

    /** Gets the numeric GID of the file. */
    public int getGroup () {
      return unpackL(116, 8);
    }

    /** Sets the numeric GID of the file. */
    public void setGroup (int val) {
      packL(116, 8, val);
    }

    /** Gets the group of the file owner (USTAR ONLY). */
    public String getGroupName () {
      return (isUstar())? unpackS(297, 32): Integer.toString(getGroup());
    }

    /** Sets the group of the file owner (USTAR ONLY). */
    public void setGroupName (String val) {
      packS(297, 32, val);
    }

    /** Gets the size of the file. */
    public long getSize () {
      return unpackL(124, 12);
    }

    /** Sets the size of the file. */
    public void setSize (int val) {
      packL(val, 124, 12);
    }

    /** Returns creation time of file in seconds since 1970   
        @since NeXtMidas 3.3.0 
     */
    public long getTime1970 () {
      return unpackX(136, 12);
    }

    /** Gets the file creation time. */
    public Time getTime () {
      double j1970 = ((double)unpackX(136, 12));
      return new Time(j1970 + Time.J1970TOJ1950);
    }

    /** Sets the file creation time. */
    public void setTime (Time t) {
      long j1950 = (long)(t.getWSec() - Time.J1970TOJ1950);
      packX(136, 12, j1950);
    }

    /** Gets the header checksum. */
    public String getCheckSum () {
      return unpackS(148, 8);
    }

    /** Sets the header checksum. */
    public void setCheckSum (int val) {
      packL(148, 8, val);
    }

    /** Gets the link indicator char. */
    public char getLinkIndicator () {
      return (char)buf[156];
    }

    /** Sets the link indicator char. */
    public void setLinkIndicator (char val) {
      buf[156] = (byte)val;
    }

    /** Gets the link indicator name. */
    public String getLinkIndicatorName () {
      char c = getLinkIndicator();
      switch (c) {
        case LINK_NORMAL_ALT:         return "Normal";
        case LINK_NORMAL:             return "Normal";
        case LINK_HARD_LINK:          return "HardLink";
        case LINK_SYMBOLIC_LINK:      return "SymbolicLink";
        case LINK_CHARACTER_SPECIAL:  return "CharacterSpecial";
        case LINK_BLOCK_SPECIAL:      return "BlockSpecial";
        case LINK_DIRECTORY:          return "Directory";
        case LINK_FIFO:               return "FIFO";
        case LINK_CONTIGUIOUS_FILE:   return "ContiguousFile";
        default: throw new MidasException("Unknown link type '"+c+"'.");
      }
    }

    /** Indicates if the file is a link (hard link or symbolic link). */
    public boolean isLink () {
      char c = getLinkIndicator();
      return (c == LINK_HARD_LINK) || (c == LINK_SYMBOLIC_LINK);
    }

    /** Gets the name of the linked file. */
    public String getLinkedFile () {
      String str = null;
      char c = getLinkIndicator();

      if ((c == LINK_HARD_LINK) || (c == LINK_SYMBOLIC_LINK)) {
        str = unpackS(157, 100);
      }
      return str;
    }

    /** Sets the name of the linked file. */
    public void setLinkedFile (String val) {
      packS(157, 100, val);
    }

    /** Unpacks a SL from the byte array. This handles the OCTAL to int conversion. */
    private int unpackL (int off, int len) {
      return (int)unpackX(off, len);
    }

    /** Packs a SL in the byte array. This handles the int to OCTAL conversion. */
    private void packL (int off, int len, int val) {
      packX(off, len, (long)val);
    }

    /** Unpacks a SX from the byte array. This handles the OCTAL to long conversion. */
    private long unpackX (int off, int len) {
      String str = unpackS(off, len);
      try {
        return (str.length() == 0)? 0L : Long.parseLong(str, 8);
      } catch (NumberFormatException e) {
        throw new MidasException("Error converting octal number '"+str+"' to decimal.");
      }
    }

    /** Packs a SX in the byte array. This handles the long to OCTAL conversion. */
    private void packX (int off, int len, long val) {
      String str = "";
      while (val > 0L) {
        str = "" + (val % 0x7) + str;
        val = val / 8;
      }
      packS(off, len, str);
    }

    /** Unpacks a string from the byte array. This will trim off any padding characters. */
    private String unpackS (int off, int len) {
      if ((off > 256) && !isUstar()) return null;

      while ((len > 0) && (buf[off+len-1] == 0)) len--;

      char[] chars = new char[len];
      for (int i = 0; i < len; i++) {
        chars[i] = (char)buf[i+off];
      }
      return new String(chars);
    }

    /** Packs a string from the byte array. This will add any required padding characters. */
    private void packS (int off, int len, String str) {
      if ((off > 256) && !isUstar()) {
        throw new IllegalArgumentException("TarFile: Can not set string value "
                                          + "'"+str+"' at off="+off+" not tar file "
                                          + "is not in in USTAR format.");
      }
      if (str.length() > len) {
        throw new IllegalArgumentException("TarFile: String value '"+str+"' "
                                         + "exceeds maximum length of "+len+".");
      }
      for (int i = 0; i < len; i++) {
        buf[off++] = (byte)((i < str.length())? str.charAt(i) : '\0');
      }
    }
  }
}
