package nxm.sys.lib;

import nxm.sys.inc.*;
import static nxm.sys.inc.Units.CARTESIAN;
import static nxm.sys.inc.Units.SPHERICAL;
import static nxm.sys.inc.Units.CYLINDRIC;
import static nxm.sys.inc.Units.ELLIPSOID;
import static nxm.sys.inc.Units.GEODETIC;

/** A class to handle reference frame and coordinate system transformations. <br>
    <br>
    Reference Frames:
    <pre>
      {@link #NONE}     - <i>No reference frame specified.</i>
      {@link #ECR}      - <b>Earth-centered rotational (ECR)</b>, which has x and y in the equatorial plane at
                 0 and 90 degrees east longitude, respectively, and z north along the
                 earth's rotational axis. This is the same as the Earth-Centered, Earth-Fixed (ECEF or ECF)
                 Cartesian coordinate system.
      {@link #ECI}      - <b>Earth-centered inertial (ECI)</b>, which has x toward the vernal equinox, y 90
                 degrees east of x, and z north along the earth's rotational axis.
      {@link #TOP}      - <b>Modified topocentric</b>, with x due east in the local horizon plane, y
                 north in the horizon plane, and z perpendicular to the horizon plane
                 in the upward direction.
      {@link #TOPO}     - <b>Topocentric</b>, which has x due south in the local horizon plane, y
                 east in the horizon plane, and z perpendicular to the horizon plane
                 in the upward direction.
      TOPOCENT - <i>Same as TOPO</i>
    </pre>
    Coordinate Systems (SI units):
    <pre>
      {@link Units#CARTESIAN}  - { x, y, z } in meters
      {@link Units#SPHERICAL}  - { r, a, b }
      {@link Units#CYLINDRIC}  - <i>Not Supported in Transform</i>
      {@link Units#ELLIPSOID}  - <i>Not Supported in Transform</i>
      {@link Units#GEODETIC}   - { alt, lat, lon } alt in meters above sea level, lat in deg, lon in deg

      ALL (AltLatLon) - GEODETIC parameters in { alt, lat, lon } order (used in files/Transform)
      LLA (LonLatAlt) - GEODETIC parameters in { lon, lat, alt } order (used in PLOT)
    </pre>

    @author Jeff Schoen / Neon Ngo
    @version $Id: Transform.java,v 1.35 2012/12/07 23:53:31 jxc Exp $
 */
public class Transform extends Object implements Constants {
  private static final double KNOTS2MS   = NM2METERS / 3600.0; // knots to m/s
  private static final double GFORCE2MS2 = 9.80665;            // 1g to m/s^2
  
  /** List of Reference Frame names. */
  public static final String referenceFrameList = "None,ECR,ECI,TOP,TOPO";

  /** Reference Frame NONE (see {@link Transform} for details. */
  public static final int NONE=1;

  /** Reference Frame ECR (see {@link Transform} for details. */
  public static final int ECR=2;

  /** Reference Frame ECI (see {@link Transform} for details. */
  public static final int ECI=3;

  /** Reference Frame TOP (see {@link Transform} for details. */
  public static final int TOP=4;

  /** Reference Frame TOPO or TOPOCENT (see {@link Transform} for details. */
  public static final int TOPO=5;

  /** List of Data Modes. */
  public static final String dataModeList = "Pos,PosVel,PosVelAcc";

  /** Data Mode: Position (POS) only. */
  public static final int POS=1;

  /** Data Mode: Position (POS) and Velocity (VEL). */
  public static final int POSVEL=2;

  /** Data Mode: Position (POS), Velocity (VEL) and Acceleration (ACC). */
  public static final int POSVELACC=3;

  // Internal use only: Common ECR transforms since they are constant (do not let users access!).
  private static final Transform Transform_ECR_CARTESIAN = new Transform(ECR, Units.CARTESIAN);
  private static final Transform Transform_ECR_GEODETIC  = new Transform(ECR, Units.GEODETIC);

  private       int    referenceFrame   = NONE;
  private       int    coordinateSystem = Units.CARTESIAN;
  private       int    unitsPos         = Units.DISTANCE_M;
  private       int    unitsVel         = Units.SPEED_M_S;
  private       int    unitsAcc         = Units.ACCELERATION_M_S2;
  private final Time   epoch            = new Time();
  private       double lat,lon,alt;
  private       double azim,elev,roll;
  
  /** Internal use only.
      @param refFrame ECR, ECI or NONE. TOP or TOPO can be used, but only if
                      setting the lon,lat,alt (and azim,elev,roll).
      @param coordSys Coordinate system
   */
  private Transform (int refFrame, int coordSys) {
    referenceFrame   = refFrame;
    coordinateSystem = coordSys;
  }
  
  /** Constructs a new instance with reference frame {@link #NONE} and a {@link Units#SCALAR}
      coordinate system.
   */
  public Transform () {
    this(NONE, Units.SCALAR); // <-- Matches NeXtMidas 3.2.0 and earlier.
  }
  
  /** Constructs a new instance, setting the referenceFrame/coordinateSystem based a Type 5000
      file (see {@link #fromDataFile(DataFile)}).
      @param df The file.
   */
  public Transform (DataFile df) {
    this();
    fromDataFile(df);
  }
  
  @Override
  public boolean equals (Object obj) { // Since NeXtMidas 3.3.0
    if (!(obj instanceof Transform)) return false;
    Transform that = (Transform)obj;
    
    return (this.referenceFrame   == that.referenceFrame)
        && (this.coordinateSystem == that.coordinateSystem)
        && (this.unitsPos         == that.unitsPos)
        && (this.unitsVel         == that.unitsVel)
        && (this.unitsAcc         == that.unitsAcc)
        && (this.epoch.equals(that.epoch))
        && (this.lat              == that.lat)
        && (this.lon              == that.lon)
        && (this.alt              == that.alt)
        && (this.azim             == that.azim)
        && (this.elev             == that.elev)
        && (this.roll             == that.roll);
  }
  
  @Override
  public int hashCode () { // Since NeXtMidas 3.3.0
    return (this.referenceFrame        )
         ^ (this.coordinateSystem <<  4)
         ^ (this.unitsPos         <<  8)
         ^ (this.unitsVel         << 16)
         ^ (this.unitsAcc         << 24)
         ^ (this.epoch.hashCode()      )
         ^ Double.valueOf(this.lat ).hashCode()
         ^ Double.valueOf(this.lon ).hashCode()
         ^ Double.valueOf(this.alt ).hashCode()
         ^ Double.valueOf(this.azim).hashCode()
         ^ Double.valueOf(this.elev).hashCode()
         ^ Double.valueOf(this.roll).hashCode();
  }
  
  /** Describes the transform. */
  @Override
  public String toString() {
    String str = "Transform Ref="+getReferenceFrame()
               + " CoordSys="+getCoordinateSystem()
               + " PosUnits="+getPosUnits()
               + " VelUnits="+getVelUnits()
               + " AccUnits="+getAccUnits();
    switch (referenceFrame) {
      case TOP: str += " alt="+alt+" lat="+lat+" lon="+lon; break;
    }
    return str;
  }

  /** Gets the reference frame.
      @return The reference frame name.
      @since NeXtMidas 2.5.3
   */
  public String getReferenceFrame () {
    return Parser.get(referenceFrameList,referenceFrame);
  }

  /** Gets the coordinate system.
      @return The coordinate system name.
      @since NeXtMidas 2.5.3
   */
  public String getCoordinateSystem () {
    return Parser.get(Units.compTypeList,coordinateSystem);
  }

  /** Gets the units used for the POS. Note that within {@link Transform} only the
      <tt>fsu2fsu(..)</tt> functions use this value, all of the other
      functions assume base-SI units. <br>
      <br>
      When the {@link #getCoordinateSystem()} is {@link Units#GEODETIC} the POS
      units describe the units for the altitude component (e.g. {@link Units#DISTANCE_M}
      or {@link Units#ALTITUDE_M}) since the latitude and longitude components are
      automatically assumed to be in degrees.
      @since NeXtMidas 3.3.0
   */
  public String getPosUnits () {
    return DataFile.getUnitsVarName(unitsPos);
  }

  /** Gets the units used for the VEL. Note that within {@link Transform} only the
      <tt>fsu2fsu(..)</tt> functions use this value, all of the other
      functions assume base-SI units.
      @since NeXtMidas 3.3.0
   */
  public String getVelUnits () {
    return DataFile.getUnitsVarName(unitsVel);
  }

  /** Gets the units used for the ACC. Note that within {@link Transform} only the
      <tt>fsu2fsu(..)</tt> functions use this value, all of the other
      functions assume base-SI units.
      @since NeXtMidas 3.3.0
   */
  public String getAccUnits () {
    return DataFile.getUnitsVarName(unitsAcc);
  }

  /** Set referenceFrame/coordinateSystem based on the QuadWords of a Type 5000 file. The coordinate
      system is set based on that used for the <tt>POS</tt> component (if available), otherwise it
      is not set. <br>
      <br>
      Prior to NeXtMidas 3.3.0 this function would "fail fast" when the reference frame was invalid
      (i.e. not one of the ones defined in the BLUE 1.0 ICD), creating a scenario where legal coordinate
      system and/or units definitions were ignored. This caused a number of down-the-line issues for
      situations where users were using {@link Units#GEODETIC} coordinates for a POS (no VEL, ACC) but
      had a file that neglected to set the reference frame. In this situation the file was "technically
      illegal" but there was nothing preventing the values from being read correctly (since the reference
      frame was irrelevant in this situation). Since NeXtMidas 3.3.0, this function explicitly allows
      files of this form (without warning) and simply sets the reference frame to ECR.
      @param df The file.
      @return true if able to load the reference frame from the file, false otherwise.
   */
  public boolean fromDataFile (DataFile df) {
    // JC4: I think we should add a check here that returns false if df is not a Type 5000 file.

    // Set referenceFrame
    //   Prior to 3.3.0 the "else" case simply returned false. This caused a strange
    //   situation where a file with an invalid reference frame (e.g. "") but legal
    //   coordinateSystem and units did not get the coordinateSystem or units loaded
    //   properly.
    String frame = df.getReferenceFrame();
         if (frame.startsWith("ECR"))  referenceFrame = ECR;
    else if (frame.startsWith("ECI"))  referenceFrame = ECI;
    else if (frame.startsWith("TOPO")) referenceFrame = TOPO;
    else if (frame.startsWith("TOP"))  referenceFrame = TOP;
    else                               referenceFrame = NONE;

    // Load QuadWords
    alt        = df.getQuadword(DataFile.Q_ALT);
    lat        = df.getQuadword(DataFile.Q_LAT);
    lon        = df.getQuadword(DataFile.Q_LON);
    azim       = df.getQuadword(DataFile.Q_AZIM);
    elev       = df.getQuadword(DataFile.Q_ELEV);
    roll       = df.getQuadword(DataFile.Q_ROLL);
    //epoch_year = df.getQuadword(DataFile.Q_YEAR);
    //epoch_sec  = df.getQuadword(DataFile.Q_SEC);
    //epoch_gha  = df.getQuadword(DataFile.Q_GHA);
    epoch.fromQuadwords(df);

    // Set default units
    unitsPos = (df.getUnits() != Units.NONE)? df.getUnits() : Units.DISTANCE_M;
    unitsVel = Units.SPEED_M_S;
    unitsAcc = Units.ACCELERATION_M_S2;

    // Set coordinateSystem and units
    int iPos = df.findRec("POS");
    int iVel = df.findRec("VEL");
    int iAcc = df.findRec("ACC");
    if (iPos >= 0) {
      coordinateSystem = df.getCompType(iPos);
      unitsPos         = df.getCompUnits(iPos);
    }
    if (iVel >= 0) unitsVel = df.getCompUnits(iVel);
    if (iAcc >= 0) unitsAcc = df.getCompUnits(iAcc);

    
    if (referenceFrame == NONE) {
      if (frame.isEmpty() && (coordinateSystem == Units.GEODETIC)
                          && (iPos >= 0) && (iVel < 0) && (iAcc < 0)) {
        // This is a strange special-case where the reference frame was not given,
        // which is technically illegal under the BLUE 1.0 ICD, but since all we
        // have is POS and since POS is in GEODETIC, we don't actually need any
        // additional information. As such, we will just flag it as ECR and go
        // on without any warning.
        referenceFrame = ECR;
      }
    }
    
    return (referenceFrame != NONE);
  }

  /** Set quadwords of a Type 5000 file to match the referenceFrame defined here.
      Note that this does NOT set the coordinate system or units!
      @param df The file.
      @since NeXtMidas 3.3.0
   */
  public void toDataFile (DataFile df) {
    if (df.getTypeCodeClass() != 5) {
      throw new MidasException("Expected a Type 5000 file.");
    }
    
    // Set QuadWords
    if (referenceFrame == ECI) {
      df.setReferenceFrame("ECI");
      df.setQuadword(DataFile.Q_ALT,  0.0); // not used
      df.setQuadword(DataFile.Q_LAT,  0.0); // not used
      df.setQuadword(DataFile.Q_LON,  0.0); // not used
      df.setQuadword(DataFile.Q_AZIM, 0.0); // not used
      df.setQuadword(DataFile.Q_ELEV, 0.0); // not used
      df.setQuadword(DataFile.Q_ROLL, 0.0); // not used
      df.setQuadword(/* unused */  7, 0.0); // not used
      df.setQuadword(/* unused */  8, 0.0); // not used
      df.setQuadword(DataFile.Q_YEAR, epoch.getYear());
      df.setQuadword(DataFile.Q_SEC,  epoch.getSoY());
      df.setQuadword(DataFile.Q_GHA,  epoch.getGHA());
    }
    else if (referenceFrame == ECR) {
      df.setReferenceFrame("ECR");
      df.setQuadword(DataFile.Q_ALT,  0.0); // not used
      df.setQuadword(DataFile.Q_LAT,  0.0); // not used
      df.setQuadword(DataFile.Q_LON,  0.0); // not used
      df.setQuadword(DataFile.Q_AZIM, 0.0); // not used
      df.setQuadword(DataFile.Q_ELEV, 0.0); // not used
      df.setQuadword(DataFile.Q_ROLL, 0.0); // not used
      df.setQuadword(/* unused */  7, 0.0); // not used
      df.setQuadword(/* unused */  8, 0.0); // not used
      df.setQuadword(DataFile.Q_YEAR, 0.0); // not used
      df.setQuadword(DataFile.Q_SEC,  0.0); // not used
      df.setQuadword(DataFile.Q_GHA,  0.0); // not used
    }
    else if ((referenceFrame == TOP) || (referenceFrame == TOPO)) {
      if (referenceFrame == TOP) df.setReferenceFrame("TOP");
      else                       df.setReferenceFrame("TOPOCENT");
      df.setQuadword(DataFile.Q_ALT,  alt );
      df.setQuadword(DataFile.Q_LAT,  lat );
      df.setQuadword(DataFile.Q_LON,  lon );
      df.setQuadword(DataFile.Q_AZIM, azim);
      df.setQuadword(DataFile.Q_ELEV, elev);
      df.setQuadword(DataFile.Q_ROLL, roll);
      df.setQuadword(/* unused */  7, 0.0); // not used
      df.setQuadword(/* unused */  8, 0.0); // not used
      df.setQuadword(DataFile.Q_YEAR, 0.0); // not used
      df.setQuadword(DataFile.Q_SEC,  0.0); // not used
      df.setQuadword(DataFile.Q_GHA,  0.0); // not used
    }
    else {
      throw new MidasException("Unknown reference frame: "+referenceFrame);
    }
  }
  
  /** Converts an arbitrary object to a transform. Supported objects:
      <pre>
        null       - Returns null
        Transform  - Returns object passed in
        DataFile   - Returns 'new Transform((DataFile)obj)'
        Number     - New transform set to match {@link #NONE}, {@link #ECR},
                     {@link #ECI}, {@link #TOP} or {@link #TOPO} same as if
                     given the string form.
        Table      - **TBD**
        "NONE"     - New transform set to NONE
        "ECR"      - New transform set to ECR (epoch=0)
        "ECI"      - New transform set to ECI (epoch=0)
        "TOP"      - New transform set to TOP (epoch=0,lat=0,lon=0,alt=0)
        "TOPCENT"  - New transform set to TOPOCENT (epoch=0,lat=0,lon=0,alt=0,
                                                    azim=0,elev=0,roll=0)
        [other]    - IllegalArgumentException
      </pre>
      @param obj The object to convert
      @return The applicable transform.
      @since NeXtMidas 3.3.0
   */
  public static Transform toTransform (Object obj) {
    if (obj == null             ) return null;
    if (obj instanceof Transform) return (Transform)obj;
    if (obj instanceof DataFile ) return new Transform((DataFile)obj);
    
    if (obj instanceof Number) {
      switch (((Number)obj).intValue()) {
        case NONE: return new Transform(NONE, Units.CARTESIAN);
        case ECR:  return new Transform(ECR,  Units.CARTESIAN);
        case ECI:  return new Transform(ECI,  Units.CARTESIAN);
        case TOP:  return new Transform(TOP,  Units.CARTESIAN);
        case TOPO: return new Transform(TOPO, Units.CARTESIAN);
      }
    }
    else if (obj instanceof Tablizable) {
      // This is TBD
    }
    else {
      obj = obj.toString();
      if ("NONE".equals(obj)      ) return new Transform(NONE, Units.CARTESIAN);
      if ("ECR".equals(obj)       ) return new Transform(ECR,  Units.CARTESIAN);
      if ("ECI".equals(obj)       ) return new Transform(ECI,  Units.CARTESIAN);
      if ("TOP".equals(obj)       ) return new Transform(TOP,  Units.CARTESIAN);
      if ("TOPOCENT".equals(obj)  ) return new Transform(TOPO, Units.CARTESIAN);
    }
    throw new IllegalArgumentException("Could not convert "+obj+" to Transform");
  }
  
  /** Sets the transform in the {@link #TOP} reference frame and {@link Units#CARTESIAN} coordinate
      system. This is the same as <tt>setTOP(alt,lat,lon,0,0,0)</tt>.
      @param alt Altitude.
      @param lat Latitude (deg).
      @param lon Longitude (deg).
   */
  public final void setTOP (double alt, double lat, double lon) {
    setTOP(alt,lat,lon,0,0,0);
  }

  /** Sets the transform in the {@link #TOP} reference frame and {@link Units#CARTESIAN} coordinate
      system.
      @param alt  Altitude.
      @param lat  Latitude (deg).
      @param lon  Longitude (deg).
      @param azim The azimuth.
      @param elev The elevation.
      @param roll The roll.
   */
  public void setTOP (double alt, double lat, double lon,
                      double azim, double elev, double roll) {
    referenceFrame   = TOP;
    coordinateSystem = Units.CARTESIAN;
    this.alt = alt;
    this.lat = lat;
    this.lon = lon;
    this.azim = azim;
    this.elev = elev;
    this.roll = roll;
  }

  /** Changes reference frame to {@link #ECR} and coordinate system to
      {@link Units#CARTESIAN}. <br>
      <br>
      Prior to NeXtMidas 3.3.0 this function neglected to set the coordinate
      system resulting in undesirable behavior when users did:
      <pre>
        Transform t = new Transform(); // defaults to SCALAR not CARTESIAN or GEODETIC
        t.setECR();
      </pre>
   */
  public void setECR () {
    referenceFrame   = ECR;
    coordinateSystem = Units.CARTESIAN;
  }

  /** Changes reference frame to {@link #ECI} and coordinate system to
      {@link Units#CARTESIAN}.
      @param epoch The ECI epoch to use.
      @since NeXtMidas 3.3.0
   */
  public void setECI (Time epoch) {
    referenceFrame   = ECI;
    coordinateSystem = Units.CARTESIAN;
    setEpoch(epoch);
  }

  /** Sets the epoch and (indirectly) the GHA.
      @since NeXtMidas 3.3.0
   */
  public void setEpoch (Time epoch) {
    this.epoch.fromTime(epoch);
  }

  /** Convert a position from {@link Units#GEODETIC} to {@link Units#CARTESIAN}. No return value
      since the internal parameters of the input position are changed as a result of the conversion.
      @param p The position to convert.
   */
  public static void geo2car (Position p) {
    double f1,f2,rlat,rlon,slat,clat,slon,clon;
    f1 = 1-EARTH_FLATTENING_FACTOR;
    f2 = (2-EARTH_FLATTENING_FACTOR)*EARTH_FLATTENING_FACTOR;
    rlat = Math.atan(Math.tan(p.lat*DEG2RAD)*f1*f1);
    rlon = p.lon*DEG2RAD;
    slat = Math.sin(rlat);
    clat = Math.cos(rlat);
    slon = Math.sin(rlon);
    clon = Math.cos(rlon);
    p.r = (EARTH_EQUATORIAL_RADIUS*f1)/Math.sqrt(1-f2*clat*clat) + p.alt;
    p.x  = p.r*clat*clon;
    p.y  = p.r*clat*slon;
    p.z  = p.r*slat;

    // Take into account velocity north and east
    if (p instanceof StateVector) {
      StateVector sv = (StateVector) p;
      double[] v = new double[9];
      v[0] = sv.x;
      v[1] = sv.y;
      v[2] = sv.z;
      v[3] = sv.ve;
      v[4] = sv.vn;
      v[5] = sv.vu;
      v[6] = sv.ae;
      v[7] = sv.an;
      v[8] = sv.au;

      Transform t =  new Transform();
      t.setTOP(p.alt, p.lat, p.lon);
      top2ecr(v, POSVELACC, 1, t);

      sv.vx = v[3];
      sv.vy = v[4];
      sv.vz = v[5];
      sv.ax = v[6];
      sv.ay = v[7];
      sv.az = v[8];
    }
  }

  /** Converts a position from {@link Units#CARTESIAN} to {@link Units#GEODETIC}. No return value
      since the internal parameters of the input position are changed as a result of the conversion.
      @param p The position to convert.
   */
  public static void car2geo (Position p) {
    double f1,f2,a,ca;
    f1 = 1-EARTH_FLATTENING_FACTOR;
    f2 = (2-EARTH_FLATTENING_FACTOR)*EARTH_FLATTENING_FACTOR;
    p.r = Math.sqrt(p.x*p.x + p.y*p.y + p.z*p.z);
    a = Math.asin( p.z/p.r );
    ca = Math.cos(a);
    p.alt = p.r-(EARTH_EQUATORIAL_RADIUS*f1)/Math.sqrt(1-f2*ca*ca);
    p.lat = RAD2DEG * Math.atan(Math.tan(a)/(f1*f1));
    p.lon = RAD2DEG * Math.atan2(p.y,p.x);

    // Take into account velocity north and east
    if (p instanceof StateVector) {
      StateVector sv = (StateVector) p;
      double[] v = new double[9];
      v[0] = sv.x;
      v[1] = sv.y;
      v[2] = sv.z;
      v[3] = sv.vx;
      v[4] = sv.vy;
      v[5] = sv.vz;
      v[6] = sv.ax;
      v[7] = sv.ay;
      v[8] = sv.az;

      Transform t =  new Transform();
      t.setTOP(p.alt, p.lat, p.lon);
      ecr2top(v, POSVELACC, 1, t);

      sv.ve = v[3];
      sv.vn = v[4];
      sv.vu = v[5];
      sv.ae = v[6];
      sv.an = v[7];
      sv.au = v[8];
    }
  }

  /** Calculate the great arc distance between two positions. Note that this method calls
      {@link #geo2car(Position)} on both <tt>p</tt> and <tt>q</tt> prior to computing the distance.
      @param p The 1st position must initially be in {@link Units#GEODETIC} mode.
      @param q The 2nd position must initially be in {@link Units#GEODETIC} mode.
      @return The great arc distance in meters.
   */
  public static double greatArc (Position p, Position q) {
    geo2car(p);
    geo2car(q);
    double r = 0.0;
    double r2 = p.r*q.r;
    if (r2>1e-20) r = (p.x*q.x+p.y*q.y+p.z*q.z)/r2;
    r = Math.min(1.0,Math.max(-1.0,r));
    r = Math.acos(r) * EARTH_EQUATORIAL_RADIUS;
    return r;
  }

  /** Convert data from {@link Units#SPHERICAL} to {@link Units#CARTESIAN} coordinate system. There
      is no return value since the values are changed within the array.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param nrec The number of records to convert (1..N).
   */
  public static void sph2car (double[] v, int mpva, int nrec) {
    double ca,cb,sa,sb,r,a,b,rr,aa,bb,ri,rrr,aaa,bbb,rri;
    int isk = 3*mpva;
    for (int i=0,n=0; n<nrec; n++,i+=isk) {
      r = v[i+0];
      a = v[i+1]*DEG2RAD;
      b = v[i+2]*DEG2RAD;
      ca = Math.cos(a);  sa = Math.sin(a);
      cb = Math.cos(b);  sb = Math.sin(b);
      v[i+0] = r*cb*ca;
      v[i+1] = r*cb*sa;
      v[i+2] = r*sb;
      if (mpva<=1) continue;
      rr = v[i+3];
      aa = v[i+4]*DEG2RAD;
      bb = v[i+5]*DEG2RAD;
      ri = rr/r;
      v[i+3] = v[i+0]*ri - v[i+2]*bb*ca - v[i+1]*aa;
      v[i+4] = v[i+1]*ri - v[i+2]*bb*sa + v[i+0]*aa;
      v[i+5] = v[i+2]*ri + r*bb*cb;
      if (mpva<=2) continue;
      rrr = v[i+6];
      aaa = v[i+7]*DEG2RAD;
      bbb = v[i+8]*DEG2RAD;
      rri = rrr/r;
      v[i+6] = (v[i+3]-v[i+0]*ri)*ri + v[i+0]*rri - (v[i+5]*bb + v[i+2]*bbb)*ca
                        + v[i+2]*bb*aa*sa - v[i+1]*aaa - v[i+4]*aa;
      v[i+7] = (v[i+4]-v[i+1]*ri)*ri + v[i+1]*rri - (v[i+5]*bb + v[i+2]*bbb)*sa
                        - v[i+2]*bb*aa*ca - v[i+0]*aaa + v[i+3]*aa;
      v[i+8] = (v[i+5]-v[i+2]*ri)*ri + v[i+1]*rri + (rr*bb+r*bbb)*ca + v[i+2]*bb*bb;
    }
  }

  /** Convert data from {@link Units#GEODETIC} to ECR {@link Units#CARTESIAN} coordinate system. There
      is no return value since the values are changed within the array.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param nrec The number of records to convert (1..N).
   */
  public static void geo2car (double[] v, int mpva, int nrec) {
    double ca,cb,sa,sb,r,a,b,ee,t;
    ee = EARTH_ECCENTRICITY*EARTH_ECCENTRICITY;
    int isk = 3*mpva;
    for (int i=0,n=0; n<nrec; n++,i+=isk) {
      r = v[i+0];
      a = v[i+1]*DEG2RAD;
      b = v[i+2]*DEG2RAD;
      ca = Math.cos(a);  sa = Math.sin(a);
      cb = Math.cos(b);  sb = Math.sin(b);
      t = EARTH_EQUATORIAL_RADIUS/Math.sqrt(1-ee*sa*sa);
      v[i+0] = (r+t)*cb*ca;
      v[i+1] = (r+t)*ca*sb;
      v[i+2] = (r+(1-ee)*t)*sa;
      if (mpva<=1) continue;
    }
  }

  /** <b>Not implemented yet:</b>
      Convert data from {@link Units#CYLINDRIC} to ECR {@link Units#CARTESIAN} coordinate system. There
      is no return value since the values are changed within the array.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param nrec The number of records to convert (1..N).
   */
  public static void cyl2car (double[] v, int mpva, int nrec) {
    Shell.warning("Transform.cyl2car is not yet coded");
  }

  /** <b>Not implemented yet:</b>
      Convert data from {@link Units#ELLIPSOID} to ECR {@link Units#CARTESIAN} coordinate system. There
      is no return value since the values are changed within the array.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param nrec The number of records to convert (1..N).
   */
  public static void ell2car (double[] v, int mpva, int nrec) {
    Shell.warning("Transform.ell2car is not yet coded");
  }

  /** Convert {@link Units#CARTESIAN} data from {@link #ECR} (ECEF) to {@link #ECI} reference frame.
      There is no return value since the values are changed within the array.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param gha   Greenwich hour angle vector (in radians)
      @param nrec The number of records to convert (1..N).
      @since NeXtMidas 2.7.0
   */
  public static void ecr2eci (double[] v, int mpva, int nrec, double[] gha) {
    eci2ecr(v, mpva, nrec, gha, false);
  }

  /** Convert {@link Units#CARTESIAN} data from {@link #ECR} (ECEF) to {@link #ECI} reference frame.
      There is no return value since the values are changed within the array.
      This is the same as {@link #ecr2eci(double[], int, int, double[])}.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param gha   Greenwich hour angle vector (in radians)
      @param nrec The number of records to convert (1..N).
      @since NeXtMidas 2.7.0
   */
  public static void ecef2eci (double[] v, int mpva, int nrec, double[] gha) {
    eci2ecr(v, mpva, nrec, gha, false);
  }

  /** Convert {@link Units#CARTESIAN} data from {@link #ECI} to {@link #TOP}.
      There is no return value since the values are changed within the array. <br>
      <br>
      <i>This conversion is currently implemented as a call to {@link #eci2ecr} followed
      by a call to {@link #ecr2top}, but may be optimized in the future.</i>
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param nrec The number of records to convert (1..N).
      @param from {@link Transform} specifying the reference frame parameters for the input values.
      @param to   {@link Transform} specifying the reference frame parameters for the output values.
      @param gha  Greenwich hour angle vector (in radians)
      @since NeXtMidas 3.3.0
   */
  public static void eci2top (double[] v, int mpva, int nrec, Transform from, Transform to, double[] gha) {
    eci2ecr(v, mpva, nrec, gha);
    ecr2top(v, mpva, nrec, to);
  }

  /** Convert {@link Units#CARTESIAN} data from {@link #ECI} to {@link #ECR} (ECEF) reference
      frame. There is no return value since the values are changed within the array.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param gha   Greenwich hour angle vector (in radians)
      @param nrec The number of records to convert (1..N).
      @since NeXtMidas 2.7.0
   */
  public static void eci2ecr (double[] v, int mpva, int nrec, double[] gha) {
    eci2ecr(v, mpva, nrec, gha, true);
  }

  /** Convert {@link Units#CARTESIAN} data from {@link #ECI} to {@link #ECR} (ECEF) reference
      frame. There is no return value since the values are changed within the array.
      This is the same as {@link #eci2ecr(double[], int, int, double[])}.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param gha   Greenwich hour angle vector (in radians)
      @param nrec The number of records to convert (1..N).
      @since NeXtMidas 2.7.0
   */
  public static void eci2ecef (double[] v, int mpva, int nrec, double[] gha) {
    eci2ecr(v, mpva, nrec, gha, true);
  }

  /** Convert {@link Units#CARTESIAN} data from/to {@link #ECI} to/from {@link #ECR} (ECEF)
      reference frame. There is no return value since the values are changed within the array.
      Note, this conversion does not account for the earth's polar motion.
      @param v     Array of position vectors.
      @param mpva  Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param nrec  The number of records to convert (1..N).
      @param gha   Greenwich hour angle vector (in radians)
      @param toECR true if ECI to ECR, false for ECR to ECI
      @since NeXtMidas 2.7.0, Added java port of M$ECI_TO_ECR in type5000.for
   */
  static void eci2ecr (double[] v, int mpva, int nrec, double[] gha, boolean toECR) {
    int dir = (toECR)? 1 : -1; // 1 for ECI to ECR, -1 for ECR to ECI
    int isk = 3*mpva;
    double eav,eavt,eav2,x,y,xx,yy,xxx,yyy,ca,sa;
    eav  = dir*EARTH_ANGULAR_VELOCITY;
    eavt = eav+eav;
    eav2 = eav*eav;
    for (int i=0,n=0; n<nrec; n++,i+=isk) {
      sa = Math.sin(gha[n]);
      ca = Math.cos(gha[n]);
      if (dir<0) sa = -sa;
      x      = v[i+0];
      y      = v[i+1];
      v[i+0] =  ca*x + sa*y;
      v[i+1] = -sa*x + ca*y;
      if (mpva<=1) continue;
      xx = v[i+3];
      yy = v[i+4];
      v[i+3] =  ca*xx + sa*yy + eav*(-sa*x+ca*y);
      v[i+4] = -sa*xx + ca*yy + eav*(-ca*x-sa*y);
      if (mpva<=2) continue;
      xxx = v[i+6];
      yyy = v[i+7];
      v[i+6] =  ca*xxx + sa*yyy + eavt*(-sa*xx+ca*yy) + eav2*(-ca*x-sa*y);
      v[i+7] = -sa*xxx + ca*yyy + eavt*(-ca*xx-sa*yy) + eav2*(+sa*x-ca*y);
    }
  }

  /** Convert {@link Units#CARTESIAN} data from {@link #TOP} to {@link #ECI}.
      There is no return value since the values are changed within the array. <br>
      <br>
      <i>This conversion is currently implemented as a call to {@link #top2ecr} followed
      by a call to {@link #ecr2eci}, but may be optimized in the future.</i>
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param nrec The number of records to convert (1..N).
      @param from {@link Transform} specifying the reference frame parameters for the input values.
      @param to   {@link Transform} specifying the reference frame parameters for the output values.
      @param gha  Greenwich hour angle vector (in radians)
      @since NeXtMidas 3.3.0
   */
  public static void top2eci (double[] v, int mpva, int nrec, Transform from, Transform to, double[] gha) {
    top2ecr(v, mpva, nrec, from);
    ecr2eci(v, mpva, nrec, gha);
  }

  /** Convert {@link Units#CARTESIAN} data from {@link #TOP} to {@link #ECR} reference frame. There
      is no return value since the values are changed within the array.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param nrec The number of records to convert (1..N).
      @param from {@link Transform} specifying the reference frame parameters for the input values.
   */
  public static void top2ecr (double[] v, int mpva, int nrec, Transform from) {
    TMatrix tm = top2ecrTMatrix(from);
    matrixVectorMultiply (tm,v,mpva,nrec);
  }

  /** Convert {@link Units#CARTESIAN} data between two {@link #TOP} reference frames with different
      reference frame parameters. There is no return value since the values are changed within the array.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param nrec The number of records to convert (1..N).
      @param from {@link Transform} specifying the reference frame parameters for the input values.
      @param to   {@link Transform} specifying the reference frame parameters for the output values.
   */
  public static void top2top (double[] v, int mpva, int nrec, Transform from, Transform to) {
    TMatrix tm = top2ecrTMatrix(from);
    TMatrix tn = top2ecrTMatrix(to);
    tn.transpose();
    tn.multiply(tm);
    matrixVectorMultiply (tn,v,mpva,nrec);
  }

  /** Convert {@link Units#CARTESIAN} data from {@link #ECR} to {@link #TOP} reference frame. There
      is no return value since the values are changed within the array.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param nrec The number of records to convert (1..N).
      @param to   {@link Transform} specifying the reference frame parameters for the output values.
   */
  public static void ecr2top (double[] v, int mpva, int nrec, Transform to) {
    TMatrix tm = top2ecrTMatrix(to);
    tm.transpose();
    matrixVectorMultiply (tm,v,mpva,nrec);
  }

  /** Convert data from {@link Units#CARTESIAN} to {@link Units#SPHERICAL} coordinate system. There
      is no return value since the values are changed within the array.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param nrec The number of records to convert (1..N).
   */
  public static void car2sph (double[] v, int mpva, int nrec) {
    double x,y,z,xx,yy,zz,xxx,yyy,zzz,r,ri,si,si2;
    int isk = 3*mpva;
    for (int i=0,n=0; n<nrec; n++,i+=isk) {
      x = v[i+0];
      y = v[i+1];
      z = v[i+2];
      r = Math.sqrt(x*x+y*y+z*z);
      ri = 1.0/r;
      v[i+0] = r;
      v[i+1] = Math.atan2(y,x)*RAD2DEG;
      v[i+2] = Math.asin(z*ri)*RAD2DEG;
      if (mpva<=1) continue;
      xx = v[i+3];
      yy = v[i+4];
      zz = v[i+5];
      si2 = Math.sqrt(x*x+y*y);
      si = Math.sqrt(si2);
      v[i+3] = (x*xx+y*yy+z*zz)*ri;
      v[i+4] = (x*yy-xx*y)*si2*RAD2DEG;
      v[i+5] = (zz-z*v[i+3]*ri)*si*RAD2DEG;
      if (mpva<=2) continue;
      xxx = v[i+6];
      yyy = v[i+7];
      zzz = v[i+8];
      v[i+6] = (xx*xx+yy*yy+zz*zz + x*xxx+y*yyy+z*zzz - v[i+3]*v[i+3])*ri;
      v[i+7] = ((x*yyy-xxx*y)*RAD2DEG-2*v[i+4]*(x*xx+y*yy))*si2;
      v[i+8] = (zzz-zz*v[i+3]*ri-z*(v[i+6]*ri-(v[i+3]*ri)*(v[i+3]*ri)))
                     *si*RAD2DEG - v[i+5]*(x*xx+y*yy)*si2;
    }
  }

  /** @deprecated Since NeXtMidas 3.3.0: This function was never implemented and
      has been superceded by {@link #fsu2fsu(double[],int,int,Transform,Transform,double[])}.
   */
  @Deprecated
  public static void fsu2fsu (double[] v, int mpva, int nrec, Transform from, Transform to) {
    Shell.warning("Transform.fsu2fsu is not yet coded");
  }

  /** Convert reference frame, coordinate system, and units between two different transforms.
      This is a variant of {@link #fsu2fsu(double[],int,int,Transform,Transform,double[])}
      that is intended for use when there is a single record to transform.
      @param v     Array of position vectors.
      @param mpva  Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param from  {@link Transform} specifying the reference frame parameters for the input values.
      @param to    {@link Transform} specifying the reference frame parameters for the output values.
      @param epoch The epoch defining the Greenwich hour angle vector (in radians) to use (see
                   {@link Time#getGHA()}. Required for any conversions involving ECI, can be omitted
                   otherwise.
   */
  public static void fsu2fsu (double[] v, int mpva, Transform from, Transform to, Time epoch) {
    fsu2fsu(v, mpva, 1, from, to, null, epoch);
  }

  /** Convert reference frame, coordinate system, and units between two different transforms. There
      is no return value since the values are changed within the array. This function does not
      support ALL possible conversions, only those most common involving transforms.
      <pre>
        Supported Units:
          {@link Units#DISTANCE_M}    {@link Units#SPEED_M_S}    {@link Units#ACCELERATION_M_S2}
          {@link Units#DISTANCE_FT}   {@link Units#SPEED_FT_S}   {@link Units#ACCEL_FT_S2}
          {@link Units#DISTANCE_NMI}  {@link Units#SPEED_NMI_S}  {@link Units#ACCEL_NMI_S2}
                                      {@link Units#SPEED_KNOTS}  {@link Units#ACCEL_KNOTS_S}

        Supported Reference Frames:
          {@link Units#CARTESIAN}  {@link Units#SPHERICAL}  {@link Units#CYLINDRIC}
          {@link Units#ELLIPSOID}  {@link Units#GEODETIC}

        Supported Coordinate Systems:
          "ECR"  "ECI"  "TOP"  "TOPOCENT" (or "TOPO")
      </pre>
      The algorithm used:
      <pre>
        1. Convert from input units to base-SI units
        2. Convert from input reference frame to {@link Units#CARTESIAN}
        3. Convert reference frame
        4. Convert from {@link Units#CARTESIAN} to output reference frame
        5. Convert from base-SI units to output units
      </pre>
      This function is specifically optimized to return immediately (with no changes to the values)
      if either transform is null or if the two transforms are equal.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param nrec The number of records to convert (1..N).
      @param from {@link Transform} specifying the reference frame parameters for the input values.
      @param to   {@link Transform} specifying the reference frame parameters for the output values.
      @param gha  Greenwich hour angle vector (in radians). Required for any conversions involving
                  ECI, can be omitted otherwise. Length of array must match <tt>nrec</tt>.
   */
  public static void fsu2fsu (double[] v, int mpva, int nrec,
                              Transform from, Transform to, double[] gha) {
    fsu2fsu(v, mpva, nrec, from, to, gha, null);
  }
  
  /** Internal Use Only: This version takes GHA as array or single value. */
  @SuppressWarnings("fallthrough")
  private static void fsu2fsu (double[] v, int mpva, int nrec,
                               Transform from, Transform to,
                               double[] gha, Time epoch) {
    if ((from == null) || (to == null) || from.equals(to)) return; // nothing to do
    int fromRef = from.referenceFrame;
    int toRef   = to.referenceFrame;
    
    // CONVERT TO SI UNITS
    switch (mpva) {
      case POSVELACC: fixUnits(v, mpva, nrec, 6, from.unitsAcc, true); // FALLTHROUGH
      case POSVEL:    fixUnits(v, mpva, nrec, 3, from.unitsVel, true); // FALLTHROUGH
      case POS:       fixUnits(v, mpva, nrec, 0, from.unitsPos, true); // FALLTHROUGH
    }

    // CONVERT TO CARTESIAN
    switch (from.coordinateSystem) {
      case CARTESIAN: /* nothing to do */                    break;
      case SPHERICAL: sph2car(v, mpva, nrec); fromRef = ECR; break;
      //TBD: case CYLINDRIC: cyl2car(v, mpva, nrec); fromRef = ECR; break;
      //TBD: case ELLIPSOID: ell2car(v, mpva, nrec); fromRef = ECR; break;
      case GEODETIC:  geo2car(v, mpva, nrec); fromRef = ECR; break;
      default: throw new MidasException("Unsupported conversion "+from.getCoordinateSystem()
                                       +" to CARTESIAN to "+to.getCoordinateSystem());
    }

    // PREPARE FOR CONVERT FROM CARTESIAN
    switch (to.coordinateSystem) {
      case CARTESIAN: /* no change */ break;
      case SPHERICAL: toRef = ECR;    break;
      //TBD: case CYLINDRIC: toRef = ECR;    break;
      //TBD: case ELLIPSOID: toRef = ECR;    break;
      case GEODETIC:  toRef = ECR;    break;
      default: throw new MidasException("Unsupported conversion "+from.getCoordinateSystem()
                                       +" to CARTESIAN to "+to.getCoordinateSystem());
    }

    // CONVERT REFERENCE
    if ((fromRef != toRef) || (fromRef == TOP) || (fromRef == TOPO)) {
      // Only do this if converting between two reference frames or between two
      // TOP/TOPO references. We skip this for everything else since it is
      // possible to convert coordinate systems without a "valid" reference.
      if ((gha == null) && (epoch != null)) {
        gha = new double[] { epoch.getGHA() };
      }
      switch (fromRef + (toRef * 100)) {
        case ECR +(ECI *100): ecr2eci(v, mpva, nrec,           gha); break;
        case ECR +(TOP *100): ecr2top(v, mpva, nrec,       to     ); break;
        case ECR +(TOPO*100): ecr2top(v, mpva, nrec,       to     ); break;
        case ECI +(ECR *100): eci2ecr(v, mpva, nrec,           gha); break;
        case ECI +(TOP *100): eci2top(v, mpva, nrec, from, to, gha); break;
        case ECI +(TOPO*100): eci2top(v, mpva, nrec, from, to, gha); break;
        case TOP +(ECR *100): top2ecr(v, mpva, nrec, from         ); break;
        case TOP +(ECI *100): top2eci(v, mpva, nrec, from, to, gha); break;
        case TOP +(TOP *100): top2top(v, mpva, nrec, from, to     ); break;
        case TOP +(TOPO*100): top2top(v, mpva, nrec, from, to     ); break;
        case TOPO+(ECR *100): top2ecr(v, mpva, nrec, from         ); break;
        case TOPO+(ECI *100): top2eci(v, mpva, nrec, from, to, gha); break;
        case TOPO+(TOP *100): top2top(v, mpva, nrec, from, to     ); break;
        case TOPO+(TOPO*100): top2top(v, mpva, nrec, from, to     ); break;
        default: throw new MidasException("Unsupported conversion "+from.getReferenceFrame()
                                         +" to "+to.getReferenceFrame());
      }
    }

    // CONVERT FROM CARTESIAN
    //   Note: We intentionally omit "case CARTESIAN:" and "default:" as there
    //   is nothing further to do for those cases.
    switch (to.coordinateSystem) {
      case SPHERICAL: car2sph(v, mpva, nrec); break;
      case CYLINDRIC: car2cyl(v, mpva, nrec); break;
      case ELLIPSOID: car2ell(v, mpva, nrec); break;
      case GEODETIC:  car2geo(v, mpva, nrec); break;
    }

    // CONVERT FROM SI UNITS
    switch (mpva) {
      case POSVELACC: fixUnits(v, mpva, nrec, 6, to.unitsAcc, false); //FALLTHROUGH
      case POSVEL:    fixUnits(v, mpva, nrec, 3, to.unitsVel, false); //FALLTHROUGH
      case POS:       fixUnits(v, mpva, nrec, 0, to.unitsPos, false); //FALLTHROUGH
    }
  }

  /** <b>Internal use only:</b> Convert reference frame, coordinate system, and
      units between two different transforms. There is no return value since the
      values are changed within the table. The table is expected to have the
      following keys:
      <pre>
        POS=(x,y,x)   Position
        VEL=(x,y,x)   Velocity      <i>[Must be omitted if POS is omitted.]</i>
        ACC=(x,y,x)   Acceleration  <i>[Must be omitted if VEL is omitted.]</i>
      </pre>
      @param tbl  The table with the entries. The output will be as an updated
                  <tt>double[]</tt> entries in the table.
      @param from {@link Transform} specifying the reference frame parameters for the input values.
      @param to   {@link Transform} specifying the reference frame parameters for the output values.
      @param gha  Greenwich hour angle vector (in radians). Required for any conversions involving
                  ECI, can be omitted otherwise.
      @see #fsu2fsu(double[], int, int, Transform, Transform, double[])
      @since NeXtMidas 3.3.0
   */
  @InternalUseOnly
  @SuppressWarnings("fallthrough")
  public static void fsu2fsu (Table tbl, Transform from, Transform to, Double gha) {
    int      mpva = POSVELACC;
    double[] acc  = Convert.o2da(tbl.get("ACC"));  if (acc == null) mpva = POSVEL;
    double[] vel  = Convert.o2da(tbl.get("VEL"));  if (vel == null) mpva = POS;
    double[] pos  = Convert.o2da(tbl.get("POS"));  if (pos == null) return; // nothing to do
    double[] _gha = (gha == null)? null : new double[] { gha };
    double[] v    = new double[9];
    
    switch (mpva) {
      case POSVELACC: System.arraycopy(acc, 0, v, 6, 3); //FALLTHROUGH
      case POSVEL:    System.arraycopy(vel, 0, v, 3, 3); //FALLTHROUGH
      case POS:       System.arraycopy(pos, 0, v, 0, 3);
    }
    
    fsu2fsu(v, mpva, 1, from, to, _gha);
    
    switch (mpva) {
      case POSVELACC: System.arraycopy(v, 6, acc, 0, 3);  tbl.put("ACC", acc); //FALLTHROUGH
      case POSVEL:    System.arraycopy(v, 3, vel, 0, 3);  tbl.put("VEL", vel); //FALLTHROUGH
      case POS:       System.arraycopy(v, 0, pos, 0, 3);  tbl.put("POS", pos);
    }
  }

  /** Converts to/from SI units.
      @param v     The array.
      @param mpva  POS, POSVEL, or POSVELACC.
      @param nrec  The number of records.
      @param off   The offset (0=POS, 3=VEL, or 6=ACC).
      @param val   The value to multiply by.
      @param units The units to convert to/from.
      @param toSI  true=to SI units, false=from SI units
   */
  private static void fixUnits (double[] v, int mpva, int nrec, int off, int units, boolean toSI) {
    double mult = 0;
    switch (units) {
      // DIST/SPEED/ACCEL
      case Units.DISTANCE_M:        return; // nothing to do
      case Units.DISTANCE_FT:       mult = FEET2METERS; break;
      case Units.DISTANCE_NMI:      mult = NM2METERS;   break;
      case Units.SPEED_M_S:         return; // nothing to do
      case Units.SPEED_FT_S:        mult = FEET2METERS; break;
      case Units.SPEED_NMI_S:       mult = NM2METERS;   break;
      case Units.SPEED_KNOTS:       mult = KNOTS2MS;    break;
      case Units.ACCELERATION_M_S2: return; // nothing to do
      case Units.ACCEL_FT_S2:       mult = FEET2METERS; break;
      case Units.ACCEL_NMI_S2:      mult = NM2METERS;   break;
      case Units.ACCEL_KNOTS_S:     mult = KNOTS2MS;    break;
      case Units.ACCEL_G:           mult = GFORCE2MS2;  break;
      // OTHER
      case Units.ALTITUDE_M:        return; // nothing to do
      case Units.ALTITUDE_FT:       mult = FEET2METERS; break;
    }

    if (mult == 0) {
      String name = DataFile.getUnitsVarName(units);
      if (toSI) throw new MidasException("Can not convert "+name+" to SI units");
      else      throw new MidasException("Can not convert SI units to "+name);
    }

    if (!toSI) mult = 1/mult;

    for (int i = 0; i < nrec; i++) {
      v[off++] *= mult;
      v[off++] *= mult;
      v[off++] *= mult;
      off += (mpva-1)*3;
    }
  }

  /** Transform data for plotter coordinates (Lon,Lat,Alt) in EER.
      @param v    Array of position vectors.
      @param nrec The number of records to convert (1..N).
      @return Always returns <tt>nrec</tt>.
   */
  public int toLonLatAlt (double[] v, int nrec) {
    // TODO: Replace the code here with the following since the current implementation
    // lacks a LOT of error checking and will give bad results if things aren't
    // quite right.
    //    fsu2fsu(v, POS, nrec, this, Transform_ECR_GEODETIC, null, null);
    //    all2lla(v,nrec);  // plotter needs (Lon,Lat,Alt) not (Alt,Lat,Lon)
    //    return nrec;
    if (referenceFrame==TOP||referenceFrame==TOPO) top2ecr(v,POS,nrec,this);
    if (coordinateSystem==Units.CARTESIAN) car2geo(v,POS,nrec);
    all2lla(v,nrec);  // plotter needs (Lon,Lat,Alt) not (Alt,Lat,Lon)
    return nrec;
  }

  /** Transform data from plotter coordinates (Lon,Lat,Alt) in EER. This is the
      reverse of {@link #toLonLatAlt}.
      @param v    Array of position vectors.
      @param nrec The number of records to convert (1..N).
      @return Always returns <tt>nrec</tt>.
      @since NeXtMidas 3.3.0
   */
  public int fromLonLatAlt (double[] v, int nrec) {
    lla2all(v,nrec);
    fsu2fsu(v, POS, nrec, Transform_ECR_GEODETIC, this, null, null);
    return nrec;
  }

  /** Transform AltLatLon vector (as given in a BLUE file) to LonLatAlt vector for plotters.
      @param v    Array of position vectors.
      @param nrec The number of records to convert (1..N).
      @return Always returns <tt>nrec</tt>.
   */
  public static int all2lla (double[] v, int nrec) {
    double tmp;
    for (int n=0,i=0; n<nrec; n++,i+=3) {
      tmp=v[i+0]; v[i+0]=v[i+2]; v[i+2]=tmp;
    }
    return nrec;
  }

  /** Transform LonLatAlt vector to AltLatLon vector.
      @param v    Array of position vectors.
      @param nrec The number of records to convert (1..N).
      @return Always returns <tt>nrec</tt>.
   */
  private static int lla2all (double[] v, int nrec) {
    return all2lla(v, nrec); // <-- the math is the same (i.e. swap v[0] with v[2])
  }

  /** Clip at earth intersect lines from point to origin.
      @param v     The position vector (has a length that is a multiple of three
                   such that v[i+0]=x, v[i+1]=y, v[i+2]=z makes up a
                   single record/point).
      @param ioff  The offset into the position vector.
      @param nrec  The number of records/points to act on.
      @param vo    The reference point (has a length of three such that
                   v[0]=x, v[1]=y, v[2]=z makes up a record/point).
      @param mode  The clipping mode when clipping at the Earth's surface. <pre>
                   mode = 1 - Cone extends to horizon.
                   mode = 0 - Clip cone at a plane perpendicular to q.</pre>
      @return Always returns <tt>nrec</tt>.
   */
  public static int clipAtEarth (double[] v, int ioff, int nrec, double[] vo, int mode) {

    double xo,yo,zo,xd,yd,zd;       // origin and delta vectors
    double xe,ye,ze,xei,yei,zei;    // eccentricity vector and inverse
    double p,qA,qB,qC,qR,qS,qSR;    // quadratic components
    double poo,pod,pdd;             // inner products

    xe = EARTH_EQUATORIAL_RADIUS;
    ye = EARTH_EQUATORIAL_RADIUS;
    ze = xe*Math.sqrt(1.0-(EARTH_ECCENTRICITY*EARTH_ECCENTRICITY));
    xei = 1.0/xe;
    yei = 1.0/ye;
    zei = 1.0/ze;
    xo = xei*vo[0];
    yo = yei*vo[1];
    zo = zei*vo[2];
    for (int n=0,i=ioff; n<nrec; n++,i+=3) {
      xd = xei*v[i+0] - xo;
      yd = yei*v[i+1] - yo;
      zd = zei*v[i+2] - zo;
      pdd = xd*xd + yd*yd + zd*zd;
      pod = xo*xd + yo*yd + zo*zd;
      poo = xo*xo + yo*yo + zo*zo;
      qA = pdd;
      if (qA<1.0e-30) continue;
      qB = 2.0*pod;
      qC = poo - 1.0;
      qR = qB / (qA*2.0);
      qS = qC / qA;
      if (qR*qR > qS) {   // has 2 solutions
        qSR = Math.sqrt( qR*qR - qS );
        p = -qR - qSR;
        if (mode==1 || (p>0.0 && p<1.0)) {  // use it
          v[i+0] = (xo + p*xd) * xe;
          v[i+1] = (yo + p*yd) * ye;
          v[i+2] = (zo + p*zd) * ze;
        }
      } else if (mode==1) {
        qA = pod*pod - pdd;
        qB = 2.0*pod*(poo-1.0);
        qC = poo*(poo-1.0);
        qR = qB / (qA*2.0);
        qS = qC / qA;
        qSR = qR*qR - qS;
        if (qSR>0) qSR = Math.sqrt(qSR); else qSR=0;
        if (qR>0) p = -qR + qSR;
        else p = -qR - qSR;
        v[i+0] = (xo + p*xd) * xe;
        v[i+1] = (yo + p*yd) * ye;
        v[i+2] = (zo + p*zd) * ze;
      }
    }
    return nrec;
  }

  /** Convert data from ECR {@link Units#CARTESIAN} to {@link Units#GEODETIC} coordinate system. There
      is no return value since the values are changed within the array.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param nrec The number of records to convert (1..N).
   */
  public static void car2geo (double[] v, int mpva, int nrec) {
    double x,y,z,a,b,c,eer,ee,ee1,ee1i,r,t,f1,f2,sa,ca,h,da,dh;
    ee = EARTH_ECCENTRICITY*EARTH_ECCENTRICITY;
    ee1 = (1-ee); ee1i = 1/ee1;
    eer = EARTH_EQUATORIAL_RADIUS;
    int isk = 3*mpva;
    for (int i=0,n=0; n<nrec; n++,i+=isk) {
      x = v[i+0];
      y = v[i+1];
      z = v[i+2];
      if (x==0 && y==0) {
        v[i+0] = Math.abs(z) - eer*Math.sqrt(ee1);
        v[i+1] = (z<0)? -90 : 90;
        v[i+2] = 0;
        continue;
      }
      r = Math.sqrt(x*x+y*y+z*z);
      b = Math.sqrt(x*x+y*y);
      c = Math.asin(z/r);
      a = Math.atan( Math.tan(c)*ee1i );
      sa = Math.sin(a);
      h = r - eer/Math.sqrt(1+ee*sa*sa*ee1i);
      do {
        sa = Math.sin(a); ca = Math.cos(a);
        t = 1./Math.sqrt(1-ee*sa*sa);
        f1 = (h + eer*t)*ca - b;
        f2 = (h + eer*ee1*t)*sa - z;
        da = (-f1*sa+f2*ca)/(h+eer*ee1*t*t*t);
        dh = f1*ca+f2*sa;
        a -= da; h -= dh;
      } while (da>.0001 || dh>.0001);
      v[i+0] = h;
      v[i+1] = RAD2DEG*a;
      v[i+2] = RAD2DEG*Math.atan2(y,x);
      if (mpva<=1) continue;
    }
  }

  /** <b>Not implemented yet:</b>
      Convert data from {@link Units#ELLIPSOID} to {@link Units#CYLINDRIC} coordinate system. There
      is no return value since the values are changed within the array.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param nrec The number of records to convert (1..N).
   */
  public static void car2cyl (double[] v, int mpva, int nrec) {
    Shell.warning("Transform.car2cyl is not yet coded");
  }

  /** <b>Not implemented yet:</b>
      Convert data from {@link Units#ELLIPSOID} to {@link Units#ELLIPSOID} coordinate system. There
      is no return value since the values are changed within the array.
      @param v    Array of position vectors.
      @param mpva Mode: {@link #POS}, {@link #POSVEL}, {@link #POSVELACC}
      @param nrec The number of records to convert (1..N).
   */
  public static void car2ell (double[] v, int mpva, int nrec) {
    Shell.warning("Transform.car2ell is not yet coded");
  }

  private static TMatrix top2ecrTMatrix (Transform t) {
    double t1,t2,slat,clat,slon,clon,x,y,z;
//JEFF    if (t.lat>90 || t.lat<-90) return null;
    slat = Math.sin(t.lat*DEG2RAD); clat = Math.cos(t.lat*DEG2RAD);
    slon = Math.sin(t.lon*DEG2RAD); clon = Math.cos(t.lon*DEG2RAD);
    t1 = EARTH_ECCENTRICITY*EARTH_ECCENTRICITY;
    t2 = EARTH_EQUATORIAL_RADIUS/Math.sqrt(1-t1*slat*slat);
    x = (t.alt+t2)*clat*clon;
    y = (t.alt+t2)*clat*slon;
    z = (t.alt+(1-t1)*t2)*slat;
    TMatrix tm = new TMatrix();
    tm.rotate(t.roll,TMatrix.ROLL);          // local ROLL
    tm.rotate(t.elev,TMatrix.ELEV);          // local ELEVation
    tm.rotate(t.azim,TMatrix.AZIM);          // local AZIMuth
    tm.rotate(90.0-t.lat,TMatrix.ROLL);      // topo lat roll
    tm.rotate(90.0+t.lon,TMatrix.AZIM);      // topo lon AZIMuth
    tm.offset(x,y,z);                        // topo origin
    return tm;
  }

  private static void matrixVectorMultiply (TMatrix tm, double[] v, int mpva, int nrec) {
    switch (mpva) {
      case 0:
        matrixVectorMultA(tm,v,0,3,nrec);
        break;
      case 1:     // POS
        matrixVectorMultAB(tm,v,0,3,nrec);
        break;
      case 2:     // POS,VEL
        matrixVectorMultAB(tm,v,0,6,nrec);
        matrixVectorMultA(tm,v,3,6,nrec);
        break;
      case 3:     // POS,VEL
        matrixVectorMultAB(tm,v,0,9,nrec);
        matrixVectorMultA(tm,v,3,9,nrec);
        matrixVectorMultA(tm,v,6,9,nrec);
        break;
    }
  }

  private static void matrixVectorMultA (TMatrix tm, double[] v, int ioff, int isk, int nrec) {
    double x,y,z; int i,n;
    for (i=ioff,n=0; n<nrec; i+=isk,n++) {
      x=v[i+0]; y=v[i+1]; z=v[i+2];
      v[i+0] = x*tm.a11 + y*tm.a12 + z*tm.a13;
      v[i+1] = x*tm.a21 + y*tm.a22 + z*tm.a23;
      v[i+2] = x*tm.a31 + y*tm.a32 + z*tm.a33;
    }
  }

  private static void matrixVectorMultAB (TMatrix tm, double[] v, int ioff, int isk, int nrec) {
    double x,y,z; int i,n;
    for (i=ioff,n=0; n<nrec; i+=isk,n++) {
      x=v[i+0]; y=v[i+1]; z=v[i+2];
      v[i+0] = x*tm.a11 + y*tm.a12 + z*tm.a13 + tm.b1;
      v[i+1] = x*tm.a21 + y*tm.a22 + z*tm.a23 + tm.b2;
      v[i+2] = x*tm.a31 + y*tm.a32 + z*tm.a33 + tm.b3;
    }
  }
}
