package nxm.sys.lib;

import java.net.*;
import nxm.sys.inc.ShellParent;

/** ShellParent implementation that is used when running within a WebStart
    context.  The methods provide the particulars of running from within this
    context such as getting the list of option trees and the directories.

    @author J. Czechowski IV
    @version $Id: Webstart.java,v 1.21 2012/08/24 15:02:04 ntn Exp $
*/
public class Webstart implements ShellParent {
  private URL codebase = null; // The codebase where this information is served from.

  /** Command line interpreter for webstart applications. Instantiates a new
      instance of a Webstart and invokes the run method with the command line
      arguments specified in args.

      @param args Any command-line arguments.
  */
  public static void main (String[] args) {
    Shell.runNew ( new Webstart(), StringUtil.join(args," "), false);
    System.exit(0);
  }

  /** Initialize a default instance of the object. */
  public Webstart () {
    initCodebase();
  }

  /** Initialize the codebase attribute of the application.  In order to avoid
      having to have the jnlp.jar in the classpath while compiling we'll use
      dynamic method invocation to retrieve the values.  Upon returning from
      this method, the codebase member will be set.
   */
  protected void initCodebase () {
    codebase = null;

    try {
      Object basicSvc = KeyObject.invokeStaticMethod(
          "javax.jnlp.ServiceManager", "lookup",
          new Object[] { "javax.jnlp.BasicService" }
        );

      codebase = (URL)KeyObject.invokeInstanceMethod(
        basicSvc, "getCodeBase", new Object[0]
        );

    } catch (Exception e) {
      e.printStackTrace(); // Can't use normal Midas exception handling since it is not initialized yet
    }
  }

  /** Return the arguments to the AUX command for configuring the
      application's aux path. The values returned by this method are retrieved
      from the AUX parameter specified in the JNLP file for the application. If
      that parameter does not exist then a default aux path is returned.
      @return The arguments to the AUX command for configuring the aux path.
   */
  public String getAux () {
    final String defAux = "HOME HOME|RAM|DAT";
    String aux = System.getProperty("AUX");
    if (aux==null) aux = defAux;
    return aux;
  }

  /** Return the URL to the codebase that this application is served from.
      This is the full URL including the protocol, hostname, and path.
      @return The URL that this web application was initially served from.
   */
  public URL getCodeBase () {
    return codebase;
  }

  /** Return the value of NMROOT for this application.  For Webstart deployed
      applications this will always be jar:nxm-sys.jar
   */
  public String getNmRoot () {
    String nmroot = OptionTree.findOptRoot("SYS");
    return nmroot;
  }

  /** Return the option trees that are to be loaded by the Shell on startup.
      The list of these is taken from the OPTS JNLP property. The String is a
      list of comma separated option tree names.
      @return A comma separated list of option trees to configure on startup
              or null if there are none.
   */
  public String getOpts () {
    String opts = Shell.getProperty("OPTS");
    return OptionTree.getOptsHelper(opts, this, true);
  }

  /** Return the current working directory.
      This is simply the full codebase minus any query string that may exist.
      @return The current working directory.
   */
  public String getCWD () {
    String file = codebase.toString();
    int i = file.indexOf("?");
    if (i>=0) file = file.substring(0,i);
    return file;
  }

  public String getNMS () {
    // return null; // prior to 3.1.3: TODO: cleanup in 3.3.x or later beta series
    return Shell.getProperty("NMSTARTUP.USER"); // since 3.1.3 as part of BUG 1447
  }

  /** Return the value that should be used for the home directory
      @return null - Tells Shell to set this the "normal" way.
   */
  public String getHome () {
    return null;
  }

  /** Return the hostname where the execution of this Shell originated.  The
      hostname is determined simply from the host property of the codebase.
      @return The name of the origin host where this code is served from.
   */
  public String getServer () {
    return codebase.getHost();
  }

  /** Gets theme based on "ENV.THEME" system property, otherwise returns null
      to use system defaults.
      @since NeXtMidas 2.9.3
   */
  public String getTheme () {
    return Shell.getProperty("ENV.THEME");
  }

  /** Gets style based on "ENV.STYLE" system property, otherwise returns null
      to use system defaults.
      @since NeXtMidas 2.9.3
   */
  public String getStyle () {
    return Shell.getProperty("ENV.STYLE");
  }

  /**  @since NeXtMidas 3.3.0 */
  public Terminal getTerminal () {
    return null; // return null to let Shell determine terminal type
  }

  /** Return a string describing the object.
      @return A string describing the object.
   */
  public String toString () {
    return "Webstart\n" +
      "\tcodebase : " + getCodeBase() + "\n" +
      "\tcwd      : " + getCWD();
  }
}
