package nxm.sys.lib;

import nxm.sys.inc.*;
import java.util.Stack;
import java.util.regex.*;

/** Defines XML functions useful throughout NeXtMidas and NeXtMidas applications.

    @author  J. Czechowski IV
    @version $Id: XmlFile.java,v 1.29 2012/01/20 22:17:59 ntn Exp $
    @since   NeXtMidas 2.1.2
 */
public class XmlFile extends BaseFile {
  private static final Pattern XML_LT  = Pattern.compile("&lt;");  // Matches "&lt;" in the XML
  private static final Pattern XML_GT  = Pattern.compile("&gt;");  // Matches "&gt;" in the XML
  private static final Pattern XML_AMP = Pattern.compile("&amp;"); // Matches "&amp;" in the XML

  /** Opens a given text file and then reads it in as an XML file converting the XML to a Table.
      This is the same as <code>fastXmlFileToTable(new TextFile(ref,url))</code>.
      @param ref The reference to use when opening the {@link TextFile} (usually a Midas instance).
      @param url The URL of the file.
      @return The table.
   */
  public static Table fastXmlFileToTable(Object ref, String url) {
    return fastXmlFileToTable(new TextFile(ref,url));
  }

  /** Opens a given text file and then reads it in as an XML file converting the XML to a Table.
      This is the same as <code>fastXmlFileToTable(tf,null,null,1,true)</code>.
      @param tf The TextFile to read from.
      @return The table.
   */
  public static Table fastXmlFileToTable(TextFile tf) {
    return fastXmlFileToTable(tf,null,null,1,true);
  }

  /** Opens a given text file and then reads it in as an XML file converting the XML to a Table.
      @param tf       The TextFile to read from.
      @param delim    The delimiter between tag name and tag count (null=useDefault).
      @param cname    The name for the character data blocks (null=useDefault).
      @param start    The index of the first tag (start<0 is special case).
      @param comp     Compress individual text records.
      @return The table.
   */
  public static Table fastXmlFileToTable(TextFile tf, String delim, String cname, int start, boolean comp) {
    tf.open(TextFile.INPUT|TextFile.FORCEABORT);
    String xml = tf.readAll();
    tf.close();
    return fastXmlToTable(xml, delim,cname, start, comp);
  }

  /** Quickly converts a block of XML to a Table.
      @param xml      The XML packed in a string.
      @param delim    The delimiter between tag name and tag count (null=useDefault).
      @param cname    The name for the character data blocks (null=useDefault).
      @param start    The index of the first tag (start<0 is special case).
      @param comp     Compress individual text records.
      @return The table.
   */
  public static Table fastXmlToTable(String xml, String delim, String cname, int start, boolean comp) {
    XmlToTableHandler handler = new XmlToTableHandler(delim,cname,start,comp);
    parseXmlFast(xml, handler);

    return handler.getTable();
  }


  /** Parses an XML string. This is a "fast" parser in that it does little or no error checking
      and does no XML validation. If an error occurs it will either go undetected or will cause
      an exception (of some form). This parser is intended for use in situations where the validity
      of the XML is guaranteed and speed is essential.
      @param xml     The raw XML.
      @param handler Handler that will be called to handle any tags/text found.
   */
  public static void parseXmlFast(String xml, SimpleXmlHandler handler) {
    int     cdataStart = 0;
    boolean upcaseTags = false;
    boolean upcaseKeys = false;
    char[]  chars      = xml.toCharArray(); // Using chars[i] is faster than using xml.charAt(i)
    int     tableMode  = Table.KV;

    if (handler instanceof NotSoSimpleXmlHandler) {
      upcaseTags = ((NotSoSimpleXmlHandler)handler).getUpcaseAllTags();
      upcaseKeys = ((NotSoSimpleXmlHandler)handler).getUpcaseAllKeys();
      tableMode  = ((NotSoSimpleXmlHandler)handler).getTableMode();
    }

    for(int i = 0; i < xml.length(); i++) {
      char thisChar = chars[i];

      if (thisChar == '<') {
        char nextChar = chars[i+1];

        if (cdataStart != i) {
          String cdata = xml.substring(cdataStart,i).trim();

          if (cdata.length() > 0) {
            handler.gotCharData(removeEscapeChars(cdata));
          }
        }

        if (nextChar == '/') {
          i+=2;
          int tagStart = i;
          i++;
          while (chars[i] != '>') { i++; }
          String tag = xml.substring(tagStart,i);
          if (upcaseTags) tag = toUpper(tag);
          handler.gotCloseTag(tag);

        } else if (nextChar == '!') {
          if ((i+6 < xml.length()) && (chars[i+2] == '-') && (chars[i+3] == '-')) {
            // An XML comment "<!-- ... -->" block
            i+=3;
            while ((chars[i]!='-') || (chars[i+1]!='-') || (chars[i+2]!='>')) { i++; }
            i+=2;

          } else if ((i+9 < xml.length()) && xml.substring(i,i+9).equals("<![CDATA[")) {
            // A CDATA section ("<![CDATA[ ... ]]>") that contains no XML or escape characters
            i+=9;
            cdataStart = i;
            while ((chars[i]!=']') || (chars[i+1]!=']') || (chars[i+2]!='>')) { i++; }
            handler.gotCharData(xml.substring(cdataStart,i));
            i+=2;

          } else if ((i+9 < xml.length()) && xml.substring(i,i+9).equals("<!DOCTYPE")) {
            // XML Document Type Definition ("<!DOCTYPE ... >")
            while ((chars[i] != '>')) {
              if (chars[i] == '[') { // Allows embedded DTD declarations inside the brackets.
                while ((chars[i] != ']')) { i++; }
              }
              i++;
            }
          } else {
            throw new RuntimeException("Error unknown tag beginning with '<!'.");
          }

        } else if (nextChar == '?') {
          // An XML Processing Instruction "<? ... ?>"
          i+=2;
          while ((chars[i]!='?') || (chars[i+1]!='>')) { i++; }
          i++;

        } else {
          i++;
          int tagStart = i;
          while (isTag(chars[i])) { i++; }
          String  tag  = xml.substring(tagStart,i);
          boolean end  = false;
          boolean done = false;
          Table   args = new Table(tableMode, Table.ADD);

          while (!end && !done) {
            while (isWhiteSpace(chars[i])) { i++; }
            int keyStart = i;

            if (chars[i] == '/') {
              end = true;
              i++;
            } else if (chars[i] == '>') {
              done = true;
            } else {
              while (chars[i] != '=') { i++; }
              String key = xml.substring(keyStart,i);
              i++;
              char tic = chars[i]; // Either tic (') or quote (")
              i++;
              int valStart = i;
              while (chars[i] != tic) { i++; }
              String val = xml.substring(valStart,i);
              i++;
              if (upcaseKeys) key = toUpper(key);
              args.putx(key, val, 0); // this is faster than args.put(key,val) and works in this simple case
            }
          }
          args.setFlags(0);
          if (upcaseTags) tag = toUpper(tag);
          handler.gotOpenTag(tag, args);
          if (end) { handler.gotCloseTag(tag); }
        }
        cdataStart = i+1;
      } else if ((cdataStart == i) && (isWhiteSpace(thisChar))) {
        cdataStart = i+1;
      }
    }
  }

  /** Fast toUpperCase(..) method for ASCII only. */
  private static String toUpper (String str) {
    char[]  chars = str.toCharArray();
    boolean mod   = false;

    for (int i = 0; i < chars.length; i++) {
      char c = chars[i];
      if ((c >= 'a') && (c <= 'z')) {
        chars[i] = (char)(c - 'a' + 'A');
        mod = true;
      }
    }
    return (mod)? new String(chars) : str;
  }

  /** Indicates if whitespace. Only used by parseXmlFast(String). */
  private static boolean isWhiteSpace(char c) {
    return (c == ' ') || (c == '\t') || (c == '\n');
  }

  /** Indicates if not part of a tag name. Only used by parseXmlFast(String). */
  private static boolean isTag(char c) {
    return (!isWhiteSpace(c)) && (c != '/') && (c != '>');
  }

  /** Removes escape characters from the XML. */
  private static String removeEscapeChars(String text) {
    // JC4: Pre-compiling the pattern is faster... especially when this is called often.
    //return text.replaceAll("&lt;",  "<")
    //           .replaceAll("&gt;",  ">")
    //           .replaceAll("&amp;", "&");
    text = XML_LT.matcher(text).replaceAll("<");
    text = XML_GT.matcher(text).replaceAll(">");
    text = XML_AMP.matcher(text).replaceAll("&");
    return text;
  }

  /** Converts a NeXtMidas table to XML. This is the same as
      <code>fastTableToXml(tab,null,null,null,false,newlines)</code>.
      @param tab      The table.
      @param newlines Should the XML be generated in a more readable format (newlines and
                      indentation), if this is false no extra whitespace will be added.
      @return The XML string.
      @since NeXtMidas 2.5.0
   */
  public static String fastTableToXml(Keyable tab, boolean newlines) {
    return fastTableToXml(tab,null,null,null,false,newlines);
  }

  /** Converts a NeXtMidas table to XML. This assumes that the table is using a similar mapping to
      XML that {@link #fastXmlToTable(String,String,String,int,boolean)} uses. This will just throw
      an exception is something looks wrong. Note that any entries that are {@link Keyable} will be
      expanded as sub-tables, everything else will be converted to a string.
      @param tab      The table.
      @param outerTag The desired outer XML tag (null = use default)
      @param delim    The delimiter used for duplicate tags (null = use default).
      @param cname    The name to use for CDATA sections (null = use default).
      @param compress Have CDATA sections been compressed (if true, it is not possible to go back,
                      all table entries are assumed to be attributes on the XML tags).
      @param newlines Should the XML be generated in a more readable format (newlines and
                      indentation), if this is false no extra whitespace will be added.
      @return The XML string.
      @since NeXtMidas 2.5.0
   */
  public static String fastTableToXml(Keyable tab, String outerTag, String delim,
                                      String cname, boolean compress, boolean newlines) {
    if (outerTag == null) outerTag = "XML";

    TableToXml tableToXml = new TableToXml(delim, cname, compress, newlines);

    return tableToXml.toXml(outerTag, tab);
  }

  /** Opens a given text file and then writes a table to it as XML. This is the same as
      <code>fastTableToXmlFile(new TextFile(ref,url),tab,newlines)</code>.
      @param ref The reference to use when opening the {@link TextFile} (usually a Midas instance).
      @param url The URL of the file.
      @param tab      The table.
      @param newlines Should the XML be generated in a more readable format (newlines and
                      indentation), if this is false no extra whitespace will be added.
      @since NeXtMidas 2.5.0
   */
  public static void fastTableToXmlFile(Object ref, String url, Keyable tab, boolean newlines) {
    fastTableToXmlFile(new TextFile(ref,url),tab,newlines);
  }

  /** Opens a given text file and then writes a table to it as XML. This is the same as
      <code>fastTableToXmlFile(tf,tab,null,null,null,false,newlines)</code>.
      @param tf       The TextFile to write to.
      @param tab      The table.
      @param newlines Should the XML be generated in a more readable format (newlines and
                      indentation), if this is false no extra whitespace will be added.
      @since NeXtMidas 2.5.0
   */
  public static void fastTableToXmlFile(TextFile tf, Keyable tab, boolean newlines) {
    fastTableToXmlFile(tf,tab,null,null,null,false,newlines);
  }

  /** Opens a given text file and then writes a table to it as XML.
      @param tf       The TextFile to write to.
      @param tab      The table.
      @param outerTag The desired outer XML tag (null = use default)
      @param delim    The delimiter used for duplicate tags (null = use default).
      @param cname    The name to use for CDATA sections (null = use default).
      @param compress Have CDATA sections been compressed (if true, it is not possible to go back,
                      all table entries are assumed to be attributes on the XML tags).
      @param newlines Should the XML be generated in a more readable format (newlines and
                      indentation), if this is false no extra whitespace will be added.
      @since NeXtMidas 2.5.0
   */
  public static void fastTableToXmlFile(TextFile tf, Keyable tab, String outerTag, String delim,
                                        String cname, boolean compress, boolean newlines) {
    tf.open(TextFile.OUTPUT|TextFile.FORCEABORT);
    tf.write(fastTableToXml(tab, outerTag, delim, cname, compress, newlines));
    tf.close();
  }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Nested Classes
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** Internal class used to convert a table to XML. */
  private static final class TableToXml {
    private final String INDENT = "  ";

    private StringBuffer xml;
    private String       delim;
    private String       cname;
    private boolean      compress;
    private String       newline;

    public TableToXml(String delim, String cname, boolean compress, boolean newlines) {
      this.xml = new StringBuffer();
      this.delim = (delim != null) ? delim: "_";
      this.cname = (cname != null) ? cname: "CDATA";
      this.compress = compress;
      this.newline  = (newlines)? "\n" : "";
    }

    /** COnverts a table to XML. */
    public String toXml(String outerTag, Keyable tab) {
      return toXml(outerTag, tab, "").toString();
    }

    /** Recursive method... only called internally. */
    private StringBuffer toXml(String topLevel, Keyable tab, String indent) {
      StringBuffer xml       = new StringBuffer(); // The body xml.
      StringBuffer attr      = new StringBuffer(); // The attributes in the open tag.
      String[]     keys      = tab.getKeys();      // Table entries
      String       indt1     = "";                 // First-level indent
      String       indt2     = "";                 // Second-level indent
      boolean      hasNested = false;              // Are there nested keys?

      if (newline.length() > 0) {
        indt1 = newline + indent;
        indt2 = indt1 + INDENT;
      }
      attr.append(indt1).append("<").append(topLevel); // Add in "<topLevel" to front of attrs.
      int lenDelim = delim.length();
      for (int i = 0; i < keys.length; i++) {
        String key   = keys[i];
        Object value = tab.getKey(key); // Do this BEFORE stripping off the delim
        int idelim = key.lastIndexOf(delim);
        if (idelim > 0) { // Strip off the one-up counter necessary for a table, but not xml
          String count = key.substring(idelim+lenDelim);
          if (StringUtil.isInteger(count)) {
            key = key.substring(0,idelim);
          } else {
            throw new MidasException("XmlFile: Error converting table key '"+keys[i]+"' to XML, "
                                   + "when using delim='"+delim+"', invalid count="+count);
          }
        }

        if (value == null) {
          // do nothing
        } else if (key.equals(cname) && !compress) {
          value = StringUtil.toHTML(value.toString()); // Change '<' to '&lt;', '>' to '&gt;', etc.
          xml.append(indt2).append(value);

        } else {
          if (value instanceof Keyable) {
            xml.append(toXml(key, (Keyable)value, indent+INDENT));
            hasNested = true;
          } else {
            // Append the tag value: <tag>='<value>' to the top tag.
            attr.append(" ").append(key).append("=\"").append(value).append("\"");
          }

        }
      }

      if (xml.length() == 0) {
        attr.append("/>");
      } else if (!hasNested && (xml.length() < 40)) { // Tighten up when xml is just a small CDATA
        attr.append(">");
        attr.append(xml.substring(indt2.length()));
        attr.append("</").append(topLevel).append(">");
      } else {
        attr.append(">");
        attr.append(xml);
        attr.append(indt1).append("</").append(topLevel).append(">");
      }
      return attr;
    }
  }

  /** Defines the interface the XML parser will use to process the XML tags. */
  public static interface SimpleXmlHandler {
    /** This method is called when the XML parser finds an open tag.
        @param tag  The name of the tag.
        @param args The arguments associated with the tag (if any).
     */
    public void gotOpenTag(String tag, Table args);

    /** This method is called when the XML parser finds a close tag.
        @param tag  The name of the tag.
     */
    public void gotCloseTag(String tag);

    /** This method is called when the parser finds character data.
        @param cdata The character data.
     */
    public void gotCharData(String cdata);
  }

  /** Defines the interface the XML parser will use to process the XML tags.
      @since NeXtMidas 2.5.5
   */
  public static abstract class NotSoSimpleXmlHandler implements SimpleXmlHandler {
    /** Indicates if all keys should be converted to uppercase before calling
        {@link #gotOpenTag} should be passed a table where all keys have been
        converted to upper case.
        @since NeXtMidas 2.5.5
     */
    public boolean getUpcaseAllKeys() {
      return false;
    }

    /** Indicates if the tags should be converted to uppercase before calling
        {@link #gotOpenTag} or {@link #gotCloseTag}.
        @since NeXtMidas 2.5.5
     */
    public boolean getUpcaseAllTags() {
      return false;
    }


    /** Indicates the table mode to use with the table passed to {@link #gotOpenTag}.
        @since NeXtMidas 2.5.5
     */
    public int getTableMode() {
      return Table.KV;
    }
  }

  /** An implementation of SimpleXmlHandler that fills in a {@link Table}. */
  public static class XmlToTableHandler extends NotSoSimpleXmlHandler {
    private final Stack<Table> tables = new Stack<Table>(); // The stack of tables.
    private final Stack<Table> counts = new Stack<Table>(); // The stack of tag counts.
    private final Stack<String> names = new Stack<String>(); // The stack of names.
    private final String  delim;                  // Delimiter
    private final String  cname;                  // Name of CDATA field.
    private final int     start;                  // Start index (usually 1 or 0; -1 and -2 are special)
    private final boolean compress;               // Compress single entries?
    private final int     mode;                   // Table mode to use.
    private       String  cdata    = null;        // Any character data.
    private       Table   root     = null;        // Root table

    /** Creates a new instance using XmlToTableHandler(null,null,1,false,Table.KV). */
    public XmlToTableHandler() {
      this(null,null,1,false,Table.KV);
    }

    /** Creates a new instance using XmlToTableHandler(delim,cname,start,false,Table.KV).
        @param delim   The delimiter between tag name and tag count (null=useDefault).
        @param cname   The name for the character data blocks (null=useDefault).
        @param start   The index of the first tag.
     */
    public XmlToTableHandler(String delim, String cname, int start) {
      this(delim,cname,start,false,Table.KV);
    }

    /** Creates a new instance using XmlToTableHandler(delim,cname,start,compress,Table.KV).
        @param delim   The delimiter between tag name and tag count (null=useDefault).
        @param cname   The name for the character data blocks (null=useDefault).
        @param start   The index of the first tag.
     */
    public XmlToTableHandler(String delim, String cname, int start, boolean compress) {
      this(delim,cname,start,compress,Table.KV);
    }

    /** Creates a new instance. The <code>start</code> value indicates how duplicate elements are
        handled since all Table entries must be unique.
        <pre>
          start = -2    - 0-based counter (always use counter)
          start = -1    - First has no counter, others start with 1 (legacy
                          functionality that matches old XML2TBL conversions)
          start =  0    - 0-based counter (use only if duplicates are found)
          start =  1    - 1-based counter (use only if duplicates are found)
          start =  N    - N-based counter (where N > 0)
        </pre>
        Compression will take any single record with just character data and compress it. For
        example:
        <pre>
          &lt;XML&gt;
            &lt;greeting&gt;Hello&lt;/greeting&gt;
          &lt;/XML&gt;

          Without Compression:  {GREETING={CDATA="Hello"}}
          With Compression:     {GREETING="Hello"}
        </pre>
        @param delim    The delimiter between tag name and tag count (null=useDefault).
        @param cname    The name for the character data blocks (null=useDefault).
        @param start    The index of the first tag.
        @param compress Compress individual text records?
        @param mode     The default mode to use for tables ({@link Table#KV} or {@link Table#HT}).
        @since NeXtMidas 2.5.5
     */
    public XmlToTableHandler(String delim, String cname, int start, boolean compress, int mode) {
      this.delim    = (delim == null)? "_": delim;
      this.cname    = (cname == null)? "CDATA": cname.toUpperCase();
      this.start    = start;
      this.compress = compress;
      this.mode     = mode;
    }

    public boolean getUpcaseAllKeys () { return true; }
    public boolean getUpcaseAllTags () { return true; }
    public int     getTableMode ()     { return mode; }

    /** Clears and resets the handler so it can be used again. JC4: I added this method
        since I was parsing hundreds of XML messages a second and wanted to just "clear"
        the previous run and reuse the same handler (to eliminate the extra object creation
        time).
        @since NeXtMidas 2.3.0
     */
    public void clear() {
      tables.clear();
      counts.clear();
      names.clear();
      cdata = null;
      root  = null;
    }

    public void gotOpenTag(String tag, Table args) {
      if (!tables.empty()) tag = addUnique(tag, args); // Add if not first in file.

      tables.push(args);
      counts.push(new Table(Table.HT));
      names.push(tag);
    }

    public void gotCloseTag(String tag) {
      Table  top  = tables.pop();
      Table  cnt  = counts.pop();
      String name = names.pop().toString();

      if (compress && (top.size() == 1)) { // If only one entry, compress it from Table to record.
        String key = top.getKeys()[0];
        Object val = top.get(key);

        if (key.equals(cname)) {
          if (!tables.empty()) {
            Table prev = tables.peek();
            prev.remove(name);
            prev.put(name, val);
          } else {
            top.rename(top.getKeys()[0], name);
          }
        }
      }
      if (tables.empty()) { root = top; } // Only occurs at end of file.
    }

    public void gotCharData(String cdata) {
      addUnique(cname, cdata);
    }

    /** Returns the table. This should only be called after XML parsing is done. */
    public Table getTable() {
      return root;
    }

    /** Adds a unique entry to the table on the top of the stack.
        @param key The key.
        @param val The value.
        @return The unique key used.
     */
    private String addUnique(String key, Object val) {
      Table  table = tables.peek();
      Table  count = counts.peek();
      int    index = count.getL(key, start);
      String ukey  = key + delim + index;

      if (index == -1) {               // Used by XML2TBL in CONVERT (no counter on first one)
        ukey = key;
        index = 0;
      } else if (index == -2) {        // Always add 0-based counter
        ukey  = key + delim + 0;
        index = 0;
      } else if (index == start) {     // Don't add one-up-counter until there are duplicates
        ukey = key;
      } else if (index == (start+1)) { // Found first duplicate, rename old one.
        table.rename(key, (key + delim + start));
      }

      count.put(key,  index+1);
      table.put(ukey, val);

      return ukey;
    }
  }
}
