package nxm.sys.libg;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.image.ImageObserver;
import java.awt.image.BufferedImage;
import java.awt.image.ImageProducer;
import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.ArrayList;
import javax.imageio.ImageIO;

/** This class provides utility methods designed to help in exporting an image.
    The following shows how to export an image from an {@link MPlot} using
    either graphical prompts displayed in the MPlot window (which is an
    extension of {@link MWindow}) or text-based prompts displayed in the
    terminal:

    <table border="1">
      <tr>
        <td>Graphical (GUI) Prompts:
            <pre>
              MPlot mplot;
              ...
              ExportImage.exportFile(mplot, mplot);
            </pre>
        </td>
      </tr>
      <tr>
        <td>Terminal (Text-Only) Prompts:
            <pre>
            MPlot    mplot;
            Terminal term;
            ...
            ExportImage.exportFile(mplot, term);
            </pre>
        </td>
      </tr>
    </table>

    @see ExportsImage
    @see MPlot
    @see Terminal
    @see MWindow
    @since   NeXtMidas Version 1.7.5
    @author  J. Czechowski IV
    @version $Id: ExportImage.java,v 1.7 2012/01/24 16:56:31 ntn Exp $
*/
public class ExportImage extends Image {
  private static final String[] SUGGESTED_TYPES = { "png", "jpeg", "jpg", "gif" }; // Suggested types (in order)
  private static final String   DEFAULT_TYPE    = SUGGESTED_TYPES[0];

  /** This is the underlying image that is being created. */
  protected BufferedImage image;

  /** This will create a new imager to use to create an image. This will use
      the default image type of {@link BufferedImage#TYPE_INT_RGB}.
    @param width  The width of the image.
    @param height The height of the image.
  */
  public ExportImage(int width, int height) {
    this(width, height, BufferedImage.TYPE_INT_RGB);
  }

  /** This will create a new imager to use to create an image.
    @param width     The width of the image.
    @param height    The height of the image.
    @param imageType The type of image to create (see {@link BufferedImage#BufferedImage}).
  */
  public ExportImage(int width, int height, int imageType) {
    this.image = new BufferedImage(width, height, imageType);
  }

  /** Gets the underlying {@link BufferedImage} that is represented by this class.
      @return The image that is the core image of this class.
  */
  public BufferedImage getBufferedImage() {
    return this.image;
  }

  /** This will get the list of the file types that can be written to. Any of
      the elements of this array would be a valid <code>fileType</code> to
      pass to {@link #writeToFile(String,String)}.
      @return The file types that can be written to.
      @see #writeToFile(String,String)
  */
  public static String[] getFileTypes() {
    return getFileTypes(false);
  }

  /** This will get the list of the file types that can be written to. Any of
      the elements of this array would be a valid <code>fileType</code> to
      pass to {@link #writeToFile(String,String)}.
      @param lowercaseOnly Only include lower-case versions of names. When this is false, it is
                           common to see many names that differ only in case (e.g. "JPG" vs "jpg").
      @return The file types that can be written to.
      @see #writeToFile(String,String)
      @since NeXtMidas 2.5.5
  */
  public static String[] getFileTypes (boolean lowercaseOnly) {
    String[] types = ImageIO.getWriterFormatNames();
    if (lowercaseOnly) {
      ArrayList<String> lower = new ArrayList<String>();
      for (String ty : types) {
        String lc = ty.toLowerCase();

        if (ty.equals(lc)) lower.add(ty);
      }
      types = lower.toArray(StringUtil.EMPTY_STRING_ARRAY);
    }
    return types;
  }

  /** This will get a suggest a type to use.
      @return A suggested type from {@link #getFileTypes()} or null if none found.
      @since NeXtMidas 2.5.5
   */
  public static String getSuggestedType() {
    String[] types = getFileTypes();
    Arrays.sort(types);

    for (String suggestedType : SUGGESTED_TYPES) {
      if (Arrays.binarySearch(types, suggestedType) >= 0) return suggestedType;
    }
    return (types.length == 0)? null : types[0];
  }


  /** This will write the image to a file. The extension of the file is taken
      from the filename. If there is no extension on the filename, the default
      is used from {@link #DEFAULT_TYPE}.
      @param fileName The name of the output file.
      @throws IOException
      @see #getFileTypes()
   */
  public void writeToFile (String fileName) throws IOException {

    if (fileName.indexOf(".") < 0) {
      Shell.warning("ExportImage: No file type or extension provided, assuming default of "+DEFAULT_TYPE);
    }

    writeToFile(fileName, DEFAULT_TYPE);
  }

  /** This will write the image to a file.<br>
      Note 1: If an extension is desired on the file, that extension must be
      already added to the file name given.<br>
      Note 2: If the given file type does not match the extension provided on
      the file name, the file type will override the extension on the name,
      since it is explicit.
      @param fileName The name of the output file.
      @param fileType The file type to use for the image file (this is usually
                      the same as the extension of the fileName).
      @throws IOException If there is an error in writing the file.
      @see #getFileTypes()
  */
  public void writeToFile(String fileName, String fileType) throws IOException {

    if (fileName.indexOf(".") < 0) {} // do nothing (allow for intentional "blank" extension
    else {
      String ext = fileName.substring(fileName.indexOf("."));
      if (!ext.substring(1).equalsIgnoreCase(fileType)) {
        // remove extension from filename and use passed file type
        Shell.warning("Extension in file name ("+ext+") does not match given file type ("
            +fileType+"). Using "+fileType);
      }
    }

    if (!ImageIO.write(this.image, fileType, new File(fileName))) {
      throw new MidasException("Could not find image writer for "
                                 + "file type (" + fileType + ").");
    }
  }

  /** This will export a Component to an image file. For example, to create a
      <code>.jpeg</code> file from an existing {@link javax.swing.JButton},
      named <code>okButton</code>, the following would be used:
      <pre>
          exportToFile(okButton, "button.jpeg", "jpeg")
      </pre>
      Note: <i>This method may fail to export the entire image if the
            {@link Component} uses complex paint methods.</i>

      @param src      The component to make an image of.
      @param fileName The name of the output file.
      @param fileType The file type to use for the image file (this is usually
                      the same as the extension of the fileName).

      @throws IOException If there is an error in writing the file.
      @see #writeToFile(String,String)
      @see #getFileTypes()
      @see #export(Component)
  */
  public static void exportToFile (Component src,String fileName,String fileType) throws IOException {
    export(src).writeToFile(fileName, fileType);
  }

  /** This will export a Component to an image. For example, to create an
          image from an existing {@link javax.swing.JButton}, named
          <code>okButton</code>, the following would be used:
          <pre>
              ExportImage image = ExportImage.export(okButton);
          </pre>
          Note: <i>This method may fail to export the entire image if the
                {@link Component} uses complex paint methods.</i>
    @param src The component to make an image of.
    @see #export(ExportsImage)
    @return the exported ExportImage
  */
  public static ExportImage export(Component src) {
    ExportImage compImgr = new ExportImage(src.getWidth(), src.getHeight());
    src.print(compImgr.getGraphics());
    return compImgr;
  }

  public void flush() {
    this.image.flush();
  }

  public Graphics getGraphics() {
    return this.image.getGraphics();
  }

  public int getHeight(ImageObserver observer) {
    return this.image.getHeight(observer);
  }

  public Object getProperty(String name, ImageObserver observer) {
    return this.image.getProperty(name, observer);
  }

  public ImageProducer getSource() {
    return this.image.getSource();
  }

  public int getWidth(ImageObserver observer) {
    return this.image.getWidth(observer);
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Deprecated since NeXtMidas 2.5.5
  //////////////////////////////////////////////////////////////////////////////////////////////////

  /** @deprecated Since NeXtMidas 2.5.5: Use {@link #exportToFile(Component,String,String)} */
  @Deprecated public static void exportFile(final ExportsImage export, final MWindow window) {
    final String[] fileTypes = getFileTypes();
    (new Thread() {
      public void run() {
        String fileName;
        String fileType;

        if (fileTypes.length == 0) {
          new GDialog(window, "Can not export plot: Java can not find an"
                            + "appropriate file export library.", 0, 0, 0,null);

          return;
        }

        fileType = showOptionSet(window, "Export Format", fileTypes);
        fileName = showTextEntry(window, "Export File Name","myplot." + fileType.toLowerCase(), 32);
        try {
          /* Create the image and write it out to the specified file. */
          exportToFile(export, fileName, fileType);

        } catch(MidasException exception) {
          new GDialog(window, "Can not export plot: File type (" + fileType
                             + ") is not supported.", 0, 0, 0, null);

        } catch(java.io.IOException exception) {
          new GDialog(window, "Can not export plot: Error while writing "
                            + "file (" + fileName + ").", 0, 0, 0, null);

        }
      }
    }).start();
  }

  /** @deprecated Since NeXtMidas 2.5.5: Use {@link #exportToFile(Component,String,String)} */
  @Deprecated public static void exportFile(final ExportsImage export, final Terminal terminal) {
    final String[] fileTypes = getFileTypes();

    (new Thread() {
      public void run() {
        String fileName;
        String fileType;

        if (fileTypes.length == 0) {
          terminal.M.error("Can not export plot: Java can not find "
                         + "appropriate file export library.");
          return;
        }

        StringBuffer availableTypes = new StringBuffer();
        availableTypes.append("File types available for export: ");
        availableTypes.append(fileTypes[0]);

        for(int i = 1; i < fileTypes.length; i++) {
          availableTypes.append(", ");
          availableTypes.append(fileTypes[i]);
        }


        terminal.M.info(availableTypes.toString());
        fileType = terminal.getInput("File Type: ", fileTypes[0]);

        fileName = terminal.getInput("File Name: ",
                                     "myplot." + fileType.toLowerCase());

        try {
          /* Create the image and write it out to the specified file. */
          exportToFile(export, fileName, fileType);

        } catch(Exception exception) {
          // This line discards any nested exception and makes debugging difficult.
          // terminal.M.error("Can not export plot: File type " + fileType
          //               +" name "+ fileName);
          throw new MidasException("Can not export plot: File type=" + fileType +
              " name="+ fileName, exception);
        }
      }
    }).start();
  }

  /** @deprecated Since NeXtMidas 2.5.5: Use {@link #exportToFile(Component,String,String)} */
  @Deprecated public static void exportToFile(ExportsImage src, String fileName, String fileType) throws IOException {
    export(src).writeToFile(fileName, fileType);
  }

  /** @deprecated Since NeXtMidas 2.5.5: Use {@link #export(Component)} */
  @Deprecated public static ExportImage export(ExportsImage src) {
    ExportImage compImgr = new ExportImage(src.getExportWidth(), src.getExportHeight());
    src.exportTo(compImgr.getGraphics()); // Paint the component to the image same as display

    return compImgr;
  }

  /** @deprecated Since NeXtMidas 2.5.5: This class is no longer required. */
  @Deprecated private static class DelayedMessageHandler implements MessageHandler {
    /** This is the message. */
    private Message message;

    /** This indicates if the message has been set yet. */
    private boolean messageSet;

    /** Creates a new instance and sets the message set indicator to false. */
    public DelayedMessageHandler() {
        this.message    = null;
        this.messageSet = false;
    }

    public synchronized int processMessage (Message msg) {
        this.message    = msg;
        this.messageSet = true;

        return 1;
    }

    public Message getMessage() {
      while(!this.messageSet) {
        Time.sleep(0.5);
      }
      return this.message;
    }
  }

  /** @deprecated Since NeXtMidas 2.5.5: This method is no longer required and will be removed. */
  @Deprecated public static String showOptionSet(Object parent, String title, String options) {
    DelayedMessageHandler msgHandler = new DelayedMessageHandler();

    new GMenu(parent, title,options, 0, 0, msgHandler);

    return msgHandler.getMessage().getData().toString();
  }

  /** @deprecated Since NeXtMidas 2.5.5: This method is no longer required and will be removed. */
  @Deprecated public static String showOptionSet(Object parent, String title, String[] options) {
    StringBuffer optList = new StringBuffer(32);

    for (int i = 0; i < options.length; i++) {
      if (i != 0) {
        optList.append(",");
      }
      optList.append(options[i]);
    }

    return showOptionSet(parent, title, optList.toString());
  }

  /** @deprecated Since NeXtMidas 2.5.5: This method is no longer required and will be removed. */
  @Deprecated public static String showTextEntry(Object parent, String title, String text, int length) {
    DelayedMessageHandler msgHandler = new DelayedMessageHandler();

    new GPrompt(parent, title, text, length, 0, msgHandler);

    return msgHandler.getMessage().getData().toString();
  }
}
