package nxm.sys.libg;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

import java.awt.Graphics;
import java.lang.Math;

/** A class to hold and plot a feature. Note, a feature can be of multiple types because
    of the type option masking. Hence a feature can be a box, circle, ellipse, and so on.
    <br>
    <br>
    There are two modes for determining the position of the feature:
    <pre>
      MODE 1: dt &lt; 0 (default)
        Feature is plotted based on center point (x,y,z) and
        extents (dx,dy,dz). The point (x,y,z) is based on the
        relative coordinates of the plot (for static plots and
        the held-value in real time plots the relative and the
        absolute values are often the same -- this is not true
        for real-time raster plots after the first pass).

        After the feature has been plotted it will remain in
        the same position even after the plot has "wrapped"
        (in the case of dynamic plots).

      MODE 2: dt &gt;= 0
        Feature is plotted based on center point (x,y',z) and
        extents (dx,dt,dz). Where y' is derived based on absolute
        time position t. (Note that dy and dt are both relative
        time positions and thus dy = dt.)

        A feature will only be plotted within when within the
        time range of the plot. After the plot wraps the
        feature will be plotted over.
    </pre>

    @author Jeff Schoen
    @version $Id: Feature.java,v 1.94 2012/02/02 14:36:51 jwb Exp $
*/
public class Feature extends Object implements MessageHandler,Chainable,Drawable {
  private static final boolean FAST_DT2DY = true; // Fast dy=dt conversion (true in NXM 2.4.1)

  /** List of type of options supported. */
  public static final String typeList = "Symbol,Data,HLine,VLine,Box,Circ,Oval,Elps,"
                                      + "Ltag,Rtag,Ttag,Btag,Dtag,Peak,IPlot,VRide,"
                                      + "OSymbol,Text,Inverse,Tri,Filled,XLine,YLine";

  /** Type Option: Draw a fixed-size symbol at (x,y). */
  public static final int SYMBOL  = 0x000001;
  /** Type Option: Highlight enclosed data.           */
  public static final int DATA    = 0x000002;
  /** Type Option: Draw a horizontal line.            */
  public static final int HLINE   = 0x000004;
  /** Type Option: Draw a vertical line.              */
  public static final int VLINE   = 0x000008;
  /** Type Option: Draw a box centered at (x,y).      */
  public static final int BOX     = 0x000010;
  /** Type Option: Draw a circle centered at (x,y).   */
  public static final int CIRC    = 0x000020;
  /** Type Option: Draw an oval centered at (x,y).    */
  public static final int OVAL    = 0x000040;
  /** Type Option: Draw a geodetic ellipse.           */
  public static final int ELPS    = 0x000080;
  /** Type Option: Show text/tag along left axis.    */
  public static final int LTAG    = 0x000100;
  /** Type Option: Show text/tag along right axis.   */
  public static final int RTAG    = 0x000200;
  /** Type Option: Show text/tag along top axis.     */
  public static final int TTAG    = 0x000400;
  /** Type Option: Show text/tag along bottom axis.  */
  public static final int BTAG    = 0x000800;
  /** Type Option: NOT USED.                         */
  public static final int DTAG    = 0x001000;
  /** Type Option: NOT USED.                         */
  public static final int PEAK    = 0x002000;
  /** Type Option: Show text/tag inside axis.        */
  public static final int IPLT    = 0x004000;
  /** Type Option: Vertical position follows data.   */
  public static final int VRIDE   = 0x008000;
  /** Type Option: Fixed size circle. (Overlay)      */
  public static final int OSYMBOL = 0x010000;
  /** Type Option: Show text.                        */
  public static final int TEXT    = 0x020000;
  /** Type Option: Use inverse coloring.             */
  public static final int INVERSE = 0x040000;
  /** Type Option: Fixed size triangle (DEPRECATED). */
  public static final int TRI     = 0x080000;
  /** Type Option: Fill box/circle/oval feature.     */
  public static final int FILLED  = 0x100000;
  /** Type Option: Feature that draws a line along the X Data points.     */
  public static final int XLINE   = 0x200000;
  /** Type Option: Feature that draws a line along the Y data points.     */
  public static final int YLINE   = 0x400000;

  /** @deprecated Since NeXtMidas 2.3.0, use TEXT    */
  @Deprecated public static final int LABEL   = TEXT;

  /** List of anchors supported. */
  public static final String anchorList = "Data,Screen,Pixel";

  /** Anchor Option: Data Coordinates */
  public static final int A_DATA    = 1;
  /** Anchor Option: Fraction of screen from top left */
  public static final int A_SCREEN  = 2;
  /** Anchor Option: Pixels from top left */
  public static final int A_PIXEL   = 3;

  /* The allowable gap/variation in dt during feature validation. */
  final double DT_GAP = 0.100;

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // these fields should be protected
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** Feature name (both for drawing on display and for referencing in MPlot's list of features). */
  protected String name;
  /** Enable flags. */
  protected  int enable;
  /** For internal usage by MPlot only; true if this feature needs to be refreshed.
      As of NextMidas 2.7.1, set to false since MPlot.addFeature() sets it to true.
   */
  protected boolean needRefresh=false;
  /** Line that hold attributes like color. */
  protected Line line;
  /** Plot to which this feature belongs. */
  protected MPlot MP;
  /** Deprecated since NeXtMidas 2.5.0: this is for internal usage only. */
  private Object ref;  // reference object used when opening files (can be Midas, Command, etc.)

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // these fields should be private
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** Layer to which this feature belongs. */
  private Layer layer;
  private int type;                  // mask holding all the types of this feature
  private int anchor;                // anchor option for this feature
  private String text;               // The text to go along with feature
  /** x, y, z are the Real World Coordinates (RWC) of the center point*/
  private double x, y, z, t, dx, dy, dz, dt=-1;
  private double semiMajor=10000;    // Ellipse Semi-major axis in meters
  private double semiMinor=1000;     // Ellipse Semi-minor axis in meters
  private double tilt=30;            // Ellipse tilt in degrees
  private MPoint pix = new MPoint();

  /** Creates a new feature. */
  public Feature() {
    super();
    type = SYMBOL;
    anchor = A_DATA;
    line = new Line(this);
    line.setSymbol("CROSS"); // default symbol type
    enable = 0xFB;           // 0XFB = |Global|Labels|Thin|Base|Zoom|FastBase|FastZoom|
  }

  /** Creates a new feature on the given layer.  <br>
      <b>Intended for internal use in Layer and its subclasses only.</b> This will
      create a Feature applicable to its given Layer. The constructor is public to
      facilitate use by user-defined Layers, but should not be used outside one of
      the Layer subclasses.<br>
      <br>
      Note that this constructor sets the Feature type to {@link #DATA} by default.
      This is backwards-compatible to support older versions of NeXtMidas, but most
      Layers will not want or need this set. This can be "fixed" by calling
      {@link #setType(int)} with a value of 0 right after calling the constructor
      or by using the ddefault {@link Layer#getFeature()} implementation.
      @param layer The Layer invoking this constructor
   */
  public Feature (Layer layer) {
    this();
    type = DATA;
    this.layer = layer;
    this.MP = layer.MP;
  }

  /** Creates a new feature based on a table of keys.  */
  public Feature (Table t) {
    this();
    fromTable(t);
  }

  public boolean getNeedRefresh () {
    return needRefresh || line.getNeedRefresh();
  }

  public void clearNeedRefresh () {
    needRefresh = false;
    line.clearNeedRefresh();
  }

  public void draw (int mode) {
    if (MP.gc==null) return;
    int ix,iy,ix1,iy1,ix2,iy2;
    double top=0.0;           // JMM: top might not be integral
    // Note:  x, y, z = Real World Coord.
    //       ix,iy,iz = Pixel Positions (on screen).
    clearNeedRefresh();
    if (MP==null) return;
    // ORIG before 2.6.0: Layer lay = (layer!=null)? layer : (Layer)MP.getLayers().get(1);
    Layer lay = getLayer();
    Line  lin = line.copy();

    /* Added for BUG 2166: Allow plotting of time-based Features on Type 1000 files */
    boolean lay1d = (lay != null && lay.pf instanceof DataFile && lay.typeClass==1);
    boolean lay2d = (lay != null && lay.pf instanceof DataFile && lay.typeClass==2);

    // resolve the tag name
    String tag = "";
    if (text!=null && text.length() != 0) tag=text;
    else if (name!=null) tag=name;

    // Get the y position from vertical ride
    if ((type&VRIDE)!=0 && lay!=null && lay.ndata>0) {
      try { y = lay.getNearestYValue(x); }
      // generated by x values within plot BUT outside of data set
      catch (ArrayIndexOutOfBoundsException e) { } // ignore
      catch (ArithmeticException e) { }            // ignore
    }

    // Get the y position using time line if dt >= 0
    if ((lay!=null) && (dt>=0)) {
      // code in this block re-written for Bug 2355
      double tmpVal;
      double tmpDelta;
      double startAdj = lay.pf.getStart();

      tmpVal =  (lay2d)? lay.getYPosAtTime(t) : lay.getPosAtTime(t);
      tmpVal += startAdj;
      top    = (lay2d)? lay.getYPosAtTime(t-dt/2) : lay.getPosAtTime(t-dt/2);
      top    += startAdj;

      if (FAST_DT2DY) {
        tmpDelta = dt;
      }
      else  {
        tmpDelta =  (lay2d)? lay.getYPosAtTime(t+dt/2) : lay.getPosAtTime(t+dt/2);
        tmpDelta += startAdj - top;
      }

      if (lay1d && MP.xvy) {
        x  = tmpVal;   // type 1000 time-based so set the x values
        dx = tmpDelta;
      } else {
        y  = tmpVal;
        dy = tmpDelta;
      }
    }

    /* Get the pixel position of the feature (center point) */
    MP.getPoint(x,y,z,pix);
    ix=pix.x; // X Pixel position
    iy=pix.y; // Y Pixel position

    // check for in range to display
    if (dx==0 && dy==0) {   // Box of size zero
      ix1=ix2=ix; iy1=iy2=iy;
    } else {  // Determine the size of the box from dx and dy
      MP.getPoint(x-dx/2,y-dy/2,z-dz/2,pix);
      ix1=pix.x; iy1=pix.y;
      MP.getPoint(x+dx/2,y+dy/2,z+dz/2,pix);
      ix2=pix.x; iy2=pix.y;
      // swap values so that ix1 <= ix2; iy1 <= iy2
      if (ix2<ix1) { int i=ix1; ix1=ix2; ix2=i; }
      if (iy2<iy1) { int i=iy1; iy1=iy2; iy2=i; }
    }

    // Apply our range checks
    boolean xok = (ix2>=MP.ix1 && ix1<=MP.ix2);
    boolean yok = (iy2>=MP.iy1 && iy1<=MP.iy2);

    // We should always render features if inside their limits, b/c they could be filled.
    if (isType(ELPS)) {
      // NTN: for faster calculation, using max of either semiMajor and semiMinor as radius,
      //      so we don't have to worry about rotation (also convert from meters to RWC length)
      int ex1, ex2, ey1, ey2; // ellipse pixel boundary
      double len = Math.max(semiMajor, semiMinor);
      double radius = Constants.METERS2DEG * len;
      MP.getPoint(x-radius,y-radius,z-dz/2,pix); ex1=pix.x; ey1=pix.y;
      MP.getPoint(x+radius,y+radius,z+dz/2,pix); ex2=pix.x; ey2=pix.y;
      // swap values so that ex1 <= ex2; ey1 <= ey2
      if (ex2<ex1) { int i=ex1; ex1=ex2; ex2=i; }
      if (ey2<ey1) { int i=ey1; ey1=ey2; ey2=i; }

      // update range check to include ellipse boundary
      xok = xok || (ex2>=MP.ix1 && ex1<=MP.ix2);
      yok = yok || (ey2>=MP.iy1 && ey1<=MP.iy2);
    }

    // FIX BUG 836 to still display circle when zoom in too much
    if (isType(CIRC)) {
      int cx1, cx2, cy1, cy2; // circle pixel boundary
      double radius = Math.max(dx,dy);
      MP.getPoint(x-radius,y-radius,z,pix); cx1=pix.x; cy1=pix.y;
      MP.getPoint(x+radius,y+radius,z,pix); cx2=pix.x; cy2=pix.y;
      // swap values so that ex1 <= ex2; ey1 <= ey2
      if (cx2<cx1) { int i=cx1; cx1=cx2; cx2=i; }
      if (cy2<cy1) { int i=cy1; cy1=cy2; cy2=i; }
      // update range check to include circle boundary
      xok = xok || (cx2>=MP.ix1 && cx1<=MP.ix2);
      yok = yok || (cy2>=MP.iy1 && cy1<=MP.iy2);
    }

    // XLINE and VLINE added for Bug 2191
    boolean yLineToHLine = (type&YLINE)!=0 && !MP.xvy;   // Draw YLINE as Horizontal Line
    boolean yLineToVLine = (type&YLINE)!=0 &&  MP.xvy;   // Draw YLINE as Vertical Line
    boolean xLineToHLine = (type&XLINE)!=0 &&  MP.xvy;   // Draw XLINE as Horizontal Line
    boolean xLineToVLine = (type&XLINE)!=0 && !MP.xvy;   // Draw XLINE as Vertical Line

//System.out.println("Feat draw: name="+name+" top="+top+" x="+x+" y="+y+" ix="+ix+" iy="+iy+" ix1,2="+ix1+","+ix2+" iy1,2="+iy1+","+iy2+" dx="+dx+" dt="+dt+" dy="+dy+" xok="+xok+" yok="+yok);
    if (!xok && (type&(HLINE))==0      && !xLineToHLine && !yLineToHLine) return;
    if (!yok && (type&(VLINE|DATA))==0 && !xLineToVLine && !yLineToVLine) return;

    // if time, draw only if top is valid ?
    if (dt >= 0) {
      double min =  (lay1d) ? MP.orx1 : MP.ory1;
      double max =  (lay1d) ? MP.orx2 : MP.ory2;
      if (top < min || top > max) return;
    }

    Graphics gc = MP.gc;
    Graphics gt = MP.gc; // on axis (not in clip window)
    lin.updateGraphics(gc); // set color, font, stroke, etc for line

    if ((type&INVERSE)!=0) {
      gc.setXORMode(MP.theme.cfg);
      gt.setXORMode(MP.theme.cfg);
    }
    if (((type&HLINE)!=0 || yLineToHLine || xLineToHLine) && yok) {
      gc.drawLine(MP.ix1,iy1,MP.ix2-2,iy1);
      if (iy2!=iy1) gc.drawLine(MP.ix1,iy2,MP.ix2-2,iy2);
      if ((type&IPLT)!=0) {
        if ((type&LTAG)!=0) gt.drawString(tag,MP.ix1,iy);
        if ((type&RTAG)!=0) gt.drawString(tag,MP.ix2-MP.tw*(1+tag.length()),iy);
      } else {
        if ((type&LTAG)!=0) gt.drawString(tag,MP.ix1-MP.tw*(1+tag.length()),iy);
        if ((type&RTAG)!=0) gt.drawString(tag,MP.ix2+MP.tw,iy);
      }
    }
    if (((type&VLINE)!=0 || yLineToVLine || xLineToVLine) && xok) {
      gc.drawLine(ix1,MP.iy1,ix1,MP.iy2-2);
      if (ix2!=ix1) gc.drawLine(ix2,MP.iy1,ix2,MP.iy2-2);
      int iix = ix;
      if (dx>0) iix = ix - (MP.tw*tag.length())/2;
      if ((type&IPLT)!=0) {
        if ((type&TTAG)!=0) gt.drawString(tag,iix,MP.iy1+MP.th-2);
        if ((type&BTAG)!=0) gt.drawString(tag,iix,MP.iy2);
      } else {
        if ((type&TTAG)!=0) gt.drawString(tag,iix,MP.iy1-5);
        if ((type&BTAG)!=0) gt.drawString(tag,iix,MP.iy2+MP.th);
      }
    }
    if ((type&DATA)!=0 && xok) {
      if (lay instanceof Layer1D && lay.ndata>0) { // BUG 1392 - prevent ClassCastException if not Layer1D
        int i1 = Math.max(0,(int)((x-dx/2-lay.xstart)/lay.xdelta+.5));
        int i2 = Math.min(lay.ndata,(int)((x+dx/2-lay.xstart)/lay.xdelta+.5));
        int i12 = i2-i1; if (i12>0)
        MP.drawLine (((Layer1D)lay).fbuf,i1,1,i12,lin,lay.xstart,lay.xdelta);
      }
    }

    if (xok && yok) {
      int idx = ix2-ix1;
      int idy = iy2-iy1;
      boolean filled = (type&FILLED)!=0;

      if ((type&BOX)!=0)  {
        gc.drawRect(ix1,iy1,idx,idy);
        if (filled) gc.fillRect(ix1,iy1,idx,idy);
      }
      if ((type&CIRC)!=0) {
        int idc = Math.max(idx,idy);
        int halfRadius = idc/2;
        gc.drawOval(ix-halfRadius,iy-halfRadius,idc,idc);
        if (filled) gc.fillOval(ix-halfRadius,iy-halfRadius,idc,idc);
      }
      if ((type&OVAL)!=0) {
        gc.drawOval(ix1,iy1,idx,idy);
        if (filled) gc.fillOval(ix1,iy1,idx,idy);
      }
      // Support for inverted filled Triangle, probably belongs elsewhere
      // but I put it here until features are refactored
      if ((type&TRI)!=0) drawTriangle (ix,iy,(int)dx,gc);
      if ((type&ELPS)!=0) MP.drawEllipse (x,y,z,semiMajor,semiMinor,90.0-tilt,-1,lin);
      if ((type&SYMBOL)!=0) lin.getSymbol().draw(ix,iy,gc);
      if ((type&OSYMBOL)!=0 ) gc.drawOval(ix-2,iy-2,4,4);
      if ((type&TEXT)!=0)  drawText(gc,tag,ix,ix1,ix2,iy,iy1,iy2);
    }
    gt.setColor(MP.theme.cfg);
    if ((type&INVERSE)!=0) {
      gc.setPaintMode();
      gt.setPaintMode();
    }
  }

  /** Draw the text label tag at the position(s) specified in options (Left, Right, Top, Bottom, or Default)
      @param gc The Graphics context to draw on
      @param tag The text to be displayed
      @param ix The X pixel position of the feature
      @param ix1 The X min pixel position of the feature
      @param ix2 The X max pixel position of the feature
      @param iy The Y pixel position of the feature
      @param iy1 The Y min pixel position of the feature
      @param iy2 The Y max pixel position of the feature
      @since NeXtMidas 2.7.4 (BUG 2008)
   */
  private void drawText (Graphics gc,String tag, int ix, int ix1, int ix2,  int iy, int iy1, int iy2) {
    int posx = ix; // starting x pos of label
    int posy = iy; // starting y pos of label
    final int CENTER_OFFSET = 2;   // amount that the center point needs to be adjusted
    int centerx = (ix1 + (ix2-ix1) / 2) + CENTER_OFFSET;   // center X pix pos
    int centery = (iy1 + (iy2-iy1) / 2) - CENTER_OFFSET;   // cenetr Y pix pos
    int charWidth  = MP.tw;          // width of a single character
    int textWidth  = charWidth*(1+tag.length());  // width of all the text
    int textHeight = MP.th;          // height of a single character

    if      ((type&LTAG)!=0) {     // left side of feature
      posx = ix1 - textWidth;
      posy = centery +(textHeight /2);
      gc.drawString(tag, posx, posy);
    }
    if ((type&RTAG)!=0) {     // right side of feature
      posx = ix2 + charWidth;
      posy = centery + (textHeight /2) ;
      gc.drawString(tag, posx, posy);
    }
    if ((type&TTAG)!=0) {    // top of feature
      posx = centerx - (textWidth/2);
      posy = iy1 - 2;
      gc.drawString(tag, posx, posy);
    }
    if ((type&BTAG)!=0) {    // bottom of feature
      posx = centerx - (textWidth/2);
      posy = iy2 + textHeight;
      gc.drawString(tag, posx, posy);
    }
    if ((type&BTAG)==0 && (type&TTAG)==0 && (type&RTAG)==0 && (type&LTAG)==0) {  // default
      posx = ix + CENTER_OFFSET;
      posy = iy - CENTER_OFFSET;
      gc.drawString(tag, posx, posy);
    }
  }

  private void drawTriangle (int ix, int iy, int size, Graphics gc) {
    int xpoints [] = new int [4];
    int ypoints [] = new int [4];
    xpoints[0] = ix-size;    ypoints[0] = iy;
    xpoints[1] = ix;         ypoints[1] = iy+size;
    xpoints[2] = ix+size;    ypoints[2] = iy;
    xpoints[3] = xpoints[0]; ypoints[3] = ypoints[0]; //Close the polygon
    gc.fillPolygon (xpoints,ypoints,xpoints.length);
  }

  private final void undraw () {
// NTN: the below optimization no longer works as MPlot.gc is null after calls to MPlot.drawDataArea()
//    if (MP==null) return;
//    if ((type&INVERSE)!=0 && drawn) { MP.drawOverlay(0,this); drawn=false; }
  }

  public void refresh () {
    //if (!drawn && MP!=null) MP.drawOverlay(0,this);
    //else if (layer!=null) layer.needRefresh=true;
    //else needRefresh=true;
    if (isEnabled(0)) needRefresh=true;
  }

  public boolean reset () {
    return false; // not applicable for a Feature
  }

  /** Get the tag NAME of the feature */
  public String getName () { return name; }
  /** Get the TEXT of the feature*/
  public String getText () { return text; }
  /** Get the TYPE name (String) of the feature*/
  public String getType () { return Parser.mask2s(typeList,type); }
  /** Get the Anchor mode of the feature*/
  public String getAnchor () { return Parser.get(anchorList,anchor); }
  /** Get the X POSITION of the feature*/
  public double getX () { return x; }
  /** Get the Y POSITION of the feature*/
  public double getY () { return y; }
  /** Get the Z POSITION of the feature*/
  public double getZ () { return z; }
  /** Get the time of the feature. */
  public double getT () { return t; }
  /** Get the X DELTA of the feature (the width). */
  public double getDX () { return dx; }
  /** Get the Y DELTA of the feature (the height). */
  public double getDY () { return dy; }
  /** Get the Z DELTA of the feature (the elevation). */
  public double getDZ () { return dz; }
  /** Get the time delta of the feature. */
  public double getDT () { return dt; }
  /** Get the ellipse semi-major axis of the feature, in meters. */
  public double getMajor () { return semiMajor; }
  /** Get the ellipse semi-major axis of the feature, in meters. */
  public double getSemiMajor () { return semiMajor; }
  /** Get the ellipse semi-minor axis of the feature, in meters. */
  public double getMinor () { return semiMinor; }
  /** Get the ellipse semi-minor axis of the feature, in meters. */
  public double getSemiMinor () { return semiMinor; }
  /** Get the ellipse tilt angle of the feature, in degrees. (0=North)*/
  public double getTilt () { return tilt; }

  public void setAnchor (String value) {
    setAnchor( Parser.find(anchorList,value,A_DATA) );
  }

  public void setName (String value) {
    if (value==null || value.length()==0 || value.equals(name)) return;
    value = value.toUpperCase();
    if (MP!=null && name!=null) MP.features.rename (name,value);
    name = value;
    refresh();
  }

  /** Set the text of the Feature.  This can be moved to left, right, top or bottom of feature. */
  public void setText (String value) {
    if ((text != null && value != null && text.equals(value)) ||
        (text == null && value == null)) {
      return;
    }
    undraw();
    text = StringUtil.isNull(value)? null : value;
    refresh();
  }

  public void setEnable (int value) {
    if (enable == value) return;  // no change
    boolean wasDisplayed = (enable&MPlot.EN_GLOBAL) !=0;
    undraw();
    enable = value;
    refresh();
    boolean nowDisplayed = (enable&MPlot.EN_GLOBAL) !=0;
    if ((wasDisplayed && !nowDisplayed)) needRefresh=true;   // ensure feature is removed when un-enabled (Bug 2036)
  }

  // The following get/setEnable methods match those in Layer.java
  public void   setEnable (String value) { setEnable(Parser.mask(MPlot.enableList,value,enable)); }
  public int    getEnable ()             { return enable; }
  public String getEnableString ()       { return Parser.mask2s(MPlot.enableList, enable); }

  /** set list of feature types to enable */
  public void setType (String value) {
    String temp = "|"+value.toUpperCase()+"|";
    if (temp.indexOf("|LABEL|") >=0) {
      Shell.deprecate("FEATURE: TYPE=|LABEL| is deprecated, use TYPE=|TEXT| in place of it.");
      value = temp.replaceAll("[|]LABEL[|]", "|TEXT|");
    }
    setType( Parser.mask(typeList,value,type) );
  }

  /** Sets the type of feature to plot.
      @param value The mask value indicating the feature's type.
   */
  public void setType (int value) {
    int prevType = type;
    Layer lay = getLayer();
    // fix for Bug 1392 - if TYPE=|DATA| and not on Layer1D (non-null) then warn
    if ( ((value&DATA) != 0) && (lay != null) && !(lay instanceof Layer1D) ) {
      Shell.warning("Features of TYPE=|Data| are only supported on Layer1D. Feature "
          +this.getName()+" on layer " +lay.getName()+ " is of class "
          +lay.getClass().getSimpleName()+ " and has a specified TYPE="
          +Parser.mask2s(typeList, value));
    }
    type = value;
    if (type != prevType) { // only need to refresh if type changed
      undraw();
      refresh();
    }
  }

  public void setAnchor (int value) {
    if (anchor == value) return;
    undraw();
    anchor = value;
    refresh();
  }

  /** Sets the X-coordinate (RWC) of the center point */
  public void setX (double value) {
    if (x == value) return;
    undraw();
    x = value;
    refresh();
  }

  /** Sets the Y-coordinate (RWC) of the center point */
  public void setY (double value) {
    if (y == value) return;
    undraw();
    y = value;
    refresh();
  }

  /** Sets the Z-coordinate (RWC) of the center point */
  public void setZ (double value) {
    if (z == value) return;
    undraw();
    z = value;
    refresh();
  }

  public void setT (double value) {
    if (t == value) return;
    undraw();
    t = value;
    refresh();
  }

  /** Sets the width of the feature in RWC. Note that the {@link #CIRC} feature
      uses the the larger of dx or dy when drawing.
   */
  public void setDX (double value) {
    if (dx == value) return;
    undraw();
    dx = value;
    refresh();
  }

  /** Sets the height of the feature in RWC. Note that the {@link #CIRC} feature
      uses the the larger of dx or dy when drawing.
   */
  public void setDY (double value) {
    if (dy == value) return;
    undraw();
    dy = value;
    refresh();
  }

  public void setDZ (double value) {
    if (dz == value) return;
    undraw();
    dz = value;
    refresh();
  }

  public void setDT (double value) {
    if (dt == value) return;
    undraw();
    dt = value;
    refresh();
  }

  /** Sets X and DX using min and max values. This is computed as:
      <pre>
        dx = max - min;
        x  = min + dx/2;
      </pre>
      @param min Minimum X value.
      @param max maximum X value.
      @since NeXtMidas 2.5.5
   */
  public void setXMinMax (double min, double max) {
    if (min > max) { double temp = min; min = max; max = temp; }
    boolean doRefresh = false;
    if (dx != (max-min) || (x != (min + ((max-min)/2)) )) doRefresh = true;
    if (!doRefresh) return;
    undraw();
    dx = max - min;
    x  = min + dx/2;
    refresh();
  }

  /** Sets Y and DY using min and max values. This is computed as:
      <pre>
        dy = max - min;
        y  = min + dy/2;
      </pre>
      @param min Minimum Y value.
      @param max maximum Y value.
      @since NeXtMidas 2.5.5
   */
  public void setYMinMax (double min, double max) {
    if (min > max) { double temp = min; min = max; max = temp; }
    boolean doRefresh = false;
    if (dy != (max-min) || (y != (min + ((max-min)/2)) )) doRefresh = true;
    if (!doRefresh) return;
    undraw();
    dy = max - min;
    y  = min + dy/2;
    refresh();
  }

  /** Sets T and DT using min and max values. This is computed as:
      <pre>
        dt = max - min;
        t  = min + dt/2;
      </pre>
      @param min Minimum T value.
      @param max maximum T value.
      @since NeXtMidas 2.5.5
   */
  public void setTMinMax (double min, double max) {
    if (min > max) { double temp = min; min = max; max = temp; }
    if (dt == (max-min) && (t == (min + ((max-min)/2)) )) return;
    undraw();
    dt = max - min;
    t  = min + dt/2;
    refresh();
  }

  /** Gets the minimum X value.
      @return <tt>x - dx/2</tt>
      @since NeXtMidas 2.5.5
   */
  public double getXMin() { return x - dx/2; }

  /** Gets the maximum X value.
      @return <tt>x + dx/2</tt>
      @since NeXtMidas 2.5.5
   */
  public double getXMax() { return x + dx/2; }

  /** Gets the minimum Y value.
      @return <tt>y - dy/2</tt>
      @since NeXtMidas 2.5.5
   */
  public double getYMin() { return y - dy/2; }

  /** Gets the maximum Y value.
      @return <tt>y + dy/2</tt>
      @since NeXtMidas 2.5.5
   */
  public double getYMax() { return y + dy/2; }

  /** Gets the minimum T value.
      @return <tt>t - dt/2</tt>
      @since NeXtMidas 2.5.5
   */
  public double getTMin() { return t - dt/2; }

  /** Gets the maximum T value.
      @return <tt>t + dt/2</tt>
      @since NeXtMidas 2.5.5
   */
  public double getTMax() { return t + dt/2; }


  public void setMajor (double value) { setSemiMajor(value); }

  /** Set the semi-major ellipse axis in meters */
  public void setSemiMajor (double value) {
    if (semiMajor == value) return;
    undraw();
    semiMajor = value;
    refresh();
  }

  public void setMinor (double value) { setSemiMinor(value); }

  /** Set the semi-minor ellipse axis in meters */
  public void setSemiMinor (double value) {
    if (semiMinor == value) return;
    undraw();
    semiMinor = value;
    refresh();
  }

  /** Set the tilt of an ellipse (0=North).
      @param value The value in degrees
   */
  public void setTilt  (double value) {
    if (tilt == value) return;
    undraw();
    tilt = value;
    refresh();
  }

  public void delete () {
    if (name!=null) MP.removeFeature (this.name);
    else if (layer!=null) layer.feature = null;
  }

  public boolean isEnabled (int mode) {
    // Implements Drawable.isEnabled(..), see Drawable.java for notes about implementation.
    if ((enable&MPlot.EN_GLOBAL)==0) return false;
    if (mode==0) return true;
    else return (mode>0) ^ ((enable&MPlot.EN_BSTORE)!=0);
  }

  /** issue configuration menus */
  public void configure () {
    // JC4: 13DEC06 Removed GoToLoc from list... never implemented.
    new GMenu (MP,"Feature.cfg","Rename,Delete,Enable,Type,Line,Anchor,Dialog,Query",0,0,this);
  }

  public int processMessage (Message msg) {
    int   status = 1; // JC4: Why do we use 1/0 rather than NORMAL/NOOP???

    if (msg.name.equals("FEATURE.CFG")) {
      String text = (String)msg.data;
      if      (text.equals("RENAME")) new GPrompt (MP,"Name",name,0,this);
      else if (text.equals("DELETE")) MP.removeFeature (name);
      else if (text.equals("ENABLE")) new GMenu(MP,"Enable",MPlot.enableList,enable,GMenu.TOGGLE,this);
      else if (text.equals("TYPE"))   new GMenu(MP,"Type",typeList,type, GMenu.TOGGLE, this);
      else if (text.equals("LINE"))   line.configure();
      else if (text.equals("ANCHOR")) new GMenu(MP,"Anchor",anchorList,anchor, 0, this);
      else if (text.equals("DIALOG")) dialog();
      else if (text.equals("QUERY"))  new GQuery(MP,"Feature "+name, this, 0, this);
      else { Shell.warning("FEATURE: Unknown message name="+msg.getName()+" data="+msg.getData()); }
    }
    else if (msg.name.equals("ENABLE"))   { enable = msg.info; refresh(); }
    else if (msg.name.equals("TYPE"))     { type   = msg.info; refresh(); }
    else if (msg.name.equals("NAME") ||
             msg.name.equals("RENAME") ||
             msg.name.equals("NEWNAME"))  { setName((String)msg.data);    }
    else if (msg.name.equals("FEATURE.DIALOG")) {
      Table t = (Table)msg.data;
      String action = t.getS("DIALOG.VALUE");
      setX (t.getD("X.VALUE"));
      setY (t.getD("Y.VALUE"));
      setText (t.getString("TEXT.VALUE"));
      if (action.equals("ZOOMTO")) {
        double zr = t.getD("ZOOMRANGE.VALUE");
        double[] zoom = new double[] { x-zr,x+zr,y-zr,y+zr };
        MP.mh.processMessage( new Message("ZOOM",0,zoom) );
      }
    }
    else {
      Shell.warning("FEATURE: Unknown message name="+msg.getName()+" data="+msg.getData());
      status = 0;
    }
    return status;
  }

  /** present a configuration dialog */
  public void dialog () {
    String sticky = "Apply,ZoomTo", actions = "Submit,Cancel,"+sticky;
    String ptext = (text==null)? "null" : text;
    boolean geo = (MP.view==MPlot.V_LATLON) || (MP.view==MPlot.V_MERCATOR) || (MP.view==MPlot.V_GEODETIC);
    GDialog gd = new GDialog (MP, "Feature.Dialog", 250,75, actions,sticky, 0,this);
    GPrompt gl = new GPrompt (gd, "Text", ptext, GWidget.NOMSG, this);
    GValue  gx = new GValue (gd, "X  ", x, MP.orx1, MP.orx2, 1.0, GWidget.NOMSG, this);
    GValue  gy = new GValue (gd, "Y  ", y, MP.ory1, MP.ory2, 1.0, GWidget.NOMSG, this);
    GValue  gz = new GValue (gd, "Z  ", z, MP.orz1, MP.orz2, 1.0, GWidget.NOMSG, this);
    double  zr = Math.min(Math.abs(MP.orx2-MP.orx1),Math.abs(MP.ory2-MP.ory1));
    GValue  gr = new GValue (gd, "ZoomRange", zr/10, zr/100, zr, zr/100, GWidget.NOMSG, this);
    if (geo) { gx.setTitle("Lon"); gy.setTitle("Lat"); gz.setTitle("Alt"); }
    gd.resize(1);
  }

  public Object getNextLink () {
    return line;
  }

  public Object getPrevLink () {
    return MP;
  }

  /** Sets the plot this feature is part of */
  public void setMPlot (MPlot plot) { setPlot(plot); }

  /** Sets the plot this feature is part of */
  public void setPlot (MPlot plot) {
    this.MP = plot;
    line.setPlot(plot);
  }

  /** Sets the layer this feature is on */
  public void setLayer (Layer layer) {
    this.layer = layer;
    line.setParent(layer);
  }

  /** Gets the plot this feature is part of */
  public MPlot getMPlot () {
    return getPlot();
  }

  /** Gets the plot this feature is part of */
  public MPlot getPlot () {
    return MP;
  }

  /** Gets the layer this feature is on. If none specified, then get the first
      Layer from the Plot to which this feature belongs. */
  public Layer getLayer () {
    Layer result = layer;
    if (result==null && MP!=null) {
      KeyVector layers = MP.getLayers();
      if (layers!=null && layers.size()>0) result = (Layer)layers.get(1);
    }
    return result;
    // ORIG before 2.6.0: return layer;
  }

  /** Gets the line this feature uses */
  public Line getLine () {
    return line;
  }

  /** get feature parameters as a table (for export) */
  public Table toTable () {
    Table ft = new Table();
    java.awt.Color color = line.getColor();
    ft.put("NAME",name);
    ft.put("TYPE",getType());
    if (getType().contains("Symbol")) ft.put("SYMBOL", line.getSymbol().getName());  // added for Bug 2264
    if (anchor!=A_DATA) ft.put("ANCHOR",getAnchor());
    if (color!=null) {
      // ft.put("COLOR",Convert.l2x(color.getRGB()));  // replaced with line below for Bug 2263
      ft.put("COLOR",MColor.toString(color));
    }
    if (x!=0)  ft.put("X",x);
    if (y!=0)  ft.put("Y",y);
    if (z!=0)  ft.put("Z",z);
    if (t!=0)  ft.put("T",t);
    if (dx!=0) ft.put("DX",dx);
    if (dy!=0) ft.put("DY",dy);
    if (dz!=0) ft.put("DZ",dz);
    if (dt!=0) ft.put("DT",dt);
    if (text!=null) {
      ft.put("TEXT",text);
      Shell.deprecate("FEATURE.toTable, LABEL key deprecated, use TEXT key. Since NeXtMidas 2.3.0");
      ft.put("LABEL",text);
    }
    return ft;
  }

  /** set feature parameters from a table */
  public void fromTable (Table t) {
    if (t.containsKey("LABEL")) {
      Shell.deprecate("FEATURE.fromTable: LABEL key deprecated, use TEXT key. Since NeXtMidas 2.3.0");
      t.put("TEXT", t.getS("LABEL"));
    }
    KeyObject.setKeys(this,t);
  }

  /** Checks if this feature is of the following type option (feature) mask. Note, a feature can be
      of multiple types (b/c of the masking, hence it can be a box, circle, ellipse, and so on.
      @param typeOptionMask Type option (feature type) to test if this feature is of that type.
      @return true, if this feature is of the specified type option mask, otherwise false
      @since NeXtMidas 2.7.1; was previously private since 2.3.1
  */
  public final boolean isType (int typeOptionMask) {
    return (type&typeOptionMask) !=0;
  }

  //NTN: the alternate form of isType(String) might be useful for usage in a macro (2009-04-13)
  ///** Checks if this feature is of the following type list (e.g. "CIRC|ELPS").
  //    Note, a feature can be of multiple types (b/c of the masking, hence it can
  //    be a box, circle, ellipse, and so on.
  //    @see #isType(int)
  // */
  //public final boolean isType (String typeList) {
  //  return isType(Parser.mask(typeList, typeList, 0));
  //}

  // The intend ref object is used when opening files (can be Midas, Command, etc.),
  // but is currently not used in Feature (it was probably copied from Layer.java).
  /** Deprecated since NeXtMidas 2.5.0: only for internal usage by nxm.sys.intr.Feature. */
  public void setRef (Object ref) { this.ref = ref; }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Deprecated Fields/Methods
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** @deprecated Since NeXtMidas 2.3.0, use getText */
  @Deprecated public String getLabel () {
    Shell.deprecate("FEATURE: getLabel deprecated since 2.3.0, use getText(String)");
    return text;
  }

  /** @deprecated Since NeXtMidas 2.3.0, use setText */
  @Deprecated public void setLabel (String value) {
    Shell.deprecate("FEATURE: setLabel deprecated since 2.3.0, use setText(String)");
    setText(value);
  }
}
