package nxm.sys.libg;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

import java.awt.*;

import javax.swing.JOptionPane;

/** A class to hold and plot a data layer.

    @author Jeff Schoen
    @version $Id: Layer.java,v 1.133 2012/05/14 17:30:47 jwb Exp $
 */
public class Layer extends Object implements MessageHandler,Chainable,Drawable {

  public    int           ndata;                 // Number of data elements being processed
  public    int           frame;                 // Frame size of the layer
  public    int           size;                  // Number of elements in a single transfer length
  public    int           scaleable = 0xF;       // Scalability flags (see MPlot.scaleList)
  public    int           enable = MPlot.EN_DEFAULT; // Enable flags (see MPlot.enableList)
  public    double        xstart;                // x-axis start value
  public    double        xdelta;                // x-axis delta
  private   int           xunits;                // x-axis units (see {@link Units})
  public    double        ystart;                // y-axis start value
  public    double        ydelta;                // y-axis delta
  private   int           yunits;                // y-axis units (see {@link Units})
  public    double        zstart;                // z-axis start value
  public    double        zdelta;                // z-axis delta
  private   int           zunits;                // z-axis units (see {@link Units})
  public    double        time;                  // current time for animation (used by LayerGeo)
  public    double        fseek;                 // Current seek position in the plot file
  public    double        x1     = 0;            // Minimum x-axis value (set by findRange())
  public    double        x2     = 1;            // Maximum x-axis value (set by findRange())
  public    double        y1     = 0;            // Minimum y-axis value (set by findRange())
  public    double        y2     = 1;            // Maximum y-axis value (set by findRange())
  public    double        z1     = 0;            // Minimum z-axis value (set by findRange())
  public    double        z2     = 1;            // Maximum z-axis value (set by findRange())
  public    double        zd     = 0.1;          // Delta   z-axis for contours
  public    double        a1     = 0;            // Minimum actual auto-scale value (set by findRange())
  public    double        a2     = 0;            // Maximum actual auto-scale value (set by findRange())
  public    double        fx1    = 0;            // Minimum Full x-axis value (set by findFullRange())
  public    double        fx2    = 1;            // Maximum Full x-axis value (set by findFullRange())
  public    double        fy1    = 0;            // Minimum Full y-axis value (set by findFullRange())
  public    double        fy2    = 1;            // Maximum Full y-axis value (set by findFullRange())
  public    double        fz1    = 0;            // Minimum Full z-axis value (set by findFullRange())
  public    double        fz2    = 1;            // Maximum Full z-axis value (set by findFullRange())
  public    boolean       realtime;              // Is this a real-time plot layer
  public    boolean       isPaged;               // Is the layer paged (used by Layer2D, et.al.)
  public    boolean       needRefresh;           // Indicates that the layer needs to be refreshed
  public    boolean       isReset;               // Has the layer been reset (set by reset())
  public    boolean       isDone;                // Is the plot done?
  public    boolean       zaxis;                 // True if data is represented on Z axis
  protected boolean       isBlank;               // Is the layer is blank (i.e. don't need to draw)
  private   boolean       needsRescale;          // Indicates if this layer needs to be rescaled
  public    int           ncolors;               // Number of colors in the color map (usually colors.length-1)
  public    Color[]       colors;                // The colors in the color map (colors[0] is special)
  public    int           cmap;                  // The color map ID (e.g. MColot.MULTI)
  public    int           typeClass;             // Type class of the plot file (7=images)
  public    int           defmode=Drawable.DATA; // Default draw mode (overridden by some layers)
  public    int           stype=Surface.FILL;    // The surface type (usually used by Layer2D)
  public    MPlot         MP;                    // The plot the layer is being drawn on
  public    Feature       feature;               // The feature being used by this layer
  public    PlotFile      pf;                    // The file being plotted (can be null)
  protected PlotFileNH    pfnh;                  // pf cast to PlotFileNH (if applicable)
  public    Data          data;                  // The data buffer being used
  public    String        comment;               // A comment about the layer
  protected Object        ref;                   // Midas reference (for results resolving and when opening files)
  public    Line          line;                  // The line used for drawing the layer
  public    String        name;                  // The name of the layer
            float[]       ffbuf;                 // Fast float buffer (for direct IO reads) - package-private (used only in libg)
  protected GLegend.Entry legendEntry;           // Legend entry for this layer.

  /** List of supported configuration options */
  public static final String configureList =
    "Rename,Delete,Enable,Scale,Line,Symbol,Surface,Feature,Color,Colors,ColorMap,Query,Effects";

  // Effects variables below added for Bug 2196 in NeXtMidas 2.8.2
  /** List of effects that can be applied to the Layer (usually to the ColorMap). Val= {@value #effectsList}*/
  public static final String effectsList = "InvertColors,ReverseColors";
  /** Effect: Invert the colormap. val= {@value #E_INVERT_COLORS} */
  public static final int E_INVERT_COLORS  = 0x01;
  /**  Effect: Reverse the colormap. Val= {@value #E_REVERSE_COLORS} */
  public static final int E_REVERSE_COLORS = 0x02;
  /**  Default effect. (0 = none) */
  public static final int DEF_EFFECTS = 0;  // none

  /** Mask of current effects on the Layer */
  public int effects = DEF_EFFECTS;
  
  private boolean timeAlign = false;  // whether to align type100 time plots off the base layer. Since 3.1.2 (Bug 2419)

  /** Creates a new layer. */
  public Layer() {
    super();
    line = new Line(this);
    isBlank = true;
    legendEntry = new GLegend.Entry(this);
  }

  /** Determine necessary layer type from the table of layer flags and return instance. This looks
      for the <code>LAYERTYPE=</code> or (more commonly) the <code>LT=</code> flag that specifies
      the layer type. <code>LT=</code> should be the fully-qualified name of a layer (e.g.
      <code>LT=nxm.sys.libg.LayerWMS</code>; however, if the name begins with
      <code>nxm.sys.libg.Layer</code> such can be omitted (e.g. <code>LT=WMS</code> will be
      converted to <code>LT=nxm.sys.libg.LayerWMS</code> automatically).
      @return A new Layer (usually a subclass of Layer) or null if type could not be determined.
   */
  public static Layer getLayer (Table tab) {
    if (tab==null) return null;
    // Give the user the option to use longhand or shorthand to set the layer class name.
    String className = tab.getS("LAYERTYPE");
    if (className==null) className = tab.getS("LT");
    if (className==null) return null;
    if (className.indexOf('.')<0) className="nxm.sys.libg.Layer"+className;
    try {
      Class<?> c = Class.forName(className);
      Layer lay = (Layer)c.newInstance();
      return lay;
    }
    catch (Exception e) {
      throw new MidasException("Err loading Layer class: "+className,e);
    }
  }

  /** set the file input - if necessary - should be over-ridden */
  public boolean setFile (PlotFile pf, Table tab, Args MA) {
    setFile(pf);
    setSize((int)pf.getSize());
    setData(pf.getDataBuffer(size));
    return true;
  }

  /** Gets the file being used. */
  public PlotFile getFile () {
    return pf;
  }

  /** Closes the layer. This should only be called from {@link MPlot#removeLayer}. Subclasses should
      override this method as needed, but should always include a call to <tt>super.close()</tt>.
      @since NeXtMidas 2.5.5
   */
  public void close () {
    if ((pf != null) && pf.isOpen()) pf.close();
  }

  /** Get the next available data from the input source.
      This should only be called by the plot primitive.
      @param mode always 0
      @return the drawing mode to use, see {@link Drawable}
   */
  public int getData (int mode) {
    if (pf==null || !pf.isOpen() || data==null || isDone) return Drawable.FULL;
    int stat = Drawable.FULL;
    int ndo = (int)pf.avail();
    if (ndo>0) {         // check data range
      ndo = Math.min(ndo,data.size);
      fseek = pf.seek(); // for getTimeAt reference
      if (isPaged) {
        double s=pf.getStart(), d=pf.getDelta();
        if (typeClass==2) {
          if (d>0) ndo = Math.min(ndo,(int)((MP.ry2-s)/d-fseek+2));
          if (d<0) ndo = Math.min(ndo,(int)((MP.ry1-s)/d-fseek+2));
          setYStart(s+fseek*d);
        } else {
          if (d>0) ndo = Math.min(ndo,(int)((MP.rx2-s)/d-fseek+2));
          if (d<0) ndo = Math.min(ndo,(int)((MP.rx1-s)/d-fseek+2));
          setXStart(s+fseek*d);
        }
      }
    }
    if (ndo>0) {       // process new data
      if (ffbuf!=null) {
        data.buf = Native.uncastF(ffbuf);
        ndo = pf.read(data,ndo);
        ffbuf = Native.castF(data.buf);
      } else {
        ndo = pf.read(data,ndo);
      }
      if (ndo<=0) return Drawable.FULL;
      process(ndo);
      if (isReset) stat = Drawable.DATA;
      else if (isPaged) stat = Drawable.DATAX;
      else stat = defmode;
    }
    else if (isPaged) {  // stop page reads
      pf.seek(pf.getSize());
      if (typeClass==2) setYStart(pf.getStart());
      else setXStart(pf.getStart());
      isDone=true;
      ndata = 0;
    }
    else if (!realtime) stat=Drawable.FULL;
    else if (pf.isStream() && !pf.isStreaming()) stat = Drawable.DONE;
    return stat;
  }

  /** Indicates if the layer is paged.
      @param dir The "directionality" of the plot file (use 2 for Type 2000 files
                 that use Y, 1 for all others that use X).
   */
  public boolean isPaged (int dir) {
    if (dir==1) return isPaged && typeClass != 2;
    if (dir==2) return isPaged && typeClass == 2;
    return isPaged;
  }

  /** Reread the data for the layer, if possible, this will attempt to reopen the file. */
  public boolean reread () {
    if (pf==null || !pf.isOpen() || realtime) return false;
    if (pf instanceof DataFile) {
      DataFile df = (DataFile)pf;
      pf = (PlotFile)df.reOpen();
    }
    if ((MP.options&MPlot.O_BSTORE)!=0 && (enable&MPlot.EN_BSTORE)!=0) MP.reset(-1);
    pf.seek(0.0);
    isDone=false;
    return true;
  }

  /** Reset the layer. This implies a full re-drawing of the entire layer (but
      not necessarily a re-read of the data, see {@link #reread()}) that
      is within the bounds of the plot. This method should only be called by
      {@link MPlot} when it needs to "reset" the layer (e.g. for a "FULL" draw
      such as in response to a ZOOM or when redrawing all of the layers in
      response to external events). (This method may also be used internally
      by the individual layers.) As such, expect this method to be called with
      relative frequency since the plot is usually operating in a dynamic
      environment where layers are changing, users are panning/zooming and
      the window manager is triggering refreshes (e.g. window resized, window
      moved to front, etc).<br>
      <br>
      This resets the internal Layer parameters to match the current parameters in
      {@link MPlot}. In the case of a plot that wraps around or shifts (e.g. {@link Layer2D}),
      this will take into account the RWC bounds the plot is currently showing. For other
      layers, this will typically reset the offset to 0.0 (since they typically have a
      plot window that "starts" at the start of the file). <br>
      <br>
      For the case of a paged file, this resets the the parameters so everything is
      reread and redrawn starting at the "start" of the plot. In the case of a pipe,
      the layer has some flexibility with respect to what is "redrawn," if it has
      some cached data available, it may choose to include it; otherwise it just
      starts over using the new data from the pipe.
      @return {@inheritDoc}
   */
  public boolean reset () {
    if (isPaged && pf!=null && pf.isOpen()) {
      double s=pf.getStart(), d=pf.getDelta(), offset=0.0;
      // Recompute offset into file (computed based on the RWC bounds of the plot)
      if (typeClass==2) {
        if (d>0) offset = Math.floor((MP.ry1-s)/d);
        if (d<0) offset = Math.floor((MP.ry2-s)/d);
      } else {
        if (d>0) offset = Math.floor((MP.rx1-s)/d);
        if (d<0) offset = Math.floor((MP.rx2-s)/d);
      }
      offset = Math.max(0.0,Math.min(pf.getSize(),offset));
      if (fseek<0 || isDone || (offset!=pf.seek())) {
        pf.seek(offset);
        isReset=true;
        isDone=false;
      }
    }
    return isReset;
  }

  /** Clears all data in the layer. The default implementation in {@link Layer} does nothing since
      it holds no persistent information. Layers that hold persistent information will need to
      override this method and to clear that information.
      @return true If anything was cleared, false if there is nothing to clear.
      @since NeXtMidas 2.1.2
   */
  public boolean clear () {
    return false;
  }

  /** Set the Layer Type. This is really handled by {@link #getLayer(Table)}. This method is only
      here to prevent errors when the flags are set via {@link KeyObject}. Any value passed to this
      method will be ignored.
   */
  public final void setLayerType (String value) {
    // this actually is handled in getLayer()
  }

  /** Gets the current layer type. This always returns the fully-qualified name of the layer, even
      if the layer type was initially set via an abbreviation in {@link #getLayer(Table)}.
      @since NeXtMidas 2.5.0
   */
  public final String getLayerType () {
    return getClass().getName();
  }

  /** Set the file being plotted by this layer. */
  public void setFile (PlotFile pf) {
    this.pf   = pf;
    this.pfnh = FileUtil.asPlotFileNH(pf);  // <-- returns null if not an instance of PlotFileNH

    typeClass = pf.getType()/1000;
    fseek = -1;

    setXUnits(pf.getXUnits());
    setYUnits(pf.getYUnits());
    setZUnits(Units.NONE);
  }

  /** Set the plotter being used. Only one plotter may be used at any given time. */
  public void setMPlot (MPlot value) {
    MP = value;
    if (line!=null) line.setPlot(MP);
    if (feature!=null) feature.MP = MP;
  }

  /** Sets the name of the layer. */
  public void setName (String value) {
    if (value==null || value.length()==0) return;
    value = value.toUpperCase();
    if (MP!=null && name!=null) MP.layers.rename(name,value);
    name = value;
  }

  /** Sets the data buffer being used by the layer. */
  public void setData (Data value) { data = value; }

  /** Resync with MPlots mode/view settings */
  public void update () {
    if (ndata==0 && isPaged && data!=null) ndata = data.size;
    if (ndata>0) process();
  }

  public void setEnable (int value) {
    if (enable != value) {
      enable = value;
      needRefresh=true; // special case, force update since we may be changing the ENABLE flag
      refresh();
    }
  }

  /** Toggles the layer's visibility. This is the same as toggling the GLOBAL enable flag. */
  public void toggle () {
    setEnable(enable^MPlot.EN_GLOBAL);
  }

  // The following get/setEnable methods match those in Feature.java
  public void    setEnable (String value) { setEnable(Parser.mask(MPlot.enableList,value,enable)); }
  public int     getEnable ()             { return enable; }
  public String  getEnableString ()       { return Parser.mask2s(MPlot.enableList, enable); }

  /** Sets the auto-scale options flag. Outside of the Layers and {@link MPlot} the
      {@link #setScale(String)} method should be used in place of this one.
      @param value The auto-scale options.
   */
  public void setScale (int value) {
    scaleable = value;
  }

  /** Sets the auto-scale options flag.
      @param value The auto-scale options. The available flags are given in {@link MPlot#scaleList}.
      @since NeXtMidas 2.1.0
   */
  public final void setScale (String value) {
    setScale(Parser.mask(MPlot.scaleList,value,scaleable));
  }

  /** Get the auto-scale options
      @return {@link #scaleable}
      @since NeXtMidas 2.7.1
   */
  public int getScale () { return scaleable; }

  /** Sets the X-Axis Start value.
      @param value The start value.
      @since NeXtMidas 2.1.0
   */
  public void setXStart (double value) { xstart = value; }

  /** Gets the X-Axis start value
      @return {@link #xstart}
      @since NeXtMidas 2.7.1
   */
  public double getXStart () { return xstart; }

  /** Sets the X-Axis Delta value.
      @param value The delta value.
      @since NeXtMidas 2.1.0
   */
  public void setXDelta (double value) { xdelta = value; }

  /** Gets the X-Axis delta value
      @return {@link #xdelta}
      @since NeXtMidas 2.7.1
  */
  public double getXDelta () { return xdelta; }

  /** Sets the X-Axis Start and Delta values.
      @param start The start value.
      @param delta The delta value.
   */
  public void setXAxis (double start, double delta) {
    setAxis (start, delta, 'X');
  }

  /** Sets the Y-Axis Start value.
      @param value The start value.
      @since NeXtMidas 2.1.0
   */
  public void setYStart (double value) { ystart = value; }

  /** Get the Y Axis Start value
      @return {@link #ystart}
      @since NeXtMidas 2.7.1
   */
  public double getYStart () { return ystart;  }

  /** Sets the X-Axis Delta value.
      @param value The delta value.
      @since NeXtMidas 2.1.0
   */
  public void setYDelta (double value) { ydelta = value;  }

  /** Get the Y Axis delta value
      @return {@link #ydelta}
      @since NeXtMidas 2.7.1
   */
  public double getYDelta () { return ydelta; }

  /** Sets the Y-Axis Start and Delta values.
      @param start The start value.
      @param delta The delta value.
   */
  public void setYAxis (double start, double delta) {
    setAxis(start, delta, 'Y');
 }

  /** Sets the Y-Axis Start value.
      @param value The start value.
      @since NeXtMidas 2.1.0
   */
  public void setZStart (double value) { zstart = value;  }

  /** Get the Z Axis Start value
      @return {@link #zstart}
      @since NeXtMidas 2.7.1
   */
  public double getZStart () { return zstart; }

  /** Sets the Z-Axis Delta value.
      @param value The delta value.
      @since NeXtMidas 2.1.0
   */
  public void setZDelta (double value) { zdelta = value;  }

  /** Get the Z Axis delta value
      @return {@link #zdelta}
      @since NeXtMidas 2.7.1
   */
  public double getZDelta () { return zdelta; }

  /** Sets the Z-Axis Start and Delta values.
      @param start The start value.
      @param delta The delta value.
   */
  public void setZAxis (double start, double delta) {
    setAxis(start, delta, 'Z');
  }

  /** Sets the Axis values for the given Axis type. <br>
      The /INDEX switch and axis values precision is checked to ensure no loss of precision. <br>
      If /INDEX or the values fail the precision check, the values are set to /INDEX mode
      (START=0, DELTA=1). <br>
      @param start The Axis start value
      @param delta The Axis delta (change in values in the file)
      @param c The type of Axis being set (only valid values are 'X', 'Y', or 'Z')
   */
  private void setAxis (double start, double delta, char c) {
    boolean precision = true;  // whether the values maintain precision

    if (MP.isAxisIndex()) precision = false;   // INDEX switch has precedence
    else if ((delta==0) || (start == Double.MAX_VALUE)) { }    // Value is just a holding value
    else if ((start+delta) != start) precision = true;  // precision is fine
    else                             precision = false;

    // If no precision, set to INDEX values and notify user
    if (!precision)  {

      // create a popup to notify the user we are changing the axis values
      if (!MP.isAxisIndex()) {
        String text = "WARNING: "+c+" Axis START / DELTA values combination is    \n"+
                      "beyond precision: \n     START = "+start+
                      " \n     DELTA = " +delta+" \n"+
                      "Setting to INDEX mode: \n     START = 0 \n     DELTA = 1";
        MJPopup popup = MJPopup.createOk(MP.getPanel(), "Beyond Precision",
            JOptionPane.WARNING_MESSAGE,text );
        popup.display(0);

        Shell.warning(text);  // Also notify user via terminal
      }

      // Set the start and delta to /INDEX values
      start = 0;
      delta = 1;

      MP.setAxisIndex(true);    // Set /INDEX to true since we set the values that way
    }

    // Set the axis values
     switch (c) {
      case 'X' : xstart = start; xdelta = delta; break;
      case 'Y' : ystart = start; ydelta = delta; break;
      case 'Z' : zstart = start; zdelta = delta; break;
      default  : Shell.warning("Invalid axis type: " +c);
    }
  }

  public void setFrame (int value) { frame = value; }
  public int  getFrame () { return frame; }

  public void setSize (int value) { size = value; }
  public int  getSize () { return size; }

  public void   setTime (double value) { time = value; }
  public double getTime () { return time; }

  public void   setRef (Object ref) { this.ref = ref; }
  public Object getRef () { return ref; }

  /** Sets the real-time mode flag. Note that setting the plot to real-time (rt=true) causes the
      use of BackingStore (enable=+BSTORE) to be turned off.
      @param rt Indicates if the plot should be done in real-time (true) or if it is static (false).
   */
  public void setRealTime (boolean rt) {
    realtime = rt;
    if (rt) enable &= ~MPlot.EN_BSTORE;
  }

  /** Gets the readout for the plot at a given cursor position.
      @param mp The position on the plot.
      @return The readout string or null if the layer does not support readout.
   */
  public String getReadOut (Position mp) { return null; }

  /** Indicates if the layer needs to be refreshed. Subclasses should override this method if they
      have multiple {@link Line}s or if they have other fields to check. Subclasses should override
      this method as follows:
      <pre>
        public boolean getNeedRefresh () {
          return super.getNeedRefresh() || [other things to check];
        }
      </pre>
      If this class uses multiple line objects (other than just <tt>Layer.line</tt>) it should
      include a {@link Line#getNeedRefresh()} check for each line used, if any line has changed,
      this method should return true.
      @return {@inheritDoc}
      @since NeXtMidas 2.5.4
   */
  public boolean getNeedRefresh () {
    return needRefresh || line.getNeedRefresh();
  }

  /** Called by {@link nxm.sys.libg.MPlot} to clear the refresh flag(s) without actually drawing
      the drawable. This should clear all of the flags checked in {@link #getNeedRefresh()}. This
      method is also called at the top of {@link nxm.sys.libg.Layer#draw(int)}. Any subclass that
      overrides {@link #getNeedRefresh()} should also override this method.
      @since NeXtMidas 2.5.4
   */
  public void clearNeedRefresh () {
    needRefresh = false;
    line.clearNeedRefresh();
  }

  public void refresh () {
    if ((enable&MPlot.EN_GLOBAL)!=0) needRefresh=true;
    if (isEnabled(-1) && MP!=null) MP.reset(-1);
  }

  /** Refresh the Layer. This method does NOT set {@link #needRefresh} to true.
      This method provides a mechanism for the Layer to call {@link MPlot#refresh(int)}
      with a user-defined drawmode (usually {@link Drawable#DATA}) to prevent the
      redrawing of the entire plot (including any backing store)
      @param drawmode The type of refresh. See {@link Drawable#DONE}, {@link Drawable#FULL},
      {@link Drawable#DATA}, {@link Drawable#DATAX}, {@link Drawable#FAST})
      @since NeXtMidas 2.9.3
   */
  public void refresh (int drawmode) {
    MP.refresh(drawmode);
  }

  /** process the plot data for this frame */
  public void process () {
    isBlank = false;
  }

  /** process the plot data for specified number of frame elements */
  public void process (int elem) {
    int osize = data.size;
    if (elem<data.size) data.setSize(elem);
    process();
    if (elem!=osize) data.setSize(osize);
  }

  /* find layer bounds - should be overridden */
  public void findRange () {
    x1 = 0; x2 = 1;
    y1 = 0; y2 = 1;
    z1 = 0; z2 = 1;
    a1 = 0; a2 = 1;
  }

  /* find full range layer bounds */
  public void findFullRange () {
    fx1 = x1; fx2 = x2;
    fy1 = y1; fy2 = y2;
    fz1 = z1; fz2 = z2;
  }

  /* find full range layer bounds */
  public void findFullRange (PlotFile pf) {
    fx1 = x1; fx2 = x2;
    fy1 = y1; fy2 = y2;
    fz1 = z1; fz2 = z2;
    if (pf==null || realtime);
    else if (this instanceof Layer2D) {
      fy1 = pf.getYStart();
      double fsize = pf.getSize();
      if (pf instanceof DataFile) { // make sure it is a DataFile before casting
        DataFile df = (DataFile)pf;
        if ((df.flags&BaseFile.WRAP)!=0) fsize = df.getOutByte(0)/df.dbpe;  // use expected size if WRAP
      }
      fy2 = fy1 + (fsize-1)*pf.getYDelta();
      if (fy1>fy2) { double t=fy1; fy1=fy2; fy2=t; }
    } else {
      fx1 = pf.getXStart();
      fx2 = fx1 + (pf.getSize()-1)*pf.getXDelta();
      if (fx1>fx2) { double t=fx1; fx1=fx2; fx2=t; }
    }
  }

  /** Whether or not to align Type 1000 time-based plots off the base layer
      @since NeXtMidas 3.1.2
   */
  public boolean isTimeAlign () {
    return timeAlign;   // added for Bug 2419
  }
  
  /** Set whether or not to align Type 1000 time-based plots off the base layer
      @param val true to align time-based plots off the base layer (Type 1000 only)
      @since NeXtMidas 3.1.2
   */
  public void setTimeAlign (boolean val) {
    timeAlign = val;
  }
  
  /** Draws the layer. Subclasses should override this method to handle layer-specific drawing, note
      that in doing so they should always include a call to <tt>super.draw(flag)</tt>, for example:
      <pre>
        public void draw (int flag) {
          super.draw(flag);
          ...
        }
      </pre>
      When drawing, subclasses should make a copy of the line being used when drawing using
      {@link Line#copy()} this prevents threading issues if an external thread changes line
      parameters (e.g. color) during the draw.
      @param flag {@inheritDoc}
   */
  public void draw (int flag) {
    clearNeedRefresh();
    isReset = false;
  }

  /** issuer configuration menus */
  public void configure () {
    new GMenu (MP,"Layer.cfg", configureList, 0, 0, this);
  }

  /** process a message from parent process or config menu */
  public int processMessage (Message msg) {
    if (msg.name.equals("LAYER.CFG")) {
      String text = (String)msg.data;
      if (text.equals("RENAME"))
        new GPrompt (MP,"Name",name,0,this);
      else if (text.equals("DELETE"))
        MP.removeLayer (name);
      else if (text.equals("ENABLE"))
        new GMenu (MP,"Enable", MPlot.enableList, enable, GMenu.TOGGLE, this);
      else if (text.equals("QUERY"))
        new GQuery (MP,"Layer "+name, this, 0, this);
      else if (text.equals("SCALE"))
        new GMenu (MP,"Scale",MPlot.scaleList,scaleable,GMenu.TOGGLE,this);
      else if (text.equals("LINE"))
        line.configure();
      else if (text.equals("SYMBOL"))
        new GMenu  (MP,"Symbol",Symbol.symbolList,0,0,line);
      else if (text.equals("SURFACE"))
        new GMenu (MP,"Surface",Surface.typeList,stype,GMenu.TOGGLE,this);
      else if (text.equals("FEATURE"))
        { getFeature().configure(); }  // updated in 3.1.1 to prevent Data features on non Layer1D plots (Bug 2549)
      else if (text.equals("COLOR")) {
        String colorString = MColor.toString(getColor()); // Map color to a name if possible
        int len = (colorString != null) ? Math.max(colorString.length(), 20) : 20;
        new GPrompt (MP,"Color",colorString, len, 0, this);
      }
      else if (text.equals("COLORS")) {
        int len = Math.max(Math.min(getColorMap().length(),500), 20);
        new GPrompt (MP,"Colors",getColorMap(), len, 0, this);
      }
      else if (text.equals("COLORMAP"))
        new GMenu (MP,"ColorMap",MColor.colorMapList+",Custom", cmap, 0, this);
      else if (text.equals("EFFECTS"))
        new GMenu (MP,"Effects",effectsList, effects,GMenu.TOGGLE,this);
    }
    else if (msg.name.equals("COLORMAP")) {
      if (msg.data.toString().equalsIgnoreCase("CUSTOM")) {
        new GPrompt(MP,"Colors",getColorMap(), 0, this);
      } else {
        setColorMap(msg.info);
      }
    }
    else if (msg.name.equals("EFFECTS")) {
      if (msg.data.toString().equals("REVERSECOLORS")) {
        reverseColorMap();
      } else if (msg.data.toString().equals("INVERTCOLORS")) {
        invertColorMap();
      }
    }
    else if (msg.name.equals("COLOR")) setColor ((String)msg.data);
    else if (msg.name.equals("COLORS")) setColors ((String)msg.data);
    else if (msg.name.equals("ENABLE")) setEnable(msg.info);
    else if (msg.name.equals("SCALE")) setScale (msg.info);
    else if (msg.name.equals("SURFACE")) setSurface (msg.info);
    else if (msg.name.equals("NAME") || msg.name.equals("RENAME") ||
             msg.name.equals("NEWNAME")) setName((String)msg.data);
    else return 0;
    return 1;
  }

  public boolean isEnabled (int flag) {
    // Implements Drawable.isEnabled(..), see Drawable.java for notes about implementation.
    if ((enable&MPlot.EN_GLOBAL)==0) return false;
    if (flag==0) return true;
    else return (flag>0) ^ ((enable&MPlot.EN_BSTORE)!=0);
  }

  /** Is the layer blank? */
  public boolean isBlank () {
    return isBlank;
  }

  /** Indicates if the layer is visible.
      @return <code>true</code> if the layer is visible, <code>false</code> if not.
   */
  public boolean isVisible () {
    return ((enable&MPlot.EN_GLOBAL) != 0);
  }

  /** set the X bounds */
  public void setX1   (double r1) { setXMin(r1); }
  public void setXMin (double r1) { setXBounds(r1,r1+x2-x1); }
  public void setX2   (double r2) { setXMax(r2); }
  public void setXMax (double r2) { setXBounds(x1,r2); }
  public void setX    (double r1, double r2) { setXBounds(r1,r2); }
  public void setXBounds (double r1, double r2) {
    x1 = Math.min(r1,r2); x2 = Math.max(r1,r2);
  }


  /** set the Y bounds */
  public void setY1   (double r1) { setYMin(r1); }
  public void setYMin (double r1) { setYBounds(r1,r1+y2-y1); }
  public void setY2   (double r2) { setYMax(r2); }
  public void setYMax (double r2) { setYBounds(y1,r2); }
  public void setY    (double r1, double r2) { setYBounds(r1,r2); }
  public void setYBounds (double r1, double r2) {
    y1 = Math.min(r1,r2); y2 = Math.max(r1,r2);
  }

  /** set the Z bounds */
  public void setZ1   (double r1) { setZMin(r1); }
  public void setZMin (double r1) { setZBounds(r1,r1+z2-z1); }
  public void setZ2   (double r2) { setZMax(r2); }
  public void setZMax (double r2) { setZBounds(z1,r2); }
  public void setZ    (double r1, double r2) { setZBounds(r1,r2); }
  public void setZBounds (double r1, double r2) {
    z1 = Math.min(r1,r2); z2 = Math.max(r1,r2);
  }

  public double getX1   () { return getXMin(); }
  public double getXMin () { return x1; }
  public double getX2   () { return getXMax(); }
  public double getXMax () { return x2; }
  public double getY1   () { return getYMin(); }
  public double getYMin () { return y1; }
  public double getY2   () { return getYMax(); }
  public double getYMax () { return y2; }
  public double getZ1   () { return getZMin(); }
  public double getZMin () { return z1; }
  public double getZ2   () { return getZMax(); }
  public double getZMax () { return z2; }

  /** get the time at given position */
  public double getTime (int px, int py) {
    return time;
  }

  /** Gets the position at a given time.
      @param time The time to get the position for (in MJS).
      @return The position that corresponds with the given time (in pixels) or a value less than
              zero is such can not be computed for the given layer.
      @since NeXtMidas 1.9.3
   */
  public double getPosAtTime (double time) {
    return -1;
  }

  /** Gets the Y position at a given time.
      @param time The time to get the position for (in MJS).
      @return The Y offset that corresponds with the given time or a negative value
              if the position is out-of-range or cannot be computed for the given layer.
      @since NeXtMidas 2.7.1
   */
  public double getYPosAtTime (double time) {
    return -1;
  }

  /** Given x, return nearest y value
      @exception  ArithmeticException
   */
  public double getNearestYValue (double x)
    throws ArithmeticException {
    if (xdelta==0.0) throw new ArithmeticException();
    int i = (int)((x-xstart)/xdelta+0.5);
    if (i<0) i=0;
    if (i>=ndata) i=ndata-1;
    return getNearestYValue(i);
  }
  public double getNearestYValue (int i) {
    return 0.0;
  }

  public Feature getFeature () {
    if (feature!=null) return feature;
    feature = new Feature(this);
    //feature.setType(Feature.DATA);  // changed to line below in 3.1.1 since most layers do not support it (Bug 2549)
    feature.setType(0);
    feature.setName("_LAYER_"+getName()+"_");  // added in 3.1.1
    feature.setEnable(0);
    return feature;
  }

  /** Set surface rendering mode. See {@link Surface#typeList} for available options. */
  public void setSurface (int value) { stype=value; refresh(); }

  /** Set surface rendering mode. See {@link Surface#typeList} for available options. */
  public void setSurface (String value) { setSurface( Parser.mask(Surface.typeList,value,stype) ); }

  /** Sets the color of the layer.
      @param value The name of the color.
   */
  public void setColor (String value) { line.setColor(value); }

  /** Sets the color of the layer based on indexing into the a pre-defined set of colors.
      This does NOT take in the RGB value of a color.
      @param value The index of the color.
   */
  public void setColor (int value) { line.setColor(value); }

  /** Sets the line color of the layer.
      @param value The color.
   */
  public void setColor (Color value) { line.setColor(value); }

  /** Gets the line color of the layer.  <br>
      Be careful, this may not be what you want if a color map is being used
      (i.e. there is no single color for the layer).
  */
  public Color getColor () {
    return line.getColor();
  }

  /** Sets the color map colors being used in the color map.
      @param value The name of the color map or a list of colors.
      @see #setColorMap(String)
   */
  public void setColors (String value) {
    setColorMap(value);
  }

  /** Sets the color map being used. <br>
      This is the same as <code>setColorMap(value,-1)</code>.
      @param value The color map ID number.
   */
  public void setColorMap (int value) { setColorMap(value,-1); }

  /** Sets the color map being used.
      @param value   The color map ID number.
      @param ncolors The number of colors to use from the color map (-1 = use all colors).
   */
  public void setColorMap (int value, int ncolors) {
    if (value==-1) value = cmap;
    if (value<1 || value>MColor.MAXTYPE) return;
    if (value==cmap && ncolors==this.ncolors) return;
    cmap = value;
    setColorMap(MColor.getColorMap(value,ncolors));
  }

  /** Sets the actual color map. All of the other <code>setColor(..)</code> methods in {@link Layer}
      end up calling this method. Subclasses should override this if they need to update other
      fields (e.g. a backing store or image), but should remember to include a call to
      <code>super.setColorMap(colors)</code> at the top of the method.
      @param colors The colors that make up the color map.
   */
  public void setColorMap (Color[] colors) {
    if (effects != 0) colors = applyEffects(colors, effects);
    this.colors  = colors;
    this.ncolors = colors.length-1;
    refresh();
  }

  /** Sets the color map being used. <br>
      <br>
      Example usage:
      <pre>
        Use the "Ramp" color map defined in {@link MColor}:
          setColorMap("RAMP")

        Define a custom color map:
          setColorMap("RED,YELLOW,GREEN,BLUE")
            or
          setColorMap("(RED,YELLOW,GREEN,BLUE)")

      </pre>
      @param value The name of the color map or a list of colors.
      @since NeXtMidas 2.1.0
   */
  public void setColorMap (String value) {
    if (value.length()==0) return;
    setColorMap(MColor.getColorMap(value,-1));
  }

  /** Gets the color map being used. <br>
      Note that this intentionally omits color number zero from the list. This prevents the color
      map from changing if <code>setColorMap(..)</code> is later called with the string returned
      from this method.
      @return The list of colors that make up the color map being used.
      @since NeXtMidas 2.1.0
   */
  public String getColorMap () {
    StringBuffer str = new StringBuffer();
    if (colors == null) {return str.toString();}
    for (int i = 1; i < colors.length; i++) {
      if (i != 1) { str.append(','); }
      str.append(MColor.toString(colors[i]));
    }
    return str.toString();
  }

  /** Set the effects on the layer.
      @param arg The effects to set
      @return True if something was set, else false (if the effects were already that value)
      @see #effectsList
      @since NeXtMidas 2.8.2
   */
  public boolean setEffects (int arg) {
    if (arg == effects) return false;
    effects = arg;
    setColorMap(cmap);
    return true;
  }
  /** Set the effects on the layer.
      @param arg The effects to set
      @return True if something was set, else false (if the effects were already that value)
      @see #effectsList
      @since NeXtMidas 2.8.2
   */
  public boolean setEffects (String arg) {
    return setEffects(Parser.mask(effectsList, arg, effects));
  }

  /** Get whether or not the effect(s) given is set in {@link #effects}
      @since NeXtMidas 2.8.2
   */
  public boolean getEffect (int mask) {
    return (effects&mask)!=0;
  }

  /** Apply effects to an array of Colors
      @param colors The array of Colors to adjust
      @param effects Must be a subset of {@link #effectsList}
      @return The modified Color array
      @since NeXtMidas 2.8.2
   */
  private Color[] applyEffects (Color[] colors, int effects) {
    if ((effects&E_INVERT_COLORS) != 0 )  colors = MColor.invertColorsMap(colors);
    if ((effects&E_REVERSE_COLORS) != 0 ) colors = MColor.reverseColorMap(colors);
    return colors;
  }

  /** Reverses the current ColorMap. Doing this twice will revert colors back to normal.
      @since NeXtMidas 2.8.2
  */
  public void reverseColorMap () {
    if (colors==null) {
      Shell.warning("No color map in use, therefore cannot reverse it for "+this);
      return;
    }
    effects ^= E_REVERSE_COLORS; // Toggle
    setColorMap(cmap);
  }

  /** Reverse the current colorMap. This is a convenience method to allow for calling
      {@link #reverseColorMap()} on the plot command line using the {LAYER={ReverseColorMap=TRUE}} qualifier.
      @param val TRUE to reverse the colorMap, otherwise does nothing
      @since NeXtMidas 2.8.2
   */
  public void setReverseColorMap (boolean val) {
    if (val) reverseColorMap();   // FALSE does nothing
  }

  /** Invert the colors (to create photo negative effect). Doing this twice will
      revert colors back to normal.
      @since NeXtMidas 2.9.0
  */
  public void invertColorMap () {
    if (colors==null) {
      Shell.warning("No color map in use, therefore cannot invert it for "+this);
      return;
    }
    effects ^= E_INVERT_COLORS;  // Toggle
    setColorMap(cmap);
  }

  /** This is a convenience method to allow for calling {@link #invertColorMap()}
      on the plot command line using the {LAYER={InvertColorMap=TRUE}} qualifier.
      @param val TRUE to invert the colors, otherwise does nothing
      @since NeXtMidas 2.9.0
   */
  public void setInvertColorMap (boolean val) {
    if (val) invertColorMap(); // FALSE does nothing
  }

  public void setHiMode (int value) {
    if (value==1) value = Feature.DATA;
    else if (value==2) value = Feature.VLINE;
    else if (value==3) value = Feature.DATA|Feature.VLINE;
    getFeature().setType(value);
  }
  public void setHiX (double value) { getFeature().setX(value); }
  public void setHiXD (double value) { getFeature().setDX(value); }
  public void setHiColor (String value) { getFeature().getLine().setColor(value); }

  /** Get the Highlight X value*/
  public double getHiX () { return getFeature().getX(); }

  /** Get the Highlight deltax value*/
  public double getHiXD () { return getFeature().getDX(); }

  public int getx1i () {
    int i = (int)( (MP.rx1-xstart)/xdelta );
    return Math.max(1,Math.min(frame,i+1));
  }

  public int getx2i () {
    int i = (int)( (MP.rx2-xstart)/xdelta );
    return Math.max(1,Math.min(frame,i+1));
  }

  /** @deprecated Since NeXtMidas 2.5.4 - this was for internal usage and may be
                 removed in a future release.
   */
  @Deprecated protected void prepDATAX (double x1, double x2, boolean delta, boolean marker) {
    if (MP.gc==null) return;
    int ic = 0;
    int ixc1 = MP.getix(x1);
    int ixc2 = MP.getix(x2)+1;
    if (delta) {
      MP.gc.clearRect(ixc1+1,MP.iy1,ixc2-ixc1+ic,MP.iy21);
    }
    if (marker && ixc2>MP.ix1 && ixc2<MP.ix2-1) {
      MP.gc.drawLine(ixc2+1,MP.iy1+3,ixc2+1,MP.iy2-4);
    }
  }

  protected static float setScale (Data data, int mode) {
    float scale = 1;
    if (mode==MPlot.M_10LOG) scale=(float)(10.0/Math.log(10.0));
    if (mode==MPlot.M_20LOG) scale=(float)(20.0/Math.log(10.0));
    if (data.spa!=1) scale = scale/2;
    if (mode==MPlot.M_PHASE) scale=(float)(180.0/Math.PI);
    return scale;
  }

  @SuppressWarnings("fallthrough")
  protected float[] data2fa (Data data, int ndata, float[] fbuf, int offset, int mode) {
    int i=offset,i2=i+ndata,spa=data.spa;
    float scale = setScale(data, mode);
    if (ffbuf!=null) fbuf=ffbuf;
    else {
      if (fbuf==null || fbuf.length<i2*spa){
        try {
          fbuf = new float[i2*spa];
        } catch (OutOfMemoryError e) {  // for when user tries to allocate a buffer bigger than memory will hold (Bug 885)
          throw new MidasException("Not enough memory to allocate a (float) buffer of size "+(i2*spa*4)+" bytes. "
              +"Current free memory is "+Runtime.getRuntime().freeMemory()+" bytes.",e);
        }
      }
      Convert.bb2ja (data.buf,0,data.type, fbuf,i*spa,Data.FLOAT, ndata*spa);
    }
    if (data.spa==1) {
      switch (mode) {
      case MPlot.M_10LOG: case MPlot.M_20LOG:
        for (;i<i2;i++) fbuf[i] = scale *
            (float)Math.log(Math.max(10.e-11,Math.abs((double)fbuf[i])));
        break;
      default:
        break;
      }
    }
    else if (data.spa==2) {
      int j=i;
      switch (mode) {
      case MPlot.M_MAG:
        for (;i<i2;i++,j+=2) fbuf[i] = (float)Math.sqrt((double)(fbuf[j]*fbuf[j]+fbuf[j+1]*fbuf[j+1]));
        break;
      case MPlot.M_PHASE:
        for (;i<i2;i++,j+=2) fbuf[i] = scale * (float)Math.atan2((double)fbuf[j+1],(double)fbuf[j]);
        break;
      case MPlot.M_IMAG: case MPlot.M_VY:
        j+=1; // fallthrough
      case MPlot.M_REAL: case MPlot.M_VX:
        for (;i<i2;i++,j+=2) fbuf[i] = fbuf[j];
        break;
      case MPlot.M_10LOG:
      case MPlot.M_20LOG:
        for (;i<i2;i++,j+=2) fbuf[i] = scale * (float)Math.log(Math.max(1.e-20,(double)(fbuf[j]*fbuf[j]+fbuf[j+1]*fbuf[j+1])));
        break;
      default:
      }
    }
    else if (data.spa==3) {
      int j=i;
      switch (mode) {
      case MPlot.M_MAG:
          for (;i<i2;i++,j+=3) fbuf[i] = (float)Math.sqrt((double)(fbuf[j]*fbuf[j]+fbuf[j+1]*fbuf[j+1]+fbuf[j+2]*fbuf[j+2]));
        break;
      case MPlot.M_VZ:
          j+=1; // fallthrough
      case MPlot.M_VY: case MPlot.M_IMAG:
          j+=1; // fallthrough
      case MPlot.M_VX: case MPlot.M_REAL:
        for (;i<i2;i++,j+=3) fbuf[i] = fbuf[j];
        break;
      default:
      }
    }
    return fbuf;
  }

  /** Read specified amount of data (ndata) from Data and return it back as array of int
      @param data Data object to read data from
      @param ndata number of data elements
      @param buf buffer to put read data into, unless it does not have enough space.
      @param offset offset into buf that data is copied into
      @since NeXtMidas 2.5.0
   */
  protected int[] data2la (Data data, int ndata, int[] buf, int offset) {
    int i=offset,i2=i+ndata;
    if ( data.getFormat().equals("SL") ) {
      if (buf==null||buf.length<i2) buf = new int[i2];
      Convert.bb2ja(data.buf,0,data.type, buf,i,DataTypes.LONG, ndata);
    } else if (data.getFormat().equals("QB")) {
      if (buf==null||buf.length<i2) buf = new int[i2];
      // if data is not the same endian is OS, then swap the bytes
      if (!Convert.sameEndian(data.rep, Shell.rep)) Convert.swap4(data.buf, 0, ndata);

      Convert.bb2ja(data.buf, 0, DataTypes.LONG, buf, i, DataTypes.LONG, ndata);
    } else {
      Shell.getMidasContext().warning("Layer.data2la: data format "+data.getFormat()
                                     +" NOT supported for this conversion.");
    }

    return buf;
  }

  /** Chainable interface next=Line, also see {@link #getPrevLink()} */
  public Object getNextLink () { return line; }
  /** Chainable interface previous=MPlot, also see {@link #getNextLink()} */
  public Object getPrevLink () { return MP; }

  /** Gets the X-axis units for the layer. This defaults to the X-Units field in a data file's
      header or {@link Units#NONE} if not applicable.
   */
  public int getXUnits () { return xunits; }

  /** Gets the Y-axis units for the layer. This defaults to the Y-Units field in a data file's
      header or {@link Units#NONE} if not applicable.
   */
  public int getYUnits () { return yunits; }

  /** Gets the Z-axis units for the layer. This defaults to the Z-Units field in a data file's
      header or {@link Units#NONE} if not applicable.
   */
  public int getZUnits () { return zunits; }

  /** Gets the X-axis units for the layer. */
  public String getXUnitsName () { return DataFile.getUnitsVarName(getXUnits()); }

  /** Gets the Y-axis units for the layer. */
  public String getYUnitsName () { return DataFile.getUnitsVarName(getYUnits()); }

  /** Gets the Z-axis units for the layer. */
  public String getZUnitsName () { return DataFile.getUnitsVarName(getZUnits()); }

  /** Sets the X-axis units for the layer. This will not alter the header of the file being plotted.
      @param units The X-axis units for the layer (see {@link Units}).
   */
  public void setXUnits (String units) {
    setXUnits(DataFile.getUnitsID(units)); }

  /** Sets the Y-axis units for the layer. This will not alter the header of the file being plotted.
      @param units The Y-axis units for the layer (see {@link Units}).
   */
  public void setYUnits (String units) { setYUnits(DataFile.getUnitsID(units)); }

  /** Sets the Z-axis units for the layer. This will not alter the header of the file being plotted.
      @param units The Z-axis units for the layer (see {@link Units}).
   */
  public void setZUnits (String units) { setZUnits(DataFile.getUnitsID(units)); }

  /** Sets the X-axis units for the layer. This will not alter the header of the file being plotted.
      @param units The X-axis units for the layer (see {@link Units}).
   */
  public void setXUnits (int units) { xunits = units; }

  /** Sets the Y-axis units for the layer. This will not alter the header of the file being plotted.
      @param units The Y-axis units for the layer (see {@link Units}).
   */
  public void setYUnits (int units) { yunits = units; }

  /** Sets the Z-axis units for the layer. This will not alter the header of the file being plotted.
      @param units The Z-axis units for the layer (see {@link Units}).
   */
  public void setZUnits (int units) { zunits = units; }

  /** Gets the line used for drawing this layer. */
  public Line getLine () { return line; }

  /** Gets the symbol used for drawing this layer.
      @return The symbol used in drawing this layer. This will return <code>null</code> when the
              symbols are turned off, although the line (see {@link #getLine()}) may still have a
              pointer to the symbol.
   */
  public Symbol getSymbol () {
    if (getLine().isSymbolOn()) {
      return getLine().getSymbol();
    } else {
      return null;
    }
  }

  /** Gets the name of this layer.
      @return The name of this layer.
   */
  public String getName () { return name; }

  /** Gets a description of the layer.
      @return A description of the layer.
   */
  public String toString () {
    String className = getClass().getName();

    if (className.startsWith("nxm.sys.libg.")) {
      className = className.substring(13);
    }

    return className + ": " + getName();
  }

  /** Processes mouse movement messages. The default implementation of this does nothing with the
      messages, however some layers do (such as LayerT5 which uses it to support AutoSelect).
      @param msg  The message.
      @param type The message type ({@link nxm.sys.prim.plot#ONCLICK},
                                    {@link nxm.sys.prim.plot#ONDRAG},
                                    {@link nxm.sys.prim.plot#ONMOTION}).
      @since NeXtMidas 2.1.0
   */
  public void processMouseMessage (Message msg, int type) {
    /* Default implementation does nothing. */
  }

  /** Indicates if the layer supports a given view. By default this returns <code>true</code>
      for all views but some layers (most often image layers) will override this to return
      <code>false</code> for some views.
      @param view The view (see {@link MPlot#viewList}).
      @return <code>true</code> if the layer supports the given view, <code>false</code> if not.
      @since NeXtMidas 2.1.1
   */
  public boolean isViewSupported (int view) {
    return true;
  }

  /** Sets the comment for this layer. */
  public void setComment (String comment) {
    this.comment = comment;
  }

  /** Rescales the layer, ONLY called by MPlot. */
  void rescale () {
    setNeedsRescale(false);
    findRange();
  }

  /** Indicates if this layer needs to be rescaled. Only the individual layer can set this value. */
  public boolean getNeedsRescale () {
    return needsRescale;
  }

  /** Indicates if this layer needs to be rescaled. Only the individual layer can set this value. */
  protected void setNeedsRescale (boolean rescale) {
    needsRescale = rescale;
  }

  /** Gets the legend entry for the layer. Subclasses should not need to override this method.
      @return The top-level entry for the layer.
      @see #getLegendSubEntries()
      @since NeXtMidas 2.5.5
   */
  public GLegend.Entry getLegendEntry () {
    return legendEntry;
  }

  /** Gets the legend sub-entries for the layer. Subclasses should override this method if they have
      any sub-entries to add to the legend.
      @return The sub-entries for the layer. If there are no sub-entries, this method may return
              either null or an array with length 0.
      @see #getLegendEntry()
      @since NeXtMidas 2.5.5
   */
  public GLegend.Entry[] getLegendSubEntries () {
    return null;
  }
}
