package nxm.sys.libg;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

import java.util.*;
import java.util.Stack;
import java.awt.*;
import java.awt.image.BufferedImage;

import javax.imageio.IIOException;
import javax.imageio.ImageIO;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import org.xml.sax.*;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;

import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;


/**
  A class to hold and display a KML layer.
  For complete documentation on KML documents, see the Google KML reference.
  <br><br>
  This class consists of the Layer itself, the Parser, a class to hold the
  collection of KML objects parsed, and classes for each of the KML objects
  themselves. <br>
  <br>
  The process flow is as follows:<br>
    1. The user plots a *.kml file using the plot command.<br>
    2. Plot (and MPlot) open the file and call {@link #setFile(PlotFile, Table, Args)}<br>
    3. <code>setFile</code> calls the parser on the file.<br>
    4. The parser (which uses the Java SAX XML parser) reads the the file line
       by line. When it encounters a start tag for an element which is
       supported by the Layer, it does one of two things: <br>
       1) If the element is a 'parent' element, which means it has its own kml
       object class and contains child nested data, the parser creates a new kml
       object of that element type and puts it on the stack.<br>
       2) If the element only contains data (such as name or description),it
       creates a StringBuffer to hold the data, and that StringBuffer is put on
       the stack. Also, a flag is set to indicate that the next data to arrive
       needs to be stored. Data is characterized as any text not contained
       within element tags. This data is stored in the StringBuffer on the stack.<br>
    5. When the parser encounters a end tag for a supported element, it does one
       of two things: <br>
       1) If the element is a 'parent' element, with no parent above it (such as
       a document tag), it is popped off the stack and the object is placed in
       the kmlObjects class.<br>
       2) If the element has a parent element above it, the child element is
       popped off the stack and added to the parent kml Object (which is right
       above it on the stack) via a 'set' method.<br>
    6. When the parser reaches the end of the kml file (the closing kml tag), it
       sends the collection of kml Objects back to the Layer.<br>
    7. The Layer draw method is called automatically by MPlot. The Layer draw
       initiates the processing to break down the collection of kml Objects.
       Some objects are stored (such as kml Styles) for later use, while the
       drawable objects (kml Geometry objects such as linnear rings, polygons,
       points, etc) are drawn.<br>
       <br>
       <br>
       Current supported elements are found in {@link #supportedElements}.
       <br>

  @author James Bauer
  @since NeXtMidas 2.7.0
  @version $Id: LayerKML.java,v 1.55 2013/01/04 14:33:02 ntn Exp $
 */
public class LayerKML extends Layer {

  /** Coloring mode: Monochrome (1-color) (see {@link #coloringModes}). */
  public static final int COLORS_MONOCHROME = 1;
  /** Coloring mode: Mod 5 (NXM200 default) (see {@link #coloringModes}). */
  public static final int COLORS_MOD_5      = 2;
  /** Coloring mode: Mod N (N = num colors) (see {@link #coloringModes}). */
  public static final int COLORS_MOD_N      = 3;
  /** Coloring mode: Minimize adj with same color (see {@link #coloringModes}). */
  public static final int COLORS_MIN_ADJ    = 4;
  /** Coloring mode: Color selected item only (see {@link #coloringModes}). */
  public static final int COLORS_ON_SELECT  = 5;
  /** Default coloring mode. This may change between releases. */
  public static final int COLORS_DEFAULT = COLORS_MIN_ADJ;
  /** The coloring modes. */
  public static final String coloringModes = "Mono,Mod5,ModN,MinAdj,OnSelect";
  /** Alpha */
  public static final int OPAQ             = 255;
  /** The default border color */
  public static final Color DEFAULT_BORDER_COLOR = java.awt.Color.lightGray;
  /** The default point color */
  public static final Color DEFAULT_POINT_COLOR = java.awt.Color.WHITE;
  private Color colorBorder = DEFAULT_BORDER_COLOR;      /* Shape polygon border color */

  /** List of currently supported KML element tags.  */
  public String supportedElements = "kml,coordinates,point,polygon,placemark,linearring,"  // TODO: organize this better
    + "linestring,name,description,document,style,linestyle,"
    + "polystyle,color,width,bgcolor,extrude,url,"
    + "altitudemode,heading,iconstyle,icon,scale,"
    + "outerboundaryis,fill,outline,multigeometry,"
    + "visibility,networklinkcontrol,folder,styleurl,"
    + "networklink,link,refreshinterval,groundoverlay,"
    + "overlay,lookat,minrefreshperiod,refreshmode,"
    + "longitude,latitude,rotation,range,heading,"
    + "tilt,range,altitude,href,north,south,east,west,"
    + "labelstyle,balloonstyle,text,bgcolor,textcolor,displaymode,"
    + "viewformat";

  /* Types of Points */
  /** Point Type of Symbol = 1 */
  private final int SYMBOL = 1;
  /** Point Type of Image = 2 */
  private final int IMAGE  = 2;

  private final int      MAX_ICON_SIZE              = 32; // Max height/width (in pixels) for icon images
  private final int      DEFAULT_LINE_WIDTH         = 2;
  private final int      DEFAULT_TEXT_SIZE          = 12;
  private final int      DEFAULT_POINT_SIZE         = 8;
  private final float    DEFAULT_ICON_SCALE         = 1.0f;
  private final int      DEFAULT_ICON_HEADING       = 0;
  private final String   DEFAULT_POINT_SYMBOL       = "Cross";
  private final int      DEFAULT_NON_ZOOM_ICON_SIZE = 20;
  private final int      DEFAULT_ICON_ZOOM_SIZE     = 64;
  private final float    DEFAULT_COLOR_BLEND_RATIO  = .8f;
  private final String[] LABEL_POSITIONS            = {"TOPR","TOP","BOTTOMR","BOTTOM","RIGHT"};
  private final String   DEFAULT_LABEL_POS          = "TOPR";

  /** Max height/width (in pixels) for non-Zoomed icon images */
  private int  iconSize     = DEFAULT_NON_ZOOM_ICON_SIZE;
  /** Max height/width (in pixels) for Zoomed icon images */
  private int  iconZoomSize = DEFAULT_ICON_ZOOM_SIZE;

  /** Ratio of Original color to blended color when coloring Images.
      Must be between 0 and 1. Use a low number (e.g. 0.1f) to heavily
      blend (tint) the original and a high number (e.g. 0.9f) to highly blend.*/
  private float   colorBlendRatio = DEFAULT_COLOR_BLEND_RATIO;

  /** Where to place the label for points.
      <pre> Acceptable values are:
            TOPR    - Label text starts near top right corner of point. (default)
            TOP     - Label text is centered above point.
            BOTTOMR - Label text starts near bottom right corner of plot.
            BOTTOM  - Label text is centered below point.
            RIGHT   - Label text starts to the right of the point.
      </pre>
   */
  private String  labelPos = DEFAULT_LABEL_POS;

  private int     coloringMode     = COLORS_DEFAULT; // The coloring mode being used.
  private int     contrast         = 50;             // Percentage of contrast (0..100).
  private int     brightness       = 50;             // Percentage of brightness (0..100).
  private int     alpha            = -1;             // True alpha level (0..255). May be computed from the opaqueness.

  private   KmlParser        kmlParser;              // Instance of the parser
  protected kmlObjects       objects;                // (Supported) KML objects in the file
  protected Vector<kmlStyle> styles;                 // Keep track of kmlStyles for later reference

  /* The four variables below are for MPLOT's drawLine method, which is used for most drawing */
  private double[] buf;          // Holds the coordinates to be plotted
  private int      off  = 0;     // Starting point offset in buf. Usually 0.
  private int      dim  = 3;     // 3 = (x,y,z), 2 = (x,y). We usually use 3 (altitude default to 0).
  private int      npts;         // Number of points. This is (buf.length / 3)

  private boolean  polygonFill    = true;
  private boolean  polygonOutline = true;
  private Color    polygonColor   = DEFAULT_POINT_COLOR;
  private Color    lineColor      = DEFAULT_POINT_COLOR;
  private int      lineWidth      = DEFAULT_LINE_WIDTH;;
  private int      textSize       = DEFAULT_TEXT_SIZE;    // Size of the text to display next to the point (label)
  private float    labelScale     = 1;                    // Scale multiplier for labels next to points
  private Color    labelColor     = DEFAULT_POINT_COLOR;  // Color of text next to points
  private int      pointType      = SYMBOL;               // Default point type is SYMBOL
  private String   pointSymbol    = DEFAULT_POINT_SYMBOL; // When kml Icon is a known Symbol
  private BufferedImage             pointImage;           // When kml Icon is an image
  private int      pointSize      = DEFAULT_POINT_SIZE;
  private float    iconHeading    = DEFAULT_ICON_HEADING; // Default to North
  private float    iconScale      = DEFAULT_ICON_SCALE;   // Multiplier (x pointSize) for displaying Symbol/Icon
  private Color    iconColor      = DEFAULT_POINT_COLOR;
  private boolean  symbolFill     = true;
  private boolean  showLabel      = true;         // whether or not to draw the label next to the point
  private double[] pollRate;                      // 0.0 means no polling
  private double[] lastPoll;                      // time when linked file was last polled
  private double   minRefreshPeriod = 0;          // override using pollrate
  private Table    points;                        // All points that have been drawn on. Used in MousePopups
  private int      pointCount     = 0;            // index of kmlPoints
  private String   currDesc;                      // description of current feature being drawn
  private String   currPath;                      // path to where the KML File resides
  private kmlBalloonStyle balloonStyle;           // the current kmlBalloonStyle
  private boolean  allowPopups    = true;         // allow balloon popups for KMML points
  private FileName fn;

  /** files used when using network polling */
  private KmlFile[] linkedFile     = null;
  /** ignore poll interval and pollrate for the first draw per file */
  private boolean[]  isFirstPass;
  /** Refresh the plot even if the file modified date has not changed (e.g. cgi scripts) */
  private boolean forceRefresh   = false;
  protected static final WeakHashMap<String,Cache.TimeStampedValue> SHARED_VCACHE = new WeakHashMap<String,Cache.TimeStampedValue>();  // Shared image cache.
  protected final  Cache            imgCache        = getDefImgCache();    // The image cache to use.
  private          int              imgCacheSize    = -1;                  // Size of image cache (-1=UseDefault)

  /** Default constructor. */
  public LayerKML() {
    super();

    /* Set up the Line to be used for drawing */
    line.setRadius(3);
    line.setSymbol(getPointSymbol());
    line.getSymbol().setSize(getPointSize());
    line.setType(Line.SOLID);
    line.setColor(coloringMode);

    /* Initialize objects */
    kmlParser    = new KmlParser();
    objects      = new kmlObjects();
    styles       = new Vector<kmlStyle>();
    currDesc     = new String();

    setCacheSize(-1);  // set cache size default to -1

    points = new Table();

    // make sure there is at least one boolean in the array
    isFirstPass = new boolean[1];
    isFirstPass[0] = true;

    // Set initial values (over-ridden by coordinate values)
    x1 =  180;   // max possible x value
    x2 = -180;   // min possible x value
    y1 =  90;    // max possible y value
    y2 = -90;    // min possible y value
    z1 = 0.0;    // default
    z2 = 10000;  // default
  }

  /** Creates a new instance.
      @param cmap The color map to use.
   */
  public LayerKML (int cmap) {
    this();
    setColorMap(cmap);
    name = "KML";
  }

  /** Extracts the filename and calls the parser */
  public boolean setFile (PlotFile pf, Table tab, Args MA) {
    boolean status = true;
    fn = FileName.toFileName(pf.getName());
    currPath =  fn.getPath();  // save path for later use when making URLs for local files

    /* Check to see if the file is a KML file. */
    status = checkFileType(pf);

    /* If it is a kml file, parse it */
    if (status) {
      super.setFile(pf);
      if (tab != null) {
        int iconSize = (tab.getL("ICONSIZE", 0));
        if (iconSize > 0) this.setIconSize(iconSize);

        int iconZoomSize = (tab.getL("ICONZOOMSIZE", 0));
        if (iconZoomSize > 0) this.setIconZoomSize(iconZoomSize);

        float ratio = (tab.getF("COLORBLENDRATIO", -1.0f));
        if (ratio >= 0) this.setColorBlendRatio(ratio);

        String pos = (tab.getS("LABELPOS", null));
        if (pos != null) setLabelPos(pos);

        boolean forceRefresh = (tab.getState("FORCEREFRESH",this.forceRefresh));
        setForceRefresh(forceRefresh);

        boolean allowPopups = (tab.getState("ALLOWPOPUPS",this.allowPopups));
        setAllowPopups(allowPopups);
      }
      pf.open(1);  // Open as input
      InputStream is = ((KmlFile)pf).getResource().getInputStream(); // Get file data as InputStream
      parseFile(is);                      // Parse the file

    } else {
      Shell.warning("File " +pf.getName()+ " is not recognized as a KML file");
    }
    return status;
  }

  /** Check the file type to see if it is KML */
  private boolean checkFileType (PlotFile pf) {
    if (pf instanceof KmlFile) return true;
    else return false;
  }

  /** Calls the kmlParser on the file */
  private void parseFile (InputStream fname) {
    try {
      getKmlParser().parseKmlFile(fname);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /** Processes mouse movement messages to create balloon popus.
      @param msg  The Mouse message
      @param type Type of message
      @since NeXtMidas 2.7.2
   */
  public void processMouseMessage(Message msg, int type) {
    Position pos = (Position)msg.getData();
    if (msg.name.equals("HOVER")) {
       if (!testForPopup(pos)) super.processMouseMessage(msg, type);
    }
  }

  public void setAllowPopups (boolean val) { allowPopups = val; }

  public boolean isAllowPopups () { return allowPopups; }

  /** Gets the closest point to the given Position and displays a <code>Popup</code> for it */
  private boolean testForPopup (Position pos) {
    if (!allowPopups) return false;
    int range = 2;  // how close the mouse must be to the point
    Table tmp;
    for (float step = 0; step <= range; step += .1) {   // spiral out from point of mouse click
      for (int index = 1; index <= pointCount; index++) {
        tmp = points.getTable(Integer.toString(index));
        if (tmp == null ) continue;

        double x = tmp.getD("x");
        if (x > (pos.x - range) && x < (pos.x + range)) {
          double y = tmp.getD("y");
          if (y > (pos.y - range) && y < (pos.y + range)) {

            // Do not popup if the BalloonStyle displayMode is set to 'hide'
            kmlBalloonStyle style = (kmlBalloonStyle)tmp.getO("balloonstyle");
            if (style != null) {
              String mode = style.getDisplayMode();
              if (mode != null && mode.equalsIgnoreCase("hide")) return false;
            }
            // The next two lines find the mouse pos
            MPoint p = MP.getPoint(pos);
            Point p2 = MP.panel.getLocationOnScreen();

            Popup pu = new Popup(tmp, (p.x+p2.x), (p.y+p2.y));  // popup a frame showing info about the kmlPoint
            return true;  // only get the closest point to the mouse
          }
        }
      }
    }
    return false;  // no point in range
  }

  /** Check {@link #linkedFile} to see if it has been modified. */
  private boolean linkedFileModified (KmlFile linkedFile, double pollRate, int index) {
    if (isFirstPass[index]) { return true; }
    long lastMod = -1;
    URL url = null;
    try {
      url = new URL(linkedFile.getURL());
    } catch (MalformedURLException e1) {
      e1.printStackTrace();
    }
    if (url != null) lastMod     = FileUtil.lastModified(url);

    if ((now() - lastMod) < pollRate){
      return true;
    } else return false;
  }

  /** Gets the current time after adjusting for 1970/1950 epoch */
  private double now () {
    return Time.current() - Time.J1970TOJ1950;
  }

  /** Calls setFile (which will parse and plot a file) on the file set in
      {@link #handleKmlNetworkLink(kmlNetworkLink, int)}
   */
  private void pollData (KmlFile linkedFile, double pollRate, int index) {
    if (linkedFile != null) {
      if (linkedFileModified(linkedFile, pollRate, index) || forceRefresh) {
        if (index>0) {    // not the 'base' kml file
          boolean newLayer = true;
          Layer   lay;
          PlotFile file = null;
          for (int idx=0; idx<MP.layers.size(); idx++) {
            lay = (Layer)MP.layers.get(idx);
            if (lay != null) file = lay.getFile();
            if (file != null && linkedFile.equals(file)) newLayer = false;
          }
          if (newLayer) {
            linkedFile.setQualifier("FG", "KML");   // tell Plot to open as a KML file ...
            MP.mh.processMessage(new Message("OPENFILE",0,linkedFile.getName().toString()+linkedFile.getQualifiers()));
          }
        } else setFile(linkedFile, null, null);
        isFirstPass[index] = false;
        super.needRefresh = true;  // refreshes the display with the new data
      }
    }
  }

  /** Checks the Poll Rate and calls pollData if the rate has expired */
  public int getData (int flag) {
    if (linkedFile != null) {
      for (int idx=0; idx<linkedFile.length; idx++) {
        if (Time.current() >= (pollRate[idx] + lastPoll[idx])) {
          pollData(linkedFile[idx], pollRate[idx], idx);
          lastPoll[idx] = Time.current();
        }
      }
    }
    return Drawable.FULL;
  }

  /** Draw the KML objects from the file
      This method is called by MPLOT and represents the 'Launch Pad' for the
       drawing and handling of KML Objects.
   */
  public void draw (int mode) {
    super.draw(mode);
    points = new Table();        // clear out references to previously drawn kmlPoints
    handleKmlObjects(objects);   // Break apart the KML objects and draw them
  }

  /** This method is called by the <code>drawPolygon</code> methods since the
       setup for the Line is the same for each of them.
      @return The Line to draw.
   */
  private Line getPolyLine () {
    Line myLine = line.copy();

    Color oldColor = myLine.getColor();        // Current Color
    Color newColor = getLineColor();           // Color specified in kml LineStyle, [def=white]

    /* Check for NULL (should not happen because def=white) and check against
        current color (to avoid needless refresh)  */
    if (newColor != null) {
      if (!oldColor.equals(newColor)) myLine.setColor(newColor);
    }

    oldColor = myLine.getFillColor();     // Current fill color
    newColor = getPolygonColor();         // Fill color specified in kml PolyStyle, [def=white]

    /* Check for NULL (should not happen because def=white) and check for
        polygonFill is true before setting fill color and Line type to 'FILL' */
    if (newColor != null && isPolygonFill() == true) {
      if (!oldColor.equals(newColor)) myLine.setFillColor(newColor);
      myLine.setType(myLine.getType()+"Fill");   // Fill in the polygons if polygonFill is true
    }

    /* Set the thickness */
    if (myLine.getThickness() != getLineWidth()) myLine.setThickness(getLineWidth());

    return myLine;
  }

  /** Draw a kml Point */
  private int drawPoint (kmlPoint point) {
    int status = 0;
    npts = 1;   // Only one point at a time

    /* Load the coordinates of the point into a double[] */
    buf    = new double[3];
    buf[0] = point.getCoords().getLon();
    buf[1] = point.getCoords().getLat();
    buf[2] = point.getCoords().getAltitude();

    Line myLine = line.copy();                   // Copy the Line for drawing

    /* Line and Symbol setup. These are adjusted in the kml source file using the
        LineStyle and IconStyle tags.  */
    myLine.setType("Symbol");                    // Set the type to Symbol
    BufferedImage drawableImage = null;    // mods done to new image to preserve old one
    if (pointType == IMAGE && pointImage==null) {
      setPointType(SYMBOL);
      setPointSymbol(DEFAULT_POINT_SYMBOL);
    }
    if (pointType == IMAGE) {
      drawableImage = getDrawableImage();  // mods done to new image to preserve old one
      if (iconColor != null) applyPointImageColor(drawableImage);
      if (iconScale != 1.0f) drawableImage = scalePointImage(drawableImage, iconScale);  // apply scale to image icon, if necessary

      if (MP.getZoomLevel() == 0) drawableImage = adjustPointImageSize(drawableImage, iconSize);
      else                        drawableImage = adjustPointImageSize(drawableImage, iconZoomSize);
      myLine.setSymbol(drawableImage);          // If an icon, just set the symbol to the image
      setPointSize(drawableImage.getHeight(null));
    } else {
      myLine.setSymbol(getPointSymbol());          // Set the Symbol type
      myLine.setSymbolFill(isSymbolFill());
      myLine.setColor(getIconColor());
      myLine.setThickness(getLineWidth());
      int symbolSize = (int)(getPointSize() * getIconScale());
      if (symbolSize > MAX_ICON_SIZE) symbolSize = MAX_ICON_SIZE;  // prevent giant symbols
      myLine.getSymbol().setSize(symbolSize); // Set the size of the Symbol
    }

    myLine.setSymbolHeading(getIconHeading());

    MP.drawLineLonLat(buf, off, dim, npts, myLine);  // Draw the point

    /* Display the Name of the point next to it */
    String label = point.getParentFeature().getName();      // The 'name' tag is one level higher
    if (isShowLabel() && label != null && MP.getZoomLevel() > 0) {
      MPoint pix = new MPoint();
      MP.getPoint(buf[0], buf[1], buf[2], pix);
      //MP.drawString(label, pix.x, pix.y, textSize, MP.RIGHT);

      // Use Graphics2D to work around bug where label would not display if
      // any map was plotted underneath it (which there usually is).
      Graphics2D g2d = (Graphics2D)MP.gc;
      g2d.setColor(this.getLabelColor());
      Font f = new Font (label, MP.font.getStyle(), (int)(textSize * labelScale));
      g2d.setFont(f);
      pix = adjustLabelPosition(g2d, f, label, labelPos, pix, myLine.getSymbolSize());
      g2d.drawString(label, pix.x, pix.y);
    }

    saveKmlPointData (buf[0],buf[1],label,myLine,currDesc,balloonStyle,pointType == IMAGE? pointImage: null);

    return status;
  }

  /** Save the information about a drawn kmmlPoint to display later in a popup
      @param x The lat
      @param y The lon
      @param label The name for the point
      @param myLine The Line used when drawing the point
      @param balloonStyle The kmlBalloonStyle with data about the point
      @param img The BufferedImage of the point, if any
      @since NeXtMidas 2.7.2
   */
  private void saveKmlPointData (double x, double y, String label,
      Line myLine, String desc, kmlBalloonStyle balloonStyle, BufferedImage img) {

    boolean addData = true;
    for (int index = 0; index < pointCount; index++) {
      Table tmp = points.getTable(Integer.toString(index));
      if (tmp != null && tmp.getD("x") == x && tmp.getD("y") == y) {
        addData = false;    // avoid adding the point twice (happens when zooming)
        break;
      }
    }

    // Put the point data into the points table for later use
    if (addData) {
      Table data = new Table();
      data.put("x",     buf[0]);
      data.put("y",     buf[1]);
      data.put("label", label);
      data.put("line",  myLine);
      data.put("img",   img);
      data.put("desc",  desc);
      data.put("balloonstyle", balloonStyle);
      pointCount++;      // point index becomes the key for the table
      String key = Integer.toString(pointCount);
      points.put(key, data);
    }
  }

  /** Draws a polygon from whatever is in the buffer */
  private void drawPolygon () {
    drawPolygon(buf);
  }

  /** Draw a Polygon on the layer when given coordinate set
      This is the simplest drawPolygon method. No checks are made on the data,
       so it should be called after the checks are made.
      NOTE: This method assumes an offset of 0 and dim of 3.
   */
  private void drawPolygon (double[] coords) {

    buf      = coords;
    npts     = (coords.length / 3);
    off      = 0;
    dim      = 3;

    Line myLine = getPolyLine();  // Get the Line to draw

    MP.drawLineLonLat(buf, off, dim, npts, myLine);  // draw the Polygon
  }

  /** Draw a kml Polygon
      This is the first method called when handling the drawing of a kml polygon.
      This method makes some decisions about the polygon and then passes it on
       to other draw methods.
   * */
  private void drawPolygon (kmlPolygon polygon) {

    int rings = polygon.innerBoundary.size();   // Number of KML Linear Rings in the Inner Polygons

    if (!isPolygonFill()) {
      drawPolygonOutlines(polygon, rings);   // Polygon not filled. (Contours or rings)
    } else {
      if (rings > 1) {
        buf = combinePolygons(polygon, rings);        // Multiple inner rings
      } else if (rings > 0) {
        buf = this.combineSinglePolygons(polygon);    // Single inner ring
      } else {
        buf = polygon.getOuterBoundary().getCoordsAsDouble();  // No inner rings
      }

      if (polygon.isExtrude()) {
        drawExtrudedPolygon(buf);   // extruded polygon
      } else  drawPolygon (buf);     // non-extruded polygon (kml default)
    }
  }

  /** Draw KML Polygon that is not filled.
      Since the polygon is not filled, we can actually draw the inner rings as
       separate polygons, which eliminates the line connecting the outer and
       inner rings found in Filled polygons with inner rings.
     @param polygon The KML Polygon
     @param innerRings Number of inner rings (holes)
   */
  private void drawPolygonOutlines (kmlPolygon polygon, int innerRings) {

    /* Parameters for the drawLine method */
    buf = polygon.getOuterBoundary().getCoordsAsDouble();

    /* Draw the outer polygon */
    if (polygon.isExtrude()) {
      drawExtrudedPolygon(buf);  // Extruded
    } else  drawPolygon(buf);    // Non-extruded (default)

    /* Draw the inner polygons */
    if (innerRings > 0) {
      for (int xx=0; xx<innerRings; xx++) {
        buf = ((polygon.innerBoundary.get(xx))).getCoordsAsDouble();
        if (polygon.isExtrude()) {
          drawExtrudedPolygon(buf);   // extruded
        }
        else drawPolygon(buf);        // non-extruded (default)
      }
    }
  }

  /** This method uses an algorithm that combines the coordinate sets but leaves
       a 'path' to retrace. This path reduces the chance that a polygon will
       draw over top of an existing inner ring. It is definitely not a perfect
       algorithm, but should eliminate most occurrences.
      NOTE: If the line color (set in the kml LineStyle object) is a different
       color than the fill (set in the kml PolyStyle object), then you will
       see a line connecting the outer and inner polygons.
     @param polygon The KML Polygon
     @param innerRings The number of inner rings
     @return The double[] array of (non-extruded) coordinates to draw.
   */
  private double[] combinePolygons (kmlPolygon polygon, int innerRings) {

    /* Outer Polygon (linear ring) coordinates */
    double[] outerCoords    =  polygon.getOuterBoundary().getCoordsAsDouble();
    int outerLength =  outerCoords.length;

    if (innerRings == 0) return outerCoords;    // If no inner rings, just need outer

    double[] coords;
    int totalLength    = 0;
    int tmpPos         = 0;
    int innerLength    = 0;
    int endPathCounter = 0;

    /* Determine the combined length (# of coords) for all the inner rings */
    for (int xx=0; xx< innerRings; xx++) {
      innerLength += (polygon.innerBoundary.get(xx)).getCoordsAsDouble().length;
    }

    /* Add an extra coord set between the inner rings */
    totalLength = outerLength + innerLength + (3 * (innerRings-1));

    coords = new double[totalLength];  // Set the return coords size

    double[] endPath = new double[(innerRings-1) * 3];

    for (int xx=0; xx<outerLength; xx++, tmpPos++) {
      coords[tmpPos] =  outerCoords[xx];            // Append the outer coordinates
    }

    for (int xx=0; xx< innerRings; xx++) {
      double[] tmpCoords = (polygon.innerBoundary.get(xx)).getCoordsAsDouble();
      for (int yy=0; yy<tmpCoords.length; yy++, tmpPos++) {
        coords[tmpPos] = tmpCoords[yy];        // Append the inner coordinates
      }
      if (xx < (innerRings -1)) {
        endPathCounter = ((endPath.length) - (3 * xx) -3);       // Set the retrace path
        endPath[endPathCounter]    = coords[tmpPos-3];  // lon
        endPath[endPathCounter +1] = coords[tmpPos-2];  // lat
        endPath[endPathCounter +2] = coords[tmpPos-1];  // alt
      }
    }

    for (int xx=0; xx< endPath.length; xx++, tmpPos++) {
      coords[tmpPos] =  endPath[xx];                 // Append the retrace path
    }

    return coords;
  }

  /** This method is used for the case of a polygon with a single
       inner ring, since that case allows for a faster algorithm.
      NOTE: If the line color (set in the kml LineStyle object) is a different
       color than the fill (set in the kml PolyStyle object), then you will
       see a line connecting the outer and inner polygons.
      @param polygon The <code>kmlPolygon</code>
      @return The combined coordinates of the two rings (outer + inner)
   */
  private double[] combineSinglePolygons (kmlPolygon polygon) {
    double[] coords;
    int innerRings       = polygon.innerBoundary.size();

    /* Send to other method if multiple inner rings */
    if (innerRings > 1) {
      coords = combinePolygons(polygon, innerRings);
    } else {
      /* Combine the two polygons, starting with the outer */
      double[] outerCoords = polygon.getOuterBoundary().getCoordsAsDouble();
      double[] innerCoords = (polygon.innerBoundary.get(0)).getCoordsAsDouble();
      int length = (outerCoords.length + innerCoords.length);
      coords = new double[length];
      int counter = 0;
      for (int xx=0; xx<outerCoords.length; xx++, counter++) {
        coords[counter] = outerCoords[xx];
      }
      for (int xx=0; xx<innerCoords.length; xx++,counter++) {
        coords[counter] = innerCoords[xx];
      }
    }
    return coords;
  }

  /** This method provides no real attempt to prevent the polygon from filling
       incorrectly. This method simply combines all the polygon rings into
       a single polygon.
      NOTE: If the line color (set in the kml LineStyle object) is a different
       color than the fill (set in the kml PolyStyle object), then you will
       see a line connecting the outer and inner polygons.
     @param polygon The <code>kmlPolygon</code>
     @param innerRings Number of inner boundaries
     @return Single double array of coordinates to draw.
   */
  private double[] combinePolygonsFast (kmlPolygon polygon, int innerRings) {
    double[] coords;

    if (innerRings == 0) return polygon.getOuterBoundary().getCoordsAsDouble();

    /* Combine the two polygons, starting with the outer */
    double[] outerCoords = polygon.getOuterBoundary().getCoordsAsDouble();
    int innerLength = 0;
    int innerCounter = 0;
    for (int xx=0; xx< innerRings; xx++) {
      innerLength +=  (polygon.innerBoundary.get(xx)).getCoordsAsDouble().length;
    }
    double[] innerCoords = new double[innerLength];
    for (int xx=0; xx<innerRings; xx++) {
      double[] tmpCoords = (polygon.innerBoundary.get(xx)).getCoordsAsDouble();
      for (int yy=0; yy< tmpCoords.length; yy++, innerCounter++) {
        innerCoords[innerCounter] = tmpCoords[yy];
      }
    }

    int length = (outerCoords.length + innerLength);  // 3 added to return to start point
    coords = new double[length];
    int counter = 0;
    for (int xx=0; xx<outerCoords.length; xx++, counter++) {
      coords[counter] = outerCoords[xx];
    }
    for (int xx=0; xx<innerCoords.length; xx++,counter++) {
      coords[counter] = innerCoords[xx];
    }
    return coords;
  }

  /** Draw a kml Line String */
  private int drawLineString (kmlLineString lineString) {
    int status = 1;
    buf = lineString.getCoordsAsDouble();  // Get the Line String coordinates in double[] form

    npts = (buf.length / 3);           // 3 values (x,y,z) per point
    dim = 3;
    off = 0;
    Line myLine = line.copy();             // Line to draw
    Color currColor = myLine.getColor();
    Color newColor = getLineColor();
    if (!currColor.equals(newColor)) myLine.setColor(newColor);
    if (myLine.getThickness() != getLineWidth()) myLine.setThickness(getLineWidth());

    MP.drawLineLonLat(buf, off, dim, npts, myLine);    // Draw the lines
    return status;
  }

  /** Draw a kml Linear Ring */
  private int drawLinearRing (kmlLinearRing linearRing) {
    int status =1;
    return status;
  }

  /** Set the poll rate used when polling external file */
  public void setPollRate (double rate) {
    setPollRate(0,rate);
  }

  /** Set the poll rate used when polling external files
      @param index The index of the file (if it's a linked file)
      @param rate  The interval between poll cycles
   */
  public void setPollRate (int index, double rate) {
    if (rate >= minRefreshPeriod) pollRate[index] = rate;
  }

  /** Get the poll rate used when polling external file */
  public double getPollRate () {
    return getPollRate(0);   // base file
  }

  /** Get the poll rates used when polling external file */
  public double[] getPollRates () {
    return pollRate;
  }

  /** Get the poll rate used when polling external file
      @param index The index of the file to get the rate for
   */
  public double getPollRate (int index) {
    return pollRate[index];
  }

  /** Set whether to just draw the outline of the polygon */
  public void setPolygonOutline (boolean outline) {
    polygonOutline = outline;
  }

  /** Get whether to just draw the outlines of the polygons */
  public boolean getPolygonOutline () {
    return polygonOutline;
  }

  /** Convenience method */
  public boolean isPolygonOutline () {
    return getPolygonOutline();
  }

  /** set whether to fill the polygon */
  public void setPolygonFill (boolean doFill) {
    polygonFill = doFill;
  }

  /** get whether to fill the polygon */
  public boolean getPolygonFill () {
    return polygonFill;
  }

  /** Convenience method */
  public boolean isPolygonFill () {
    return getPolygonFill();
  }

  /** Sets the color of the <code>kmlPolygon</code> fill.
      NOTE: This method does not call refresh(), since setting the color of
       the polygon is called when the <code>kmlPolyStyle</code> tag is
       handled, which is before the actual drawing of the objects.
      @param color The new color
   */
  public void setPolygonColor (Color color) {
    polygonColor = color;
  }

  /** Get the color of the fill for <code>kmlPolygon</code>s. */
  public Color getPolygonColor () {
    return polygonColor;
  }

  /** Set the color of the lines (or borders for polygons)
      NOTE: This method does not call refresh(), since setting the color of
       the lines is called when the <code>kmlLineStyle</code> tag is
       handled, which is before the actual drawing of the objects.
   */
  public void setLineColor (Color color) {
    lineColor = color;
  }

  /** Get the color of the lines (or borders for polygons) */
  public Color getLineColor () {
    return lineColor;
  }

  /** Sets the width of the lines (or borders)
      NOTE: This method does not call refresh(), since setting the color of
       the lines is called when the <code>kmlLineStyle</code> tag is
       handled, which is before the actual drawing of the objects.
   */
  public void setLineWidth (int width) {
    lineWidth = width;
  }

  public int getLineWidth () {
    return lineWidth;
  }

  /** Set the scale multiplier of the text displayed next to points */
  public void setLabelScale (float scale) { labelScale = scale; }

  public float getLabelScale () { return labelScale; }

  public void setLabelColor (Color color) { labelColor = color; }

  public Color getLabelColor () { return labelColor; }

  /** Set the size of the text displayed next to points */
  public void setTextSize (int size) { textSize = size; }

  /** Get the size of the text displayed next to points */
  public int getTextSize () { return textSize; }

  public boolean isShowLabel () { return showLabel; }

  public void setShowLabel (boolean show) { showLabel = show; }

  public void setPointType (int type) { pointType = type; }
  public int getPointType ()  { return pointType; }

  /** Get the type of kmlPoint (SYMBOL or IMAGE) in the given kmlStyle id
      @param styleId The id attribute of the Style
      @return The type of point (IMAGE or SYMBOL)
   */
  private int getPointType (String styleId) {
    if (this.styles.size() > 0) {
      for (int xx=0; xx<styles.size(); xx++) {
        if ((styles.get(xx)).getId().equalsIgnoreCase(styleId)) {
          String href = (styles.get(xx)).iconStyle.icon.getHref().toLowerCase();
          if (href.startsWith("http")  ||
              href.startsWith("file")  ||
              isImage(href))     {
            return IMAGE; //TODO Make this smarter
          }
        }
      }
    }
    return SYMBOL;
  }

  private boolean isImage(String filename) {
    filename = filename.toLowerCase();
    if (filename.endsWith("gif")  ||
        filename.endsWith("tif")  ||
        filename.endsWith("tiff") ||
        filename.endsWith("jpg")  ||
        filename.endsWith("jpeg") ||
        filename.endsWith("bmp")  ||
        filename.endsWith("wbmp") ||
        filename.endsWith("pdf")) {
      return true;
    } else {
      return false;
    }
  }

  /** Get the current Symbol type being used */
  public String getPointSymbol () {
    return pointSymbol;
  }

  /** Set the Symbol and Symbol Type type */
  public void setPointSymbol (String symbol) {
    if (Symbol.symbolList.indexOf(symbol) >= 0) {
      pointSymbol = symbol;
      setPointType(SYMBOL);
    } else {
      Shell.warning("Invalid symbol: " +symbol+ ". Symbol must be in " +Symbol.symbolList);
    }
  }

  /** Get the Image being used for kml Points */
  public Image getPointImage () {
    return pointImage;
  }

  /** Set the Image being used for kml Points */
  public void setPointImage (URL url) {
    pointImage = addToImgCache(url.toString());
    setPointType(IMAGE);  // Set the type to IMAGE
  }

  /** Set the Image being used for kml Points */
  public void setPointImage (String urlString) {
    pointImage = addToImgCache(urlString);
    setPointType(IMAGE);  // Set the type to IMAGE
  }

  /** Checks the image cache to see if the image already exists there. If not, it
      adds it.
      @param urlString The url of the image (in String form)
      @return The BufferedImage
      @since NeXtMidas 2.7.3
   */
  private BufferedImage addToImgCache (String urlString) {
    BufferedImage img = (BufferedImage)imgCache.get(urlString);  // first check to see if it is already there

    if (img == null) {
      URL url = getURL(urlString);
      try {
        img = ImageIO.read(url);
        if (img != null) imgCache.put(urlString, img);
      } catch (IIOException ioe) {
        Shell.warning("Unable to retrieve image from url: ["+urlString+"]");
      } catch (Exception e) {
        Shell.printStackTrace(e);
      }
    }
    return img;
  }

  /** Set the max size (in pixels) that an un-zoomed Icon can be.
      The same number applies to both icon Height and Width.
      @since NeXtMidas 2.7.1
   */
  public void setIconSize (int size) { iconSize = size; }

  public int getIconSize () { return iconSize; }

  /** Set the max size (in pixels) that a zoomed Icon can be.
      The same number applies to both icon Height and Width.
      @since NeXtMidas 2.7.1
   */
  public void setIconZoomSize (int size) { iconZoomSize = size; }

  public int getIconZoomSize () { return iconZoomSize; }

  /** Set whether or not to ignore re-draws ONLY on the file changing .
      @since NextMidas 2.7.1
   */
  public void setForceRefresh ( boolean val) { forceRefresh = val; }
  public boolean getForceRefresh () { return forceRefresh; }

  /** This is used internally to reduce the size of the icon if zoomed out. */
  private BufferedImage adjustPointImageSize (BufferedImage img, int maxSize) {
    if (img != null) {
      float h = img.getHeight();
      float w = img.getWidth();

      boolean adjustH = (h > maxSize);
      boolean adjustW = (w > maxSize);

      // Do the check below to avoid scaling the icon twice if height and width are both over the max
      if (adjustH && !adjustW) img = scalePointImage(img, ((float)maxSize / h));
      else if (adjustW)        img = scalePointImage(img, ((float)maxSize / w));
    }
    return img;
  }

  /** Blend the color given in iconStyle with color already in the image
      @param img The BufferedImage to apply the icon color to
      @since NeXtMidas 2.7.1
   */
  private void applyPointImageColor (BufferedImage img) {
    applyPointImageColor(img, iconColor);
  }

  /** Blend the color given in iconStyle with color already in the image
      @param img The BufferedImage to apply the icon color to
      @param color The color from iconStyle
      @since NeXtMidas 2.7.1
   */
  private void applyPointImageColor (BufferedImage img, Color color) {
    if (color != null && img != null) {
      MColor.colorImage(img, color, colorBlendRatio);
    }
  }

  public MPoint adjustLabelPosition (Graphics g2d, Font f, String label, String pos, MPoint pix, int iconSize) {
    int posx = pix.x; // x pos of label
    int posy = pix.y; // y pos of label
    FontMetrics metrics = g2d.getFontMetrics(f);
    int textWidth  = metrics.stringWidth(label);
    int textHeight = metrics.getHeight();
    if (pos.equals("TOP")) {
      posx = pix.x - (textWidth/2);
      posy = pix.y - (iconSize/2) - 1;
    } else if (pos.equals("TOPR")) {
      posx = pix.x + (iconSize/2);
      posy = pix.y - (iconSize/2) - 1;
    } else if (pos.equals("BOTTOM")) {
      posx = pix.x - (textWidth/2);
      posy = pix.y + (iconSize/2) + textHeight;
    } else if (pos.equals("BOTTOMR")) {
      posx = pix.x + (iconSize/2);
      posy = pix.y + (iconSize/2) + textHeight;
    } else if (pos.equals("RIGHT")) {
      posx = pix.x + (iconSize/2);
      posy = pix.y + textHeight/2;
    }
    MPoint newPoint = new MPoint(posx, posy);
    return newPoint;
  }

  public String getLabelPos () { return labelPos; }

  public void setLabelPos (String pos) {
    if (pos == null) return;
    pos = pos.toUpperCase();
    for (int index=0; index < LABEL_POSITIONS.length; index++) {
      if (LABEL_POSITIONS[index].equals(pos)) { labelPos=pos; return; }
    }
    String labelString = "'";
    for (int index=0; index < LABEL_POSITIONS.length; index++) {
      labelString += (LABEL_POSITIONS[index] + "', '");
    }
    Shell.warning("Invalid Label Position: " +pos+". Position must be in "
        +labelString.substring(0, labelString.length()-3)+".");
  }

  /** Makes a copy of the current point image. This is used to prevent over-
      blending on re-draws if using a network link file
      @return The image to draw
   */
  private BufferedImage getDrawableImage () {
    if (pointImage==null) return null;  // Bug 2360
    int w = pointImage.getWidth();
    int h = pointImage.getHeight();

    BufferedImage newImg = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
    Graphics2D g = newImg.createGraphics();
    g.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
        RenderingHints.VALUE_INTERPOLATION_BILINEAR);
    g.drawImage(pointImage, 0, 0, w, h, 0, 0, w, h, null);
    g.dispose();
    return newImg;
  }

  /** Apply a given scale to an Image .
      @param scale The factor to scale
   */
  private BufferedImage scalePointImage (BufferedImage img, float scale) {
    if (scale != 1.0f && img != null) {
      int oldW = img.getWidth();
      int oldH = img.getHeight();
      int newH = (int)(oldH * scale);
      int newW = (int)(oldW * scale);
      BufferedImage scaled = new BufferedImage(newW, newH, BufferedImage.TYPE_INT_ARGB);
      Graphics2D g = scaled.createGraphics();
      g.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
          RenderingHints.VALUE_INTERPOLATION_BILINEAR);
      g.drawImage(img, 0, 0, newW, newH, 0, 0, oldW, oldH, null);
      g.dispose();
      return scaled;
    } else {
      Shell.warning("Unable to scale image due to NULL image or scale factor of 100%");
      return img;
    }
  }


  /** Apply the scale factor to an Image Icon */
  public void scalePointImage () {
    if (iconScale != 1.0f && pointImage != null) {
      pointImage = scalePointImage(pointImage, iconScale);
    }
  }

  /** Get the size of the KML Points (Symbols) */
  public int getPointSize () {
    return pointSize;
  }

  public void setColorBlendRatio (float ratio) {
    if (ratio <= 1.0f && ratio >= 0.0f) colorBlendRatio = ratio;
    else Shell.warning("Invalid Image Color ratio of " +ratio
        +". Ratio must be between 0 and 1");
  }

  public float getColorBlendRatio () { return colorBlendRatio; }

  /** Set the size of the KML Points (Symbols) */
  public void setPointSize (int pointSize) {
    this.pointSize = pointSize;
  }

  /** Get the direction the Symbol is facing */
  public float getIconHeading () {
    return iconHeading;
  }

  /** Set the direction the Symbol is facing */
  public void setIconHeading (float heading) {
    iconHeading = heading;
  }

  /** Get the scale (Magnification) of the current Symbol */
  public float getIconScale () {
    return iconScale;
  }

  /** Set the scale (Magnification) of the current Symbol */
  public void setIconScale (float scale) {
    iconScale = scale;
  }

  public void setIconColor (Color color) { iconColor = color; }

  public Color getIconColor () { return iconColor; }

  /** Gets the appropriate image cache to use for the layer. */
  private Cache getDefImgCache () {
    return new Cache(Cache.SLRU, 4, 3600.0, SHARED_VCACHE);
  }

  /** Sets the size of the local image cache.
      @param size The cache size or -1 if the size should be set automatically.
   */
  public void setCacheSize (int size) {
    setCacheSize(size,0,4);
  }

  /** Sets the size of the local image cache.
      @param size The cache size or -1 if the size should be set automatically.
      @param min  The minimum allowable size.
      @param def  The default size to use when <tt>size=-1</tt> (ignored if <tt>def&lt;min</tt>).
   */
  protected final void setCacheSize (int size, int min, int def) {
    if (size == -1) {
      imgCache.setMaxSize(Math.max(def, min));
      imgCacheSize = -1;
    } else {
      if (size < min) {
        size = min;
      }
      imgCache.setMaxSize(size);
      imgCacheSize = imgCache.getMaxSize(); // cache may internally restrict size.
    }
  }

  /** Gets the size of the local image cache.
      @return The cache size or -1 if the size is set automatically.
      @see #getCurrentCacheSize()
   */
  public final int getCacheSize () {
    return imgCacheSize;
  }

  /** Gets the current size of the local image cache.
      @return The cache size currently in use (if the cache size is set automatically, this method
              will return the current size of the cache).
      @see #getCacheSize()
   */
  public final int getCurrentCacheSize () {
    return imgCache.getMaxSize();
  }

  public boolean isSymbolFill () {
    return symbolFill;
  }

  /** Convenience method */
  public boolean getSymbolFill () {
    return isSymbolFill();
  }

  public void setSymbolFill (boolean fill) {
    symbolFill = fill;
  }

  /** Get the KML Parser */
  public KmlParser getKmlParser () {
    return kmlParser;
  }

  /** Set the KML parser to use */
  public void setKmlParser (KmlParser parser) {
    kmlParser = parser;
  }

  /** Get the vector with the KML Objects being used */
  public kmlObjects getKmlObjects () {
    return objects;
  }

  /** Set the Vector of KML objects */
  public void setKmlObjects (kmlObjects objs) {
    objects = objs;
  }

  /** Break apart the objects in the kml file into their components
      Usually called by the draw method.

      Note: The number of objects passed to this method will usually be quite
       small, since they are rolled up to a parent object. The parent object is
       usually a <code>kmlDocument</code>, but this is not strictly enforced.
   */
  private void handleKmlObjects (kmlObjects objects) {

    boolean hasDocuments          = !objects.documents.isEmpty();
    boolean hasPlacemarks         = !objects.placemarks.isEmpty();
    boolean hasFolders            = !objects.folders.isEmpty();
    boolean hasGroundOverlays     = !objects.groundOverlays.isEmpty();
    boolean hasNetworkLinkControl = !objects.networkLinkControls.isEmpty();
    boolean hasNetworkLink        = !objects.networkLinks.isEmpty();
    //boolean hasRegions       = !objects.regions.isEmpty();

    int numObjs = 0; // temporary int holder
    int idx     = 0; // temporary int holder

    if (hasNetworkLink) {
      numObjs = objects.networkLinks.size();
      int adjIndex = createOrUpdateLinkedFiles(numObjs);
      for (idx=0; idx<numObjs; idx++) {
        handleKmlNetworkLink(objects.networkLinks.get(idx), idx+adjIndex);
      }
    }

    if (hasNetworkLinkControl) {
      handleKmlNetworkLinkControl(objects.networkLinkControls.get(0));  // only gets first one
    }

    if (hasDocuments) {
      numObjs = objects.documents.size();
      for (idx=0; idx<numObjs; idx++) {
        handleKmlDocument(objects.documents.get(idx));
      }
    }

    if (hasPlacemarks) {
      numObjs = objects.placemarks.size();
      for (idx=0; idx< numObjs; idx++) {
        handleKmlPlacemark(objects.placemarks.get(idx));
      }
    }

    if (hasFolders) {
      numObjs = objects.folders.size();
      for (idx=0; idx< numObjs; idx++) {
        handleKmlFolder(objects.folders.get(idx));
      }
    }

    if (hasGroundOverlays) {
      numObjs = objects.groundOverlays.size();
      for (idx=0; idx< numObjs; idx++) {
        handleKmlGroundOverlay(objects.groundOverlays.get(idx));
      }
    }

    /*
    if (hasRegions) {
      int numRegions = objects.regions.size();
      for (int xx=0; xx<numRegions; xx++) {
        handleKmlRegion((kmlRegion)objects.regions.get(xx));
      }
    }
     */
  }

  /** Common method to either create a new {@link #linkedFile} object and associated objects
      ({@link #pollRate}, {@link #isFirstPass}, and {@link #lastPoll}) or, if they
      already exist, extend them to include new elements. This occurs when the NetworkLinks
      are in different folders or levels of the kml hierarchy.
      @param numLinks Number of elements to create or add in the new arrays.
      @return The size of the current {@link #linkedFile} array, or 0 if {@link #linkedFile} has not been created yet.
      @since NeXtMidas 2.9.0 (Bug 2237)
   */
  private int createOrUpdateLinkedFiles (int numLinks) {
    int adjIndex = 0;
    if (linkedFile==null) {    // folder is the top-level container
      linkedFile    = new KmlFile[numLinks];
      pollRate      = new double[numLinks];
      lastPoll      = new double[numLinks];
      isFirstPass   = new boolean[numLinks];
    } else {                  // Network Links are in different folders / levels
      adjIndex = linkedFile.length;
      KmlFile[] tmpF = new KmlFile[adjIndex+numLinks];
      for (int xx=0; xx<adjIndex; xx++) { tmpF[xx] = linkedFile[xx]; }
      linkedFile = tmpF;
      double[] tmpD = new double[linkedFile.length];
      for (int xx=0; xx<adjIndex; xx++) { tmpD[xx] = pollRate[xx]; }
      pollRate = tmpD;
      double[] tmpD2 = new double[linkedFile.length];
      for (int xx=0; xx<adjIndex; xx++) { tmpD2[xx] = lastPoll[xx]; }
      lastPoll = tmpD2;
      boolean[] tmpB = new boolean[linkedFile.length];
      for (int xx=0; xx<adjIndex; xx++) { tmpB[xx] = isFirstPass[xx]; }
      isFirstPass = tmpB;
    }
    return adjIndex;
  }

  /** Check to see if a given String is in URL format. If not, make it so */
  private URL getURL (String href) {
    URL url = null;
    FileName fn = new FileName(href);
    String path = fn.getPath();
    if (path.equals("")) {
      fn.setPath(currPath);
    }
    try {
      return fn.toURL(IOResource.DIR_INPUT);
    } catch (MalformedURLException e) {
      e.printStackTrace();
    }
    return url;
  }

  /** Get the image and location for the overlay and draw it */
  private void handleKmlGroundOverlay (kmlGroundOverlay overlay) {
    kmlIcon icon = overlay.getIcon();       // get the kmlIcon used by the Overlay
    Color color = overlay.getColor();
    int h;                                 // img height
    int w;                                 // img width
    String href        = icon.getHref();
    double north       = overlay.getNorth();  // north lat
    double south       = overlay.getSouth();  // south lat
    double east        = overlay.getEast();   // east lon
    double west        = overlay.getWest();   // west lon

    // viewformat needs to get added to the end of the url
    if ((href.startsWith("http") || href.startsWith("HTTP"))  && icon.getViewFormat()!=null) {
      href = href + "&" + icon.parseViewFormat(north, south, east, west, MP.panel.getWidth(), MP.panel.getHeight());
    }

    BufferedImage img = addToImgCache(href); // get the image from the cache
    if (img == null) return;    // href did not work

    if (img != null) {
      int alpha = color.getAlpha();
      h = img.getHeight();
      w = img.getWidth();
      if (alpha != OPAQ && alpha != -1) {
        img = MColor.adjustImage(img, 0, 0, w, h, 50, 50, alpha, false);
      }
    }

    /* Adjust for rotation, if necessary */
    double rotation = overlay.getRotation();
    if (rotation != 0) {
      double rads     = Math.toRadians(rotation);
      img = GraphicsUtil.rotateImage(img, rads);   // Adjust the rotation of the img
    }

    MP.drawImage(img, north, south, east, west);
  }

  /** Break apart a KML Document Object. */
  private void handleKmlDocument (kmlDocument document) {
    String docName          = document.getName();
    String docDescription   = document.getDescription();
    kmlRegion region        = document.getRegion();
    int numFeatures         = document.features.size();
    String styleId          = document.getStyleUrl();
    Vector<kmlStyle> styles = document.getStyles();


    if (styles.size() > 0) {
      for (int xx=0; xx<styles.size(); xx++) {
        this.styles.add(styles.get(xx));
      }
    }

    if (numFeatures > 0) {
      for (int xx=0; xx < numFeatures; xx++) {
        handleKmlFeature(document.features.get(xx));
      }
    }

    if (styleId != null) {
      handleKmlStyle(styleId);
    } else {
      setStyleDefaults();
    }

  }

  /** Sets Style (color, width, etc) back to defaults so the next object being
       drawn does not inherit unique style from the previous object.
   */
  private void setStyleDefaults () {
    setPolygonFill    (true);
    setPolygonOutline (true);
    setLineWidth      (DEFAULT_LINE_WIDTH);
    setLineColor      (DEFAULT_POINT_COLOR);
    setPolygonColor   (DEFAULT_POINT_COLOR);
    setIconHeading    (DEFAULT_ICON_HEADING);
    setIconScale      (DEFAULT_ICON_SCALE);
    setPointSymbol    (DEFAULT_POINT_SYMBOL);
    setPointSize      (DEFAULT_POINT_SIZE);
  }

  /** Looks up the Style based on the given id and sends it to
       {@link #handleKmlStyle(kmlStyle)}.
      @param styleId
   */
  private void handleKmlStyle (String styleId) {
    for (int xx=0; xx<styles.size(); xx++) {
      if (styleId.equalsIgnoreCase((styles.get(xx)).getId())) {
        handleKmlStyle(styles.get(xx));
      }
    }
  }

  /** Break down the <code>kmlStyle</code> element into the respective styles
       and handle them. This usually involves setting certain attributes (such
       as color, width, etc) of the objects to be drawn prior to the drawing.
     @param style The <code>kmlStyle</code.
   */
  private void handleKmlStyle (kmlStyle style) {
    setStyleDefaults();

    if (style.hasLineStyle) {
      handleKmlLineStyle(style.getLineStyle());
    }
    if (style.hasPolyStyle) {
      handleKmlPolyStyle(style.getPolyStyle());
    }
    if (style.hasIconStyle) {
      handleKmlIconStyle(style.getIconStyle());
    }
    if (style.hasLabelStyle) {
      handleKmlLabelStyle(style.getLabelStyle());
    }

    // Load the balloon Style (null will use defaults)
    handleKmlBalloonStyle(style.getBalloonStyle());

  }

  /** Load the balloon style info for later use. null==defaults */
  private void handleKmlBalloonStyle (kmlBalloonStyle balloonStyle) {
    if (balloonStyle==null) this.balloonStyle = new kmlBalloonStyle();  // null == defaults
    else this.balloonStyle = balloonStyle;

     // if no text was set for the balloon style, use the description field
    String text = balloonStyle.getText();
    if ((text==null || text.length()< 1 || text.equals("")) && currDesc != null) text = currDesc;
    this.balloonStyle.setText(text);
  }

  /** Set Width and line color, if given */
  private void handleKmlLineStyle (kmlLineStyle lineStyle) {
    setLineWidth((int)lineStyle.getWidth());
    setLineColor(lineStyle.getColor());
  }

  private void handleKmlPolyStyle (kmlPolyStyle polyStyle) {
    setPolygonFill(polyStyle.isFill());
    setPolygonOutline(polyStyle.isOutline());
    setPolygonColor(polyStyle.getColor());
  }

  private void handleKmlIconStyle (kmlIconStyle iconStyle) {
    setIconHeading (iconStyle.getHeading());
    setIconScale   (iconStyle.getScale());     // Adjusts size displayed
    setIconColor   (iconStyle.getColor());
    handleKmlIcon  (iconStyle.getIcon());      // Symbol or image
  }

  private void handleKmlIcon (kmlIcon icon) {
    String url = icon.getHref();

    // TODO: Add logic to append parseViewFormat to the url...requires reference to parent geometry

    if (Symbol.symbolList.indexOf(url) >= 0) {
      setPointSymbol(url);
    } else if (url.toLowerCase().startsWith("http") || url.toLowerCase().startsWith("file")) {
      setPointImage(url);
    } else {
      setPointImage(url);
    }
  }

  private void handleKmlLabelStyle (kmlLabelStyle labelStyle) {
    setLabelColor (labelStyle.getColor());
    setLabelScale (labelStyle.getScale());
  }

  private void handleKmlFolder (kmlFolder folder) {
    int numSubFolders = folder.getFolders().size();
    int numFeatures   = folder.features.size();
    int numLinks      = folder.networkLinks.size();
    int idx            = 0;              // temp int holder

    if (numSubFolders > 0) {
      for (idx=0; idx<numSubFolders; idx++) {
        handleKmlFolder(folder.getFolders().get(idx));
      }
    }

    if (numLinks > 0) {
      int adjIndex = createOrUpdateLinkedFiles(numLinks);   // adjust index for multiple links but not at same level
      for (idx=0; idx<numLinks; idx++) {
        handleKmlNetworkLink(folder.networkLinks.get(idx), idx+adjIndex);
      }
    }

    if (folder.getStyleUrl() != null) {
      handleKmlStyle(folder.getStyleUrl());
    } else {
      setStyleDefaults();
    }

    if (numFeatures > 0) {
      for (idx=0; idx<numFeatures; idx++) {
        handleKmlFeature(folder.features.get(idx));
      }
    }
  }

  /** Break apart KML Placemark objects and draw them */
  private void handleKmlPlacemark (kmlPlacemark placemark) {
    String styleId = placemark.getStyleUrl();
    if (styleId != null) {
      handleKmlStyle(styleId);
    } else if (placemark.getStyles().size() > 0) {
      for (int xx=0; xx< placemark.getStyles().size(); xx++) {
        handleKmlStyle(placemark.getStyles().get(xx));
      }
    } else {
      setStyleDefaults();
    }

    handleKmlGeometry(placemark.getGeometry());
  }

  /** Take a Generic kmlGeometry object and send to the precise handler for it
      @param geometry The kmlGeometry object
   */
  private void handleKmlGeometry (kmlGeometry geometry) {

    if (geometry instanceof kmlLineString) {
      handleKmlLineString((kmlLineString)geometry);
    } else if (geometry instanceof kmlLinearRing) {
      handleKmlLinearRing((kmlLinearRing)geometry);
    } else if (geometry instanceof kmlPoint) {
      handleKmlPoint((kmlPoint)geometry);
    } else if (geometry instanceof kmlPolygon) {
      handleKmlPolygon((kmlPolygon)geometry);
    } else if (geometry instanceof kmlMultiGeometry) {
      handleKmlMultiGeometry((kmlMultiGeometry)geometry);
    }
  }

  /** Break apart a KML Line String object and draw it */
  private void handleKmlLineString (kmlLineString lineString) {

    /* If the LineString extrudes, it becomes a polygon */
    if (lineString.isExtrude()) {
      drawExtrudedPolygon(lineString.getCoordsAsDouble());
    } else {
      drawLineString(lineString);
    }
  }

  /** Used when Polling a file. Sets the layer poll rate and file to be polled. */
  private void handleKmlNetworkLink (kmlNetworkLink networkLink, int index) {
    String mode   =  networkLink.link.getRefreshMode().toLowerCase();
    if (mode.equals("oninterval")) {
      pollRate[index]      =  networkLink.link.getRefreshInterval();
      if (pollRate[index] < minRefreshPeriod) pollRate[index] = minRefreshPeriod;  // keep pollrate within bounds
    } else {
      pollRate[index]      = 3600;    // check every hour even if not specified
    }
    linkedFile[index]    = new KmlFile(networkLink.link.getHref());
    isFirstPass[index]   = true;
  }

  /** sets some limits on what the network Link is allowed to do */
  private void handleKmlNetworkLinkControl (kmlNetworkLinkControl control) {
    minRefreshPeriod = control.getMinRefreshPeriod();
    if (pollRate==null) { pollRate = new double[1]; pollRate[0]=3600; return; }  // polling done by base/1st KML layer
    for (int idx=0; idx<pollRate.length; idx++) {
      if (pollRate[idx] < minRefreshPeriod) pollRate[idx] = minRefreshPeriod;
    }
  }

  /** Creates and returns the kml Coordinates for extruded objects */
  private double[] getExtrudeCoords (double[] coords) {
    int length = ((coords.length * 2) + 3);
    double[] polyCoords = new double[length];
    int coordsCounter = 0;

    for (int xx=0; xx<coords.length; xx++, coordsCounter++) {
      polyCoords[coordsCounter] = coords[xx];
    }

    for (int xx=(coords.length -1); xx>=0; xx -= 3) {
      polyCoords[coordsCounter++] = coords[xx - 2];
      polyCoords[coordsCounter++] = coords[xx - 1];
      polyCoords[coordsCounter++] = 0;
    }

    /* Close the loop by returning to the starting point */
    polyCoords[coordsCounter++] = coords[0];
    polyCoords[coordsCounter++] = coords[1];
    polyCoords[coordsCounter++] = coords[2];

    return polyCoords;
  }

  private void drawExtrudedPolygon (double[] points) {
    int length = points.length;

    drawPolygon(points);

    for (int xx=0; xx<length-3; xx+=3) {
      double[] poly = {points[xx],   points[xx+1], points[xx+2],
          points[xx+3], points[xx+4], points[xx+5],
          points[xx+3], points[xx+4], 0,
          points[xx],   points[xx+1], 0,
          points[xx],   points[xx+1], points[xx+2]};

      drawPolygon(poly);
    }

  }

  /** Break apart a KML Linear Ring object and draw it */
  private void handleKmlLinearRing (kmlLinearRing linearRing) {
    if (linearRing.isExtrude()) {
      drawExtrudedPolygon(linearRing.getCoordsAsDouble());
    } else {
      drawLinearRing(linearRing);
    }
  }

  /** Break apart a KML Point object and draw it */
  private void handleKmlPoint (kmlPoint point) {
    drawPoint(point);
  }

  /** Break apart a KML Polygon object and draw it */
  private void handleKmlPolygon (kmlPolygon polygon) {
    drawPolygon(polygon);
  }

  /** Break apart a kmlMultiGeometry into separate kmlGeometrys and send to
       {@link #handleKmlGeometry(kmlGeometry)}.
      @param multiGeometry The kmlMultiGeometry object
   */
  private void handleKmlMultiGeometry (kmlMultiGeometry multiGeometry) {
    int geoCount = multiGeometry.geometrys.size();
    for (int xx=0; xx<geoCount; xx++) {
      handleKmlGeometry(multiGeometry.geometrys.get(xx));
    }
  }

  /** Break apart a KML Feature object and draw it */
  private void handleKmlFeature (kmlFeature feature) {
    String           name                  = feature.getName();
    boolean          visibility            = feature.getVisibility();
    boolean          open                  = feature.getOpen();
    String           address               = feature.getAddress();
    String           phoneNumber           = feature.getAddress();
    String           snippet               = feature.getSnippet();
    int              snippetMaxLines       = feature.getSnippetMaxLines();
    String           description           = feature.getDescription();
    kmlTimePrimitive timePrimitive         = feature.getTimePrimitive();
    kmlStyleSelector styleSelector         = feature.getStyleSelector();
    kmlRegion        region                = feature.getRegion();
    kmlExtendedData  extendedData          = feature.getExtendedData();

    String featureType = feature.getClass().getName();

    currDesc = description; // save the description for later use

    //System.out.println("Feature type: " +featureType);

    if (featureType.indexOf("kmlPlacemark") >= 0) {
      handleKmlPlacemark ((kmlPlacemark)feature);
    } else if (featureType.indexOf("kmlFolder") >= 0) {
      handleKmlFolder ((kmlFolder)feature);
    } else if (featureType.indexOf("kmlDocument") >= 0) {
      handleKmlDocument((kmlDocument)feature);
    } else if (featureType.indexOf("kmlGroundOverlay") >= 0) {
      handleKmlGroundOverlay((kmlGroundOverlay)feature);
    }
  }

  /** Break apart a KML Region object and draw it */
  private void handleKmlRegion (kmlRegion region) {
    // to be added later as required
  }

  /** Break apart a KML Overlay object and draw it. Not implemented yet...*/
  private void handleKmlOverlay (kmlOverlay overlay) {
    // to be added later as required
  }

  /** Check the given value and if less than the current X Min, set as X Min.
      If greater than the current X Max, then set as X Max.
   */
  protected void updateXRange (double xval) {
    if (xval < x1) x1 = xval;
    if (xval > x2) x2 = xval;
  }

  /** Check the given value and if less than the current Y Min, set as Y Min.
      If greater than the current Y Max, then set as Y Max.
   */
  protected void updateYRange (double yval) {
    if (yval < y1) y1 = yval;
    if (yval > y2) y2 = yval;
  }

  /** Since NeXtMidas 2.9.0, does nothing. Range values set when reading coordinates.
      @see #updateXRange(double)
      @see #updateYRange(double)
    */
  public void findRange() {

  }

  public void findFullRange() { findFullRange(pf); }

  /** Not implemented yet */
  private void adjustColors () {

  }

  /** Sets the contrast to the given percent then redraws.
      @param percent Percentage of contrast (0..100)
   */
  public void setContrast (int percent) {
    if (percent != contrast) {
      contrast = Math.max(0,Math.min(100,percent));
      //needRefresh = true;
      //refresh();
    }
  }

  /** Gets the contrast of the layer.
      @return Percentage of contrast (0..100).
   */
  public float getContrast() { return contrast; }

  /** Sets the brightness to the given percent then redraws.
      @param percent Percentage of brightness (0..100)
   */
  public void setBrightness (int percent) {
    if (percent != brightness) {
      brightness = Math.max(0,Math.min(100,percent));
      //needRefresh = true;
      //refresh();
    }
  }

  /** Gets the brightness of the layer.
      @return Percentage of brightness (0..100).
      @since NeXtMidas 2.7.0
   */
  public float getBrightness() { return brightness; }

  /** Sets the opaqueness to the given percent then sets the alpha.
      @param percent Percentage of opaqueness (0..100).
   */
  public void setOpaqueness (int percent) {
    setAlpha((percent * 255) / 100); // convert from % (0..100) to 0..255
  }

  /** Gets the opaqueness of the layer, which is the percentage of alpha.
      @return Percentage of opaqueness (0..100).
   */
  public int getOpaqueness() { return (int)(alpha / 2.55); }

  /** Sets the alpha either directly or called from setOpaqueness.
      @param newAlpha New alpha value (0..255).
   */
  public void setAlpha (int newAlpha) {
    if (newAlpha != alpha) {
      alpha = Math.max(0,Math.min(255,newAlpha));
      //needRefresh = true;
      //refresh();
    }
  }

  /** Gets the alpha of the layer.
      @return Alpha value (0..255).
      @since NeXtMidas 2.7.0
   */
  public int getAlpha() { return alpha; }

  //////////////////////////////////////////////////////////////////////////////
  /////////////////////// NESTED CLASSES ///////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////

  /** KML Parser
      This parser is a basic JAVA SAX XML parser.
      As it parses through the KML, it reads the type of tag, then sends the data
       within the tag to the appropriate handler.
   <pre>
   To add support for a new element or object, do the following:

    1. Add the tag for the element to {@link #tagList}.

    2. Determine if the element(tag) is an object that will have nested tags.
       If so, add the element to the "IF" list in
       {@link #startElement(String, String, String, Attributes)}. This pushes
       the object to the stack when the start tag is found. If the element just
       contains data (e.g. coordinates), then you do not need to do this step.

    3. Next, (for <b>ALL</b> tags) add the element to the "IF" list in
       {@link #endElement(String, String, String)}. Then add the necessary
       'addTo' method to add the object to its parent. This is where you pop
       the object or data off the stack and add it to its parent object. Make
       sure you have the appropriate "SET" methods in parent objects.
       Note: you must handle all possible parent cases, for casting purposes.
   </pre>
   */
  public class KmlParser extends DefaultHandler {

    int DEBUG = 0;       // for debugging purposes

    /** Flag to indicate we should capture the data in the characters method.
       This flag is set when the startElement method recognizes an element that
        contains data we are interested in (e.g. coordinates).
     */
    boolean processData;

    //String myUri        = new String();
    String kmlFile;

    LayerKML layer         = LayerKML.this;  // The layer

    /**  tag is the current element. */
    String tag          = new String();

    /** List of currently supported elements.  */
    String tagList = layer.supportedElements;

    /* The stack is used to keep track of nested elements */
    private Stack<Object> stack;
    private Locator locator;

    /* kmlObjs is an object view of the elements and data in the file */
    private kmlObjects       kmlObjs;
    boolean hasKmlStartTag = false;
    boolean hasKmlEndTag   = false;
    boolean objects        = false;
    String  parent;   // Current parent object. Used when keeping track of nested objects.

    /** Constructor   */
    public KmlParser () {
      super();
      stack           = new Stack<Object>();   // Reset stack
      processData     = false;         // Set to true when we want to capture element data
      parent          = new String();
    }

    /** Parse a kml file. Called by <code>LayerKML.setFile</code> method.
        This method invokes the SAX XML parser methods and calls the SAX parser
         on the file. When parsing the file, the SAX parser automatically calls
         a number of methods (some of which we override).
        @param filename The name of the file to parse
        @throws Exception
     */
    public void parseKmlFile (InputStream filename) throws Exception {

      /* SAX parser invocation */
      XMLReader xr = XMLReaderFactory.createXMLReader();      // create XML reader
      KmlParser myHandler = new KmlParser();                  // create Handler
      xr.setContentHandler(myHandler);                        // set Content handler
      xr.setErrorHandler(myHandler);                          // set error Handler

      try {
        Reader in = new InputStreamReader(filename);
        if (in != null) xr.parse(new InputSource(in));
      }
      catch (Exception e) {
        String popupMsg = "Bad KML: "+e.getMessage()+" Click OK to continue Plot (excluding this KML) or Cancel to abort";
        String boo = MJPopup.showOkCancel(MP.getPanel(), "Bad KML found.", JOptionPane.OK_CANCEL_OPTION, popupMsg);
        if (boo == null){         // null means they hit cancel button
          MP.status = MWindow.EXIT;
          MP.MW.refresh();
        }
        e.printStackTrace();
      }
    }

    /** Called automatically by the SAX parser when the <code>xml</code>
        tag is encountered.
     */
    public void startDocument () {
      if (DEBUG > 3) System.out.println("Start document");
    }

    /** Called automatically by the SAX parser when the <code></xml></code>
         tag is encountered or document ends. If it has reached the end of
         the file without encountering the closing <code>kml</code> tag,
         it is assumed that the file does not have one and the "wrapper"
         <code>kmlObjects</code> object is created. Since the call to this method
         signifies that the file is done parsing, we can then examine the
         objects created from the file and plot them.
     */
    public void endDocument () {
      if (!hasKmlEndTag ) {
        Object obj = null;
        if (!stack.isEmpty()) obj= stack.pop();
        if (obj instanceof kmlObjects) {
          Shell.warning("Missing KML tags in file " +kmlFile);
          addToKml((kmlObjects)obj);
        }
      }
      layer.setKmlObjects(getKmlObjects());     // Sets the kml Objects in LayerKML class
    }

    /** Get the kmlObjs object.
        @return kmlObjs The kmlObjects.
     */
    public kmlObjects getKmlObjects () { return kmlObjs; }

    /** Set the Locator   */
    public void setDocumentLocator( Locator rhs ) { locator = rhs; }

    /** Called automatically by the SAX parser when it encounters an opening
         tag/element. For our purposes, we then look at the element to
         determine if we support it. If so, we create a new kml object for
         'parent' objects such as <code>Placemark</code> and <code>Point</code>
         or a StringBuffer for actual data, such as <code>name</code>. All of
         the objects are put on the stack for future use. For data
         elements we set the {@link #processData} flag to true.
        @param uri         The xml uri
        @param localName   The short name of the element
        @param qName       The full name of the element
        @param attributes  Any attributes associated with the element
     */
    public void startElement (String uri, String localName, String qName, Attributes attributes) {

      if (DEBUG > 3) System.out.println("startElement: uri=" +uri+ ", localName=" +localName);

      /* For standardization purposes, convert all element names to lower case */
      String element = localName.toLowerCase();

      /* If the element is one of the tags we parse, capture the data
         Complex (nested) elements first) */
      if (tagList.indexOf(element) >= 0) {
        if (!hasKmlStartTag || (element.equalsIgnoreCase("kml"))) {
          stack.push(new kmlObjects());               // Used to hold the kml Objects
          hasKmlStartTag = true;
        }
        if (element.equalsIgnoreCase("kml")) {} // do nothing. Handled above
        else if (element.equalsIgnoreCase("networklinkcontrol")) {
          stack.push(new kmlNetworkLinkControl()); // kmlNetworkLinkControl
        } else if (element.equalsIgnoreCase("networklink")) {
          stack.push(new kmlNetworkLink());        // kmlNetworkLink
        } else if (element.equalsIgnoreCase("link")) {
          stack.push(new kmlLink());              // kmlLink
        } else if (element.equals("document")) {
          stack.push(new kmlDocument());           // new kmlDocument object
        } else if (element.equalsIgnoreCase("folder")) {
          stack.push(new kmlFolder());             // kmlFolder
        } else if (element.equals("placemark")) {
          stack.push(new kmlPlacemark());          // kmlPlacemark object
        } else if (element.equals("multigeometry")) {
          stack.push(new kmlMultiGeometry());      // new kmlMultiGeometry object
        } else if (element.equals("point")) {
          Object obj = stack.peek();
          if (!(obj instanceof kmlFeature)) {
            Object tmpObj = stack.pop();
            obj = stack.peek();        // Go up another level to get the Feature
            stack.push(tmpObj);
          }
          stack.push(new kmlPoint(obj));
        } else if (element.equals("polygon")) {
          stack.push(new kmlPolygon());            // new kmlPolygon object
        } else if (element.equals("linearring")) {
          stack.push(new kmlLinearRing());         // new kmlLinearRing object
        } else if (element.equals("linestring")) {
          stack.push(new kmlLineString());         // new kmlLineString object
        } else if (element.equals("outerboundaryis")) {
          ((kmlPolygon)stack.peek()).setOuter(true);  // set outer ring flag to true
        } else if (element.equalsIgnoreCase("groundoverlay")) {
          stack.push(new kmlGroundOverlay());        // kmlGroundOverlay
        } else if (element.equals("lookat")) {
          stack.push(new kmlLookAt());             // kmlLookAt
        } else if (element.equals("polystyle")) {
          stack.push(new kmlPolyStyle());          // new kmlPolyStyle object
        } else if (element.equals("style")) {
          stack.push(new kmlStyle());              // new kmlStyle object
          for (int xx=0; xx< attributes.getLength(); xx++) {
            if (attributes.getLocalName(xx).equalsIgnoreCase("id")) {
              ((kmlStyle)stack.peek()).setId(attributes.getValue(xx));  // Set the Style Id
            }
          }
        } else if (element.equals("labelstyle")) {
          stack.push(new kmlLabelStyle());          // new kmlLineStyle object
        } else if (element.equals("linestyle")) {
          stack.push(new kmlLineStyle());          // new kmlLineStyle object
        } else if (element.equals("iconstyle")) {
          stack.push(new kmlIconStyle());          // new kmlIconStyle object
        } else if (element.equals("balloonstyle")) {
          stack.push(new kmlBalloonStyle());      // new kmlBalloonStyle
        } else if (element.equals("icon")) {
          stack.push(new kmlIcon());               // new kmlIcon object
        } else if (element.equals("latlonbox")) {
          // recognize tag but do nothing
        }
        else {
          stack.push(new StringBuffer());          // Used to hold element data
          processData = true;                      // Indicates that next set of data is relevant
        }
      }
    }

    /** Called automatically by the SAX parser when it encounters the closing
         tag for an element, such as <code>/coordinates</code>. For our
         purposes, we can then add the nested objects to their parents, such as
         <code>coordinates</code> to their <code>point</code>.
     */
    public void endElement (String uri, String localName, String rawName) {
      processData     = false;  // Set to false since data element is closing

      /* For standardization purposes, convert all element names to lower case */
      String element = localName.toLowerCase();
      if (tagList.indexOf(element) >= 0 && !element.equals("outerboundaryis")) {

        Object obj = stack.pop();    // Get the newest object off the stack
        obj = StringUtil.trimIfString(obj); // fix for Bug 1825 (to trim String values within tags)

        if (!element.equals("kml"))  {
          parent = stack.peek().getClass().getName();
          int i  = parent.indexOf("$");
          parent = parent.substring(i+4).toLowerCase();
        }

        //System.out.println("Popped a "+localName+" nested in "+ parent+ " from the stack");

        /* The IF statements below evaluate the tag and put the resulting child
              (nested) data on the parent object. Parent objects themselves are
              added to the appropriate Vector of objects in the kmlObjects class.
         */
        if (element.equals("kml"))                  { hasKmlEndTag=true; addToKml((kmlObjects)obj); }
        else if (element.equals("networklink"))       addToContainer(element, obj);
        else if (element.equals("style"))             addToFeature(element, obj);
        else if (parent.equals("objects"))            addToKmlObjects(obj);
        else if (parent.equals("networklinkcontrol")) addToNetworkLinkControl(element, obj);
        else if (parent.equals("networklink"))        addToNetworkLink(element, obj);
        else if (parent.equals("link"))               addToLink(element, obj);
        else if (parent.equals("document"))           addToDocument(element, obj);
        else if (parent.equals("folder"))             addToFolder(element, obj);
        else if (parent.equals("placemark"))          addToPlacemark(element, obj);
        else if (parent.equals("multigeometry"))      addToMultiGeometry(element, obj);
        else if (parent.equals("point"))              addToPoint(element, obj);
        else if (parent.equals("polygon"))            addToPolygon(element, obj);
        else if (parent.equals("linearring"))         addToLinearRing(element, obj);
        else if (parent.equals("linestring"))         addToLineString(element, obj);
        else if (parent.equals("groundoverlay"))      addToGroundOverlay(element, obj);
        else if (parent.equals("lookat"))             addToLookAt(element, obj);
        else if (parent.equals("polystyle"))          addToPolyStyle(element, obj);
        else if (parent.equals("style"))              addToStyle(element, obj);
        else if (parent.equals("labelstyle"))         addToLabelStyle(element, obj);
        else if (parent.equals("linestyle"))          addToLineStyle(element, obj);
        else if (parent.equals("iconstyle"))          addToIconStyle(element, obj);
        else if (parent.equals("balloonstyle"))       addToBalloonStyle(element, obj);
        else if (parent.equals("icon"))               addToIcon(element, obj);

        else { } // Should not enter this but if so...
      }
    }

    private void addToKml (kmlObjects obj) {
      kmlObjs = obj;
      objects = true;
    }

    private void addToKmlObjects (Object obj) {
      ((kmlObjects)stack.peek()).addKmlObject(obj);
    }

    private void addToNetworkLinkControl (String element, Object obj) {
      if (element.equals("minrefreshperiod")) {
        ((kmlNetworkLinkControl)stack.peek()).setMinRefreshPeriod(Float.parseFloat(obj.toString()));
      } else if (element.equals("lookat")) {
        ((kmlNetworkLinkControl)stack.peek()).setAbstractView((kmlLookAt)obj);
      }
    }

    private void addToNetworkLink (String element, Object obj) {
      if (element.equals("link")) {
        ((kmlNetworkLink)stack.peek()).setLink((kmlLink)obj);
      } else if (element.equals("styleurl")) {
        ((kmlNetworkLink)stack.peek()).setStyleUrl(obj.toString());
      } else if (element.equals("name")) {
        ((kmlNetworkLink)stack.peek()).setName(obj.toString());
      }
    }

    private void addToLink (String element, Object obj) {
      if (element.equals("refreshinterval")) {
        ((kmlLink)stack.peek()).setRefreshInterval(Float.parseFloat(obj.toString()));
      } else if (element.equals("refreshmode")) {
        ((kmlLink)stack.peek()).setRefreshMode((obj.toString()));
      } else if (element.equals("href")) {
        ((kmlLink)stack.peek()).setHref(obj.toString());
      }
    }

    private void addToDocument (String element, Object obj) {
      if (element.equals("folder")) {
        ((kmlDocument)stack.peek()).addFeature((kmlFolder)obj);
      } else if (element.equals("placemark")) {
        ((kmlDocument)stack.peek()).addFeature((kmlPlacemark)obj);
      } else if (element.equals("groundoverlay")) {
        ((kmlDocument)stack.peek()).addFeature((kmlGroundOverlay)obj);
      } else if (element.equals("visibility")) {
        boolean booTmp = (Integer.parseInt(obj.toString()) == 0)? false: true;
        ((kmlDocument)stack.peek()).setVisibility(booTmp);
      } else if (element.equals("styleurl")) {
        String styleId = obj.toString();
        if (styleId.startsWith("#")) styleId = styleId.substring(1);
        ((kmlDocument)stack.peek()).setStyleUrl(styleId);
      } else if (element.equals("name")) {
        ((kmlDocument)stack.peek()).setName(obj.toString());
      } else if (element.equals("description")) {
        ((kmlDocument)stack.peek()).setDescription(obj.toString());
      }
    }

    private void addToFolder (String element, Object obj) {
      if (element.equals("folder")) {
        ((kmlFolder)stack.peek()).addFolder((kmlFolder)obj);
      } else if (element.equals("document")) {
        ((kmlFolder)stack.peek()).addFeature((kmlDocument)obj);
      } else if (element.equals("placemark")) {
        ((kmlFolder)stack.peek()).addFeature((kmlPlacemark)obj);
      } else if (element.equals("groundoverlay")) {
        ((kmlFolder)stack.peek()).addFeature((kmlGroundOverlay)obj);
      } else if (element.equals("visibility")) {
        boolean booTmp = (Integer.parseInt(obj.toString()) == 0)? false: true;
        ((kmlFolder)stack.peek()).setVisibility(booTmp);
      } else if (element.equals("styleurl")) {
        String styleId = obj.toString();
        if (styleId.startsWith("#")) styleId = styleId.substring(1);
        ((kmlFolder)stack.peek()).setStyleUrl(styleId);
      } else if (element.equals("name")) {
        ((kmlFolder)stack.peek()).setName(obj.toString());
      }
    }

    private void addToPlacemark (String element, Object obj) {
      if (element.equals("multigeometry")) {
        ((kmlPlacemark)stack.peek()).setGeometry((kmlMultiGeometry)obj);
      } else if (element.equals("polygon")) {
        ((kmlPlacemark)stack.peek()).setGeometry((kmlPolygon)obj);
      } else if (element.equals("point")) {
        ((kmlPlacemark)stack.peek()).setGeometry((kmlPoint)obj);
      } else if (element.equals("linearring")) {
        ((kmlPlacemark)stack.peek()).setGeometry((kmlLinearRing)obj);
      } else if (element.equals("linestring")) {
        ((kmlPlacemark)stack.peek()).setGeometry((kmlLineString)obj);
      } else if (element.equals("styleurl")) {
        String styleId = obj.toString();
        if (styleId.startsWith("#")) styleId = styleId.substring(1);
        ((kmlPlacemark)stack.peek()).setStyleUrl(styleId);
      } else if (element.equals("name")) {
        ((kmlPlacemark)stack.peek()).setName(obj.toString());
      } else if (element.equals("description")) {
        ((kmlPlacemark)stack.peek()).setDescription(obj.toString());
      }
    }

    private void addToMultiGeometry (String element, Object obj) {
      if (element.equals("polygon")) {
        ((kmlMultiGeometry)stack.peek()).addGeometry((kmlPolygon)obj);
      } else if (element.equals("point")) {
        ((kmlMultiGeometry)stack.peek()).addGeometry((kmlPoint)obj);
      } else if (element.equals("linearring")) {
        ((kmlMultiGeometry)stack.peek()).addGeometry((kmlLinearRing)obj);
      } else if (element.equals("linestring")) {
        ((kmlMultiGeometry)stack.peek()).addGeometry((kmlGeometry)obj);
      }
    }

    private void addToPoint (String element, Object obj) {
      if (element.equals("extrude")) {
        boolean boo = (Integer.parseInt(obj.toString()) == 1)?  true: false;         // anything but "1" is FALSE
        ((kmlPoint)stack.peek()).setExtrude(boo);
      }  else if (element.equals("coordinates")) {
        ((kmlPoint)stack.peek()).setCoordinates(obj.toString());
      }
    }

    private void addToPolygon (String element, Object obj) {
      if (element.equals("linearring")) {
        if (((kmlPolygon)stack.peek()).isOuter()) {
          ((kmlPolygon)stack.peek()).setOuterBoundary((kmlLinearRing)obj);
          ((kmlPolygon)stack.peek()).setOuter(false);
        } else {
          ((kmlPolygon)stack.peek()).addInnerBoundary((kmlLinearRing)obj);
        }
      } else if (element.equals("extrude")) {
        boolean boo = (Integer.parseInt(obj.toString()) == 1)?  true: false;         // anything but "1" is FALSE
        ((kmlPolygon)stack.peek()).setExtrude(boo);
      } else if (element.equals("altitudemode")) {
        ((kmlPolygon)stack.peek()).setAltitudeMode(obj.toString());
      } else if (element.equals("tessellate")) {
        boolean boo = (Integer.parseInt(obj.toString()) == 1)? true: false;         // anything but "1" is FALSE
        ((kmlPolygon)stack.peek()).setTessellate(boo);
      } else if (element.equals("coordinates")) {
        ((kmlPolygon)stack.peek()).addCoordinates(obj.toString());
      }
    }

    private void addToLinearRing (String element, Object obj) {
      if (element.equals("extrude")) {
        boolean boo = (Integer.parseInt(obj.toString()) == 1)?  true: false;         // anything but "1" is FALSE
        ((kmlLinearRing)stack.peek()).setExtrude(boo);
      } else if (element.equals("coordinates")) {
        ((kmlLinearRing)stack.peek()).setCoordsVector(obj.toString());
      }
    }

    private void addToLineString (String element, Object obj) {
      if (element.equals("extrude")) {
        boolean boo = (Integer.parseInt(obj.toString()) == 1)? true: false;         // anything but "1" is FALSE
        ((kmlLineString)stack.peek()).setExtrude(boo);
      } else if (element.equals("coordinates")) {
        ((kmlLineString)stack.peek()).setCoordsVector(obj.toString());
      }
    }

    private void addToGroundOverlay (String element, Object obj) {
      if (element.equals("north")) {
        ((kmlGroundOverlay)stack.peek()).setNorth(Double.parseDouble(obj.toString()));
      } else if (element.equals("south")) {
        ((kmlGroundOverlay)stack.peek()).setSouth(Double.parseDouble(obj.toString()));
      } else if (element.equals("east"))  {
        ((kmlGroundOverlay)stack.peek()).setEast(Double.parseDouble(obj.toString()));
      } else if (element.equals("west")) {
        ((kmlGroundOverlay)stack.peek()).setWest(Double.parseDouble(obj.toString()));
      } else if (element.equals("rotation")) {
        ((kmlGroundOverlay)stack.peek()).setRotation(Double.parseDouble(obj.toString()));
      } else if (element.equals("icon")) {
        ((kmlOverlay)stack.peek()).setIcon((kmlIcon)obj);
      }  else if (element.equals("color")) {
        Color newColor = toColor(obj);
        ((kmlGroundOverlay)stack.peek()).setColor(newColor);
      }
    }

    private void addToLookAt (String element, Object obj) {
      if (element.equals("longitude")) {
        ((kmlLookAt)stack.peek()).setLon(Double.parseDouble(obj.toString()));
      } else if (element.equals("latitude")) {
        ((kmlLookAt)stack.peek()).setLat(Double.parseDouble(obj.toString()));
      } else if (element.equals("altitude")) {
        ((kmlLookAt)stack.peek()).setAlt(Double.parseDouble(obj.toString()));
      } else if (element.equals("heading")) {
        ((kmlLookAt)stack.peek()).setHeading(Double.parseDouble(obj.toString()));
      } else if (element.equals("range")) {
        ((kmlLookAt)stack.peek()).setRange(Double.parseDouble(obj.toString()));
      } else if (element.equals("tilt")) {
        ((kmlLookAt)stack.peek()).setTilt(Double.parseDouble(obj.toString()));
      }
    }

    /** Converts input KML-formatted color string (ABGR) to Color (ARGB).
        @param obj Should be String, StringBuffer, StingBuilder object (cannot be null)
        @return Color (ARGB)
        @todo JWB: Move to MColor like getColor(String, ABGR)
     */
    private Color toColor(Object obj) {
      /* Check formatting of color to ensure it is recognized */
      String color = obj.toString().toUpperCase();
      if (!color.startsWith("#") && !color.startsWith("0X")) {
        color = "0x" + color;   // Make Color recognizable as Hex
      }
      Color newColor = convertColor(MColor.getColor(color));
      return newColor;
    }

    private void addToPolyStyle (String element, Object obj) {
      if (element.equals("fill")) {
        boolean booTmp = (Integer.parseInt(obj.toString()) == 0)? false: true;
        ((kmlPolyStyle)stack.peek()).setFill(booTmp);
      } else if (element.equals("outline")) {
        boolean booTmp = (Integer.parseInt(obj.toString()) == 0)? false: true;
        ((kmlPolyStyle)stack.peek()).setOutline(booTmp);
      } else if (element.equals("color")) {
        Color newColor = toColor(obj);
        ((kmlPolyStyle)stack.peek()).setColor(newColor);
      }
    }

    private void addToStyle (String element, Object obj) {
      if (element.equals("linestyle")) {
        ((kmlStyle)stack.peek()).setLineStyle((kmlLineStyle)obj);
      } else if (element.equals("labelstyle")) {
        ((kmlStyle)stack.peek()).setLabelStyle((kmlLabelStyle)obj);
      } else if (element.equals("polystyle")) {
        ((kmlStyle)stack.peek()).setPolyStyle((kmlPolyStyle)obj);
      } else if (element.equals("iconstyle")) {
        ((kmlStyle)stack.peek()).setIconStyle((kmlIconStyle)obj);
      } else if (element.equals("balloonstyle")) {
        ((kmlStyle)stack.peek()).setBalloonStyle((kmlBalloonStyle)obj);
      }
    }

    private void addToLabelStyle (String element, Object obj) {
      if (element.equals("color")) {
        Color newColor = toColor(obj);
        ((kmlLabelStyle)stack.peek()).setColor(newColor);
      } else if (element.equals("scale")) {
        ((kmlLabelStyle)stack.peek()).setScale(Float.parseFloat(obj.toString()));
      }
    }

    private void addToLineStyle (String element, Object obj) {
      if (element.equals("color")) {
        Color newColor = toColor(obj);
        ((kmlLineStyle)stack.peek()).setColor(newColor);
      } else if (element.equals("width")) {
        ((kmlLineStyle)stack.peek()).setWidth(Float.parseFloat(obj.toString()));
      }
    }

    private void addToIconStyle (String element, Object obj) {
      if (element.equals("icon")) {
        ((kmlIconStyle)stack.peek()).setIcon((kmlIcon)obj);
      } else if (element.equals("color")) {
        Color newColor = toColor(obj);
        ((kmlIconStyle)stack.peek()).setColor(newColor);
      } else if (element.equals("heading")) {
        ((kmlIconStyle)stack.peek()).setHeading(Float.parseFloat(obj.toString()));
      } else if (element.equals("scale")) {
        ((kmlIconStyle)stack.peek()).setScale(Float.parseFloat(obj.toString()));
      }
    }

    private void addToBalloonStyle (String element, Object obj) {
      if (element.equals("bgcolor")) {
        Color color = toColor(obj);
        ((kmlBalloonStyle)stack.peek()).setBgColor(color);
      } else if (element.equals("textcolor")) {
        Color color = toColor(obj);
        ((kmlBalloonStyle)stack.peek()).setTextColor(color);
      } else if (element.equals("text")) {
        ((kmlBalloonStyle)stack.peek()).setText(obj.toString());
      } else if (element.equals("displaymode")) {
        ((kmlBalloonStyle)stack.peek()).setDisplayMode(obj.toString());
      }
    }

    private void addToIcon (String element, Object obj) {
      if (element.equals("href")) {
        ((kmlIcon)stack.peek()).setHref(obj.toString());
      } else if (element.equals("viewformat")) {
        ((kmlIcon)stack.peek()).setViewFormat(obj.toString());
      }
    }

    private void addToContainer (String element, Object obj) {
      if (element.equals("networklink")) {
        if (stack.peek() instanceof kmlContainer) {
          ((kmlContainer)stack.peek()).addNetworkLink((kmlNetworkLink)obj);
        } else if (stack.peek() instanceof kmlObjects) {
          ((kmlObjects)stack.peek()).addNetworkLink((kmlNetworkLink)obj);
        }
      }
    }

    private void addToFeature (String element, Object obj) {
      if (element.equals("style")) {
        ((kmlFeature)stack.peek()).addStyle((kmlStyle)obj);
      } else if (element.equals("lookat")) {
        ((kmlFeature)stack.peek()).setAbstractView((kmlLookAt)obj);
      }
    }

    /** Convert a KML-formatted color (aabbggrr) to standard (aarrggbb)
        @param kmlColor The color in the KML file to be converted
        @return The KML color in standard aarrggbb format
        @todo move this to MColor (should be public static)
     */
    public int convertColor(int kmlColor) {
      int aa = (kmlColor >> 24) & 0xFF;    // alpha
      int bb = (kmlColor >> 16) & 0xFF;    // blue
      int gg = (kmlColor >>  8) & 0xFF;    // green
      int rr = (kmlColor >>  0) & 0xFF;    // red

      return (aa << 24) | (rr << 16) | (gg << 8) | (bb << 0);
    }

    /** Convert a KML-formatted color (aabbggrr) to standard (aarrggbb)
        @param kmlColor The color in the KML file to be converted
        @return The KML color in standard aarrggbb format
        @todo move this to MColor (should be public static)
     */
    public Color convertColor(Color kmlColor) {
      int bgr = kmlColor.getRGB();
      //      int aa  = (bgr >> 24) & 0xFF;
      //      int bb  = (bgr >> 16) & 0xFF;
      //      int gg  = (bgr >>  8) & 0xFF;
      //      int rr  = (bgr >>  0) & 0xFF;
      //
      //      int rgb = (aa << 24) | (rr << 16) | (gg << 8) | (bb << 0);
      //      boolean hasAlpha = ((rgb & 0xFF000000) != 0);
      //
      //      return new Color(rgb, hasAlpha);

      int argb = convertColor(bgr);
      boolean hasAlpha = ((bgr & 0xFF000000) != 0);

      return new Color(argb, hasAlpha);
    }

    /** Called automatically by the SAX parser when it encounters data outside
         of an element tag, which it assumes to be data relevant to the most
         recent open element. In order to avoid processing data that we do not
         want or support, we use the flag {@link #processData} to indicate
         whether the data should be captured. If so, we load that data into the
         current item on the stack, which should always be a <code>StringBuffer
         </code>.
     */
    public void characters (char[] ch, int start, int length) {

      if (processData) {
        if (DEBUG > 1) System.out.println("Adding char data to StringBuffer");
        ((StringBuffer)stack.peek()).append( ch, start, length );
      } else {}    // ignore unwanted data
    }

    /** Gets the number of coordinate sets (points) in a given String.
        NOTE: The input String MUST follow the kml convention of whitespace
              characters ONLY between coordinate sets. Commas (with no
              whitespace) separate x,y,z values.

     */
    private int getNumPoints (String coords) {
      String[] tmp =  coords.split("\r\n|\r|\n|\\s");

      return tmp.length;
    }

    /** Future use    */
    private void processElementData (String tag, String data) {
      if (DEBUG > 1) System.out.println("Processing " +tag+ "=" +data);
    }

  }
  ////////////////////////////////////////////////////////////////////
  /** This class is used to store the objects created during the parsing of the
       kml file. It contains Vectors for each of the objects and methods to add
       the parsed objects to their associated Vector.
       <pre>
      Note: Most of the Vectors will be empty, since most kml objects are
             nested. Only the Vector for the top-most parent objects
             (such as <code>Placemarks</code> and <code>Documents</code>) will
             have data. But since this is not enforced in kml, all the objects
             are available.
     </pre>
   */
  public class kmlObjects {

    /* Supported kml Objects */
    private Vector<kmlPlacemark>          placemarks;
    private Vector<kmlFolder>             folders;
    private Vector<kmlDocument>           documents;
    private Vector<kmlNetworkLinkControl> networkLinkControls;
    private Vector<kmlGroundOverlay>      groundOverlays;
    private Vector<kmlNetworkLink>        networkLinks;

    /** Constructor instantiates the Vectors   */
    public kmlObjects () {
      placemarks            = new Vector<kmlPlacemark>();
      folders               = new Vector<kmlFolder>();
      documents             = new Vector<kmlDocument>();
      networkLinkControls   = new Vector<kmlNetworkLinkControl>();
      groundOverlays        = new Vector<kmlGroundOverlay>();
      networkLinks          = new Vector<kmlNetworkLink>();
    }

    /** generic method to add a Kml object */
    public void addKmlObject (Object obj) {
      if      (obj instanceof kmlPlacemark)          addPlacemark((kmlPlacemark)obj);
      else if (obj instanceof kmlDocument)           addDocument((kmlDocument)obj);
      else if (obj instanceof kmlFolder)             addFolder((kmlFolder)obj);
      else if (obj instanceof kmlNetworkLinkControl) addNetworkLinkControl((kmlNetworkLinkControl)obj);
      else if (obj instanceof kmlGroundOverlay)      addGroundOverlay((kmlGroundOverlay)obj);
      else if (obj instanceof kmlNetworkLink)        addNetworkLink((kmlNetworkLink)obj);
      else  Shell.warning("Unknown object " +obj.getClass().getSimpleName()+". Not adding to List");
    }

    public void addPlacemark (kmlPlacemark p) { placemarks.add(p);  }

    public void addFolder (kmlFolder folder) { folders.add(folder);  }

    public void addDocument (kmlDocument d) {  documents.add(d);   }

    public void addGroundOverlay (kmlGroundOverlay overlay) { groundOverlays.add(overlay); }

    public void addNetworkLinkControl (kmlNetworkLinkControl control) {
      networkLinkControls.add(control);
    }

    public void addNetworkLink (kmlNetworkLink link) {
      networkLinks.add(link);
    }

    // public void addRegion (kmlRegion r) { regions.add(r); }

  }

  ///////////////////////////////////////////////////////////////////

  /** KML Abstract Object */
  public class kmlObject {
    String id;          // optional
    String targetId;    // optional

    public String getId () { return id; }
    public void setId (String i) { id = i; }

    public String getTargetId () { return targetId; }
    public void setTargetId (String t) { targetId = t; }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Feature */
  public class kmlFeature extends kmlObject {
    String           name;
    boolean          visibility            = true;
    boolean          open                  = false;
    String           address;
    String           phoneNumber;
    String           snippet;
    int              snippetMaxLines       = 2;
    String           description;
    kmlTimePrimitive timePrimitive;
    String           styleUrl;
    kmlStyleSelector styleSelector;
    kmlRegion        region;
    kmlExtendedData  extendedData;
    Vector<kmlStyle> styles;
    kmlAbstractView  abstractView;

    public kmlFeature () {
      name          = new String();
      address       = new String();
      phoneNumber   = new String();
      snippet       = new String();
      description   = new String();
      timePrimitive = new kmlTimePrimitive();
      styleUrl      = null;
      styleSelector = new kmlStyleSelector();
      region        = new kmlRegion();
      extendedData  = new kmlExtendedData();
      styles        = new Vector<kmlStyle>();
      abstractView  = new kmlAbstractView();
    }

    public void addStyle (kmlStyle style) {
      styles.add(style);
    }

    public Vector<kmlStyle> getStyles () {
      return styles;
    }

    public String getStyleUrl() { return styleUrl; }
    public void setStyleUrl(String styleUrl) { this.styleUrl = styleUrl;}

    public String getName () { return name; }
    public void setName (String n) { name = n; }

    public boolean getVisibility () { return visibility; }
    public void setVisibility (boolean v) { visibility = v; }

    public boolean getOpen () { return open; }
    public void setOpen (boolean o) { open = o; }

    public String getAddress () { return address; }
    public void setAddress (String a) { address = a; }

    public String getPhoneNumber () { return phoneNumber; }
    public void setPhoneNumber (String p) { phoneNumber = p; }

    public int getSnippetMaxLines () { return snippetMaxLines; }
    public void setSnippetMaxLines (int lines) { snippetMaxLines = lines; }

    public String getSnippet () { return snippet; }
    public void setSnippet ( String s) { snippet = s; }

    public String getDescription () { return description; }
    public void setDescription (String s) { description = s; }

    public kmlTimePrimitive getTimePrimitive () { return timePrimitive; }
    public void setTimePrimitive (kmlTimePrimitive tp) { timePrimitive = tp; }

    public kmlStyleSelector getStyleSelector () { return styleSelector; }
    public void setStyleSelector (kmlStyleSelector ss) { styleSelector = ss; }

    public kmlExtendedData getExtendedData () { return extendedData; }
    public void setExtendedData (kmlExtendedData ed) { extendedData = ed; }

    public kmlRegion getRegion () { return region; }
    public void setRegion (kmlRegion r) { region = r; }

    public kmlAbstractView getAbstractView() { return abstractView;  }
    public void setAbstractView(kmlAbstractView abstractView) { this.abstractView = abstractView; }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Geometry  - Abstract -  */
  public class kmlGeometry extends kmlObject {

    private kmlFeature parentFeature = new kmlFeature();

    public void setParentFeature (kmlFeature feat) {
      parentFeature = feat;
    }

    public kmlFeature getParentFeature () {
      return parentFeature;
    }
  }
  ///////////////////////////////////////////////////////////////////
  /** KML Abstract View
      Defines which view to use (Camera or LookAt)
   */
  public class kmlAbstractView extends kmlObject {

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Balloon Style
      Specifies how the description balloon for a kml Feature is drawn.
   */
  public class kmlBalloonStyle extends kmlObject {
    private Color   bgColor   = Color.white;
    private Color   textColor = Color.black;
    private String  text;

    /** Must be 'default' or 'hide' */
    private String displayMode  = "default";

    public kmlBalloonStyle () {
      text = new String();
    }

    public Color getBgColor() { return bgColor;  }
    public void setBgColor(Color bgColor) { this.bgColor = bgColor; }

    public Color getTextColor() { return textColor; }
    public void setTextColor(Color textColor) { this.textColor = textColor;  }

    public String getText() { return text;  }
    public void setText(String text) { this.text = text; }

    public String getDisplayMode() { return displayMode; }

    public void setDisplayMode(String displayMode) {
      if (displayMode.equalsIgnoreCase("default")) {
        this.displayMode = "default";
      } else if (displayMode.equalsIgnoreCase("hide")) {
        this.displayMode = "hide";
      } else {
        Shell.warning("Invalid display mode: " +displayMode+ ". Mode must be 'hide' or 'default'");
      }
    }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Camera */
  public class kmlCamera extends kmlAbstractView {
    double  lon      = 0;      // 0 to (+/-) 180 Degrees
    double  lat      = 0;      // 0 to (+/-) 90  Degrees
    double  altitude = 0;      //
    double  heading  = 0;      // 0 to 360 degrees
    double  tilt     = 0;      // 0 to 180 degrees
    double  roll     = 0;      // 0 to (+/-) 180 degrees

    /** Must be either 'clampToGround' (default), 'relativeToGround', or 'absolute'.  */
    String altitudeMode    = "clampToGround";

    public kmlCamera () {}

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Color Style */
  public class kmlColorStyle extends kmlObject {
    Color   color = Color.white;

    /** Must be 'normal' (default) or 'random' */
    String colorMode            = "normal";

    public Color getColor() { return color; }
    public void setColor(Color color) { this.color = color; }

    public String getColorMode() { return colorMode; }

    public void setColorMode(String colorMode) {
      if (colorMode.equalsIgnoreCase("normal")) {
        this.colorMode = "normal";
      } else if (colorMode.equalsIgnoreCase("random")) {
        this.colorMode = "random";
      } else {
        Shell.warning("Invalid color mode: " +colorMode+ ". Mode must be 'normal' or 'random'.");
      }
    }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Coordinates
      Creating this class enables us to get around the KML ordering of
       comma-separated lat,lon,alt and white-space-separated coordinate sets
   */
  public class kmlCoordinates extends kmlObject {
    double lat;              // 0 to (+/-) 90 degrees
    double lon;              // 0 to (+/-) 180 degrees
    double alt       = 0;    // meters

    public kmlCoordinates () {}

    /**  Convert the String coords to the doubles */
    public void setCoordinates (String coords) {

      /* split the String on the commas */
      if (coords.indexOf(",") > 0) {
        String[] tmp = coords.split(",");
        setCoordinates(tmp);
      } else {
        /* If not valid coordinate syntax, alert the user */
        Shell.warning("Invalid coordinate set= " + coords);
      }
    }

    /**  Convert the String array coords to the doubles */
    public void setCoordinates (String[] coords) {
      boolean valid = false;

      if (coords.length < 4 && coords.length > 1) {
        valid = true;
        lon = Double.parseDouble(coords[0]);
        lat = Double.parseDouble(coords[1]);

        if (coords.length == 3) {
          alt = Double.parseDouble(coords[2]);
        }
      }

      /* If not valid coordinate syntax, alert the user */
      if (!valid) Shell.warning("Invalid coordinate set: " + coords);
      else {
        // update file X and Y values ranges
        updateXRange(lon);
        updateYRange(lat);
      }
    }

    public double getLat() { return lat; }

    public void setLat (double lat) { this.lat = lat; }

    public double getLon() { return lon;  }

    public void setLon (double lon) { this.lon = lon; }

    public double getAltitude () { return alt;  }

    public void setAltitude (double altitude) { alt = altitude;  }


  }

  ///////////////////////////////////////////////////////////////////
  /** KML Container. Abstract Only */
  public class kmlContainer extends kmlFeature {
    Vector<kmlFeature>         features;
    Vector<kmlNetworkLink>     networkLinks;
    Vector<kmlFolder>          folders;

    public kmlContainer () {
      features     = new Vector<kmlFeature>();
      networkLinks = new Vector<kmlNetworkLink>();
      folders      = new Vector<kmlFolder>();
    }

    public void addFeature (kmlFeature feature) {
      features.add(feature);
    }

    /** @deprecated As of NeXtMidas 2.9.0, use {@link #getNetworkLinks()} */
    @Deprecated public kmlNetworkLink getNetworkLink () {  return networkLinks.get(0);}

    /** @deprecated As of NeXtMidas 2.9.0, use addNetworkLink(kmlNetworkLink) */
    @Deprecated public void setNetworkLink (kmlNetworkLink networkLink) { addNetworkLink(networkLink);}

    public Vector<kmlNetworkLink> getNetworkLinks () {
      return networkLinks;
    }

    public void addNetworkLink (kmlNetworkLink link) {
      networkLinks.add(link);
    }

    public void addFolder (kmlFolder folder) {
      folders.add(folder);
    }

    public Vector<kmlFolder> getFolders () {
      return folders;
    }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Document */
  public class kmlDocument extends kmlContainer {

    Vector<kmlSchema>   schemas;

    public kmlDocument () {
      schemas       = new Vector<kmlSchema>();
    }

    public void addSchema (kmlSchema schema) {
      schemas.add(schema);
    }
    public Vector<kmlSchema> getSchemas () {
      return schemas;
    }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Extended Data */
  public class kmlExtendedData extends kmlObject {
    String dataName;
    String dataDisplayName;
    String dataValue;
    String schemaDataUrl;
    String simpleDataName;

    public String getDataName()                { return dataName; }
    public void   setDataName(String dataName) { this.dataName = dataName; }

    public String getDataDisplayName()         { return dataDisplayName; }
    public void   setDataDisplayName(String dataDisplayName) { this.dataDisplayName = dataDisplayName; }

    public String getDataValue()               { return dataValue; }
    public void   setDataValue(String dataValue) { this.dataValue = dataValue; }

    public String getSchemaDataUrl()           { return schemaDataUrl; }
    public void   setSchemaDataUrl(String schemaDataUrl) { this.schemaDataUrl = schemaDataUrl; }

    public String getSimpleDataName()          { return simpleDataName; }
    public void   setSimpleDataName(String simpleDataName) { this.simpleDataName = simpleDataName; }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Folder */
  public class kmlFolder extends kmlContainer {

    public kmlFolder () {

    }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Ground Overlay */
  public class kmlGroundOverlay extends kmlOverlay {

    double           altitude             = 0;

    /** Must be either 'clampToGround' (default), 'relativeToGround', or 'absolute'.  */
    String altitudeMode    = "clampToGround";

    /* The Lat Lon Box specifies the boundaries of a bounding box for the overlay */
    double           north;             // Lat Lon Box: 0 to +/- 90 degrees
    double           south;             // Lat Lon Box: 0 to +/- 90 degrees
    double           east;              // Lat Lon Box: 0 to +/- 180 degrees
    double           west;              // Lat Lon Box: 0 to +/- 180 degrees
    double           rotation = 0;      // Lat Lon Box: 0 (North) to +/- 180 degrees

    public kmlGroundOverlay() {

    }

    public double getAltitude() {  return altitude; }
    public void setAltitude(double altitude) { this.altitude = altitude; }

    public String getAltitudeMode() { return altitudeMode; }

    public void setAltitudeMode(String altitudeMode) {
      if (altitudeMode.equalsIgnoreCase("clamptoground")) {
        this.altitudeMode = "clampToGround";
      } else if (altitudeMode.equalsIgnoreCase("relativetoground")) {
        this.altitudeMode = "relativeToGround";
      } else if (altitudeMode.equalsIgnoreCase("absolute")) {
        this.altitudeMode = "absolute";
      } else {
        Shell.warning("Invalid altitude mode: " +altitudeMode+ ". Mode must be 'clampToGround', 'relativeToGround', or 'absolute'");
      }
    }

    public double getNorth () { return north; }
    public void setNorth (double north) {
      this.north = north;
      updateYRange(north);  // update layer Y value range
    }

    public double getSouth () { return south;}
    public void setSouth (double south) {
      this.south = south;
      updateYRange(south);  // update layer Y value range
    }

    public double getEast () {  return east; }
    public void setEast (double east) {
      this.east = east;
      updateXRange(east);  // update layer X value range
    }

    public double getWest () { return west; }
    public void setWest (double west) {
      this.west = west;
      updateXRange(west);  // update layer X value range
    }

    public double getRotation () { return rotation; }
    public void setRotation (double rotation) { this.rotation = rotation; }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Icon */
  public class kmlIcon extends kmlObject {
    private String     href        = "Cross";       // default to cross

    /** Must be 'onChange', 'onInterval', or 'onExpire' */
    String     refreshMode       = "onChange";

    float      refreshInterval   = 4;

    /** Must be  'never', 'onStop', 'onRequest', or 'onRegion' */
    String     viewRefreshMode   = "never";

    float      viewRefreshTime   = 4;
    float      viewBoundScale    = 1;
    String     viewFormat;
    String     httpQuery;

    public String getHref () {  return href; }
    public void setHref (String url) { this.href = url; }

    public String getRefreshMode() { return refreshMode; }

    public void setRefreshMode (String mode) {
      if (mode.equalsIgnoreCase("onchange")) {
        this.refreshMode = "onChange";
      } else if (mode.equalsIgnoreCase("oninterval")) {
        this.refreshMode = "onInterval";
      } else if (mode.equalsIgnoreCase("onexpire")) {
        this.refreshMode = "onExpire";
      } else {
        Shell.warning("Invalid Refresh Mode: " +mode+ ". Mode must be 'onChange', 'onInterval', or 'onExpire'.");
      }
    }

    public float getRefreshInterval () { return refreshInterval; }
    public void setRefreshInterval (float refreshInterval) { this.refreshInterval = refreshInterval; }

    public String getViewRefreshMode () { return viewRefreshMode;  }

    public void setViewRefreshMode (String mode) {
      if (mode.equalsIgnoreCase("never")) {
        this.viewRefreshMode = "never";
      }  else if (mode.equalsIgnoreCase("onstop")) {
        this.viewRefreshMode = "onStop";
      } else if (mode.equalsIgnoreCase("onrequest")) {
        this.viewRefreshMode = "onRequest";
      } else if (mode.equalsIgnoreCase("onregion")) {
        this.viewRefreshMode = "onRegion";
      } else {
        Shell.warning("Invalid View Refresh Mode: " +mode+ ". Mode must be 'never', 'onStop', 'onRequest', or 'onRegion'.");
      }
    }

    public float getViewRefreshTime () { return viewRefreshTime; }
    public void setViewRefreshTime (float viewRefreshTime) { this.viewRefreshTime = viewRefreshTime; }

    public float getViewBoundScale () { return viewBoundScale; }
    public void setViewBoundScale (float viewBoundScale) { this.viewBoundScale = viewBoundScale; }

    public String getViewFormat () { return viewFormat; }
    public void setViewFormat (String viewFormat) { this.viewFormat = viewFormat; }

    public String getHttpQuery () { return httpQuery;}
    public void setHttpQuery (String httpQuery) { this.httpQuery = httpQuery; }

    /** Parse the Icon's viewFormat, if it exists, replacing the parameters given.<br><br>
        <b>Note: This method may change during the beta series to incorporate more view format elements</b>
        @param north North side of BBOX
        @param south South side of BBOX
        @param east  East side of BBOX
        @param west  West side of BBOX
        @param width Width of panel
        @param height Height of panel
        @return A String (in http format) of the viewFormat
        @since NeXtMidas 2.9.0
     */
    public String parseViewFormat (double north, double south, double east, double west, int width, int height) {
      String raw = getViewFormat();
      if (raw==null) return null;  // no view format
      StringBuffer result = new StringBuffer();
      String[] elems = StringUtil.split(raw, "&");  // split apart the viewFormat
      for (String elem : elems) {
        int indexEQ = elem.indexOf('=');    // make key/value pairs
        if (indexEQ>0) {                                     // have equal sign (get value after it)
          String key = elem.substring(0,indexEQ).trim();     // trim
          String value = elem.substring(indexEQ + 1).trim(); // +1 to skip equal sign

          if (key.equalsIgnoreCase("BBOX")) {   // handle BBOX=
            String[] bbelems = StringUtil.split(value, ",");
            for (int i=0; i<bbelems.length; i++) {
              String bbelem = bbelems[i];
              if (bbelem.startsWith("[")) {
                String dir = bbelem.substring(5,9);
                if      (dir.equalsIgnoreCase("west")) bbelems[i] = ""+west;
                else if (dir.equalsIgnoreCase("sout")) bbelems[i] = ""+south;
                else if (dir.equalsIgnoreCase("east")) bbelems[i] = ""+east;
                else if (dir.equalsIgnoreCase("nort")) bbelems[i] = ""+north;
              }
            }
            result.append("BBOX="+StringUtil.join(bbelems, "%2C"));  // %2C is to comma-serarate the value
          } else if (key.equalsIgnoreCase("width")) {   // handle width=
            if (value.equalsIgnoreCase("[horizPixels]")) value = ""+width;
            result.append("&width="+value);
          } else if (key.equalsIgnoreCase("height")) {  // handle height=
            if (value.equalsIgnoreCase("[vertPixels]")) value = ""+height;
            result.append("&height="+value);
          }
        }
      } // end FOR loop of elements
      return result.toString();
    }
  }
  //////////////////////////////////////////////////////////////////
  /** KML Icon Style <br>
      For NeXtMidas purposes, we treat this as a Symbol
   */
  public class kmlIconStyle extends kmlColorStyle {
    float     scale     = 1;
    float     heading   = 0;
    kmlIcon   icon;
    double    hotSpotX  = .5;
    double    hotSpotY  = .5;

    /** Must be 'fraction', 'pixels', 'insetPixels' */
    String    hotSpotXUnits = "fraction";
    /** Must be 'fraction', 'pixels', 'insetPixels' */
    String    hotSpotYUnits = "fraction";

    public kmlIconStyle () {
      icon = new kmlIcon();
    }

    public float getScale () {  return scale;  }
    public void setScale (float scale) {  this.scale = scale;  }

    public float getHeading () { return heading; }
    public void setHeading (float heading) { this.heading = heading; }

    public kmlIcon getIcon () { return icon; }
    public void setIcon (kmlIcon icon) { this.icon = icon; }

    public double getHotSpotX () {  return hotSpotX;  }
    public void setHotSpotX (double hotSpotX) { this.hotSpotX = hotSpotX; }

    public double getHotSpotY () { return hotSpotY; }
    public void setHotSpotY (double hotSpotY) { this.hotSpotY = hotSpotY; }

    public String getHotSpotXUnits () { return hotSpotXUnits; }

    public void setHotSpotXUnits (String hotSpotXUnits) {
      this.hotSpotXUnits = validateUnits(hotSpotXUnits);
    }

    public String getHotSpotYUnits () {  return hotSpotYUnits; }

    public void setHotSpotYUnits (String hotSpotYUnits) {
      this.hotSpotYUnits = validateUnits(hotSpotYUnits);
    }

    private String validateUnits (String units) {
      String results = new String();;

      if (units.equalsIgnoreCase("fraction")) {
        results = "fraction";
      } else if (units.equalsIgnoreCase("pixels")) {
        results = "pixels";
      } else if (units.equalsIgnoreCase("insetpixels")) {
        results = "insetPixels";
      } else {
        Shell.warning("Invalid units: " +units+ ". Setting Units to default (fraction).");
        results = "fraction";
      }
      return results;
    }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Label Style */
  public class kmlLabelStyle extends kmlColorStyle {
    float scale  = 1 ;     // scale factor applied to the label

    public float getScale() { return scale;  }

    public void setScale(float scale) { this.scale = scale; }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Linear Ring */
  public class kmlLinearRing extends kmlGeometry {

    /** Specifies whether to connect to the ground.
        To extrude, altitudeMode shall be either 'relativeToGround',
         or 'absolute', and the altitude should be > 0.
     */
    boolean   extrude = false;

    /** Specifies whether to drape over the terrain.
        To enable, altitudeMode shall be 'clampToGround'
     */
    boolean   tessellate = false;

    /** Must be either 'clampToGround' (default), 'relativeToGround', or 'absolute'.  */
    String altitudeMode    = "clampToGround";

    Vector<kmlCoordinates> coordsVector;
    String                 coordsString;

    public kmlLinearRing () {
      coordsVector = new Vector<kmlCoordinates>();
      coordsString = new String();
    }

    public kmlLinearRing (kmlFeature feat) {
      this();
      setParentFeature(feat);
    }

    public boolean isExtrude() { return extrude; }
    public void setExtrude(boolean extrude) { this.extrude = extrude;  }

    public boolean isTessellate() { return tessellate;   }
    public void setTessellate(boolean tessellate) { this.tessellate = tessellate; }

    public String getAltitudeMode() {  return altitudeMode;  }

    public void setAltitudeMode(String altitudeMode) {
      if (altitudeMode.equalsIgnoreCase("clamptoground")) {
        this.altitudeMode = "clampToGround";
      } else if (altitudeMode.equalsIgnoreCase("relativetoground")) {
        this.altitudeMode = "relativeToGround";
      } else if (altitudeMode.equalsIgnoreCase("absolute")) {
        this.altitudeMode = "absolute";
      } else {
        Shell.warning("Invalid altitude mode: " +altitudeMode+ ". Mode must be 'clampToGround', 'relativeToGround', or 'absolute'");
      }
    }

    public void setCoordsString (String coords) {  coordsString = coords; }

    public String getCoordsString () { return coordsString;  }

    /** Converts all the sets of coordinates for the LineString into a single
         double array. This is done for performance reasons when drawing the
         lines.
        @return All the coordinates as a single double array.
     */
    public double[] getCoordsAsDouble () {
      double[] coords=null;

      if (coordsVector.isEmpty()) {                 // make sure it is not empty
        Shell.warning("Coordinates not set");
        return null;
      } else {
        int size = coordsVector.size();    // get the number of coordinate sets
        coords = new double[size*3];       // x3 (Lon/Lat/Alt in each set)
        int coordsIndex = 0;

        for (int xx=0; xx< size; xx++) {
          coords[coordsIndex] = coordsVector.get(xx).getLon();
          coordsIndex++;
          coords[coordsIndex] = coordsVector.get(xx).getLat();
          coordsIndex++;
          coords[coordsIndex] = coordsVector.get(xx).getAltitude();
          coordsIndex++;
        }
        return coords;
      }
    }

    /** Set the coordinates Vector from String form.
        This method converts the coordinates from their raw from into
         the Vector (lon,lat,alt) used by the class, stripping away carriage
         returns and whitespace.
     */
    public void setCoordsVector (String coordinates) {
      String[] tmp = coordinates.split("\r\n|\r|\n|\\s");
      int counter = 0;
      for (int xx=0; xx< tmp.length; xx++) {
        tmp[xx] = tmp[xx].trim();
        if (!tmp[xx].equals("")) {
          kmlCoordinates coords = new kmlCoordinates();
          coords.setCoordinates(tmp[xx]);
          coordsVector.add(coords);
          counter++;
        }
      }
      //System.out.println("Number of coord sets=" + counter);
    }
  }
  ///////////////////////////////////////////////////////////////////
  /** KML Line String */
  public class kmlLineString extends kmlGeometry {

    /** Specifies whether to connect to the ground.
        To extrude, altitudeMode shall be either 'relativeToGround',
         or 'absolute', and the altitude should be > 0.
     */
    boolean   extrude    = false;

    /** Specifies whether to drape over the terrain.
        To enable, altitudeMode shall be 'clampToGround'
     */
    boolean   tessellate = false;

    /** Must be either 'clampToGround' (default), 'relativeToGround', or 'absolute'.  */
    String altitudeMode    = "clampToGround";

    String                 coordsString = new String();
    Vector<kmlCoordinates> coordsVector;   // Holds as many coordinate sets as needed

    public kmlLineString () {
      coordsVector = new Vector<kmlCoordinates>();
    }

    public kmlLineString (kmlFeature feat) {
      this();
      setParentFeature(feat);
    }

    public boolean isExtrude() { return extrude; }
    public boolean getExtrude() { return isExtrude(); }   // convenience method
    public void setExtrude(boolean extrude) { this.extrude = extrude; }

    public boolean isTessellate() { return tessellate; }
    public boolean getTessellate() { return isTessellate(); }
    public void setTessellate(boolean tessellate) { this.tessellate = tessellate;  }

    public String getAltitudeMode () { return altitudeMode; }

    public void setAltitudeMode(String altitudeMode) {
      if (altitudeMode.equalsIgnoreCase("clamptoground")) {
        this.altitudeMode = "clampToGround";
      } else if (altitudeMode.equalsIgnoreCase("relativetoground")) {
        this.altitudeMode = "relativeToGround";
      } else if (altitudeMode.equalsIgnoreCase("absolute")) {
        this.altitudeMode = "absolute";
      } else {
        Shell.warning("Invalid altitude mode: " +altitudeMode+ ". Mode must be 'clampToGround', 'relativeToGround', or 'absolute'");
      }
    }

    public void setCoordsString (String coords) { coordsString = coords;  }

    public String getCoordsStrikmlColorStyleng () {  return coordsString;  }

    /** Converts all the sets of coordinates for the LineString into a single
         double array. This is done for performance reasons when drawing the
         lines.
        @return All the coordinates as a single double array.
     */
    public synchronized double[] getCoordsAsDouble () {
      double[] coords=null;

      if (coordsVector.isEmpty()) {                 // make sure it is not empty
        Shell.warning("Coordinates not set");
        return null;
      } else {
        int size = coordsVector.size();    // get the number of coordinate sets
        coords = new double[size*3];       // x3 (Lon/Lat/Alt in each set)
        int coordsIndex = 0;

        for (int xx=0; xx< size; xx++) {
          coords[coordsIndex++] = coordsVector.get(xx).getLon();
          coords[coordsIndex++] = coordsVector.get(xx).getLat();
          coords[coordsIndex++] = coordsVector.get(xx).getAltitude();
        }

        return coords;
      }
    }

    /** Set the coordinates Vector from String form.
        This method converts the coordinates from their raw form into
         the Vector (lon,lat,alt) used by the class, stripping away carriage
         returns and whitespace.
     */
    public synchronized void setCoordsVector (String coordinates) {
      String[] tmp = coordinates.split("\r\n|\r|\n|\\s");
      for (int xx=0; xx< tmp.length; xx++) {
        tmp[xx] = tmp[xx].trim();
        if (!tmp[xx].equals("")) {
          kmlCoordinates coords = new kmlCoordinates();
          coords.setCoordinates(tmp[xx]);
          coordsVector.add(coords);
        }
      }
      //System.out.println("Number of coord sets=" + tmp.length);
    }
  }
  ///////////////////////////////////////////////////////////////////
  /** KML Line Style */
  public class kmlLineStyle extends kmlColorStyle {
    float width = 1;     // width in pixels

    public void setWidth (float w) { width = w; }
    public float getWidth () {return width; }
  }
  ///////////////////////////////////////////////////////////////////
  /** KML Link <br>
      Note: refreshInterval deals with time (i.e. poll rate), while
       viewRefreshInterval deals with the "camera" view.
   */
  public class kmlLink extends kmlObject {
    String href;                           // URL to a location

    /** Time-based refresh mode. MUST be 'onChange' (default), 'onInterval', or 'onExpire' */
    String refreshMode       = "onChange";
    float  refreshInterval   = 4;          // Refresh the file every N seconds

    /** Must be 'never' (default), 'onStop', 'onRequest', 'onRegion' */
    String viewRefreshMode   = "never";
    float  viewRefreshTime   = 4;          // Seconds to wait after camera stops to refresh
    float  viewBoundScale    = 1;          // Scale of the bounding box.
    //  < 1 = smaller than current view
    //  > 1 = larger then current view
    String viewFormat;
    String httpQuery;

    public kmlLink () {
      href       = new String();
      viewFormat = new String();
      httpQuery  = new String();
    }

    public String getHref() { return href;  }
    public void setHref(String href) {  this.href = href; }

    public String getRefreshMode() { return refreshMode; }

    public void setRefreshMode (String mode) {
      if (mode.equalsIgnoreCase("onchange")) {
        this.refreshMode = "onChange";
      } else if (mode.equalsIgnoreCase("oninterval")) {
        this.refreshMode = "onInterval";
      } else if (mode.equalsIgnoreCase("onexpire")) {
        this.refreshMode = "onExpire";
      } else {
        Shell.warning("Invalid Refresh Mode: " +mode+ ". Mode must be 'onChange', 'onInterval', or 'onExpire'.");
      }
    }

    public float getRefreshInterval () { return refreshInterval; }
    public void setRefreshInterval (float refreshInterval) { this.refreshInterval = refreshInterval; }

    public String getViewRefreshMode () { return viewRefreshMode;  }

    public void setViewRefreshMode (String mode) {
      if (mode.equalsIgnoreCase("never")) {
        this.viewRefreshMode = "never";
      }  else if (mode.equalsIgnoreCase("onstop")) {
        this.viewRefreshMode = "onStop";
      } else if (mode.equalsIgnoreCase("onrequest")) {
        this.viewRefreshMode = "onRequest";
      } else if (mode.equalsIgnoreCase("onregion")) {
        this.viewRefreshMode = "onRegion";
      } else {
        Shell.warning("Invalid View Refresh Mode: " +mode+ ". Mode must be 'never', 'onStop', 'onRequest', or 'onRegion'.");
      }
    }

    public float getViewRefreshTime () { return viewRefreshTime; }
    public void setViewRefreshTime (float viewRefreshTime) { this.viewRefreshTime = viewRefreshTime; }

    public float getViewBoundScale () { return viewBoundScale; }
    public void setViewBoundScale (float viewBoundScale) { this.viewBoundScale = viewBoundScale; }

    public String getViewFormat () { return viewFormat; }
    public void setViewFormat (String viewFormat) { this.viewFormat = viewFormat; }

    public String getHttpQuery () { return httpQuery;}
    public void setHttpQuery (String httpQuery) { this.httpQuery = httpQuery; }


  }
  ///////////////////////////////////////////////////////////////////
  /** KML List Style */
  public class kmlListStyle extends kmlObject {
    Color  bgColor = Color.white;   // Default to white

    /** Must be  'check' (default), 'checkOffOnly', 'checkHideChildren', or
         'radioFolder' */
    String listItemType  = "check";

    Vector<kmlIcon> icons;

    public kmlListStyle () {
      icons = new Vector<kmlIcon>();
    }

    public void addIcon (kmlIcon icon) {
      icons.add(icon);
    }

    public Color getBgColor() {  return bgColor; }
    public void setBgColor(Color bgColor) { this.bgColor = bgColor; }

    public String getListItemType() { return listItemType; }

    /** Must be  'check' (default), 'checkOffOnly', 'checkHideChildren', or
    'radioFolder' */
    public void setListItemType(String listItemType) {
      if (listItemType.equalsIgnoreCase("check")) {
        this.listItemType = "check";
      } else if (listItemType.equalsIgnoreCase("checkoffonly")) {
        this.listItemType = "checkOffOnly";
      } else if (listItemType.equalsIgnoreCase("checkHideChildren")) {
        this.listItemType = "checkHideChildren";
      } else if (listItemType.equalsIgnoreCase("radioFolder")) {
        this.listItemType = "radioFolder";
      } else {
        Shell.warning("Invalid List Item Type: " +listItemType+
        ". Type must be 'check' (default), 'checkOffOnly', 'checkHideChildren', or 'radioFolder'");
      }
    }

    public Vector<kmlIcon> getIcons() { return icons; }


  }
  ///////////////////////////////////////////////////////////////////
  /** KML LookAt */
  public class kmlLookAt extends kmlAbstractView {
    double lon     = 0;       // 0 to +/- 180 degrees
    double lat     = 0;       // 0 to +/-  90 degrees
    double alt     = 0;       // meters
    double heading = 0;       // 0 to 360 degrees
    double tilt    = 0;       // 0 to 90 degrees
    double range;

    /** Must be 'clampToGround' (default), 'relativeToGround', or 'absolute' */
    String altitudeMode = "clampToGround";

    public kmlLookAt () {}

    public double getLon() {  return lon; }
    public void setLon(double lon) { this.lon = lon; }

    public double getLat() {  return lat;  }
    public void setLat(double lat) { this.lat = lat; }

    public double getAlt() { return alt; }
    public void setAlt(double alt) { this.alt = alt; }

    public double getHeading() { return heading;  }
    public void setHeading(double heading) { this.heading = heading; }

    public double getTilt() { return tilt; }
    public void setTilt(double tilt) { this.tilt = tilt; }

    public double getRange() { return range; }
    public void setRange(double range) { this.range = range; }

    public String getAltitudeMode() { return altitudeMode; }
    public void setAltitudeMode(String altitudeMode) { this.altitudeMode = altitudeMode; }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Model */
  public class kmlModel extends kmlGeometry {

    /** Must be 'clampToGround' (default), 'relativeToGround', or 'absolute' */
    String altitudeMode = "clampToGround";

    kmlCoordinates coords;                      // location

    /* Orientation */
    double heading           = 0;               // 0 to 360 degrees
    double tilt              = 0;               // 0 to 360 degrees
    double roll              = 0;               // 0 to 360 degrees

    /* Scale */
    float scaleX            = 1;
    float scaleY            = 1;
    float scaleZ            = 1;

    kmlLink link;

    /* Resource Map Alias */
    /* TODO: NEEDS TO BE OWN CLASS TO ALLOW FOR MULTIPLE MAPS */
    String targetHref;
    String sourceHref;

    public kmlModel () {
      link = new kmlLink();
    }

    public kmlModel (kmlFeature feat) {
      this();
      setParentFeature(feat);
    }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Multi Geometry */
  public class kmlMultiGeometry extends kmlGeometry {

    Vector<kmlGeometry> geometrys;     // vector to hold the kmlGeometry objects

    public kmlMultiGeometry () {
      geometrys = new Vector<kmlGeometry>();    // instantiate the Vector
    }

    public kmlMultiGeometry (kmlFeature feat) {
      this();
      setParentFeature(feat);
    }

    /** Add a kmlGeometry object to the Vector
        @param geometry The kmlGeometry object
     */
    public void addGeometry (kmlGeometry geometry) {
      geometrys.add(geometry);
    }

  }
  ///////////////////////////////////////////////////////////////////
  /** KMl Network Link */
  public class kmlNetworkLink extends kmlFeature {
    boolean refreshVisibility  = false;
    boolean flyToView          = false;
    kmlLink link;

    public kmlNetworkLink () {
      link = new kmlLink();
    }

    public void setLink (kmlLink link) {
      this.link = link;
    }
  }
  ///////////////////////////////////////////////////////////////////
  /** KMl Network Link Control */
  public class kmlNetworkLinkControl {

    float minRefreshPeriod = 0;      // seconds between fetches of the file
    float maxSessionLength = -1;     // seconds to stay connected. '-1' = do not disconnect
    String cookie;
    String message;
    String linkName;
    String linkDescription;
    String linkSnippet;
    int linkSnippetMaxLines = 2;
    Date expires;

    /** Must be 'Change', 'Create', or 'Delete' */
    String update = new String();

    /** Will be a <code>kmlLookAt</code> or <code>kmlCamera</code>.  */
    kmlAbstractView abstractView;

    public kmlNetworkLinkControl () {
      cookie           = new String();
      message          = new String();
      linkName         = new String();
      linkDescription  = new String();
      linkSnippet      = new String();

      abstractView     = new kmlAbstractView();
    }

    public float getMinRefreshPeriod() {  return minRefreshPeriod; }
    public void setMinRefreshPeriod (float minRefreshPeriod) { this.minRefreshPeriod = minRefreshPeriod; }

    public float getMaxSessionLength () { return maxSessionLength; }
    public void setMaxSessionLength (float maxSessionLength) {this.maxSessionLength = maxSessionLength;}

    public String getCookie () { return cookie;}
    public void setCookie (String cookie) { this.cookie = cookie; }

    public String getMessage () { return message; }
    public void setMessage (String message) { this.message = message; }

    public String getLinkName () { return linkName; }
    public void setLinkName (String linkName) { this.linkName = linkName;  }

    public String getLinkDescription () {  return linkDescription;  }
    public void setLinkDescription (String linkDescription) { this.linkDescription = linkDescription; }

    public String getLinkSnippet () {  return linkSnippet; }
    public void setLinkSnippet (String linkSnippet) { this.linkSnippet = linkSnippet; }

    public int getLinkSnippetMaxLines () { return linkSnippetMaxLines; }
    public void setLinkSnippetMaxLines (int linkSnippetMaxLines) { this.linkSnippetMaxLines = linkSnippetMaxLines; }

    public Date getExpires () { return expires; }
    public void setExpires (Date expires) { this.expires = expires;  }

    public String getUpdate () {  return update; }

    public void setUpdate (String update) {
      if (update.equalsIgnoreCase("change")){
        this.update = "Change";
      } else if (update.equalsIgnoreCase("create")) {
        this.update = "Create";
      } else if (update.equalsIgnoreCase("delete")) {
        this.update = "Delete";
      } else {
        Shell.warning("Invalid Update entry: " +update+ ". Update must be 'Change', 'Create', or 'Delete'");
      }
    }

    public kmlAbstractView getAbstractView () { return abstractView; }
    public void setAbstractView (kmlAbstractView abstractView) { this.abstractView = abstractView; }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Overlay - Abstract Element */
  public class kmlOverlay extends kmlFeature {

    private  Color    color;
    private  int      drawOrder;
    private  kmlIcon  icon;          // Image associated with the overlay

    public kmlOverlay () {
      color     = Color.white;    // white by default
      drawOrder = 0;
      icon      = new kmlIcon();
    }

    public Color getColor() { return color;  }
    public void setColor(Color color) { this.color = color;  }

    public int getDrawOrder() {  return drawOrder;  }
    public void setDrawOrder(int drawOrder) { this.drawOrder = drawOrder; }

    public kmlIcon getIcon() { return icon; }
    public void setIcon(kmlIcon icon) { this.icon = icon; }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Photo Overlay */
  public class kmlPhotoOverlay extends kmlOverlay {

    /** Must be  'rectangle' (default), 'cylinder', or 'sphere' */
    String shape = "rectangle";

    /* View Volume (How much of the current scene is visible) */
    double  leftFov   = 0;          // Left Field of View. 0 to 180 degrees
    double  rightFov  = 0;          // Right Field of View. 0 to 180 degrees
    double  bottomFov = 0;          // Right Field of View. 0 to 90 degrees
    double  topFov    = 0;          // Top Field of View. 0 to 90 degrees
    double  near      = 0;          // Meters from the camera

    float  roll;                   // Adjusts the photo inside the field of view. 0 to 180 degrees

    /* Image Pyramid (Heirarchical set of images, increasing in resolution */
    int    tileSize   = 256;       // Size in pixels. Must be a square (power of 2)
    int    maxWidth;
    int    maxHeight;

    /** Must be 'lowerLeft' (default) or 'upperLeft' */
    String gridOrigin = "lowerleft"; //

    kmlPoint point;                //

    public kmlPhotoOverlay() {
      point = new kmlPoint();
    }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Placemark */
  public class kmlPlacemark extends kmlFeature {

    kmlGeometry geometry;

    public kmlPlacemark () {
      geometry = new kmlGeometry();
    }

    public kmlGeometry getGeometry() { return geometry; }
    public void setGeometry(kmlGeometry geometry) { this.geometry = geometry; }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Point */
  public class kmlPoint extends kmlGeometry {

    /** Specifies whether to connect to the ground.
        To extrude, altitudeMode shall be either 'relativeToGround',
         or 'absolute', and the altitude should be > 0.
     */
    boolean   extrude    = false;

    /** Must be either 'clampToGround' (default), 'relativeToGround', or 'absolute'.  */
    String altitudeMode    = "clampToGround";

    kmlCoordinates coords;

    /** Default constructor */
    public kmlPoint () {
      coords = new kmlCoordinates();   // instantiate kmlCoordinates
    }

    /** Keep track of the feature so we can call things like the name and desc */
    public kmlPoint (Object obj) {
      this();
      if (obj instanceof kmlFeature) {
        setParentFeature((kmlFeature)obj);
      } else if (obj instanceof kmlMultiGeometry) {
        // possible future use
      }
    }

    public boolean getExtrude () { return isExtrude(); }
    public boolean isExtrude () { return extrude;  }

    public void setExtrude (boolean extrude) { this.extrude = extrude; }

    public String getAltitudeMode () { return altitudeMode; }

    public void setAltitudeMode (String altitudeMode) {
      if (altitudeMode.equalsIgnoreCase("clamptoground")) {
        this.altitudeMode = "clampToGround";
      } else if (altitudeMode.equalsIgnoreCase("relativetoground")) {
        this.altitudeMode = "relativeToGround";
      } else if (altitudeMode.equalsIgnoreCase("absolute")) {
        this.altitudeMode = "absolute";
      } else {
        Shell.warning("Invalid altitude mode: " +altitudeMode+ ". Mode must be 'clampToGround', 'relativeToGround', or 'absolute'");
      }
    }

    public kmlCoordinates getCoords () { return coords; }

    /** Set the kmlCoordinates for the point based on an input String
        @param string The coordinates in String form, which is how they appear
                      to the parser.
     */
    public void setCoordinates (String string) {
      //System.out.println("Coords= " +string);    // Debug Statement
      coords.setCoordinates(string);
    }
  }
  ///////////////////////////////////////////////////////////////////
  /** KML Polygon */
  public class kmlPolygon extends kmlGeometry {

    /** Specifies whether to connect to the ground.
        To extrude, altitudeMode shall be either 'relativeToGround',
         or 'absolute', and the altitude should be > 0.
     */
    boolean   extrude    = false;

    /** Specifies whether to drape over the terrain.
        To enable, altitudeMode shall be 'clampToGround'
     */
    boolean   tessellate = false;

    /** Must be either 'clampToGround' (default), 'relativeToGround', or 'absolute'.  */
    String altitudeMode    = "clampToGround";

    kmlLinearRing         outerBoundary;  // only 1
    boolean               outer;          // flag to show whether the current linear ring is the outer ring
    Vector<kmlLinearRing> innerBoundary;  // can be 0 or more

    public kmlPolygon () {
      outerBoundary = new kmlLinearRing();
      innerBoundary = new Vector<kmlLinearRing>();
    }

    public kmlPolygon (kmlFeature feat) {
      this();
      setParentFeature(feat);
    }

    public void addInnerBoundary (kmlLinearRing linearRing) {
      innerBoundary.add(linearRing);
    }

    public void setOuterBoundary (kmlLinearRing outerBoundary) {
      this.outerBoundary = outerBoundary;
    }

    public kmlLinearRing getOuterBoundary () {
      return outerBoundary;
    }

    public boolean isOuter () {return outer;}
    public boolean getOuter () { return outer; }
    public void setOuter (boolean outer) { this.outer = outer; }

    public boolean isExtrude () { return extrude;  }
    public boolean getExtrude () { return isExtrude(); }
    public void setExtrude (boolean extrude) {  this.extrude = extrude;  }

    public boolean isTessellate () { return tessellate; }
    public void setTessellate (boolean tessellate) { this.tessellate = tessellate; }

    public String getAltitudeMode () { return altitudeMode; }

    public void setAltitudeMode(String altitudeMode) {
      if (altitudeMode.equalsIgnoreCase("clamptoground")) {
        this.altitudeMode = "clampToGround";
      } else if (altitudeMode.equalsIgnoreCase("relativetoground")) {
        this.altitudeMode = "relativeToGround";
      } else if (altitudeMode.equalsIgnoreCase("absolute")) {
        this.altitudeMode = "absolute";
      } else {
        Shell.warning("Invalid altitude mode: " +altitudeMode+ ". Mode must be 'clampToGround', 'relativeToGround', or 'absolute'");
      }
    }

    public void addCoordinates(String coords) {
      kmlLinearRing linearRing = new kmlLinearRing();
      linearRing.setCoordsVector(coords);
      innerBoundary.add(linearRing);
    }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Poly Style */
  public class kmlPolyStyle extends kmlColorStyle {
    boolean fill         = true;    // Whether to fill the polygon
    boolean outline      = true;    // Whether to outline the polygon

    public boolean isFill () { return fill; }
    public void setFill (boolean fill) { this.fill = fill; }

    public boolean isOutline () { return outline; }
    public void setOutline (boolean outline) { this.outline = outline; }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Screen Overlay */
  public class kmlScreenOverlay extends kmlOverlay {

    /* OverlayXY (Point on the overlay mapped to screen coordinates) */
    float  overlayX;
    float  overlayY;
    /** MUST be 'fraction' (default), 'pixels', or 'insetPixels' */
    String    overlayXUnits    = "fraction";
    /** MUST be 'fraction' (default), 'pixels', or 'insetPixels' */
    String    overlayYUnits    = "fraction";   //

    /* ScreenXY (Point relative to the screen that the overlay is mapped to) */
    float  screenX;
    float  screenY;
    /** MUST be 'fraction' (default), 'pixels', or 'insetPixels' */
    String    screenXUnits     = "fraction";   //
    /** MUST be 'fraction' (default), 'pixels', or 'insetPixels' */
    String    screenYUnits     = "fraction";   //

    /* RotationXY (Point relative to the screen about which the overlay is rotated) */
    float  rotationX;
    float  rotationY;
    /** MUST be 'fraction' (default), 'pixels', or 'insetPixels' */
    String    rotationXUnits   = "fraction";   //
    /** MUST be 'fraction' (default), 'pixels', or 'insetPixels' */
    String    rotationYUnits   = "fraction";   //

    /** Size (Size of the image for the overlay)
       '-1' = Use the native dimension
       '0'  = Maintain the aspect ratio
        n   = Sets the value to the dimension to n
     */
    float  sizeX;
    float  sizeY;
    /** MUST be 'fraction' (default), 'pixels', or 'insetPixels' */
    String    sizeXUnits       = "fraction";   //
    /** MUST be 'fraction' (default), 'pixels', or 'insetPixels' */
    String    sizeYUnits       = "fraction";   //

    float  rotation         = 0;   // 0 to +/- 180 degrees

  }

  ///////////////////////////////////////////////////////////////////
  /** KML Schema */
  public class kmlSchema {

    String name;
    String schemaID;
    String simpleFieldType;
    String simpleFieldName;
    String displayName;

    public kmlSchema () {
      name            = new String();
      schemaID        = new String();
      simpleFieldType = new String();
      simpleFieldName = new String();
      displayName     = new String();

    }

    public String getName() { return name;  }
    public void setName(String name) { this.name = name; }

    public String getTargetID() { return schemaID; }
    public void setTargetID(String targetID) { this.schemaID = targetID;  }

    public String getSimpleFieldType() { return simpleFieldType; }
    public void setSimpleFieldType(String simpleFieldType) { this.simpleFieldType = simpleFieldType; }

    public String getSimpleFieldName() { return simpleFieldName; }
    public void setSimpleFieldName(String simpleFieldName) { this.simpleFieldName = simpleFieldName;  }

    public String getDisplayName() { return displayName; }
    public void setDisplayName(String displayName) { this.displayName = displayName;  }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Style */
  public class kmlStyle extends kmlStyleSelector {

    /* The boolean variables are used to check for existence of styles later */
    private kmlBalloonStyle balloonStyle;
    private boolean         hasBalloonStyle = false;
    private kmlIconStyle    iconStyle;
    private boolean         hasIconStyle    = false;
    private kmlLabelStyle   labelStyle;
    private boolean         hasLabelStyle   = false;
    private kmlLineStyle    lineStyle;
    private boolean         hasLineStyle    = false;
    private kmlListStyle    listStyle;
    private boolean         hasListStyle    = false;
    private kmlPolyStyle    polyStyle;
    private boolean         hasPolyStyle    = false;

    public kmlStyle () {
      balloonStyle  = new kmlBalloonStyle();
      iconStyle     = new kmlIconStyle();
      labelStyle    = new kmlLabelStyle();
      lineStyle     = new kmlLineStyle();
      listStyle     = new kmlListStyle();
      polyStyle     = new kmlPolyStyle();
    }

    public kmlBalloonStyle getBalloonStyle() {  return balloonStyle; }
    public void setBalloonStyle(kmlBalloonStyle balloonStyle) {
      this.balloonStyle = balloonStyle;
      hasBalloonStyle = true;
    }

    public kmlIconStyle getIconStyle() { return iconStyle; }
    public void setIconStyle(kmlIconStyle iconStyle) {
      this.iconStyle = iconStyle;
      hasIconStyle = true;
    }

    public kmlLabelStyle getLabelStyle() { return labelStyle; }
    public void setLabelStyle(kmlLabelStyle labelStyle) {
      this.labelStyle = labelStyle;
      hasLabelStyle = true;
    }

    public kmlLineStyle getLineStyle() { return lineStyle; }
    public void setLineStyle(kmlLineStyle lineStyle) {
      this.lineStyle = lineStyle;
      hasLineStyle = true;
    }

    public kmlListStyle getListStyle() { return listStyle; }
    public void setListStyle(kmlListStyle listStyle) {
      this.listStyle = listStyle;
      hasListStyle = true;
    }

    public kmlPolyStyle getPolyStyle() { return polyStyle; }
    public void setPolyStyle(kmlPolyStyle polyStyle) {
      this.polyStyle = polyStyle;
      hasPolyStyle = true;
    }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Style Selector */
  public class kmlStyleSelector extends kmlObject {

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Style Map */
  public class kmlStyleMap extends kmlStyleSelector {
    /** Must be 'normal' (default) or 'highlight' */
    String normalKey    = "normal";
    String highlightKey = "highlight";

    String normalStyleUrl;
    String highlightStyleUrl;
  }

  ///////////////////////////////////////////////////////////////////
  /** KML Region
      A Region is 'active' when the bounding box is within view and the user's
       LOD requirements are met.
      Features associated with a Region are only drawn if a Region is active.
   */
  public class kmlRegion extends kmlObject {

    /* Bounding Box (Area of Interest) */
    double  north;                            // 0 to 90 degrees
    double  south;                            // 0 to 90 degrees
    double  east;                             // 0 to 180 degrees
    double  west;                             // 0 to 180 degrees
    double  minAltitude  = 0;
    double  maxAltitude  = 0;

    /** Must be 'clampToGround' (default), 'relativeToGround', or 'absolute' */
    String altitudeMode = "clampToGround";

    /* Lod (Validity Range in Terms of Projected Screen Size) */
    float minLodPixels  = 0;
    float maxLodPixels  = -1;
    float minFadeExtant = 0;                // Opaqueness / distance
    float maxFadeExtant = 0;

    public kmlRegion () {}

    public double getNorth() { return north; }
    public void setNorth(double north) {
      this.north = north;
      updateYRange(north);   // update layer Y value range
    }

    public double getSouth() {  return south;  }
    public void setSouth(double south) {
      this.south = south;
      updateYRange(south);  // update layer Y value range
    }

    public double getEast() {  return east; }
    public void setEast(double east) {
      this.east = east;
      updateXRange(east);  // update layer X value range
    }

    public double getWest() {  return west; }
    public void setWest(double west) {
      this.west = west;
      updateXRange(west);  // update layer X value range
    }

    public double getMinAltitude() {  return minAltitude;  }
    public void setMinAltitude(double minAltitude) { this.minAltitude = minAltitude; }

    public double getMaxAltitude() {  return maxAltitude; }
    public void setMaxAltitude(double maxAltitude) { this.maxAltitude = maxAltitude; }

    public String getAltitudeMode() { return altitudeMode;  }

    public void setAltitudeMode(String altitudeMode) {
      if (altitudeMode.equalsIgnoreCase("clamptoground")) {
        this.altitudeMode = "clampToGround";
      } else if (altitudeMode.equalsIgnoreCase("relativetoground")) {
        this.altitudeMode = "relativeToGround";
      } else if (altitudeMode.equalsIgnoreCase("absolute")) {
        this.altitudeMode = "absolute";
      } else {
        Shell.warning("Invalid altitude mode: " +altitudeMode+ ". Mode must be 'clampToGround', 'relativeToGround', or 'absolute'");
      }
    }

    public float getMinLodPixels() { return minLodPixels;  }
    public void setMinLodPixels(float minLodPixels) { this.minLodPixels = minLodPixels;  }

    public float getMaxLodPixels() { return maxLodPixels; }
    public void setMaxLodPixels(float maxLodPixels) { this.maxLodPixels = maxLodPixels;}

    public float getMinFadeExtant() { return minFadeExtant; }
    public void setMinFadeExtant(float minFadeExtant) { this.minFadeExtant = minFadeExtant; }

    public float getMaxFadeExtant() { return maxFadeExtant; }
    public void setMaxFadeExtant(float maxFadeExtant) { this.maxFadeExtant = maxFadeExtant; }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML Time Primitive */
  public class kmlTimePrimitive extends kmlObject {

    public kmlTimePrimitive () {}
  }
  ///////////////////////////////////////////////////////////////////
  /** KML Time Span */
  public class kmlTimeSpan extends kmlTimePrimitive {
    // redo with NXM timestamp, if applicable
    Time begin;    // time to begin
    Time end;      // time to end

    public kmlTimeSpan () {}

    public Time getBegin() { return begin; }
    public void setBegin(Time begin) { this.begin = begin;  }

    public Time getEnd() { return end;  }
    public void setEnd(Time end) { this.end = end; }

  }
  ///////////////////////////////////////////////////////////////////
  /** KML TimeStamp */
  public class kmlTimeStamp extends kmlTimePrimitive {
    Time when;    // REDO with NXM timeStamp, if applicable

    public kmlTimeStamp () {}

    public Time getWhen() {  return when; }

    public void setWhen(Time when) { this.when = when; }
  }
  ///////////////////////////////////////////////////////////////////

  /** A class to create a popup panel, which is used to display
      detailed information about <code>kmlPoint</code>s
      @since NeXtMidas 2.7.2
   */
  public class Popup {

    private static final int DEFAULT_XPOS = 200;
    private static final int DEFAULT_YPOS = 200;
    JFrame frame;
    String label;

    public Popup (Table tab) {
      this(tab, DEFAULT_XPOS, DEFAULT_YPOS);
    }

    public Popup (Table tab, int xPos, int yPos) {
      label    = tab.getS("label");
      frame    = new JFrame(label);

      frame.setDefaultCloseOperation(JFrame.HIDE_ON_CLOSE);
      PopupPanel myPopup = new PopupPanel(tab);

      // keep the popup size between 250 - 600 pixels
      int w = Math.min(600, Math.max(200, myPopup.getPreferredSize().width));
      int h = Math.min(600, Math.max(250, myPopup.getPreferredSize().height));
      frame.setSize(w,h);

      frame.setLocation(xPos - (w/2), yPos - (h/2));  // where to popup on the screen
      frame.setContentPane(myPopup);
      frame.setVisible(true);
    }

    /** A class to popup a panel on the screen with information about a <code>kmlPoint</code>
        @since NeXtMidas 2.7.2
     */
    @SuppressWarnings("serial")
    public class PopupPanel extends JPanel {

      /** size (in pixels) to draw Images */
      private static final int DEFAULT_SIZE = 100;
      /** max size allowable for popup images */
      private static final int MAX_SIZE = 130;
      /** min size allowable for popup images */
      private static final int MIN_SIZE = 10;

      private Symbol        symbol;            // the Symbol that was drawn
      private BufferedImage img;               // the Image used in imageIcon
      private Line          line;              // Line used to draw the point
      private ImageIcon     icon = null;       // The ImageIcon to draw
      private JScrollPane   scrollPane;        // The scrollpane used to display the text
      private String        text;              // The text to display

      public PopupPanel(Table tab) {
        this(tab, DEFAULT_SIZE);
      }

      public PopupPanel(Table tab, int symbolSize) {
        setLayout(new GridLayout(1,1));

        // Get the other values from the input Table
        String label  = tab.getS("label");
        double xPos   = tab.getD("x");    // x Position of the point
        double yPos   = tab.getD("y");    // y Position of the point
        String desc   = tab.getS("desc"); // The description of the point
        kmlBalloonStyle style = (kmlBalloonStyle) tab.getO("balloonstyle");

        line   = (Line)tab.getO("line");
        symbol = line.getSymbol();
        float heading = symbol.getHeading();

        // Set Up the look and fell of the popup based on data in the kmlBalloonStyle
        setupStyle(style, label, desc);

        // Setup the image of the point to draw
        setupImage(tab.getO("img"), heading, symbolSize);

        // Setup the text to display under the point image
        String labelText = setupText(label, xPos, yPos, heading, text);

        // display the image or symbol if the style is not set to "hide'
        if (!style.getDisplayMode().equalsIgnoreCase("hide")) display (labelText);
      }

      private void setupStyle (kmlBalloonStyle style, String name, String desc) {
        text = style.getText();
        text = text.replaceAll("\\$\\[name\\]", name);
        text = text.replaceAll("\\$\\[description\\]", desc);
        if (text == null || text.equals("")) text = "None";
      }

      /** Prepare the image to be displayed.
          If the image is already provided (i.e. the point uses an ImageIcon), then
          we simply use that image and rotate and resize it. if there is no existing
          image, we get one from the Symbol class. Note that images obtained from
          Symbol do not need rotating. */
      private void setupImage (Object obj, float heading, int newSize) {
        if (obj != null && obj instanceof BufferedImage) {
          img = (BufferedImage)obj;
          if (heading != 0) {
            double rads     = Math.toRadians(heading);
            img = GraphicsUtil.rotateImage(img, rads);   // Adjust the rotation of the img
          }
        }
        else {
          int h = symbol.getIconHeight();
          int w = symbol.getIconWidth();
          img   = symbol.getImage(w, h);
        }

        resize(newSize);  // resize (usually enlarge) the image
      }

      /** Resize the existing image to the size given.
          This method will look at the Height and width of the image or symbol,
          find the largest value, and scale the image or symbol based on the size given.
          Note that size is still restrained by {@link #MAX_SIZE} & {@link #MIN_SIZE}
          @param newSize The size to make the popup image or symbol
       */
      private void resize (float newSize) {
        if (newSize > MAX_SIZE) newSize = MAX_SIZE; // keep sizes in check
        else if (newSize < MIN_SIZE) newSize = MIN_SIZE;
        float ratio;  // scale difference between current size and new size

        float oldH = img.getHeight();
        float oldW = img.getWidth();

        // use the max dimension to calc scale multiplier
        ratio = (oldH > oldW)? (newSize / oldH): (newSize / oldW);
        int newH = (int) (oldH * ratio); // the height of the new Image (in pixels)
        int newW = (int) (oldW * ratio); // the width  of the new Image (in pixels)

        //Get a scaled version of the image and set the imageIcon to it
        Image i = ((Image)img).getScaledInstance(newW, newH, Image.SCALE_SMOOTH);
        icon = new ImageIcon(i);

        //System.out.println("LayKML: newSize=" +newSize+", oldH= " +oldH+", newH=  " +newH+ ", ratio= "+ratio+", Img h = " +img.getHeight());
      }

      /** Format the text to be displayed below the image */
      private String setupText (String label, double xPos, double yPos, double heading, String text) {

        // Put the common items in the string first
        String displayText = "<html>Name: <font color=red><b>"+label+"</b></font>"+
        "<p>Lat: <font color=red><b>"+xPos+"</b></font>"+
        "<p>Lon: <font color=red><b>"+yPos+"</b></font>"+
        "<p>Heading: "+"<font color=red><b>"+heading+"</b></font>"+
        "<p>Description: ";

        // Add the description or text field of the style
        if (text.startsWith("<html") || text.startsWith("<HTML")) {
          displayText += text.substring(text.indexOf(">")+1);
        } else {
          displayText += "<font color=red><b>"+ text + "</b></font>"+
          "</b></font></html>";
        }

        return  displayText;
      }

      /** Display the ImageIcon or Symbol.
          When using <code>ImageIcons</code>, we can simply make a single <code>JLabel</code>
          that contains the </code>ImageIcon</code> and the text to display beneath it.
          When just using the <code>Symbol</code>, we need to draw the <code>Symbol</code>
          first and then add the text beneath it on a seperate label
       */
      private void display (String labelText) {
        JLabel textDisplay = new JLabel(labelText, icon, JLabel.CENTER);
        textDisplay.setHorizontalTextPosition(JLabel.CENTER);
        textDisplay.setVerticalTextPosition(JLabel.BOTTOM);
        scrollPane = new JScrollPane(textDisplay);
        add(scrollPane);

        setVisible(true);
      }
    } // end class PopupPanel

  } // end class Popup

} // end class LayerKML
