package nxm.sys.libg;

import nxm.sys.lib.*;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Point;
import java.awt.image.BufferedImage;
import java.io.*;
import java.net.*;
import javax.swing.*;

import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

/** A class that plots images loaded from a Web Map Server (WMS) compliant with the OpenGIS WMS
    protocol specifications.

    @author  J. Czechowski IV
    @version $Id: LayerWMS.java,v 1.33 2013/01/03 16:01:50 ntn Exp $
    @since   NeXtMidas 2.3.1
*/
public class LayerWMS extends LayerMap {
  // Good test cases for version 1.1.1:
  //    http://iq.usgs.gov/cgi-bin/nuiwms.exe
  //    http://webhosts.cr.usgs.gov/wmsconnector/com.esri.wms.Esrimap/nm_overview (Bad SRS)
  //    http://www.geographynetwork.ca/wmsconnector/com.esri.wms.Esrimap/OBM_Full_I
  //    http://wms.co.fulton.ga.us/wms/public   (Has Styles)
  //
  // Good test cases for version 1.3.0:
  //    http://alice.techma.com:8888/wms/mapServer
  //


  // KEY NAMES /////////////////////////////////////////////////////////////////////////////////////
  //   - These names used in local layers tables
  //   - Package private to allow config window to access
  //   - May change without notice!!!
  static final String CRS_KEY     = "LOCAL_CRS";
  static final String STYLE_KEY   = "LOCAL_STYLE";
  static final String STYLES_KEY  = "LOCAL_STYLES";
  static final String LAYERS_KEY  = "LOCAL_LAYERS";
  static final String ENABLED_KEY = "LOCAL_ENABLED";

  // INTERNAL DEFINITIONS //////////////////////////////////////////////////////////////////////////
  private static final String  DEF_BGCOLOR = "0x000066"; // Default background color.
  private static final String  WMS_111     = "1.1.1";    // Identifier for protocol version 1.1.1
  private static final String  WMS_130     = "1.3.0";    // Identifier for protocol version 1.3.0
  private static final String  EX_XML      = "XML";      // Exception format: XML (required)
  private static final String  EX_INIMAGE  = "INIMAGE";  // Exception format: INIMAGE (optional)
  private static final String  EX_BLANK    = "BLANK";    // Exception format: BLANK (optional)
  private static final String  EX_MIMEPRE  = "application/vnd.ogc.se_";      // MIME type prefix used in 1.1.1
  private static final String  XLINK_XLINK = "http://www.w3.org/1999/xlink"; // Required xlink:xlink in 1.1.1 and 1.3.0
  private static final String  XLINK_TYPE  = "simple";   // Required xlink:type in 1.1.1 and 1.3.0
  private static final int     MAX_TRYS    = 1;          // Max retries

  private static final String[] INHERIT_REP_KEYS         // Inherited keys, used in getWmsLayer(..)
          = { "EX_GEOGRAPHICBOUNDINGBOX", "BOUNDINGBOX", "DIMENSION", "ATTRIBUTION",
              "MINSCALEDENOMINATOR", "MAXSCALEDENOMINATOR", "QUERYABLE", "CASCADED",
              "OPAQUE", "NOSUBSETS", "FIXEDWIDTH", "FIXEDHEIGHT" };
  private static final String[] INHERIT_ADD_KEYS         // Inherited keys, used in getWmsLayer(..)
          = { STYLES_KEY /*, "AUTHORITYURL"*/ };

  // EXTERNAL DEFINITIONS //////////////////////////////////////////////////////////////////////////

  /** The list of OpenGIS WMS protocol versions supported in increasing order
      (<code>value={@value}</code>). The OpenGIS WMS specs assume that the newest version
      supported by both the client and server is the best choice.
   */
  public static final String WMS_VERSIONS = WMS_111+","+WMS_130;

  /** The list of image types supported by this client in order of preference, with the most
      preferable one last (<code>value={@value}</code>).
   */
  public static final String IMAGE_TYPES = "image/jpeg,image/gif,image/png";

  /** The list of exception types supported by this client in order of preference, with the most
      preferable one last (<code>value={@value}</code>).
   */
  public static final String EXCEPTION_TYPES = EX_BLANK+","+EX_XML+","+EX_INIMAGE;


  // VARIABLES /////////////////////////////////////////////////////////////////////////////////////
  //   Variable naming scheme:
  //     supXxxs     - List of options supported by client.
  //     wmsXxx      - Option selected to match server, or option that matches server capabilities
  //     [other]     - Other variables
  //
  protected String  supVersions        = WMS_VERSIONS;      // WMS protocol versions supported
  protected String  supImageTypes      = IMAGE_TYPES;       // Image types supported
  protected String  supExceptionTypes  = EXCEPTION_TYPES;   // Exception types supported

  protected String  wmsVersion         = null;              // WMS protocol version (null=NotConnected)
  protected String  wmsFormat          = null;              // Image format (MIME type) to use
  protected String  wmsExceptions      = null;              // Exception format to use
  protected String  wmsGetMapUrl       = null;              // URL for GetMap requests (ends in '?')
  protected Color   wmsBgColor         = null;              // Background color
  protected boolean wmsTransparent     = false;             // Should layers be transparent?
  protected int     wmsMaxWidth        = Integer.MAX_VALUE; // Maximum width in pixels
  protected int     wmsMaxHeight       = Integer.MAX_VALUE; // Maximum width in pixels

  protected Table   capabilities       = null;              // The server's capabilities
  protected Table   allLayersTable     = null;              // The layers available from the server
  protected Vector<Table> allLayersList= null;              // A "flat" view of allLayersTable (in poriority order)
  protected Vector<Table> allLocalCrs  = null;              // All locally-valid CRS options
  protected int     hiResMultiplier    = 2;                 // Multiplier to use in high-res mode
  protected String  status             = null;              // Status message
  protected LayerWmsConfig configPanel = null;              // Configuration Panel
  protected JFrame  configFrame        = null;              // Frame holding configPanel
  protected String  pendingLayers      = null;              // Layers to use (waiting to be set)

  private GraphicsDevice gdev   = null;                     // Where to put the config menu (Bug 2580) (Since 3.1.1)
  private boolean        popped = false;                    // whether the config menu was set for a popped panel

  /** Creates a new instance. */
  public LayerWMS () {
    super("WMS");
    setCoordRefSystems(null);
    setDownloadThreads(4);
    status = "Not connected, URL not set yet.";
  }

  public Table getConfig () {
    Table  tbl = super.getConfig();
    String lay = getLayers();
    String bg  = (wmsBgColor == null)? null: MColor.toString(wmsBgColor);

    if (lay != null   ) { tbl.put("LAYERS",      lay           ); }
    if (bg  != null   ) { tbl.put("BGCOLOR",     bg            ); }
    if (wmsTransparent) { tbl.put("TRANSPARENT", wmsTransparent); }

    return tbl;
  }

  /** Sets the URL of the applicable map server. */
  public void setURL (String url) {
    super.setURL(url);
    if (mapServer == null) {
      status = "Not Connected";
    } else {
      status = "Connecting to "+getHostPort();
      wmsVersion = null;
      loadCapabilities();
      checkInit();
      setNeedsRescale(true);
    }
  }

  /** Status string used in config window. */
  String getStatus () {
    return status;
  }

  /** Gets the high-resolution multiplication factor.
      You should NEVER need to use this method!
   */
  public int getHiResMultiplier () {
    return hiResMultiplier;
  }

  /** Sets the high-resolution multiplication factor.
      You should NEVER need to use this method, it always issues a warning.
   */
  public void setHiResMultiplier (int mult) {
    warn("HiResMultiplier has been changed from "+hiResMultiplier+" to "+mult+". "
        +"You should NEVER need to use this method!");
    hiResMultiplier = mult;
  }

  /** Override some of the internal parameters of this class. Overridable parameters include:
      <pre>
        WMS_VERSIONS    - Override {@link #WMS_VERSIONS}
        IMAGE_TYPES     - Override {@link #IMAGE_TYPES}
        EXCEPTION_TYPES - Override {@link #EXCEPTION_TYPES}
      </pre>
      @deprecated Override some of the internal parameters of this class. This has always been
                  deprecated. It is only here for use in debug the client, and may be
                  removed/changed at any time without notice!
   */
  @Deprecated
  public void setInternalParams (Table tbl) {
    warn("************************ WARNING ************************");
    warn("* LayerWMS: Internal parameters are being overridden.   *");
    warn("*   This may cause LayerWMS to malfunction, use at your *");
    warn("*   own risk!                                           *");
    warn("*********************************************************");
    supVersions       = tbl.getS("WMS_VERSIONS",supVersions);
    supImageTypes     = tbl.getS("IMAGE_TYPES",supImageTypes);
    supExceptionTypes = tbl.getS("EXCEPTION_TYPES",supExceptionTypes);
  }

  /** Sets the list of locally-valid Coordinate Reference Systems (CRS) allowed. See
      <code>nxm/sys/cfg/wms_coord_sys.xml</code> for more info.
      @param table A table of the valid coordinate systems (or null to use the defaults).
   */
  public synchronized void setCoordRefSystems (Object table) {
    if (table == null) {
      table = XmlFile.fastXmlFileToTable(midas, "nxm.sys.cfg.wms_coord_sys.xml");
    }
    Table tbl = Convert.o2t(table, midas);

    if (tbl == null) {
      throw new MidasException("LayerWMS: Unable to load CRS from "+table);
    }

    allLocalCrs  = new Vector<Table>();
    String[] keys = tbl.getKeys();
    for (int i = 0; i < keys.length; i++) {
      allLocalCrs.add(tbl.getTable(keys[i]));
    }
  }

  /** Gets a COPY of the list of all locally-valid Coordinate Reference Systems (CRS) allowed. */
  public synchronized Table getCoordRefSystems () {
    Table tbl = new Table();
    for (int i = 0; i < allLocalCrs.size(); i++) {
      Table crs = allLocalCrs.get(i);
      tbl.put("CRS_"+i, crs.copy());
    }
    return tbl;
  }

  /** Indicates if the background should be transparent.
      @since NeXtMidas 2.4.2
   */
  public void setTransparent (boolean transparent) {
    if (wmsTransparent != transparent) {
      wmsTransparent = transparent;
      refresh();
    }
  }

  /** Indicates if the background should be transparent.
      @since NeXtMidas 2.4.2
   */
  public boolean getTransparent () {
    return wmsTransparent;
  }

  /** Sets the background color for the plot (usually used for the oceans, null=UseDefault). */
  public void setBgColor (Object color) {
    Color bgcolor = MColor.getColor(color);
    if (bgcolor != wmsBgColor) {
      wmsBgColor = bgcolor;
      refresh();
    }
  }

  /** Gets the background color for the plot (usually used for the oceans, null=UseDefault). */
  public String getBgColor () {
    return (wmsBgColor == null)? DEF_BGCOLOR: MColor.toString(wmsBgColor);
  }

  /** Gets the background color as a 6-digit hex value (i.e. 0xRRGGBB). This method follows the
      OpenGIS specs, note that the 'x' must be lower-case.
   */
  private String getBgColorHex () {
    String bgcolor = DEF_BGCOLOR;
    if (wmsBgColor != null) {
      int rgb = 0x00FFFFFF & wmsBgColor.getRGB();
      bgcolor = "0x"+StringUtil.padLeft(Integer.toHexString(rgb), 6, '0');
    }
    return bgcolor;
  }

  public boolean isConnected () {
    return (wmsVersion != null);
  }

  /** Get the capabilities for the server. */
  public Table getCapabilities () {
    return capabilities.copy();
  }

  /** Sets the range appropriate for a map of the world. */
  public void findRange () {
    super.findRange(); // Start with default values.

    if (allLayersList != null) {
      boolean rangeSet = false;
      for (Table info : allLayersList) {
        Table boundingBox = info.getTable("EX_GEOGRAPHICBOUNDINGBOX");

        if (boundingBox != null) {
          if (!rangeSet) {
            rangeSet = true;
            x1 = boundingBox.getD("WESTBOUNDLONGITUDE",x1);
            x2 = boundingBox.getD("EASTBOUNDLONGITUDE",x2);
            y1 = boundingBox.getD("SOUTHBOUNDLATITUDE",y1);
            y2 = boundingBox.getD("NORTHBOUNDLATITUDE",y2);
          } else {
            x1 = Math.min(x1, boundingBox.getD("WESTBOUNDLONGITUDE",x1));
            x2 = Math.max(x2, boundingBox.getD("EASTBOUNDLONGITUDE",x2));
            y1 = Math.min(y1, boundingBox.getD("SOUTHBOUNDLATITUDE",y1));
            y2 = Math.max(y2, boundingBox.getD("NORTHBOUNDLATITUDE",y2));
          }
        }
      } // end for loop
    }
  }

  /** Get the information about all layers available from the server.
      @return The layers in a standard tree structure. This returns null if info has
              not yet been loaded from server.
   */
  Table getAllLayersTable () {
    return allLayersTable;
  }

  /** Get the information about all layers available from the server.
      @return The layers in a "flat" priority order. This returns null if info has
              not yet been loaded from server.
   */
  List<Table> getAllLayersList () {
    return allLayersList;
  }

  /** Get the information about a single layer available from the server. Must provide the layer
      access key.
   */
  Table getAvailableLayer (String key) {
    return (key.equals(""))? getAllLayersTable(): getAllLayersTable().getTable(key);
  }

  /** Get the information about a single layer available from the server. Must provide the layer
      access key.
   */
  synchronized Table findAvailableLayer (String name) {
    Table layer = (name.equals(""))? getAllLayersTable(): null;

    if (layer == null) {
      int index = findLayer(name);
      if (index >= 0) {
        layer = allLayersList.get(index);
      }
    }

    return layer;
  }

  /** Get the information about a single layer available from the server. Must provide the layer
      access key.
   */
  Table findAvailableLayer (int index) {
    Table layer = null;
    if ((allLayersList != null) && (index >= 0) && (index < allLayersList.size())) {
      layer = allLayersList.get(index);
    }
    return layer;
  }


  /** Get the information about a single layer available from the server. Must provide the layer
      access key.
   */
  private int findLayer (String name) {
    int index = -1;

    for (int i = 0; (index < 0) && (i < allLayersList.size()); i++) {
      Table  lay     = allLayersList.get(i);
      String layName = lay.getS("NAME");
      if (name.equals(layName)) {
        index = i;
      }
    }
    return index;
  }

  /** Translates layer name into key name for that layer (note: all layers have a key, but not
      all layers have a name).
   */
  String findLayerKey (String name) {
    String key = null;

    if (name == null) {
      key = null;
    } else if (name.equals("")) {
      key = "";
    } else {
      key = findLayerKey(name, allLayersTable, "");
    }
    return key;
  }

  /** Translates layer name into key name. */
  private synchronized String findLayerKey (String name, Table layer, String prefix) {
    Table  layers  = (layer == null)? null: layer.getTable(LayerWMS.LAYERS_KEY);
    String key     = null;
    String layName = layer.getS("NAME", null);

    if ((layName != null) && layName.equals(name)) {
      key = prefix;
    } else if (layers != null) {
      String[] keys = layers.getKeys();
      for (int i = 0; (key == null) && (i < keys.length); i++) {
        String tag = LAYERS_KEY + "." + keys[i];
        String dot = (prefix.length() == 0)? "": ".";

        key = findLayerKey(name, layers.getTable(keys[i]), (prefix+dot+tag));
      }
    }
    return key;
  }

  /** Shifts the order of one of the layers.
      @param name  The layer to move.
      @param index The new index of the layer.
      @return true if moved, false otherwise.
   */
  synchronized boolean moveAvailableLayer (String name, int index) {
    int     from = findLayer(name);
    int     to   = (index < from)? index: (index-1);
    boolean move = (from >= 0) && (from != index);

    if (move) {
      Table layer = allLayersList.remove(from);
      allLayersList.insertElementAt(layer,to);
      refresh();
    }
    return move;
  }

  /** Set the list of layers enabled and their styles. The layers are specified in order (first is
      bottom, same as <tt>PLOT</tt>) Since NeXtMidas 2.4.2, this method accepts the following
      syntax when specifying a layer:
      <pre>
         [layer]          Adds the named layer with no style setting
        +[layer]          Adds the named layer with no style setting
        -[layer]          Removes the named layer
         [layer]/[style]  Adds the named layer with the given style
        +[layer]/[style]  Adds the named layer with the given style
        -[layer]/[style]  Removes the named layer (style is ignored)
        &nbsp;*                Adds all layers with no style setting
        +*                Adds all layers with no style setting
        -*                Removes all layers
      </pre>
      If the first layer in the list is NOT specified using a "+" or "-" all previous settings
      are discarded. Some examples:
      <pre>
        1. Show ONLY the "shapePolitical" layer:

             setLayers("shapePolitical")

        2. Show the "shapePolitical" layer and the "worldRivers" layer, with
           "worldRivers" on top:

             setLayers("shapePolitical,worldRivers")

        3. Add the "worldRivers" layer on top of any existing layers, but have
           it use the "grayscale" style:

             setLayers("+worldRivers/grayscale")

        4. Include all layers, except the "worldRivers" layer:

             setLayers("*,-worldRivers")
      </pre>
      After a connection has been made to the server, it is possible to get a list of all available
      layers and all available styles for the given layer. From a macro this would look like:
      <pre>
        if reg.myplot.layers.wms.isConnected() then
          invoke allLayers reg.myplot.layers.wms.getAllLayers()

          foreach layer inList allLayers
            invoke allStyles reg.myplot.layers.wms.getAllStyles("^layer")

            info "layer='^{layer}' styles='^{allStyles}'"
          endfor
        else
          warn "Not connected"
        endif
      </pre>
      @param layerList The list of layers (in order). If a layer has a style set it should be
             listed as "[layer]/[style]". NOTE: layer names and styles are case-sensitive (this
             is a "limitation" of the WMS protocol).
      @see #getLayers()
      @see #getAllLayers()
      @see #getAllStyles(String)
   */
  public void setLayers (String layerList) {
    if (layerList == null) { layerList = ""; }

    if (!checkInit()) {
      pendingLayers = layerList;
    } else {
      doSetLayers(layerList);
      refresh();
    }
  }

  /** Set the list of layers enabled.
      @param layerList The comma-separated list of layers (in order).
   */
  private void doSetLayers (String layerList) {
    // Implementation notes:
    //   (1) This can not use the "standard" NeXtMidas mask parser since it assumes order is not
    //       important. However, it is important for map layers since it defines stacking order.
    //   (2) Prior to NeXtMidas 2.7.0, "+<layer>" and "-<layer>" could alter the stacking order,
    //       this was changed in 2.7.0 (when "~<layer>" was added) to prevent situations where
    //       "+MyLayer" followed by "-MyLayer" (or "~MyLayer" followed by "~MyLayer") resulted
    //       in a reordering.
    layerList.replace('|', ','); // <-- for the benefit of users who try a traditional mask string
    Parser parser = new Parser(layerList);

    for (int i = 1; i <= parser.elements(); i++) {
      String  layer   = parser.get(i);
      String  style   = "";
      boolean reorder = true;
      boolean enabled = true;
      boolean toggle  = false;

      if (layer.startsWith("-")) {
        reorder = false;
        enabled = false;
        layer   = layer.substring(1);
      } else if (layer.startsWith("+")) {
        reorder = false;
        enabled = true;
        layer   = layer.substring(1);
      } else if (layer.startsWith("~")) {
        reorder = false;
        toggle  = true;
        layer   = layer.substring(1);
      } else if (i == 1) {
        // Defining all layers from scratch, clear out current settings then enable.
        doSetAllLayers(false,"");
      }

      int index = layer.indexOf('/');
      if (index > 0) {
        style = layer.substring(index+1);
        layer = layer.substring(0,index);
      }

      if (layer.equals("*")) {
        doSetAllLayers(enabled,style);
      } else {
        Table info = findAvailableLayer(layer);
        if (info == null) {
          warn("Unknown layer '"+layer+"' in '"+layerList+"', note that layer names are case-sensitive.");
        }
        else {
          if (toggle) {
            enabled = !info.getState(ENABLED_KEY, false);
          }
          info.put(ENABLED_KEY, enabled);
          info.put(STYLE_KEY,   style);
          if (enabled && reorder) {
            allLayersList.remove(info);
            allLayersList.add(info);
          }
        }
      }
    }
  }

  /** Clears any set layers. */
  private void doSetAllLayers (boolean enabled, String style) {
    for (int i = 0; i < allLayersList.size(); i++) {
      Table info = allLayersList.get(i);
      info.put(ENABLED_KEY, enabled);
      info.put(STYLE_KEY,   style);
    }
  }

  /** Get the list of available styles for a given layer.
      @param layer The name of the layer (case-sensitive). If the layer is listed as "[name]/[style]"
                   the style part is ignored.
      @return The list of styles as a comma-separated list. If no styles are available the list
              will be blank.
      @see #getLayers()
      @since NeXtMidas 2.4.2
   */
  public String getAllStyles (String layer) {
    if (layer.indexOf('/') > 0) {
      layer = layer.substring(0,layer.indexOf('/'));
    }

    String styles = null;
    Table  info   = findAvailableLayer(layer);

    if (info == null) {
      if (isConnected()) {
        warn("Unknown layer='"+layer+"', note that layers are case-sensitive.");
      }
    } else {
      Table        tbl  = info.getTable(STYLES_KEY);
      String[]     keys = tbl.getKeys();
      StringBuffer list = new StringBuffer();

      for (int i = 0; i < keys.length; i++) {
        if (i != 0) list.append(',');
        list.append(tbl.getTable(keys[i]).getS("NAME"));
      }
      styles = list.toString();
    }
    return styles;
  }

  /** Get the information about all layers available from the server.
      @return The list of layers as a comma-separated list. If a layer has a style set it will
              be listed as "[layer]/[style]". This returns null if info has not yet been loaded
              from server.
      @see #getLayers()
   */
  public String getAllLayers () {
    return (checkInit())? getLayers(true): null;
  }

  /** Get the list of layers enabled.
      @return The list of layers as a comma-separated list. If a layer has a style set it will
              be listed as "<layer>/<style>".
   */
  public String getLayers () {
    return getLayers(false);
  }

  /** Get the list of layers enabled. */
  private String getLayers (boolean all) {
    StringBuffer layers = new StringBuffer(); // Lowest priority listed last

    if (!checkInit()) {
      layers.append(pendingLayers);
    } else {
      for (int i = allLayersList.size()-1; i >= 0; i--) {
        Table   layer   = allLayersList.get(i);
        String  name    = layer.getS("NAME");
        String  style   = layer.getS(STYLE_KEY, "");
        boolean enabled = layer.getState(ENABLED_KEY);

        if ((name != null) && (all || enabled)) {
          if (layers.length() > 0) { layers.append(','); }
          layers.append(name);

          if ((style != null) && (style.length() > 0)) {
            layers.append("/").append(style);
          }
        }
      }
    }
    return layers.toString();
  }

  /** Get the list of layers enabled and styles enabled.
      @return An array holding { layers, styles, crs, layerList }.
   */
  private synchronized String[] getLayersAndStyles () {
    StringBuffer layers    = new StringBuffer(); // Lowest priority listed last
    StringBuffer styles    = new StringBuffer(); // Same order as layers
    Vector<Table> validCrs = new Vector<Table>(allLocalCrs);

    if (checkInit()) {
      for (int i = 0; i < allLayersList.size(); i++) {
        Table   layer   = allLayersList.get(i);
        String  name    = layer.getS("NAME");
        String  style   = layer.getS(STYLE_KEY, "");
        String  crs     = layer.getS(CRS_KEY, "");
        boolean enabled = layer.getState(ENABLED_KEY);

        if ((name != null) && enabled) {
          if (style == null) { style = ""; }

          if (layers.length() > 0) {
            layers.append(',');
            styles.append(',');
          }
          layers.append(name);
          styles.append(style);
          updateValidCrs(validCrs, crs);
        }
      }
    } else {
      layers.append(pendingLayers);
    }
    return new String[] { layers.toString(), styles.toString(), getFirstValid(validCrs) };
  }


  /** Used by getLayersAndStyles() */
  private void updateValidCrs (List<Table> allValid, String layValid) {
    for (int i = 0; i < allValid.size(); i++) {
      Table crs = allValid.get(i);
      if (crs != null) {
        String  name  = crs.getS("NAME",  null);
        boolean allow = crs.getState("ALLOW", false);
        int     index = -1;

        if (allow && (name != null)) {
          index = Parser.find(layValid, name, -1) - 1;
        }
        if (index < 0) {
          allValid.set(i,null);
        }
      }
    }
  }

  /** Used by getLayersAndStyles() */
  private String getFirstValid (List<Table> allValid) {
    String name = null;
    for (Table crs : allValid) {
      if (crs != null) {
        name = crs.getS("NAME", null);
        if (name != null) break; // found it so break out of loop
      }
    }
// NTN TODO Cleanup orig pre Java 5 code
//    for (int i = 0; (name == null) && (i < allValid.size()); i++) {
//      Table crs = allValid.get(i);
//      if (crs != null) {
//        name = crs.getS("NAME", null);
//      }
//    }
    return name;
  }

  /** Checks initialization of layers, styles, etc and updates them as necessary.
      @return true if layer info has been loaded from server, false if not.
   */
  private synchronized boolean checkInit () {
    if (pendingLayers != null) {
      doSetLayers(pendingLayers);
      pendingLayers = null;
    }
    return (allLayersList != null);
  }

  /** Checks the height and width; if out of range, re-scale image. */
  private Dimension getImageDimension (int width, int height, boolean hiRes, boolean warn) {
    Dimension dim;

    if ((width > wmsMaxWidth) || (height > wmsMaxHeight)) {
      double ratio  = ((double)height) / ((double)width);
      double limitW = (ratio / wmsMaxHeight);
      double maxW   = Math.min((double)wmsMaxWidth, (wmsMaxHeight / ratio));

      int w = (int)Math.max((double)wmsMaxWidth,  maxW);
      int h = (int)(ratio * w);
      dim = new Dimension(w,h);

      if (warn) {
        warn("Image size of "+width+"x"+height+" pixels exceeds server limit of "
           + wmsMaxWidth+"x"+wmsMaxHeight+" using "+w+"x"+h);
      }
    } else if (hiRes) {
      dim = getImageDimension(width*hiResMultiplier, height*hiResMultiplier, false, false);
    } else {
      dim = new Dimension(width,height);
    }
    return dim;
  }

  protected int drawArea (double minLon, double maxLon, double minLat, double maxLat,
                          int width, int height, boolean hiRes, String pos) {
    if (wmsVersion == null) {
      // No connection to server.
    } else {
      Dimension dim             = getImageDimension(width, height, hiRes, debug);
      String    bounds          = ""+minLon+","+minLat+","+maxLon+","+maxLat;
      String[]  layersAndStyles = getLayersAndStyles();
      boolean   blank           = (layersAndStyles[0].length() == 0);

      if (blank) {
        if (debug) { info("GetMap url = <blank>"); }
      } else {
        String crs         = layersAndStyles[2];
        String url         = null;
        String bgcolor     = getBgColorHex();
        String transparent = (wmsTransparent)? "TRUE" : "FALSE"; // Must be exactly "TRUE" or "FALSE"
        BufferedImage img;

        if (crs == null) {
// JC4: Need a better way of handling this.
warn("Warning no matching CRS for layers "+layersAndStyles[0]);
          crs = "none";
        }

        if (wmsVersion.equals(WMS_130)) {
          url = wmsGetMapUrl                                          // wmsGetMapUrl ends in a '?'
              +  "VERSION="     + wmsVersion
              + "&REQUEST="     + "GetMap"
              + "&CRS="         + crs
              + "&BBOX="        + bounds
              + "&WIDTH="       + dim.width
              + "&HEIGHT="      + dim.height
              + "&LAYERS="      + layersAndStyles[0]
              + "&STYLES="      + layersAndStyles[1]
              + "&FORMAT="      + wmsFormat
              + "&EXCEPTIONS="  + wmsExceptions
              + "&BGCOLOR="     + bgcolor
              + "&TRANSPARENT=" + transparent;
        } else if (wmsVersion.equals(WMS_111)) {
          url = wmsGetMapUrl                                          // wmsGetMapUrl ends in a '?'
              +  "VERSION="     + wmsVersion
              + "&REQUEST="     + "GetMap"
              + "&SRS="         + crs                                 // <-- This differs from 1.3.0
              + "&BBOX="        + bounds
              + "&WIDTH="       + dim.width
              + "&HEIGHT="      + dim.height
              + "&LAYERS="      + layersAndStyles[0]
              + "&STYLES="      + layersAndStyles[1]
              + "&FORMAT="      + wmsFormat
              + "&EXCEPTIONS="  + getExceptionMimeType(wmsExceptions) // <-- This differs from 1.3.0
              + "&BGCOLOR="     + bgcolor
              + "&TRANSPARENT=" + transparent;
        } else {
          throw new MidasException("Unknown WMS protocol version "+wmsVersion);
        }

        img = requestImage(url,MAX_TRYS,false,dim.width,dim.height);
        if (img != null) {
          drawImage(img, minLon, maxLon, minLat, maxLat);
        }
      }
    }
    return 0;
  }

  /** Gets the MIME type for an exception type as used in WMS 1.1.1. */
  private static String getExceptionMimeType (String type) {
    return EX_MIMEPRE + type.toLowerCase();
  }

  /** Gets the name for an exception from its MIME type as used in WMS 1.1.1. */
  private static String getExceptionName (String mimeType) {
    return mimeType.substring(EX_MIMEPRE.length()).toUpperCase();
  }

  /** Gets the name for an exception from its MIME type as used in WMS 1.1.1. */
  private static String[] getExceptionNames (String[] mimeTypes) {
    String[] names = new String[mimeTypes.length];
    for (int i = 0; i < names.length; i++) {
      names[i] = getExceptionName(mimeTypes[i]);
    }
    return names;
  }

  protected HttpURLConnection openConnection (String url) throws IOException {
    HttpURLConnection conn = super.openConnection(url);
    Object content = conn.getHeaderFields().get("Content-Type");
    String type    = (content == null)? null : content.toString();

    if (wmsExceptions.equals(EX_XML) && (type == null)) {
      throw new MapServerException("Content-Type is not set.", false);

    } else if (type.equals("[text/xml]") || type.startsWith("[text/xml;")
                                         || type.equals("[application/vnd.ogc.se_xml]")) {
      try {
        BufferedReader in   = new BufferedReader(new InputStreamReader(conn.getInputStream()));
        StringBuffer   msg  = new StringBuffer();
        String         line = null;
        while ((line = in.readLine()) != null) {
          msg.append('\n').append(line);
        }
        throw new MapServerException(msg.toString(), false);
      } catch (MapServerException e) {
        throw e;
      } catch (Exception e) {
        throw new MapServerException("Unable to load XML exception info.", e, false);
      }
    } else if (type.equals("[application/vnd.ogc.se_inimage]")) {
      throw new MapServerException("Exception described in image.", true);
    } else if (type.equals("[application/vnd.ogc.se_blank]")) {
      throw new MapServerException("Unknown server exception, image is blank.", false);
    }
    return conn;
  }

  /** Loads the capabilities from the server and negotiates a common protocol to use.
      This initializes many fields, including wmsVersion. If an error occurs, the
      wmsVersion will be set to null indicating that an connection can not be established
      or that the options provided by the server do not match those provided by the client.
   */
  private void loadCapabilities () {
//    Midas   midas     = Shell.getMidasContext();
    Parser  versions  = new Parser(supVersions);
    int     index     = versions.elements();
    String  serverVer = null;

    try {
      // CHECK SERVER VERSION
      //   This follows the "Version number negotiation" specs for version 1.3.0 (section 6.2.4)
      while ((index > 0) && (wmsVersion == null)) {
        String localVer = versions.get(index);

        if ((serverVer == null) || (StringUtil.compareVersions(localVer, serverVer) < 0)) {
          String url = mapServer+"?VERSION="+localVer+"&REQUEST=GetCapabilities&SERVICE=WMS";
          String timeoutQual = "{CONNECTTIMEOUT="+connectTimeout+",TIMEOUT="+readTimeout+"}";

          if (debug) { info("GetCapabilities url = "+url+" timeouts="+timeoutQual); }
          capabilities = XmlFile.fastXmlFileToTable(midas, url+timeoutQual);
          serverVer    = capabilities.getS("VERSION");
        }
        if (serverVer == null) {
          index = -1;
        } else {
          int comp = StringUtil.compareVersions(localVer, serverVer);
          if (comp == 0) {
            wmsVersion = localVer;
          } else if (comp < 0) {
            index = -1; // All local versions are too old
          } else {
            index--; // This local version is too old, try a newer one
          }
        }
      }
      if (serverVer == null) {
        throw new MidasException("Server does not support the same OpenGIS WMS protocol version "
                               + "as the client.");
      }

      // CHECK IMAGE FORMATS
      Table getMap = getCapTable("CAPABILITY.REQUEST.GETMAP");
      if (getMap == null) {
        throw new MidasException("Could not find GETMAP section in the server capabilities.");
      }
      wmsFormat = getBest(supImageTypes, getAllS(getMap, "FORMAT"));

      if (wmsFormat == null) {
        throw new MidasException("Server does not support any of the image formats supported "
                               + "by the client.");
      }

      // GET URL FOR HTTP/GET
      Table httpGet = getCapTable("CAPABILITY.REQUEST.GETMAP.DCPTYPE.HTTP.GET.ONLINERESOURCE");
      if (httpGet == null) {
        warn("Could not find online resource for GetMap");
        wmsGetMapUrl = mapServer+"?";
      } else {
        String xlink = httpGet.getS("XLINK:XLINK",XLINK_XLINK); // Fixed value in 1.1.1/1.3.0
        String href  = httpGet.getS("XLINK:HREF", null       ); // Required parameter
        String type  = httpGet.getS("XLINK:TYPE", XLINK_TYPE ); // Fixed value in 1.1.1/1.3.0

        wmsGetMapUrl = checkGetMapHref(href);
        if (!xlink.equals(XLINK_XLINK)) { // This is the only supported value in ver 1.1.1/1.3.0
          warn("Invalid xlink for GetMap, xlink:xlink="+xlink);
        }
        if (!type.equals(XLINK_TYPE)) {   // This is the only supported value in ver 1.1.1/1.3.0
          warn("Invalid type for GetMap, xlink:type="+type);
        }
      }

      // GET ALL LAYERS
      loadWmsLayers(getCapTable("CAPABILITY.LAYER"));

      // Get resolution limits
      Table service = getCapTable("SERVICE");
      if (service != null) {
        wmsMaxWidth  = service.getL("MAXWIDTH",  Integer.MAX_VALUE);
        wmsMaxHeight = service.getL("MAXHEIGHT", Integer.MAX_VALUE);
      }

      // CHECK ERROR TYPES
      Table    exception = getCapTable("CAPABILITY.EXCEPTION");
      String[] formats   = getAllS(exception,"FORMAT");
      boolean isWMS111_server = WMS_111.equals(wmsVersion);
      if (isWMS111_server) {
        formats = getExceptionNames(formats); // convert WMS1.1.1 to WMS1.3.0 exception format
      }
      wmsExceptions = getBest(supExceptionTypes,formats);

      // 2010-10-19 NTN: fall back to WMS 1.1.1 exception format (BUG 2267)
      if ((wmsExceptions == null) && !isWMS111_server) {
        formats = getExceptionNames(formats); // convert WMS1.1.1 to WMS1.3.0 exception format
        wmsExceptions = getBest(supExceptionTypes,formats);
        if (WMS_130.equals(wmsVersion)) {
          warn("Server claims to be WMS 1.3.0 but is using WMS 1.1.1 exception format: "+mapServer);
        }
      }

      if (wmsExceptions == null) {
        throw new MidasException("Server does not support any of the exception formats supported "
                               + "by the client.");
      }

    } catch(Exception e) {
      warn("Unable to connect to server on "+mapServer,e);
      wmsVersion = null; // Indicate that we are unable to connect to this server.
      status     = "Unable to connect to "+getHostPort();
    }

    if (wmsVersion != null) {
      status = "Connected to "+getHostPort()+" (WMS "+wmsVersion+")";
    }
  }

  /** Checks the HREF given for use with a GetMap request. As per the WMS specs (1.1.1 and 1.3.0),
      the href should end with "?" or "?key1=val1&...&keyN=valN&", some omit the last "?" or
      "&" in violation of the WMS spec, so add it with a warning. <br>
      <br>
      If we can't get to the server, try using the same server as GetCapabilities. This does not
      follow the WMS specs (which say that the server will always give a valid href), but use it
      as a fall-back after issuing a warning.
   */
  private String checkGetMapHref (String href) {
    String url = null;

    if (href != null) {
      href = href.replaceAll("&amp;", "&"); // JC4: I've seen this before...

      if (href.endsWith("?") || (href.indexOf("?")>0 && href.endsWith("&"))) {
        url = href;
      } else {
        url = (href.indexOf("?")>0)? (href+"&") : (href+"?");
        warn("GetCapabilities for '"+mapServer+"' specified invalid href for "
           + "GetMap requests (xlink:href='"+href+"'), will try using '"+url+"'.");
      }
    }

    try {
      checkReachable(url);
    } catch (MidasException e) {
      url  = mapServer+"?";
      warn(e.getMessage());
      warn("Unable to use URL specified in GetCapabilities for GetMap requests, will try using '"+url+"'.");
    }

    return url;
  }

  public void configure () {
    setConfigPanelVisible(true);
  }

  public void close () {
    super.close();
    setConfigPanelVisible(false);
  }

  /** Sets the configuration panel to visible/invisible. */
  public void setConfigPanelVisible (boolean visible) {
    if (isConnected()) {
      if (configFrame!=null && ((MP.isPopped() != popped) ||
          (MP.isPopped()&& gdev != MP.getPopDevice()))) {
        configFrame = null;  // if plot was popped then replaced, need to renew config panel to put it on the right screen (Bug 2580)
      }
      if ((configFrame == null) && visible) {
        checkInit();
        configPanel = new LayerWmsConfig(this);
       // configFrame = new JFrame("LayerWMS Configuration");  // pre 3.1.1
        if (MP.isPopped()) {  // if plot is popped, put the config panel on the right screen (Bug 2580)
          gdev = MP.getPopDevice();
          if (gdev==null) gdev =  GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice();
          configFrame = new JFrame("LayerWMS Configuration", gdev.getDefaultConfiguration());
          popped = true;
        } else {
          configFrame = new JFrame("LayerWMS Configuration");
          popped = false;
        }

        configFrame.getContentPane().add(configPanel);
        configFrame.pack();
      }
      if ((configFrame != null) && (configFrame.isVisible() != visible)) {
        if (visible && (MP != null) && (MP.getPanel() != null)) {
          Point point = MP.getPanel().getLocationOnScreen();
          point.x += 20;
          point.y += 20;
          configFrame.setLocation(point);
        }
        configFrame.setVisible(visible);
      }
    } else {
      if (configFrame != null) {
        configFrame.setVisible(false);
      }
      if (visible) {
        String  message = "LayerWMS: "+getStatus();
        MJPopup mjpopup = MJPopup.createOk(MP.getPanel(), "LayerWMS Configuration",
                                           JOptionPane.ERROR_MESSAGE, message);
        mjpopup.display(0);
      }
    }
  }

  public void refresh () {
    super.refresh();
    if ((configFrame != null) && configFrame.isVisible()) {
      configPanel.refreshConfig();
    }
  }

  /** Reads the WMS layers out of a table. This will take care of any inheritance issues and
      convert the flat layer structure...
      <pre>
        LAYER_1={NAME=<name1>,...}
        LAYER_2={NAME=<name2>,...}
        STYLE_1={NAME=<name1>,...}
        STYLE_2={NAME=<name2>,...}
      </pre>
      ...into a nested one.
      <pre>
        LAYERS={
          <NAME1>={NAME=<name1>,...}
          <NAME2>={NAME=<name1>,...}
        }
        STYLES={
          <NAME1>={NAME=<name1>,...}
          <NAME2>={NAME=<name1>,...}
        }
      </pre>
   */
  private void loadWmsLayers (Table tbl) {
    allLayersList  = new Vector<Table>();
    allLayersTable = getWmsLayer(tbl,null,allLayersList);
  }


  /** Gets WMS Layer info. Only used by loadWmsLayers(..) */
  private Table getWmsLayer (Table tbl, Table inheritFrom, Vector<Table> flatView) {
    final String[] inheritAddKeys = { STYLES_KEY /*, "AUTHORITYURL"*/  };

    if (tbl == null) { return null; }

    Table   layer   = new Table();
    Table   layers  = new Table();
    Table   styles  = new Table();
    String  crs     = "";
    String  tempCrs = "";

    // Insert fields and take care of merging.
    flatView.add(layer);
    layer.put(LAYERS_KEY, layers);
    layer.put(STYLES_KEY, styles);
    layer.put(ENABLED_KEY,false);
    layer.put(STYLE_KEY,  "");
    if (inheritFrom != null) {
      copyTableValues(INHERIT_REP_KEYS, inheritFrom, layer); // Items that replace
      copyTableTables(INHERIT_ADD_KEYS, inheritFrom, layer); // Items that add (by merging)
      crs = inheritFrom.getS(CRS_KEY); // Items that add (by appending, see below)
    }
    layer.merge(tbl);

    // Handle the CRS (ver 1.3.0) vs SRS (ver 1.1.1) issue; plus handle merging them.
    crs = mergeLists(crs, extractStrings("CRS",layer)); // WMS version 1.3.0
    crs = mergeLists(crs, extractStrings("SRS",layer)); // WMS version 1.1.1
    layer.put(CRS_KEY,crs);

    // Convert ver 1.1.1 bounding box to ver 1.3.0
    Table bounds = layer.getTable("LATLONBOUNDINGBOX");
    if (bounds != null) {
      layer.rename("LATLONBOUNDINGBOX", "EX_GEOGRAPHICBOUNDINGBOX");
      bounds.rename("MINY", "SOUTHBOUNDLATITUDE");
      bounds.rename("MAXY", "NORTHBOUNDLATITUDE");
      bounds.rename("MINX", "WESTBOUNDLONGITUDE");
      bounds.rename("MAXX", "EASTBOUNDLONGITUDE");
    }

    // Convert ver 1.1.1 scale hints to ver 1.3.0 scale denominators
    Table scaleHint = layer.getTable("SCALEHINT");
    if (scaleHint != null) {
      layer.remove("SCALEHINT");
      layer.put("MINSCALEDENOMINATOR", 1.0/scaleHint.getD("MIN")); // <-- JC4: Check this computation
      layer.put("MAXSCALEDENOMINATOR", 1.0/scaleHint.getD("MAX")); // <-- JC4: Check this computation
    }

    // Move individual styles to table.
    extractFixAndCopy("STYLE", "NAME", layer, styles);

    // Move individual layers to table. (***RECURSIVE***)
    Object[] allLayers = extractAll(layer,"LAYER");
    for (int i = 0; i < allLayers.length; i++) {
      Table  subLay = getWmsLayer((Table)allLayers[i],layer,flatView);
      String name   = subLay.getS("NAME");
      layers.put(fixName(name), subLay);
    }

    return layer;
  }

  /** Merges two lists. */
  private String mergeLists (String first, String last) {
    String str;
    if (first.length() == 0) {
      str = last;
    } else if (last.length() == 0) {
      str = first;
    } else {
      Parser parser = new Parser(last);
      str = first;
      for (int i = 1; i <= parser.elements(); i++) {
        String elem = parser.get(i);
        if (Parser.find(str,elem,-1) > 0) {
          str = str + "," + elem;
        }
      }
    }
    return str;
  }

  /** Copies nested values from one table into another. Used by getWmsLayer(..). */
  private void copyTableValues (String[] keys, Table src, Table dest) {
    for (int i = 0; i < keys.length; i++) {
      Object val = src.get(keys[i]);
      if (val != null) { dest.put(keys[i], val); }
    }
  }

  /** Merges nested tables from one table into another. Used by getWmsLayer(..). */
  private void copyTableTables (String[] keys, Table src, Table dest) {
    for (int i = 0; i < keys.length; i++) {
      Table srcTable  = src.getTable(keys[i]);
      Table destTable = dest.getTable(keys[i]);
      if (destTable != null) {
        srcTable.merge(destTable);
      }
    }
  }

  /** Merges nested tables from one table into another. Used by getWmsLayer(..). */
  private void copyTableString (String[] keys, Table src, Table dest) {
    for (int i = 0; i < keys.length; i++) {
      String sStr = src.getS(keys[i],"");
      String dStr = dest.getS(keys[i],"");
      src.put(keys[i], (sStr + dStr));
    }
  }

  /** Extract, fix and copy. Used by getWmsLayer(..). */
  private void extractFixAndCopy (String tableKey, String nameKey, Table src, Table dest) {
    Object[] tables = extractAll(src,tableKey);

    for (int i = 0; i < tables.length; i++) {
      Table  tbl  = (Table)tables[i];
      String name = tbl.getS(nameKey);
      dest.put(fixName(name),tbl);
    }
  }

  /** Fixes a name so it can be used as a key in a table. */
  private static String fixName (String name) {
    return (name==null)? null: name.replaceAll("[^A-Za-z0-9_]","_").toUpperCase();
  }

  /** Gets all strings in a table matching a given key. */
  private String[] getAllS (Table tab, String key) {
    ArrayList<String> values = new ArrayList<String>();

    String val = tab.getS(key);
    if (val != null) {
      values.add(val);
    } else {
      for (int i = 1; (i == 1) || (val != null); i++) {
        val = tab.getS(key+"_"+i);
        if (val != null) {
          values.add(val);
        }
      }
    }
    return values.toArray(StringUtil.EMPTY_STRING_ARRAY);
  }

  /** Removes all objects in a table matching a given key. */
  private Object[] extractAll (Table tab, String key) {
    ArrayList<Object> values = new ArrayList<Object>();

    Object val = tab.get(key);
    if (val != null) {
      values.add(val);
      tab.remove(key);
    } else {
      for (int i = 1; (i == 1) || (val != null); i++) {
        String tag = key+"_"+i;
        val = tab.get(tag);
        if (val != null) {
          values.add(val);
          tab.remove(tag);
        }
      }
    }
    return values.toArray();
  }

  /** Removes all objects in a table matching a given key. */
  private String extractStrings (String tableKey, Table src) {
    return StringUtil.join(extractAll(src,tableKey), ",");
  }

  /** For a given list of options, choose the best option available. The list of options should
      be in increasing order of preference ("Low,...,High").
      @param list    The list of supported options. (client side)
      @param options The list of available options. (server side)
      @return The name of the best option or null if none match.
   */
  private String getBest (String list, String[] options) {
    String value = null;
    int    index = -1;

    for (int i = 0; i < options.length; i++) {
      int idx = Parser.find(list, options[i], -1);
      if (idx > index) {
        index = idx;
        value = options[i];
      }
    }
    return value;
  }

  /** Get a named table from the list of capabilities. */
  private Table getCapTable (String key) {
    return (capabilities == null)? null: capabilities.getTable(key);
  }

  /** Get a named string from the list of capabilities. */
  private String getCapS (String key) {
    return (capabilities == null)? null: capabilities.getS(key);
  }

}
