package nxm.sys.libg;

import java.awt.*;
import java.awt.event.*;

import javax.swing.JPanel;

/** A class to handle a text area (SWING) using a NeXtMidas theme.
    This can be used to implement Area, List, or Tree functions.
    Since it is often used as a long menu, it is one-based. Line 0 is the input.<br>
    @since NeXtMidas 2.7.1
    @author  Jeff Schoen / Jim Bauer
    @version $Id: MJText.java,v 1.19 2011/08/10 12:57:18 ntn Exp $
*/
@SuppressWarnings("serial")
public class MJText extends JPanel
  implements MouseListener, MouseMotionListener, MouseWheelListener, KeyListener, FocusListener {

  public static final String flagsList =
    "Multiple,Fifo,Tree,Columns,NoSelect,NoHilite,BackLite,Input,Text,Wrap,Color,Banner,AddPlus,Logger";

  public static final int MULT    =    0x1;
  public static final int FIFO    =    0x2;
  public static final int TREE    =    0x4;
  public static final int COLS    =    0x8;
  public static final int NOSEL   =   0x10;
  public static final int NOHIL   =   0x20;
  public static final int BACKL   =   0x40;
  public static final int INPUT   =   0x80;
  public static final int TEXT    =  0x100;
  public static final int WRAP    =  0x200;
  public static final int COLOR   =  0x400;
  public static final int BANNER  =  0x800;
  public static final int ADDPLUS = 0x1000;  // Show the '+' ad the end of a wrapped line (BUG 1563)
  public static final int LOGGER  = 0x2000;  // indicates the MJText is from MSGLOG (BUG 1563)

  int rows;
  int cols;
  int items;
  int flags;
  Font font;
  Theme theme;
  int tw=10,th=20,ta=10,bh=0;
  int maxlen=0;
  int selecti1=-1, selecti2=-1, selectc1=-1, selectc2=-1;
  int selecti1_, selecti2_, selectc1_, selectc2_;
  int nc,nr,cw,inrow;
  int si=0,sr=0,mr=1,sv=5, sw=sv*3;
  boolean hasScroll,dragScroll,dragSelect;
  ActionListener actionListener;
  boolean refreshI,refreshO;
  int[] row2line;
  String[] line;    // where all the output data is stored
  String banner,prompt,input;  // where all the input data is stored
  int cursor=-1;
  int posi,posc;
  int wraplen;
  MWindow MW;

  /** Constructor for generic text list display
      @param lines The number of text lines to setup for.
      @param rows The number of rows to initially display.
      @param cols The number of columns to initially display.
      @param fontsize The fontsize in pixels to use
      @param theme The color theme to use
      @param flags The options to enable
   */
  public MJText (int lines, int rows, int cols, int fontsize, Theme theme, int flags) {
    setFlags(flags);
    setFontSize(fontsize);
    setTheme(theme);
    setSize (cols*tw,rows*th);
    setLines(lines);
    addMouseListener(this);
    addMouseMotionListener(this);
    addKeyListener(this);
    addFocusListener(this);
    addMouseWheelListener(this);
  }

  public void addActionListener (ActionListener obj) {
    actionListener = obj;
  }

  private void refreshI() {
    refreshI = true;

    // JWB - The line below was changed to eliminate glitches when using double-buffering
    if (isVisible()) repaint();
  }

  private void refreshO() {
    refreshO = true;

    // JWB - The line below was changed to eliminate glitches when using double-buffering
    if (isVisible()) repaint();
  }

  public void setMWindow (MWindow mw) {
    MW = mw;
    theme = MW.theme;
  }

  public void setFontSize (int size) {
    font = new Font ("Monospaced",Font.PLAIN,size);
    FontMetrics fm = getFontMetrics(font);
    tw = fm.stringWidth("0");
    th = fm.getHeight();
    ta = fm.getAscent();
  }

  public void setLines (int lines) {
    String[] oline = line;
    line = new String[lines+1];
    if (oline != null) for (int i=0; i<=items; i++) line[i]=oline[i];
    items = 0;
    maxlen = 0;
  }

  public void setPrompt (String text) {
    prompt = text;
    refreshI();
  }

  public void setInput (String text) {
    input = text;
    cursor = text.length();
    refreshI();
  }

  public String getInput() {
    return input;
  }

  public void setBanner (String text) {
    banner = text;
    refreshO();
  }

  public String getBanner() {
    return banner;
  }

  public void addText (String text) {
    int lf = text.indexOf('\n');
    if (items==0) line[++items]="";
    while (lf>=0) {
      line[items] += text.substring(0,lf);
      add("");
      text = text.substring(lf+1);
      lf = text.indexOf('\n');
    }
    line[items] += text;
    refreshO();
  }

  public void add (String text) {
    boolean fifo = (flags&FIFO) != 0;
    boolean full = items+1 >= line.length;
    boolean scroll = fifo && (si >= items-rows-1);
    if (full && fifo) {        // shift data
      for (int i=0; i<items; i++) line[i] = line[i+1];
      setLine(items,text);
      return;
    }
    if (full) setLines(line.length*2);    // expand array
    int promptSize = (flags&LOGGER)!=0 ? 1: 2;  // BUG 1563: Shellgui needs 2 lines, Msglog needs 1
    if (scroll) si=sr=Math.max(0,items-rows+promptSize);
    setLine(++items,text);
  }

  public String getLine (int i) {
    if (i<0 || i>items) return null;
    return line[i];
  }

  public void replace (String text) {
    setLine(items,text);
  }

  public void setLine (int i, String text) {
    if (i<0 || i>items) return;
    line[i] = text;
    maxlen = Math.max(maxlen,text.length());
    refreshO();
  }

  public void removeAll() {
    for (int i=1; i<=items; i++) line[i]=null;
    items = si = sr = maxlen = 0;
    selecti1 = selecti2 = selectc1 = selectc2 = -1;
    refreshO();
  }

  public void setFlags (int flags) {
    this.flags = flags;
    refreshO();
  }

  public int getFlags() {
    return flags;
  }

  public void setTheme (Theme theme) {
    this.theme = theme;
    repaint();
  }

  public void select (int item) {
    select (item,-1);
  }

  /** select at item and offset */
  public void select (int item, int ich) {
    if ((flags&NOSEL)!=0) return;
    if (item<0 || item>items) return;
    String text = line[item];
    int len = text.length();
    if (item==selecti1 && (ich<0 || ich>=len || (ich>=selectc1 && ich<selectc2))) {
      if (selectc1==0 && selectc2==len) { selecti1=selecti2=-1; }  // deselect
      else { selectc1=0; selectc2=len; }  // select entire line
    }
    else {
      selecti1=selecti2=item;
      selectc1=0; selectc2=len;
      // check for word select
      if ((flags&TEXT) != 0 && ich>=0 && ich<len && text.charAt(ich)>' ') {
        for (selectc1=ich; selectc1>0 && text.charAt(selectc1-1)>' '; selectc1--);
        for (selectc2=ich; selectc2<len && text.charAt(selectc2)>' '; selectc2++);
      }
    }
    if (actionListener != null) {
      ActionEvent e = new ActionEvent(this,0,"SELECT",selecti1);
      actionListener.actionPerformed(e);
    }
    copy(1);
    refreshO();
  }

  /** Copy the selected text to the system clipboard */
  public void copy() { copy(0); }

  /** Copy selected text to either selection buffer (when mode = 1) or the
   *  system clipboard buffer (when mode != 0).
   *  @param mode set to 1 to copy to selection buffer (X-Windows),
   *              otherwise copy to system clipboard.
   */
  private void copy (int mode) {
    if ((flags&TEXT)==0);
    else if (mode==1) GraphicsUtil.toSystemSelection(getSelectedItem());
    else GraphicsUtil.toSystemClipboard(getSelectedItem());
  }

  public void deselect (int item) {
    selecti1 = selecti2 = -1;
    refreshO();
  }

  public String getSelectedItem() {
    String s=null;
    if (selecti1<0);
    else if (selecti1==selecti2) s=line[selecti1].substring(selectc1,selectc2);
    else for (int i=selecti1; i<=selecti2; i++) {
           if (i==selecti1) s = line[i].substring(selectc1) + "\n";
      else if (i==selecti2) s += line[i].substring(0,selectc2);
      else s += line[i] + "\n";
    }
    return s;
  }

  public String[] getSelectedItems() {
    return null;
  }

  public int getRows() { return rows; }

  public int getColumns() { return cols; }

  public void update (Graphics g) {
    paintComponent(g);
  }

  public void paintComponent (Graphics g) {
    if (g != null && (Theme.getOption(Theme.SWINGDB))) super.paintComponent(g); // Swing Double-buffering requires full repaint
    bh = ((flags&BANNER) != 0)? th+6 : 0;
    boolean isInput = (flags&INPUT) != 0;
    boolean inputOnly = isInput && refreshI && !refreshO;
    refreshI = false;
    refreshO = false;
    int w = getWidth();
    int h = getHeight();
    cols = Math.max(0, (w-sw) / tw);
    rows = Math.max(0, (h-bh-5) / th);
    if (isInput && rows>0) rows--;
    nr = Math.max(1,rows);    // number of rows
    nc = Math.max(1,Math.min((items-1)/nr+1,cols/(maxlen+2)));  // number of columns
    if ((flags&COLS)==0) nc=1;
    mr = Math.max(0,(items-1)/nc)+1;
    hasScroll = items > nc*nr;
    if (!hasScroll) si=sr=0;
    if (si>0) { sr=si/nc; si=sr*nc; }
    cw = (w - (hasScroll?sw:0)) / nc;
    int alen = Math.max(2,cw/tw-1);
    wraplen = alen-2;
    if (g==null) return; // just calculating rows for setScroll
    g.setFont(font);
    if (inputOnly) { paintInput(g); if (!Theme.getOption(Theme.SWINGDB)) return; }   // input only
    g.setColor(theme.cwbg);
    g.drawLine(0,1,w-1,1);
    g.drawLine(0,2,w-1,2);
    if (hasScroll) paintScroll(g);     // scroll bar
    if (bh>0) paintBanner(g);       // banner
    // loop through text lines
    if (row2line==null || row2line.length!=nr) row2line = new int[nr];
    int ir=0, wrap=0;
    for (int i=si+1,ic=0; i<=items; i++) {
      int x = ic*cw;
      int y = ir*th + th + bh;
      paintLine (g,i,x,y,alen,wrap);
      if (ic==0) row2line[ir] = i;
      if (++ic >= nc) { ic=0; ir++; }
      if (ir == nr) break;
      if ((flags&WRAP)!=0 && line[i]!=null && wrap+alen<line[i].length()) { wrap += wraplen; i--;}
      else wrap=0;
    }
    int y = ir*th +th + bh - ta;
    g.setColor(theme.cwbg);      // fill space
    g.fillRect(0,y,w-1-(hasScroll?sw:0),h-y);
    inrow = ir;
    if (isInput) paintInput(g);
    // frame border
    g.setColor(theme.cwms);
    g.drawLine(1,0,w-3,0);
    g.drawLine(1,h-2,w-3,h-2);
    // column dividers
    for (int i=1; i<nc; i++) g.drawLine(i*cw,bh+tw,i*cw,h-tw-1);
    // place holders
    for (;ir<row2line.length;ir++) row2line[ir]=-1;
  }

  /** Paint a line of text
      @param g The Graphics context
      @param i The line number to draw
      @param x The X Position
      @param y The Y Position
      @param alen The width of the display window
      @param off Offset. 0 unless this is a wrapped line.
   */
  private void paintLine (Graphics g, int i, int x, int y, int alen, int off) {
    String text = line[i];   // get the text string for that line
    g.setColor(theme.cwbg);
    g.fillRect(x,y-ta,cw-1,th);
    if (text==null) return;
    x += tw;

    // unless disabled, color the text of certain lines
    Color c = theme.cwfg;
    if ((flags&COLOR)==0);
    else if (text.startsWith("INFO:"))  c = Color.green;
    else if (text.startsWith("WARN:"))  c = Color.yellow;
    else if (text.startsWith("ERROR:")) c = Color.red;

    if (off>0) text = text.substring(off);    // if this is the continuation of a previous line
    int tl = text.length();
    if (tl>alen) {                            // if length of line is longer than width of screen
      tl=alen-2;                              // go back 2 chars and then get the substring up to that point
      text = text.substring(0,tl);
      if ((flags&ADDPLUS)!=0) text += " +";   // add a plus sign unless it is turned off
    }

    // selected / highlighted text
    if (i>=selecti1 && i<=selecti2) {
      int c1 = Math.max(0,Math.min(tl,(i==selecti1)? selectc1-off : 0));
      int c2 = Math.max(0,Math.min(tl,(i==selecti2)? selectc2-off : tl));
      if ((flags&BACKL)!=0) {   // backlit selected text
        g.setColor(theme.cwfh);
        g.fillRect(x+c1*tw,y-ta,(c2-c1)*tw,th);
        g.setColor(c);
        g.drawString(text,x,y);
      } else {
        g.setColor(c);
        g.drawString(text,x,y);
        g.setColor(theme.cwfh);
        g.drawString(text.substring(c1,c2),x+c1*tw,y);
      }
    }
    else {
      if (i==0) { // input prompt
        int pl = Math.min(Math.min(alen,prompt.length()), text.length());
        g.setColor(theme.cwfh);
 //System.out.println("MJtext calling drawString with alen="+alen+" prompt length="+prompt.length()+" pl="+pl);
        g.drawString(text.substring(0,pl),x,y);
        text = text.substring(pl); x += tw*pl;
      }
      g.setColor(c);
      g.drawString(text,x,y);  // draw the text
    }
  }

  private void paintBanner (Graphics g) {
    int x = tw;
    int y = th;
    int w = getWidth() - (hasScroll?sw-2:2);
    int h = bh;
    if (banner==null) banner = "Null Banner";
    g.setColor(theme.cwms);
    g.fillRect(2,0,w-2,h);
    g.setColor(theme.cwfh);
    g.drawString(banner,x,y);
  }

  private void paintInput (Graphics g) {
    line[0] = prompt;
    if (input != null) line[0] += input;
    if (line[0]==null) return;
    int lp = prompt.length();
    int x = 0;
    int y = inrow*th + th + bh;
    int w = getWidth() - x - tw - (hasScroll?sw:4);
    int alen = (w-2)/tw - 1;
    paintLine (g,0,x,y,alen,0);
    if (!hasFocus()) return;
    x += (cursor+lp+1)*tw;
    g.setColor(theme.cwfh);
    if (x<w) g.drawLine(x,y-2*th/3,x,y);
    g.setColor(theme.cwfg);
  }

  private void paintScroll (Graphics g) {
    int x=getWidth()-sw/2-2, y=0, w=sw, h=getHeight()-2;
    double f=(double)sr/(mr-nr);
    int flags = MWindow.SLIDER_INSET | MWindow.SLIDER_NOTRACK |
                (dragScroll? MWindow.SLIDER_ACTIVE : 0);
    MWindow.drawVSlider(x,y,w,h, f,flags, g,theme);
  }

  public void setScroll (int index) {
    paintComponent((Graphics)null); // just calculate rows
    if (index<0) index = Math.max(0,items-rows);
    else if (index>=items) index = items-1;
    si = sr = index;
    refreshO();
  }

  public void keyPressed (KeyEvent e) {
    if ((flags&INPUT)==0) return;
    int keycode = e.getKeyCode();
    char keychar = e.getKeyChar();
    boolean cc = e.isControlDown();
//  boolean ak = e.isAltDown(); // unused variable
    String action = null;
    int modifier = 0;

    if (keycode == KeyEvent.VK_ENTER) {
      action = "ENTER";
      si=sr=Math.max(0,items-rows+1);
    }
    else if (keycode == KeyEvent.VK_BACK_SPACE && cursor>0) {
      input = input.substring(0,cursor-1) + input.substring(cursor);
      cursor--;
      refreshI();
    }
    else if (keycode == KeyEvent.VK_DELETE && selecti1==0) {
      int i1 = Math.max(0,selectc1-prompt.length());
      int i2 = Math.max(0,selectc2-prompt.length());
      input = input.substring(0,i1) + input.substring(i2);
      selecti1 = selecti2 = -1;
      cursor = i1;
      refreshI();
    }
    else if (keycode == KeyEvent.VK_DELETE && cursor<input.length()) {
      input = input.substring(0,cursor) + input.substring(cursor+1);
      refreshI();
    }
    else if (keycode == KeyEvent.VK_RIGHT || (cc && keycode == KeyEvent.VK_F)) {  // right arrow
      cursor = Math.min(input.length(),cursor+1);
      refreshI();
    }
    else if (keycode == KeyEvent.VK_LEFT) {  // left arrow
      cursor = Math.max(0,cursor-1);
      refreshI();
    }
    else if (keycode == KeyEvent.VK_UP) {  // up arrow
      modifier = e.isShiftDown()? -2 : -1;
      action = "POPCODE";
    }
    else if (keycode == KeyEvent.VK_DOWN) {  // down arrow
      modifier = e.isShiftDown()? 2 : 1;
      action = "POPCODE";
    }
    else if ((keycode == KeyEvent.VK_END) || (cc && keycode == KeyEvent.VK_E)) {  // end of line
      cursor = input.length();
      refreshI();
    }
    else if ((keycode == KeyEvent.VK_HOME) || (cc && (keycode == KeyEvent.VK_A || keycode == KeyEvent.VK_H))) {  // beginning of line
      cursor = 0;
      refreshI();
    }
    else if (cc && keycode == KeyEvent.VK_U) {  // delete line
      setInput("");
    }
    else if (cc && keycode == KeyEvent.VK_D) {  // interrupt
      action = "INTERRUPT";
    }
    else if (cc && keycode == KeyEvent.VK_P) {  // pop
      if (MW == null) action = "POP"; // JWB - 27 Aug 09: This is usually caught by MWindow.keyPressed
    }
    else if (cc && keycode == KeyEvent.VK_C) {  // copy
      String selectedText = getSelectedItem();
      if (selectedText==null || selectedText.length() <= 0 ) {
        // print info message to user to use CTRL-D instead of CTRL-C for BUG 1109
        addText("INFO: Use Ctrl-D=Interrupt, Ctrl-C=CopySelectedText\n"); // BUG 2218: always display message in this MJText/SHELLGUI terminal
      }
      else copy();
    }
    else if (cc && keycode == KeyEvent.VK_V) {  // paste
      paste();
    }
    else if (keychar>=' ' && keychar<='~' && !cc) {
      input = input.substring(0,cursor) + keychar + input.substring(cursor);
      cursor++;
      if (cursor==1) action = "KEYPRESS";
      refreshI();
    }
    if (actionListener != null && action != null) {
      ActionEvent ae = new ActionEvent(this,0,action,modifier);
      actionListener.actionPerformed(ae);
    }
  }

  public void addInput (String text) {
    input = input.substring(0,cursor) + text + input.substring(cursor);
    cursor += text.length();
    refreshI();
  }

  public void keyReleased (KeyEvent e) {
    // Do nothing
  }

  public void keyTyped (KeyEvent e) {
    // Do nothing
  }

  public void mouseEntered (MouseEvent e) {
  }

  public void mouseExited (MouseEvent e) {
  }

  public void focusGained (FocusEvent e) {
    refreshI();
  }
  public void focusLost (FocusEvent e) {
    refreshI();
  }

  private boolean updateScroll (int y) {
    dragScroll = true;
    int h = getHeight()-sw-6;
    if (h<th) return false;
    int r = (y-sw/2-3)*mr/h;
    int i = r*nc;
    if (i<0 || i>=items) return false;
    si = i;
    sr = r;
    return true;
  }

  private boolean hit (int x, int y) {
    int ic = Math.max(0,Math.min(nc-1, (x-tw)/cw ));
    int ir = (y-bh)/th;
    int c  = (x-tw-ic*cw)/tw;
    int i = -1;
    posi = posc = i;
    if ((flags&INPUT)!=0 && ir==inrow && c<line[0].length()) i=0;
    else if (ir>=0 && ir<nr) i = row2line[ir] + ic;
    if (i<0 || i>items) return false;
    for (;ir>0 && row2line[ir-1]==row2line[ir]; ir--) c+=wraplen;
    posi = i;
    posc = Math.max(0,Math.min(line[i].length(), c));
    return true;
  }

  private void startSelect (int x, int y) {
    if (hit(x,y)) select(posi,posc);
  }

  private boolean updateSelect (int x, int y) {
    boolean paint = false;
    hit(x,y);
    if (posi<0 || posi>items || line[posi]==null) return paint;
    if (dragSelect) {
      paint = (posi!=selecti2_ || posc!=selectc2_);
      selecti2_ = posi;
      selectc2_ = posc;
    } else {
      selecti1_ = selecti2_ = posi;
      selectc1_ = selectc2_ = posc;
      dragSelect = true;
    }
    if (selecti2_<selecti1_ || (selecti2_==selecti1_ && selectc2_<=selectc1_)) {
      selecti1 = selecti2_; selectc1 = selectc2_;
      selecti2 = selecti1_; selectc2 = selectc1_;
    } else {
      selecti1 = selecti1_; selectc1 = selectc1_;
      selecti2 = selecti2_; selectc2 = selectc2_;
    }
    if (paint) copy(1);
    return paint;
  }

  public void mousePressed (MouseEvent e) {
    int x = e.getPoint().x;
    int y = e.getPoint().y;
    int w = getWidth();
    if (event2button(e)!=1) return;
    if (hasScroll && x>w-sw) {
      if (updateScroll(y)) repaint();
    }
  }

  public void mouseReleased (MouseEvent e) {
    if (dragScroll) { dragScroll=false; repaint(); }
    dragSelect = false;
  }

  public void mouseClicked (MouseEvent e) {
    int x = e.getPoint().x;
    int y = e.getPoint().y;
    int button = event2button(e);
    if (button==1) {
      if (y<bh) {
        ActionEvent ae = new ActionEvent(this,0,"BANNER",button);
        if (actionListener!=null) actionListener.actionPerformed(ae);
      }
      else if (x<cw*nc) {
        startSelect (x,y);
      }
    }
    else if (button==2 && (flags&INPUT)!=0) {
      paste(1);
    }
    else {
      ActionEvent ae = new ActionEvent(this,0,"BUTTON",button);
      if (actionListener!=null) actionListener.actionPerformed(ae);
    }
  }

  public void paste() { paste(0); }
  private void paste (int mode) {
    if ((flags&INPUT)==0) return;
    String text = (mode==1)? GraphicsUtil.fromSystemSelection()
                           : GraphicsUtil.fromSystemClipboard();
    if (text != null) addInput(text);
  }

  public void mouseMoved (MouseEvent e) {
    if (MW==null) return;
    Rectangle pos = getBounds();
    MW.px = pos.x + e.getPoint().x;
    MW.py = pos.y + e.getPoint().y;
  }

  private void mouseMovedAction() {
  }

  public void mouseDragged (MouseEvent e) {
    if (event2button(e) != 1) return;
    int x = e.getPoint().x;
    int y = e.getPoint().y;
    if (dragScroll) {
      if (updateScroll(y)) repaint();
    }
    else if ((flags&TEXT)!=0) {
      if (updateSelect(x,y)) repaint();
    }
  }

  /** Allow use of Mouse Scroll Wheel to scroll through lists
      @since NeXtMidas 2.7.2
   */
  public void mouseWheelMoved (MouseWheelEvent e) {
    if (e.getWheelRotation() > 0) {
      if (si < (mr-nr)) { si++; repaint(); }  // scroll down
    } else {
      if (si > 0)       { si--; repaint(); } // scroll up
    }
  }

  /** Convert event to button number. */
  private int event2button (AWTEvent e) {
    int button;
    int modifiers = ((InputEvent)e).getModifiers();
         if ((modifiers&MouseEvent.BUTTON1_MASK)!=0 ) button = 1;
    else if ((modifiers&MouseEvent.BUTTON2_MASK)!=0 ) button = 2;
    else if ((modifiers&MouseEvent.BUTTON3_MASK)!=0 ) button = 3;
    else button = 1;
    if ((modifiers&MouseEvent.SHIFT_MASK)!=0 ) button = button+100;
    if ((modifiers&MouseEvent.CTRL_MASK)!=0 ) button = button+200;
    return button;
  }

}
