package nxm.sys.libg;

import nxm.sys.inc.*;
import nxm.sys.libg.*;

/** A class to transform 2D views. <br>
    <br>
    Transformation matrix:
    <pre>
      x' = mxx*x + mxb    x = (x'-mxb)/mxx
    </pre>
    @author Jeff Schoen
    @version $Id: View.java,v 1.14 2012/06/12 17:47:45 jwb Exp $
 */
public class View extends Object implements Constants {

  /** the MPlot associated with this View */
  protected MPlot MP;

  /** Construct a new View
      @param MP the MPlot associated with this View
   */
  public View (MPlot MP) {
    this.MP = MP;
  }

  /** Set the Geo Perspective parameters. The default implementation does nothing since the
      default view is independent of the observer. Subclasses may override this as needed.
      @param arg the array of arguments
      @return true
   */
  public boolean setObserver (double[] arg) {
    return true;
  }

  /** Get the Geo Perspective parameters. The default implementation simply sets the array to
      <tt>{ 0, 0, 0 }</tt> since the default view is independent of the observer.
      Subclasses may override this as needed.
      @param arg the array of arguments to be set
      @return true
   */
  public boolean getObserver (double[] arg) {
    arg[0]=arg[1]=arg[2]=0;
    return true;
  }

  /** Rotate the view. The default implementation does nothing since the default view is
      independent of the observer. Subclasses may override this as needed.
      @param arg the array of arguments
      @return true
   */
  public boolean rotate (double[] arg) {
    return true;
  }

  /** Initialize parameters for this view */
  public void setup () {
  }

  /** Transform from real world coordinates to pixel
      @param rwc an array of real world coordinates. Passing in a null here will do nothing.
      @param pix the MPoint which holds the pixel coordinates
      @return 1
   */
  public int rwc2pix (double[] rwc, MPoint pix) {
    if (rwc==null) return 1;   // Added for Bug 1975
    if (MP.xvy) {
      pix.y = (int)(rwc[0]*MP.mxx + MP.mxb);
      pix.x = (int)(rwc[1]*MP.myy + MP.myb);
    } else {
      pix.x = (int)(rwc[0]*MP.mxx + MP.mxb);
      pix.y = (int)(rwc[1]*MP.myy + MP.myb);
    }
    return 1;
  }

  /** Transform pixel to real world coordinates
      @param pix the pixel coordinates. Passing in a null here will do nothing.
      @param rwc the array which holds the real world coordinates
      @return 0
   */
  public int pix2rwc (MPoint pix, double[] rwc) {
    if (pix==null) return 0;  // Added for Bug 1975
    if (MP.xvy) {
      rwc[0] = (((double)pix.y-MP.mxb)/MP.mxx);
      rwc[1] = (((double)pix.x-MP.myb)/MP.myy);
    } else {
      rwc[0] = (((double)pix.x-MP.mxb)/MP.mxx);
      rwc[1] = (((double)pix.y-MP.myb)/MP.myy);
    }
    rwc[2] = 0;
    return 0;
  }

  /** Convert real world data to pixels, x values are computed on fixed start and delta parameters.
      @param buf    the buffer. Passing in a null here will do nothing.
      @param off    the offset (this should usually be 0)
      @param dim    the dimension (usually 1 for an array of y values)
      @param npts   the number of points
      @param pix    the pixel buffer
      @param xstart the beginning x coordinate
      @param xdelta the change in position for the x coordinate
      @return The number of points to draw. Note that additional information is provided via
              <tt>pix.off</tt> and <tt>pix.n</tt> when {@link Pix#CLIP} is set in <tt>pix.flags</tt>
              (only {@link MPlot} should be using the {@link Pix#CLIP} flag).
   */
  public int rwc2pix (float[] buf, int off, int dim, int npts, Pix pix, double xstart, double xdelta) {
    if (buf==null) return 0;  // Added for BUG 1975
    int    i;
    int    j    = off;
    int    n    = pix.setSize(npts);
    int    xoff = off/dim; 
    int[]  pax  = pix.x;
    int[]  pay  = pix.y;
    double mxx  = MP.mxx;
    double myy  = MP.myy;
    double mxb  = MP.mxb;
    double myb  = MP.myb;
    double x    = xstart+xoff*xdelta;
    if (dim==1) {
      for (i=0; i<n; i++,j++) {
        pax[i] = (int)( x*mxx + mxb); x+=xdelta;
        pay[i] = (int)( (buf[j])*myy + myb);
      }
    } else {
      for (i=0; i<n; i++,j+=dim) {
        pax[i] = (int)( x*mxx + mxb); x+=xdelta;
        pay[i] = (int)( (buf[j])*myy + myb);
      }
    }
    return n;
  }

  /** Convert real world data to pixels. This is identical to the version that uses a double[]
      buffer, both versions are provided for performance reasons.
      @param buf  the buffer. Passing in a null here will do nothing.
      @param off  the offset (this should usually be 0)
      @param dim  the dimension, use 2 or 3 to indicate type of array (2=(x,y), 3=(x,y,z))
      @param npts the number of points
      @param pix  the pixel buffer
      @return The number of points to draw. Note that additional information is provided via
              <tt>pix.off</tt> and <tt>pix.n</tt> when {@link Pix#CLIP} is set in <tt>pix.flags</tt>
              (only {@link MPlot} should be using the {@link Pix#CLIP} flag).
   */
  public int rwc2pix (float[] buf, int off, int dim, int npts, Pix pix) {
    // THIS METHOD IS IDENTICAL TO THE ONE THAT USES A double[] BUFFER FOR INPUT
    if (buf==null) return 0;  // Added for BUG 1975
    int    i;
    int    j   = off;
    int    n   = pix.setSize(npts); 
    int[]  pax = pix.x; 
    int[]  pay = pix.y;
    double mxx = MP.mxx; 
    double myy = MP.myy; 
    double mxb = MP.mxb; 
    double myb = MP.myb;
    if ((pix.flags&Pix.CLIP)!=0) {
     if (MP.view == MP.V_LATLON) {
       return rwc2pixClip(buf,off,dim,npts,pix,false); // Special handling for CLIP
     } else {
      int dx,dy,dxmax=MP.ix21/2, dymax=MP.iy21/2;
      for (i=0; i<n; i++,j+=dim) {
        pax[i] = (int)(buf[j+0]*mxx + mxb);
        pay[i] = (int)(buf[j+1]*myy + myb);
        if (i==0) continue;
        dx=pax[i]-pax[i-1]; dy=pay[i]-pay[i-1]; 
        if (dx<-dxmax||dx>dxmax||dy<-dymax||dy>dymax) {
          pix.off=i; pix.n=i; return i;
        }
      }
     }
    } else if (dim==2) {
      for (i=0; i<n; i++) {
        pax[i] = (int)(buf[j++]*mxx + mxb);
        pay[i] = (int)(buf[j++]*myy + myb);
      }
    } else {
      for (i=0; i<n; i++,j+=dim) {
        pax[i] = (int)(buf[j+0]*mxx + mxb);
        pay[i] = (int)(buf[j+1]*myy + myb);
      }
    }
    return n;
  }

  /** Convert real world data to pixels. This is identical to the version that uses a float[]
      buffer, both versions are provided for performance reasons.
      @param buf  the buffer. Passing in a null here will do nothing.
      @param off  the offset (this should usually be 0)
      @param dim  the dimension, use 2 or 3 to indicate type of array (2=(x,y), 3=(x,y,z))
      @param npts the number of points
      @param pix  the pixel buffer
      @return The number of points to draw. Note that additional information is provided via
              <tt>pix.off</tt> and <tt>pix.n</tt> when {@link Pix#CLIP} is set in <tt>pix.flags</tt>
              (only {@link MPlot} should be using the {@link Pix#CLIP} flag).
   */
  public int rwc2pix (double[] buf, int off, int dim, int npts, Pix pix) {
    // THIS METHOD IS IDENTICAL TO THE ONE THAT USES A float[] BUFFER FOR INPUT
    if (buf==null) return 0;  // Added for BUG 1975
    int    i;
    int    j   = off; 
    int    n   = pix.setSize(npts); 
    int[]  pax = pix.x;
    int[]  pay = pix.y;
    double mxx = MP.mxx; 
    double myy = MP.myy;
    double mxb = MP.mxb;
    double myb = MP.myb;
    if ((pix.flags&Pix.CLIP)!=0) {
     if (MP.view == MP.V_LATLON) {
       return rwc2pixClip(buf,off,dim,npts,pix,false); // Special handling for CLIP
     } else {
      int dx,dy,dxmax=MP.ix21/2, dymax=MP.iy21/2;
      for (i=0; i<n; i++,j+=dim) {
        pax[i] = (int)(buf[j+0]*mxx + mxb);
        pay[i] = (int)(buf[j+1]*myy + myb);
        if (i==0) continue;
        dx=pax[i]-pax[i-1]; dy=pay[i]-pay[i-1]; 
        if (dx<-dxmax||dx>dxmax||dy<-dymax||dy>dymax) {
          pix.off=i; pix.n=i; return i;
        }
      }
     }
    } else if (dim==2) {
      for (i=0; i<n; i++) {
        pax[i] = (int)(buf[j++]*mxx + mxb);
        pay[i] = (int)(buf[j++]*myy + myb);
      }
    } else {
      for (i=0; i<n; i++,j+=dim) {
        pax[i] = (int)(buf[j+0]*mxx + mxb);
        pay[i] = (int)(buf[j+1]*myy + myb);
      }
    }
    return n;
  }

  /** Special version of rwc2pix used with CLIP and LatLon or Mercator view.
      @param buf  the buffer. Passing in a null here will do nothing.
      @param off  the offset (this should usually be 0)
      @param dim  the dimension, use 2 or 3 to indicate type of array (2=(x,y), 3=(x,y,z))
      @param npts the number of points
      @param pix  the pixel buffer
      @param mer  whether or not it is mercator
      @return     The number of points to draw
   */
  int rwc2pixClip (float[] buf, int off, int dim, int npts, Pix pix, boolean mer) {
    // THIS METHOD IS IDENTICAL TO THE ONE THAT USES A double[] BUFFER FOR INPUT
    if (buf==null) return 0;  // Added for BUG 1975
    int i;
    int    j   = off;
    int    n   = pix.setSize(npts); 
    int[]  pax = pix.x;
    int[]  pay = pix.y;
    double mxx = MP.mxx;
    double mxb = MP.mxb;
    double myy = MP.myy;
    double myb = MP.myb;
    if (mer) {
      myy = (double)(MP.iy1-MP.iy2) / (ViewMercader.lat2mer(MP.ry2) - ViewMercader.lat2mer(MP.ry1));
      myb = MP.iy2 - ViewMercader.lat2mer(MP.ry1)*myy;
    }
    double rx = 0;
    double ry = 0;
    for (i=0; i<n; i++,j+=dim) {
      double rxp = rx; // previous rx
      double ryp = ry; // previous ry
      rx = buf[j+0];
      ry = buf[j+1];
      while (rx>+180) rx -= 360;
      while (rx<-180) rx += 360;
      while (ry> +90) ry -= 180;
      while (ry< -90) ry += 180;

      pax[i] = (int)(rx*mxx + mxb);
      pay[i] = (mer)? (int)(ViewMercader.lat2mer(ry)*myy + myb)
                    : (int)(ry*myy + myb);
      if (i==0) continue;
      // Check for wrap at +180/-180 longitude
      if ((rx>+90.0 && rxp<-90.0) || (rx<-90.0 && rxp>+90.0)) {
        // Split into two line segments, to simulate +180/-180 wrap,
        // put the first in the pax/pay buffers and put the second in
        // the pix.sx/pix.sy buffers to be plotted separately, do not
        // update to i++ since next segment will resume at that point.
        double rxn;  // new rx
        double ryn;  // rew ry
        if (rx < rxp) {
          double dx1 = rx - rxp + 360;      // old dx
          double dy1 = ry - ryp;            // old dy
          double dx2 = 180 - rxp;           // new dx
          rxn = -180;                       // new rx (right)
          ryn = (dx2 / (dx1 / dy1)) + ryp;  // new ry
        }
        else {
          double dx1 = rx - rxp - 360;      // old dx
          double dy1 = ry - ryp;            // old dy
          double dx2 = -180 - rxp;          // new dx
          rxn = 180;                        // new rx (left)
          ryn = (dx2 / (dx1 / dy1)) + ryp;  // new ry
        }

        pix.sx[0] = (int)(rxn*mxx + mxb);
        pix.sx[1] = pax[i];
        pix.sy[0] = (mer)? (int)(ViewMercader.lat2mer(ryn)*myy + myb)
                         : (int)(ryn*myy + myb);
        pix.sy[1] = pay[i];
        pax[i]    = (int)(-rxn*mxx + mxb);
        pay[i]    = pix.sy[0];
        pix.save  = true;
        pix.off   = i;
        pix.n     = i;
        return i+1;
      }
    }
    return n;
  }
  
  /** Convert real world map coordinates to pixels. This method takes in two pixel buffers, a primary one
      and another that is used in the event that the coords cross the dateline. This is done in the event that the
      lines are making a polygon that may need to be filled. This differs from {@link #rwc2pixClip(double[], int, int, int, Pix, boolean)}
      in that the MPlot is passed in and this method does not return when crossing the date line.
      The effect of that is that for a crossed date line, the buffer continues and any line being drawn will 
      draw along the edge of the map border instead of breaking.
      @param buf     The real world coords
      @param off     The initial offset into the buffer
      @param dim     2 for x,y or 3 for x,y,x
      @param npts    How many points
      @param MP      The MPlot 
      @param pixMain The pixel buffer for the main set of pixels
      @param pixAlt  The pixel buffer for the set of pixels used when the coords have crossed the date line
      @since NeXtMidas 3.1.3
   */
  int rwc2pixClip (double[] buf, int off, int dim, int npts, MPlot MP, Pix pix) {
    if (buf==null) return 0;  
    double rx=0, rxp;       // current and previous x pos
    double ry=0, ryp;       // current  y pos
    int x, y;               // the pixel coords
    double  mxx = MP.mxx;
    double  mxb = MP.mxb;
    double  myy = MP.myy;
    double  myb = MP.myb;
    boolean mer = !(MP.view==MPlot.V_LATLON);  // checked this way to mimic existing methods
    
    if (mer) {
      myy = (double)(MP.iy1-MP.iy2) / (ViewMercader.lat2mer(MP.ry2) - ViewMercader.lat2mer(MP.ry1));
      myb = MP.iy2 - ViewMercader.lat2mer(MP.ry1)*myy;
    }
    
    int i1 = 0;  // index into main buffer
    int i2 = 0;  // index into alt buffer

    boolean drawMain = true;  // start filling with a single buffer
    
    Pix pixMain = new Pix(npts,pix.flags);
    Pix pixAlt  = new Pix(npts,pix.flags);
    
    for (int i=0; i<npts; i++,off+=dim) {
      rxp = rx; // previous rx
      ryp = ry; // previous ry
      rx = buf[off+0];
      ry = buf[off+1];
      while (rx>+180) rx -= 360;
      while (rx<-180) rx += 360;
      while (ry> +90) ry -= 180;
      while (ry< -90) ry += 180;
      
      x =                                                    (int)(rx*mxx + mxb);
      y = (mer)? (int)(ViewMercader.lat2mer(ry)*myy + myb) : (int)(ry*myy + myb);
      
      if ((rx>+90.0 && rxp<-90.0) || (rx<-90.0 && rxp>+90.0)) {  // crossing the dateline
        
        // the code below adds two points to the arrays. One on each side of the dateline
        int pt = (rx < rxp) ? 180 : -180;
        int newY = getExtrapolatedYPix (rx, rxp, ry, ryp, MP);
        if (!drawMain) pt = -pt;  // reverse the order if not filling the main array
        checkSize(pixMain,i1+2);  // check size since we are adding two points for every border cross
        checkSize(pixAlt,i2+2);
        pixMain.x[i1]   = (int)(pt*mxx + mxb); 
        pixMain.y[i1++] = newY;
        pixAlt.x[i2]    = (int)(-pt*mxx + mxb);
        pixAlt.y[i2++]  = newY;
        
        drawMain = !drawMain;
        pix.save = true;
      } 
      
      if (drawMain) {
        pixMain.x[i1]   = x;
        pixMain.y[i1++] = y;
      } else {
        pixAlt.x[i2]   = x;
        pixAlt.y[i2++] = y;
      }
    }
    // keep track of how many points we put in each array
    pix.n = i2;
    pix.x = pixMain.x;
    pix.y = pixMain.y;
    pix.sx = pixAlt.x;
    pix.sy = pixAlt.y;
    pix.off = npts;
    return i1;
  }
  
  /** Get the location of a Y pixel point between two other points, one on each side of the dateline  */
  private int getExtrapolatedYPix (double rx, double rxp, double ry, double ryp, MPlot MP) {
    // this method added in 3.1.3 - Bug 2595
    double myy   = MP.myy;
    double myb   = MP.myb;
    boolean mer  = (MP.view==MPlot.V_MERCATOR);
    if (mer) {
      myy = (double)(MP.iy1-MP.iy2) / (ViewMercader.lat2mer(MP.ry2) - ViewMercader.lat2mer(MP.ry1));
      myb = MP.iy2 - ViewMercader.lat2mer(MP.ry1)*myy;
    }
    
    int base = (rx < rxp) ? 180 : -180;
    
    double dx1 = rx - rxp + (base*2);      // old dx
    double dy1 = ry - ryp;            // old dy
    double dx2 = base - rxp;           // new dx
    double ryn = (dx2 / (dx1 / dy1)) + ryp;  // new ry
    
    int newY = (mer)? (int)(ViewMercader.lat2mer(ryn)*myy + myb) : (int)(ryn*myy + myb);
    
    return newY;
  }
  
  /** Ensure there is room to add another point. If it is too small, we double in size to avoid future copies */
  private void checkSize (Pix pix, int size) {
    // this method added in 3.1.3 - Bug 2595
    if (pix.x.length <= size || pix.y.length <= size) {
      int[] tmpx = new int[pix.x.length*2];
      int[] tmpy = new int[pix.y.length*2];
      System.arraycopy(pix.x, 0, tmpx, 0, pix.x.length);
      pix.x = tmpx;
      System.arraycopy(pix.y, 0, tmpy, 0, pix.y.length);
      pix.y = tmpy;
    }
  }
  
  /** Special version of rwc2pix used with CLIP and LatLon or Mercator view.
      @param buf  the buffer. Passing in a null here will do nothing.
      @param off  the offset (this should usually be 0)
      @param dim  the dimension, use 2 or 3 to indicate type of array (2=(x,y), 3=(x,y,z))
      @param npts the number of points
      @param pix  the pixel buffer
      @param mer  whether or not it is mercator
      @return     The number of points to draw
   */
  int rwc2pixClip (double[] buf, int off, int dim, int npts, Pix pix, boolean mer) {
    // THIS METHOD IS IDENTICAL TO THE ONE THAT USES A float[] BUFFER FOR INPUT
    if (buf==null) return 0;  // Added for BUG 1975
    int i;
    int j      = off;
    int n      = pix.setSize(npts); 
    int[]  pax = pix.x, pay=pix.y;
    double mxx = MP.mxx;
    double mxb = MP.mxb;
    double myy = MP.myy;
    double myb = MP.myb;
    if (mer) {
      myy = (double)(MP.iy1-MP.iy2) / (ViewMercader.lat2mer(MP.ry2) - ViewMercader.lat2mer(MP.ry1));
      myb = MP.iy2 - ViewMercader.lat2mer(MP.ry1)*myy;
    }
    double rx = 0;
    double ry = 0;
    for (i=0; i<n; i++,j+=dim) {
      double rxp = rx; // previous rx
      double ryp = ry; // previous ry
      rx = buf[j+0];
      ry = buf[j+1];
      while (rx>+180) rx -= 360;
      while (rx<-180) rx += 360;
      while (ry> +90) ry -= 180;
      while (ry< -90) ry += 180;

      pax[i] = (int)(rx*mxx + mxb);
      pay[i] = (mer)? (int)(ViewMercader.lat2mer(ry)*myy + myb)
                    : (int)(ry*myy + myb);
      if (i==0) continue;
      // Check for wrap at +180/-180 longitude
      if ((rx>+90.0 && rxp<-90.0) || (rx<-90.0 && rxp>+90.0)) {
        // Split into two line segments, to simulate +180/-180 wrap,
        // put the first in the pax/pay buffers and put the second in
        // the pix.sx/pix.sy buffers to be plotted separately, do not
        // update to i++ since next segment will resume at that point.
        double rxn;  // new rx
        double ryn;  // rew ry
        if (rx < rxp) {
          double dx1 = rx - rxp + 360;      // old dx
          double dy1 = ry - ryp;            // old dy
          double dx2 = 180 - rxp;           // new dx
          rxn = -180;                       // new rx (right)
          ryn = (dx2 / (dx1 / dy1)) + ryp;  // new ry
        }
        else {
          double dx1 = rx - rxp - 360;      // old dx
          double dy1 = ry - ryp;            // old dy
          double dx2 = -180 - rxp;          // new dx
          rxn = 180;                        // new rx (left)
          ryn = (dx2 / (dx1 / dy1)) + ryp;  // new ry
        }

        pix.sx[0] = (int)(rxn*mxx + mxb);
        pix.sx[1] = pax[i];
        pix.sy[0] = (mer)? (int)(ViewMercader.lat2mer(ryn)*myy + myb)
                         : (int)(ryn*myy + myb);
        pix.sy[1] = pay[i];
        pax[i]    = (int)(-rxn*mxx + mxb);
        pay[i]    = pix.sy[0];
        pix.save  = true;
        pix.off   = i;
        pix.n     = i;
        return i+1;
      }
    }
    return n;
  }
}
