package nxm.sys.net;

import nxm.sys.lib.*;

import java.util.Date;
import java.util.Hashtable;
import java.net.HttpURLConnection;

/** A class to return the contents of a file to a Browser.
    @author Jeff Schoen / Neon Ngo
    @version $Id: HFile.java,v 1.18 2011/02/03 17:58:36 ntn Exp $
*/
public class HFile extends HSource {

  private BaseFile bf;
  private static boolean debug = false; // enable to turn on debug messages

  public HFile (String name, BaseFile bf) {
    this.name = name;
    this.bf = bf;
    if (debug) Shell.writeln("HFile() name="+name+" bf="+bf);
  }

  /** @since NeXtMidas 2.7.1 to turn on debugging */
  public static void setDebug(boolean val) { debug = val;  }
  /** @since NeXtMidas 2.7.1 to see if debugging is enabled */
  public static boolean isDebug()          { return debug; }

  public boolean canHandleRequest (String uri) {
    String rname = getName();
    return uri.equals(rname);
  }

  public void handleRequest (String uri, HPage hp) {
    handle (bf, hp, false);
  }

  public static void handle (BaseFile bf,  HPage hp, boolean subs) {
    if (debug) Shell.writeln("HFile.handle() "+hp.getMethod()+" bf.url="+bf.getURL());
    boolean isHeadRequest = hp.getMethod().equals("HEAD");
    bf.open(BaseFile.INPUT|BaseFile.NOABORT);
    if (!bf.isOpen) {
      hp.setStatus(HttpURLConnection.HTTP_NOT_FOUND);
      hp.open();
      if (!isHeadRequest) { hp.writeNotFound(); }
      hp.close();
      return;
    }
    int bytes=0, bufsize=4096;
    double tbytes = bf.getSize();
    double offset = hp.getHeaderD("Range"); // NOTE: we only support first-byte-pos, last-byte-pos is ignored
    String mime   = BaseFile.getMimeType(null,bf.getName());
    long lastModified = bf.getResource().lastModified();
    if (offset>0) { // got Range (partial content) request
      bf.seek(offset);
      hp.setContentRange((long)offset, (long)(tbytes-1), (long)tbytes); // set Content-Range and status of 206 (Partial Content)
      tbytes -= offset;
    }
    // NOTE: all response header fields (must be set before call to HPage.open()
    if (lastModified>0) hp.setLastModified(new Date(lastModified)); // set last-modified if available
    hp.setContentType(mime);
    hp.setContentLength((long)tbytes);
    hp.open(); // <-- this sends HTTP response headers back to client
    if (isHeadRequest) { } // DO NOT return content back for HEAD request
    else if (mime.equals("application/x-java-jnlp-file") || subs) {
      bytes = (int)tbytes;
      byte[] buf = new byte[bytes];
      bf.read(buf,0,bytes);
      String sbuf = new String(buf);
      if (hp.server!=null) {
        sbuf = sbuf.replaceAll("MYHOST",hp.server.getHost());
        sbuf = sbuf.replaceAll("MYPORT",""+hp.server.getPort());
        sbuf = hp.server.parent.MA.evaluateCarets(sbuf);
      }
      hp.write(sbuf);
    }
    else {
      byte[] buf = new byte[bufsize];
      for (; tbytes>0; tbytes-=bytes) {
        bytes = (int)Math.min(tbytes,(double)bufsize);
        int bytesRead = bf.read(buf,0,bytes);
        if (bytesRead<0) break;
        if (hp.write(buf,0,bytesRead)<0) break;
      }
    }
    hp.close();
    bf.close();
  }

  public static void handlePipe (DataFile bf,  HPage hp) {
    bf.open(BaseFile.INPUT|BaseFile.NOABORT);
    boolean isHeadRequest = hp.getMethod().equals("HEAD");
    if (!bf.isOpen) {
      hp.setStatus(HttpURLConnection.HTTP_NOT_FOUND);
      hp.open();
      if (!isHeadRequest) { hp.writeNotFound(); }
      hp.close();
      return;
    }
    int bytes=0, bufsize=1024;
    hp.setContentType(binaryContentType);
    hp.setContentLength((long)(100*1024*1024));
    hp.open();
    if (!isHeadRequest) {
      bf.setDataStart(512.0);	// fake out HttpResource
      bf.setDataSize(1024.0*1024.0);	// fake out HttpResource
      hp.write(bf.hb,0,512);	// write header
      byte[] buf = new byte[bufsize];
      for (;;) {
        bytes = bufsize;
        int bytesRead = bf.read(buf,0,bytes);
        if (bytesRead<0) break;
        if (hp.write(buf,0,bytesRead)<0) break;
      }
    }
    hp.close();
    bf.close();
  }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Deprecated methods
  //////////////////////////////////////////////////////////////////////////////////////////////////
  private static final String htmlContentType = "text/html";
  private static final String textContentType = "text/plain";
  private static final String binaryContentType = "application/octet-stream";
  /** @deprecated Since NeXtMidas 2.7.0: Use {@link BaseFile#getMimeType()} */
  @Deprecated public static Hashtable<String,String> types;

  /** @deprecated Since NeXtMidas 2.7.0: Use {@link BaseFile#getMimeType()} */
  @Deprecated public static String getContentType (String ext) {
    ext = ext.toLowerCase();
    if (types==null) {
      Hashtable<String,String> typesMap = new Hashtable<String,String>();
      typesMap.put("html",htmlContentType);
      typesMap.put("htm" ,htmlContentType);
      typesMap.put("mm"  ,textContentType);
      typesMap.put("key" ,textContentType);
      typesMap.put("txt" ,textContentType);
      typesMap.put("java",textContentType);
      typesMap.put("c",   textContentType);
      typesMap.put("h",   textContentType);
      typesMap.put("tmp" ,binaryContentType);
      typesMap.put("prm" ,binaryContentType);
      typesMap.put("pkt" ,binaryContentType);
      typesMap.put("det" ,binaryContentType);

      // Extensions associated with Java Web Start
      // Specified under "Servlet Configuration" -> "Configuring file extensions and MIME types"
      // in "Packaging JNLP Applications in a Web Archive / Java TM Web Start 1.4.2"
      // http://java.sun.com/javase/6/docs/technotes/guides/javaws/downloadservletguide.html
      typesMap.put("jnlp",    "application/x-java-jnlp-file");
      typesMap.put("jar",     "application/x-java-archive");
      typesMap.put("jardiff", "application/x-java-archive-diff");
      types = typesMap;
    }
    String type = types.get(ext);
    if (type==null) type = binaryContentType;
    return type;
  }

}
