package nxm.sys.prim;

import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.sys.libm.Waveform;

/**
  Creates signal waveforms of various shapes.

  @author Jeff Schoen
  @version $Id: waveform.java,v 1.47 2012/10/04 12:46:25 jgs Exp $
*/
public final class waveform extends Primitive {

  private int shape, throttle, lrs=1, cur=0;
  private double size, amp, freq, poff, phase, chirp, mean, var, sdev, a2s, dt, tstart;
  private boolean dble, exact;	// double precision and exact phase
  private String form;
  private DataFile hcb;
  private Data data;

  /** The waveform shapes this primitive can generate {@value} */
  public  static final String shapeList =
    "Sine,Cosine,Square,Triangle,SawTooth,Pulse,Constant,Zero,White,LRS,Ramp";
  private static final int SIN=1,COS=2,SQU=3,TRI=4,SAW=5,PUL=6,CON=7,ZER=8,WHI=9,LRS=10,RAM=11;

  /** The throttle mode {@value}.
      FULL - the primitive runs as fast as it can.
      REALTIME - the data is produced at the defined sample rate.
      BLOCK - no data is produced. */
  public  static final String throttleList = "RealTime,Full,Block";
  private static final int REALTIME=1,FULL=2,BLOCK=3;

  public int open() {

    // Get parameters
    shape = MA.getChoice("SHAPE",shapeList,0);
    form  = MA.getFormat("FORM","S#,C#");
    freq  = MA.getD("FREQ");
    amp   = MA.getD("AMP"); 
    poff  = MA.getD("PHASE");
    size  = MA.getD("ELEM");
    chirp = MA.getD("CHIRP",0.0);
    mean  = MA.getD("MEAN",0.0);
    var   = MA.getD("VARIANCE",Math.pow(amp/LNTEN,2));
    sdev  = MA.getD("SDEV",Math.sqrt(var));

    // Open output file
    hcb = MA.getDataFile("OUT","1000","SF",0);
    hcb.setFormat (form);
    hcb.setXStart (MA.getD("START"));
    hcb.setXDelta (MA.getD("DELTA"));
    hcb.setXUnits (Units.TIME);
    hcb.setSize   (size);
    hcb.open(hcb.OUTPUT);

    // Setup throttling mode
    throttle = MA.getState("/RT")? REALTIME : FULL;
    throttle = MA.getChoice("/THROTTLE",throttleList,throttle); // Override any /RT input
    if (throttle==REALTIME) hcb.setTimeAt(Time.current());

    // Setup processing
    dble  = MA.getState("/DP",(hcb.getFormatType()==DOUBLE));
    exact = MA.getState("/EXACT");
    xfer  = MA.getL("/TL",4096);
    data  = hcb.getDataBuffer(xfer,dble?DOUBLE:FLOAT);
    dt    = hcb.getXDelta();
    a2s   = sdev/amp; // set user ratio of sdev/amplitude

    todo(size);

    Waveform.setSeed(MA.getL("/SEED",-1));

    return NORMAL;
  }

  public int process() {

    // handle data throttling
    if (throttle>0) {
      if (throttle==REALTIME) {
        if (tstart==0) tstart = Time.current()-hcb.seek()*dt;
        if (Time.current() < tstart+hcb.seek()*dt) return NOOP;
      }
      else tstart=0;
      if (throttle==BLOCK) return NOOP;
    }

    int ndo = todo();
    if (ndo==0) return FINISH;

    double p  = phase+poff;             // offset phase
    double dp = freq*dt;                // delta phase

    if (poff!=0) p -= Math.floor(p);    // modulo 1.0 if necessary
    if (dp<0 && shape!=SIN) dp = -dp;   // no negative freqs except for SINs

    int mode = (exact && (shape==SIN || shape==COS))? -hcb.spa : hcb.spa;

    if (dble) {
      double[] dbuf = data.castD(false);  // get array
      switch (shape) {                    // branch on wave shape
        case SIN: Waveform.sincos(dbuf,amp,p,dp,ndo,mode); break;
        case COS: Waveform.sincos(dbuf,amp,p+.25,dp,ndo,mode); break;
        case SQU: Waveform.square(dbuf,amp,p,dp,ndo,mode); break;
        case TRI: Waveform.triangle(dbuf,amp,p,dp,ndo,mode); break;
        case SAW: Waveform.sawtooth(dbuf,amp,p,dp,ndo,mode); break;
        case PUL: Waveform.pulse(dbuf,amp,p,dp,ndo,mode); break;
        case CON: Waveform.constant(dbuf,amp,ndo,mode); break;
        case ZER: Waveform.constant(dbuf,0.0,ndo,mode); break;
        case WHI: Waveform.whitenoise(dbuf,sdev,ndo,mode); break;
        case LRS: lrs=Waveform.lrs(dbuf,amp,ndo,mode,lrs); break;
        case RAM: cur=Waveform.ramp(dbuf,amp,ndo,mode,cur); break;
      }
      if (mean!=0) for (int i=0; i<ndo*hcb.spa; i++) dbuf[i]+=mean;
      data.uncast(dbuf,true);
    }
    else {
      float fmean=(float)mean;
      float[] fbuf = data.castF(false);
      switch (shape) {                    // branch on wave shape
        case SIN: Waveform.sincos(fbuf,amp,p,dp,ndo,mode); break;
        case COS: Waveform.sincos(fbuf,amp,p+.25,dp,ndo,mode); break;
        case SQU: Waveform.square(fbuf,amp,p,dp,ndo,mode); break;
        case TRI: Waveform.triangle(fbuf,amp,p,dp,ndo,mode); break;
        case SAW: Waveform.sawtooth(fbuf,amp,p,dp,ndo,mode); break;
        case PUL: Waveform.pulse(fbuf,amp,p,dp,ndo,mode); break;
        case CON: Waveform.constant(fbuf,amp,ndo,mode); break;
        case ZER: Waveform.constant(fbuf,0.0,ndo,mode); break;
        case WHI: Waveform.whitenoise(fbuf,sdev,ndo,mode); break;
        case LRS: lrs=Waveform.lrs(fbuf,amp,ndo,mode,lrs); break;
        case RAM: cur=Waveform.ramp(fbuf,amp,ndo,mode,cur); break;
      }
      if (fmean!=0) for (int i=0; i<ndo*hcb.spa; i++) fbuf[i]+=fmean;
      data.uncast(fbuf,true);
    }

    hcb.write(data,ndo);

    phase += dp*ndo;			// increment phase
    phase -= Math.floor(phase);		// modulo 1.0

    if (chirp!=0) freq += chirp*dt*ndo;

    return NORMAL;
  }

  public int close() {
    hcb.close();
    return NORMAL;
  }

  public int restart() {
    hcb.close();
    hcb = MA.getDataFile("OUT",hcb,0);
    hcb.setFormat (form);
    hcb.setSize (todo);
    hcb.open();
    data = hcb.getDataBuffer(xfer,FLOAT);
    tstart = 0;
    return NORMAL;
  }

  public int processReady() {
    if (state!=PROCESS) return 0;
    return hcb.processReady(xfer);
  }

  // set run-time properties
  public void setShape (String value) { shape = Parser.find(shapeList,value,shape,0,Parser.E_WARNING); }
  public void setMean (double value) { mean = value; }
  public void setSDev (double value) { sdev = value; }
  public void setVariance (double value) { sdev = Math.sqrt(value); }
  public void setAmplitude (double value) { amp = value; sdev = amp*a2s; }
  public void setFrequency (double value) { freq = value; }
  public void setChirp (double value) { chirp = value; }
  public void setPhase (double value) { phase = value; }
  public void setThrottle (String value) { throttle = Parser.find(throttleList,value,throttle,0,Parser.E_WARNING); }
  public void setFormat (String format) { if (form.equals(format)) return; form = format; setState(RESTART); }

  // get run-time properties
  public String getShape () { return Parser.get(shapeList,shape); }
  public double getMean() { return mean; }
  public double getSDev () { return sdev; }
  public double getVariance () { return sdev*sdev; }
  public double getAmplitude () { return amp; }
  public double getFrequency () { return freq; }
  public double getChirp () { return chirp; }
  public double getPhase () { return phase; }
  public String getThrottle () { return Parser.get(throttleList,throttle); }
  public String getFormat () { return MA.getS("FORM"); }
}
