package nxm.ucl.intr;

import nxm.sys.inc.Units;
import nxm.sys.lib.*;

import static java.lang.Math.max;
import static java.lang.Math.min;
import static nxm.sys.lib.StateVector.computeMeasurements;

/** This is intended as a drop-in-replacement for the X-Midas DUAL5000 command
    that can be used by users migrating to NeXtMidas; all of the underlying
    computations are done by {@link StateVector#computeMeasurements}.
    @author  J. Czechowski IV
    @version $Id: Dual5000.java,v 1.2 2012/11/14 22:25:09 jxc Exp $
    @since   NeXtMidas 3.3.0
 */
public class Dual5000 extends Intrinsic {
  @Override
  public int process () {
    // SORT SOME THINGS OUT BASED ON THE INPUT =================================
    DataFile    in1 = MA.getDataFile("IN1", "5000", "NH|VD", DataFile.INPUT); in1.open();
    DataFile    in2 = MA.getDataFile("IN2", "5000", "NH|VD", DataFile.INPUT); in2.open();
    DataFile    out;
    double      start;
    double      delta;
    double      end;
    StateVector sv1;
    StateVector sv2;
    
    // Although it is tempting to use Position rather than StateVector in the
    // case where only a single point is given, this is not valid. The reason
    // is that X-Midas does not assume that a single-entry file is necessarially
    // a fixed point (i.e. it can have VEL=). Although this is a fairly silly
    // use case, it is presumably done for the benefit of users who found they
    // could using DUAL5000 to compare multiple unrelated configurations in a
    // single call (i.e. compare A->B,A->C,A->D) rather than making multiple
    // calls.
    if ((in1.getSize() == 1) && (in2.getSize() == 1)) {
      // This case is not in X-Midas, since X-Midas makes the implicit assumption
      // (as suggested in the explain file) that no more than 1 of the inputs
      // may be a single point.
      out   = MA.getDataFile("OUT", in1, DataFile.OUTPUT);
      start = in1.getTimeAt(0.0);
      delta = in1.getDelta();
      end   = start;
      sv1   = new StateVector(this, in1.getName());
      sv2   = new StateVector(this, in2.getName());
    }
    else if (in1.getSize() == 1) {
      out   = MA.getDataFile("OUT", in2, DataFile.OUTPUT);
      start = in2.getTimeAt(0.0);
      delta = in2.getDelta();
      end   = in2.getTimeAt(in2.getSize()-1);
      sv1   = new StateVector(this, in1.getName());
      sv2   = new StateVector(this, in2.getName());
    }
    else if (in2.getSize() == 1) {
      out   = MA.getDataFile("OUT", in1, DataFile.OUTPUT);
      start = in1.getTimeAt(0.0);
      delta = in1.getDelta();
      end   = in1.getTimeAt(in1.getSize()-1);
      sv1   = new StateVector(this, in1.getName());
      sv2   = new StateVector(this, in2.getName());
    }
    else {
      out   = MA.getDataFile("OUT", in1, DataFile.OUTPUT);
      start = max(in1.getTimeAt(0.0), in2.getTimeAt(0.0));
      delta = in1.getDelta();
      end   = min(in1.getTimeAt(in1.getSize()-1), in2.getTimeAt(in2.getSize()-1));
      sv1   = new StateVector(this, in1.getName());
      sv2   = new StateVector(this, in2.getName());
    }
    
    double rf  = -1;
    Object val = out.getKeyWord("RF");
    if (val != null) {
      rf = Convert.o2d(val);
    }
    
    // Although X-Midas allows the user to override RF, it does not permit the
    // start/delta to be overridden and presumes that both files will have the
    // same delta.
    rf    = MA.getD("/RF",    rf);
    start = MA.getD("/START", start);
    delta = MA.getD("/DELTA", delta);
    end   = MA.getD("/END",   end);
    
    if (out.getTimeCode() == 0) {
      out.setXStart(start);
      out.setXDelta(delta);
    }
    else if (out.getXStart() == 0) {
      out.setTimeCode(0.0);
      out.setXDelta(delta);
    }
    else {
      out.setTimeCode(0.0);
      out.setXStart(start);
      out.setXDelta(delta);
    }
    
    if ((out.getXUnits() != Units.TIME) || (out.getXUnits() != Units.TIMECODE)) {
      out.setXUnits(Units.TIME);
    }
    
    
    if (rf < 0) {
      // JC4: I'd prefer to handle this differently, but the 1 MHz with a
      // warning message matches X-Midas
      M.warning("No RF specified for doppler calc - using 1 MHz");
      rf = 1e6;
    }
    
    in1.close();
    in2.close();
    
    // CONFIGURE THE OUTPUT ====================================================
    out.setComponents("");
    for (int i = 4; i <= MA.getNArgs(); i++) {
      String key = MA.getS(i);
           if (key.equals("R"   )) out.addComp("R",    "SD", Units.UNDEFINED, Units.DISTANCE_M);
      else if (key.equals("RR"  )) out.addComp("RR",   "SD", Units.UNDEFINED, Units.SPEED_M_S);
      else if (key.equals("DELT")) out.addComp("DELT", "SD", Units.UNDEFINED, Units.DELAY_S);
      else if (key.equals("DELF")) out.addComp("DELF", "SD", Units.UNDEFINED, Units.DOPPLER_HZ);
      else if (key.equals("SPRF")) out.addComp("SPRF", "SD", Units.UNDEFINED, Units.DIMENSIONLESS);
      else if (key.equals("AZ"  )) out.addComp("AZ",   "SD", Units.UNDEFINED, Units.ANGLE_DEG);
      else if (key.equals("EL"  )) out.addComp("EL",   "SD", Units.UNDEFINED, Units.ANGLE_DEG);
      else                         throw new MidasException("Unsupported function: "+key);
    }
    
    if (out.getNumberComponents() == 0) {
      // Unlike X-Midas (which would output an empty file in this case), if no
      // output components are specified, we include all of them, however we
      // do not guarentee order.
      out.addComp("R",    "SD", Units.UNDEFINED, Units.DISTANCE_M);
      out.addComp("RR",   "SD", Units.UNDEFINED, Units.SPEED_M_S);
      out.addComp("DELT", "SD", Units.UNDEFINED, Units.DELAY_S);
      out.addComp("DELF", "SD", Units.UNDEFINED, Units.DOPPLER_HZ);
      out.addComp("SPRF", "SD", Units.UNDEFINED, Units.DIMENSIONLESS);
      out.addComp("AZ",   "SD", Units.UNDEFINED, Units.ANGLE_DEG);
      out.addComp("EL",   "SD", Units.UNDEFINED, Units.ANGLE_DEG);
    }
    
    // WRITE OUT THE OUTPUT ====================================================
    Table tbl = new Table();
    
    out.open();
    double timeCode = start;
    // We use "timeCode=start+delta*i" rather than "timeCode+=delta" to avoid
    // divergance due to rounding issues
    for (int i = 0; timeCode <= end; i++,timeCode=start+delta*i) {
      sv1.setTimeCode(timeCode);
      sv2.setTimeCode(timeCode);
      
      computeMeasurements(sv1, sv2, rf, tbl, true);
      out.setTimeAt(timeCode);
      out.writeDataTable(tbl);
    }
    out.close();
    
    return FINISH;
  }
}
