//******************************************************************************
//  Description: Test cases for MIDAS Utility routines (UTILITIES)
//
//  Author : Neon Ngo
//  Version: $Id: test_utilities.cc,v 1.14 2012/06/26 20:12:48 ntn Exp $
//
//******************************************************************************
#include <primitive.h>
#include <sstream>        // for ostringstream
//#define _DEBUG_NXM_TEST // uncomment to enable test framework debugging
#include "nxm_test.hh"    // XBC C++ test framework
#include <qmessages.h>    // included only to make sure qmessages.h and
                          //   generated: qmessdecl.h, qmessdef.hh builds cleanly
#include <xmtime.h>        // included only to make sure xmtime.h build cleanly

// functions declarations (keep in alphabetical order)
void test_m_between();
void test_m_d2a();
void test_m_head_m_tail_m_root_m_ext();
void test_m_lowercase();
void test_m_now_m_pause(real_4 pause_secs);
void test_m_promote_format();
void test_m_spa();
void test_m_trims();

// global variables
bool_4 _debug; // to enabled additional debugging messages

//******************************************************************************
// main C++ host primitive entry point
//******************************************************************************
void mainroutine () {
  real_8 pause;
  int_4  niter;

  pause  = m_get_dswitch_def("PAUSE", 0.3);
  niter  = m_get_switch_def("ITER",1);
  _debug = m_get_sswitch_def("DEBUG", FALSE);
//cout << "DEBUG: [" << Mu->id << "] test_utilities: pause=" << pause << " debug=" << _debug << endl;

  nxm_assert_init("DEBUG","STATUS");


  while (!Mc->break_ && (niter-- > 0)) {
    test_m_now_m_pause(pause);
  }

  test_m_promote_format();
  test_m_lowercase();
  test_m_spa();
  test_m_head_m_tail_m_root_m_ext();
  test_m_trims();
  test_m_between();
  test_m_d2a();

  nxm_assert_auto_report();
}

//******************************************************************************
// test case for m_now (BUG 2272) and m_pause (BUG 2271)
//******************************************************************************
void test_m_now_m_pause (real_4 pause_secs) {
  real_8 wsecs_before = -1;
  real_8 fsecs_before = -1;
  real_8 wsecs_after  = 0;
  real_8 fsecs_after  = 0;
  ostringstream ostr; // output stringstream for debug messages
  ostr.precision(14); // set floating point precision for this outstream (to display up to picoseconds, 10E-12 (12) + 2 for "0.")

  m_now(wsecs_before, fsecs_before); // get time before pausing
  if (_debug) {
    ostr.str(""); // clear/empty the ostringstream buffer
    ostr << "m_now() before: wsecs=" << wsecs_before << " fsecs=" << fsecs_before
         << "  pausing for " << pause_secs << " seconds...";
    m_info(ostr.str());
  }
  nxm_assert("m_now() before wsec", (wsecs_before >  0));
  nxm_assert("m_now() before fsec", (fsecs_before >= 0));

  m_pause(pause_secs);

  m_now(wsecs_after, fsecs_after);   // get time after pausing
  if (_debug) {
    ostr.str(""); // clear/empty the ostringstream buffer
    ostr << "m_now() after : wsecs=" << wsecs_after << " fsecs=" << fsecs_after;
    m_info(ostr.str());
  }
  nxm_assert("m_now() after wsec", (wsecs_after >  0));
  nxm_assert("m_now() after fsec", (fsecs_after >= 0));

  real_8 tolerance = pause_secs/100;
  real_8 elpasedTime = (wsecs_after - wsecs_before) + (fsecs_after - fsecs_before) + tolerance;
  nxm_assert("m_pause(.) elpased time", (elpasedTime >= pause_secs));
} // end test_m_now_m_pause()

//******************************************************************************
// test case for m_promote_format (BUG 2498)
//******************************************************************************
void test_m_promote_format () {
  const string expFormat = "CF";
  string format = m_promote_format("CI", "SF", "SL");
  nxm_assert_eqs("m_promote_format(..) output format", expFormat, format);
}

//******************************************************************************
// test case for m_lowercase (BUG 2498) and xmlibc.c f2jstr (BUG 2501)
//******************************************************************************
void test_m_lowercase () {
  const string lowercase = "this long mixed cased string should get lowercased! 123456789012345__";
  string       teststr   = "This Long Mixed Cased string SHOULD get LOWERCASED! 123456789012345__";
  m_lowercase(teststr);
  nxm_assert_eqs("m_lowercase(.) output", lowercase, teststr);

  // test case for f2jstr on input string longer than 160 chars (BUG 2501)
  const string expLongStr = lowercase + lowercase + lowercase;
  string       longstr    = teststr   + teststr   + teststr;
  nxm_assert("m_lowercase(.) input long string >= 160 chars", (longstr.length() >= 160));
  m_lowercase(longstr);
  nxm_assert_eqs("m_lowercase(.) output of long string >= 160 chars", expLongStr, longstr);
}

//******************************************************************************
// test case for m_spa (BUG 2523)
//******************************************************************************
void test_m_spa () {
  int    expSPA[]  = { 1,    2,    3,    4,    9,    10,   16,   1};
  string formats[] = {"SB", "CI", "VL", "QX", "MF", "XD", "TB", "UI"};
  int    size      = sizeof(expSPA) / sizeof(int); // no easy way to get size of array

  for (int ii=0; ii<size; ii++) {
    string format = formats[ii];
    int result = m_spa(format);
    nxm_assert_eq("m_spa(.) "+format, expSPA[ii], result);
  }
}

//******************************************************************************
// test case for m_head, m_tail, m_root, and m_ext (BUG 1374), since 3.1.2
//******************************************************************************
void test_m_head_m_tail_m_root_m_ext () {
  const string fname = "/tmp/path/to/filename.ext";
  const string expectedHead = "/tmp/path/to/";
  const string expectedTail = "filename.ext";
  const string expectedRoot = "filename";
  const string expectedExt  = "ext";
  string teststr;
  int_4 outlen;

  outlen = m_head(fname, teststr);
  nxm_assert_eq("m_head(.) outstr len", expectedHead.length(), outlen);
  nxm_assert_eqs("m_head(.) outstr equals ", expectedHead, teststr);

  outlen = m_tail(fname, teststr);
  nxm_assert_eq("m_tail(.) outstr len", expectedTail.length(), outlen);
  nxm_assert_eqs("m_tail(.) outstr equals", expectedTail, teststr);

  outlen = m_root(fname, teststr);
  nxm_assert_eq("m_root(.) outstr len", expectedRoot.length(), outlen);
  nxm_assert_eqs("m_root(.) outstr equals", expectedRoot, teststr);

  outlen = m_ext(fname, teststr);
  nxm_assert_eq("m_ext(.) outstr len", expectedExt.length(), outlen);
  nxm_assert_eqs("m_ext(.) outstr equals", expectedExt, teststr);
}

//******************************************************************************
// test case for m_trims (BUG 1374), since 3.1.2
//******************************************************************************
void test_m_trims () {
  string outStr = "test";
  const string substr1        = "This is a ";
  const string substr2        = " testcase.";
  const string expectedOutStr = " m_trims";
  string teststr              = "This is a m_trims testcase.";
  int_4 outlen;

  outlen = m_trims(teststr, outStr, substr1, substr2);
  nxm_assert_eq("m_trims(.) outstr len", expectedOutStr.length(), outlen);
  nxm_assert_eqs("m_trims(.) outstr equals ", expectedOutStr, outStr);
}

//******************************************************************************
// test case for m_between (BUG 1374), since 3.1.2
//******************************************************************************
void test_m_between () {
  string outStr         = "";
  string teststr        = "This is a string";
  int_4 outlen;
  int_4 min             = 1;
  int_4 max             = 7;
  string expectedOutStr = "This is";

  outlen = m_between(teststr, outStr, min, max);
  nxm_assert_eq("m_between(.) outstr len", expectedOutStr.length(), outlen);
  nxm_assert_eqs("m_between(.) outstr equals ", expectedOutStr, outStr);

  outlen = m_between(teststr, outStr, max, min);
  nxm_assert_eq("m_between(.) outstr len", expectedOutStr.length(), outlen);
  nxm_assert_eqs("m_between(.) outstr equals ", expectedOutStr, outStr);

  min = 4;
  max = -100;
  expectedOutStr = "This";

  outlen = m_between(teststr, outStr, min, max);
  nxm_assert_eq("m_between(.) outstr len", expectedOutStr.length(), outlen);
  nxm_assert_eqs("m_between(.) outstr equals ", expectedOutStr, outStr);

  outlen = m_between(teststr, outStr, max, min);
  nxm_assert_eq("m_between(.) outstr len", expectedOutStr.length(), outlen);
  nxm_assert_eqs("m_between(.) outstr equals ", expectedOutStr, outStr);

  min = -5;
  max = 0;
  expectedOutStr = "string";

  outlen = m_between(teststr, outStr, min, max);
  nxm_assert_eq("m_between(.) outstr len", expectedOutStr.length(), outlen);
  nxm_assert_eqs("m_between(.) outstr equals ", expectedOutStr, outStr);

  outlen = m_between(teststr, outStr, max, min);
  nxm_assert_eq("m_between(.) outstr len", expectedOutStr.length(), outlen);
  nxm_assert_eqs("m_between(.) outstr equals ", expectedOutStr, outStr);
}

//******************************************************************************
// since 3.1.2 for testing BUG 2606
//******************************************************************************
void test_m_d2a () {
  const double dval   =  0.123456;
  const string expStr = "0.1234560";
  string ostr;
  int_4 len = m_d2a(dval, ostr);
  nxm_assert_eq("m_d2a(.) outstr len", expStr.length(), len);
  nxm_assert_eqs("m_d2a(.) outstr equals ", expStr, ostr);
}
