package tools;

import java.io.*;
import java.util.*;
import java.lang.reflect.Method;

/** This class represents the build properties for a specific option tree.
    <br>
    <br>
    The properties held by this class are a combination of the following things:
    <ol>
      <li>
        The property <code>nmroot</code> is set to the path to the
        <code>$NMROOT</code> directory as determined by the system
        property <code>NMROOT</code>.
      </li>
      <li>
        The property <code>nmlibpath</code> is a Java classpath that includes
        all <code>.jar</code> files in <code>$NMROOT/jars</code>,
        <code>$NMROOT/nxm/sys/jars</code> and <code>$NMROOT/jars</code> that do
        not begin with <code>nxm-</code> (the prefix for the NeXtMidas WebStart
        Jars).
      </li>
      <li>
        The properties file for a specified option tree (located
        at <code>$NMROOT/nxm/<i>option</i>/cfg/build.props</code> where
        <code><i>option</i></code> is the name of the appropriate option
        tree other than <code>SYS</code>). If no such file exists, it is
        ignored.
      </li>
      <li>
        The properties file for the <code>SYS</code> option tree (located
        at <code>$NMROOT/nxm/sys/cfg/build.props</code>).
      </li>
      <li>
        The system properties (via the {@link System#getProperties()} method).
      </li>
    </ol>
    If any of the above properties include duplicate definitions of the same
    property the highest ranked property (lowest number on above list) will
    be used. This enables an arbitrary option tree to override the properties
    defined in the <code>SYS</code> option tree.<br>
    <br>
    <b>Implementation note:</b> This class does not require access to any
    other classes outside of the <code>java.*</code> package. This permits
    the class to be compiled and run independent of other NeXtMidas classes.

    @author  J. Czechowski IV / Neon Ngo
    @version $Id: BuildProps.java,v 1.17 2013/01/07 19:05:03 ntn Exp $
    @since   NeXtMidas 1.9.2
 */
public final class BuildProps extends Properties {
  /** NOTE: this should be changed if the structure/fields of this class changes. */
  private static final long serialVersionUID = 2011051700292L;

  /** max number of nested levels allowed to expand variables in properties values.
      used to prevent StackOverflowError from recursive property variable expansion.
   */
  private static final int MAX_NESTED_LEVEL = 10;

  /** The set of all currently loaded properties files.
      Properties files are cached in memory until {@link #clearCache()} or
      when the JVM exits.
      <pre>
        KEY   - The name of the option tree (in uppercase).
        VALUE - The instance of BuildProps for the specified option tree.
      </pre>
   */
  private static Hashtable<String,BuildProps> allBuildProps = new Hashtable<String,BuildProps>();

  /** Last modified time of property file when it was loaded (used to reload file if it has changed) */
  private long lastModified = 0; // using zero, since File.lastModified() return that if file does not exists or on IO error
  /** Property file reference (used to reload file if it has changed since we last read it) */
  private File propertyFile;

  /** This will instantiate the class for the specified option tree.

      @param opt The option tree to get the properties for.
      @param ref The reference object (a pointer to Shell if available).

      @throws RuntimeException If the properties file cannot be read.
   */
  private BuildProps(String opt, Object ref) {
    super(opt.equalsIgnoreCase("SYS")? System.getProperties():
                                       BuildProps.getPropsFor("SYS")); // Use SYS BuildProps as defaults for all other option trees
    try {
      String fileName = BuildProps.getNmRoot()
                      + File.separator
                      + "nxm"
                      + File.separator
                      + opt.toLowerCase()
                      + File.separator
                      + "cfg"
                      + File.separator
                      + "build.props";
      if (ref != null) {
        Method getOptPath = ref.getClass().getMethod("getOptPath", new Class<?>[] {String.class,String.class});
        fileName = "" + getOptPath.invoke(ref, (Object[])new String[] { opt, "cfg" }) + "build.props";
      }

      propertyFile = new File(fileName);

      if (propertyFile.exists()) {
        lastModified = propertyFile.lastModified();
        this.load(new FileInputStream(propertyFile));
      }
    } catch(Exception exception) {
      throw new RuntimeException("Unable to load properties file for the ["
                               + opt.toUpperCase()
                               + "] option tree.", exception);
    }

    this.put("nmroot",    BuildProps.getNmRoot());
    this.put("nmlibpath", BuildProps.getNmLibPath());

    allBuildProps.put(opt.toUpperCase(), this);
  }

  /** Clear (reset) all cached properties files (build.props) from memory.
      @since NeXtMidas 2.7.0
   */
  public static void clearCache() {
    allBuildProps.clear();
  }

  /** This will attempt to determine the NMROOT variable.
      @return The NMROOT variable.
   */
  private static String getNmRoot() {
    String nmroot = System.getProperty("NMROOT");

    /* If nmroot is still null then NMROOT is probably not specified. This
       usually means the user started something from the commandline (such as
       BaselineBuild) and did not specify the value of NMROOT. Therefore just
       output a message explaining this and exit.
     */
    if (nmroot == null) {
      System.err.println("ERROR: The value of NMROOT is undefined. Try running "
                       + "this application again with '-DNMROOT=???' specified "
                       + "(where ??? is the path to your NMROOT).");

      System.exit(-1);

      throw new RuntimeException("This line should be unreachable.");
    } else {
      return nmroot;
    }
  }

  /** This will attempt to determine the CLASSPATH variable that includes all of
      the NeXtMidas libraries.

      @return The CLASSPATH variable for the NeXtMidas libraries.

      @see #getNmLibs()
   */
  private static String getNmLibPath() {
    final File[] jars = getNmLibs();
    final StringBuffer classpath  = new StringBuffer();

    if (jars.length > 0) {
      classpath.append(jars[0]);
      for (int i = 1; i < jars.length; i++) {
        classpath.append(File.pathSeparator);
        classpath.append(jars[i]);
      }
    }

    return classpath.toString();
  }

  /** This will attempt to determine the CLASSPATH variable that includes all of
      the NeXtMidas libraries. <br>
      <br>
      By default this will include all <code>.jar</code> files that do not begin
      with <code>nxm-</code> in <code>$NMROOT/jars</code>,
      <code>$NMROOT/nxm/sys/jars</code> and <code>$JAVAHOME/lib</code>.

      @return The library files to include in the classpath for NeXtMidas.
   */
  private static File[] getNmLibs() {
    final File           nxmDir  = new File(getNmRoot()+File.separator+"nxm");
    final File[]         optDirs = nxmDir.listFiles();
    final ArrayList<String>      dirs = new ArrayList<String>();
    final Hashtable<String,File> jars = new Hashtable<String,File>();
    final FilenameFilter filter  = new FilenameFilter() {
        public boolean accept(File dir, String name) {
          return name.endsWith(".jar") && (!name.startsWith("nxm-"));
        }
    };

    dirs.add(getNmRoot() + File.separator + "jars");
    for (File dir : optDirs) {
      dirs.add(dir.getAbsolutePath() + File.separator + "jars");
    }
    dirs.add(System.getProperty("javahome") + File.separator + "lib");

    for (String dir : dirs) {
      final File jarDir = new File(dir);

      if (jarDir.exists() && jarDir.isDirectory()) {
        final File[] files = jarDir.listFiles(filter);

        for (File file : files) {
          jars.put(file.getName(), file);
        }
      }
    }

    // ORIG return (File[])new Vector(jars.values()).toArray(new File[0]);
    return jars.values().toArray(new File[0]);
  }

  /** Reload the property file if it is newer than the current cached one.
      @since NeXtMidas 2.9.2
   */
  private void checkForChanges () {
    try {
      if (propertyFile.lastModified() > lastModified) {
        if (propertyFile.exists()) {
          lastModified = propertyFile.lastModified();
          load(new FileInputStream(propertyFile));
        }
      }
    } catch (Exception e) {
      warn("BuildProps.checkForChanges: unable to check for changes to properties file: "+propertyFile);
      e.printStackTrace(System.err);
    }
  }

  /** This will get a build property for the <code>SYS</code> option tree.
      <br>
      <br>
      This is identical to <code>BuildProps.getPropFor("SYS",prop)</code>.

      @param prop The name of the property to get.

      @return The specified property for the <code>SYS</code> option tree,
              or <code>null</code> if such property does not exist.

      @see #getPropFor(String,String)
   */
  public static final String getSysProp(String prop) {
    return BuildProps.getPropFor("SYS", prop);
  }

  /** This will get a build property for the <code>SYS</code> option tree as
      a path for the current operating system.
      <br>
      <br>
      This is identical to <code>BuildProps.getPathFor("SYS",prop)</code>.

      @param prop The name of the property to get.

      @return The specified property for the <code>SYS</code> option tree,
              or <code>null</code> if such property does not exist.

      @see #getPathFor(String,String)
   */
  public static final String getSysPath(String prop) {
    return BuildProps.getPathFor("SYS", prop);
  }

  /** This will get the build property for a given option tree as
      a path for the current operating system. <br>
      <br>
      This is identical to
      <code>BuildProps.getPropFor(opt).getProperty(prop)</code>.

      @param prop The name of the property to get.

      @return The specified property for the <code>SYS</code> option tree,
              or <code>null</code> if such property does not exist.

      @see #getPropsFor(String)
      @see #getProperty(String)
   */
  public static final String getPropFor(String opt, String prop) {
      return getPropFor(opt, prop, null);
  }

  /** This will get the build property for a given option tree as
      a path for the current operating system. <br>
      <br>
      This is identical to
      <code>BuildProps.getPropsFor(opt).getProperty(prop)</code>.

      @param prop The name of the property to get.

      @return The specified property for the <code>SYS</code> option tree,
              or <code>null</code> if such property does not exist.

      @see #getPropsFor(String)
      @see #getProperty(String)
   */
  public static final String getPropFor(String opt, String prop, Object ref) {
    return BuildProps.getPropsFor(opt,ref).getProperty(prop);
  }

  /** This will get the build property for a given option tree. <br>
      <br>
      This is identical to
      <code>BuildProps.getPropFor(opt).getPath(prop)</code>.

      @param prop The name of the property to get.

      @return The specified property for the <code>SYS</code> option tree,
              or <code>null</code> if such property does not exist.

      @see #getPropsFor(String)
      @see #getPath(String)
   */
  public static final String getPathFor(String opt, String prop) {
      return getPathFor(opt, prop, null);
  }

  /** This will get the build property for a given option tree. <br>
      <br>
      This is identical to
      <code>BuildProps.getPropFor(opt).getPath(prop)</code>.

      @param prop The name of the property to get.

      @return The specified property for the <code>SYS</code> option tree,
              or <code>null</code> if such property does not exist.

      @see #getPropsFor(String)
      @see #getPath(String)
   */
  public static final String getPathFor(String opt, String prop, Object ref) {
    return BuildProps.getPropsFor(opt,ref).getPath(prop);
  }

  /** This will get the build properties for the specified option tree. <br>
      <br>
      If the properties file for the given option tree has not yet been loaded
      it will be loaded at this time. If the properties file for the option
      tree has already been loaded it will use the existing instance.

      @param opt The option tree to get the properties for.

      @return The build properties for the specified option tree.
   */
  public static final BuildProps getPropsFor(String opt) {

    return getPropsFor(opt, null);
  }

  /** This will get the build properties for the specified option tree. <br>
      <br>
      If the properties file for the given option tree has not yet been loaded
      it will be loaded at this time. If the properties file for the option
      tree has already been loaded it will use the existing instance.

      @param opt The option tree to get the properties for.

      @return The build properties for the specified option tree.
   */
  public static final BuildProps getPropsFor(String opt, Object ref) {
    BuildProps props = allBuildProps.get(opt.toUpperCase());

    if (props != null) {
      props.checkForChanges();
      return props;
    } else {
      return new BuildProps(opt, ref);
    }
  }

  /** This will expand variables specified via ("${varname}") in a string. This means that each
      variable surrounded by a dollar and braces will get replaces with the corresponding
      property's value.
      For example, if the following properties are set: <pre>
      x = joey
      test.joey = bella
      then
      y = Hello ${x}.       ==> Hello joey.
      z = Hi ${test.${x}}!  ==> Hi bella!  (e.g. nested variables are expanded since 2.9.2)
      @param value The original string with any ${varname} in it resolved (expanded).
      @param nestedLevel recursive nesting level (after #MAX_NESTED_LEVEL levels,
               a Runtime Exception will be thrown since this is probably b/c of a recursive
               property variable expansion.
      @return The string with any variables resolved.
      @throws RuntimeException if a variable to expand is NOT found (i.e. not defined) or
              if braces are unbalanced.
      @since NeXtMidas 2.9.2
   */
  private String expandVariables (String value, int nestedLevel) {
    if (value==null) return value;
    int iDollar, iEnd, lenValue;

    // recursively check for embedded variables and replace them (working from backwards from end of string)
    for (int jj=lenValue=value.length(); (iDollar=value.lastIndexOf('$',jj))>=0; jj=iDollar-1) {
      String expandedValue = null;
      if ((iDollar+1 < lenValue) && value.charAt(iDollar+1)=='{') {
        iEnd = value.indexOf('}',iDollar);
        // If the curly braces do not match (i.e. "{ ... }") then it is an error.
        //   For example, "hello ${A" would cause this method to throw an exception.
        if (iEnd < iDollar+1) {
          throw new RuntimeException("Error loading property. Syntax Error: Can not evaluate variable: Missing '}' character in ( " + value + " ).");
        }
        String varName = value.substring(iDollar+2, iEnd);
        if (nestedLevel >= MAX_NESTED_LEVEL) // since 3.3.1
          throw new RuntimeException("Error loading property. Too many nested levels [" + nestedLevel
            + "] trying to expand variable(s) in property [" + varName + "] from value [" + value
            + "]. Check for any recursive property variables to expand.");

        expandedValue = getProperty(varName, nestedLevel+1);

        if (expandedValue == null) {
          throw new RuntimeException("Error loading property. Unknown variable '${" + varName + "}'.");
        }

        iEnd++; // point to just after right (closing) brace
      }
      else {
        if (jj<1) continue;
        iEnd = iDollar+jj+1;
      }
      if (expandedValue==null) continue;
      if (iEnd<lenValue) {
        expandedValue = expandedValue + value.substring(iEnd); // append value to right of expanded variable's value
      }
      if (iDollar==0) return expandedValue; // at start, so just return expanded value
      value = value.substring(0, iDollar) + expandedValue; // process again with left of variable + expanded variable's value
      lenValue = value.length();
    } // end for loop
    return value;
  }

  /** Gets the specified property. <br>
      <br>
      When loading the property the following translations will be performed:
      <ul>
        <li>
          Any embedded variables will be replaced. (For example a variable
          defined as <code>"${nmroot}/nxm/sys/mcr"</code> in the properties
          file will have <code>${nmroot}</code> replaced with the property it
          represents.)
        </li>
      </ul>

      @param key The name of the property to get. If <code><i>key</i></code>
                 starts with <code>"env."</code> it will automatically get
                 the property from the system properties (via
                 {@link System#getProperties()}). <i>(This feature is for
                 compatibility with Ant which can load the system properties
                 to match this.)</i> Cannot be null.
      @return  The value in this property list with the specified key value. Returns
               null if the property is not found at the specified key.
      @throws RuntimeException If the property has an embedded variable that can
                               not be resolved.

      @see #getProperty(String,String)
   */
  public final String getProperty (String key) {
    return getProperty(key, 0);
  }

  /** @see #getProperty(String)
      @since NeXtMidas 3.3.1
   */
  private final String getProperty (String key, int nestedLevel) {
    String prop = null;
    try {
      if (key.startsWith("env.")) {
        prop = System.getProperty(key.substring(4));
      } else {
        checkForChanges();
        prop = super.getProperty(key);
      }
    } catch (RuntimeException exception) {
      warn("Unable to get '"+key+"' property. "+exception.getMessage());
      // Unable to get property, just make it null.
      prop = null;
    }

    return expandVariables(prop, nestedLevel);
  }

  /** legacy version on expanding variables in property values (nested
      variables are NOT supported). This is original behavior in getProperty(..)
      prior to 2.9.2.
      TODO: remove in 3.1.x or later
      @see #getPropertyLegacy(String)
   */
  private String expandVariablesLegacy (String prop) {
    if (prop==null) return prop;

    // recursively check for embedded variables and replace them
    while ( prop.indexOf ( "${", 0 ) > -1 ) {
      int start = prop.indexOf ( "${", 0 );
      int end   = prop.indexOf ( '}', start );

      String varName  = prop.substring ( start + 2, end );
      String varValue = this.getProperty( varName );

      if (varValue != null) {
        prop = prop.substring( 0, start )
             + varValue
             + prop.substring( end + 1 );

      } else {
        throw new RuntimeException("Error loading property. Unknown "
                                 + "variable '${" + varName + "}'.");
      }
    }
    return prop;
  }

  /** legacy version of getProperty(String key) prior to NeXtMidas 2.9.2.
      This is only here to allow someone to fall back this this old behavior
      without nested variable expansion support (just in case it is causing
      problems for them).
      TODO: remove in 3.1.x or later
      @deprecated since initial version in NeXtMidas 2.9.2
      @see  #getProperty(String)
   */
  @Deprecated public final String getPropertyLegacy (String key) {
    String prop = null;
    try {
      if (key.startsWith("env.")) {
        prop = System.getProperty(key.substring(4));
      } else {
        checkForChanges();
        prop = super.getProperty(key);
      }
    } catch(RuntimeException exception) {
      /* Unable to get property, just make it null. */
      prop = null;
    }

    return expandVariablesLegacy(prop);
  }

// 2011-05-18 NTN: This method is defined the same in the Properties's super class.
//                 So there is no need to redefine it here.
// TODO: remove in 3.1.x or later
//  /** Gets the specified property. <br>
//      <br>
//      This is identical to <code>getProperty(key)</code> except that this
//      will return <code><i>def</i></code> in place of <code>null</code>.
//
//      @param key The name of the property to get.
//      @param def The default value to return if the key is not found.
//
//      @return The value of the property matching the given key or
//              <code><i>def</i></code> the key is not found.
//
//      @see #getProperty(String)
//   */
//  public final String getProperty(String key, String def) {
//    String value = this.getProperty(key);
//
//    return ((value != null)? value : def);
//  }

  /** Get value of property at specified key. If prependSuffixes is specified, this
      method will search for properties at "key + prependSuffixes[n]" and prepend it to
      the property value (space separated). Similarly, if appendSuffixes is specified, this
      method will search for properties at "key + appendSuffixes[n]" and append
      it to the property value (spaces separated). <br>
      For example, given following property key/values
      <pre>mykey = ABC
      mykey.32bit = options for 32-bit
      mykey.64bit = options for 64-bit
      mykey.append = ZZZ
      mykey.prepend = JJJ

      and the following code:
      String[] prependSuffixes = {".prepend"};
      String[] appendSuffixes = {".32bit", ".anothersuffix", ".append"};
      String propValue  = getProperty("mykey", prependSuffixes, appendSuffixes);
      ----
      The above gets propValue will include all the values from keys at:
        "mykey.prepend" + "mykey", "mykey.32bit", "mykey.anothersuffix", "mykey.append",
      (if any of those properties exists in that order).
      This results in
      propValue ==> "JJJ ABC options for 32-bit ZZZ".</pre>

      @param key             The name of the property to get.
      @param prependSuffixes String array of suffixes (e.g. {".prepend"}) to prepend to key
                             to search for additional property values to prepend to final returned
                             property value. Ignored if null.
      @param appendSuffixes  String array of suffixes (e.g. {".64bit", ".append"}) to append to key
                             to search for additional property values to append to final returned
                             property value. Ignored if null.
      @see #getProperty(String)
      @since NeXtMidas 2.9.2
   */
  public final String getProperty(String key, String[] prependSuffixes, String[] appendSuffixes) {
    String value = getProperty(key);
    StringBuilder sb = new StringBuilder();

    if (prependSuffixes != null) {
      for (String suffix : prependSuffixes) { // prepend to property the following key suffixes
        String val = getProperty(key + suffix);
        if (val != null && val.length()>0) {  // ignore null and empty values
          sb.append(val).append(" ");
        }
      }
    }

    if (value != null && value.length()>0) {  // ignore null and empty values
      sb.append(value);
    }

    if (appendSuffixes != null) {             // append to property the following key suffixes
      for (String suffix : appendSuffixes) {
        String val = getProperty(key + suffix);
        if (val != null && val.length()>0) {  // ignore null and empty values
          sb.append(" ").append(val);
        }
      }
    }

    return (sb.length() == 0)? null : sb.toString();
  }

  /** Gets the specified property as a path for the current
      operating system. <br>
      <br>
      This is identical to <code>getPath(key,null)</code>

      @param key The name of the property to get.

      @see #getProperty(String)
   */
  public final String getPath(String key) {
    return this.getPath(key,null);
  }

  /** Gets the specified property as a path for the current
      operating system. <br>
      <br>
      This is the same as {@link #getProperty(String)} except that any path
      separators in the property (either <code>'\'</code> (DOS) or
      <code>'/'</code> (UNIX)) will be converted to the path separator in use
      by the current operating system.

      @param key The name of the property to get.
      @param def The default value to return if the key is not found.

      @return The value of the property matching the given key or
              <code><i>def</i></code> the key is not found.

      @see #getProperty(String)
   */
  public final String getPath(String key, String def) {
    String prop = this.getProperty(key, def);

    if (prop != null) {
      /* Change any occurrence of '\' or '/' into the proper file separator. */
      if (!File.separator.equals("\\")) {
        prop = prop.replaceAll("[\\\\/]", File.separator);
      } else {
        prop = prop.replaceAll("[\\\\/]", "\\\\"); /* Must use escape char. */
      }

      return prop;
    } else {
      return prop;
    }
  }

  /** Display specified warning message with "WARN: " pre-pended using System.err.println (to STDERR).
      @param msg warning message to display
   */
  private static final void warn (CharSequence msg) { // since 3.1.3
    System.err.println("WARN: "+msg);
  }

  /** Command-line interface that permits external scripts to read from the
      <code>build.props</code> file. <br>
      <br>
      Usage:
      <pre>
        BuildProps usage: BuildProps [-p] PROPERTY [OPT [VNAME]]
          Gets the property PROPERTY from the build.props
          file of the OPT option tree (default is SYS).

          -p  Translates any path separators in the prop to
              that of the current system.

          If VNAME is specified it will print out the property in a
          "set VNAME=" syntax (used by the Windows shell scripts).
      </pre>

      @param args The command-line arguments.
   */
  public static void main(String[] args) {
    boolean error   = false;
    boolean convert = false;
    String  prop    = null;
    String  opt     = "sys";

    if ((args == null) || (args.length == 0)) {
      System.err.println("ERROR: No property specified.");
      error = true;

    } else if (args.length == 1) {
      prop    = args[0];

    } else if (args.length == 2) {
      if (args[0].equals("-p")) {
        convert = true;
        prop    = args[1];

      } else {
        prop    = args[0];
        opt     = args[1];
      }

    } else if (args.length == 3) {
      if (args[0].equals("-p")) {
        convert = true;
        prop    = args[1];
        opt     = args[2];
      } else {
        prop    = args[0];
        opt     = args[1];
        System.out.print("set " + args[2] + "=");
      }
    } else if ((args.length == 4) && (args[0].equals("-p"))) {
      convert = true;
      prop    = args[1];
      opt     = args[2];
      System.out.print("set " + args[3] + "=");
    } else {
      System.err.println("ERROR: Invalid number of arguments.");
      error = true;
    }



    if (!error) {
      if (convert) {
        System.out.println(BuildProps.getPathFor(opt, prop));
      } else {
        System.out.println(BuildProps.getPropFor(opt, prop));
      }

    } else {
      System.err.println();
      System.err.println("BuildProps usage: BuildProps "
                       + "[-p] PROPERTY [OPT [VNAME]]");
      System.err.println("  Gets the property PROPERTY from the build.props");
      System.err.println("  file of the OPT option tree (default is SYS).");
      System.err.println();
      System.err.println("  -p  Translates any path separators in the prop to");
      System.err.println("      that of the current system.");
      System.err.println("");
      System.err.println("      If VNAME is specified it will print out the");
      System.err.println("      property in a 'set VNAME=' syntax (used by");
      System.err.println("      the Windows shell scripts).");

      System.exit(-1);
    }
  }
}
