package tools;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.Hashtable;

import javax.swing.JButton;
import javax.swing.JEditorPane;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

/** This provides a simple tool that will enable a user to sort through a set of
    CVS logs (in a CVS log file) to easily see which changes match which DR and
    to find which DRs were being worked on between a specified set of dates.
    <br>
    <br>
    Typical usage (as a stand-alone program):
    <pre>
      $ cd $MY_CVS_DIR                (For NeXtMidas this $MY_CVS_DIR = $NMROOT)
      $ cvs log > /tmp/cvslog
      $ java tools.DRManager /tmp/cvslog
    </pre>
    Typical usage from within another Java program:
    <pre>
      // Starts the DRManager.
      //
      // @param nameOfLogFile The name of the CVS log file to use.
      //
      // @throws IOException If there is an error in reading the log file.
      //
      public void startDRManager(String nameOfLogFile) throws IOException {
        DRManager mgr = new DRManager(nameOfLogFile);

        // false must be passed to .createGUI(...) to prevent a call to
        // System.exit(...) when the user exits the GUI.
        mgr.createGUI(false);

        mgr.setVisible(true);
      }
    </pre>
    <br>
    <pre>
    Note: This is designed to work with Java version 1.5 or later. It is
          unlikely that this will run on earlier versions of Java.
    </pre>

    @author  J. Czechowski IV
    @version $Id: DRManager.java,v 1.5 2010/08/02 23:08:07 ntn Exp $
    @since   NeXtMidas Version 1.8.0
 */
public class DRManager {
  /** This class represents a File in CVS that has both a name and a set of
      changes associated with it.
   */
  protected static class File {
    /** The file name. */
    protected String name;

    /** The changes/DRs for the file */
    protected ArrayList<Change> changes;

    /** Creates a new instance.
        @param name The file's name.
     */
    public File(String name) {
      this.changes = new ArrayList<Change>();
      this.name = name;
    }

    /** Gets the file's name.
        @return The file's name.
     */
    public String getName() {
      return this.name;
    }

    /** Adds a Change/DR to this file.
        @param change The change/DR to add.
     */
    public void addChange(Change change) {
      this.changes.add(change);
    }

    /** Gets all of the changes to the file.
        @return All of the changes/DR's for the file.
     */
    public Collection<Change> getChanges() {
      return this.changes;
    }
  }

  /** This is a committed change to the file. */
  protected static class Change implements Comparable<Change> {
    /** The date of the change. */
    protected Date date;

    /** This is the ID of the person making the change. */
    protected String author;

    /** This is the comment about the change. */
    protected String comment;

    /** The file the change is applied to. */
    protected File file;

    /** Creates a new instance and add the change to the associated file.

        @param file    The file this change is applied to.
        @param date    The date of the change.
        @param author  The ID of the person making the change.
        @param comment Any author-inserted comments regarding the change.
     */
    public Change(File file, Date date, String author, String comment) {
      this.file    = file;
      this.date    = date;
      this.author  = author;
      this.comment = comment;

      file.addChange(this);
    }

    /** Gets the change's date.
        @return The date of the change.
     */
    public Date getDate() {
      return this.date;
    }

    /** Gets the change's date.
        @return The date of the change.
     */
    public String getAuthor() {
      return this.author;
    }

    /** Gets the change's date.
        @return The date of the change.
     */
    public String getComment() {
      return this.comment;
    }

    /** Gets the file the change applies to.
        @return The file the change applies to.
     */
    public File getFile() {
      return this.file;
    }

    public int compareTo (Change obj) {
     return this.date.compareTo(obj.date);
    }
  }

  /** This is a change that includes an associated DR. */
  protected static class DRChange extends Change {
    /** The the DR the change is associated with. */
    protected DR dr;

    /** Creates a new instance and add the change to the associated file.

        @param file    The file this change is applied to.
        @param date    The date of the change.
        @param author  The ID of the person making the change.
        @param comment Any author-inserted comments regarding the change.
        @param dr      The number associated DR.
     */
    public DRChange(File file, Date date, String author,
                    String comment, DR dr) {
      super(file, date, author, comment);
      this.dr = dr;
      dr.addChange(this);
    }

    /** Gets the DR associated with the change.
        @return The DR associated with the change.
     */
    public DR getDR() {
      return this.dr;
    }
  }

  /** This is a complete DR that has many changes associated with it. */
  protected static class DR implements Comparable<DR> {
    /** The number of the DR. (This may include non-numeric characters.) */
    protected String drNum;

    /** All of the changes associated with this DR. */
    protected ArrayList<Change> changes;

    /** Creates a new instance.
        @param drNum   The number of the DR.
     */
    public DR(String drNum) {
      this.changes = new ArrayList<Change>();

      if (drNum.matches(DRManager.DR_NUM_REGEX)) {
        int length = drNum.indexOf('-');
        if (length == -1) {
          length = drNum.length();
        }

        for (int i = length; i < 8; i++) {
          drNum = " " + drNum;
        }
      } else {
        for (int i = 0; i < 10; i++) {
          drNum = " " + drNum;
        }
      }

      this.drNum = drNum;
    }

    /** Gets the change's DR number.
        @return The DR number.
     */
    public String getDRNum() {
      return this.drNum.trim();
    }

    /** This will associate a new change with this DR.
        @param change The new change to associate with this DR.
     */
    public void addChange(DRChange change) {
      this.changes.add(change);
    }

    /** Gets a list of all the changes.
        @return All of the changes associated with the DR.
     */
    public DRChange[] getChanges() {
      DRChange[] array = new DRChange[this.changes.size()];

      return this.changes.toArray(array);
    }

    public int compareTo (DR obj) {
      return this.drNum.compareTo(obj.drNum);
    }
  }

  /** This is the regular expression which indicates valid separators that may
      be used to separate a list of DR numbers. <br>
      <br>
      Currently this is set to <code>{@value}</code>.
   */
  public static final String DR_SEP_REGEX = "[ \t,:;]+";

  /** This is the regular expression which all DR numbers must match. <br>
      <br>
      Currently this is set to <code>{@value}</code>.
   */
  public static final String DR_NUM_REGEX = "[0-9]+([-][0-9]+)?";

  /** This is the set of special DR codes that are permitted. These are
      considered to be case-insensitive.<br>
      <br>
      Currently this is set to
      <code>{"Docs", "Release", "New", "Test", "Fix", "Cleanup", "Deprecate"}</code>.
   */
  public static final String[] DR_CODES
    = {"Docs", "Release", "New", "Test", "Fix", "Cleanup", "Deprecate"};

  /** This is the regular expression matching the set of special DR codes that
      are permitted.

      @see #DR_CODES
      @see #DR_REGEX
   */
  public static final String DR_CODE_REGEX = makeRegex(DR_CODES);

  /** This is the regular expression which all DR numbers must match.

      @see #DR_NUM_REGEX
      @see #DR_CODE_REGEX
   */
  public static final String DR_REGEX
    = "((" + DR_NUM_REGEX + ")|(" + DR_CODE_REGEX + "))";

  /** This is the date format used by CVS. <br>
      <br>
      Currently this is set to match <code>"yyyy/MM/dd HH:mm:ss"</code>.
   */
  public static final SimpleDateFormat CVS_DATE_FORMAT
    = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");


  /** This is the date format used by the users. <br>
      <br>
      Currently this is set to match <code>"yyyy/MM/dd"</code>.
   */
  public static final SimpleDateFormat USER_DATE_FORMAT
    = new SimpleDateFormat("yyyy/MM/dd");

  /** This is the collection of all known files.
      <pre>
        key   - File name (equal to value.getName()).
        value - The file.
      </pre>
   */
  protected Hashtable<String,Object> files;

  /** This is the collection of all known DR changes.
      <pre>
        key   - The DR number. (equal to value.getDRNum()).
        value - The DR.
      </pre>
   */
  protected Hashtable<String,DR> drs;

  /** This is the DR representing all changes without a DR number. */
  protected DR UNKNOWN_DR;

  /** This is the DR representing all changes with an invalid DR number. */
  protected DR INVALID_DR;


  /** Creates a new instance of DRManager. */
  public DRManager() {
    this.files = new Hashtable<String,Object>();
    this.drs   = new Hashtable<String,DR>();

    for (String drCode : DR_CODES) {
      this.drs.put(drCode.toUpperCase(), new DR(drCode));
    }

    this.UNKNOWN_DR = new DR("Not Specified");
    this.INVALID_DR = new DR("Invalid DR Number");
    this.drs.put(this.UNKNOWN_DR.getDRNum().toUpperCase(), this.UNKNOWN_DR);
    this.drs.put(this.INVALID_DR.getDRNum().toUpperCase(), this.INVALID_DR);
  }

  /** Creates a new instance of DRManager and parses the specified log file.

      @param logFile The name of the log file to parse.

      @throws IOException If there is an error in reading the file.

      @see #parseLogFile(String)
   */
  public DRManager(String logFile) throws java.io.IOException {
    this();
    this.parseLogFile(logFile);
  }

  /** Creates a new instance of DRManager and parses the specified log file.

      @param logFile The log file to parse.

      @throws IOException If there is an error in reading the file.

      @see #parseLogFile(BufferedReader)
   */
  public DRManager(BufferedReader logFile) throws java.io.IOException {
    this();
    this.parseLogFile(logFile);
  }

  /** This will read in and parse a CVS log file.

      @param logFile The name of the log file to parse.

      @throws IOException If there is an error in reading the file.

      @see #parseLogFile(BufferedReader)
   */
  public void parseLogFile(String logFile) throws java.io.IOException {
    this.parseLogFile(new BufferedReader(new FileReader(logFile)));
  }

  /** This will read in and parse a CVS log file.

      @param logFile The log file to parse.

      @throws IOException If there is an error in reading the file.
   */
  public void parseLogFile(BufferedReader logFile)  throws java.io.IOException {
    final String DR_SET = "(" + DR_REGEX + ")"
                        + "(" + DR_SEP_REGEX + DR_REGEX + ")*";

    String line = logFile.readLine();

    File   file          = null;
    String drNum         = null;
    String changeAuthor  = null;
    String changeDate    = null;
    String changeComment = null;

    while (line != null) {
      if (line.startsWith("Working file:")) {
        file = new File(line.substring("Working file: ".length()));

        drNum         = null;
        changeAuthor  = null;
        changeDate    = null;
        changeComment = null;

      } else if (line.startsWith("date:")) {
        String[] parts = line.split(";");


        changeDate    = parts[0].substring(parts[0].indexOf(':') + 1);
        changeAuthor  = parts[1].substring(parts[1].indexOf(':') + 1);
        changeComment = ""; /* Indicator that the change comment is next. */

      } else if (changeComment != null) {

        if (line.matches("[Dd][Rr][:].*")) {
          drNum = line.substring(3).trim();

        } else if (line.matches(DR_SET)) {
          drNum = line;

        } else if (line.matches("((-|=)*)")) {
          /* Fromat and HTML-ize the comment. */
          changeComment = changeComment.trim();
          changeComment = changeComment.replaceAll("&", "&amp;");
          changeComment = changeComment.replaceAll("<", "&lt;");
          changeComment = changeComment.replaceAll(">", "&gt;");

          if(changeComment.matches("[Cc][Oo][Mm][Mm][Ee][Nn][Tt][:].*")) {
            changeComment = changeComment.substring("Comment:".length()).trim();
          }

          /* Format the Date */
          Date date = null;

          try {
            date = CVS_DATE_FORMAT.parse(changeDate.trim());
          } catch(java.text.ParseException exception) {
            date = null;
          }

          /** Format the DR number(s). */
          String[] drNums = (drNum == null)? new String[] {""}
                                           : drNum.trim().split("[ ,;:]");

          this.addDRChange(file, date, changeAuthor, changeComment, drNums[0]);

          for (int i = 1; i < drNums.length; i++) {
            drNums[i] = drNums[i].trim();

            if (!drNums[i].equals("")) {
              this.addDRChange(file,date,changeAuthor,changeComment,drNums[i]);
            }
          }

          drNum         = null;
          changeAuthor  = null;
          changeDate    = null;
          changeComment = null;

        } else {
          changeComment += "\n" + line;
        }
      }

      line = logFile.readLine();
    }
  }

  /** This will add a DR change. <br>
      <br>
      Prior to addition, the comments will be filtered and any special HTML
      characters (<code>&amp;</code>, <code>&lt;</code>, and <code>&gt;</code>)
      will be converted their HTML encoding.

      @param file    The file the change is associated with.
      @param date    The date of the change.
      @param author  The ID of the person making the change.
      @param comment The comment describing the change.
      @param drNum   The DR number.

      @see #getDR(String)
   */
  protected void addDRChange (File file, Date date, String author,
                              String comment, String drNum) {
    new DRChange(file, date, author, comment, this.getDR(drNum));
  }

  /** Gets a DR based on its DR number. If the DR does not exist within
      {@link #drs} then a new one will be created and added to {@link #drs}.
      If <code>number</code> is <code>null</code> or equal to <code>""</code>
      then it will be replaced with <code>"Not Specified"</code>. If
      <code>number</code> does not match {@link #DR_REGEX} then it will be
      set to <code>"Invalid DR Number"</code>.

      @param number The DR number.
      @return The associated DR.
   */
  protected DR getDR (String number) {
    number = (number == null) ? "" : number.trim().toUpperCase();

    if(number.equals("")) {
      return this.UNKNOWN_DR;

    } else if(!number.matches(DR_REGEX)) {
      return this.INVALID_DR;

    } else {
      DR dr = this.drs.get(number);

      if(dr == null) {
        dr = new DR(number);
        this.drs.put(dr.getDRNum(), dr);
      }

      return dr;
    }
  }

  /** Searches through all of the DRs changes that match the specified DR.

      @param drNum The DR number.

      @return The results as a HTML-formatted page.
   */
  public String getDRByNumber (String drNum) {
    DR dr = this.drs.get(drNum.trim().toUpperCase());
    if(dr == null) {
      return "<html><body><font size=\"+1\"><i>No changes listed for DR Number "
           + "\"" + drNum.trim() + "\".</i></font></body></html>";

    } else {
      StringBuilder results = new StringBuilder();
      DRChange[]   changes  = dr.getChanges();
      ArrayList<DRChange> allChanges = new ArrayList<DRChange>(changes.length);

      for (DRChange change : changes) {
        allChanges.add(change);
      }

      Collections.sort(allChanges);
      Collections.reverse(allChanges);

      results.append("<html><body>\n");
      results.append("<center><b><u><font size=\"+2\">DR Number \"");
      results.append(dr.getDRNum());
      results.append("\"</font></u></b></center>\n");
      results.append("\n");
      results.append("<center><table border=\"1\" width=\"95%\">\n");
      results.append("  <tr><th width=\"200\">File Modified</th>\n");
      results.append("      <th>Modified By</th>\n");
      results.append("      <th width=\"100\">Date</th>\n");
      results.append("      <th>Comment</th></tr>\n");

      int count = 0;
      for (DRChange change : allChanges) {
        if (count++ > 1000) {
          break;
        }

        results.append("  <tr><td>");
        results.append(change.getFile().getName().trim());
        results.append("</td>\n");
        results.append("      <td>");
        results.append(change.getAuthor().trim());
        results.append("</td>\n");
        results.append("      <td>");
        results.append(USER_DATE_FORMAT.format(change.getDate()));
        results.append("</td>\n");
        results.append("      <td>");
        results.append(change.getComment().trim());
        results.append("</tr>\n");
      }
      results.append("</table></center>\n");

      if (count >= 1000) {
        results.append("  <br><center><i><b>");
        results.append((changes.length - count));
        results.append(" more changes are not listed here.</b></i></center>");
      }

      results.append("</body></html>\n");

      return results.toString();
    }
  }


  /** Searches through all of the DRs to find any whose changes fall within the
      specified date range. If either date is <code>null</code> or equal to
      <code>""</code>, then it will be replaced by <code>"1900/01/01"</code>
      (for <code>start</code>) or <code>"2900/01/01"</code> (for
      <code>end</code>).

      @param start The start date.
                   Must be in the form <code>"YYYY/MM/DD"</code>.
      @param end   The end date.
                   Must be in the form <code>"YYYY/MM/DD"</code>.

      @return The results as a HTML-formatted page.

      @see #getDRByDate(Date,Date)
   */
  public String getDRByDate (String start, String end) {
    start = start.trim();
    end   = end.trim();

    if (start.equals("")) {
      start = "1900/01/01";
    }

    if (end.equals("")) {
      end = "2900/01/01";
    }

    try {
      return this.getDRByDate(USER_DATE_FORMAT.parse(start),
                              USER_DATE_FORMAT.parse(end));

    } catch(java.text.ParseException exception) {
      return "Can not process search, invalid date entered";
    }
  }

  /** Searches through all of the DRs to find any whose changes fall within the
      specified date range.

      @param start The start date.
      @param end   The end date.

      @return The results as a HTML-formatted page.
   */
  public String getDRByDate (Date start, Date end) {
    StringBuilder results = new StringBuilder();
    ArrayList<DR> allDRs  = new ArrayList<DR>();

    allDRs.addAll(this.drs.values());
    Collections.sort(allDRs);

    results.append("<html><body>\n");
    results.append("<center><b><u><font size=\"+2\">All DR's From ");
    results.append(USER_DATE_FORMAT.format(start));
    results.append(" to ");
    results.append(USER_DATE_FORMAT.format(end));
    results.append("</font></u></b></center>\n");
    results.append("\n");
    results.append("<center><table border=\"1\" width=\"95%\">\n");
    results.append("  <tr><th>DR Number</th>\n");
    results.append("      <th>Modification Dates</th>\n");
    results.append("      <th>Number of Changes</th></tr>\n");

    for (DR dr : allDRs) {
      DRChange[] changes = dr.getChanges();

      if(changes.length > 0) {
        Date min = changes[0].getDate();
        Date max = changes[0].getDate();

        for (int i = 1; i < changes.length; i++) {
          if (changes[i].getDate().before(min)) {
            min = changes[i].getDate();
          }

          if (changes[i].getDate().after(max)) {
            max = changes[i].getDate();
          }
        }

        if((min.after(start) && min.before(end))
           || (max.after(start) && max.before(end))
           || (min.before(start) && max.after(end))) {

          results.append("  <tr><td>");
          results.append(dr.getDRNum());
          results.append("</td>\n");
          results.append("      <td>");
          results.append(USER_DATE_FORMAT.format(min));
          results.append(" to ");
          results.append(USER_DATE_FORMAT.format(max));
          results.append("</td>\n");
          results.append("      <td>");
          results.append(changes.length);
          results.append("</td></tr>\n");
        }
      }
    } // end for loop

    results.append("</table></center>\n");
    results.append("</body></html>\n");

    return results.toString();
  }

  /** Creates the graphical interface to the DR manager.


      @param headless Indicates if the GUI will run in headless mode. If this
                      is <code>true</code> then the program will exit (via
                      {@link System#exit}) when the window is closed.

      @return The graphical interface to the DR manager.
   */
  public JFrame createGUI (boolean headless) {
    return new GUI(headless);
  }

  /** This is the graphical user interface to the DRManager. */
  protected class GUI extends JFrame implements ActionListener {
    /** NOTE: this should be changed if the structure/fields of this class changes. */
    private static final long serialVersionUID = 2010080200290L; // for Serializable interface

    /** This is the help page displayed when the help button is pressed. */
    public final String HELP_PAGE
      = "<html>\n"
      + "  <body>\n"
      + "    <center><font size=\"+1\"><b>DRManager Help</b></font></center>\n"
      + "    <p>\n"
      + "      The DRManager is designed to provide an easy way to search\n"
      + "      through a set of DRs. The buttons above provide easy access\n"
      + "      to the DRs."
      + "      <ul>\n"
      + "        <li><b>Search By DR Number</b>: A prompt will come up asking\n"
      + "            for the number of the DR to search for, upon entering a\n"
      + "            DR number the search is made over all the changes to all\n"
      + "            the files in CVS and a listing is provided summarizing\n"
      + "            the changes that match the provided DR number.</li>\n"
      + "        <li><b>Search By Date</b>: A prompt will come up asking\n"
      + "            for the start date and the end dates to search for.\n"
      + "            A search is then made that will list all of the DRs for\n"
      + "            which a file was changed between the specified dates.\n"
      + "            <i>(If the start date is left blank, \n"
      + "            <code>\"1900/01/01\"</code> will be used. If the end\n"
      + "            date is left blank, <code>\"2900/01/01\"</code> will be\n"
      + "            used. Basically, leaving both dates blank will list all\n"
      + "            DR changes made.)</i></li>\n"
      + "        <li><b>Help</b>: Displays this message.</li>"
      + "        <li><b>Exit</b>: Exits the DRManager GUI.</li>"
      + "      </ul>"
      + "    </p>\n"
      + "    <p>\n"
      + "      Special DR Codes Permitted:\n"
      + "      <ul>"
      + "        " + listDRCodes()
      + "      </ul>"
      + "    </p>\n"
      + "  </body>\n"
      + "</html>\n"
      ;

    /** The text pane which displays any results. */
    protected JEditorPane textPane;

    /** The "search by DR number" button. */
    protected JButton byNumberButton;

    /** The "search by date" button. */
    protected JButton byDateButton;

    /** The exit button. */
    protected JButton exitButton;

    /** The help button. */
    protected JButton helpButton;

    /** Indicates if the GUI is running in headless mode. */
    protected boolean headless;

    /** Creates a new instance. <br>
        <br>
        This is the same as <code>GUI(false)</code>.
     */
    public GUI() {
      this(false);
    }

    /** Creates a new instance.

        @param headless Indicates if this is running in headless mode. If this
                        is <code>true</code> then the program will exit (via
                        {@link System#exit}) when the window is closed.
     */
    public GUI (boolean headless) {
      this.headless = headless;

      textPane = new JEditorPane("text/html", this.HELP_PAGE);
      textPane.setPreferredSize(new java.awt.Dimension(800,600));
      textPane.setMinimumSize(textPane.getPreferredSize());
      textPane.setMaximumSize(textPane.getPreferredSize());
      textPane.setSize(textPane.getPreferredSize());

      JPanel buttonPanel = new JPanel();

      byNumberButton = new JButton("Search By DR Number");
      byDateButton   = new JButton("Search By Date");
      exitButton     = new JButton("Exit");
      helpButton     = new JButton("Help");

      byNumberButton.setSize(50, 30);
      byDateButton.setSize(50, 30);
      exitButton.setSize(50, 30);

      buttonPanel.add(byNumberButton);
      buttonPanel.add(byDateButton);
      buttonPanel.add(helpButton);
      buttonPanel.add(exitButton);

      byNumberButton.addActionListener(this);
      byDateButton.addActionListener(this);
      exitButton.addActionListener(this);
      helpButton.addActionListener(this);

      add(buttonPanel, BorderLayout.PAGE_START);
      add(new JScrollPane(textPane, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                                    JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS),
          BorderLayout.CENTER);

      this.addWindowListener(new WindowAdapter() {
        public void windowClosing(WindowEvent evt) {
          setVisible(false);

          if (isHeadless()) {
            System.exit(0);
          }
        }
      });

      this.setSize((int)textPane.getPreferredSize().getWidth()  + 50,
                   (int)textPane.getPreferredSize().getHeight() + 125);
      this.setResizable(false);
    }

    /** Indicates if the window is headless.

        @return If the GUI is headless.
     */
    public boolean isHeadless() {
      return this.headless;
    }

    public void actionPerformed( ActionEvent event) {
      if(event.getSource() == this.byNumberButton) {
        String drNum = JOptionPane.showInputDialog("Enter the DR Number to "
                                                 + "search for.");
        if (drNum == null) {
          return;
        }

        this.textPane.setText("SEARCHING...");
        this.textPane.setText(getDRByNumber(drNum));

      } else if (event.getSource() == this.byDateButton) {
        String startDate
          = JOptionPane.showInputDialog("Enter the STARTING date to search for "
                                      + "(This must be in the form of "
                                      + "'YYYY/MM/DD'.)");
        if (startDate == null) {
          return;
        }

        String endDate
          = JOptionPane.showInputDialog("Enter the ENDING date to search for "
                                      + "(This must be in the form of "
                                      + "'YYYY/MM/DD'.)");
        if (endDate == null) {
          return;
        }

        this.textPane.setText("SEARCHING...");
        this.textPane.setText(getDRByDate(startDate, endDate));

      } else if (event.getSource() == this.exitButton) {
        this.setVisible(false);

        if (this.isHeadless()) {
          System.exit(0);
        }
      } else if(event.getSource() == this.helpButton) {
        this.textPane.setText(this.HELP_PAGE);
      }
    }


  }

  /** Runs the GUI for the DRManager. <br>
      <br>
      This method does little more than call
      <code>(new DRManager(args[0])).createGUI(true).setVisible(true)</code>.

      @param args Any command line arguments. This must have exactly one value
                  which represents the name of the CVS log file to use.
   */
  public static void main(String[] args) {
    try {

      if (args.length != 1) {
        System.err.println("CVS Log file must be specified.");

      } else {
        JFrame frame = (new DRManager(args[0])).createGUI(true);
        frame.setResizable(true);
        frame.setVisible(true);
      }

    } catch(IOException exception) {
      System.err.println("Error in reading from CVS log file (" + args[0] + "): "
                       + exception);
    }
  }

  /** Transforms an array of strings into a case-insensitive regular expression.
      @param words The array of strings;
      @return The regular expression.
   */
  private static final String makeRegex(String[] words) {
    StringBuilder regex = new StringBuilder();

    regex.append('(');

    for (int i = 0; i < words.length; i++) {
      if(i != 0) {
        regex.append('|');
      }
      regex.append('(');

      for(int j = 0; j < words[i].length(); j++) {
        regex.append(("([" + words[i].charAt(j)).toUpperCase());

        regex.append(("]|[" + words[i].charAt(j)).toLowerCase());

        regex.append("])");
      }

      regex.append(')');
    }

    regex.append(')');

    return regex.toString();
  }


  /** Simply lists all of the special DR codes.
      @return The special DR codes.
   */
  private String listDRCodes() {
    StringBuilder codes = new StringBuilder();

    for (String drCode : DR_CODES) {
      codes.append("<li><code>");
      codes.append(drCode);
      codes.append("</code></li>");
    }

    return codes.toString();
  }
}
